//-----------------------------------------------------------------------------
//
// MN_DEF.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __MN_DEF_H__
#define __MN_DEF_H__

#include "p_data.h"
#include "dd_input.h"

// Macros

#define LEFT_DIR		 0
#define RIGHT_DIR		 1

#define LINEHEIGHT		16
#define LINEHEIGHT_B	13 // R3D - was 15
#define LINEHEIGHT_A	8

// Types

typedef enum
{
	ITT_EMPTY,
	ITT_EFUNC,
	ITT_CPFUNC,
	ITT_LRFUNC,
	ITT_SETMENU,
	ITT_INERT
} ItemType_e;

typedef enum
{
	MENU_MAIN,
	MENU_EPISODE,
	MENU_SKILL,
	MENU_OPTIONS,
	MENU_GAMEPLAY,
	MENU_HUD,
	MENU_CONTROLS,
	MENU_INPUT,
	MENU_LOAD,
	MENU_SAVE,
	MENU_MULTIPLAYER,
	MENU_GAMESETUP,
	MENU_PLAYERSETUP,
	MENU_NONE,

	MENU_DISPLAY = 1,
	MENU_VIDEO,
	MENU_SOUND
} MenuType_e;

typedef struct
{
	ItemType_e	type;
	char		*mntext;
	void		(*func)(int option);
	int			option;
	char		*mnlumpname;
	int			mnlumpnum;
	int			id;
} MenuItem_t;

typedef struct
{
	int			x;
	int			y;
	void		(*drawFunc)(void);
	int			itemCount;
	MenuItem_t	*items;
	int			lastOn;
	MenuType_e	prevMenu;
	dpatch_t	*font;				// Font for menu items.
	int			itemHeight;
	// For multipage menus.
	int			firstItem;
	int			numVisItems;
} Menu_t;


extern int MenuTime;
extern RDBOOL shiftdown;
extern Menu_t *currentMenu;

void Set_Menu(MenuType_e menu);
void M_DrawTitle(char *text, int y);
void M_WriteText(int x, int y, char *string);
void M_WriteText2(int x, int y, char *string, dpatch_t *font, 
				  float	red, float green, float	blue);
void M_WriteMenuText(Menu_t *menu, int index, char *text);

extern Menu_t ControlsDef;

#endif // __MENU_DEFS_H_