//-----------------------------------------------------------------------------
//
// CON_CONSOLE.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __CON_CONSOLE_H__
#define __CON_CONSOLE_H__

#include <stdio.h>
#include "con_functions.h"
#include "dd_input.h"

#define MAX_ARGS	256

typedef struct
{
	char	cmdLine[2048];
	int		argc;
	char	*argv[MAX_ARGS];
} cmdargs_t;

// A console buffer line.
typedef struct
{
	int		len;	// This is the length of the line (no term).
	char	*text;	// This is the text.
	int		flags;
} cbline_t;

void	Con_Init();
void	Con_Shutdown();
void	Con_MaxLineLength(void);
void	Con_Open(int yes);
void	Con_UpdateKnownWords(void);
void	Con_Ticker(void);
RDBOOL	Con_Responder(event_t *event);
void	Con_Drawer(void);
void	Con_DrawRuler(int y, int lineHeight, float alpha);
void	Con_Printf(char *format, ...);
void	Con_Printfo(char *format, ...);
void	Con_FPrintf(int flags, char *format, ...);
void	Con_playFPrintf(RDBOOL always, int flags, char *format, ...);
void	Con_SetFont(ddfont_t *cfont);
cbline_t *Con_GetBufferLine(int num);
int		Con_Execute(char *command, int silent);
int		Con_Executef(int silent, char *command, ...);

void	Con_Message(char *message, ...);
void	Con_Error(char *error, ...);
void	Con_ErrorD(char *error);
void	Con_ErrorDev (char *error, ...);
cvar_t* Con_GetVariable(char *name);
int		Con_GetInteger(char *name);
float	Con_GetFloat(char *name);
byte	Con_GetByte(char *name);
char*	Con_GetString(char *name);

void	Con_SetInteger(char *name, int value);
void	Con_SetFloat(char *name, float value);
void	Con_SetString(char *name, char *text);
void	Con_SaveDefaults(void);
int		Con_ParseCommands(char *fileName, RDBOOL silent);

char	*plural(int val);

extern int	set_light_defaults;
extern int	bufferLines;
extern int	consoleDump;
extern char	configFileName[256];

// checked in dd_loop.c
extern float	ConsoleY;

#endif