//-----------------------------------------------------------------------------
//
// DEDFILE.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __DEDFILE_H__
#define __DEDFILE_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "dd_types.h"

#include "dd_dfdat.h"

#define MAX_FRAME_MODELS	8

#define DED_VERSION			5

#define DED_SPRITEID_LEN	4
#define DED_STRID_LEN		7
#define DED_STRINGID_LEN	47
#define DED_PATH_LEN		MAXPATH - 1
#define DED_FLAGS_LEN		160

#define DED_PTC_STAGES		16

// Note: string lengths have +1 for NULL terminator
typedef char			ded_strid_t[DED_STRID_LEN + 1];
typedef ded_strid_t		ded_mapid_t;
typedef char			ded_stringid_t[DED_STRINGID_LEN+1];
typedef ded_stringid_t	ded_string_t;
typedef ded_stringid_t	ded_mobjid_t;
typedef ded_stringid_t	ded_stateid_t;
typedef ded_stringid_t	ded_soundid_t;
typedef ded_stringid_t	ded_musicid_t;
typedef ded_stringid_t	ded_funcid_t;
typedef char			ded_flags_t[DED_FLAGS_LEN+1];

typedef struct ded_count_s { int num, max; } ded_count_t;

typedef struct
{
	char path[DED_PATH_LEN + 1];
} ded_path_t;

typedef struct
{
	char id[DED_SPRITEID_LEN + 1];
} ded_sprid_t;

typedef struct
{
	char str[DED_STRINGID_LEN + 1];
} ded_str_t;

typedef struct
{
	ded_stringid_t	id;
	int				value;
} ded_flag_t;

typedef struct
{
	ded_mobjid_t	id;	
	ded_string_t	name;
    ded_flags_t     flags[NUM_MOBJ_FLAGS];
	
	ded_stateid_t	spawnstate;
	ded_stateid_t	seestate;
	ded_stateid_t	painstate;
	ded_stateid_t	meleestate;
	ded_stateid_t	missilestate;
	ded_stateid_t	deathstate;
	ded_stateid_t	xdeathstate;
	ded_stateid_t	raisestate;

	ded_soundid_t	seesound;
	ded_soundid_t	attacksound;
	ded_soundid_t	painsound;
	ded_soundid_t	deathsound;
	ded_soundid_t	activesound;

	RDBOOL			cache;					// b211-0

	float			speed;
	float			radius;
	float			height;

	int				doomednum;
	int				reactiontime;
	int				painchance;
	int				spawnhealth;
	int				mass;
	int				damage;
	int				weapon_offset;			// v220-4
	short			scplightlevel;			// v220-13

} ded_mobj_t;

typedef struct
{	
	ded_stateid_t	id;				// ID of this state.
	ded_sprid_t		sprite;
	ded_flags_t		flags;
	int				frame;
	int				tics;
	ded_funcid_t	action;
	ded_stateid_t	nextstate;
	int				misct[NUM_MISCST];
} ded_state_t;

typedef struct
{
	ded_stateid_t	state;
	float			xoffset;		// Origin offset in world coords
	float			yoffset;		// Zero means automatic
	float			size;			// Zero: automatic
	float			color[3];		// Red Green Blue (0,1)
	ded_flags_t		flags_string;	
	int				flags;			// Runtime
} ded_light_t;

typedef struct
{
	ded_path_t		filename;
	ded_string_t	frame;
	int				framerange;
	ded_flags_t		flags;			// ASCII string of the flags.
	int				skin;
	int				skinrange;
	float			offset[3];		// Offset XYZ within model.
	float			alpha;
	float			parm;			// Custom parameter.
	unsigned char	selskinbits[2];	// Mask and offset.
	unsigned char	selskins[8];	
	ded_string_t	shinyskin;
	float			shiny;
	float			shinycolor[3];
} ded_submodel_t;

typedef struct
{
	ded_stateid_t	state;
	int				off;
	ded_sprid_t		sprite;			// Only used by autoscale.
	int				spriteframe;	// Only used by autoscale.
	ded_flags_t		group;
	int				selector;
	ded_flags_t		flags;
	float			intermark;
	float			interrange[2];	// 0-1 by default.
	int				skintics;		// Tics per skin in range.
	float			scale[3];		// Scale XYZ
	float			resize;
	float			offset[3];		// Offset XYZ
	float			shadowradius;	// Radius for shadow (0=auto).
	ded_submodel_t	sub[MAX_FRAME_MODELS];
	char			model_path[DED_PATH_LEN];
} ded_model_t;

// Although much of this is redundant it has been left in place
// for compatibility with existing deds - R3D
typedef struct
{
	ded_soundid_t	id;			// ID of this sound, refered to by others.
	ded_string_t	lumpname;	// Actual lump name of the sound (DS not included).
	ded_string_t	name;		// A tag name for the sound.
	ded_flags_t		flags;		// Flags (SF_*).
	int				group;		// Exclusion group (only used with chainsaw).

	// The following, apart from dedmaxchan, are ignored
	ded_soundid_t	link;		// Link to another sound.
	int				link_pitch;
	int				link_volume;
	int				priority;	// Priority classification.
	int				dedmaxchan;	// Max number of channels to occupy.
	ded_path_t		ext;		// External sound file (WAV).
} ded_sound_t;

typedef struct
{
	ded_musicid_t	id;			// ID of this piece of music.
	ded_string_t	lumpname;	// Lump name.
	ded_path_t		path;		// External file (not a normal MUS file).
} ded_music_t;

typedef struct
{
	ded_flags_t		flags;
	ded_string_t	texture;
	float			offset;
	float			color_limit;
	float			skyAngleIncrement;
	int				grpnum;
} ded_skylayer_t;

#define NUM_SKY_LAYERS		2

typedef struct
{
	ded_mapid_t		id;				// ID of the map (e.g. E2M3 or MAP21).
	ded_string_t	mapname;		// Name of the map.
	ded_string_t	author;			// Author of the map.
	ded_string_t	epitext;		// Text file for a pwad doom episode
	ded_flags_t		flags;			// Flags.
	ded_path_t		music;			// Music to play.
	int				partime;		// Partime, in seconds.	// v220-4 all partime refs changed from float to int
	byte			fog_color[4];	// Fog color (RGBA)
	float			fog_start;
	float			fog_end;
	float			fog_density;
	float			ambient;	// Ambient light level.
	float			gravity;	// 1 = normal.
	float			sky_height;
	float			horizon_offset;
	ded_skylayer_t	sky_layers[NUM_SKY_LAYERS];

	// b211-0 R3D vars (not read in from defs)
	char			bexbgflat[10];	// +1 for \0 +1 to align
	char			nextmap[8];		// ID of the next map.
	char			secretnext[8];	// ID of the next secret map.
	int				cluster;
	int				R3Dflags;
	int				titlepatch;
	int				nexttitlepatch;
	int				secrettitlepatch;
	int				mapinfomuslump;	// MAPINFO musiclump
	int				muslump;		// WAD music lump
} ded_mapinfo_t;

typedef struct
{
	ded_stringid_t	id;
	char			*text;
} ded_text_t;

typedef struct
{
	ded_stringid_t	id;
	ded_count_t		count;
	ded_str_t		*textures;
} ded_tenviron_t;

typedef struct
{
	char			*id;	
	char			*text;
} ded_value_t;

typedef struct
{
	ded_stringid_t	debrief;
	ded_stringid_t	brief;
	int				game;
	char			*script;
} ded_finale_t;

typedef struct
{
	ded_string_t	wall;			// Name of a wall texture.
	ded_string_t	flat;			// Name of a flat.
	ded_string_t	det_lumpname;	// The lump with the detail texture.
	float			scale;
	float			strength;
	float			maxdist;
} ded_detailtexture_t;

typedef struct
{
	ded_string_t	type;
	int				tics;
	float			variance;		// Stage variance (time).
	float			radius_variance;		// Stage variance (time).
	float			color[4];		// RGBA
	float			radius;
	ded_flags_t		flags;
	float			bounce;
	float			resistance;		// Air resistance.
	float			gravity;
} ded_ptcstage_t;

typedef struct
{
	ded_stateid_t	state;			// Triggered by this state (if mobj-gen).
	ded_string_t	flat;			// Triggered by this flat.
	int				flat_num;
	ded_mobjid_t	type;			// Triggered by this type of mobjs.
	ded_mobjid_t	type2;			// Also triggered by this type.
	int				type_num;
	int				type2_num;
	ded_mobjid_t	damage;			// Triggered by mobj damage of this type.
	int				damage_num;
	ded_string_t	map;			// Triggered by this map (or empty string).
	ded_flags_t		flags_string;
	int				flags;
	float			speed;			// Particle spawn velocity.
	float			spd_variance;	// Spawn speed variance (0-1).
	float			vector[3];		// Particle launch vector.
	float			vec_variance;	// Launch vector variance (0-1). 1=totally random.
	float			center[3];		// Offset to the mobj (relat. to source).
	float			min_spawn_radius; // Spawn uncertainty box.
	float			spawn_radius;	
	float			maxdist;		// Max visibility for particles.
	int				spawn_age;		// How long until spawning stops?
	int				max_age;		// How long until generator dies?
	int				particles;		// Maximum number of particles.
	float			spawn_rate;		// Particles spawned per tic.
	float			spawn_variance;
	int				presim;			// Tics to pre-simulate when spawned.
	int				alt_start;
	float			alt_variance;	// Probability for alt start.
	float			force;			// Radial strength of the sphere force.
	float			force_radius;	// Radius of the sphere force.
	float			force_axis[3];	// Rotation axis of the sphere force (+ speed).
	float			force_origin[3]; // Offset for the force sphere.
	ded_ptcstage_t	stages[DED_PTC_STAGES];
} ded_ptcgen_t;

// The ded_t structure encapsulates all the data one definition file 
// can contain. This is only an internal representation of the data.
// An ASCII version is written and read by DED_Write() and DED_Read() 
// routines.

// It is VERY important not to sort the data arrays in any way: the
// index numbers are important. The Game DLL must be recompiled with the
// new constants if the order of the array items changes.

typedef struct ded_s
{
	int			version;			// DED version number.
	char		mobj_prefix[10];	// Prefix for mobj type constants.
	char		state_prefix[10];	// Prefix for state constants.
	char		sprite_prefix[10];	// Prefix for sprite constants.
	char		model_path[260];	// Directories for searching MD2s.
	ded_flags_t	model_flags;		// Default values for models.
	float		model_scale;		
	float		model_offset;		
	char		sfx_prefix[10];		// Prefix for sfx constants.
	char		mus_prefix[10];		// Prefix for mus constants.
	char		text_prefix[10];	// Prefix for text constants.

	struct ded_counts_s {
		ded_count_t includes;
		ded_count_t	flags;
		ded_count_t	mobjs;
		ded_count_t	states;
		ded_count_t	cstates;
		ded_count_t	sprites;
		ded_count_t	lights;
		ded_count_t	models;
		ded_count_t	sounds;
		ded_count_t	music;
		ded_count_t	mapinfo;
		ded_count_t	text;
		ded_count_t	tenviron;
		ded_count_t	values;
		ded_count_t	details;
		ded_count_t	ptcgens;
		ded_count_t	finales;
	} count;

	// Include other DED files.
	ded_path_t	*includes;

	// Flag values (for all types of data).
	ded_flag_t	*flags;

	// Map object information.
	ded_mobj_t	*mobjs;

	// States.
	ded_state_t	*states;

	// States copy
	ded_state_t	*cstates;

	// Sprites.
	ded_sprid_t	*sprites;

	// Lights.
	ded_light_t *lights;

	// Models.
	ded_model_t	*models;

	// Sounds.
	ded_sound_t	*sounds;

	// Music.
	ded_music_t	*music;

	// Map information.
	ded_mapinfo_t *mapinfo;

	// Text.
	ded_text_t	*text;

	// Aural environments for textures.
	ded_tenviron_t *tenviron;

	// Free-from string values.
	ded_value_t	*values;

	// Detail texture assignments.
	ded_detailtexture_t *dettex;

	// Particle generators.
	ded_ptcgen_t *ptcgens;

	// Finales.
	ded_finale_t *finales;
	
} ded_t;

// Routines for managing DED files.
void DED_Init(ded_t *ded);
void DED_Destroy(ded_t *ded);
int DED_Read(ded_t *ded, char *sPathName, int bDestroyOld);
int DED_ReadLump(ded_t *ded, int lump, int bDestroyOld);

int DED_AddInclude(ded_t *ded, char *inc);
int DED_AddFlag(ded_t *ded, char *name, int value);
int DED_AddMobj(ded_t *ded, char *idstr);
int DED_AddState(ded_t *ded, char *id);
int DED_AddSprite(ded_t *ded, const char *name);
int DED_AddLight(ded_t *ded, const char *stateid);
int DED_AddModel(ded_t *ded, char *spr);
int DED_AddSound(ded_t *ded, char *id);
int DED_AddMusic(ded_t *ded, char *id);
int DED_AddMapInfo(ded_t *ded, char *str);
int DED_AddText(ded_t *ded, char *id);
int DED_AddTexEnviron(ded_t *ded, char *id);
int DED_AddValue(ded_t *ded, const char *id);
int DED_AddDetail(ded_t *ded);
int DED_AddPtcGen(ded_t *ded, const char *state);
int DED_AddFinale(ded_t *ded);

void *DED_NewEntries(void **ptr, ded_count_t *cnt, int elem_size, int count);
void *DED_NewEntry(void **ptr, ded_count_t *cnt, int elem_size);
void DED_DelEntry(int index, void **ptr, ded_count_t *cnt, int elem_size);
void DED_DelArray(void **ptr, ded_count_t *cnt);
void DED_ZCount(ded_count_t *c);

extern char	dedReadError[];

#ifdef __cplusplus
}
#endif

#endif