//-----------------------------------------------------------------------------
//
// GL_TEX.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __GL_TEX_H__
#define __GL_TEX_H__

#include "r_data.h"
#include "rend_list.h"
#include "rend_sky.h"

extern texscale_t	*ctex;

extern detailinfo_t *rendtexdetail;
extern int		rendtexflags;	
extern RDBOOL	rendtexmask;
	
extern uint		curtex;
extern byte		*palbuffer;
extern byte		*iwadpalbuffer;

extern DGLuint	dltexname, glowtexname;

void		PalIdxToRGB(byte *pal, int idx, byte *rgb);
void		GL_InitTextureManager(void);
void		GL_ShutdownTextureManager(void);
void		GL_SetHudsprites(void);
void		GL_LoadSystemTextures(void);
void		GL_DownMipmap32(byte *in, int width, int height, int comps);
void		GL_SetFlat(int idx);
void		GL_SetSkyTopColor(skylayer_t *slayer);
void		GL_SetSprite(int splnum, int *hsindex, int callnum);
RDBOOL		GL_GetHSsize(int index, int *w, int *h, int *f, float *ttc);
void		GL_SetTranslatedSprite(int splnum, byte *trans);
void		GL_GetFlatColor(int fnum, unsigned char *rgb);
void		GL_NewSplitTex(int lump, DGLuint part2name);
void		GL_SetNoTexture(void);
DGLuint		GL_PrepareTexture(int idx); 
DGLuint		GL_PrepareTextureNoAlpha(int idx);		// b211-0
void		GL_DeleteAllSecondaryAlphaTex(void);	// b211-0
RDBOOL		CheckFlatSet(int idx);	// v220-15
DGLuint		GL_PrepareFlat(int idx);
RDBOOL		GL_Prepare3DFlat(rendpoly_t *poly, int lumpnum, int snum);
DGLuint		GL_PrepareLightTexture(void);	// The dynamic light map.
RDBOOL		GL_SetPatch(int lump, RDBOOL nearest);	// v220-8 nearest added.
uint		GL_PrepareSky(int idx);
uint		GL_PrepareSprite(int splnum, int *hsindex, int callnum);
int			CeilPow2(int num);
int			GL_ValidTexHeight2(int width, int height);
int			GL_GetSkinTexIndex(const char *skin);
byte		*GL_LoadImage(char *imagefn, int *width, int *height,
				int *pixelsize, RDBOOL usemodelpath, RDBOOL ignorealpha);

void	GL_PrepareSkyBox(byte *buffer, int width, int height, RDBOOL alpha, DGLuint *address);	// v220-3

// Only for textures (not for flats, sprites, etc.)
void	GL_DeleteTexture(int texidx);
void	GL_DeleteTexIndex(DGLuint *texIndex, char *type);

// Only for per-level 3D scaled textures
void	GL_Delete3DTextures(void);

// Load the skin texture and prepare it for rendering.
uint	GL_PrepareSkin(struct model_s *mdl, int skin, ushort *skinalpha);
uint	GL_PrepareShinySkin(struct modeldef_s *md, int sub);

#endif
