//-----------------------------------------------------------------------------
//
// M_MISC.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __M_MISC__
#define __M_MISC__

#include "p_data.h"
#include "sys_file.h"

#define FORMAT_CHAR(c)	((c) == 0 || (c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r')

// limit extension changes R3D
#define MAPBS(a,b)		( (int)((((__int64)(a) - (__int64)(b)) >> MAPBLOCKSHIFT)) )
#define MAPBSF(a,b)		( (int)((((__int64)(a) - (__int64)(b)) >> MAPBTOFRAC)) )
#define MAPBS3(a,b,c)	( (int)((((__int64)(a) - (__int64)(b) + (__int64)(c)) >> MAPBLOCKSHIFT)) )

#define DOTPROD2(a, b)	(a[0]*b[0] + a[1]*b[1])
#define DOTPROD3(a, b)	(a[0]*b[0] + a[1]*b[2] + a[2]*b[1])

// Note c must not be 0 when calling this
#define MULDIV32(a,b,c) ( (DWORD)((((__int64)(a)*(__int64)(b))/(__int64)(c))) )

// macros to change from little endian to big endian format (i.e. Intel CPUs)
#define DWORDSWAP(dw) ( (((dw)>>24)&0x000000FFL)|(((dw)>> 8)&0x0000FF00L)|\
						(((dw)<< 8)&0x00FF0000L)|(((dw)<<24)&0xFF000000L))

#define WORDSWAP(w) ((((w)>>8)&0x00FF)|(((w)<<8)&0xFF00))

#define MAXMODELBOX	128		// Can be increased to the number of
							// 7800 + 7700 types possible
							// but 128 is a more sensible limit
typedef struct {
	int		sidenum;
	fixed_t	texoff;
	fixed_t	rowoff;
} analysesides_t;

extern analysesides_t *analysesides;
extern int maxannumsides;

// v220-33
#define MAXDAMAGENUM	16

extern float playerdamagefactor[MAXDAMAGENUM];
extern short playerdamagetype[MAXDAMAGENUM];
extern int playerdamageidx;
extern int enemyheadshottype[MAXDAMAGENUM][3];
extern int enemyheadshotidx;

// Debug routine
char *Def_IDfromType(int type);	// v220-33 for development purposes

// File system routines.
RDBOOL		M_WriteFile(char *name, void *source, int length);
void		M_ExtractFileBase(char *path, char *dest);
RDBOOL		M_CheckPath(char *path);
int			M_FileExists(char *file);
void		M_TranslatePath(char *path, char *translated);
void		M_PrependBasePath(char *path, char *newpath);
RDBOOL		M_PrependBaseToFile(char *path);
void		M_ReadLine(char *buffer, int len, DFILE *file);
void		M_CopyFile(char *infile, char *outfile);

// Bounding boxes.
void		M_ClearBox (fixed_t *box);
void		M_AddToBox(fixed_t *box, fixed_t x, fixed_t y);
void		M_AddLineToBox(fixed_t *box, line_t *ld);
RDBOOL		M_CompareBox(fixed_t *inner, fixed_t *outer);
RDBOOL		M_FitBox(fixed_t *inner, fixed_t *outer);
RDBOOL		M_SameBox(fixed_t *inner, fixed_t *outer);
RDBOOL		M_ComparexyBox(fixed_t x, fixed_t y, fixed_t *outer);
RDBOOL		M_CompareFPxyBox(float x, float y, float *outer);
RDBOOL		M_CompareFPBox(float *inner, float *outer);
RDBOOL		M_FitFPBox(float *inner, float *outer);
RDBOOL		M_FitSSECBox(fixed_t *box, fixed_t fx, fixed_t fy, fixed_t fr);
void		M_SetFPBoxFromIntBox(float *fpbox, fixed_t *intbox);
int			M_AreaBox(fixed_t *box);
float		M_AreaFPBox(float *box);
float		M_FPAreaBox(fixed_t *box);
void		M_setBoxOffsets(int *box, int x, int y, int radius);
void		M_getBlockOffsets(int *data, int adjust, int *xl, int *xh, int *yl, int *yh);
int			M_AddLIMIT(__int64 a, __int64 b);

// Text utilities.
char *		M_SkipWhite(char *str);
char *		M_FindWhite(char *str);
char *		M_SkipPlus(char *str);
char *		M_FindPlus(char *str);
void		M_WriteCommented(FILE *file, char *text);
void		M_WriteTextEsc(FILE *file, char *text);
RDBOOL		M_IsComment(char *text);
RDBOOL		strstrCI(char *str, char *charset);

// Random numbers.
int			M_Random(void);
int			M_RandomD(void);
float		M_FRandom(void);

// Math/vector routines.
float		M_AccurateDistance (fixed_t dx, fixed_t dy);
float		M_AccurateDistanceFP(float dx, float dy);
float		M_AccurateLength(line_t *ld);
float		M_ApproxDistancef(float dx, float dy);
float		M_PointLineDistance(float *a, float *b, float *c);
float		M_PointUnitLineDistance(float *a, float *b, float *c);
float		M_Normalize(float *a);
float		M_DotProduct(float *a, float *b);
void		M_LineUnitVector(line_t *line, float *unitvec);
void		M_CrossProduct(float *a, float *b, float *out);
void		M_PointCrossProduct(float *v1, float *v2, float *v3, float *out);
void		M_RotateVector(float vec[3], float degYaw, float degPitch);
void		M_ProjectPointOnLinef(fixed_t *point, fixed_t *linepoint, fixed_t *delta, float gap, float *result);

// Other utilities.
RDBOOL		M_DoScreenShot (void);
RDBOOL		M_ValidateID(char *id);			// R3D v211-0
RDBOOL		M_GroupVis(subsector_t *ss);	// R3D v220-4
void		M_SetGlobalYaw(void);			// R3D v220-3
void		M_strcatQuoted(char *dest, char *src);

#endif 