//-----------------------------------------------------------------------------
//
// P_DATA.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __P_DATA_H__
#define __P_DATA_H__

#include "dd_types.h"
#include "p_think.h"
#include "dd_share.h"
#include "doom/mbf_data.h"
#include "doom/mbf_defs.h"

// v220-16 moved here from doom\p_local.h
#define USERANGE		64*FRACUNIT
#define MELEERANGE		64*FRACUNIT
#define MEDIUMRANGE		16*64*FRACUNIT
#define STANDARDRANGE	32*64*FRACUNIT	// v220-14
#define MISSILERANGE	64*64*FRACUNIT	// doubled - was 32*64*FRACUNIT
#define MAXBULLETRANGE	128*64*FRACUNIT	// v220-16 quadrupled

// Max. distance to move in one call to P_XYMovement.
#define MAXMOVE			(30*FRACUNIT)

#define DEFAULT_FRICTION	0xe800

// State flags - do not change
#define ST_LITSKIN		0x00000001	// v220-4
#define ST_DESYNC		0x00000002	// v220-11
// v220-35
// Used with A_WeapSound. Sounds in Abbspack.wad
#define ST_SND1			0x00000004
#define ST_SND2			0x00000008
#define ST_SND3			0x00000010 
#define ST_SND4			0x00000020
#define ST_SND5			0x00000040
#define ST_SND6			0x00000080
#define ST_SND7			0x00000100
#define ST_SND8			0x00000200
#define ST_SND9			0x00000400
#define ST_SND10		0x00000800
#define ST_SND11		0x00001000
#define ST_SND12		0x00002000
#define ST_SNDP1		0x00004000	// punch sound 
#define ST_SNDP2		0x00008000	// melee punch sound 
// Damage flags
#define ST_GD1			0x00010000
#define ST_GD2			0x00020000
#define ST_GD4			0x00040000
#define ST_GD8			0x00080000 
// Used with A_WeapSound. Sounds in Risen3D,wad
#define ST_DS1			0x00100000
#define ST_DS2			0x00200000 
#define ST_DS3			0x00400000
#define ST_DS4			0x00800000

// internal use
#define MAXSTSOUNDS		13			// ST_SND4 to ST_SND12 plus ST_DS1 to ST_DS4
#define ST_DMGSHIFT		16			// v220-35
#define ST_HASMODEL		0x01000000	// v220-10
#define ST_NOMODEL		0x02000000	// v220-10
#define ST_NOTPSPR		0x04000000	// v220-16

#define ST_GDMGMASK		0x000F0000	// v220-35
#define ST_CHNGMASK		0x000FFFFC	// v220-35
#define ST_FLAGMASK		0x00FFFFFF	// v220-35

// mobj_t flags
#define MF_SPECIAL		0x00000001	// Call P_SpecialThing when touched.
#define MF_SOLID		0x00000002	// Blocks.
#define MF_SHOOTABLE	0x00000004	// Can be hit.
#define MF_NOSECTOR		0x00000008	// Don't use the sector links (invisible but touchable).
#define MF_NOBLOCKMAP	0x00000010	// Don't use the blocklinks (inert but displayable)
#define MF_AMBUSH		0x00000020	// Not to be activated by sound, deaf monster.
#define MF_JUSTHIT		0x00000040	// Will try to attack right back.
#define MF_JUSTATTACKED	0x00000080	// Will take at least one step before attacking.
#define MF_SPAWNCEILING	0x00000100	// On level spawning (initial position), hang from ceiling instead of stand on floor.
#define MF_NOGRAVITY	0x00000200	// Don't apply gravity (every tic), that is, object will float, keeping current height or changing it actively.
#define MF_DROPOFF		0x00000400	// This allows jumps from high places.
#define MF_PICKUP		0x00000800	// For players, will pick up items.
#define MF_NOCLIP		0x00001000	// Allow Thing to clip through walls
#define MF_SLIDE		0x00002000	// Player: keep info about sliding along walls.
#define MF_FLOAT		0x00004000	// Allow moves to any height, no gravity. For active floaters, e.g. cacodemons, pain elementals.
#define MF_TELEPORT		0x00008000	// Don't cross lines ??? or look at heights on teleport.
#define MF_MISSILE		0x00010000	// Don't hit same species, explode on block. Player missiles as well as fireballs of various kinds.
#define MF_DROPPED		0x00020000	// Dropped by a demon, not level spawned. E.g. ammo clips dropped by dying former humans.
#define MF_SHADOW		0x00040000	// Use fuzzy draw (shadow demons or spectres), temporary player invisibility powerup.
#define MF_NOBLOOD		0x00080000	// Flag: don't bleed when shot (use puff),  barrels and shootable furniture shall not bleed.
#define MF_CORPSE		0x00100000	// Don't stop moving halfway off a step, that is, have dead bodies slide down all the way
#define MF_INFLOAT		0x00200000	// Floating to a height for a move, ??? don't auto float to target's height.
#define MF_COUNTKILL	0x00400000	// On kill, count this enemy object towards intermission kill total.
#define MF_COUNTITEM	0x00800000	// On picking up, count this item object towards intermission item total.
#define MF_SKULLFLY		0x01000000	// Special handling: skull in flight. Neither a cacodemon nor a missile.
#define MF_NOTDMATCH	0x02000000	// Don't spawn this object in death match mode (e.g. key cards).

	// Player sprites in multiplayer modes are modified using
	// an internal color lookup table for re-indexing.
	// If 0x4 0x8 or 0xc, use a translation table for player colormaps
#define MF_TRANSLATION		0x0c000000
#define MF_TRANSSHIFT		26			// Bits to shift.
#define MF_LOCAL			0x10000000
#define MF_BRIGHTEXPLODE	0x20000000
#define MF_BRIGHTSHADOW		0x40000000
#define MF_VIEWALIGN		0x80000000


// R3D Used with mobj->flags2
#define MF2_ALWAYSLIT		0x00000001
#define MF2_LOGRAV			0x00000002	// alternate gravity setting
#define MF2_NOTELEPORT		0x00000004	// does not teleport
#define MF2_PUSHABLE		0x00000008	// can be pushed by player or other pushable Thing
#define MF2_CANSLOPE		0x00000010	// set if a THING must be orientated with a slope
#define MF2_NOSHADOW		0x00000020	// indicate not to have a shadow
#define MF2_NOFADE			0x00000040	// indicate not to be faded
#define MF2_RANDOM			0x00000080	// Ignore spot orientation and spawn at a random angle
#define MF2_UNCLIPPED		0x00000100	// used to place all sectors a Thing touches in visseclist
#define MF2_OCCLUDE			0x00000200	// indicates a scripted mobj should have any block lines set to occlude when reset
#define MF2_HULLCOLLIDE		0x00000400	// used with sliding door models
#define MF2_EVENLIGHT		0x00000800	// v220-35 use even light by ignoring vtx normals
#define MF2_NOSTEPDOWN		0x00040000	// v220-14 prevent mobj from stepping down > than its height
#define MF2_NOOCCLUDE		0x40000000	// v220-9 set if a model should not be tested for occlusion
#define MF2_LOADMASK		0x40040FFF

// Reserved for internal use only
#define MF2_NORENDER		0x00001000	// not drawn but otherwise treated as a normal mobj
#define MF2_TORCH			0x00002000	// marked for use with dynlights
#define MF2_3DTELEPLINE		0x00004000	// marked as having a z-offset set by a LNF2_3DTELEPLINE type
#define MF2_TELEPCORPSE		0x00008000	// allows a corpse to be teleported
#define MF2_NOTLIT			0x00010000	// inhibits loading DL_AddLuminous
#define MF2_GLUECEIL		0x00020000	// fix amodel to ceiling
#define MF2_DONTDRAW		0x00080000	// used to prevent a Thing being drawn
#define MF2_HEIGHTSET		0x00100000	// used when checking in analyser
#define MF2_NOCHECK			0x00200000	// used to prevent P_ZMovement check with Z set 16xxx Things 
#define MF2_MELEE			0x00400000	// v220-6 set if in melee attack mode
#define MF2_SPECNOMOVE		0x00800000	// v220-5 set with pickup items - cleared if moved from spawn position
#define MF2_CORPSEFADE		0x01000000	// indicate if in suspended fade state
#define MF2_ONSLOPE			0x02000000	// set if player is on a slope
#define MF2_NOAIRSUPPLY		0x04000000	// set only with players and if the air supply is less than 0
#define MF2_NOFITBOTTOM		0x08000000	// set if thing referenced to floor
#define MF2_VDOOR			0x10000000	// vertical door
#define MF2_NOSEESOUND		0x20000000	// prevent seesound playing
#define MF2_FLAGSSET		0x80000000	// indicates that flags2 have been set
#define	MF2_HSETMASK		0x08900000	// (MF2_SPECNOMOVE|MF2_HEIGHTSET|MF2_NOFITBOTTOM)

// R3D Used with mobj->flags3
// v220-4 reorganised
#define MF3_OFFSETPOS		0x00000001	// b211-0 add distance offset to vissprite
#define MF3_OFFSET5			0x00000002	// Increase vmo distance by 5
#define MF3_OFFSET32		0x00000004	// Increase vmo distance by 32
#define MF3_OFFSETMASK		0x00000007

#define MF3_PAIREDSPAWN		0x00000008	// Spot occupation check allowed if both types have this flag set
#define MF3_ALWAYSBLOCK		0x00000010	// Prevents a Things height being changed or the resetting of the MF_SOLID flag on death
#define MF3_NODYNLIGHT		0x00000020	// Prevents a Thing being dynlit (if a model)
#define MF3_NOGLIGHT		0x00000040	// Prevents global light being applied to the Thing (if a model)
#define MF3_USETRUERADIUS	0x00000080	// Use the models true radius (not the ded value)
#define MF3_CHECKPOS		0x00000100	// Used where needed with changing sector height

// Custom specials - flag should be defined in R3D_addon\name\name_addon xx.wad with embedded dd_defn script
#define MF3_SPECIAL			0x00000200	// Custom special pickup (should only be used with Things MISC87 to MISC90)
#define MF3_NEARCLIP		0x00000400	// Only used with scripted map
#define MF3_SWINGDOOR		0x00000800	// Used for swing doors to differentiate from sliding door
#define MF3_MODULATE		0x00001000	// v220-5 modulated water flag

// b220-0 The remainder are reserved for internal use

// spare flag (gold dust)
#define MF3_SPARE1			0x00008000	// 

// R3D specific unless stated otherwise
#define MF3_AMODEL			0x00002000
#define MF3_STEALTH			0x00004000
#define MF3_STATIC			0x00010000
#define MF3_FALLING			0x00020000	// killough $dropoff_fix
#define MF3_DYNLIGHT		0x00040000	// v220-31 identifies mobj as being type 5000, 5500 - 5505
#define MF3_SECRET			0x00080000	// if a thing 15508 is in the map then pickups with the deaf flag set are counted as secret
#define MF3_GLOBALAMBIENT	0x00100000
#define MF3_NOROTATE		0x00200000
#define MF3_ARMED			0x00400000	// killough: Object is armed (for MF_TOUCHY objects)
#define MF3_ONGROUND		0x00800000	// Player is on ground
#define MF3_RESURRECT		0x01000000	// One time resurrect monster spawned in dead state
#define MF3_HASCENTRE		0x02000000

#define MF3_INTERREACT		0x04000000
#define MF3_DEHAMBIENT		0x08000000

// Set by the ded code (not a ded flag)
#define MF3_TOUCHY			0x10000000	// killough 11/98: dies when solids touch it
#define MF3_BOUNCES			0x20000000	// killough 7/11/98: for beta BFG fireballs
#define MF3_FRIEND			0x40000000	// killough 7/18/98: friendly monsters
#define MF3_TRANSLUCENT		0x80000000	// v220-9

#define MF3_LOADMASK		0x30001FFF	// v220-35 TOUCHY and BOUNCY can be set in a ded
#define MF3_TRANSFERMASK	0xFC001FFF	// MF3_LOADMASK + MF3_INTERREACT + MF3_DEHAMBIENT + MF3_TOUCHY etc.

// R3D Used with mobj->flags4 (short)
// COD engine compatibility
#define MF4_LOSTSOUL		0x0001	// spawn a lost soul on death
#define MF4_BADASSZOMBIE	0x0002	// spawn a rocket / drop a rocket on death
#define MF4_BOSSSPID		0x0004	// spawn an arachnotron on death
#define MF4_TRANSPAIN		0x0008	// stealth vile spawns stealth monsters
#define MF4_VILESPAWNER		0x0010	// stealth vile spawns stealth monsters
#define MF4_CODENDTYPE		0x0020	// type used to indicate a special action
// Rest to 0x8000 SPARE

// mobj->misc masks
// v220-4 reorganised
#define MJMISC_FLAGS		0xff000000
#define MJMISC_DEHKEY		0x00ff0000
#define MJMISC_NEGOFF		0x0000ff00

// v220-10 Note: transient flags in range 0xff are not saved/restored with savegame
#define	MJMISC_TRANS_TC		0x00000001	// v220-10 If touching a carry sector
#define	MJMISC_LIT3D		0x00000002	// v220-35 if needing to be 3d lit

// v220-10 Redundant $halo
//#define MJMISC_HALOF		0x000000ff	// initially retained for backward compatibility
//#define MJMISC_HALOVAL	0x0000007f

// internal MJMISC_FLAGS msb flags (used with mobj.misc)
#define MJF_UNLIT			0x01000000	// (prior to v220-31 was MJF_LIFT3D) can only be set with a wadanalyse entry
#define MJF_LINE3D			0x02000000	// thing is on a 3d line
#define MJF_FLAT3D			0x04000000	// thing is on a 3d flat
#define MJF_ONLIFT			0x08000000	// indicates thing on a change height sector needs z-smoothing
#define MJF_FLAMEWALL		0x10000000
#define MJF_SCP_PICKUP		0x20000000
#define MJF_SCP_SPLASH		0x40000000
#define MJF_SCP_3DFLOOR		0x80000000

// killough $dropoff_fix
// For torque simulation:
#define OVERDRIVE 6
#define MAXGEAR (OVERDRIVE+16)


extern int		numthings;

#define MOBJ_ENTRIES  65536

#define MAX_DIVS		 64

#define DEHKEY_SHIFT	16
#define NEGOFF_SHIFT	 8
#define HALOF_SHIFT		 0

#pragma pack(4)

// Tick Command raw data
typedef struct
{
	byte	buttons;
	byte	time;			// The tick this cmd was built.
	char	forwardmove;	// *2048 for move
	char	sidemove;		// *2048 for move
	int		cmd_angle;		// for angle delta
	short	cmd_lookdir;	// view pitch
	short	use;			// usekey active
	short	align;			// usekey active
} ticcmd_t;

// each sector has a degenmobj_t in it's center for sound origin purposes
typedef struct
{
	thinker_t		thinker; // not used
	fixed_t			x,y,z;
} degenmobj_t;

typedef struct
{
	fixed_t		v1[2], v2[2];	// Vector to calculate height (x, y)
	int			mapangle;
	int			dx, dy;
	int			cosfrac, sinfrac;
	int			len;
	int			gradient;
	int			bbox[4];
	ushort		slopeheight;
} subslope_t;

typedef struct
{
	subslope_t slope[2];
} slope_t;

typedef struct vertex_s {
	fixed_t		x, y;
} vertex_t;

typedef struct fvertex_s {
	float		x, y;
} fvertex_t;

typedef struct zvertex_s {
	float		x, y;
	int			zptr;
} zvertex_t;

typedef struct ssecvtxz_s {
	float		z[2];
} ssecvtxz_t;

extern ssecvtxz_t	*ssecvtxz;	// ssec vertices z heights in the ss->verts list

// v220-13 $rotate removed
/*
typedef struct rtparm_s {
	float		len1, len2;
	float		stang1, stang2;
} rtparm_t;

extern rtparm_t *rtparm;
*/

// Thing definition, position, orientation and type,
// plus skill/visibility flags and attributes.
typedef struct
{
    short	x;
    short	y;
    short	angle;
    short	doomednum;
    short	options;
} mapthing_t;

// We'll use the base mobj template directly as our mobj.
typedef struct mobj_s
{
	//----------------------------------------------------------------------
	// Note: these entries must be in this order (see linkmobj_s)
	thinker_t		thinker; 			// thinker node.
	fixed_t			x,y,z;				// position.

	struct mobj_s	*bnext, *bprev;		// links in blocks (if needed).

	//----------------------------------------------------------------------

	nodeindex_t		lineroot;			// lines to which this is linked.

	struct mobj_s	*snext, **sprev;	// links in sector (if needed).
	struct mobj_s	*target;			// Thing being chased/attacked (or NULL) or missile
	struct mobj_s	*tracer;			// Thing being chased/attacked for tracers.
	struct mobj_s	*corelink;			// flame link model.
	struct mobj_s	*onmobj;			// the mobj this one is on top of.
    struct mobj_s	*lastenemy;			// last known enemy -- killough 2/15/98

	// Additional info record for player avatars only. Only valid if type == MT_PLAYER
	struct player_s		*player;		// NULL if not a player mobj.
	struct ddplayer_s	*dplayer;		// ditto

	struct subsector_s	*subsector;		// subsector in which this resides.
	struct state_s		*state;			// ptr to current st

	int				lodstate;			// v220-5 (was 4 byte adjust_sector) v220-9 (was spare_int) use as lodstate;
	spritenum_t		sprite;				// used to find patch_t and flip value.
	mapthing_t		spawnpoint;			// For nightmare respawn.

	mobjinfo_t		*info;				// &mobjinfo[mobj->type]
	int				health;

	fixed_t		momx, momy, momz;	
	fixed_t		radius;
	fixed_t		height;	
    fixed_t		floorz;		// highest contacted floor.
    fixed_t		ceilingz;	// lowest contacted ceiling.
	angle_t		angle;
	short		wallhit;	// the mobj is hitting a wall.
	short		srvo[3];	// short-range visual offset (xyz).
	short		visangle;	// visual angle ("angle-servo").

	int		frame;			// might be or'ed with FF_FULLBRIGHT.
	int		ddflags;		// DD mobj flags (DDMF_*).
	int		valid;			// if == valid, already checked.
	int		type;			// mobj type.
	int		tics;			// state tic counter.
	int		selector;		// multipurpose info.
	int		validcount;		// used when iterating.
	int		flags;			// doom flags
	int		flags2;			// R3D flags
	int		flags3;			// R3D flags
	short	flags4;			// R3D flags

	// Movement direction, movement generation (zig-zagging).
	short	movedir;		// 0-7
	short	movecount;		// when 0, select a new dir
    short	strafecount;    // killough 9/8/98: monster strafing

	// Reaction time: if non 0, don't attack yet.
	// Used by player to freeze a bit after teleporting.
	int		reactiontime;   
	int		threshold;		// If >0, the target will be chased no matter what (even if shot)

	fixed_t	dropoffz;		// killough $dropoff_fix
	short	gear;			// killough 11/98: used in torque simulation
    short	pursuecount;	// killough 9/9/98: How long a monster pursues a target.

	// R3D - Controlling lift sector if mobj on linked 3D lift sector
	//int	liftOffset;		// v220-31 replaced with next two entries (previously liftOffset was unused)
	short	spawnz;			// v220-31 used to store z offset defined in the SCP_ZOFF script at spawn time
	// v220-31 used to store flats offset [0] at spawn time and [1] in game. At spawn time if spawned on a
	//         flats3D structure a level calculated from a value sent in the SCP_ZOFF script will be stored
	//         in [1] and copied to [0]. Since all the initial z info is lost when respawning this can now
	//         be used to ensure respawning is on the correct flats level, especially important if on a lift.
	char	flat3Dlevel[2];

	int		servoz;			// Remember height difference on a slope
	int		lastlook;		// Player number last looked for.
	int		splashbase;		// Splashbase to hook splashes in fake sectors to the surface
	int		iden_nums;		// hi word stores thing num, low word identifier num
	int		turntime;		// $visangle-facetarget
	int		inverseAlpha;	// $mobjalpha
	int		corpsetics;		// $vanish: how long has this been dead?
	int		old_x, old_y;	// Remember previous missile position
	int		old_z;			// v220-4 Remember previous z position
	ushort	startangle;

	// see MJMISC_ flags
	int		misc;			// arranged as follows;
	//byte	flags;			// Flags use by scripted maps
	//byte	neg_offset;		// Value transferred if a neg value set in SCP_ZOFF
	//byte	dehkey;			// Value set if a dehacked key
	//byte	halofactor;		// halo flags (kept for backward ded read compatibility reasons)

	// R3D - modelbox parms
	short	modelboxnum;	// index to current model bounds for sliding door

	// R3D - Scripted model stuff
	short	t79resurrectid; // Used for savegame purposes
	short	scplightlevel;	// If !0 use value else use seclight
	short	scpzoff_ptr;	// If > 0 will be used to access scpzoff[scpzoff_ptr - 1].intensity
	int		lineocclude;	// Line number for script if mf2_occlude flag is set with an AMODEL

	// Flame model handling
	short	torchtic;		// R3D - used to desync torches
	short	skintic;
	float	zcenter;

	// v220-5 scripted rotation
	float	rotateperiod;
	uint	rotateprev;

	// v220-10	last scroll dx, dy values applied
	fixed_t	lastdx, lastdy;

	// v220-16 radius to be used with pickups
	fixed_t	pickupRadius;

	int		mjvalidcount;	// R3D - Additional valid flag (separate logical/rendering checks)

} mobj_t;

typedef struct seg_s 
{
	vertex_t		*v1 , *v2;
	int				rtp_num;
	int				zptr1, zptr2;
	float			fv1[2], fv2[2];
	float			length;			// Accurate length of the segment (v1 -> v2).
	fixed_t			offset;
	struct side_s	*sidedef;
	struct line_s	*linedef;
	struct sector_s	*frontsector;
	struct sector_s	*backsector;	// NULL for one sided lines
	ushort			sgflags;

} seg_t;

typedef struct subsector_s
{
	struct sector_s *sector;
	int			linecount;
	int			firstline;

	// Sorted edge vertices for rendering floors and ceilings.
	short		numverts;		// made short to align - was char (negative numbers allowed ??)
	zvertex_t	*verts;			// A list of edge vertices.
	zvertex_t	midpoint;		// Center of vertices.
	fixed_t		ssbox[4];		// fixed_t extent of bounding box
	int			ssflags;

	// R3D
	int			sslightsec;
	int			*flats3DArray;
	int			floorsec, ceilsec;
	int			lfsec;
	int			lcsec;
	int			viewgroup;
	int			ssvalidcount;
	fixed_t		s_waterplane;	// v220-4
	fixed_t		dlfloorheight;
	fixed_t		dlceilheight;
	short		ssfloorpic;
	short		ssceilpic;
	short		dynlight;
	float		maxdimension;
	int			*xlines;		// v220-25

} subsector_t;


// WATERSEC v220-5

typedef struct {
	fixed_t *z, startz;

} wmo_t;

typedef struct {
	int			secn;
	int			waterthing_count;
	float		start_val;
	int			period;
	fixed_t		floorht;
	wmo_t		*wmo;

} watersec_t;

extern watersec_t	*wsec;
extern int	watersec_count;


typedef struct {
	float	fog_density;
	float	fogrgba[4];
	fixed_t modulate;
	int		period;
	int		phase;
	int		opacity;
} watervar_t;
	
typedef struct sector_s
{
    fixed_t		floorheight, ceilingheight;
    short		floorpic   , ceilingpic;
    short		lightlevel;

    int			validcount;			// if == validcount, already checked
    int			scvalidcount;		// used with P_ThingSectorsIterator

    mobj_t			*thinglist;		// linked list of mobjs in sector
    int				linecount;		// number of lines in linelist
    struct line_s	**lines;		// linelist

	fixed_t 	flooroffx, flooroffy;	// floor texture offsets
	fixed_t		ceiloffx , ceiloffy;	// ceiling texture offset
	int	 		idxf;					// v220-4 index to floor smoothed texture offsets
	int		 	idxc;					// v220-4 index to ceiling smoothed texture offsets

    int			blockbox[4];		// mapblock bounding box for height changes

	float		reverb[NUM_REVERB_DATA];
    int			soundtraversed;		// 0 = untraversed, 1,2 = sndlines -1
    mobj_t		*soundtarget;		// thing that made a sound (or null)
	degenmobj_t	soundorg;			// for any sounds played by the sector

    short	tag;
    short	special;

	int		nexttag,firsttag;		// killough 1/30/98: improves searches for tags.
	short	oldspecial;				// jff 2/16/98 remembers if sector WAS secret (automap)

	/* killough 8/28/98: friction is a sector property, not an mobj property.
	* these fields used to be in mobj_t, but presented performance problems
	* when processed as mobj properties. Fix is to make them sector properties.
	*/
	int		friction,movefactor;
	int		heightsec;			//  killough 3/7/98: support faked flats, other sector, or -1 if no other sector
	int		floorlightsec, ceilinglightsec;	// killough 4/11/98: support for lightlevels coming from another sector
	int		bottommap, midmap, topmap;		// killough 4/4/98: dynamic colormaps
	int		sidesky;						// killough 10/98: support skies coming from sidedefs

    // thinker_t's for reversable actions
	void	*floordata;    // jff 2/22/98 make thinkers on
	void	*ceilingdata;  // floors, ceilings, lighting,
	void	*lightingdata; // independent of one another

	// jff 2/26/98 lockout machinery for stairbuilding
	int		stairlock;   // -2 on first locked -1 after thinker done 0 normally
	int		prevsec;     // -1 or number of sector for previous step
	int		nextsec;     // -1 or number of next step sector

	// Risen3D special conditions
	// For falling rising flats rendered at fixed height
	int		R3Dfloorsec   , R3Dceilsec;
	int		R3Dfloorheight, R3Dceilheight;

	// Risen3D stuff;
	int		*listsec;			// For composite 3D lift structures
	int		liftsec3D;			// Controlling sector used for 3D lift
	int		sector3D;			// Sector containing flats for 3D lines
	short	floorlight3D, ceillight3D;	// 3D flats lightlevels

	// v220-13 $rotate removed
/*	int		rotateX, rotateY;	// For 3D rotating structures
	short	rotateinc;			// and amount to rotate*/

	int		specFlags;			// Analysis flags1
	int		specFlags2;			// Analysis flags2 (v220-24 changed from short to int)

	// Scripted stuff
	short	scatter;			// bullet scatter (used with scripted vdoors)
	short	sightgroup;			// If > 0 will be used to determine whether amodel Things that emit cyclic sounds in linked sectors can be heard by the player
	int		ceiloffset;			// For rising door models

	// slope stuff
	float	slopeangles[2];		// slope pitch for THING rotation

	int		skyfix;				// Offset to ceiling height rendering w/sky.
	fixed_t	sky_offy;			// v220-5 sky offset from referencing lnedef special 271/272
	int		skyvar;				// v220-5 index to sky parms - if not needed kept at -1

	short	walllight;			// Remember sector lightlevel (may be needed when glow's applied)
	int		floodline;			// Line reference for a sector that can flood if moving down 

	// Dynamically set by renderer for more 'esoteric' map structures.
	// In this case lines with same sector each side but having a hi-tex
	// which is expected to be rendered if in a sector with a missing tex
	// causing a change to the ceiling height.  E.g AV sec 188 MAP 19
	int		ceildiff;
	int		stptr;						// Tilted sectors
	fixed_t	smoothfloor, smoothceil;	// Smoothed values
	watervar_t	*wvar;					// v220-3	water values
	int			oversec;
	int			R3Dlowsec;				// v220-10
	short		texwithheight[2];		// $v16added
	fixed_t		withheight;				// $v16added

	// v220-10 secinfo stuff moved here
	float	sibounds[4];
				// Bounding box for the sector.
	// height smoothing
	fixed_t oldfloor[2], oldceiling[2];
	fixed_t	visflooroffset, visceilingoffset;

	// v220-33
	float	scalingValueX;
	float	scalingValueY;


	ushort	siflags;
	// End Risen3D stuff

	plane_t	planes[2];				// Stores info on moving planes

} sector_t;

// v220-35	separated from sector_t for savegame compatibility
typedef struct skyline_s
{
	int	sky_line;
}
skyline_t;

typedef struct side_s
{
    sector_t	*sector;			// Sector the SideDef is facing.

    fixed_t		textureoffset;		// used to adjust texture horizontally
    fixed_t		rowoffset;			// used to adjust texture vertically

    // Indexes to texArray
    short		toptexture;
    short		bottomtexture;
    short		midtexture;

	//---------- R3D Parms ------------

	// Used when editing to check for changes to offsets
	short		orig_textureoffset;
	short		orig_rowoffset;
	int			smoothoff;			// v220-4 Smoothed offsets

	short		texFlags;			// Used mainly by analysis code

    fixed_t		rowoffadj;			// Compatibility fiddle factor

	// for LOS algo, edit, dynlights etc. Dynamically set whilst rendering
	fixed_t		ll, lh, ml, mh, hl, hh;
	fixed_t		sl0, sh0, sl1, sh1, editml, editmh;	// v220-33 parms added

	ushort		flatl, flath;		// Edit flat substitution for missing tex
	short		lsiflags;

} side_t;


typedef struct line3d_s
{
	// 3D stuff
	fixed_t		totalheight, offset3D;
	fixed_t		lineheightlow, lineheightmid;
	fixed_t		scale3Dlow, scale3Dmid;
	short		walllowlightlevel;	// Light level from 3D lines
	short		wallmidlightlevel;	// Light level from 3D lines
	short		wallhilightlevel;	// Light level from 3D lines

} line3d_t;

typedef struct line_s
{
    sector_t	*frontsector, *backsector;
    vertex_t	*v1, *v2;			// Line vertices

	fixed_t		bbox[4];			// lines bounding box
    fixed_t		dx, dy;				// Precalculated v2 - v1 for side checking.
    slopetype_e	slopetype;			// To aid move clipping (nothing to do with R3D slopes).

    short		ml_flags;			// ML_* flags
    short		special, tag;		// Animation related.
	int			firsttag, nexttag;	// killough 4/17/98: improves searches for tags.

    int			validcount;			// if == validcount, already checked
    int			sidenum[2];			// sidenum[1] will be -1 if one sided

	// containers for scripted effects
	ushort		linesound;
	char		*ltext;				// Used with scripted 67nn/77nn types

	// flag or type containers used for analysis and/or rendering
	byte		specType;			// R3D_* identifier
	byte		ambNumLink;			// Link for ambnum
	byte		bv;
	byte		scpFlags;			// v220-3 Extra flags
	int			adjacentsec;		// For rendering door tracks with masked door textures

	int			lnflags2;			// General flags
	short		xtraflags;			// Need yet more flags v220-31

	byte		mtflags, spflags, tyflags, ovflags;

	line3d_t	*d3d;				// data for 3d lines

} line_t;

typedef struct
{
	fixed_t	x,y,dx,dy;		// partition line
	fixed_t	bbox[2][4];		// bounding box for each child
	int	children[2];		// if NFx_SUBSECTOR it's a subsector

} node_t;

// Player sprites.
typedef struct						
{
	state_t	*stateptr;
	int		tics;
	float	plight;			
	float	alpha;			
	float	x, y;	
	int		flags;
	int		offx, offy;
} ddpsprite_t;

typedef struct ddplayer_s
{
	mobj_t		*mo;					// pointer to a (game specific) mobj
	fixed_t		viewz;					// focal origin above r.z
	fixed_t		viewheight;				// base height above floor for viewz
	fixed_t		deltaviewheight;	
	float		lookdir;				// It's now a float, for mlook.
	int			fixedcolormap;			// can be set to REDCOLORMAP, etc
	int			extralight;				// so gun flashes light up areas
	int			ingame;					// is this player in game?
	int			flags;		
	int			filter;					// RGBA filter for the camera
	int			clAngle;				// client side
	float		clLookDir;				// client side
	angle_t		lastangle;				// For calculating turndeltas.
	ddpsprite_t	psprites[DDMAXPSPRITES];// Player sprites.
	void		*extradata;				// Pointer to any game-specific data.
} ddplayer_t;

typedef struct intercept_s
{
	fixed_t		frac;		// along trace line
	RDBOOL		isaline;
	union {
		mobj_t	*thing;		
		line_t	*line;		
	} d;
} intercept_t;

typedef RDBOOL (*traverser_t) (intercept_t *in);

typedef struct
{
	int		sidenum;
	ushort	offsetlow;
	ushort	offsethigh;
	byte	type[2];

} slsid_t;

typedef struct{
	fixed_t	ox, oy, oz;
} mposv_t;

typedef struct{
	mposv_t	mp[MOBJ_ENTRIES];
	float	mt, tm;
} mpos_t;

extern mpos_t mpos;

#pragma pack()

extern slsid_t	*slsid;

void	P_LoadReject(int lump);
void 	P_MobjThinker (mobj_t* mobj);

extern mobj_t	mobjarray[MOBJ_ENTRIES];
extern int		useBulletRange;	// v220-16

extern int			numvertexes;
extern vertex_t		*vertexes;

extern int			numsegs;
extern seg_t		*segs;

extern int			numsectors;
extern sector_t		*sectors;
extern skyline_t	*skl;

extern int			numsubsectors;
extern subsector_t	*subsectors;

extern int			numnodes;
extern node_t		*nodes;

extern int			numlines;
extern line_t		*lines;

extern int			numsides;
extern side_t		*sides;

// Slopetypes (tilted sectors)
extern slope_t		*slopes;
extern int			numslopes, maxnumslopes;

extern slsid_t		*slsidarray;
extern int			numsidarray, maxnumsidarray;

// Gravity for the current map.
extern fixed_t		mapgravity;

#endif
