//-----------------------------------------------------------------------------
//
// P_MAPUTIL.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __P_MAPUTIL_H__
#define __P_MAPUTIL_H__

typedef struct
{
	fixed_t x, y, dx, dy;
} divline_t;

extern divline_t trace;

typedef struct {
	mobj_t	**mobj_list;
	int		listcnt;
	int		maxlistcnt;
} linklist_t;

fixed_t	P_ApproxDistance (fixed_t dx, fixed_t dy);
fixed_t P_ApproxDistance3(fixed_t dx, fixed_t dy, fixed_t dz);
float	P_MobjSoundDistancef(mobj_t *end, float *fixpoint, float attenuation);
int		P_PointOnNodeSide (fixed_t x, fixed_t y, const node_t *node);
int		P_PointOnLineSide (fixed_t x, fixed_t y, const line_t *line);
int		P_CheckOnLineSide (fixed_t x, fixed_t y, const line_t *line);
int		P_PointOnViewSide (float x, float y, float dx, float dy);
int		P_BoxOnLineSide (fixed_t *tmbox, line_t *ld);
fixed_t	P_InterceptVector (const divline_t *v2, const divline_t *v1);
void	P_LineOpening (line_t *linedef, int moheight, int x, int y);
void	P_LinkThing (mobj_t *thing, int flags);
void	P_UnlinkThing (mobj_t *thing);
RDBOOL	P_BlockLinesIterator (int x, int y, RDBOOL(*func)(line_t*,void*), void*);
RDBOOL	P_BlockThingsIterator (int x, int y, RDBOOL(*func)(mobj_t*,void*), void* );
RDBOOL	P_BlockSpecialIterator(int x, int y, RDBOOL(*func)(mobj_t*,void*), void*);
RDBOOL	P_ThingLinesIterator (mobj_t *thing, RDBOOL (*func)(line_t*,void*), void*);
RDBOOL	P_ThingSectorsIterator (mobj_t *thing, RDBOOL (*func)(sector_t*,void*), void *data);
RDBOOL	P_MoveSectorThingsIterator (sector_t *sector, RDBOOL (*func)(mobj_t*,void*), void *data);
void	P_SectorTouchingThingsList(sector_t *sector, RDBOOL add_amodel);
RDBOOL	P_PathTraverse (fixed_t x1, fixed_t y1, fixed_t x2, fixed_t y2, RDBOOL usedist, int flags, RDBOOL (*trav) (intercept_t *));

// v220-3 transferred from r_util.h
angle_t		R_PointToAngle (fixed_t x, fixed_t y);
angle_t		R_PointToAngle2 (fixed_t px, fixed_t py, fixed_t x, fixed_t y);
subsector_t *R_PointInSubsector (fixed_t x, fixed_t y);

// R3D - put line opening info into structure and add fields for R3D extensions
typedef struct
{
	sector_t	*front, *back;

	fixed_t		shootz, dist, frac;
	fixed_t		topslope, bottomslope; // slopes to top and bottom of target
	fixed_t		aimslope, attackrange;
	fixed_t		editz;
	fixed_t		origslope;		// R3D store starting aim slope
	int			autoaim;		// R3D for better autoaim determination
	fixed_t		lastx, lasty, lastz;
	RDBOOL		has3Dflats;

} lineopen_t;

extern lineopen_t	lop;

extern RDBOOL		make_noway_sound;
extern RDBOOL		is_sliding;
extern mobj_t		*shootthing;
extern mobj_t		*slidemo;

#endif
