//-----------------------------------------------------------------------------
//
// P_PARTICLE.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __P_PARTICLE_H__
#define __P_PARTICLE_H__

#define MAX_ACTIVE_PTCGENS	96
#define MAX_PTC_STAGES		DED_PTC_STAGES
#define MAX_PTC_TEXTURES	32			// Maximum # of textures in particle system

// Generator flags
#define PGF_STATIC				0x00000001	// Can't be replaced by anything.
#define PGF_RELATIVE_VELOCITY	0x00000002	// Particles inherit source's velocity.
#define PGF_SPAWN_ONLY			0x00000004	// Generator is spawned only when source is being spawned.
//#define PGF_ANGLE_SPAWN		0x00000008	// Rotate spawn vector w/mobj angle.
#define PGF_ADD_BLEND			0x00000010	// Render using additive blending.
#define PGF_FLOOR_SPAWN			0x00000020	// Flat-trig: spawn on floor.
#define PGF_CEILING_SPAWN		0x00000040	// Flat-trig: spawn on ceiling.
#define PGF_SPACE_SPAWN			0x00000080	// Flat-trig: spawn in air.
#define PGF_PARTS_PER_128		0x00000100	// Definition specifies a density.
#define PGF_MODEL_ONLY			0x00000200	// Only spawn if source is a 3D model.
#define PGF_SCALED_RATE			0x00000400	// Spawn rate affected by a factor.
#define PGF_DIFFUSE_NEAR		0x00000800	// R3D - Fixes rend_particle_diffuse at 1.6f max
#define PGF_BUBBLES				0x00001000	// R3D - Player bubbles id
#define PGF_SMOKE				0x00002000	// Could be overhead
#define PGF_SUSPEND				0x00004000	// Generator can be suspended

// Runtime generator flags:
#define PGF_VISIBLE				0x01000000	// v220-4 Should be rendered.
#define PGF_UNTRIGGERED			0x02000000
#define PGF_SUSPENDED			0x04000000	// v220-4 Should be suspended.

// Particle types
#define PTC_NONE	  0
#define PTC_POINT	  1
#define PTC_LINE	  2
// New types can be added here.
#define PTC_TEXTURE 100

// Particle flags
#define PTCF_STAGE_TOUCH		0x0001	// Touching ends current stage.
#define PTCF_DIE_TOUCH			0x0002	// Dies from first touch.
#define PTCF_BRIGHT				0x0004	// Fullbright.
#define PTCF_SHADING			0x0008	// Pseudo-3D.
#define PTCF_PLANE_FLAT			0x0010	// Touches a plane => render as flat.
#define PTCF_STAGE_WALL_TOUCH	0x0020	// Touch a wall => end stage.
#define PTCF_STAGE_FLAT_TOUCH	0x0040	// Touch a flat => end stage.
#define PTCF_WALL_FLAT			0x0080	// Touches a wall => render as flat.
#define PTCF_SPHERE_FORCE		0x0100

typedef struct
{
	char			stage;			// -1 => particle doesn't exist
	short			tics;
	fixed_t			pos[3];			// Coordinates.
	fixed_t			mov[3];			// Momentum.
	float			lowlimit;		// Lowest z value allowed
	subsector_t		*subsector;		// Updated when needed.
	line_t			*contact;		// Updated when lines hit/avoided.
} particle_t;

typedef struct
{
	short			type;
	short			flags;
	int				resistance;
	int				bounce;
	int				radius;
	int				gravity;
} ptcstage_t;

// Particle Generator
typedef struct ptcgen_s
{
	thinker_t		thinker;		// Func = P_PtcGenThinker
	sector_t		*sector;		// Flat-triggered.
	int				plane;			// Flat-triggered.0 - floor 1 - ceil
	float			area;			// Rough estimate of sector area.
	ded_ptcgen_t	*def;			// The definition of this generator.
	mobj_t			*source;		// If mobj-triggered.
	thid_t			srcid;			// Source mobj ID.
	int				type;			// Type-triggered; mobj type number (-1=none).
	int				type2;			// Type-triggered; alternate type.
	fixed_t			center[3];		// Used by untriggered/damage gens.
	fixed_t			vector[3];		// Converted from the definition.
	int				flags;
	float			spawncount;
	int				spawncp;		// Spawn cursor.
	int				age;
	int				count;			// Number of particles.
	particle_t		*ptcs;			// List of particles.
	ptcstage_t		stages[MAX_PTC_STAGES];
} ptcgen_t;

void	P_PtcGenThinker(ptcgen_t *gen);
void	P_SpawnParticleGen(ded_ptcgen_t *def, mobj_t *source);
void	P_SpawnTypeParticleGens(void);
void	P_SpawnMapParticleGens(void);
void	P_SpawnDamageParticleGen(mobj_t *mo, mobj_t *inflictor, int amount);
void	P_CheckPtcPlanes(void);
void	P_InitParticledefSectors(void);

float	P_GetParticleRadius(ded_ptcstage_t *stage_def, int ptc_index);
fixed_t	P_GetParticleZ(particle_t *pt);

#endif
