//-----------------------------------------------------------------------------
//
// P_THINK.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// Revision 2.1.1 build 0 07/Dec/2006 gmj (b211-0)
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __P_THINK_H__
#define __P_THINK_H__

#include "dd_types.h"

// think_t is used to signal that a thinker exists but is no longer valid
typedef void (*think_t) ();

// Doubly linked list of mobjs.
typedef struct thinker_s
{
	struct		thinker_s	*prev, *next;
	think_t		function;
	
	// killough 8/29/98: we maintain thinkers in several equivalence classes,
	// according to various criteria, so as to allow quicker searches.
	struct		thinker_s *cnext, *cprev; // Next, previous thinkers in same class
	
	// killough 11/98: count of how many other objects reference
	// this one using pointers. Used for garbage collection.
	uint		mobjreferences;

	thid_t		id;			// Only used for mobjs (zero is not an ID).
	thid_t		spawnid;	// b211-0 If grenade etc. holds id of mobj that spawned it
} thinker_t;

// killough 8/29/98: threads of thinkers, for more efficient searches
#define th_misc		0
#define th_friends	1
#define th_enemies	2
#define NUMTHCLASS	3

extern thinker_t thinkerclasscap[NUMTHCLASS];

extern thinker_t thinkercap;
extern thinker_t fnthinkcap;

void	P_RunThinkers(void);
void	P_InitThinkers(void);
void	P_InitParticleThinkers(void);
void	P_AddThinker(thinker_t *thinker);
void	P_AddFNThinker(thinker_t *thinker);
void	P_AddParticleThinker(thinker_t *thinker);
void	P_RemoveThinker(thinker_t *thinker);

// Extended thinker functions - killough
void	P_UpdateThinker(thinker_t *thinker);

#endif

