//-----------------------------------------------------------------------------
//
// R3D_COMPAT.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __R3D_COMPAT_H__
#define __R3D_COMPAT_H__

#include <windows.h>

#include "p_data.h"
#include "r3d_defs.h"
#include "r3d_arrays.h"

extern HWND			hWndMain;
extern HDC			mainHdc;	// v220-8 made global
extern HINSTANCE	hInstApp;


// letter used for service release with minor changes
// Previous				"3"					// build 3  29/Oct/2008 (v220-3)
// Previous				"4"					// build 4  23/Jul/2009 (v220-4)
// Previous				"5"					// build 5  20/Dec/2009 (v220-5)
// Previous				"6"					// build 6  19/Mar/2010 (v220-6)
// Previous				"7"					// build 7  05/Jul/2010 (v220-7)
// Previous				"8"					// build 8  14/Aug/2010 (v220-8)
// Previous				"9"					// build 9  18/Dec/2010 (v220-9)
// Previous				"10"				// build 10 14/May/2011 (v220-10)
// Previous				"11"				// build 11 08/Aug/2011 (v220-11)
// Previous				"12"				// build 12 01/Oct/2011 (v220-12)
// Previous				"13"				// build 13 15/Mar/2012 (v220-13)
// Previous				"14"				// build 14 01/May/2012 (v220-14)
// Previous				"15"				// build 15 08/Jun/2012 (v220-15)
// Previous				"16"				// build 16 28/Sep/2012 (v220-16)
// Previous				"17"				// build 17 27/Oct/2012 (v220-17)
// Previous				"18"				// build 18 05/Jan/2013 (v220-18)
// Previous				"19"				// build 19 21/Jan/2013 (v220-19)
// Previous				"20"				// build 20 31/Mar/2013 (v220-20)
// Previous				"21"				// build 21 02/Jun/2013 (v220-21)
// Previous				"22"				// build 22 14/Dec/2013 (v220-22)
// Previous				"23"				// build 23 10/Jan/2014 (v220-23)
// Previous				"24"				// build 24 20/Mar/2014 (v220-24)
// Previous				"25"				// build 25 13/May/2014 (v220-25)
// Previous				"26"				// build 26 05/Jul/2014 (v220-26)
// Previous				"27"				// build 27 09/Sep/2014 (v220-27)
// Previous				"28"				// build 28 04/Oct/2014 (v220-28)
// Previous				"29"				// build 29 22/Nov/2014 (v220-29)
// Previous				"30"				// build 30 03/Dec/2014 (v220-30)
// Previous				"31"				// build 31 22/Dec/2016 (v220-31)
// Previous				"32"				// build 32 18/Mar/2017 (v220-32)
// Previous				"33"				// build 33 08/Jul/2018 (v220-33)
// Previous				"34"				// build 34 12/Nov/2018 (v220-34)
#define BUILDTEXT		"35"
#define VERSIONNUM		"22035"				// build 35 06/Dec/2021 (v220-35)
#define VERSIONSTR		"r3dv2.2.35"		// (v220-35)
#define VERSIONTEXT		"Risen3D V2.2.0"

#define DOOMWADSUPPLIER	"www.idsoftware.com or www.steampowered.com"
#define RISEN3DLINK		"http://risen3d.drdteam.org/"

#define RATIO			(85.0f / 110.0f)	// v220-3 establish common reference value
#define RATIO_FACTOR	(1.0f/1.2f)			// Used to compensate for the Doom 1:1.2 ratio

#define TIDNUM(x)	((int)(x->iden_nums & 0xFFFF))		// thing identifier
#define MAPNUM(x)	((int)((uint)(x->iden_nums) >> 16))	// thing mapnum

#define WADV	"wrong version of Risen3D.wad or *.deds installed"

#define MAX_SEC_PER_VTX		64
#define MAX_MODEL_LIGHTS	10

#define MAXZLISTPTR			4	// see zdistlist

#define PLR_VIEWHT 41*FRACUNIT	// v220-5 player view height fixed
// v220-8 melee heights
#define PLR_VIEWMINHT FRACUNIT
#define PLR_VIEWMAXOFFSET 10*FRACUNIT

// v220-3 underwater damage timeouts
#define		UWMAXTIME		30
#define		UWMAXTICS		(UWMAXTIME * TICRATE)

// b211-0 HUD sprite type flags - used with hsflags
#define		HS_IWADNAME		0x0001
#define		HS_PWADNAME		0x0002
#define		HS_CUSTONAME	0x0004
#define		HS_PNGTEX		0x0008
#define		HS_JPGTEX		0x0010
#define		HS_DOOMTEX		0x0020
#define		HS_IGNORE		0x0040

// b211-0 ded select switch flags
// revised v220-35
#define SWF_NC				0x00000001
#define SWF_NM				0x00000002
#define SWF_XA				0x00000004		// No models but CFX
#define SWF_SPZ				0x00000008		// Sitter's pistol guy models
#define SWF_AM				0x00000010		// Sitters and Abbs models
#define SWF_FP				0x00000020		// Fullpack weapons
#define SWF_AP				0x00000040		// Alternative weapons
#define SWF_FX				0x00000080		// Extra particle effects
#define SWF_GND				0x00000100		// (launcher) grenade  replaces rocket
#define SWF_PBM				0x00000200		// (launcher) pipebomb replaces rocket
#define SWF_DEH				0x00000400		// DEH loaded
#define SWF_NOAMWDS			0x00000800		// Disable model replacement
#define SWF_NUM				12				// Number of SWF_ definitions

// For dehacked sprites and things marking.
#define DH_NUMOBJECTS	MT_STALAG		// Limit for DEHACKED Things
#define DH_NUMSTATES	S_GRENADEX		// Limit for DEHACKED states search DEH NUM STATES
#define DH_CHECKNUM		DH_NUMSTATES

#define MAX_MOVESEC		5		// Usually only one

// mthing option flags.
#define	MTF_EASY		0x0001
#define	MTF_NORMAL		0x0002
#define	MTF_HARD		0x0004
#define	MTF_AMBUSH		0x0008	// Deaf monsters (do not react to sound).
#define	MTF_MULTIPLAYER 0x0010
#define	MTF_DEAD		0x0020	// MTF_DEAD == deathmatch
#define	MTF_ALIVE		0x0040	// MTF_ALIVE == co-op
#define	MTF_FRIEND		0x0080	// MTF_FRIEND	MBF flag

#define	MTF_NOKILL		0x8000	// R3D specific internal only

// Scripted Model defs - much of the way this is designed is to allow for savegame support

// Used with zoflags
#define	ZO_OFFSET5			0x0001
#define	ZO_LIFTSPEED		0x0002

// Used with setflags
#define	T_SET				0x0001	// set if spot has been loaded
#define	T_MULTISET			0x0002	// set if spot has more than one map instance
#define	T_UNSET				0x0004	// set to disable a 78xx spot
#define	T_ENABLE			0x0008	// set to re-enable a disabled 78xx spot

// Used with t79flags
#define	T79_TRIGGER			0x00000001	// set if t77nn has MTF_HARD set
#define	T79_MOVSEC			0x00000002	// set if t77nn line accesses a mov sector
#define	T79_RESET			0x00000004	// set if t79nn's model is to be reset
#define	T79_INITDEAD		0x00000008	// set if t79nn's model is initialised killed
#define	T79_MESSAGESET		0x00000010	// set if t79nn's text message isto be shown
#define	T79_MESSAGEMANY		0x00000020	// set if t79nn's text message is always shown
#define	T79_DELAYACTIVE		0x00000040	// set if t79nn's mo is to be reset after a delay
#define	T79_NOAMBIENT		0x00000080	// set if t79nn's ambient was stopped
#define	T79_REMOVE			0x00000100	// set if t79nn's amodel is to be removed
#define	T79_TEXCHANGE		0x00000200	// set if t77nn's line's texture is to be changed
#define	T79_VDOOR			0x00000400	// set if a vertical door type
#define	T79_VDOORCRUSH		0x00000800	// set if a vdoor's falling sector should crush anything under it
#define	T79_VDOORREVERSE	0x00001000	// set if a vdoor's falling sector should reverse if hitting anything under it
#define	T79_TRIGRESURRECT	0x00002000	// set if an amodel is to be resurrected from a line trigger
#define	T79_ENDLEVELTYPE	0x00004000	// set if an amodel is an endlevel type
#define	T79_SPAWNONCE		0x00008000	// set if ondeath spawn should only be done once where the type can be resurrected
#define	T79_SPAWNDONE		0x00010000	// set if ondeath spawn has been done
#define	T79_VDOORWAIT		0x00020000	// wait for a gate to open or close
#define	T79_INTERREACT		0x00040000	// set if an AMODEL requires lines to be created to use the blocking code (certain types only)
#define	T79_TRANSFER		0x00080000	// set if t79nn's mo is to be reset after a delay
#define	T79_ACTIVATE		0x00100000	// set if a 74xx thing's line is to be activated
#define	T79_CHECKACTIVATE	0x00200000	// set if t79nn's amodel is to be removed irrespective of any other commands
#define	T79_CHECKLIST		0x00400000	// set if useed with CHECKU79LIST
#define	T79_SWAPTRIGGER		0x00800000	// set if line specials for mov sectors need to be specifically selected
#define	T79_NULLTRIGGER		0x01000000	// set if t79's thingref78 points to a 78nn thing's lineref78 and disables block
#define	T79_SETTRIGGER		0x02000000	// set if t79's thingref78 points to a 78nn thing's lineref78 and enables block
#define	T79_SETTOGGLE		0x04000000	// set if t79's thingref78 points to a 78nn thing's lineref78 and toggles block

// Used with t79keys
#define	T79_PK_BCARD		0x0001
#define	T79_PK_RCARD		0x0002
#define	T79_PK_YCARD		0x0004
#define	T79_PK_BSKULL		0x0008
#define	T79_PK_RSKULL		0x0010
#define	T79_PK_YSKULL		0x0020
#define T79_CARDKEYMASK		0x0007
#define T79_SKULLKEYMASK	0x0038
#define T79_ALLKEYMASK		0x003F
	// Monster can't open if set and player doesn't have the card
#define	T79_MK_BCARD		0x0100
#define	T79_MK_RCARD		0x0200
#define	T79_MK_YCARD		0x0400
#define	T79_MK_BSKULL		0x0800
#define	T79_MK_RSKULL		0x1000
#define	T79_MK_YSKULL		0x2000
#define T79_ALLNMYKEYMASK	0x3F00

#define	T79_REMOVEHUDKEY	0x4000	// set if t79nn's key is not to be seen in the hud following use
#define	T79_NOHUDKEY		0x8000	// set if t79nn's key should not be shown in the hud

// Used with th7400 flags
#define T74_FOG				0x0001
#define T74_FOGSOUND		0x0002
#define T74_CSECLIGHT		0x0004
#define T74_FSECLIGHT		0x0008
#define T74_CEILPIC			0x0010
#define T74_FLOORPIC		0x0020
#define T74_PLAYERONLY		0x0040
#define T74_TRIGLINEONLY	0x0080

// used with smflags
#define SM_ENDSET			0x0001
#define SM_MASSACRE			0x0002
#define SM_NOSPAWN			0x0004
#define SM_ENDLEVEL			0x0008
#define SM_DONTENDLEVEL		0x0010

// used with sspflags
#define SSPF_TOUCHY			0x0001
#define SSPF_BOUNCES		0x0002
#define SSPF_FRIEND			0x0004
#define SSPF_MASK			0x0007
#define SSPF_NOFRIENDCOUNT	0x0010

// used with boundflags
#define MBOUND_LINESSET		0x0001

#define MIN_SCR_WIDTH		640
#define MIN_SCR_HEIGHT		480
#define SZWDATA				 28	// 7 ints
#define SZNMODE				 20	// 5 ints
#define UIHZ_DESKTOP		  1
#define UIHZ_CURRENT		  2
#define UIHZ_DEFAULT		  4

// v220-35
#define MAXVDOORNUM		100
// v220-35
extern mobj_t	*vdoormo[MAXVDOORNUM][2];
extern int		vdoornum;

// v220-4 stats
typedef struct {
	// b211-0 stress parms
	int		global_mobj_count;
	int		missile_mobj_count;

	// for profiling purposes
	uint	fpsnum;
	uint	framecount;

	int		num_mobjs;

} stats_t;

extern stats_t stats;

// v220-3 Resolution support
typedef struct {
	uint	width;
	uint	height;
	uint	freq;
	uint	xpos;
	uint	ypos;
} dispmode_t;

extern dispmode_t *modeNlist;

extern int		modeNcount;
extern int		current_w;
extern int		current_h;
extern uint		scr_max_width;
extern uint		scr_max_height;
extern RDBOOL	valid_screen_ratio;
extern float	screen_ratio;

// screen handling
typedef struct {
	// *DO NOT CHANGE THE ORDER*
	// current values
	int		width;
	int		height;
	int		freq;
	int		xpos;
	int		ypos;
	int		bpp;
	int		fullscreen;

	// default values
	int		d_width;
	int		d_height;
	int		d_freq;
	int		d_xpos;
	int		d_ypos;
	int		d_bpp;
	int		d_fullscreen;

	// temporary values
	int		t_width;
	int		t_height;
	int		t_freq;
	int		t_xpos;
	int		t_ypos;
	int		t_bpp;
	int		t_fullscreen;

	// desktop values
	int		max_width;
	int		max_height;
	int		w_initialfreq;
	int		w_initialbpp;

	// frame values
	int		szCaption;
	int		szBorder;

	// scaling values
	float	scalex;
	float	scaley;
	float	weaponFOV;
	float	automap_factor;	// v220-15 not used
	float	scr_aspect;

	// resolution parms
	RDBOOL	w_setdefaults;
	int		w_checkwin;

} currentmode_t;

extern currentmode_t	cMode;

extern int init_FOV;

// SCRIPTING =========================

// v220-4
extern char *music_list[20];	// 0x1FCC - 0x1FB8

// Scripted soundthing support
typedef struct {
	short	type;
	int		soundid;
	int		oldsoundid;	// v220-33
	float	volume;
	float	attenuation;

} soundthing_t;

extern soundthing_t *sthing;

// Scripted reverb support
#define MAX_REVERBTHING	128

typedef struct {
	ushort		num;		// Tie to script
	int			thcount;	// For scp_debug stats
	int			seccount;
	int			maxseccount;
	int			*secnum;	// list of sectors things with same num are in
} reverb_t;

extern reverb_t	reverbThing[MAX_REVERBTHING];
extern int		fakeidentifier[100];

#define NUM7400SPOTS	100
#define NUMSPOTS		400
#define NUMRANGES		4		// currently 7900, 6900, 5900 and 4900

#pragma pack(2)		// Align to short

// NOTE: first five entries must be identical to mapthing_t
typedef struct {
    short	x;
    short	y;
    short	angle;
	short	type;		// Differs from mapthing_t in that this holds mt_type
	short	options;

	//-----------------
    short	z_off;		// For scripted z-offset
	short	z_nofall;
	short	spawn_fog;
	short	sspflags;
	short	set;
	mobj_t	*mo;

} spawnspot_t;

#define MAXDM_NUM	10	// v220-19 increased from 5 to 10

typedef struct {
	int		type;
	line_t	*linenum;
} deathmon_t;

typedef struct {
	line_t	*refline;
	ushort	flags;
	ushort	ptrnum;
	ushort	delay;			// added v220-31
	line_t	*lptr[10];
	// added v220-35 to support door open/close toggle
	int		msect;
	ushort	triga, trigb;

} th7400_t;

// NOTE: first five entries must be identical to mapthing_t
typedef struct {
    short	x;
    short	y;
    short	angle;
	short	type;
	short	options;

	//-----------------
	int		refline78;		// v 35 used for active blocking -->
	int		secdamage;
	mobj_t	*refmo;
	line_t	*blocklines[5];
	char	*blocktextmsg;
	char	*unblocktextmsg; // --<
	short	spawn[10][5];
	short	spawn_count;
	short	setflags;
} th7800_t;

// NOTE: first five entries must be identical to mapthing_t
typedef struct {
    short	x;
    short	y;
    short	angle;
	short	type;
	short	options;

	//-----------------
	int		t77lnum;
	short	t77special;
	// v220-35
	short	t77dummyspecial;
	short	t77dummytag;
} t77info_t;

typedef struct {
	int			maxspawn_count;
	int			spawn_count;
	short		delay;			// applies to all lines
	short		blocked;
	int			direcset[10][2];
	short		direccount;
	t77info_t	*info;
} th7700_t;

typedef struct {
	int		*lnum;
	short	*spec;
	int		count;
	int		*monum;
	int		mocount;
} redirect_t;

extern redirect_t	redir77[NUMSPOTS];

// NOTE: first five entries must be identical to mapthing_t
//       t77info_t cannot be shared
typedef struct {
    short	x;
    short	y;
    short	angle;
	short	type;
	short	options;

	//-----------------
	uint		t79flags;
	short		t79keys;

	short		killthing[10][3];
	short		killcount;

	short		replacespot;
	ushort		replacetype[2];

	ushort		replacemo[10][3];
	short		replacemocount;

	short		spawn78nn[10];
	short		spawndef_num;

	short		stopnum;		// 79num to stop sound with

	short		resurrect;		// delay time to resurrect (ignored if 0)

	short		delay;			// Delay to unblocking lines
	short		resetdelay;		// Delay before resetting model

	short		killambient;

	short		doortics;		// tics to change plat height or open a model door
	short		offset;			// Amount to change plat height

	ushort		wait;			// time to wait before closing (if 0 then will stay open)
	ushort		falltics;
	int			startheight;

	int			thingref78;

	struct scriptedsound_s *ss;

	int			soundlist[10];
	int			u79sndid;
	int			locksndid;
	int			soundid;
	float		volume;
	
	int			activesec;
	short		tag;

	int			activate;

	int			vdoorsec;
	fixed_t		vdoorsecheight;
	char		*vdoorwaitmsg;

	short		cotrigger;
	short		list79num;	// v220-23 if >= 0 then a list of 79xx things to check
	short		setflags;

	int			changemsgnum;
	int			witht79num;

	char		*textmsg;
	char		*lockmsg;
	char		*u79msg;
	mobj_t		*mo;
	mobj_t		*pickupmo;
	degenmobj_t	dmo1;
	degenmobj_t	dmo2;

} th7900_t;

#define MAXQUEUEMSG	10

typedef struct {
	char	*endlevelMessage;
	char	*scriptMessage;
	char	*queuemsg[MAXQUEUEMSG];
	char	info[9][16];
	short	qmsgcount;
	float	alpha;
	short	delay_to_next_text;
	short	delay_to_fade;
	short	endlevel_start_delay;
	short	endlevel_delay_to_fade;
	short	smflags;

} scriptMessage_t;

extern scriptMessage_t scpMessage;

typedef struct scriptedsound_s {
	thinker_t	thinker;
	int			soundid;
	int			delay;
	short		t79id;

} scriptedsound_t;

typedef struct {
	thinker_t	thinker;
	short		tnum;
	short		delay;
	short		occupied;

} modelkill_t;

typedef struct {
	thinker_t	thinker;
	int			spawn_count;
	int			delay_tic;
	spawnspot_t	spawn_sp;
	short		t78num;

} modelspawn_t;

typedef struct {
	thinker_t	thinker;
	int			lnum;
	short		special;
	short		delay;
	short		delay2;
	short		t79num;

} modelreset_t;

typedef struct {
	thinker_t	thinker;
	int			tics;
	ushort		t77num;		// might be used for some other parm
	ushort		t79num;

} delayline_t;

typedef struct {
	thinker_t	thinker;
	int			t79num;
	RDBOOL		block;

} lineocclude_t;

typedef struct {
	thinker_t	thinker;
	int			t79num;
	int			sector;
	ushort		active;
	ushort		wait;
	fixed_t		inc_up;
	fixed_t		inc_down;
	fixed_t		increment;

} adjustplat_t;

#pragma pack()

extern deathmon_t	deathmonitor[MAXDM_NUM];
extern th7400_t		th7400[NUM7400SPOTS];
extern th7700_t		th7700[NUMSPOTS];
extern th7800_t		th7800[NUMSPOTS];
extern th7900_t		th7900[NUMSPOTS];
extern byte			t79list[NUMSPOTS];

extern int T78_disable[NUMSPOTS][7];

int get77N(int num);
int get78N(int num);
int get79N(int num);
int AS_GetPushSound(int dmednum);

// v220-35
extern int		*doorspotlnums;
extern int		numdoorspots;
extern int		maxnumdoorspots;
extern byte		msgchangedarray[NUMSPOTS];

// End  Scripted Model defs --

// SCP_ZOFF	script -----------

typedef struct
{
	short	thingnum;
	short	z_off;
	short	lightlevel;
	short	zoflags;
	fixed_t	intensity;

} scpzoff_t;

extern scpzoff_t scpzoff[900];
extern ushort	scplinesound[100][2];

typedef enum {
	ELV_SPEED,
	ELV_SECN,
	ELV_HEIGHT,
	ELV_TAG,
	ELV_HILIMIT,
	ELV_LOLIMIT,
	ELV_NUM
} elv_e;

extern int		elevspeed[100][ELV_NUM];
extern RDBOOL	elevspeedset;
extern RDBOOL	elevheightset;
extern RDBOOL	scpofftagsset;

typedef enum {
	PF_TAG,
	PF_REFSEC,
	PF_DAMAGE,
	PF_FILTER,
	PF_NUM
} pf_e;

extern int		pseudofake[100][PF_NUM];

//-----------------------------

// R3D - smooth scroll vars
typedef struct
{
	fixed_t	sdx;
	fixed_t	sdy;
	fixed_t	dx;
	fixed_t	dy;
	fixed_t	oldx;
	fixed_t	oldy;
	int		s;

} scvar_t;

typedef struct
{
	scvar_t	*c;	// ceiling
	scvar_t	*f;	// floor
	scvar_t	*s;	// sides

	int cnum;
	int	fnum;
	int snum;

	int maxcnum;
	int maxfnum;
	int maxsnum;

} scsm_t;

extern scsm_t scsm;

typedef struct
{
	fixed_t	frac, x, y, z;
	RDBOOL	set;
	// v220-8
	short	floorpic;
	short	ceilpic;

} plane3d_t;

extern plane3d_t plane3d;

#define	AF_COD		0x0001

// Used with ge_inf.skybox
#define GE_SKYBOXUSE			0x00000001
#define GE_SKYBOXFORCE			0x00000002
#define GE_NOSKYBOXMESSAGE		0x00000004
#define GE_SKYBOXCHECK			0x00000003

// Used with ge_inf.flags
#define GEF_WATERTRANS		0x00000001
#define GEF_DYNLIGHTS		0x00000002
#define GEF_SPARE			0x00000004	// not used
#define GEF_FALLINGDAMAGE	0x00000008
#define GEF_UWDAMAGE		0x00000010
#define GEF_JUMP			0x00000020
#define GEF_TORCHLIGHT		0x00000040
#define GEF_VTXLIGHT		0x00000080
#define GEF_SINECOMPRESS	0x00000100
#define GEF_WATERDYNAMICS	0x00000200
#define GEF_FOOTSTEPS		0x00000400
#define GEF_WEAPRSOUND		0x00000800
#define GEF_WEAPSMOKE		0x00001000
#define GEF_SPLASHSPRITE	0x00002000

// One time set (true for whole wad - not remembered)
#define GEF_COMPSTAIRS		0x00010000
#define GEF_COMPMODEL		0x00020000

// Used with ma_inf.noscriptset flags
#define	NSF_SCPLIGHT		0x00000001
#define	NSF_FALLINGDAMAGE	0x00000002
#define	NSF_WATERDYNAMICS	0x00000004
#define	NSF_UWATERDAMAGE	0x00000008
#define	NSF_WATERTRANS		0x00000010


// v220-10 used for footstep type with flats.
//         This is normally decided automatically but
//         is provided where this is not feasible.
#define	MODF_WATER		0x00000001
#define	MODF_LAVA		0x00000002
#define	MODF_SLIME		0x00000003
#define	MODF_BLOOD		0x00000004
#define	MODF_NUKE		0x00000005
#define	MODF_BLKWATER	0x00000006	// v220-16 used with black animated flats

// v220-6 node build types (do not change)
#define	MODF_FACTOR12	0x00000010
#define	MODF_GWAONLY	0x00000020	// Use the normal nodes as shipped
#define	MODF_FACTOR16	0x00000040	// default, internal use only
#define	MODF_FACTOR32	0x00000080
#define	MODF_NOEDIT		0x00000100	// internal use only
#define	MODF_NOPRUNE	0x00000200	// internal use only
// v220-14
#define	MODF_ZDNODG		0x00000400	// internal use only
#define	MODF_ZDNODO		0x00000800	// internal use only
#define	MODF_MODNODES	0x00001000	// v220-25

#define	MODF_NODEMASK	0x00001FF0

// v220-10 used with compat array
#define	MODF_DMSTAIRS	0x00010000
#define	MODF_MSECTORS	0x00020000
#define	MODF_CFO		0x00040000

// v220-4 Used with ma_inf.modifier
#define	MODF_ITE		0x00100000	// v220-8 ignore thing error
#define	MODF_120		0x00200000
#define	MODF_KILL		0x00400000
#define	MODF_FLOORFIX	0x00800000
#define	MODF_DEHAMBIENT	0x01000000
#define	MODF_SCROLL		0x02000000
#define	MODF_NOSKYMASK	0x04000000	// v220-25

// v220-14
#define	MODF_ZDNODONLY	0x80000000	// internal use only

//--------------------------------

// v220-4 flags added
#define TUWF_WATERTEX		0x1
#define TUWF_HASBUBBLES		0x2
#define TUWF_SOUNDPLAY		0x4
#define TUWF_RESET			0x8

typedef struct {
	int		tics;
	int		breath_tics;
	int		airsupply;
	int		tuwflags;

} uwater_t;

extern uwater_t thinker_uwater;

// DEH stuff
enum
{
	StartHealth,
	StartBullets,
	MaxHealth,
	MaxArmor,
	GreenAC,
	BlueAC,
	MaxSoulsphere,
	SoulsphereHealth,
	MegasphereHealth,
	GodHealth,
	FAArmor,
	FAAC,
	KFAArmor,
	KFAAC,
	BFGCells,
	Infight,
	NUMMISCDEHVALS
};

extern uint miscDEHval[NUMMISCDEHVALS];
extern uint flagsDEH[DH_NUMSPRITES];

#define c_stairs		 0
#define c_torches		 1
#define c_camera		 2
#define c_nuke24		 3
#define c_model			 4
#define c_resurrect		 5
#define c_dtex			 6
#define c_moddebug		 7
#define c_nextmap		 8		// v220-5
#define c_nocodep		 9		// v220-13
#define c_nosound		10
#define c_nomusic		11
#define c_DEHamb		12
#define c_carry			13		// b220-1
#define c_skulllimit	14		// v220-3
#define c_fake			15		// v220-35
#define NUM_COMPAT_OPTIONS	16

extern short compat[NUM_COMPAT_OPTIONS];

#define ITC_MODLOAD		0
#define ITC_UNPACK		1
#define ITC_ANALYSE		2
#define ITC_SECREVERB	3
#define ITC_CACHE1		4
#define ITC_CACHE2		5
#define ITC_NODESBUILD	6
#define ITC_CACHENUM	7

extern float timecache[ITC_CACHENUM];

void Init_SlopeArrays(void);
void SetProgHeader(char *message);
void ResetProgHeader(void);
void PrintProgInfo(char *message, ...);	// Help messages
void PrintProgError(char *message, ...);

// dp_mobj.c
RDBOOL	P_SpawnScriptThing(spawnspot_t *sspot, int *monum, int tXXnum, RDBOOL chkspawn, RDBOOL fromscp);
void	DP_SetAmodelbounds(void);

// thinker id's
void	P_SetMobjID(thid_t id, RDBOOL set);
void	P_SetMpos(mobj_t *mo);
RDBOOL	P_IsUsedMobjID(thid_t id);
void	ResetArrays(RDBOOL initblockrings);
mobj_t	*NewReferencedMobj(thid_t idnum);
mobj_t	*SetReferencedMobj(thid_t idnum);

// line triggered sound
void	S_SetLineTriggeredAmbient(int num, int from);
void	S_StopAmbient(mobj_t *mo, int num);
void	S_StartAmbient(int num);

// Flats to be destroyed when exiting a 3D line map
void	GL_Delete3Dflats(void);

// DEH routine
RDBOOL	Defs_DEH(RDBOOL check_exists);
void	ChangeDEHThingSPR(char *name, int num);
int		CheckDEHThingSPR(char *name);

// r3d_scriptReader.c
char	*SR_NextSTR(RDBOOL notend);
int		SR_NumForString(RDBOOL *scperror);
float	SR_FloatForString(RDBOOL *scperror);
float	SR_NextFloat(void);
int		SR_NextNum(void);
int		SR_NextBOOL(void);
void	SR_KeepString(void);
void	SR_EndScript(void);
void	SR_SCPErr(char *message, ...);
void	SR_SCPWarn(char *message, ...);
void	SR_LoadLump(int lump, char *idname);
void	SR_RewindScript(void);
RDBOOL	loadDATAdirect(void *data, uint size, char *name);
RDBOOL	SR_LoadFile(char *fname, char *idname, int *abort_on_err);
RDBOOL	SR_LoadPWADLumpName(char *lumpname);
RDBOOL	SR_LoadPWADSCPLump(int lump);

// v220-4
RDBOOL	SR_GetInfoFromScript(char *sname, char *keyname, char *keyword, RDBOOL t1, int l1n, int *l1v, RDBOOL t2, int l2n, int *l2v);

void	FreeSSLines(void);

// Also called from EngineShutdown()
void	R3D_MAINfreeArrays(void);
void	R3D_Set_sslinesArray(void);
char	*R3D_Map_Name(void);

void	LS_PrintGeneralInfo(int preanalysed);

// Lump type verification
int		DoomLumpIdentify(int lump, int *width, int *height, int type);

// hi res custom text checking
RDBOOL	chkForExternalCustomTex(char *name, RDBOOL isflat);

// Shutdown routines
void	EngineShutdown(void);

// r3d_actionScript
void	Init_ActionScript(char *mapid);
void	AS_ResetMobjInfo(void);
RDBOOL	AS_CheckSCP_DEFN(char *mapid);
int		FetchDropItem(int thingidx);
void	AS_SetFCO_Limit(void);

// r3d_ThingSpecials
	// linked flame models
void	P_MobjLinkFlame (mobj_t* mobj);
void	TS_ResetWater(void);	// v 220-3
void	TS_CheckLinkedModels(void);
void	TS_CheckLinkSpecial(mobj_t *sp);
void	TS_CheckLinkKey(card_t card);

	// listed sounds
void TS_StartListSounds(void);

	// scripted models entry point
void	TS_SetMessage(th7900_t *t79, char *msg, RDBOOL print);
void	TS_CheckMessageQueue(void);
RDBOOL	TS_Activate(mobj_t* thing, line_t *ld, int type);
void	TS_setHoldState(line_t *ld);

	// thinkers
void	T_SFXNoWay(void);
void	SetSFXnoway(mobj_t *usething, int tics);
void	T_MAPMusic(void);
void	SetMAPMusic(mobj_t *mapthing);
void	T_Underwater(void);
void	T_ModelKill(modelkill_t *mk);
void	T_ModelSpawn(modelspawn_t *ms);
void	T_ModelReset(modelreset_t *mr);
void	T_SwitchReset(modelreset_t *mr);
void	T_SoundActivate(scriptedsound_t *ss);
void	T_DelayUnblock(delayline_t *dl);
void	T_DelayResurrect(modelreset_t *mr);
void	T_StateMonitor(modelreset_t *mr);
void	T_Endlevel(delayline_t *dl);
void	T_LineOcclude(lineocclude_t *lo);
void	T_AdjustCeil(adjustplat_t *ac);
void	T_PlayListsound(delayline_t *dl);
void	T_KillDelayed(delayline_t *md);
void	T_ActivateDelayed(modelreset_t *mr);	// v220-31

// mobj functions
RDBOOL	spotIsOccupied(fixed_t x, fixed_t y, fixed_t z, fixed_t r, int *bbox, int flags);
int		P_GetTypeFromDoomednum(int type);	// b211-0 Added

// hudkey functions
void	GL_ResetHudsprites(void);
void	GL_ClearAllHudIgnoreFlags(void);
void	GL_ClearIgnoreFlag(int index);
RDBOOL	GL_SetIgnoreFlag(int val);

// switch functios
void checkULTEX(line_t *ld, int test);

extern int scripted_hudscale;
extern int customPunch;

// Misc
void	ED_MovePlayer(int x, int y);
void	P_SetTarget(mobj_t **mop, mobj_t *targ);
void	S_LevelMusic(void);
void	P_SetMapID(char *id, int ep, int gm);
void	G_GetFriendStats(void);
void	G_DeductFriendStats(void);
int		G_GetFriendType(int n);
char	*P_GetCurrentMapid(void);

// Switches
extern int		advancedemo;
extern int		autoload;
extern int		automap_active;
extern int		automapHelp;
extern int		drawmtfog;
extern int		flying;			// R3D
extern int		musChange;		// R3D
extern int		menu_active;
extern int		nomonsters;		// checkparm of -nomonsters
extern int		paused;	
extern int		respawnmonsters;
extern int		usergame;
extern int		verbose;
extern int		waterdynamics;	// R3D
extern int		CFX;			// R3D
extern int		BDMFX;			// R3D v220-35
extern RDBOOL	con_freezegame;		// v220-28

// Cvars
// Switches that can be set from the console
extern int		r_texglow;
extern int		r_bloodglow;
extern int		bloodglowcurrent_state;	// v220-33 Used to remember state of r_bloodglow
extern int		useDynLights;
extern int		modelLights;
extern int		fastparm;		// checkparm of -fast
extern int		r_use_particles;
extern int		anisotropic;
extern float	maxAniso;

// open notepad inhibit
extern int		opennotepad;

// For use in p_saveg.c
extern char		*mainPwadname;

// For use in setThingAngles()
extern char		*mainIwadname;

// Used to check a pwad actually has the map being called for
extern int		haspwads;

// Used for DEH compatibility checks
#define	DEHF_THING		  1
#define	DEHF_SPRITE		  2
#define	DEHF_STATE		  4
#define	DEHF_THINGSET	  8
#define	DEHF_STATESET	 16

#define	DEHF_USEMODEL	128	// Used with or without a dehacked entry


extern RDBOOL	hasDEH;
extern RDBOOL	hasDEHactors;
extern byte		DEHCheck [DH_CHECKNUM];
extern ushort	DEHoldTH [DH_CHECKNUM];

// 3D scaled textures
extern int		anim_group_count;
extern int		newtexcount;
extern int		numstandardtextures;

// R3D UDIR
extern char		*udirGame;

// misc
extern RDBOOL	autoAim;		// b211-0
extern RDBOOL	underfakefloor, underwater, abovefakeceil;

extern int		waterlevel;
extern int		onsurface;
extern RDBOOL	water_trans;	// v220-3
extern int		water_opacity;

extern char		LOGoutfile[MAXPATH];	// Path\Risen3D.out
extern char		screenfile[MAXPATH];	// bin\screen_file

extern int		appStatus;
extern int		appTerminate;
extern int		curfilter;		// The current filter (0 = none).

// v220-4
extern RDBOOL	open_panel_request;
extern int		plr_viewgroup;
extern RDBOOL	dbfile_active;


// Hires texture/patch support v220-16 -----------

RDBOOL	IW_CheckIWADTexDimensions(char *name, short width, short height);
char	*IW_HiresTexFile(RDBOOL flat, char *name);
char	*IW_HiresPatchFile(char *name);
void	IW_InitIWADTextures(void);
void	IW_SetHiresTexPaths(void);

extern int		hiresPatches;
extern int		hiresTextures;

// Zooming stuff -----------
extern float	setFOV;
extern float	fieldOfView;
extern float	targetFOV;
extern int		inZoomMode;
extern int		zoomweapontype;

void	ZoomReset(int resetting);

// v220-6 all messages are now printed directly to Risen3D.out
// All instances of printf replaced with either of the
// following;
void	pr_OFinsert(char *message, ...);
void	pr_OFdirect(char *msg);
void	pr_OFrepeat(char *msg, char *rch, int num, int lfe);

//---------

// Map names
extern char		*analysename;
extern int		console_flat;
extern RDBOOL	update_gamma;

// Transient DEH mapname comparisons ->
#include "dedfile.h"

extern RDBOOL	R3D_lightextns;
extern RDBOOL	R3D_slopetypes;
extern RDBOOL	R3D_linemap;
extern RDBOOL	R3D_ScriptedMap;
extern RDBOOL	R3D_Secret_Pickups;
extern RDBOOL	R3D_scpzoff;
extern RDBOOL	shortListQuads;
extern RDBOOL	R3D_maptype;
extern RDBOOL	R3D_amodeldoors;
extern int		loadModels;
extern int		scriptedMapError;
extern int		has_detail;
extern int		r_detail;

extern RDBOOL	gl_inited;

extern RDBOOL	fi_active;

extern int		titlepic_num;
extern int		interpic_num;
extern int		pagepic_num;
extern int		commercial_game;
extern int		gametic;
extern int		r_ambient;

// b211-0
extern int		ss_validcount;
extern uint		maxsecsz;
extern byte		*vtoflags;
extern ushort	*vtocount;
extern int		maplimits[4];
extern int		maparea;
extern RDBOOL	whitefog;

// v220-6
extern uint GLVertStart;

// v220-3 fogwall vars
typedef struct {
	RDBOOL	isset;
	RDBOOL	outsidefog;
	RDBOOL	underwaterfog;
	float	fogstart;
	float	fogend;
	float	rgba[4];
	float	density;
	uint	mode;
} fogvalues_t;

extern fogvalues_t	fogval;
extern RDBOOL		outsidefog;
extern RDBOOL		underwaterfog;
extern ushort		*linefog_group;

extern int		yesno_exit;

// b220-6 forced sector light change via anifile
typedef struct {
	int *entries;	// 2 per 1) sector 2) lightlevel
	int maxentries;
	int numentries;
} aniseclight_t;

extern aniseclight_t anisl;

// v220-10
extern RDBOOL	reduceNearClip;

// v220-31 active dynlight vars

// Things (MT_STROBEx) angles
#define STROBE_E	0x00000000
#define STROBE_NE	0x20000000
#define STROBE_N	0x40000000
#define STROBE_NW	0x60000000
#define STROBE_W	0x80000000
#define STROBE_SW	0xA0000000
#define STROBE_S	0xC0000000
#define STROBE_SE	0xE0000000

#define	MAXMOBJLIST		16	// dynlight thinker mobj list size

typedef struct
{
	thinker_t	thinker;
	mobj_t*		list[MAXMOBJLIST];
	int			listcount;
	short		lightlevel;
	int		minlight;
	int		maxlight;
	int		direction;
	int		secn;

} dynglow_t;

typedef struct
{
	thinker_t	thinker;
	mobj_t*		list[MAXMOBJLIST];
	int			listcount;
	short		lightlevel;
	int		count;
	int		maxlight;
	int		minlight;
	int		secn;

} dynfireflicker_t;

typedef struct
{
	thinker_t	thinker;
	mobj_t	*mo;
	int		count;
	int		darktime;
	int		brighttime;

} dynstrobe_t;

typedef struct {
	int		val;
	int		dir;
} dynphase_t;

typedef struct {
	mobj_t	**list;
	int		count;
	int		maxcount;
	int		timer;
	int		period;
} secmo_t;

extern secmo_t *secmo;

typedef struct {
	sector_t	**seclist;
	int			count;
	int			maxcount;
} trig5505_t;

extern trig5505_t trig5505;

// for savegame support
typedef struct {
	int		timer;
} secmodummy_t;

extern secmodummy_t	*secdmo;

void	P_CreateDynLightStructures(RDBOOL regenerate);
void	P_CheckInheritance5505(void);
void	P_UpdatePhasedLights(void);
void	P_UpdateDynLights(void);
void	P_SetupDynlightList(void);
void	T_DynStrobeFlash (dynstrobe_t* flash);
void	T_DynLightFlash (dynstrobe_t* flash);
void	T_DynFlicker (dynstrobe_t* flash);
void	T_DynGlow(dynglow_t* dg);
void	T_DynFireFlicker(dynfireflicker_t* dflick);

// End active dynlight vars

extern float	maxLightDistance;
// v220-33 for hacks to help with arctic issues
extern RDBOOL	isarctic;
// v220-35 revised dynlight user selection control
extern int		forceDynLights;

#endif