//-----------------------------------------------------------------------------
//
// R3D_UTILS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
// Revision 2.1.1 build 0 07/Dec/2006 gmj (b211-0)
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//-----------------------------------------------------------------------------

#ifndef __R3D_UTILS_H__
#define __R3D_UTILS_H__

// flags used with LS_SameSecLineCondition
#define	NO_LINECROSSING	 0x01
#define	NO_SINGLEVTX1	 0x02
#define	NO_UNCONNLINE	 0x04
#define	NO_COINLINE		 0x08

// flags used with LS_LinesInSSECarray (int)
#define LSS_NONE		0x0001	// unique flag - use alone
#define LSS_HASTEX		0x0002	// unique flag - use alone
#define LSS_HASULTEX	0x0004	// unique flag - use alone
#define LSS_EXCSAMESD	0x0008
#define LSS_DIFFIDEN	0x0010
#define LSS_EXCSAMEHT	0x0020
#define LSS_DIFFSEC		0x0040
#define LSS_SSECSET		0x0080

#define	MAX_LISTCOUNT	16	// v220-16


// Completely separate test in two passes.
// Pass 1 can be used with any of the above
// Pass 2 is overridden only by LSS_NONE
#define LSS_SKYTEST		0x0080

// b211-0
typedef struct
{
	int	hitex;
	int	midtex;
	int	lotex;
	int	irhitex;
	int	irlotex;
	int	irowntex;
	int misshi;
	int misslo;
	int boundhaslo;
	int boundhasmid;
	int boundhashi;
	int boundshastex;
	int boundsmisshi;
	int boundsmisslo;
	int missingspec;
	int specceil;
	int specfloor;
	int	lines;
	int	anylines;
	int othersec;
	int	sameseclines;
	int	anysameseclines;
	int anysamemovef;
	int fmovfloor;

	int ownringcnt;

	int otherloringcnt;

	int	othersslinechk;
	int	othersslinecnt;
	int	othersslinetex;

} ssecinfo_t;

extern	ssecinfo_t	ssinf;

// v220-16 owner sectors
typedef struct {
	sector_t	*inner;
	sector_t	*outer;
	int			scount;
	int			lcount;
	int			*intList;
	int			listcnt;
	int			innerbox[4];
	int			outerbox[4];
	short		tag;
} osx_t;

extern	osx_t	osx;

extern int *ssinf_ownrcptr;
extern int *ssinf_otherlorcptr;
extern int *ssinf_otherlineptr;


typedef struct {
	int   dsdiffsec;	// double sided linecount - diff sectors each side
	int   dssamesec;	// double sided linecount - same sectors each side but not the calling sector's
	int   dssametype;	// as dssamesec but with different but identical sectors each side 
	int   sameunset;	// as dssamesec but with an unsec specSec
	int   dscompsec;	// as dssamesec but same as calling sector
	int   sslinecnt;	// single sided linecount
	int   movfsec;		// movfloor set for same sided line type
	int   movcsec;		// movceil  set for same sided line type
	short diff_f;		// used in subfn_setSSEClib for diff floor heights
	short diff_c;		// currently unused
} libox_t;

// One time initialisation routines
void		LS_SetSectorsClosedFlag(void);
void		LS_SetTexAndSamesecFlags(void);
void		LS_InitVtxarray(void);
void		LS_InitLXarray(void);			// b211-0
void		LS_InitVtflags(void);
void		LS_CheckUnmergedVtxs(void);		// b211-0
void		LS_SetMultiLoopFlag(void);		// b211-0
void		LS_SetLinesclosedFlag(void);
void		LS_SetDualMovSecFlag(void);
void		LS_SetFakeMoveFlag(void);

// Public analyser array values setting routines
RDBOOL		addSec(int secidx, int *count);
RDBOOL		findSec(int secidx, int count);
void		SET_countsec(int num);
void		SET_linearray(int num);
void		CLR_hitline(void);
void		CLR_excludeline(void);
void		CLR_vtxcheck(void);
void		CLR_conarray(void);
void		CLR_ssecarray(void);

// Setting routines
void		AddToMovsecList(int secnum);
int			getNextTag(void);
void		SetboxFloat(int *ibox, float *fbox);
void		Reset_SEC(int secnum);
void		Reset_SSEC(int ssecnum, int num);
void		Clear_SSEC(subsector_t *ss);

// Checking routines for animated textures/flats
RDBOOL		IsLiquidFloorFlat(sector_t *sec);
RDBOOL		IsLiquidCeilFlat(sector_t *sec);
RDBOOL		IsAnimTex(int picnum);
RDBOOL		IsAnimFlat(int picnum);

// Checking routines for array line vertexes, vtxowners or vtxflags
RDBOOL		Vtfl(line_t *ld);	// b211-0
RDBOOL		Vtfl1(line_t *ld);
RDBOOL		Vtfl3(line_t *ld);
RDBOOL		Vtxs1(line_t *ld);
RDBOOL		Vtxs2(line_t *ld);
RDBOOL		LSS(line_t *ld);
RDBOOL		LineHasMovSectors(line_t *ld);
RDBOOL		HitssecMidpointOutsideBox(int *bbox);
RDBOOL		idenLine(line_t *ld);
RDBOOL		LS_ChkAllVtx2(sector_t *sec, RDBOOL ctrl, RDBOOL mislotex);
RDBOOL		LineVertexConMovOrLiqSec(line_t *ld);
RDBOOL		LineVertexOwnsSec(line_t *ld, int secnum);
RDBOOL		secEquiv(sector_t *sec, line_t *ld, RDBOOL *diffsec);
RDBOOL		LineSecEquiv(line_t *ld);
RDBOOL		SetAndCheckSSECSInLoop(sector_t *sec, sector_t *os);
RDBOOL		SecCtrl(int secnum);
RDBOOL		CmpLptr(line_t *ld, int num);
RDBOOL		Spans1sLines(line_t *ld);
RDBOOL		ChkLptr(line_t *ld);
RDBOOL		chkSkyFloor(line_t *ld);
RDBOOL		ssUnset1(sector_t *sc);
RDBOOL		ssUnset2(subsector_t *ss);
RDBOOL		SSECunset(subsector_t *ss);
RDBOOL		ChkSSECunset(subsector_t *ss);
RDBOOL		SecUnset(sector_t *sec);

RDBOOL		SecACTN(int secnum);
RDBOOL		SecONE(int secnum);
RDBOOL		SecMIDC(int secnum);

// sector specFlags checking
int			MSFC(int secnum, int flags);

// Point on line checking routines
int			FindIfThingIsOnLine(mobj_t *th, RDBOOL diffsec);
int			FindIfThingIsOn3DLine(mobj_t *th);
RDBOOL		thingFitsInSector(mobj_t *iter, sector_t *sc);
RDBOOL		PointIsOnLine(fixed_t x, fixed_t y, line_t *l, int r);
RDBOOL		XYonLine(fixed_t x, fixed_t y, line_t *l, fixed_t r);
RDBOOL		VertexIsOnLine(vertex_t *v1, line_t *ld);

// General routines
int			subfn_setSSEClib(libox_t *lib, sector_t *sec, int *secstore);
void		LS_SectorsFromLinesInBox(int *box, int *secptr, int *ptr);
int			LS_LinesInBox(sector_t *sec, byte *mrkline, int *linelist, int *bbox, libox_t *lib, RDBOOL vtxinside);
void		LS_SubsecsInBox(int *bbox);
RDBOOL		LS_MissingHiTex(subsector_t *ss);
RDBOOL		LS_OnlySSLinesInSSECarray(sector_t *sec, byte *mrklinearray);
RDBOOL		LS_NoHiTexInSSECarray(void);
RDBOOL		LS_LinesInside(byte *mrklinearray, sector_t *s);
RDBOOL		LS_LinesInsideLoop(sector_t *sec, int loopcount);
RDBOOL		LS_SsecAllSegsFrontSame(sector_t *s);
RDBOOL		LS_SSEC_FloorSet(void);
RDBOOL		LS_LinesInSSECarray(byte *mrklinearray, int flags, int *count);
RDBOOL		LS_HasOsecSSLinesInSSECarray(sector_t *sec, sector_t *osec, byte *mrklinearray, line_t **ld, int *ldcount);
int			LS_MarkLinesInSSECarray(sector_t *sec, byte *mrklinearray, byte *exclude);
void		LS_ExcludeSubsecsInside(sector_t *sec, byte *sseclist);
RDBOOL		LS_SetSSECSinLoop(sector_t *s, int *secnum, int loopcnt, int *looplist, int flags);

RDBOOL		LS_CheckRingInSec(byte *markline, line_t *ld, sector_t *s);
RDBOOL		LS_CheckFillCeilArea(sector_t *sec, byte *hitline, int ring_count, int callnum);
RDBOOL		subfn_CheckRingFromArray(byte *mrklinarray, int lnum);
RDBOOL		subfn_CheckRingFromList(int *list, int elements, int *box);
RDBOOL		subfn_CheckForRing(byte *mrklinarray, sector_t *sec, int lnum, int *ring_count, int *bbox, int *linnb, int *seccheck);
int			subfn_LineRing(sector_t *sc, int count, int *box);
RDBOOL		subfn_CheckForRingInList(byte *mrklinarray, int *loopptr, int *looplist, sector_t *sec, int lnum, int *ring_count, int *bbox, int *nsec);
int			LS_ChkLineLoops(sector_t *sec, RDBOOL vtx2only, RDBOOL misslotex, RDBOOL clrhitline);

float		Get_LineLength(line_t *li);
sector_t	*GetLineSecFacingLine(line_t *ld, line_t *ref);
int			GetLptr(line_t *ld);
RDBOOL		ssecAllSameSec(sector_t *sec, int lnum);
byte		Chk0304Type(line_t *ld);
int			ChkLFGSecSet(line_t *ld);
void		subfn_setSSEC_LFG(int num, RDBOOL clear_mtflags);
RDBOOL		subfn_EquivalentSec(sector_t *a, sector_t *b, RDBOOL comptex);	// v220-10
RDBOOL		chk2LineSSEC(subsector_t *ss, sector_t *sec);	// v220-7
int			findContainingSec(sector_t *sec);

RDBOOL		LQSSLines(sector_t *s, int *ovrsec);
RDBOOL		LS_ChkMissing(line_t *ld, int type);
RDBOOL		LS_SameSecLineCondition(sector_t *sec, int condition);
RDBOOL		LS_AllSameSecLine(sector_t *sec);
RDBOOL		subfn_SecClosed(sector_t *sec, byte type);
sector_t	*UT_FindLowestCeilingSector(sector_t* sec);

sector_t	*getLoopSecFromBox(line_t *ld, int *bbox);
int			subfn_CheckLoopClosed(int lptr, int loopcnt, int *looplist, int *bbox, byte *mrkline);
int			subfn_CheckConnectedLineSlope(line_t *ld, line_t *li);
int			subfn_SegFacingMidpoint(float v1[2], float v2[2], float mx, float my);
int			subfn_SsecFacingLine(subsector_t *ss, line_t *ld);
void		subfn_resetSector(int secnum, byte type);
fixed_t		meanV(__int64 a, __int64 b);
RDBOOL		twoSKY(line_t *ld, RDBOOL hi);
int			LS_TestLineTex(line_t *ld);
RDBOOL		checkMissingTexLine(line_t *ld);
RDBOOL		invalidTEX_movfType(line_t *ld);
void		LS_SetSECDONE(line_t *ld, int num);
void		LS_SetLines(sector_t *sec, subsector_t *ss, RDBOOL noR3D_16, RDBOOL forceR3D_16, RDBOOL checkTypeR3D_16, int num);

// b211-0 Functions added
RDBOOL		LS_SetSSINF(sector_t *sec, byte *array, byte *exssecs, byte *bridgelines, RDBOOL exline, RDBOOL invalidtex);
int			dualRingNestedSector(sector_t *sec, int cnt, int *ringcount);
int			largestRingNestedSector(sector_t *sec, int cnt, int *ring, int *count, int *box);	// v220-8
int			*getSSECSinDualRingSector(sector_t *sec, int *sscnt, int test, RDBOOL set_outer, RDBOOL return_all);

// v220-16 added
RDBOOL		U_checkDoubleLoop(sector_t *sc, int type);

#endif

