//-----------------------------------------------------------------------------
//
// R_MODEL.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __R_MODEL_H__
#define __R_MODEL_H__

#include "dedfile.h"
#include "gl_model.h"

// Model frame flags (match with df_ flags in deds)
#define MFF_FULLBRIGHT			0x00000001	//										df_fullbright
#define MFF_SHADOW1				0x00000002	//										df_shadow1
#define MFF_SHADOW2				0x00000004	//										df_shadow2
#define MFF_BRIGHTSHADOW		0x00000008	//										df_brightshadow
#define MFF_MOVEMENT_PITCH		0x00000010	// Pitch aligned to movement.			df_movpitch
#define MFF_SPIN				0x00000020	// Spin base speed * 4 (default)		df_spin			1 rotation every 6 seconds
#define MFF_SKINTRANS			0x00000040	// Color translation -> skins.			df_skintrans
#define MFF_AUTOSCALE			0x00000080	// Scale to match sprite height.		df_autoscale
#define MFF_MOVEMENT_YAW		0x00000100	//										df_movyaw
#define MFF_DONT_INTERPOLATE	0x00000200	// Don't interpolate from the frame.	df_nointerpol
#define MFF_BRIGHTSHADOW2		0x00000400	//										df_brightshadow2
#define MFF_ALIGN_YAW			0x00000800	//										df_alignyaw
#define MFF_ALIGN_PITCH			0x00001000	//										df_alignpitch
#define MFF_DARKSHADOW			0x00002000	//										df_darkshadow
#define MFF_IDSKIN				0x00004000	// Mobj id -> skin in skin range		df_idskin
#define MFF_DISABLE_Z_WRITE		0x00008000	//										df_noz
#define MFF_NO_DISTANCE_CHECK	0x00010000	//										df_nomaxz
#define MFF_SELSKIN				0x00020000	//										df_selskin
#define MFF_PARTICLE_SUB1		0x00040000	// Sub1 center is particle origin.		df_ptsuboff1
#define MFF_NO_PARTICLES		0x00080000	// No particles for this object.		df_noptc
#define MFF_SHINY_SPECULAR		0x00100000	// Shiny skin rendered as additive.		df_specular
#define MFF_SHINY_LIT			0x00200000	// Shiny skin is not fullbright.		df_litshiny
#define MFF_IDFRAME				0x00400000	// Mobj id -> frame in frame range		df_idframe
#define MFF_ATTACKSKIN			0x00800000	// v220-4 use skin + 4					df_attackskin

// v220-5 spin speed flags added
// v220-7 spin flags can be combined. E.g. df_spin and df_doublespin will give an increase of 4 * 8
#define MFF_QUARTERSPIN			0x01000000	// Spin base speed		.				df_quarterspin	1 rotation every 24 seconds
#define MFF_HALFSPIN			0x02000000	// Spin base speed * 2			.		df_halfspin		1 rotation every 12 seconds
#define MFF_DOUBLESPIN			0x04000000	// Spin base speed * 8	.				df_doublespin	1 rotation every 3 seconds
#define MODEL_SPINMASK			0x07000020

// v220-9
#define MFF_NOCULL				0x10000000	// Turn off cull
#define MFF_ALPHANORMAL			0x20000000	// Use normal alpha blending

//------------------------------------------------

// v220-9
#define	RF_ALPHA		0x0001
#define	RF_ALPHAFN		0x0002
#define	RF_NOBLEND		0x0004
#define	RF_EQUAL		0x0008
#define	RF_NODEPTH		0x0010
#define	RF_NOCULL		0x0020
#define	RF_SPECULAR		0x0040
#define	RF_DARK			0x0080
#define RF_BFNSET		(RF_SPECULAR|RF_DARK)

typedef struct {
	float		vertex[3];
	float		shinytexcoord[2];
	float		color[3];
	float		normal[3];
} mvertex_t;

typedef struct
{
	short		model;
	short		frame;
	char		framerange;
	char		skinrange;
	byte		align;
	byte		alpha;
	int			flags;
	short		skin;
	float		offset[3];
	short		shinyskin;	// Skinname ID (index)
} submodeldef_t;

typedef struct modeldef_s
{
	int			order;

	state_t		*state;		// Pointer to the states list (in dd_defns.c).
	int			flags;
	int			select;
	short		skintics;
	float		intermark;	// [0,1) When is this frame in effect?
	float		interrange[2];
	float		offset[3];
	float		resize, scale[3];
	float		ptcoffset[3];
	float		visualradius;

	// R3D added; zheight, zoff, xyradius
	// v220-5 zheight removed - replaced with zmin and zmax
	fixed_t		zmin, zmax, zoff, xyradius;

	ded_model_t	*def;

	// Points to next inter-frame, or NULL.
	struct modeldef_s *internext;
	
	// Points to next selector, or NULL (only for "base" modeldefs).
	struct modeldef_s *selectnext;

	// Submodels.
	submodeldef_t	sub[MAX_FRAME_MODELS];
} modeldef_t;


extern modeldef_t	*modelDefs;
extern model_t		*modellist[MAX_MODELS];
extern float		rModelAspectMod;

void		R_InitModels(void);
void		R_AddModelPath(char *addPath, RDBOOL append);
void		R_SetSpriteReplacement(int sprite, char *modelname);
void		R_SetModelTorchCentre(mobj_t *mo);
int			R_FindGraphicFile(char *szinfilename, char *szoutfilename);
float		R_CheckModelFor(mobj_t *mo, modeldef_t **mdef, modeldef_t **nextmdef);
RDBOOL		R_CheckModelExists(mobj_t *mo);
RDBOOL		R_ModelHasSkins(modeldef_t *modef);
dmd_frame_t *R_GetVisibleFrame(modeldef_t *mf, int subnumber, int mobjid, int from);
modeldef_t	*GetStateModel(state_t *st, int select);


#endif 