//-----------------------------------------------------------------------------
//
// R_THINGS.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __R_THINGS_H__
#define __R_THINGS_H__

#include "p_data.h"
#include "r_data.h"

// Types.
#ifndef DGLubyte
typedef unsigned char	DGLubyte;
#endif
#ifndef DGLunit
typedef unsigned int	DGLuint;
#endif

#define REND_PSPRITE	0x0001
#define REND_PMODEL		0x0002

// Used with visflags
#define VIS_NONE		0x00000000
#define VIS_SPRITE		0x00000001
#define VIS_MODEL		0x00000002
#define VIS_WALL		0x00000004
#define VIS_TRANS		0x00000008
#define VIS_FOGWALL		0x00000010
#define VIS_NOFOG		0x00000020
#define VIS_PSPR		0x00000040
#define VIS_ATKLIGHT	0x00000080
#define VIS_LITSKIN		0x00000100	// v220-4	Use lit skin
#define VIS_PARTICLE	0x00000200	// v220-4
#define VIS_SPLASH		0x00000400	// v220-4
#define VIS_NOGLIGHT	0x00000800	// v220-4	suppress global light
#define VIS_FLAME		0x00001000	// v220-4	suppress dyn-light on torch flames
#define VIS_RENDERED	0x00008000	// v220-4
#define VIS_NOALPHACUT	0x00010000	// v220-9	certain textures look better if blended
#define VIS_SHADOW		0x00020000	// v220-9	set if thing type is MT_SHADOW (used with sprite only)
#define VIS_STEALTH		0x00040000	// v220-9	set if thing has MF3_STEALTH set
#define VIS_TRANSSPR	0x00080000	// v220-9	set if a sprite is to be rendered translucent
#define VIS_NOTLIT		0x00100000	// v220-13	set if a dehacked sprite should not be lit or a sprite/model with setunlit in R3D_WADanalyse
#define VIS_SHOOTABLE	0x00200000	// v220-24	set if an active enemy

// hasglow flags
#define HAS_CEILGLOW		0x0001
#define HAS_FLOORGLOW		0x0002
#define HAS_OFFSET			0x0004
#define HAS_NOFOG			0x0008	// $v68
#define HAS_FOGUNSET		0x0010	// v220-3
#define HAS_NOLIGHT			0x0020	// b220-0
#define HAS_UNDERWATER		0x0040	// v220-3 $translucent_water
#define HAS_WATERPLANE		0x0080	// v220-3 $translucent_water
#define HAS_NOZ				0x0100	// v220-4 suppress model depth rendering
#define HAS_RADIUS8			0x0200	// v220-33 flag model with S_GIBS state and small radius
#define HAS_EVENLIGHT		0x0400	// v220-35 2D vector light to model
#define HAS_3DLIGHT			0x1000	// v220-35 3D vector light to model
#define HAS_2DLIGHT			0x2000	// v220-35 2D vector light to model
#define HAS_DEBUG			0x8000	// v220-5 can be used for debugging

typedef struct vissprite_mobj_s {
	int		splnum;
	fixed_t	gx, gy;             // for line side calculation
	fixed_t	gz, gzt;            // global bottom / top for silhouette clipping
	RDBOOL	flip;				// Flip texture?
	float	v1[2], v2[2];		// The vertices (v1 is the left one).
	float	top;				// b211-0 pre-calculated sprite top
	int		flags;				// for color translation and shadow draw
	uint	think_id;
	int		selector;
	int		pclass;				// player class (used in translation)
	RDBOOL	viewaligned;		// Align to view plane.
	float	secfloor, secceil;
	short	vmoflags;			// flags
	// b211-0 added to deal with z-flickering for sprites with same x,y co-ords 
	int		height;				// Used when sorting to project highest sprite behind shortest

	// b211-0 align to int (only first 3 bytes used)
	byte	floorglow[4];		// Floor glow color.
	byte	ceilglow[4];		// Ceiling glow color.

	fixed_t intensity;			// b220-0 vary intensity of dynlight on mobj

	short	lightlevel;
	float	alpha;
	float	visoff[3];			// Last-minute offset to coords.

	struct	modeldef_s	*mf;
	struct	modeldef_s	*nextmf;

	float	yaw, pitch, zangle;	// For model rotation
	float	rotangle;			// For scripted model rotation
	float	inter;				// Frame interpolation, 0..1
#ifdef _DEBUG
	int		thing_num;			// For debugging
#endif
//	struct lumobj_s *light;		// For the halo (NULL if no halo)

} vissprite_mobj_t;

typedef struct {
	float	pos[2];				// x and y coordinates.
	byte	color[4];
} wall_vertex_t;

typedef struct vissprite_wall_s {
	int		texture;
	float	top;				// Top and bottom height.
	float	bottom;
	wall_vertex_t vertices[2];
	float texc[2][2];			// u and v coordinates.
} vissprite_wall_t;

// A vissprite_t is a thing or masked wall that will be drawn during 
// a refresh.
typedef struct vissprite_s
{
	// These must be first two defs
	int			visflags;		// Used to identify what the vis type is
	float		distance;		// Used vissprite sorting
	//------------------------

	short		torchtic;		// For desynchronising torches
	short		skintic;
	int			grp;			// v220-3 pointer to fogwall group

	vissprite_mobj_t mo;
	vissprite_wall_t wall;
} vissprite_t;


// Sprites are patches with a special naming convention so they can be
// recognized by R_InitSprites.  The sprite and frame specified by a
// thing_t is range checked at run time.

// a sprite is a patch_t that is assumed to represent a three dimensional
// object and may have multiple rotations pre drawn.  Horizontal flipping
// is used to save space. Some sprites will only have one picture used
// for all views.

// Used with sprflags
#define SPR_NOMODEL			0x01
#define SPR_LINKMODEL		0x02
#define SPR_ALIGNED			0x04	// v220-3
#define SPR_HIRES			0x10	// v220-16


typedef struct
{
	RDBOOL	rotate;         // if false use 0 for any position
	int		lump[8];        // sprite lump to use for view angles 0-7
	byte	flip[8];        // flip (1 = flip) to use for view angles 0-7
} spriteframe_t;

typedef struct {
	int             numframes;
	spriteframe_t   *spriteframes;
	short			sprflags;		// flags
	fixed_t			trueradius;	// v220-9
} spritedef_t;

typedef struct {
	DGLuint			tex;
	int				lump;			// Real lump number.
	short			width;
	short			height;
	short			slxoffset;		// renamed to aid search
	short			slyoffset;		// renamed to aid search
	short			alpha;
	float			flarex;			// Offset to flare.
	float			flarey;
	float			lumsize;
	float			tc[2];			// Texture coordinates.
	ushort			hudspr;			// Set to point to a hires hudsprite
	rgbcol_t		color;			// Average color, for lighting.

} spritelump_t;


extern  spritedef_t		*sprites;
extern  int             numsprites;
extern	spritelump_t	*spritelumps;
extern	int				numspritelumps;
extern  vissprite_t     *vissprites;
extern  vissprite_t     **vispointer;
extern	uint			visspritecnt;
extern	uint			maxvissprites;
extern	vissprite_t		vispsprites[DDMAXPSPRITES];

void	R_GetSpriteInfo(int sprite, int frame, spriteinfo_t *sprinfo, int from);
int		R_MaxRadius(mobj_t *mo, RDBOOL shadow, int from);
void	R_ProjectPlayerSprites(int *psp3d);
void	R_SortVisSprites(void);
void	R_AddSprites(subsector_t *ss, fixed_t fakefloor, short llevel); // v220-31 changed
void    R_AddPSprites(void);
void    R_InitSprites(void);
void	R_InitZDist(void);
int		GetSpriteMaxHeight(uint sprite);	// b211-0 changed
fixed_t getSpriteRadius(mobj_t *mo);		// v220-5



#endif 