//-----------------------------------------------------------------------------
//
// S_CHAN.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// First written for v220-4 - gmj
//
//-----------------------------------------------------------------------------

#ifndef __S_CHAN_H__
#define __S_CHAN_H__

#define	SFX_MAX_CHANNELS	 128

// Buffer properties.
#define SFXBP_VOLUME			0	// 0..1
#define SFXBP_FREQUENCY			1	// 1 = normal
#define SFXBP_PAN				2	// -1..1 (2D only)
#define SFXBP_MIN_DISTANCE		3	// 3D only
#define SFXBP_MAX_DISTANCE		4			
#define SFXBP_POSITION			5				
#define SFXBP_VELOCITY			6
#define SFXBP_NORMAL_MODE		7
#define SFXBP_RELATIVE_MODE		8

// Sfx Buffer Flags.
#define SFXBF_3D		0x00000001	// References a 3D EAX buffer.
#define SFXBF_PLAYING	0x00000002	// The buffer is playing.
#define SFXBF_REPEAT	0x00000004	// Buffer will repeat until stopped.
#define SFXBF_DONT_STOP	0x00000008	// Never stop until normal finish.
#define SFXBF_SUSPEND	0x00000010	// Identify sound as suspended
#define SFXBF_RELOAD	0x00000020	// Sample must be reloaded before playing.
#define SFXBF_NOATTN	0x00000040	// No attenuation, indicates a global map sound

// Sfx channel flags.
#define SFXCF_NO_ORIGIN		0x01	// Sound is coming from a mystical emitter.
#define SFXCF_NO_UPDATE		0x02	// Channel update is skipped
#define SFXCF_AMBIENT		0x04	// Ambient sound

// Listener properties.
#define SFXLP_UPDATE			0	// Not a real value (commit deferred)
#define SFXLP_UNITS_PER_METER	1
#define SFXLP_DOPPLER			2
#define SFXLP_POSITION			3
#define SFXLP_VELOCITY			4
#define SFXLP_ORIENTATION		5
#define SFXLP_REVERB			6
#define SFXLP_GAIN				7

typedef struct sfxsample_s
{
	void	*sdata;			// Actual sample data.
	uint	size;			// Size in bytes.
	int		id;				// Id number of the sound sample.
	int		numsamples;		// Number of samples.
	int		bytesper;		// Bytes per sample (1 or 2).
	int		rate;			// Samples per second.
	int		group;			// Exclusion group (0, if none).

} sfxsample_t;

typedef struct sfxbuffer_s
{
	sfxsample_t *sample;	// Source sample data.
	int		bytes;			// Bytes per sample (1 or 2).
	int		rate;			// Samples per second.
	uint	freq;			// Played samples per second (real freq).
	int		SFXBFflags;

	// D3D vars
	void	*ptr;			// Pointer to driver's own buffer object.
	void	*ptr3d;			// Pointer to driver's 3D buffer data.
	uint	d3dsize;		// Size of the buffer (bytes). Only used by D3D
	uint	cursor;			// Write cursor position (%length).
	uint	written;		// Total bytes written.

} sfxbuffer_t;

typedef struct sfxchannel_s
{
	sfxbuffer_t buffer;
	mobj_t	*emitter;		// Mobj that is emitting the sound.
	float	realpos[3];		// Set during channel update.
	float	nowpos[3];		// Where a sound source currently is.
	float	volume;			// Sound volume: 1.0 is max.
	float	frequency;		// Frequency adjustment: 1.0 is normal.
	float	attenuation;
	uint	chstarttime;	// Channel start time
	int		priority;
	int		SFXCFflags;

} sfxchannel_t;

extern sfxchannel_t	snd_channels[SFX_MAX_CHANNELS];

typedef struct sidriver_s
{
	int		(*SoundCreateBuffer)(sfxchannel_t *ch, int flags, int bytes, int rate, uint size);
	void	(*SoundDestroyBuffer)(sfxchannel_t *ch);
	RDBOOL	(*SoundRefresh)(sfxchannel_t *ch);
	RDBOOL	(*SoundLoad)(sfxchannel_t *ch, struct sfxsample_s *sample);
	void	(*SoundPlay)(sfxchannel_t *ch);
	void	(*SoundReset)(sfxchannel_t *ch);
	void	(*SoundStop)(sfxchannel_t *ch);
	void	(*SoundRestart)(sfxchannel_t *ch);
	void	(*SoundSet)(sfxchannel_t *ch, int property, float value);
	void	(*SoundSetv)(sfxchannel_t *ch, int property, float *values);
	void	(*SoundListener)(int property, float value);
	void	(*SoundListenerv)(int property, float *values);
	void	(*SoundShutdown)(void);

} sidriver_t;

#endif

