//-----------------------------------------------------------------------------
//
// SYS_INPUT.H - Header file for Risen3D [R3D] v220
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
// NOTE: Risen3D varies substantially from any ports on which it may
//       be based [see the c file(s) to which this applies] and any
//       same named header files are not commutable.
//
//-----------------------------------------------------------------------------

#ifndef __SYS_INPUT_H__
#define __SYS_INPUT_H__

// Key event types.
#define IKE_NONE		0
#define IKE_KEY_DOWN	0x1
#define IKE_KEY_UP		0x2

// Mouse buttons.
#define NUM_MOUSE_BUTTONS	4	// we're only interested in the first four

/*
#define IMB_LEFT		0x1
#define IMB_RIGHT		0x2
#define IMB_MIDDLE		0x4
#define IMB_FOUR		0x8
#define IMB_FIVE		0x10
#define IMB_SIX			0x20
#define IMB_SEVEN		0x40
#define IMB_EIGHT		0x80
*/

// Joystick.
#define IJOY_AXISMIN	-10000
#define IJOY_AXISMAX	10000
#define IJOY_MAXBUTTONS	32
#define IJOY_POV_CENTER	-1

typedef struct
{
	char event;			// Type of the event.		
	unsigned char code;	// The scancode (extended, corresponds DI keys).
} keyevent_t;

typedef struct
{
	int x, y, z;		// Relative X and Y mickeys since last call.
	int buttons;		// The buttons bitfield.
} mousestate_t;

typedef struct
{
	int		axis[3]; 
	int		rotAxis[3];
	int		slider[2];
	char	buttons[IJOY_MAXBUTTONS];
	float	povAngle;	// 0 - 359 degrees.
} joystate_t;

extern RDBOOL	joyInited;
extern RDBOOL	mouseInited;

int		I_Init(void);
void	I_Shutdown(void);
int		I_GetKeyEvents(keyevent_t *evbuf);
void	I_GetMouseState(mousestate_t *state, RDBOOL flush);
void	I_GetJoystickState(joystate_t *state);

#endif