//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 id Software, Inc., G Jackson, Jaakko Kernen 
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: d_items.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for DEH support, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code
//
//
// DESCRIPTION:  Doom console
//
//-----------------------------------------------------------------------------

// We are referring to sprite numbers.
#include <stdio.h>
#include <string.h>

#include "doom/info.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "r3d_compat.h"
#include "r3d_edit.h"

//
// PSPRITE ACTIONS for weapons.
// This struct controls the weapon animations.
//
// Each entry is:
//   ammo/amunition type
//  upstate
//  downstate
// readystate
// atkstate, i.e. attack/fire/hit frame
// flashstate, muzzle flash
//

//
// These are used if other definitions are not found.
//

weaponinfo_t	weaponinfo[NUMWEAPONS] =
{
    {
	// fist
	am_noammo,
	0,
	S_PUNCHUP,
	S_PUNCHDOWN,
	S_PUNCH,
	S_PUNCH1,
	S_NULL
    },	
    {
	// pistol
	am_clip,
	1,
	S_PISTOLUP,
	S_PISTOLDOWN,
	S_PISTOL,
	S_PISTOL1,
	S_PISTOLFLASH
    },	
    {
	// shotgun
	am_shell,
	1,
	S_SGUNUP,
	S_SGUNDOWN,
	S_SGUN,
	S_SGUN1,
	S_SGUNFLASH1
    },
    {
	// chaingun
	am_clip,
	1,
	S_CHAINUP,
	S_CHAINDOWN,
	S_CHAIN,
	S_CHAIN1,
	S_CHAINFLASH1
    },
    {
	// missile launcher
	am_misl,
	1,
	S_MISSILEUP,
	S_MISSILEDOWN,
	S_MISSILE,
	S_MISSILE1,
	S_MISSILEFLASH1
    },
    {
	// plasma rifle
	am_cell,
	1,
	S_PLASMAUP,
	S_PLASMADOWN,
	S_PLASMA,
	S_PLASMA1,
	S_PLASMAFLASH1
    },
    {
	// bfg 9000
	am_cell,
	40,
	S_BFGUP,
	S_BFGDOWN,
	S_BFG,
	S_BFG1,
	S_BFGFLASH1
    },
    {
	// chainsaw
	am_noammo,
	0,
	S_SAWUP,
	S_SAWDOWN,
	S_SAW,
	S_SAW1,
	S_NULL
    },
    {
	// super shotgun
	am_shell,
	2,
	S_DSGUNUP,
	S_DSGUNDOWN,
	S_DSGUN,
	S_DSGUN1,
	S_DSGUNFLASH1
    },	
};

int GetDefInt(char *def, int *returned_value)
{
	char *data;
	int val;

	// Get the value.
	if(!Def_Get(DD_DEF_VALUE, def, &data)) return 0; // No such value...
	// Convert to integer.
	val = strtol(data, 0, 0);
	if(returned_value) *returned_value = val;
	return val;
}

void GetDefState(char *def, int *val)
{
	char *data;

	// Get the value.
	if(!Def_Get(DD_DEF_VALUE, def, &data)) return;
	// Get the state number.
	*val = Def_GetStateNum(data);
	if(*val < 0) *val = 0;
}

// DEH values that override defs.
static void setDEHoverrideValues(mbf_player_t *p)
{
	// 1 - items
	if (miscDEHval[MaxHealth       ] != 0x80000000)
		healthlimit = miscDEHval[MaxHealth];
	if (miscDEHval[MaxArmor        ] != 0x80000000)
		armorpoints[1] = miscDEHval[MaxArmor];
	if (miscDEHval[BFGCells        ] != 0x80000000)
		weaponinfo[wp_bfg].pershot = miscDEHval[BFGCells];

	// 2 - player values
	if (miscDEHval[StartHealth     ] != 0x80000000)
		plhealth.health = miscDEHval[StartHealth];
	if (R3D_edit) return;	// Don't mess with ammo in edit mode
	if (miscDEHval[StartBullets    ] != 0x80000000)
		p->ammo[am_clip] = miscDEHval[StartBullets]; 
}

// Initialize weapon info, maxammo and clipammo.
// Set DEH override values
// This is only called once on startup
void P_InitWeaponAndDEHInfo()
{
#define PLMAX "Player|Max ammo|"
#define PLCLP "Player|Clip ammo|"
#define WPINF "Weapon Info|%i|"

	int i, k;
	char buf[80];
	char *data;
	char *ammotypes[] = { "clip", "shell", "cell", 
		"misl", "-", "noammo", 0 };

	// DEH values that override hard coded values
	// 0x80000000 indicates not set allowing normal values to prevail
	if (miscDEHval[MegasphereHealth] == 0x80000000)
		miscDEHval[MegasphereHealth] = 200;
	if (miscDEHval[MaxSoulsphere   ] == 0x80000000)
		miscDEHval[MaxSoulsphere   ] = 200;
	if (miscDEHval[SoulsphereHealth] == 0x80000000)
		miscDEHval[SoulsphereHealth] = 100;
	if (miscDEHval[GodHealth       ] == 0x80000000)
		miscDEHval[GodHealth       ] = 100;
	if (miscDEHval[KFAArmor        ] == 0x80000000)
		miscDEHval[KFAArmor        ] = MAXHEALTH;
/*	if (miscDEHval[Infight         ] == 0x80000000)
		miscDEHval[Infight] = GetDefInt("AI|Infight", NULL) ? 0xDD : 0;*/

	// Max ammo.
	GetDefInt(PLMAX"Clip" , &maxammo[am_clip ]);
	GetDefInt(PLMAX"Shell", &maxammo[am_shell]);
	GetDefInt(PLMAX"Cell" , &maxammo[am_cell ]);
	GetDefInt(PLMAX"Misl" , &maxammo[am_misl ]);

	// Clip ammo.
	GetDefInt(PLCLP"Clip" , &clipammo[am_clip ]);
	GetDefInt(PLCLP"Shell", &clipammo[am_shell]);
	GetDefInt(PLCLP"Cell" , &clipammo[am_cell ]);
	GetDefInt(PLCLP"Misl" , &clipammo[am_misl ]);

	// Weapon pershot and states
	for(i=0; i<NUMWEAPONS; i++)
	{
		sprintf(buf, WPINF"Type", i);

		if(Def_Get(DD_DEF_VALUE, buf, &data))
		{
			// Set the right type of ammo.
			for(k=0; ammotypes[k]; k++)
			{
				if(!_stricmp(data, ammotypes[k]))
				{
					weaponinfo[i].ammo = (ammotype_t)k;
					break;
				}
			}
		}

		sprintf(buf, WPINF"Per shot", i);
		GetDefInt(buf, &weaponinfo[i].pershot);
		sprintf(buf, WPINF"Up", i);
		GetDefState(buf, &weaponinfo[i].upstate);
		sprintf(buf, WPINF"Down", i);
		GetDefState(buf, &weaponinfo[i].downstate);
		sprintf(buf, WPINF"Ready", i);
		GetDefState(buf, &weaponinfo[i].readystate);
		sprintf(buf, WPINF"Atk", i);
		GetDefState(buf, &weaponinfo[i].atkstate);
		sprintf(buf, WPINF"Flash", i);
		GetDefState(buf, &weaponinfo[i].flashstate);
	}
}

void P_InitPlayerValues(mbf_player_t *p)
{
#define PLINA "Player|Init ammo|"
	int i;
	char buf[20];

	if (p != mbf_players) return;

	if (miscDEHval[StartHealth] != 0x80000000)
	{
		plhealth.health = miscDEHval[StartHealth];

		// Check something daft hasn't been set
		if (plhealth.health <   1) plhealth.health =   1;
		if (plhealth.health > 200) plhealth.health = 200;
	}
	else
		GetDefInt("Player|Health", &plhealth.health);

	// GetDefInt("Player|Weapon", &p->readyweapon);

	for(i=0; i<NUMWEAPONS; i++)
	{
		sprintf(buf, "Weapon Info|%i|Owned", i);
		GetDefInt(buf, &p->weaponowned[i]);
	}
	GetDefInt(PLINA"Clip" , &p->ammo[am_clip ]);
	GetDefInt(PLINA"Shell", &p->ammo[am_shell]);
	GetDefInt(PLINA"Cell" , &p->ammo[am_cell ]);
	GetDefInt(PLINA"Misl" , &p->ammo[am_misl ]);

	setDEHoverrideValues(p);

	p->readyweapon = p->ammo[am_clip] ? wp_pistol : wp_fist;
	p->pendingweapon = wp_pistol;
}

//=============================================================================
// P_Init
//=============================================================================
void P_Init (void)
{
	// Maximum health and armor points.

	maxhealth      = 100;
	healthlimit    = 200;
	armorpoints[0] = 100;
	armorpoints[1] = 200;

	GetDefInt("Player|Max Health",   &maxhealth);
	GetDefInt("Player|Health Limit", &healthlimit);
	GetDefInt("Player|Green Armor",  &armorpoints[0]);
	GetDefInt("Player|Blue Armor",   &armorpoints[1]);

	// Initialize weapon info using definitions.
	// Set any DEHACKED values
	// Must be called after setting the above - R3D
	P_InitWeaponAndDEHInfo();
}
