//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: d_refresh.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  moved to main engine.  Rewritten for R3D.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8 JDoom.dll code
//
//
// DESCRIPTION: mobj flag setting, drawing filters, titles
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "doom/p_local.h"
#include "doom/hu_stuff.h"
#include "doom/Mn_def.h"
#include "doom/M_menu.h"
#include "doom/hu_stuff.h"
#include "doom/st_stuff.h"
#include "doom/wi_stuff.h"
#include "doom/x_hair.h"
#include "doom/am_map.h"
#include "doom/wi_stuff.h"
#include "doom/f_infine.h"

#include "dd_main.h"
#include "dd_wad.h"
#include "gl_tex.h"
#include "gl_draw.h"
#include "gl_font.h"
#include "gl_main.h"
#include "net_main.h"
#include "r_main.h"
#include "r_data.h"
#include "r_things.h"
#include "con_functions.h"
#include "con_console.h"
#include "ui_main.h"

#include "r3d_compat.h"
#include "r3d_mapinfo.h"
#include "sys_system.h"


// OGL
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void R_LevelPatch(dpatch_t *dpmap, int lump);
void FR_TextOutXOffCentre(char *msg, int yoff);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

void R_DrawSpecialFilter(void);
void R_DrawLevelTitle(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern float	lookOffset;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

byte	*translationtables;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char		*str_completed = "[ COMPLETED ]";

// CODE --------------------------------------------------------------------

//===========================================================================
// D_Display
//		Draw current display.
//===========================================================================
void D_Display (void)
{
	mbf_player_t	*player = &mbf_players[0];
//	RDBOOL			iscam = (player->plr->flags & DDPF_CAMERA) != 0;
	
	// Do buffered drawing.
	switch (gamestate)
    {
	case GS_LEVEL:
		if (ppt.thinkers_have_run)
		{
			if (!useDynLights && forceDynLights)
			{
				useDynLights = true;
				if (!Sys_CriticalMessage("ADVISORY: This map requires dynamic lighting\n"
					"but this has not been enabled in the Launcher.\n\n"
					"Since this cannot be forced off in the Control\n"
					"Panel it has now been automatically enabled.\n\n", false))
				{
					Sys_Quit(false);
				}
				ma_inf.forcedyn = true;
			}

			// v220-35 inform user that a request to set a skybox has been overridden.
			if (ge_inf.use_sky_box & GE_NOSKYBOXMESSAGE)
			{
				ge_inf.use_sky_box &= ~GE_NOSKYBOXMESSAGE;
				P_SetMessage(ppt.viewplrmo->player, "CUSTOM SKY - SKYBOX REQUEST IGNORED");
			}

			// $democam: can be set on every frame
			if (R3D_edit || cfg.sbarscale != 5 /*|| iscam*/)
			{
				// Full screen.
				viewSCRheight = 200;
			}
			else
			{
				viewSCRheight = 200 - ST_HEIGHT;
			}

			// Draw the player view.
			viewangleoffset = (int)(ANGMAX * -lookOffset);
			curfilter = player->plr->filter;
			// How about fullbright?

			LevelFullBright = ((player->powers[pw_infrared] > 4*32)
				|| (player->powers[pw_infrared] & 8) 
				||  player->powers[pw_invulnerability] > 30);
		
			// Render the view
			R_RenderPlayerView(player->plr, player->cheats & CF_NOCLIP);

			// v220-22 If automap background is opaque then don't draw the filter
			if (automap_active && (cfg.automapAlpha < 0.f || cfg.automapAlpha == 1.f) )
			{
				fakefilter = false;
			}
			else
			{
				R_DrawSpecialFilter();
			}

			if (automap_active)
				AM_Drawer();
			else if (/*!iscam &&*/ cfg.xhairType)
				X_Drawer(cfg.xhairColor); // Crosshair

			if (!ma_inf.noleveltitle)
			{
				// Level title is shown briefly when the level starts 
				R_DrawLevelTitle();
			}

			if (/*!iscam &&*/ !R3D_edit) ST_Drawer(); // $democam

			HU_Drawer();
		}

		break;

	case GS_INTERMISSION:
		WI_Drawer();
		break;

	case GS_INFINE:
		FI_Drawer();
		break;

	case GS_DEMOSCREEN:
		if (pagepic_num >= 0) GL_DrawPagePatch(pagepic_num);
		break;

	case GS_WAITING:
		glClear(GL_COLOR_BUFFER_BIT);
		break;
	}

    // draw pause pic
	if (paused)
	{
		GL_DrawPatch(126, 4, W_GetNumForName("M_PAUSE"));
	}
}

//===========================================================================
// R_DrawSpecialFilter
//===========================================================================
void R_DrawSpecialFilter(void)
{
	int t;

	if ( (t = mbf_players[0].powers[pw_invulnerability]) )
	{
		float max = 30;
		float str, r, g, b;

		if(t < max)
			str = t/max;
		else if(t < 4*32 && !(t&8))
			str = .7f;
		else if(t > INVULNTICS-max)
			str = (INVULNTICS-t)/max;
		else
			str = 1; // Full inversion.

		r = str + str;
		g = r - .4f;
		b = r - .8f;
		if (r < 0.0f) r = 0.0f; else if (r > 1.0f) r = 1.0f;
		if (g < 0.0f) g = 0.0f; else if (g > 1.0f) g = 1.0f;
		if (b < 0.0f) b = 0.0f; else if (b > 1.0f) b = 1.0f;

		// Draw an inversing filter. 
		glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE_MINUS_SRC_COLOR);
		GL_DrawScreenFilter(r, g, b, 1);

		// Restore the normal rendering state.
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}

//===========================================================================
// R_CacheLevelTitle
//  b211-0 Determine what to show and cache titlepatch if necessary .
//===========================================================================

static	dpatch_t dr_mapname;
static	int		 title_offx;

// Note regarding the logic used: an iwad titlepatch is never drawn.
// instead the map name is drawn. This is possible as all iwad patches
// comprise just the map name.  Conversely pwad titlepatches are always
// drawn, but only at the map start never in the intermission.
int R_CacheLevelTitle(void)
{
	if (!cfg.levelTitle || !mapinfo)
		return -1;

	if (   mapinfo->titlepatch > 0
		&& !(mapinfo->R3Dflags & R3Dmapnameset)
		// If the level name has been changed but no pwad titlepatch exists
		// then displaying the iwad titlepatch will undoubtedly be wrong so
		// the new level name will be printed instead.
		&& !(mapinfo->R3Dflags & R3Dtpicfromiwad && !(mapinfo->R3Dflags & R3Dhasmapinfo)) )
	{
		if (   dr_mapname.lump != mapinfo->titlepatch
			&& lumptexinfo[dr_mapname.lump].ltitex)
		{
 			GL_DeleteTexIndex(&lumptexinfo[dr_mapname.lump].ltitex, "lumptexinfo 2");
		}
 
		if (!lumptexinfo[dr_mapname.lump].ltitex)
		{
			R_LevelPatch(&dr_mapname, mapinfo->titlepatch);

			title_offx = dr_mapname.xoffset + (SCREENWIDTH - dr_mapname.width)/2;
		}

		return 1;
	}
	return 0;
}

//===========================================================================
// R_DrawLevelTitle
// Show map name and UDIR folder (if a dedicated game type).
//===========================================================================

void R_DrawLevelTitle(void)
{
	int		res;
	float	alpha = 1;

	//-------------------------------

	if (   actual_leveltime > 3*35
		|| (res = R_CacheLevelTitle()) < 0)
	{
		ma_inf.noleveltitle = true;
		return;
	}

	if (actual_leveltime > 2*35)
		alpha = 1 - (actual_leveltime - 2*35)/35.0f;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	if (res == 1)
	{
		R_CacheLevelTitle();
		GL_DrawMapnamePatch(title_offx, 20, mapinfo->titlepatch, alpha);
	}
	else
	{
		extern int typein_time;

		char	*lname = mapinfo->mapname;
		int		old = typein_time;

		typein_time = 0xffffff;

		// Make the text a bit smaller.
		glTranslatef(160.0f, 12.0f, 0);
		glScalef(.85f, .85f, 1.0f);
		glTranslatef(-160.0f, -12.0f, 0);
		glColor4f(1.0f, 1.0f, 1.0f, alpha);

		M_WriteText2(160 - M_StringWidth(lname, hu_font_b)/2, 12, lname,
			hu_font_b, -1, -1, -1);

		typein_time = old;
	}
	
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//===========================================================================
// R_DrawMapTitle
//===========================================================================

void R_DrawMapTitle(char *lname, char *msg)
{
	char	n[96];
	char	*ptr;
	float	scale = 0.85f;
	int		y = 12;
	uint	len;

	if (!lname || !(*lname) ) return;	// nothing to print

	// remove leading spaces
	ptr = lname;
	while (isspace(*ptr)) ptr++;

	strncpy(n, ptr, 95);
	n[95] = 0;

	if ( !(len = strlen(n)) ) return;

	// remove following spaces
	ptr = n + len;
	while (ptr > n && isspace(*(--ptr))) *ptr = 0;

	if ( !(len = strlen(n)) ) return;

	n[63] = 0;	// 64 chars max

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	if (!R3D_ScriptedMap && (len = strlen(n)) > 30)
	{
		scale = 1.0f - (len - 20) * .0204f;
	}

	// Make the text a bit smaller.
	glTranslatef(160, (float)y, 0);
	glScalef(scale, scale, 1.0f);
	glTranslatef(-160, (float)(-y), 0);

	glColor4f ( cfg.menuColor[0], 
				cfg.menuColor[1],
				cfg.menuColor[2], 1.0f);

	M_WriteText2(160 - M_StringWidth(msg, hu_font_b)/2, y, msg,
		hu_font_b, -1, -1, -1);

	y += 16;

	M_WriteText2(160 - M_StringWidth(n, hu_font_b)/2, y, n,
		hu_font_b, -1, -1, -1);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

static char	*info_str[] = {
	"AGGREGATED STATS", 
	"PLAYER KILLS",
	"COLLATERAL KILLS",
	"MASSACRED",
	"ENEMIES LEFT",
	"FRIENDS LEFT",
	"ITEMS UNUSED",
	"SECRETS FOUND",
	"TIME",
	NULL
};

// b211-0 first written
// v220-3 matrix loading changed
// v220-4 modified for changed scpMessage.info
void R_DrawScriptedMapInf(int lh, float alpha, RDBOOL completed)
{
	char	*msg;
	int		i, y = 12, d = 98;

	if (*scpMessage.info[0]) d += y;
	if (*scpMessage.info[2]) d += y;
	if (*scpMessage.info[3]) d += y;
	if (*scpMessage.info[5]) d += y;

	// Go into screen projection mode.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 320, 200, 0, -1, 1);
	
	UI_DrawRectFill(40, 2, 238, d, 2, 
		UI_COL(UIC_BRD_RHI), UI_COL(UIC_BRD_RMED), UI_COL(UIC_BRD_RLOW), alpha/1.4f);

	glMatrixMode(GL_MODELVIEW);
	// Make the text a bit smaller.
	glTranslatef(160.0f, (float)y, 0);
	glScalef(.85f, .85f, 1.0f);
	glTranslatef(-160.0f, (float)(-y), 0);

	msg = mapinfo->mapname;

	if (msg[0])
	{
		int tw = M_StringWidth(msg, hu_font_b)/2;

		glColor4f(.882f, .785f, 0.42f, alpha);

		if (tw > 118)
		{
			tw = M_StringWidth(msg, hu_font_a)/2;
			M_WriteText2(160 - tw, y, msg, hu_font_a, -1, -1, -1);
		}
		else
		{
			M_WriteText2(160 - tw, y, msg, hu_font_b, -1, -1, -1);
		}

		y += lh - 4;

		msg = mapinfo->author;

		if (msg[0])
			M_WriteText2(160 - M_StringWidth(msg, hu_font_a)/2, y, msg, hu_font_a, -1, -1, -1);
	}

	y += 12;
	glColor4f(0.8f, 0.8f, 0.8f, alpha);

	if (completed)
	{
		M_WriteText2(160 - M_StringWidth(str_completed, hu_font_a)/2, y, str_completed, hu_font_a, -1, -1, -1);
	}
	else if (wminfo.nextmapid[0])
	{
		char buf[64];

		sprintf(buf, "NEXT: %s", wminfo.nextmapid);
		M_WriteText2(160 - M_StringWidth(buf, hu_font_a)/2, y, buf, hu_font_a, -1, -1, -1);
	}

	y = 50;

	glColor4f(0.8f, 0.4f, 0.4f, alpha);

	for (i = 0; info_str[i]; i++)
	{
		if (!*scpMessage.info[i]) continue;

		M_WriteText2( 50, y, info_str[i], hu_font_a, -1, -1, -1);
		msg = scpMessage.info[i];
		M_WriteText2(220 - M_StringWidth(msg, hu_font_a)/2, y, msg, hu_font_a, -1, -1, -1);
		y += 12;
	}

	// Restore original matrices
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

static int prTTout(int c, int k, char *msg)
{
	SIZE	strsize;
	int		p, sy;

	GetTextExtentPoint32(mainHdc, msg, c, &strsize);
	p = (current_w - strsize.cx)/2;
	pglWindowPos2i(p + aspect_offset_x, k - aspect_offset_y);
	glListBase(ttFontList);
	glCallLists(c, GL_UNSIGNED_BYTE, msg);

	sy = strsize.cy;
						
	return sy;
}

// v220-4 modified to use ttFontList. If unset will fall back to jfont
void DR_SCP_Drawer(void)
{
	char	*ptr;

	if (  !open_panel_request	// Don't start drawing text if the ui_panel's about to open
		&& (ptr = scpMessage.scriptMessage)
		&& *ptr
		&& scpMessage.delay_to_fade
		&& scpMessage.delay_to_next_text <= 0)
	{
		char  msg[300];
		int   len = strlen(ptr);
		int   c, i, k;

		// Go into screen projection mode.
		glMatrixMode(GL_PROJECTION);

		if (ttFontList)
		{
			glDisable(GL_TEXTURE_2D);
			k = 3*current_h/5;
		}
		else
		{
			float sx = (float)current_w / 640.0f;
			float sy = (float)current_h / 400.0f;

			FR_SetFont(1);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0, current_w, current_h, 0, -1, 1);
			glScalef(sx, sy, 1.0f);
			k = -50;
		}

		UI_ColorA(UI_COL(UIC_TEXT), scpMessage.alpha);

		if (len > 298) len = 298;

		// v220-4 support linefeeds. TS_ConPrintMsg related
		for (c = i = 0; i < len; i++)
		{
			if (ptr[i] == '\\' && ptr[i + 1] == 'n')
			{
				if (c)
				{
					msg[c] = 0;

					if (ttFontList)
					{
						k -= prTTout(c, k, msg);
					}
					else
					{
						FR_TextOutXOffCentre(msg, k);
						k += 12;
					}

				}
				c = 0; i++;
			}
			else
			{
				msg[c++] = ptr[i] & 0x7F;	// font range 0 - 127
			}	
		}

		if (c)
		{
			msg[c] = 0;

			if (ttFontList)
			{
				prTTout(c, k, msg);
			}
			else
				FR_TextOutXOffCentre(msg, k);
		}

		// Restore
		if (ttFontList)
			glEnable(GL_TEXTURE_2D);
		else
		{
			glMatrixMode(GL_PROJECTION);
			glPopMatrix();
		}
	}
}

// v220-4
void DR_DrawTextLoading()
{
	char	str[] = {"LOADING..."};
	int		s = M_StringWidth(str, hu_font_b);
	int		t = hu_font_b[0].height;
	float	y = (float)(102 - t/2);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 320, 200, 0, -1, 1);
	
	UI_DrawRectFill((float)(165 - s/2), y, s + 10, t + 4, 2.0f, 
		UI_COL(UIC_BRD_RHI), UI_COL(UIC_BRD_RMED), UI_COL(UIC_BRD_RLOW), 0.4f);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	// Go into screen projection mode.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// Make the text a bit smaller.
	glTranslatef(160.0f, y, 0);
	glScalef(.85f, .85f, 1.0f);
	glTranslatef(-160.0f, -y, 0);

	glColor3f(1.0f, 0.2f, 0.2f);

	M_WriteText2(165 - s/2, 100, str, hu_font_b, -1, -1, -1);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	GL_DoUpdate();
}

// v220-9 for backward savegame compatibility flags are changed here
void alterFlags(mobj_t *mo)
{
	if (mo->type >= MT_STEALTH_BABY && mo->type <= MT_STEALTH_POSSESSED)
	{
		mo->flags3 |= MF3_STEALTH;
		mo->flags2 |= MF2_NOSHADOW;
	}
	else 
		mo->flags3 &= ~MF3_STEALTH;
}

RDBOOL Def_GetScriptedNoInfight(int type);

// v220-13
void P_SetNoInfight(mobj_t *mo)
{
	if (R3D_ScriptedMap)
	{
		if (Def_GetScriptedNoInfight(mo->type))
		{
			if (ma_inf.NOINFIGHTALL)
				mo->ddflags |= DDF_SCPINFIGHT;
			else
				mo->ddflags |= DDMF_NOINFIGHT;
		}
	}
	else if (mo->type <= MT_WOLFSS && !(DEHCheck[mo->type] & DEHF_THING) )
	{
		if (    mo->type == MT_POSSESSED
			||  mo->type == MT_SHOTGUY
			||  mo->type == MT_CHAINGUY
			||  mo->type == MT_WOLFSS)
		{
			if (ge_inf.infightsametype & 1)
				mo->ddflags |= DDMF_NOINFIGHT;
			else
				mo->ddflags &= ~DDMF_NOINFIGHT;	// change back with savegame if no same type infighting
		}
		else if (mo->type == MT_SKULL)
		{
			if (ge_inf.infightsametype & 2)
				mo->ddflags |= DDMF_NOINFIGHT;
			else
				mo->ddflags &= ~DDMF_NOINFIGHT;	// change back with savegame if no lost soul infighting
		}
	}
}

extern byte		aggression_array[1000];

void P_SetDDFlags(mobj_t *mo)
{
	int n;

	if (mo->flags2 & MF2_DONTDRAW)
	{
		mo->ddflags |= DDMF_DONTDRAW;
		mo->flags2 |= MF2_NOSHADOW;
		return;
	}

	P_SetNoInfight(mo);

	if (  (n = mo->info->doomednum) >= 6000
		&& n <= 6099
		&& aggression_array[n - 6000])
	{
		mo->ddflags |= DDF_AGGRESSIVE;
	}

	if (mo->flags2 & MF2_ALWAYSLIT)
	{
		mo->flags2 |= MF2_NOSHADOW;
		mo->ddflags |= DDMF_ALWAYSLIT;
	}

	if (mo->flags3 & MF3_DEHAMBIENT)
		mo->flags2 |= MF2_DONTDRAW;

	// v220-14 Force model spectre or stealth monsters to fade in 1sec when dead
	if (CFX && loadModels && ((mo->flags3 & MF3_STEALTH) || mo->type == MT_SHADOWS))
	{
		spritedef_t	 *sprdef = &sprites[mo->sprite];

		if (!(sprdef->sprflags & SPR_NOMODEL))
			mo->ddflags |= DDF_FORCEFADE;
	}

	if (   mo->type == MT_SHADOWS
		|| mo->flags3 & MF3_STEALTH
		|| mo->flags & (MF_NOSECTOR|MF_NOBLOCKMAP))
	{
		mo->flags2 |= MF2_NOSHADOW;
	}

	if (mo->type >= MT_SPLASH && mo->type <= MT_GREENBLOOD)
	{
		mo->flags2 |= MF2_NOSHADOW;
	}
	else
	if (mo->type >= MT_TALLGREENFLAME && mo->type <= MT_SKULLFLAME)
	{
		mo->flags2 |= MF2_NOSHADOW;
	}
	else
	if (   mo->type == MT_LIGHTSOURCE	// dmednum 5000
		|| mo->type == MT_STROBE1		// dmednum 5500
		|| mo->type == MT_STROBE2		// dmednum 5501
		|| mo->type == MT_STROBE3		// dmednum 5502
		|| mo->type == MT_STROBE4		// dmednum 5503
		|| mo->type == MT_STROBE5		// dmednum 5504
		|| mo->type == MT_STROBEX)		// dmednum 5505
	{
		mo->ddflags |= (DDMF_ALWAYSLIT|DDMF_DONTDRAW);
		mo->flags2 |= (MF2_NOSHADOW|MF2_DONTDRAW);	// v220-14 add dontdraw
		mo->flags3 |= MF3_DYNLIGHT;
	}
	else
	if ( !(mo->flags2 & MF2_FLAGSSET) )
	{
		if (mo->flags & MF_SPAWNCEILING && !(mo->flags3 & MF3_AMODEL))
			mo->flags2 |= MF2_NOFITBOTTOM;
		
		alterFlags(mo);

		// Set MF2_NOTLIT when not sector linked for dynlight purposes
		// See DL_InitForNewFrame
		if (mo->flags & MF_NOSECTOR) mo->flags2 |= MF2_NOTLIT;
		if (mo->flags & MF_SPECIAL ) mo->flags2 |= MF2_CANSLOPE;

		switch(mo->type)
		{
		case MT_TPUNDERFAKE:
		case MT_TPOVERFAKE:
		case MT_PUSH:
		case MT_PULL:
		case MT_TRIGAMBIENT:
		case MT_FAKESOUNDSOURCE:
		case MT_UNKNOWNTHING:
		case MT_SOUNDSOURCE:
		case MT_MUSICSOURCE:
			mo->flags2 |= MF2_DONTDRAW;
			mo->ddflags |= DDMF_DONTDRAW;
			return;

		case  MT_PUFF:
			mo->flags2 |= (MF2_NOSHADOW|MF2_NOTLIT);
			break;

		case  MT_BLOOD:
		case  MT_TRACER:
		case  MT_PLASMA:
		case  MT_ROCKET:
		case  MT_SMOKE:
		case  MT_SPAWNSHOT:
		case  MT_SPAWNFIRE:
		case  MT_FATSHOT:
		case  MT_ARACHPLAZ:
		case  MT_FIRE:
		case  MT_BOSSSPIT:
		case  MT_BOSSTARGET:
			mo->flags2 |= MF2_NOSHADOW;
			break;

			// v220-33 ensure corpsefade does not prevent A_BossDeath from being called
		case MT_BABY:
		case MT_FATSO:
		case MT_BRUISER:
		case MT_SPIDER:
		case MT_CYBORG:
			mo->flags2 |= MF2_NOFADE;	// disable corpsefade
			break;


		default:
			break;
		}
	}

	// $democam: cameramen are invisible
//	if (P_IsCamera(mo)) mo->ddflags |= DDMF_DONTDRAW;
	
	if (mo->flags & MF_BRIGHTSHADOW)
	{
		mo->ddflags |= DDMF_BRIGHTSHADOW;
		mo->flags2 |= MF2_NOSHADOW;
	}
	else if (mo->flags & MF_SHADOW)
	{
		mo->ddflags |= DDMF_SHADOW;
		mo->flags2 |= MF2_NOSHADOW;
	}
	
	if (    mo->flags & MF_FLOAT 
		|| (  mo->flags & MF_VIEWALIGN  && !(mo->flags & MF_MISSILE)) 
		|| (!(mo->flags & MF_VIEWALIGN) &&   mo->flags & MF_MISSILE))
	{
		mo->ddflags |= DDMF_VIEWALIGN;
	}
	
	mo->ddflags |= (mo->flags & DDMF_TRANSLATION);
}
