//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: m_cheat.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, Moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:  Cheat sequence checking
//
//-----------------------------------------------------------------------------


#include "doom/p_local.h"
#include "doom/f_infine.h"
#include "doom/g_game.h"
#include "doom/p_inter.h"
#include "doom/m_cheat.h"
#include "doom/SndIdx.h"
#include "doom/d_englsh.h"
#include "dedfile.h"
#include "dd_wad.h"
#include "s_main.h"
#include "s_mus.h"
#include "net_main.h"
#include "r3d_compat.h"
#include "con_console.h"

#define SCRAMBLE(a) \
((((a)&1)<<7) + (((a)&2)<<5) + ((a)&4) + (((a)&8)<<1) \
 + (((a)&16)>>1) + ((a)&32) + (((a)&64)>>5) + (((a)&128)>>7))

int		flying;

// CHEAT SEQUENCE PACKAGE

// Called in st_stuff module, which handles the input.
// Returns a 1 if the cheat was successful, 0 if failed.
//
int cht_CheckCheat(cheatseq_t* cht, byte key)
{
	static unsigned char	cheat_xlate_table[256];
	static int		firsttime = 1;
    int i;
    int rc = 0;
	
    if (firsttime)
    {
		firsttime = 0;
		for (i=0;i<256;i++) cheat_xlate_table[i] = (byte)SCRAMBLE(i);
    }
	
    if (!cht->p)
		cht->p = cht->sequence; // initialize if first time or reset
	
	// v220-35 reset if key out of range
	if (key > 127) { cht->p = 0; return 0; }
	
    if (*cht->p == 0)
		*(cht->p++) = key;
    else if
		(cheat_xlate_table[(unsigned char)key] == *cht->p) cht->p++;
    else
		cht->p = cht->sequence;
	
    if (*cht->p == 1)
		cht->p++;
    else if (*cht->p == 0xff) // end of sequence character
    {
		cht->p = cht->sequence;
		rc = 1;
    }
	
    return rc;
}

void
cht_GetParam
( cheatseq_t*	cht,
 char*		buffer )
{
	
    unsigned char *p, c;
	
    p = cht->sequence;
    while (*(p++) != 1);
    
    do
    {
		c = *p;
		*(buffer++) = c;
		*(p++) = 0;
    }
    while (c && *p!=0xff );
	
    if (*p==0xff)
		*buffer = 0;
	
}

void cht_GodFunc(mbf_player_t *plyr)
{
	int i;

	if (plyr != mbf_players) return;

	if (!plyr->plr->mo || plyr->plr->mo->health <= 0)
	{
		P_SetMessage(plyr, "Resurrection ain't possible\n");
		return;
	}

	if ( (i = (gameskill == sk_nightmare && !ge_inf.devmap)) )
		plyr->cheats &= ~CF_GODMODE;
	else
		plyr->cheats ^= CF_GODMODE;

	if (plyr->cheats & CF_GODMODE)
	{
		plyr->plr->mo->health = miscDEHval[GodHealth];
		plhealth.health = miscDEHval[GodHealth];
	}

	P_SetMessage(plyr, (plyr->cheats & CF_GODMODE) ? 
		STSTR_DQDON : i ? "NIGHTMARE : God mode... Are you joking?\n" : STSTR_DQDOFF);
}

RDBOOL cht_GiveFunc(mbf_player_t *plyr, RDBOOL weapons, RDBOOL ammo,
				  RDBOOL armor, RDBOOL cards)
{
	static int ishacx = !strnicmp(analysename, "HACX", 4);
	int		i;

	if (!R3D_edit && gameskill == sk_nightmare && !ge_inf.devmap)
	{
		P_SetMessage(plyr, "NIGHTMARE: you get nothing.\n");
		return false;
	}

	if(armor)
	{
		plhealth.armorpoints = 200;
		plhealth.armortype = 2;
	}

	if (weapons)
	{
		for (i=0;i<NUMWEAPONS;i++)
			plyr->weaponowned[i] = true;
	}

	if (ammo)
	{
		for (i=0;i<NUMAMMO;i++)
			plyr->ammo[i] = plyr->maxammo[i];
	}

	if (cards)
	{
		for (i=0;i<NUMCARDS;i++)
		{
			if (i == it_blueskull && (gamemode == shareware || ge_inf.arctlev || ge_inf.batman)) break;	// v220-35
			if (i == it_yellowcard && ishacx) continue;	// v220-35

			plyr->cards[i] = true;
		}
	}
	return true;
}

RDBOOL S_CheatLevelMusic(ded_mapinfo_t *mt);

void cht_MusicFunc(mbf_player_t *plyr, char *buf)
{
	char	id[8];

	if (commercial_game)
		sprintf(id, "MAP%c%c", buf[0], buf[1]);
	else
		sprintf(id, "E%cM%c", buf[0], buf[1]);

	if (!S_CheatLevelMusic(Def_GetMapInfo(id)))
		P_SetMessage(plyr, STSTR_NOMUS);
}

void cht_NoClipFunc(mbf_player_t *plyr)
{
	plyr->cheats ^= CF_NOCLIP;
	P_SetMessage(plyr, (plyr->cheats & CF_NOCLIP)? STSTR_NCON : STSTR_NCOFF);
}

void ForceNoClipStatus()
{
	mbf_player_t *plyr = &mbf_players[0];
	if (!(plyr->cheats & CF_NOCLIP))
		cht_NoClipFunc(plyr);
}

void ForceEditDefaultStatus()
{
	mbf_player_t *plyr = &mbf_players[0];

	ForceNoClipStatus();
	flying = true;
	// Give keys
	cht_GiveFunc(plyr, 0, 0, 0, true);
}

// Modified to work properly (see also CCmdCheatWarp and p_setup.c)
RDBOOL cht_WarpFunc(mbf_player_t *plyr, char *buf)
{
	char	name[8];
	int		epsd, map;

	// v220-31 avoid warping at wrong point
	if (gamestate == GS_LOADING) return false;

	if (commercial_game)
	{
		epsd = 1;
		map = (buf[0] - '0')*10 + buf[1] - '0';
	}
	else
	{
		epsd = buf[0] - '0';
		map = buf[1] - '0';
	}

	if (gamemode == shareware
		&& ((epsd > 1) || (map > 9)))
		return false;

	if (gamemode == commercial)
		sprintf(name, "map%.2i", map);
	else
		sprintf(name, "E%iM%i" , epsd, map);
	
    if (W_CheckNumForName(name) < 0) return false;

	// So be it.
	P_SetMessage(plyr, STSTR_CLEV);
	G_DeferedInitNew(gameskill, epsd, map);
	brief_disabled = -1;
	return true;
}

void cht_PowerUpFunc(mbf_player_t *plyr, int i, char *msg)
{
	if (!plyr->powers[i])
	{
		P_GivePower( plyr, i);
		P_SetMessage(plyr, msg);
	}
	else if (i == pw_allmap)
	{
		extern int mapcheat;
		P_SetMessage(plyr, "You have already got the computer map..");
		mapcheat = 1;
	}
	else
	{
		if (i == pw_strength)
			plyr->powers[i] = 0;	// relies on doom bug's -ve countdown
		else
		{
			plyr->powers[i] = 1;	// prevent -ve countdown
			P_SetMessage(plyr, "Power toggled off");
		}
	}
}

void cht_ChoppersFunc(mbf_player_t *plyr)
{
	plyr->weaponowned[wp_chainsaw] = true;
	plyr->powers[pw_invulnerability] = INVULNTICS;
}

// Also shown in the automap
void cht_PosFunc(mbf_player_t *plyr)
{
	static char	buf[32];

	sprintf(buf, "x=%i y=%i",
		mbf_players[0].plr->mo->x>>16,
		mbf_players[0].plr->mo->y>>16);
	P_SetMessage(plyr, buf);
}


