//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_sound.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
//
// Based on: Doomsday v1.7.8 JDoom.dll code
//
//
// DESCRIPTION: sound generation
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_compat.h"

#include "p_data.h"
#include "dedfile.h"
#include "s_main.h"
#include "s_cache.h"
#include "s_mus.h"
#include "dd_wad.h"

#include "r3d_mapinfo.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

RDBOOL S_CheatLevelMusic(ded_mapinfo_t *mt)
{
	if (!mt) return false;

	if (   mt->music.path[0]
		&& PlayMapinfoFile(mt->music.path))
	{
		return true;
	}

	return CheckAndPlayFile(mt->muslump, "S_CheatLevelMusic");
}

//===========================================================================
// S_LevelMusic
//===========================================================================
void S_LevelMusic()
{
	if (compat[c_nomusic]) return;

	if (gamestate == GS_LEVEL)
	{
		global_looped = !(mapinfo->R3Dflags & R3Dmusiconce);

		if (currentMapMusLumpnum == -1)
			currentMapMusLumpnum = mapinfo->muslump;

		if (ge_inf.UseList & 2)
		{
			int idx = W_GetMapIndex(mapinfo->id);

			if (   musicLIST[idx].mlist_type
				&& PlayMusListExt(idx))
			{
				playStatus = "LIST";
				return;
			}
		}

		if (   mapinfo->music.path[0]
			&& PlayMapinfoFile(mapinfo->music.path))
		{
			return;
		}

		// Final option - this must be called even if
		// currentMapMusLumpnum < 0 in order to call Music_Stop
		if (!CheckAndPlayFile(currentMapMusLumpnum, "S_LevelMusic [1]"))
		{
			if (currentMapMusLumpnum >= 0)
			{
				// v220-9 If there's a problem with the lump returned by
				//        a savegame then default to the mapinfo entry.
				currentMapMusLumpnum = mapinfo->muslump;

				CheckAndPlayFile(currentMapMusLumpnum, "S_LevelMusic [2]");
			}
		}
	}
	else if (gamestate == GS_DEMOSCREEN && currentMapMusLumpnum >= 0)
	{
		global_looped = 0;

		CheckAndPlayFile(currentMapMusLumpnum, "S_LevelMusic [3]");
	}
}

#include "r_main.h"

//===========================================================================
// S_SectorSound
//	Doom-like sector sounds: when a new sound starts, stop any old ones
//	from the same origin.
//===========================================================================
void S_SectorSound(sector_t *sec, int id)
{
	if (   sfx_avail
		&& !(sec->specFlags & SF_IGNORE))
	{
		mobj_t *origin = (mobj_t*)&sec->soundorg;

		origin->z = ppt.viewplrmo->z;	//sec->floorheight; v220-14 $sound

		S_StopAllSounds(origin);	// Stop all sounds from this origin
		S_StartSound(id, origin);
	}
}

//===========================================================================
// S_SectorSoundRepeat
//  Used where a sound is required to play continuously (until stopped)
//  The leveltime&7 test allows the soundorg height to be upated
//  and also allows them to be sent for a demo recording as the repeat
//  flag is not supported with packetised sounds. 
//===========================================================================
RDBOOL S_SectorSoundRepeat(int id, sector_t *sec)
{
	if (sfx_avail && !(leveltime&7) && !(sec->specFlags & SF_IGNORE))
	{
		mobj_t	*origin = (mobj_t*)&sec->soundorg;

		origin->z = ppt.viewplrmo->z;	//sec->floorheight; v220-14 $sound

		S_SoundRepeat(id|HSF_REPEAT, origin, 1.0f, DEF_ATTN);
		return true;
	}
	return false;
}
