//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_switch.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Switches, buttons. Two-state animation. Exits.
//
//-----------------------------------------------------------------------------

#include "R3D_compat.h"

#include "doom/mbf_data.h"
#include "doom/p_spec.h"
#include "doom/g_game.h"

#include "con_console.h"
#include "dd_wad.h"
#include "r_data.h"
#include "r_main.h"
#include "s_main.h"
#include "r3d_sfx.h"

//
// CHANGE THE TEXTURE OF A WALL SWITCH TO ITS OPPOSITE
//
/*switchlist_t alphSwitchList[] =
{
    // Doom shareware episode 1 switches
    {"SW1BRCOM",	"SW2BRCOM",	1},
    {"SW1BRN1",		"SW2BRN1",	1},
    {"SW1BRN2",		"SW2BRN2",	1},
    {"SW1BRNGN",	"SW2BRNGN",	1},
    {"SW1BROWN",	"SW2BROWN",	1},
    {"SW1COMM",		"SW2COMM",	1},
    {"SW1COMP",		"SW2COMP",	1},
    {"SW1DIRT",		"SW2DIRT",	1},
    {"SW1EXIT",		"SW2EXIT",	1},
    {"SW1GRAY",		"SW2GRAY",	1},
    {"SW1GRAY1",	"SW2GRAY1",	1},
    {"SW1METAL",	"SW2METAL",	1},
    {"SW1PIPE",		"SW2PIPE",	1},
    {"SW1SLAD",		"SW2SLAD",	1},
    {"SW1STARG",	"SW2STARG",	1},
    {"SW1STON1",	"SW2STON1",	1},
    {"SW1STON2",	"SW2STON2",	1},
    {"SW1STONE",	"SW2STONE",	1},
    {"SW1STRTN",	"SW2STRTN",	1},
    
    // Doom registered episodes 2&3 switches
    {"SW1BLUE",		"SW2BLUE",	2},
    {"SW1CMT",		"SW2CMT",	2},
    {"SW1GARG",		"SW2GARG",	2},
    {"SW1GSTON",	"SW2GSTON",	2},
    {"SW1HOT",		"SW2HOT",	2},
    {"SW1LION",		"SW2LION",	2},
    {"SW1SATYR",	"SW2SATYR",	2},
    {"SW1SKIN",		"SW2SKIN",	2},
    {"SW1VINE",		"SW2VINE",	2},
    {"SW1WOOD",		"SW2WOOD",	2},
    
    // Doom II switches
    {"SW1PANEL",	"SW2PANEL",	3},
    {"SW1ROCK",		"SW2ROCK",	3},
    {"SW1MET2",		"SW2MET2",	3},
    {"SW1WDMET",	"SW2WDMET",	3},
    {"SW1BRIK",		"SW2BRIK",	3},
    {"SW1MOD1",		"SW2MOD1",	3},
    {"SW1ZIM",		"SW2ZIM",	3},
    {"SW1STON6",	"SW2STON6",	3},
    {"SW1TEK",		"SW2TEK",	3},
    {"SW1MARB",		"SW2MARB",	3},
    {"SW1SKULL",	"SW2SKULL",	3},
	
    {"\0",		"\0",		0}
};

int		switchlist[MAXSWITCHES * 2];
int		numswitches;*/

void P_Add3DSwitch(int texnum1, int texnum2);

int	locked_door;	// v220-7

button_t buttonlist[MAXBUTTONS];

// Replacement scheme for switches
static int numswitches		=  0;
static int switchindex		=  0;
static int oldswitchindex	=  0;
static int oldnumswitches	=  0;
static int max_numswitches	=  0;

static int *switchlist		= NULL;

//=============================================================================
//
// P_InitSwitchList()
//
// Only called at game initialization in order to list the set of switches
// and buttons known to the engine. This enables their texture to change
// when activated, and in the case of buttons, change back after a timeout.
//
// This routine modified to read its data from a predefined lump or
// PWAD lump called SWITCHES rather than a static table in this module to
// allow wad designers to insert or modify switches.
//
// Rewritten by Lee Killough to remove limit 2/8/98
// Further modified for R3D

static void checkSWarray()
{
	if (switchindex + 2 >= max_numswitches)
	{
		max_numswitches += 32;
		switchlist = (int*)W_Realloc(switchlist,
					   sizeof(switchlist_t) * max_numswitches, STATICID, 231);
	}
}

int P_GetSwitchPair(int txn)
{
	int i = 0;

	while (i < numswitches*2)
	{
		if (switchlist[i++] == txn) return switchlist[i];
		if (switchlist[i++] == txn) return switchlist[i - 2];
	}
	return 0;
}

// v220-26 revised. Since there is no guarantee that the default
// switch list references a texture in the TEXTURE1 list any
// textures in the former that are not in the latter no longer
// qualify as a missing texture entry.
void P_InitSwitchList(void)
{
extern int	blacktexnum;
extern int	notexTEXNUM;
extern int	R3D_edit;

 	uint psize;

	switchlist_t *alphSwitchList = (switchlist_t*)M_CacheLumpNum(W_GetNumForName("SWITCHES"), &psize, true);
	int entries = psize/sizeof(switchlist_t);
	int episode = (gamemode == registered || gamemode == retail) ?
					2 : commercial_game ? 3 : 1;
	int i, k, n[2];

	// ---------------------------

	checkSWarray();

	for (i = 0; i < entries; i++)
	{
		if (alphSwitchList[i].episode <= episode)
		{
			if (!alphSwitchList[i].episode)
				break;

			checkSWarray();

			n[0] = R_CheckTextureNumForName(alphSwitchList[i].name1);
			n[1] = R_CheckTextureNumForName(alphSwitchList[i].name2);

			// Fall back in the case of an unspecified map texture
			if (n[0] < 0) n[0] = R3D_edit ? notexTEXNUM : blacktexnum;
			if (n[1] < 0) n[1] = R3D_edit ? notexTEXNUM : blacktexnum;

			switchlist[switchindex++] = n[0];
			switchlist[switchindex++] = n[1];

			for (k = 0; k < 2; k++)
			{
				texArray[n[k]]->anim_group = anim_group_count;
				texArray[n[k]]->txf_flags |= TXF_SWITCH;
			}

			anim_group_count++;
		}
	}
	
	numswitches = switchindex/2;
	switchlist[switchindex] = -1;

	// v220-31 support R3D_SWCH script
	if (haspwads && episode> 1)
	{
		if ( SR_LoadPWADLumpName("R3D_SWCH") )
		{
			char	*nextstr;
			char	tex1[10], tex2[10];

			while ( (nextstr = SR_NextSTR(false)) )
			{
				strncpy(tex1, nextstr, 8); tex1[8] = 0;
				if ( !(nextstr = SR_NextSTR(false)) ) break;
				strncpy(tex2, nextstr, 8); tex2[8] = 0;

				n[0] = R_CheckTextureNumForName(tex1);
				if (n[0] < 0) { SR_SCPWarn("Bad texture %s", tex1); continue; }
				n[1] = R_CheckTextureNumForName(tex2);
				if (n[1] < 0) { SR_SCPWarn("Bad texture %s", tex2); continue; }					

				P_Add3DSwitch(n[0], n[1]);	// same code

				for (k = 0; k < 2; k++)
				{
					texArray[n[k]]->anim_group = anim_group_count;
					texArray[n[k]]->txf_flags |= TXF_SWITCH;
				}

				anim_group_count++;
			}

			SR_EndScript();
		}
	}

	// Remember these for 3D switches
	oldswitchindex = switchindex;
	oldnumswitches = numswitches;

	M_Free(alphSwitchList);
}

// Added for 3D scaled switches
void P_Add3DSwitch(int texnum1, int texnum2)
{
	checkSWarray();

	switchlist[switchindex++] = texnum1;
	switchlist[switchindex++] = texnum2;
	switchlist[switchindex  ] = -1;
	numswitches++;
}

void P_Remove3DSwitches()
{
	if (switchlist)
	{
		numswitches = oldnumswitches;
		switchindex = oldswitchindex;
		switchlist[switchindex] = -1;
	}
}

//
// Start a button counting down till it turns off.
//
void P_StartButton(line_t* line, bwhere_e w, int texture, int time)
{
    int		i;
    
    // See if button is already pressed
    for (i = 0; i < MAXBUTTONS; i++)
	{
		if (buttonlist[i].btimer && buttonlist[i].line == line)
			return;
	}
    
    for (i = 0; i < MAXBUTTONS; i++)
    {
		if (!buttonlist[i].btimer)
		{
			buttonlist[i].line = line;
			buttonlist[i].where = w;
			buttonlist[i].btexture = texture;
			buttonlist[i].btimer = time;
			buttonlist[i].soundorg = &line->frontsector->soundorg;
			return;
		}
    }
    Con_ErrorD("P_StartButton: no button slots left!");
}

//
// b211-0 Check if a line that is going to have its
//        textures deleted has a switch texture.
//        If so it is marked to ensure switch sound is made.

void checkULTEX(line_t *ld, int test)
{
	if (ld->special)
	{
		int hitex = 0, lotex = 0;

		if (test & 1) lotex = sides[ld->sidenum[0]].bottomtexture;
		if (test & 2) hitex = sides[ld->sidenum[0]].toptexture;

		if (hitex || lotex)
		{
			int i;

			for (i = 0; i < numswitches*2; i++)
			{
				if (   (hitex && switchlist[i] == hitex)
					|| (lotex && switchlist[i] == lotex) )
				{
					ld->ovflags |= OVF_SWITCH;
					return;
				}
			}
		}
	}
}

//
// Function that changes wall texture.
// Tell it if switch is ok to use again (1=yes, it's a button).
//
void P_ChangeSwitchTexture(line_t* line, int useAgain)
{
    int     texTop, texMid, texBot;
    int     i, sound;
	
    if (!useAgain) line->special = 0;
	
    texTop = sides[line->sidenum[0]].toptexture;
    texMid = sides[line->sidenum[0]].midtexture;
    texBot = sides[line->sidenum[0]].bottomtexture;
	
    sound = sfx_swtchn;

    // EXIT SWITCH?
    if (line->special == 11) sound = sfx_swtchx;
	
	if ( !(texTop | texMid | texBot) )
	{
		// Check if line had a texture that was removed
		if (line->ovflags & OVF_SWITCH)
 			S_StartSound(sound, (mobj_t*)buttonlist->soundorg);

		return;
	}

    for (i = 0; i < numswitches*2; i++)
    {
		if (switchlist[i] == texTop)
		{
			S_StartSound(sound, (mobj_t*)buttonlist->soundorg);
			sides[line->sidenum[0]].toptexture = switchlist[i^1];
			
			if (useAgain)
				P_StartButton(line,button_top,switchlist[i],BUTTONTIME);
			
			return;
		}
		else if (switchlist[i] == texMid)
		{
			S_StartSound(sound, (mobj_t*)buttonlist->soundorg);
			sides[line->sidenum[0]].midtexture = switchlist[i^1];
			
			if (useAgain)
				P_StartButton(line, button_middle,switchlist[i],BUTTONTIME);
			
			return;
		}
		else if (switchlist[i] == texBot)
		{
			S_StartSound(sound, (mobj_t*)buttonlist->soundorg);

			sides[line->sidenum[0]].bottomtexture = switchlist[i^1];
			
			if (useAgain)
				P_StartButton(line, button_bottom,switchlist[i],BUTTONTIME);
			
			return;
		}
    }
}

int P_Activate1_1801x_line(mobj_t* thing,line_t *line)
{
	if ( !(line->xtraflags & XT_TRIGGER) ) return false;
	if (thing != ppt.viewplrmo) return false;	// player trigger only

	if (line->special == 8143)	// gun triggered so could be multiple hits
	{
		static int start = 0;

		if (abs(leveltime - start) > 4)
		{
			start = leveltime;
		}
		else
			return false;
	}

	if (TS_Activate(thing, line, MS_USE))
	{
		int lnum = line - lines, lnumf, lnumt;
		int i, c, k, count = 0;

		for (i = 0; i < 100; count++, i += 4)
		{
			if ( (lnumf = ltarray[count].line0) < 0) return false;	// end of list

			if (lnum == ltarray[count].line0)
			{
				if (line->xtraflags & XT_TRIGGER1)
				{
					line->xtraflags &= ~XT_TRIGGER1;

					if ( (lnumt = ltarray[count].line2) >= 0)
						subfn_UseSpecialLine(thing, lines + lnumt, 14);

					return false;
				}
				else
				{
					line->xtraflags |= XT_TRIGGER1;

					if ( (lnumt = ltarray[count].line1) >= 0)
						subfn_UseSpecialLine(thing, lines + lnumt, 14);

					// deactivate a linked switch function
					if ( (lnumf = ltarray[count].line3) > 0)
						subfn_UseSpecialLine(thing, lines + lnumf, 14);

					// check the linked switch state
					for (c = k = 0; k < 100; c++, k += 4)
					{
						if (c == count) continue;

						if (ltarray[c].line0 < 0) return false;	// end of list

						// check if it has been triggered
						// line3 will be the linked stop line
						if (   lnumf == ltarray[c].line2
							&& lines[ltarray[c].line0].xtraflags & XT_TRIGGER1)	// waiting to detrigger
						{
							lines[ltarray[c].line0].xtraflags &= ~XT_TRIGGER1;
							TS_Activate(thing, &lines[ltarray[c].line0], MS_USE);
							break;
						}
					}
					return false;
				}
			}
		}
	}
	return false;
}

//
// P_UseSpecialLine
// Called when a thing uses a special line.
// If only side 0 facing triggering is required this must
// be verified before calling.
//
// v220-4 made global as may need to be referenced by script

// v220-35 Scripted map support
int		*doorspotlnums;
int		numdoorspots;
int		maxnumdoorspots;

RDBOOL subfn_UseSpecialLine(mobj_t* thing, line_t* line, int from)
{               
	short	special;
	int		res;

	if (!line || (special = line->special) <= 0) return false;

	if (R3D_ScriptedMap && numdoorspots)
	{
void ext_LockedMessage(line_t *ld, char *msg);
extern char	*msga8001[];
extern char	*msgb8001[];

		for (int n = 0; n < numdoorspots; n++)
		{
			if ((doorspotlnums[n] & 0x0FFFFFFF) == line - lines)
			{
				int type = (doorspotlnums[n] & 0xF0000000) >> 28;

				if (thing->player)
				{
					if (line->lnflags2 & LNF2_BLOCKALL)
					{
						if (type)
							ext_LockedMessage(line, msgb8001[type - 1]);
						else
							ext_LockedMessage(line, "Door operation has been disabled!");
					}
					else
					{
						if (type)
							ext_LockedMessage(line, msga8001[type - 1]);
						else
							ext_LockedMessage(line, "This door is switch operated!");
					}
				}
				return false;
			}
		}
	}

	if (special == 8145) return false;	// used to trigger numdoorspots test when  a door line itself
										// has no trigger through being indirectly triggered.

	if (R3D_ScriptedMap && special == 8142)
	{
		P_Activate1_1801x_line(thing, line);
		return false;
	}

	if (R3D_maptype)
	{
		if (special == 8090)	// R3D ambient use trigger - no other action
		{
			S_SetLineTriggeredAmbient(line->ambNumLink, 6);
			return false;
		}

		if (special == 8093)	// Line sound - no other action
		{
			if (line->linesound && thing->player)	// v220-14 play only if player triggered
				S_StartSound (line->linesound, thing);
			return false;
		}

		if (TS_Activate(thing, line, MS_USE)) return true;	// $modelswitch use

		// v220-31 teleport line checks
		if (line->xtraflags & XT_PLRTRIGONLY && !thing->player) return false;

		if (line->xtraflags & XT_LIGHTFLAGS)
		{
			cfg.useTorchLight =  cfg.setVtxLight = cfg.sineCompress = 0;

			if (line->xtraflags & XT_TURNTORCHON)
			{
				cfg.useTorchLight = 1;
			}
			else
			{
				if (line->xtraflags & XT_TURNVTXON ) cfg.setVtxLight = 1;
			}

			if (line->xtraflags & XT_TURNSINEON) cfg.sineCompress = 1;
		}
	}

	if ((ushort)special >= (ushort)GenLimit) return false;

	if (R3D_maptype && thing->player)
	{
		// v220-31 line sound checks
		if (line->xtraflags & XT_PLAYLSONCE)
		{
			S_StartSound (line->linesound, thing);
		}
		else
		if (line->xtraflags & XT_PLAYLSREPEAT)
		{
			for (int i = 0; i < MAXLINESOUNDS; i++)
			{
				if (line->linesound == soundstarts[i]) goto end_check;
			}

			for (i = 0; i < MAXLINESOUNDS; i++)
			{
				if (!soundstarts[i]) { soundstarts[i] = line->linesound; break; }
			}
end_check:;
			S_SoundRepeat(line->linesound|(HSF_REPEAT|HSF_NO_ATTENUATION), NULL, 1.0f, 0);
		}
	}

	//jff 02/04/98 add check here for generalized floor/ceil mover
	// pointer to line function is NULL by default, set non-null if
	// line special is push or switch generalized linedef type
	// check each range of generalized linedefs
	//

	if (special >= GenCrusherBase)
	{
		int (*linefunc)(line_t *line) = NULL;
		int type;

		// jff 2/27/98 all non-manual - generalized types require tag
		if (!line->tag && (special & 6) != 6) return false;

		if (special >= GenFloorBase)
		{
			// FloorModel is "Allow Monsters" if FloorChange is 0
			if (!thing->player && (special & FloorChange || !(special & FloorModel)) )
				return false;
			linefunc = EV_DoGenFloor;
		}
		else if (special >= GenCeilingBase)
		{
			// CeilingModel is "Allow Monsters" if CeilingChange is 0
			if (!thing->player && (special & CeilingChange || !(special & CeilingModel)) )
				return false;
			linefunc = EV_DoGenCeiling;
		}
		else if (special >= GenDoorBase)
		{
			// monsters may be disallowed, they can't open secret doors in any event
			if (!thing->player && (!(special & DoorMonster) || line->ml_flags & ML_SECRET) )
				return false;
			linefunc = EV_DoGenDoor;
		}
		else if (special >= GenLockedBase)
		{
			locked_door = true;
			// monsters disallowed from unlocking doors
			if (!thing->player || !P_CanUnlockGenDoor(line, thing->player, true) )
				return false;
			locked_door = false;
			linefunc = EV_DoGenLockedDoor;
		}
		else if (special >= GenLiftBase)
		{
			if (!thing->player && !(special & LiftMonster) )
				return false; // monsters disallowed
			linefunc = EV_DoGenLift;
		}
		else if (special >= GenStairsBase)
		{
			if (!thing->player && !(special & StairMonster) )
				return false; // monsters disallowed
			linefunc = EV_DoGenStairs;
		}
		else	// GenCrusherBase
		{
			if (!thing->player && !(special & CrusherMonster) )
				return false; // monsters disallowed
			linefunc = EV_DoGenCrusher;
		}
		
		type = (special & TriggerType) >> TriggerTypeShift;

		switch(type)
		{
			case PushOnce:
				if (linefunc(line))
				{
					line->special = 0; return true;
				}
				return false;
			case PushMany:
				return linefunc(line);
			case SwitchOnce:
			case SwitchMany:
				if (linefunc(line))
				{
					P_ChangeSwitchTexture(line, type == SwitchMany);
					return true;
				}
			default:  // if not a switch/push type, do nothing here
				return false;
		}
	}

	// Switches that other things can activate.
	if (!thing->player)
	{
		// never open secret doors
		if (line->ml_flags & ML_SECRET)
			return false;
		
		switch(special)
		{
		case 1: 		// MANUAL DOOR RAISE
		case 32:		// MANUAL BLUE
		case 33:		// MANUAL RED
		case 34:		// MANUAL YELLOW
			//jff 3/5/98 add ability to use teleporters for monsters
		case 195:		// switch teleporters
		case 174:
		case 210:		// silent switch teleporters
		case 209:
			break;
			
		default:
			return false;
			break;
		}
	}

	//jff 2/27/98 disallow zero tag on some types
	if (!P_CheckTag(line)) return false;

	// Dispatch to handler according to linedef type
	switch (special)
	{
		// Manual doors, push type with no tag
	case 1: 			// Vertical Door
	case 26:			// Blue Door/Locked
	case 27:			// Yellow Door /Locked
	case 28:			// Red Door /Locked
		
	case 31:			// Manual door open
	case 32:			// Blue locked door open
	case 33:			// Red locked door open
	case 34:			// Yellow locked door open
		
	case 117:			// Blazing door raise
	case 118:			// Blazing door open
		EV_VerticalDoor (line, thing);
		break;
		
		// Switches (non-retriggerable)
	case 7:
		// Build Stairs
		if (EV_BuildStairs(line,build8))
		{
			P_ChangeSwitchTexture(line,0);
		}
		break;
		
	case 9:
		// Change Donut
		if (EV_DoDonut(line))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 11:
		// Exit level
		// killough 10/98: prevent zombies from exiting levels
		if (thing->dplayer && thing->dplayer->mo != thing && thing->health <= 0)
			return false;

		P_ChangeSwitchTexture(line,0);
		// G_ExitLevel ();
		ma_inf.exitlevel = 0x0080000A;	// R3D add delay so switch change can be seen
		break;
		
	case 14:
		// Raise Floor 32 and change texture
		if (EV_DoPlat(line,raiseAndChange,32))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 15:
		// Raise Floor 24 and change texture
		if (EV_DoPlat(line,raiseAndChange,24))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 18:
		// Raise Floor to next highest floor
		if (EV_DoFloor(line, raiseFloorToNearest))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 20:
		// Raise Plat next highest floor and change texture
		if (EV_DoPlat(line,raiseToNearestAndChange,0))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 21:
		// PlatDownWaitUpStay
		if (EV_DoPlat(line,downWaitUpStay,0))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 23:
		// Lower Floor to Lowest
		if (EV_DoFloor(line,lowerFloorToLowest))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 29:
		// Raise Door
		if (EV_DoDoor(line,normal))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 41:
		// Lower Ceiling to Floor
		if (EV_DoCeiling(line,lowerToFloor))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 71:
		// Turbo Lower Floor
		if (EV_DoFloor(line,turboLower))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 49:
		// Ceiling Crush And Raise
		if (EV_DoCeiling(line,crushAndRaise))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 50:
		// Close Door
		if (EV_DoDoor(line,closeNormal))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 51:
		// Secret EXIT: killough 10/98: prevent zombies from exiting levels
		if (thing->dplayer && thing->dplayer->mo != thing)
		{
			S_StartSound(sfx_noway, thing);
			return false;
		}
		
		P_ChangeSwitchTexture(line,0);
	//	G_SecretExitLevel ();
		ma_inf.exitlevel = 0x0800100A;	// R3D add delay so switch change can be seen
		break;
		
	case 55:
		// Raise Floor Crush
		if (EV_DoFloor(line,raiseFloorCrush))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 101:
		// Raise Floor
		if (EV_DoFloor(line,raiseFloor))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 102:
		// Lower Floor to Surrounding floor height
		if (EV_DoFloor(line,lowerFloor))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 103:
		// Open Door
		if (EV_DoDoor(line,openNormal))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 111:
		// Blazing Door Raise (faster than TURBO!)
		if (EV_DoDoor (line,blazeRaise))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 112:
		// Blazing Door Open (faster than TURBO!)
		if (EV_DoDoor (line,blazeOpen))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 113:
		// Blazing Door Close (faster than TURBO!)
		if (EV_DoDoor (line,blazeClose))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 122:
		// Blazing PlatDownWaitUpStay
		if (EV_DoPlat(line,blazeDWUS,0))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 127:
		// Build Stairs Turbo 16
		if (EV_BuildStairs(line,turbo16))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 131:
		// Raise Floor Turbo
		if (EV_DoFloor(line,raiseFloorTurbo))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 133:
		// BlzOpenDoor BLUE
	case 135:
		// BlzOpenDoor RED
	case 137:
		// BlzOpenDoor YELLOW
		if ( (res = EV_DoLockedDoor (line,blazeOpen,thing)) )
			P_ChangeSwitchTexture(line,0);
		return res;
		
	case 140:
		// Raise Floor 512
		if (EV_DoFloor(line,raiseFloor512))
			P_ChangeSwitchTexture(line,0);
		break;
		
	//jff 1/29/98 added linedef types to fill all functions out so that
	// all possess SR, S1, WR, W1 types
			
	case 158:
		// Raise Floor to shortest lower texture
		// 158 S1  EV_DoFloor(raiseToTexture), CSW(0)
		if (EV_DoFloor(line,raiseToTexture))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 159:
		// Raise Floor to shortest lower texture
		// 159 S1  EV_DoFloor(lowerAndChange)
		if (EV_DoFloor(line,lowerAndChange))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 160:
		// Raise Floor 24 and change
		// 160 S1  EV_DoFloor(raiseFloor24AndChange)
		if (EV_DoFloor(line,raiseFloor24AndChange))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 161:
		// Raise Floor 24
		// 161 S1  EV_DoFloor(raiseFloor24)
		if (EV_DoFloor(line,raiseFloor24))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 162:
		// Moving floor min n to max n
		// 162 S1  EV_DoPlat(perpetualRaise,0)
		if (EV_DoPlat(line,perpetualRaise,0))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 163:
		// Stop Moving floor
		// 163 S1  EV_DoPlat(perpetualRaise,0)
		EV_StopPlat(line);
		P_ChangeSwitchTexture(line,0);
		break;
		
	case 164:
		// Start fast crusher
		// 164 S1  EV_DoCeiling(fastCrushAndRaise)
		if (EV_DoCeiling(line,fastCrushAndRaise))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 165:
		// Start slow silent crusher
		// 165 S1  EV_DoCeiling(silentCrushAndRaise)
		if (EV_DoCeiling(line,silentCrushAndRaise))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 166:
		// Raise ceiling, Lower floor
		// 166 S1 EV_DoCeiling(raiseToHighest), EV_DoFloor(lowerFloortoLowest)
		if (EV_DoCeiling(line, raiseToHighest) ||
			EV_DoFloor(line, lowerFloorToLowest))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 167:
		// Lower floor and Crush
		// 167 S1 EV_DoCeiling(lowerAndCrush)
		if (EV_DoCeiling(line, lowerAndCrush))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 168:
		// Stop crusher
		// 168 S1 EV_CeilingCrushStop()
		if (EV_CeilingCrushStop(line))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 169:
		// Lights to brightest neighbor sector
		// 169 S1  EV_LightTurnOn(0)
		EV_LightTurnOn(line,0);
		P_ChangeSwitchTexture(line,0);
		break;
		
	case 170:
		// Lights to near dark
		// 170 S1  EV_LightTurnOn(35)
		EV_LightTurnOn(line,35);
		P_ChangeSwitchTexture(line,0);
		break;
		
	case 171:
		// Lights on full
		// 171 S1  EV_LightTurnOn(255)
		EV_LightTurnOn(line,255);
		P_ChangeSwitchTexture(line,0);
		break;
		
	case 172:
		// Start Lights Strobing
		// 172 S1  EV_StartLightStrobing()
		EV_StartLightStrobing(line);
		P_ChangeSwitchTexture(line,0);
		break;
		
	case 173:
		// Lights to Dimmest Near
		// 173 S1  EV_TurnTagLightsOff()
		EV_TurnTagLightsOff(line);
		P_ChangeSwitchTexture(line,0);
		break;
		
	case 174:
		// Teleport
		// 174 S1  EV_Teleport
		if (EV_Teleport(line, thing))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 175:
		// Close Door, Open in 30 secs
		// 175 S1  EV_DoDoor(close30ThenOpen)
		if (EV_DoDoor(line,close30ThenOpen))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 189: //jff 3/15/98 create texture change no motion type
		// Texture Change Only (Trigger)
		// 189 S1 Change Texture/Type Only
		if (EV_DoChange(line,trigChangeOnly))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 203:
		// Lower ceiling to lowest surrounding ceiling
		// 203 S1 EV_DoCeiling(lowerToLowest)
		if (EV_DoCeiling(line,lowerToLowest))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 204:
		// Lower ceiling to highest surrounding floor
		// 204 S1 EV_DoCeiling(lowerToMaxFloor)
		if (EV_DoCeiling(line,lowerToMaxFloor))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 209:
		// killough 1/31/98: silent teleporter
		//jff 209 S1 SilentTeleport 
		if (EV_SilentTeleport(line, thing))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 241: //jff 3/15/98 create texture change no motion type
		// Texture Change Only (Numeric)
		// 241 S1 Change Texture/Type Only
		if (EV_DoChange(line,numChangeOnly))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 221:
		// Lower floor to next lowest floor
		// 221 S1 Lower Floor To Nearest Floor
		if (EV_DoFloor(line,lowerFloorToNearest))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 229:
		// Raise elevator next floor
		// 229 S1 Raise Elevator next floor
		if (EV_DoElevator(line,elevateUp))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 233:
		// Lower elevator next floor
		// 233 S1 Lower Elevator next floor
		if (EV_DoElevator(line,elevateDown))
			P_ChangeSwitchTexture(line,0);
		break;
		
	case 237:
		// Elevator to current floor
		// 237 S1 Elevator to current floor
		if (EV_DoElevator(line,elevateCurrent))
			P_ChangeSwitchTexture(line,0);
		break;
		
		
	// jff 1/29/98 end of added S1 linedef types
		
	//jff 1/29/98 added linedef types to fill all functions out so that
	// all possess SR, S1, WR, W1 types
		
	case 78: //jff 3/15/98 create texture change no motion type
		// Texture Change Only (Numeric)
		// 78 SR Change Texture/Type Only
		if (EV_DoChange(line,numChangeOnly))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 176:
		// Raise Floor to shortest lower texture
		// 176 SR  EV_DoFloor(raiseToTexture), CSW(1)
		if (EV_DoFloor(line,raiseToTexture))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 177:
		// Raise Floor to shortest lower texture
		// 177 SR  EV_DoFloor(lowerAndChange)
		if (EV_DoFloor(line,lowerAndChange))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 178:
		// Raise Floor 512
		// 178 SR  EV_DoFloor(raiseFloor512)
		if (EV_DoFloor(line,raiseFloor512))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 179:
		// Raise Floor 24 and change
		// 179 SR  EV_DoFloor(raiseFloor24AndChange)
		if (EV_DoFloor(line,raiseFloor24AndChange))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 180:
		// Raise Floor 24
		// 180 SR  EV_DoFloor(raiseFloor24)
		if (EV_DoFloor(line,raiseFloor24))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 181:
		// Moving floor min n to max n
		// 181 SR  EV_DoPlat(perpetualRaise,0)
		
		EV_DoPlat(line,perpetualRaise,0);
		P_ChangeSwitchTexture(line,1);
		break;
		
	case 182:
		// Stop Moving floor
		// 182 SR  EV_DoPlat(perpetualRaise,0)
		EV_StopPlat(line);
		P_ChangeSwitchTexture(line,1);
		break;
		
	case 183:
		// Start fast crusher
		// 183 SR  EV_DoCeiling(fastCrushAndRaise)
		if (EV_DoCeiling(line,fastCrushAndRaise))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 184:
		// Start slow crusher
		// 184 SR  EV_DoCeiling(crushAndRaise)
		if (EV_DoCeiling(line,crushAndRaise))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 185:
		// Start slow silent crusher
		// 185 SR  EV_DoCeiling(silentCrushAndRaise)
		if (EV_DoCeiling(line,silentCrushAndRaise))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 186:
		// Raise ceiling, Lower floor
		// 186 SR EV_DoCeiling(raiseToHighest), EV_DoFloor(lowerFloortoLowest)
		if (EV_DoCeiling(line, raiseToHighest) ||
			EV_DoFloor(line, lowerFloorToLowest))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 187:
		// Lower floor and Crush
		// 187 SR EV_DoCeiling(lowerAndCrush)
		if (EV_DoCeiling(line, lowerAndCrush))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 188:
		// Stop crusher
		// 188 SR EV_CeilingCrushStop()
		if (EV_CeilingCrushStop(line))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 190: //jff 3/15/98 create texture change no motion type
		// Texture Change Only (Trigger)
		// 190 SR Change Texture/Type Only
		if (EV_DoChange(line,trigChangeOnly))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 191:
		// Lower Pillar, Raise Donut
		// 191 SR  EV_DoDonut()
		if (EV_DoDonut(line))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 192:
		// Lights to brightest neighbor sector
		// 192 SR  EV_LightTurnOn(0)
		EV_LightTurnOn(line,0);
		P_ChangeSwitchTexture(line,1);
		break;
		
	case 193:
		// Start Lights Strobing
		// 193 SR  EV_StartLightStrobing()
		EV_StartLightStrobing(line);
		P_ChangeSwitchTexture(line,1);
		break;
		
	case 194:
		// Lights to Dimmest Near
		// 194 SR  EV_TurnTagLightsOff()
		EV_TurnTagLightsOff(line);
		P_ChangeSwitchTexture(line,1);
		break;
		
	case 195:
		// Teleport
		// 195 SR  EV_Teleport
		if (EV_Teleport(line, thing))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 196:
		// Close Door, Open in 30 secs
		// 196 SR  EV_DoDoor(close30ThenOpen)
		if (EV_DoDoor(line,close30ThenOpen))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 205:
		// Lower ceiling to lowest surrounding ceiling
		// 205 SR EV_DoCeiling(lowerToLowest)
		if (EV_DoCeiling(line,lowerToLowest))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 206:
		// Lower ceiling to highest surrounding floor
		// 206 SR EV_DoCeiling(lowerToMaxFloor)
		if (EV_DoCeiling(line,lowerToMaxFloor))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 210:
		// killough 1/31/98: silent teleporter
		//jff 210 SR SilentTeleport 
		if (EV_SilentTeleport(line, thing))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 211: //jff 3/14/98 create instant toggle floor type
		// Toggle Floor Between C and F Instantly
		// 211 SR Toggle Floor Instant
		if (EV_DoPlat(line,toggleUpDn,0))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 222:
		// Lower floor to next lowest floor
		// 222 SR Lower Floor To Nearest Floor
		if (EV_DoFloor(line,lowerFloorToNearest))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 230:
		// Raise elevator next floor
		// 230 SR Raise Elevator next floor
		if (EV_DoElevator(line,elevateUp))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 234:
		// Lower elevator next floor
		// 234 SR Lower Elevator next floor
		if (EV_DoElevator(line,elevateDown))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 238:
		// Elevator to current floor
		// 238 SR Elevator to current floor
		if (EV_DoElevator(line,elevateCurrent))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 258:
		// Build stairs, step 8
		// 258 SR EV_BuildStairs(build8)
		if (EV_BuildStairs(line,build8))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 259:
		// Build stairs, step 16
		// 259 SR EV_BuildStairs(turbo16)
		if (EV_BuildStairs(line,turbo16))
			P_ChangeSwitchTexture(line,1);
		break;
		
	// 1/29/98 jff end of added SR linedef types

	// Buttons (retriggerable switches)
	case 42:
		// Close Door
		if (EV_DoDoor(line,closeNormal))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 43:
		// Lower Ceiling to Floor
		if (EV_DoCeiling(line,lowerToFloor))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 45:
		// Lower Floor to Surrounding floor height
		if (EV_DoFloor(line,lowerFloor))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 60:
		// Lower Floor to Lowest
		if (EV_DoFloor(line,lowerFloorToLowest))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 61:
		// Open Door
		if (EV_DoDoor(line,openNormal))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 62:
		// PlatDownWaitUpStay
		if (EV_DoPlat(line,downWaitUpStay,1))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 63:
		// Raise Door
		if (EV_DoDoor(line,normal))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 64:
		// Raise Floor to ceiling
		if (EV_DoFloor(line,raiseFloor))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 66:
		// Raise Floor 24 and change texture
		if (EV_DoPlat(line,raiseAndChange,24))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 67:
		// Raise Floor 32 and change texture
		if (EV_DoPlat(line,raiseAndChange,32))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 65:
		// Raise Floor Crush
		if (EV_DoFloor(line,raiseFloorCrush))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 68:
		// Raise Plat to next highest floor and change texture
		if (EV_DoPlat(line,raiseToNearestAndChange,0))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 69:
		// Raise Floor to next highest floor
		if (EV_DoFloor(line, raiseFloorToNearest))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 70:
		// Turbo Lower Floor
		if (EV_DoFloor(line,turboLower))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 114:
		// Blazing Door Raise (faster than TURBO!)
		if (EV_DoDoor (line,blazeRaise))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 115:
		// Blazing Door Open (faster than TURBO!)
		if (EV_DoDoor (line,blazeOpen))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 116:
		// Blazing Door Close (faster than TURBO!)
		if (EV_DoDoor (line,blazeClose))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 123:
		// Blazing PlatDownWaitUpStay
		if (EV_DoPlat(line,blazeDWUS,0))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 132:
		// Raise Floor Turbo
		if (EV_DoFloor(line,raiseFloorTurbo))
			P_ChangeSwitchTexture(line,1);
		break;
		
	case 99:
		// BlzOpenDoor BLUE
	case 134:
		// BlzOpenDoor RED
	case 136:
		// BlzOpenDoor YELLOW
		if ( (res = EV_DoLockedDoor (line,blazeOpen,thing)) )
			P_ChangeSwitchTexture(line,1);
		return res;
		
	case 138:
		// Light Turn On
		EV_LightTurnOn(line,255);
		P_ChangeSwitchTexture(line,1);
		break;
		
	case 139:
		// Light Turn Off
		EV_LightTurnOn(line,35);
		P_ChangeSwitchTexture(line,1);
		break;

	default:
		return false;
	}
	return true;
}

static RDBOOL checkTriggerHeightValid(line_t *lt, mobj_t *mo)
{
	if (!R3D_maptype) return true;

	if (lt->backsector &&
		(lt->specType >= R3D_40 || lt->ovflags & OVF_3DSWITCH))
	{
		int height = lt->frontsector->liftsec3D >= 0 ?
			sectors[lt->frontsector->liftsec3D].floorheight : lt->backsector->floorheight;
		int bottomStep = height +  lt->d3d->offset3D;
		int topStep = bottomStep + lt->d3d->totalheight;

		return (mo->z + mo->height >= bottomStep && mo->z <= topStep);
	}

	if (lt->ml_flags & ML_TWOSIDED)
	{
		if (   mo->z + mo->height < lt->frontsector->floorheight
			|| mo->z > lt->frontsector->ceilingheight)
		{
			return false;
		}
	}

	return true;
}

RDBOOL P_UseSpecialLine(mobj_t* thing, line_t* line, int side, int from)
{
	if (!line || side || !checkTriggerHeightValid(line, thing))
		return false;

	return subfn_UseSpecialLine(thing, line, from);
}