//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 id Software, Inc., Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_view.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//
// Based on: Doomsday v1.7.8 JDoom.dll code
//
// DESCRIPTION: set player viewheight, angles, movement bobbing.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "doom/tables.h"
#include "doom/p_local.h"
#include "net_main.h"
#include "r_main.h"
#include "con_functions.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define MAXBOB	0x100000		// 16 pixels of (fixed_t) bob.

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// P_CalcHeight
//	Calculate the walking / running height adjustment.
//===========================================================================
void P_CalcHeight(mbf_player_t* player) 
{
	ticcmd_t	*cmd = &player->cmd;
	RDBOOL		airborne;
	int			lives = player->playerstate == PST_LIVE;	// v220-4
	ddplayer_t	*dplay = player->plr;
	mobj_t		*pmo = dplay->mo;
	static int	aircounter = 0;

    // Regular movement bobbing.
    // (needs to be calculated for gun swing even if not on ground)
	if (lives && !cmd->forwardmove && ppt.viewplrmo->misc & MJMISC_TRANS_TC)
	{
		player->bob >>= 1;
	}
	else if (lives)
	{
		player->bob	
			= FixedMul(pmo->momx, pmo->momx)
			+ FixedMul(pmo->momy, pmo->momy);
		player->bob >>= 2;
		if(player->bob > MAXBOB) player->bob = MAXBOB;
	}
	else	// v220-4 set bobbing to zero if dead
	{
		player->bob = 0;
	}

	airborne = pmo->z > pmo->floorz; // Truly in the air?

	// Should view bobbing be done?
	if (pmo == player[0].plr->mo)
	{
		int		angle, target, step;

		if (/*   P_IsCamera(dplay->mo) // $ democam
			||*/ player->cheats & CF_NOMOMENTUM
			|| !lives	// v220-4
			|| airborne
			|| R3D_edit
			|| waterlevel == 2)
		{
			// Reduce the bob offset to zero.
			target = 0;
		}
		else
		{
			float bobmul = FRACUNIT * cfg.bobView;

			angle = (FINEANGLES/20*leveltime) & FINEMASK;

			switch(waterlevel)	// R3D
			{
			case 0:
			case 4:	// flying
				bobmul *= 0.5f;
				break;
			case 1:	//knees
				bobmul *= 0.1f;
				break;
			case 2:	//chest or
			case 3:	//under
				bobmul = 0.0f;
				break;
			default:
				break;
			}
			target = bobmul ? FixedMul((int)bobmul, FixedMul(player->bob/2, finesine[angle])) : 0; 
		}

		// Do the change gradually.
		angle = viewzOffset;
		if(airborne || aircounter > 0)
			step = 0x40000 - (aircounter > 0? aircounter*0x35C0 : 0x38000);
		else
			step = 0x40000;

		if(angle > target)
		{
			if(angle - target > step)
				angle -= step;
			else
				angle = target;
		}
		else if(angle < target)
		{
			if(target - angle > step)
				angle += step;
			else
				angle = target;
		}
		viewzOffset = angle;
		// The aircounter will soften the touchdown after a fall.
		aircounter--;
		if(airborne) aircounter = TICSPERSEC/2;
	}

	// Should viewheight be moved? Not if dead or a camera
	if (   lives
		&& !(player->cheats & CF_NOMOMENTUM/* || P_IsCamera(pmo)*/) )
	{
		dplay->viewheight += dplay->deltaviewheight;
		
		if(dplay->viewheight > plrviewheight)
		{
			dplay->viewheight = plrviewheight;
			dplay->deltaviewheight = 0;
		}
		if (dplay->viewheight < plrviewheight/2)
		{
			dplay->viewheight = plrviewheight/2;
			if (dplay->deltaviewheight <= 0)
				dplay->deltaviewheight = 1;
		}
		if (dplay->deltaviewheight)	
		{
			dplay->deltaviewheight += FRACUNIT/4;
			if (!dplay->deltaviewheight)
				dplay->deltaviewheight = 1;
		}
	}
	
	// Set the player's eye-level Z coordinate.
    dplay->viewz = pmo->z + dplay->viewheight;
}
