//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 id Software, Inc., G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: st_lib.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Moved to main engine
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:
//	Status bar code.
//	Does the face/direction indicator animation.
//	Does palette indicators as well (red pain/berserk, bright pickup)
//
//-----------------------------------------------------------------------------


#include <ctype.h>

#include "doom/mbf_defs.h"
#include "doom/st_lib.h"

#include "dd_wad.h"
#include "gl_draw.h"

//
// Hack display negative frags.
//  Loads and store the stminus lump.
//
//patch_t*		sttminus;
int				sttminus_i;

void STlib_init(void)
{
	sttminus_i = W_GetNumForName("STTMINUS");
}


// ?
void
STlib_initNum
( st_number_t*		n,
  int			x,
  int			y,
  dpatch_t*		pl,
  int*			num,
  RDBOOL*		on,
  int			width )
{
    n->x	= x;
    n->y	= y;
    n->oldnum	= 0;
    n->width	= width;
    n->num	= num;
    n->on	= on;
    n->p	= pl;
}


// 
// A fairly efficient way to draw a number
//  based on differences from the old number.
// Note: worth the trouble?
//
void
STlib_drawNum
( st_number_t*	n,
  RDBOOL	refresh )
{
    int		numdigits = n->width;
    int		num = *n->num;
    
    int		w = n->p[0].width;
//    int		h = n->p[0].height;
    int		x = n->x;
    
    int		neg;

    n->oldnum = *n->num;

    neg = num < 0;

    if (neg)
    {
		if (numdigits == 2 && num < -9)
			num = -9;
		else if (numdigits == 3 && num < -99)
			num = -99;
		num = -num;
    }

    // clear the area
    x = n->x - numdigits*w;

/*    if (n->y - ST_Y < 0)
	I_Error("drawNum: n->y - ST_Y < 0");*/

//    V_CopyRect(x, n->y - ST_Y, BG, w*numdigits, h, x, n->y, FG);

    // if non-number, do not draw it
    if (num == 1994)
	return;

    x = n->x;

    // in the special case of 0, you draw 0
    if (!num)
	GL_DrawPatch(x - w, n->y, /*FG,*/ n->p[ 0 ].lump);

    // draw the new number
    while (num && numdigits--)
    {
		x -= w;
		GL_DrawPatch(x, n->y, /*FG,*/ n->p[ num % 10 ].lump);
		num /= 10;
    }

    // draw a minus sign if necessary
    if (neg)
		GL_DrawPatch(x - 8, n->y,/*FG, */sttminus_i);
}


//
void
STlib_updateNum
( st_number_t*		n,
  RDBOOL		refresh )
{
    if (*n->on) STlib_drawNum(n, refresh);
}


//
void
STlib_initPercent
( st_percent_t*		p,
  int			x,
  int			y,
  dpatch_t*		pl,
  int*			num,
  RDBOOL*		on,
  dpatch_t*		percent )
{
    STlib_initNum(&p->n, x, y, pl, num, on, 3);
    p->p = percent;
}




void
STlib_updatePercent
( st_percent_t*		per,
  RDBOOL refresh )
{
    if (refresh && *per->n.on)
		GL_DrawPatch(per->n.x, per->n.y, /*FG, */per->p->lump);
    
    STlib_updateNum(&per->n, refresh);
}



void
STlib_initMultIcon
( st_multicon_t*	i,
  int			x,
  int			y,
  dpatch_t*		il,
  int*			inum,
  RDBOOL*		on )
{
    i->x	= x;
    i->y	= y;
    i->oldinum 	= -1;
    i->inum	= inum;
    i->on	= on;
    i->p	= il;
}



void
STlib_updateMultIcon
( st_multicon_t*	mi,
  RDBOOL		refresh )
{
    /*int			w;
    int			h;
    int			x;
    int			y;*/

    if (*mi->on	&& (mi->oldinum != *mi->inum || refresh) && (*mi->inum!=-1))
    {
		/*if (mi->oldinum != -1)
		{
			x = mi->x - mi->p[mi->oldinum].patch->leftoffset;
			y = mi->y - mi->p[mi->oldinum].patch->topoffset;
			w = mi->p[mi->oldinum].patch->width;
			h = mi->p[mi->oldinum].patch->height;

			if (y - ST_Y < 0)
				I_Error("updateMultIcon: y - ST_Y < 0");

//			V_CopyRect(x, y-ST_Y, BG, w, h, x, y, FG);
		}*/
		GL_DrawPatch(mi->x, mi->y, /*FG, */mi->p[*mi->inum].lump);
		mi->oldinum = *mi->inum;
    }
}

// v220-16 new function. Draws both keys if both are owned
void STlib_updateMultKeyIcon(st_multicon_t *mi, int kc, int ks)
{
	int shift = (ks >= 0 && kc >= 0) ? 1 : 0;
	if (kc >= 0) GL_DrawPatch(mi->x - shift, mi->y - shift, kc);
	if (ks >= 0) GL_DrawPatch(mi->x + shift, mi->y, ks);
}


void
STlib_initBinIcon
( st_binicon_t*		b,
  int			x,
  int			y,
  dpatch_t*		i,
  RDBOOL*		val,
  RDBOOL*		on )
{
    b->x	= x;
    b->y	= y;
    b->oldval	= 0;
    b->val	= val;
    b->on	= on;
    b->p	= i;
}



void
STlib_updateBinIcon
( st_binicon_t*		bi,
  RDBOOL		refresh )
{
    /*int			x;
    int			y;
    int			w;
    int			h;*/

    if (*bi->on && (bi->oldval != *bi->val || refresh))
    {
		/*x = bi->x - bi->p->patch->leftoffset;
		y = bi->y - bi->p->patch->topoffset;
		w = bi->p->patch->width;
		h = bi->p->patch->height;

		if (y - ST_Y < 0)
			I_Error("updateBinIcon: y - ST_Y < 0");*/

//	if (*bi->val)
	    GL_DrawPatch(bi->x, bi->y, /*FG,*/ bi->p->lump);
	/*else
	    V_CopyRect(x, y-ST_Y, BG, w, h, x, y, FG);*/

		bi->oldval = *bi->val;
    }

}

