//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: st_stuff.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Reformatted for sanity's sake, Moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:
//	Status bar code.
//	Does the face/direction indicator animation.
//	Does palette indicators as well (red pain/berserk, bright pickup)
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

// OGL
#include "doom/p_local.h"
#include "doom/m_cheat.h"
#include "doom/st_stuff.h"
#include "doom/st_lib.h"
#include "doom/am_map.h"
#include "doom/hu_stuff.h"
#include "doom/Mn_def.h"
#include "doom/d_englsh.h"
#include "con_console.h"
#include "con_functions.h"
#include "gl_draw.h"
#include "gl_tex.h"
#include "m_misc.h"
#include "net_main.h"
#include "p_maputil.h"
#include "r_things.h"
#include "dd_wad.h"
#include "dd_input.h"

#include "r3d_utils.h"
#include "r3d_compat.h"

#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

// Location of status bar
#define ST_X				0
#define ST_FX  				143

// Number of status faces.
#define ST_NUMPAINFACES			5
#define ST_NUMSTRAIGHTFACES		3
#define ST_NUMTURNFACES			2
#define ST_NUMSPECIALFACES		3

#define ST_FACESX				143
#define ST_FACESY				168

#define ST_TURNOFFSET			(ST_NUMSTRAIGHTFACES)
#define ST_OUCHOFFSET			(ST_TURNOFFSET + ST_NUMTURNFACES)
#define ST_RAMPAGEOFFSET		(ST_EVILGRINOFFSET + 1)

#define ST_FACESTRIDE \
          (ST_NUMSTRAIGHTFACES+ST_NUMTURNFACES+ST_NUMSPECIALFACES)

#define ST_NUMEXTRAFACES		2

#define ST_NUMFACES \
          (ST_FACESTRIDE*ST_NUMPAINFACES+ST_NUMEXTRAFACES)

#define ST_GODFACE				(ST_NUMPAINFACES*ST_FACESTRIDE)
#define ST_DEADFACE				(ST_GODFACE+1)
#define ST_MUCHPAIN			20

#define ST_EVILGRINCOUNT		(2*TICRATE)
#define ST_STRAIGHTFACECOUNT	(TICRATE/2)
#define ST_TURNCOUNT			(1*TICRATE)
#define ST_OUCHCOUNT			(1*TICRATE)
#define ST_RAMPAGEDELAY			(2*TICRATE)
#define ST_EVILGRINOFFSET		(ST_OUCHOFFSET + 1)

// Radiation suit, green shift.
#define RADIATIONPAL		13

// HEALTH number pos.
#define ST_HEALTHX			90
#define ST_HEALTHY			171

// Ammunition counter.
#define ST_AMMO0X			288
#define ST_AMMO0Y			173
#define ST_AMMO1WIDTH		ST_AMMO0WIDTH

// AMMO number pos.
#define ST_AMMOWIDTH		3	
#define ST_AMMOX			44
#define ST_AMMOY			171
#define ST_AMMO0WIDTH		3
#define ST_AMMO0HEIGHT		6
#define ST_AMMO1X			288
#define ST_AMMO1Y			179
#define ST_AMMO2WIDTH		ST_AMMO0WIDTH
#define ST_AMMO2X			288
#define ST_AMMO2Y			191
#define ST_AMMO3WIDTH		ST_AMMO0WIDTH
#define ST_AMMO3X			288
#define ST_AMMO3Y			185

// Indicate maximum ammunition.
// Only needed because backpack exists.
#define ST_MAXAMMO0WIDTH	3
#define ST_MAXAMMO0HEIGHT	5
#define ST_MAXAMMO0X		314
#define ST_MAXAMMO0Y		173
#define ST_MAXAMMO1WIDTH	ST_MAXAMMO0WIDTH
#define ST_MAXAMMO1X		314
#define ST_MAXAMMO1Y		179
#define ST_MAXAMMO2WIDTH	ST_MAXAMMO0WIDTH
#define ST_MAXAMMO2X		314
#define ST_MAXAMMO2Y		191
#define ST_MAXAMMO3WIDTH	ST_MAXAMMO0WIDTH
#define ST_MAXAMMO3X		314
#define ST_MAXAMMO3Y		185

// Weapon pos.
#define ST_ARMSX			111
#define ST_ARMSY			172
#define ST_ARMSBGX			104
#define ST_ARMSBGY			168
#define ST_ARMSXSPACE		12
#define ST_ARMSYSPACE		10

// Frags pos.
#define ST_FRAGSX			138
#define ST_FRAGSY			171	
#define ST_FRAGSWIDTH		2

// ARMOR number pos.
#define ST_ARMORX			221
#define ST_ARMORY			171

// Key icon positions.
#define ST_KEY0X			239
#define ST_KEY0Y			171
#define ST_KEY1X			239
#define ST_KEY1Y			181
#define ST_KEY2X			239
#define ST_KEY2Y			191

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

void ST_Stop(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static st_chatstateenum_t st_chatstate;	// used when in chat 
static mbf_player_t*	plyr;			// main player in game
static uint	st_clock;					// used for timing
static int	veryfirsttime = 1;			// used to execute ST_Init() only once
static int	st_msgcounter=0;			// used for making messages go away
static int	st_randomnumber;			// a random number per tick
static int	st_fragscount;				// number of frags so far in deathmatch
static int	keyboxes[3];				// holds key-type for each key box on bar

static RDBOOL		st_statusbaron;		// whether left-side main status bar is active
static RDBOOL		st_chat;			// whether status bar chat is active
static RDBOOL		st_oldchat;			// value of st_chat before message popped up
static RDBOOL		st_cursoron;		// whether chat window has the cursor on
static RDBOOL		st_notdeathmatch;	// !deathmatch
static RDBOOL		st_armson;			// !deathmatch && st_statusbaron
static RDBOOL		st_fragson;			// !deathmatch

static dpatch_t		sbar;				// main bar left
static dpatch_t		tallnum[10];		// 0-9, tall numbers
static dpatch_t 	tallpercent;		// tall % sign
static dpatch_t 	shortnum[10];		// 0-9, short, yellow (,different!) numbers
static dpatch_t 	keys[NUMCARDS];		// 3 key-cards, 3 skulls
static dpatch_t 	faces[ST_NUMFACES];	// face status patches
static dpatch_t 	faceback;			// face background
static dpatch_t 	armsbg;				// main bar right
static dpatch_t 	arms[6][2];			// weapon ownership patches

static st_number_t	w_ready;			// ready-weapon widget
static st_number_t	w_frags;			// in deathmatch only, summary of frags stats
static st_number_t	w_ammo[4];			// ammo widgets
static st_number_t	w_maxammo[4];		// max ammo widgets
static st_multicon_t w_arms[6];			// weapon ownership widgets
static st_multicon_t w_faces;			// face status widget
static st_multicon_t w_keyboxes[3];		// keycard widgets

static st_percent_t	w_health;			// health widget
static st_percent_t	w_armor;			// armor widget

static st_binicon_t	w_armsbg;			// arms background

static int	st_oldhealth = -1;			// used to use appopriately paint face
static RDBOOL	oldweaponsowned[NUMWEAPONS]; // used for evil grin
static int	st_facecount = 0;			// count until face changes
static int	st_faceindex = 0;			// current face index, used by w_faces


// Massive bunches of cheat shit
// to keep it from being easy to figure them out.
// Yeah, right...
unsigned char	cheat_mus_seq[] =
{
    0xb2, 0x26, 0xb6, 0xae, 0xea, 1, 0, 0, 0xff
};

unsigned char	cheat_choppers_seq[] =
{
    0xb2, 0x26, 0xe2, 0x32, 0xf6, 0x2a, 0x2a, 0xa6, 0x6a, 0xea, 0xff // idchoppers
};

unsigned char	cheat_god_seq[] =
{
    0xb2, 0x26, 0x26, 0xaa, 0x26, 0xff  // iddqd
};

unsigned char	cheat_ammo_seq[] =
{
    0xb2, 0x26, 0xf2, 0x66, 0xa2, 0xff	// idkfa
};

unsigned char	cheat_ammonokey_seq[] =
{
    0xb2, 0x26, 0x66, 0xa2, 0xff	// idfa
};

// Smashing Pumpkins Into Small Piles Of Putried Debris
unsigned char	cheat_noclip_seq[] =
{
    0xb2, 0x26, 0xea, 0x2a, 0xb2,	// idspispopd
    0xea, 0x2a, 0xf6, 0x2a, 0x26, 0xff
};

unsigned char	cheat_commercial_noclip_seq[] =
{
    0xb2, 0x26, 0xe2, 0x36, 0xb2, 0x2a, 0xff	// idclip
}; 

unsigned char	cheat_powerup_seq[7][10] =
{
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0x6e, 0xff },	// idbeholdv Invincibility
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0xea, 0xff },	// idbeholds Berserk
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0xb2, 0xff },	// idbeholdi Invisibility
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0x6a, 0xff },	// idbeholdr Radiation Suit
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0xa2, 0xff },	// idbeholdl Lite-Amp Goggles
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0x36, 0xff },	// idbeholda Auto-map
    { 0xb2, 0x26, 0x62, 0xa6, 0x32, 0xf6, 0x36, 0x26, 0xff }	// idbehold
};

unsigned char	cheat_clev_seq[] =
{
    0xb2, 0x26,  0xe2, 0x36, 0xa6, 0x6e, 1, 0, 0, 0xff	// idclev
};

unsigned char	cheat_mypos_seq[] =
{
    0xb2, 0x26, 0xb6, 0xba, 0x2a, 0xf6, 0xea, 0xff		// idmypos
}; 


// Now what?
cheatseq_t	cheat_mus = { cheat_mus_seq, 0 };
cheatseq_t	cheat_god = { cheat_god_seq, 0 };
cheatseq_t	cheat_ammo = { cheat_ammo_seq, 0 };
cheatseq_t	cheat_ammonokey = { cheat_ammonokey_seq, 0 };
cheatseq_t	cheat_noclip = { cheat_noclip_seq, 0 };
cheatseq_t	cheat_commercial_noclip = { cheat_commercial_noclip_seq, 0 };

cheatseq_t	cheat_powerup[7] =
{
    { cheat_powerup_seq[0], 0 },
    { cheat_powerup_seq[1], 0 },
    { cheat_powerup_seq[2], 0 },
    { cheat_powerup_seq[3], 0 },
    { cheat_powerup_seq[4], 0 },
    { cheat_powerup_seq[5], 0 },
    { cheat_powerup_seq[6], 0 }
};

cheatseq_t	cheat_choppers = { cheat_choppers_seq, 0 };
cheatseq_t	cheat_clev = { cheat_clev_seq, 0 };
cheatseq_t	cheat_mypos = { cheat_mypos_seq, 0 };

// CODE --------------------------------------------------------------------

// STATUS BAR
void ST_refreshBackground(void)
{
    if(st_statusbaron)
		GL_DrawPatch(ST_X, ST_Y, sbar.lump);
}

// Respond to keyboard input events,  intercept cheats.
RDBOOL ST_Responder (event_t* ev)
{
	int		i;
    
	// if a user keypress...
	if (ev->type == ev_keydown)
	{
		// b. - enabled for more debug fun.
		// if (gameskill != sk_nightmare) {
		
		// 'iddqd' cheat for toggleable god mode
		if (cht_CheckCheat(&cheat_god, ev->data1))
		{
			cht_GodFunc(plyr);
		}
		// 'idfa' cheat for "frigging arsenal"
		else if (cht_CheckCheat(&cheat_ammonokey, ev->data1))
		{
			if (cht_GiveFunc(plyr, true, true, true, false))
				P_SetMessage(plyr, STSTR_FAADDED);
		}
		// 'idkfa' cheat for "keys and frigging arsenal"
		else if (cht_CheckCheat(&cheat_ammo, ev->data1))
		{
			if (cht_GiveFunc(plyr, true, true, true, true))
				P_SetMessage(plyr, STSTR_KFAADDED);
		}
		// 'idmus' cheat for changing music
		else if (cht_CheckCheat(&cheat_mus, ev->data1))
		{
			char buf[3];
			P_SetMessage(plyr, STSTR_MUS);
			cht_GetParam(&cheat_mus, buf);
			cht_MusicFunc(plyr, buf); // Might set plyr->message.
		}
		// Simplified, accepting both "noclip" and "idspispopd".
		// no clipping mode cheat
		else if ( cht_CheckCheat(&cheat_noclip, ev->data1) 
			|| cht_CheckCheat(&cheat_commercial_noclip,ev->data1) )
		{	
			cht_NoClipFunc(plyr);
		}
		// 'idbehold?' power-up cheats
		for (i = 0; i < 6; i++)
		{
			if (cht_CheckCheat(&cheat_powerup[i], ev->data1))
				cht_PowerUpFunc(plyr, i, STSTR_BEHOLDX);
		}
		
		// 'idbehold' power-up menu
		if (cht_CheckCheat(&cheat_powerup[6], ev->data1))
		{
			P_SetMessage(plyr, STSTR_BEHOLD);
		}
		// 'idchoppers' invulnerability & chainsaw
		else if (cht_CheckCheat(&cheat_choppers, ev->data1))
		{
			cht_ChoppersFunc(plyr);
			P_SetMessage(plyr, STSTR_CHOPPERS);
		}
		// 'idmypos' for player position
		else if (cht_CheckCheat(&cheat_mypos, ev->data1))
		{
			cht_PosFunc(plyr);
		}
		
		// 'clev' change-level cheat
		if (cht_CheckCheat(&cheat_clev, ev->data1))
		{
			char buf[3];
			cht_GetParam(&cheat_clev, buf);
			cht_WarpFunc(plyr, buf);
		}    
	}
	return false;
}

int ST_calcPainOffset(void)
{
    int		health;
    static int	lastcalc;
    static int	oldhealth = -1;
    
    health = plhealth.health > 100 ? 100 : plhealth.health;
	
    if (health != oldhealth)
    {
		lastcalc = ST_FACESTRIDE * (((100 - health) * ST_NUMPAINFACES) / 101);
		oldhealth = health;
    }
    return lastcalc;
}


//
// This is a not-very-pretty routine which handles
//  the face states and their timing.
// the precedence of expressions is:
//  dead > evil grin > turned head > straight ahead
//
void ST_updateFaceWidget(void)
{
    static int	lastattackdown = -1;
    static int	priority = 0;
    angle_t	badguyangle;
    angle_t	diffang;
    RDBOOL	doevilgrin;
    int		i;
	
    if (priority < 10)
    {
		// dead
		if (!plhealth.health)
		{
			priority = 9;
			st_faceindex = ST_DEADFACE;
			st_facecount = 1;
		}
    }
	
    if (priority < 9)
    {
		if (plyr->bonuscount)
		{
			// picking up bonus
			doevilgrin = false;
			
			for (i=0;i<NUMWEAPONS;i++)
			{
				if (oldweaponsowned[i] != plyr->weaponowned[i])
				{
					doevilgrin = true;
					oldweaponsowned[i] = plyr->weaponowned[i];
				}
			}
			if (doevilgrin) 
			{
				// evil grin if just picked up weapon
				priority = 8;
				st_facecount = ST_EVILGRINCOUNT;
				st_faceindex = ST_calcPainOffset() + ST_EVILGRINOFFSET;
			}
		}
    }
	
    if (priority < 8)
    {
		if (plyr->damagecount
			&& plyr->attacker
			&& plyr->attacker != plyr->plr->mo)
		{
			priority = 7;			// being attacked

			// b220-0 Doom bugfix st_oldhealth and plyr->health were transposed
			if (st_oldhealth - plhealth.health > ST_MUCHPAIN)
			{
				st_facecount = ST_TURNCOUNT;
				st_faceindex = ST_calcPainOffset() + ST_OUCHOFFSET;
			}
			else
			{
				badguyangle = R_PointToAngle2(plyr->plr->mo->x,
					plyr->plr->mo->y,
					plyr->attacker->x,
					plyr->attacker->y);
				
				if (badguyangle > plyr->plr->mo->angle)
				{
					// whether right or left
					diffang = badguyangle - plyr->plr->mo->angle;
					i = diffang > ANG180; 
				}
				else
				{
					// whether left or right
					diffang = plyr->plr->mo->angle - badguyangle;
					i = diffang <= ANG180; 
				}

				st_facecount = ST_TURNCOUNT;
				st_faceindex = ST_calcPainOffset();
				
				if (diffang < ANG45)
				{
					// head-on    
					st_faceindex += ST_RAMPAGEOFFSET;
				}
				else if (i)
				{
					// turn face right
					st_faceindex += ST_TURNOFFSET;
				}
				else
				{
					// turn face left
					st_faceindex += ST_TURNOFFSET+1;
				}
			}
		}
    }
	
    if (priority < 7)
    {
		// getting hurt because of your own damn stupidity
		if (plyr->damagecount)
		{
			if (st_oldhealth - plhealth.health > ST_MUCHPAIN)	//b220-0 Doom bugfix
			{
				priority = 7;
				st_facecount = ST_TURNCOUNT;
				st_faceindex = ST_calcPainOffset() + ST_OUCHOFFSET;
			}
			else
			{
				priority = 6;
				st_facecount = ST_TURNCOUNT;
				st_faceindex = ST_calcPainOffset() + ST_RAMPAGEOFFSET;
			}
		}
    }
	
    if (priority < 6)
    {
		// rapid firing
		if (plyr->attackdown)
		{
			if (lastattackdown==-1)
				lastattackdown = ST_RAMPAGEDELAY;
			else if (!--lastattackdown)
			{
				priority = 5;
				st_faceindex = ST_calcPainOffset() + ST_RAMPAGEOFFSET;
				st_facecount = 1;
				lastattackdown = 1;
			}
		}
		else
			lastattackdown = -1;
		
    }
	
    if (priority < 5)
    {
		// invulnerability
		if ((plyr->cheats & CF_GODMODE)
			|| plyr->powers[pw_invulnerability])
		{
			priority = 4;
			
			st_faceindex = ST_GODFACE;
			st_facecount = 1;
		}
    }
	
    // look left or look right if the facecount has timed out
    if (!st_facecount)
    {
		st_faceindex = ST_calcPainOffset() + (st_randomnumber % 3);
		st_facecount = ST_STRAIGHTFACECOUNT;
		priority = 0;
    }
	
    st_facecount--;
	
}

void ST_updateWidgets(void)
{
    static int	largeammo = 1994; // means "n/a"
    int		i;
	
    // must redirect the pointer if the ready weapon has changed.
    //  if (w_ready.data != plyr->readyweapon)
    //  {
    if (weaponinfo[plyr->readyweapon].ammo == am_noammo)
		w_ready.num = &largeammo;
    else
		w_ready.num = &plyr->ammo[weaponinfo[plyr->readyweapon].ammo];
    //{
    // static int tic=0;
    // static int dir=-1;
    // if (!(tic&15))
    //   plyr->ammo[weaponinfo[plyr->readyweapon].ammo]+=dir;
    // if (plyr->ammo[weaponinfo[plyr->readyweapon].ammo] == -100)
    //   dir = 1;
    // tic++;
    // }
    w_ready.data = plyr->readyweapon;
	
    // if (*w_ready.on)
    //  STlib_updateNum(&w_ready, true);
    // refresh weapon change
    //  }
	
    // update keycard multiple widgets
    for (i=0;i<3;i++)
    {
		keyboxes[i] = plyr->cards[i] ? i : -1;
		
		if (plyr->cards[i+3])
			keyboxes[i] = i+3;
    }
	
    // refresh everything if this is him coming back to life
    ST_updateFaceWidget();
	
    // used by the w_armsbg widget
    st_notdeathmatch = true;
    
    // used by w_arms[] widgets
    st_armson = st_statusbaron; 
	
    // used by w_frags widget
    st_fragson = 0; 
    st_fragscount = 0;
	
    // get rid of chat window if up because of message
    if (!--st_msgcounter)
		st_chat = st_oldchat;
	
}

void ST_Ticker (void)
{
    st_clock++;
    st_randomnumber = rand() % 256;
    ST_updateWidgets();
    st_oldhealth = plhealth.health;
}

void STlib_updateMultKeyIcon(st_multicon_t *mi, int kc, int ks);	// v220-16

void ST_drawWidgets(RDBOOL refresh)
{
    int		i;
	
    // used by w_arms[] widgets
    st_armson = st_statusbaron;
	
    // used by w_frags widget
    st_fragson = 0; 
	
    STlib_updateNum(&w_ready, refresh);
	
    for (i=0;i<4;i++)
    {
		STlib_updateNum(&w_ammo[i], refresh);
		STlib_updateNum(&w_maxammo[i], refresh);
    }
	
    STlib_updatePercent(&w_health, refresh);
    STlib_updatePercent(&w_armor, refresh);
	
    STlib_updateBinIcon(&w_armsbg, refresh);
	
    for (i=0;i<6;i++)
		STlib_updateMultIcon(&w_arms[i], refresh);
	
    STlib_updateMultIcon(&w_faces, refresh);
	
	// v220-16 Modified to draw both keys if both are owned
    for (i = 0; i < 3; i++)
	{
		if (w_keyboxes[i].on)
		{
			int kc = plyr->cards[i]   ? keys[i].lump   : -1;
			int ks = plyr->cards[i+3] ? keys[i+3].lump : -1;

			if (kc >= 0 || ks >= 0)
				STlib_updateMultKeyIcon(&w_keyboxes[i], kc, ks);
		}
	}
	
    STlib_updateNum(&w_frags, refresh);
}

void ST_doRefresh(void)
{
	if(cfg.sbarscale < 5)	// scale will need changing
	{
		static float xscales[3] = { 192.0f, 96.0f, 0 };

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glTranslatef(xscales[cfg.sbarscale - 2], 120.0f, 0);
		glScalef(0.4f, 0.4f, 1.0f);
	}

    // draw status bar background to off-screen buff
    ST_refreshBackground();

    // and refresh all widgets
    ST_drawWidgets(true);

	if (cfg.sbarscale < 5)
	{
		// Restore the normal modelview matrix.
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
}
/*
void ST_diffDraw(void)
{
    // update all widgets
    ST_drawWidgets(false);
}
*/
int ST_drawHUDSprite(int sprite, int x, int y, RDBOOL shiftleft)
{
	spriteinfo_t	sprInfo;
	
	R_GetSpriteInfo(sprite, 0, &sprInfo, 6);
	glColor3f(1.0f, 1.0f, 1.0f);

	return GL_DrawPSprite((float)x, (float)y, false, sprInfo.splnum, true, shiftleft);
}

void ST_doFullscreenStuff(void)
{
	mbf_player_t *plr = &mbf_players[0];
	char buf[20];
	int pos=0, spr, w;
	int	h_width  = (int)(320 / cfg.hudScale);
	int h_height = (int)(200 / cfg.hudScale);

	if (automap_active) return;

	// Setup the scaling matrix.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glScalef(cfg.hudScale, cfg.hudScale, 1.0f);

	// draw the visible HUD data, first health
	if(cfg.hudShown[HUD_HEALTH]) 
	{
extern int healthSPR;
		w = ST_drawHUDSprite(healthSPR, 2, h_height-2, false);
		sprintf(buf, "%i%%", plhealth.health);
		M_WriteText2(w + 4, h_height - 14, buf, hu_font_b,
			cfg.hudColor[0], cfg.hudColor[1], cfg.hudColor[2]);
		pos = w + 10 + strlen(buf) * 10;
	}	

	if(cfg.hudShown[HUD_AMMO]
		&& weaponinfo[plr->readyweapon].ammo != am_noammo)
	{
		spr = ammo_sprite[weaponinfo[plr->readyweapon].ammo];
		w = ST_drawHUDSprite(spr, pos + 2, h_height - 2, false);
		sprintf(buf, "%i", plr->ammo[weaponinfo[plr->readyweapon].ammo]);
		M_WriteText2(pos+w+4, h_height-14, buf, hu_font_b,
			cfg.hudColor[0], cfg.hudColor[1], cfg.hudColor[2]);
	}

	pos = h_width - 10;

	if(cfg.hudShown[HUD_ARMOR])
	{
		sprintf(buf, "%i%%", plhealth.armorpoints);
		pos -= strlen(buf)*10;
		M_WriteText2(pos, h_height - 14, buf, hu_font_b,
			cfg.hudColor[0], cfg.hudColor[1], cfg.hudColor[2]);
		spr = plhealth.armortype == 2 ? SPR_ARM2 : SPR_ARM1;
		pos -= 2;
		w = ST_drawHUDSprite(spr, pos, h_height-2, true);
		pos -= (w + 10);
	}

	// Modified to show both skull and card keys
	if(cfg.hudShown[HUD_KEYS])
	{
		if (R3D_ScriptedMap)
		{
			if (plr->cards[it_redcard    ]) pos -= ST_drawHUDSprite(SPR_RKEY, pos, h_height-2, true) + 2;
			if (plr->cards[it_yellowcard ]) pos -= ST_drawHUDSprite(SPR_YKEY, pos, h_height-2, true) + 2;
			if (plr->cards[it_bluecard   ]) pos -= ST_drawHUDSprite(SPR_BKEY, pos, h_height-2, true) + 2;
			if (plr->cards[it_redskull   ]) pos -= ST_drawHUDSprite(SPR_RSKU, pos, h_height-2, true) + 2;
			if (plr->cards[it_yellowskull]) pos -= ST_drawHUDSprite(SPR_YSKU, pos, h_height-2, true) + 2;
			if (plr->cards[it_blueskull  ]) pos -= ST_drawHUDSprite(SPR_BSKU, pos, h_height-2, true) + 2;
		}
		else
		{
			// Don't bother with loop
			if (plr->cards[it_redcard] ^ plr->cards[it_redskull])
				ST_drawHUDSprite(plr->cards[it_redskull] ? SPR_RSKU : SPR_RKEY, pos, h_height-2, true);
			else if (plr->cards[it_redcard])	// Both must be set if EOR false and it_redcard is true
			{
				ST_drawHUDSprite(SPR_RSKU, pos     , h_height-2, true);
				ST_drawHUDSprite(SPR_RKEY, pos +  6, h_height-6, true);
			}
			if (plr->cards[it_yellowcard] ^ plr->cards[it_yellowskull])
				ST_drawHUDSprite(plr->cards[it_yellowskull] ? SPR_YSKU : SPR_YKEY, pos - 20, h_height-2, true);
			else if (plr->cards[it_yellowcard])
			{
				ST_drawHUDSprite(SPR_YSKU, pos - 20, h_height-2, true);
				ST_drawHUDSprite(SPR_YKEY, pos - 14, h_height-6, true);
			}
			if (plr->cards[it_bluecard] ^ plr->cards[it_blueskull])
				ST_drawHUDSprite(plr->cards[it_blueskull] ? SPR_BSKU : SPR_BKEY, pos - 40, h_height-2, true);
			else if (plr->cards[it_bluecard])
			{
				ST_drawHUDSprite(SPR_BSKU, pos - 40, h_height-2, true);
				ST_drawHUDSprite(SPR_BKEY, pos - 34, h_height-6, true);
			}
		}
	}
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void ST_Drawer()
{
	if (cfg.sbarscale == 1) 
		ST_doFullscreenStuff();  
	else
		ST_doRefresh();
}

// dp->lump must be set and valid
// v220-8 nearest switch added
// v220-21 revised
void ST_SetPatch(dpatch_t *dp, int nearest)
{
	RDBOOL notset = !lumptexinfo[dp->lump].ltitex;

	GL_SetPatch(dp->lump, nearest);	// Precache/load it

	if (notset)
	{
		dp->width	= (int)((short)lumptexinfo[dp->lump].origw);
		dp->height	= (int)((short)lumptexinfo[dp->lump].origh);
		dp->xoffset	= (int)((short)lumptexinfo[dp->lump].offx );
		dp->yoffset	= (int)((short)lumptexinfo[dp->lump].offy );
	}
}

// v220-8 nearest switch added
void R_SetPatch(dpatch_t *dp, char *name, int nearest)
{
	// Get original doom lump patch. The width and height
	// will be needed to scale a hi-res version if found
	// by setPatch
	dp->lump = W_CheckOriginalPatch(name);

	if (dp->lump < 0)
	{
		// "FONTB091" to "FONTB095" are only patches allowed
		// not to cause an exit if not found. See INIT_STRINGS
		if (strnicmp("FONTB09", name, 7))
			Con_Error("R_SetPatch: lumpname \"%s\" not found\n", name);
		return;
	}
	
	ST_SetPatch(dp, nearest);
}

// Added for R3D (mapname display)
// lump is used if set in MAPINFO with titlepatch
void R_LevelPatch(dpatch_t *dp, int lump)
{
	if (lump < 0)
		dp->lump = -1;
	else
	{
		dp->lump = lump;
		ST_SetPatch(dp, 0);
	}
}

void ST_loadGraphics(void)
{
    int		i;
    int		j;
    int		facenum;
    
    char	namebuf[9];

    // Load the numbers, tall and short
    for (i=0;i<10;i++)
    {
		sprintf(namebuf, "STTNUM%d", i);
		R_SetPatch(&tallnum[i], namebuf, 0);

		sprintf(namebuf, "STYSNUM%d", i);
		R_SetPatch(&shortnum[i], namebuf, 0);
    }

    // Load percent key.
    // Note: why not load STMINUS here, too?
	R_SetPatch(&tallpercent, "STTPRCNT", 0);

    // key cards
    for (i=0;i<NUMCARDS;i++)
    {
		sprintf(namebuf, "STKEYS%d", i);
		R_SetPatch(&keys[i], namebuf, 0);
    }

    // arms background
	R_SetPatch(&armsbg, "STARMS", 0);

    // arms ownership widgets
    for(i=0; i<6; i++)
    {
		sprintf(namebuf, "STGNUM%d", i+2);

		// gray #
		R_SetPatch(&arms[i][0], namebuf, 0);

		// yellow #
		memcpy(&arms[i][1], &shortnum[i+2], sizeof(dpatch_t));
    }

    // face backgrounds for different color players
    sprintf(namebuf, "STFB%d", 0);
	R_SetPatch(&faceback, namebuf, 0);

    // status bar background bits
	R_SetPatch(&sbar, "STBAR", 0);

    // face states
    facenum = 0;
    for(i=0; i<ST_NUMPAINFACES; i++)
    {
		for(j=0; j<ST_NUMSTRAIGHTFACES; j++)
		{
			sprintf(namebuf, "STFST%d%d", i, j);
			R_SetPatch(&faces[facenum++], namebuf, 0);
		}
		sprintf(namebuf, "STFTR%d0", i);	// turn right
		R_SetPatch(&faces[facenum++], namebuf, 0);
		sprintf(namebuf, "STFTL%d0", i);	// turn left
		R_SetPatch(&faces[facenum++], namebuf, 0);
		sprintf(namebuf, "STFOUCH%d", i);	// ouch!
		R_SetPatch(&faces[facenum++], namebuf, 0);
		sprintf(namebuf, "STFEVL%d", i);	// evil grin ;)
		R_SetPatch(&faces[facenum++], namebuf, 0);
		sprintf(namebuf, "STFKILL%d", i);	// pissed off
		R_SetPatch(&faces[facenum++], namebuf, 0);
    }
	R_SetPatch(&faces[facenum++], "STFGOD0", 0);
	R_SetPatch(&faces[facenum++], "STFDEAD0", 0);
}

void ST_updateGraphics(void)
{
    char	namebuf[9];

    // face backgrounds for different color players
    sprintf(namebuf, "STFB%d", cfg.PlayerColor[0]);
	R_SetPatch(&faceback, namebuf, 0);
}

void ST_loadData(void)
{
    ST_loadGraphics();
}

void ST_initData(void)
{
    int		i;
	
    plyr = &mbf_players[0];
	
    st_clock = 0;
    st_chatstate = StartChatState;
	
    st_statusbaron = true;
    st_oldchat = st_chat = false;
    st_cursoron = false;
	
    st_faceindex = 0;	
    st_oldhealth = -1;

    for (i=0;i<NUMWEAPONS;i++)
		oldweaponsowned[i] = plyr->weaponowned[i];

    for (i=0;i<3;i++)
		keyboxes[i] = -1;
	
    STlib_init();
}

void ST_createWidgets(void)
{
    int i;

    // ready weapon ammo
    STlib_initNum(&w_ready,
		  ST_AMMOX,
		  ST_AMMOY,
		  tallnum,
		  &plyr->ammo[weaponinfo[plyr->readyweapon].ammo],
		  &st_statusbaron,
		  ST_AMMOWIDTH );

    // the last weapon type
    w_ready.data = plyr->readyweapon; 

    // health percentage
    STlib_initPercent(&w_health,
		      ST_HEALTHX,
		      ST_HEALTHY,
		      tallnum,
		      &plhealth.health,
		      &st_statusbaron,
		      &tallpercent);

    // arms background
    STlib_initBinIcon(&w_armsbg,
		      ST_ARMSBGX,
		      ST_ARMSBGY,
		      &armsbg,
		      &st_notdeathmatch,
		      &st_statusbaron);

    // weapons owned
    for(i=0;i<6;i++)
    {
	STlib_initMultIcon(&w_arms[i],
			   ST_ARMSX+(i%3)*ST_ARMSXSPACE,
			   ST_ARMSY+(i/3)*ST_ARMSYSPACE,
			   arms[i], 
			   (int *) &plyr->weaponowned[i+1],
			   &st_armson);
    }

    // frags sum
    STlib_initNum(&w_frags,
		  ST_FRAGSX,
		  ST_FRAGSY,
		  tallnum,
		  &st_fragscount,
		  &st_fragson,
		  ST_FRAGSWIDTH);

    // faces
    STlib_initMultIcon(&w_faces,
		       ST_FACESX,
		       ST_FACESY,
		       faces,
		       &st_faceindex,
		       &st_statusbaron);

    // armor percentage - should be colored later
    STlib_initPercent(&w_armor,
		      ST_ARMORX,
		      ST_ARMORY,
		      tallnum,
		      &plhealth.armorpoints,
		      &st_statusbaron, &tallpercent);

    // keyboxes 0-2
    STlib_initMultIcon(&w_keyboxes[0],
		       ST_KEY0X,
		       ST_KEY0Y,
		       keys,
		       &keyboxes[0],
		       &st_statusbaron);
    
    STlib_initMultIcon(&w_keyboxes[1],
		       ST_KEY1X,
		       ST_KEY1Y,
		       keys,
		       &keyboxes[1],
		       &st_statusbaron);

    STlib_initMultIcon(&w_keyboxes[2],
		       ST_KEY2X,
		       ST_KEY2Y,
		       keys,
		       &keyboxes[2],
		       &st_statusbaron);

    // ammo count (all four kinds)
    STlib_initNum(&w_ammo[0],
		  ST_AMMO0X,
		  ST_AMMO0Y,
		  shortnum,
		  &plyr->ammo[0],
		  &st_statusbaron,
		  ST_AMMO0WIDTH);

    STlib_initNum(&w_ammo[1],
		  ST_AMMO1X,
		  ST_AMMO1Y,
		  shortnum,
		  &plyr->ammo[1],
		  &st_statusbaron,
		  ST_AMMO1WIDTH);

    STlib_initNum(&w_ammo[2],
		  ST_AMMO2X,
		  ST_AMMO2Y,
		  shortnum,
		  &plyr->ammo[2],
		  &st_statusbaron,
		  ST_AMMO2WIDTH);
    
    STlib_initNum(&w_ammo[3],
		  ST_AMMO3X,
		  ST_AMMO3Y,
		  shortnum,
		  &plyr->ammo[3],
		  &st_statusbaron,
		  ST_AMMO3WIDTH);

    // max ammo count (all four kinds)
    STlib_initNum(&w_maxammo[0],
		  ST_MAXAMMO0X,
		  ST_MAXAMMO0Y,
		  shortnum,
		  &plyr->maxammo[0],
		  &st_statusbaron,
		  ST_MAXAMMO0WIDTH);

    STlib_initNum(&w_maxammo[1],
		  ST_MAXAMMO1X,
		  ST_MAXAMMO1Y,
		  shortnum,
		  &plyr->maxammo[1],
		  &st_statusbaron,
		  ST_MAXAMMO1WIDTH);

    STlib_initNum(&w_maxammo[2],
		  ST_MAXAMMO2X,
		  ST_MAXAMMO2Y,
		  shortnum,
		  &plyr->maxammo[2],
		  &st_statusbaron,
		  ST_MAXAMMO2WIDTH);
    
    STlib_initNum(&w_maxammo[3],
		  ST_MAXAMMO3X,
		  ST_MAXAMMO3Y,
		  shortnum,
		  &plyr->maxammo[3],
		  &st_statusbaron,
		  ST_MAXAMMO3WIDTH);
}


static RDBOOL	st_stopped = true;

void ST_Start (void)
{
    if (!st_stopped) ST_Stop();

    ST_initData();
    ST_createWidgets();
    st_stopped = false;

}

void ST_Stop (void)
{
    if (st_stopped) return;

    st_stopped = true;
}

void ST_Init (void)
{
    veryfirsttime = 0;
    ST_loadData();
}


//---------------------------------------------------------------------------
// CONSOLE COMMANDS 
//---------------------------------------------------------------------------

// This is the multipurpose cheat ccmd.
int CCmdCheat(int argc, char **argv)
{
	uint i;

	if(argc != 2)
	{
		// Usage information.
		Con_Printf( "Usage: cheat (cheat)\nFor example, 'cheat idclev25'.\n");
		return true;
	}
	// Give each of the characters in 2nd arg to the ST event handler.
	for(i=0; i<strlen(argv[1]); i++)
	{
		event_t ev;
		ev.type = ev_keydown;
		ev.data1 = argv[1][i];
		ev.data2 = ev.data3 = 0;
		ST_Responder(&ev);
	}
	return true;
}

int CCmdCheatGod(int argc, char **argv)
{
	if(!mbf_players[0].plr->mo) return false;
	cht_GodFunc(&mbf_players[0]);
	return true;
}

int CCmdCheatNoClip(int argc, char **argv)
{
	cht_NoClipFunc(&mbf_players[0]);
	return true;
}

// Modified to work properly (see also cht_WarpFunc and p_setup.c)
int CCmdCheatWarp(int argc, char **argv)
{
	char buf[10];

	if (   R3D_edit
		|| (gamestate != GS_WAITING && gamestate != GS_LEVEL))	// v220-31
	{
		return false;
	}

	buf[0] = 0;

	if(commercial_game)
	{
		if(argc != 2 || strlen(argv[1]) > 2) return false;
		sprintf(buf, "%.2i", atoi(argv[1]));
	}
	else
	{
		if (argc == 2 &&
			strlen(argv[1]) == 4 &&
			(argv[1][0] == 'E' || argv[1][0] == 'e') &&
			(argv[1][2] == 'M' || argv[1][2] == 'm'))
		{
			buf[0] = argv[1][1];
			buf[1] = argv[1][3];
		}
		else
		if(argc == 3 &&
			strlen(argv[1]) == 1 &&
			strlen(argv[2]) == 1)
		{
			buf[0] = argv[1][0];
			buf[1] = argv[2][0];
		}
		else return false;			
	}
	return cht_WarpFunc(&mbf_players[0], buf);
}

int CCmdCheatReveal(int argc, char **argv)
{
	extern RDBOOL check_plr_in_bounds;
	extern int    mapcheat;
	int	option;

	if (    argc != 2
		|| (option = atoi(argv[1])) < 0
		|| (ge_inf.devmap && option > 8)
		|| (!ge_inf.devmap && option > 7))
	{
		Con_Printf( "Current value: %i\n", mapcheat);
		Con_Printf( "Usage: reveal x (where x = 0 to 7)\n");
		Con_Printf( "0 normal\n");
		Con_Printf( "1 same as first use of the iddt cheat\n");
		Con_Printf( "2 same as next  use of the iddt cheat\n");
		Con_Printf( "3 show all lines + identify keys and unvisited secret sectors\n");
		Con_Printf( "4 show all lines: aid to map debug\n");
		Con_Printf( "5 show all lines and Things: aid to map debug\n");
		Con_Printf( "6 show all 1's lines only  : aid to map debug\n");
		Con_Printf( "7 show subsectors          : aid to map debug\n");
		if (ge_inf.devmap)
			Con_Printf( "8 show enemies             : aid to map debug\n");
		Con_Printf( "For full description see R3D_Docs\\Help\\R3D_AutoMap.txt\n");
		return true;
	}

	check_plr_in_bounds = ( !(mapcheat = option) );

	if (option < 3) mbf_players[0].powers[pw_allmap] = option;

	return true;
}

int CCmdCheatGive(int argc, char **argv)
{
	char	buf[100];
	int		i, k;
	mbf_player_t *plyr = &mbf_players[0];

	if (R3D_edit) return true;

	if(argc != 2)
	{
		Con_Printf( "Usage: give (stuff)\n");
		Con_Printf( "Stuff consists of one or more of:\n");
		Con_Printf( " a - ammo\n");
		Con_Printf( " b - berserk\n");
		Con_Printf( " c - ammo, weapons, armour and backpack given\n");
		Con_Printf( " g - light amplification visor\n");
		Con_Printf( " i - invulnerability\n");
		Con_Printf( " k - key cards/skulls\n");
		Con_Printf( " m - computer area map\n");
		Con_Printf( " r - armor\n");
		Con_Printf( " s - radiation shielding suit\n");
		Con_Printf( " v - invisibility\n");
		Con_Printf( " w - weapons\n");
		Con_Printf( "Example: 'give arw' corresponds to IDFA.\n");
		return true;		
	}

	strcpy(buf, argv[1]);
	strlwr(buf);

	for (i = 0; buf[i]; i++)
	{
		switch(buf[i])
		{
		case 'a':
			if (cht_GiveFunc(plyr, 0, true, 0, 0))
				P_SetMessage(plyr, "Ammo given.");
			break;

		case 'b':
			cht_PowerUpFunc(plyr, pw_strength, "Your vision blurs! Yaarrrgh!!");
			break;

		case 'c':	// v220-4 Combined cheat
			if (cht_GiveFunc(plyr, true, true, true, 0))
			{
				Con_Printf("Combined cheat: weapons, armour, backpack and ammo given.%s\n", ge_inf.maxcheat ? " Maxammo x 10" : "");

				if (!plyr->backpack)
				{
					for (k = 0 ; k < NUMAMMO; k++) plyr->maxammo[k] *= 2;

					plyr->backpack = true;
				}

				if (ge_inf.maxcheat && plyr->maxammo[0] == 400)
				{
					// v220-4 10 * ammo for map development
					for (k = 0 ; k < NUMAMMO; k++) plyr->maxammo[k] *= 10;
				}
			}
			break;

		case 'g':
			cht_PowerUpFunc(plyr, pw_infrared, "Light amplification visor given");
			break;

		case 'i':
			cht_PowerUpFunc(plyr, pw_invulnerability, "You feel invincible!");
			break;

		case 'k':
			if (cht_GiveFunc(plyr, 0, 0, 0, true))
				P_SetMessage(plyr, "Card and skull keys given");
			break;

		case 'm':
			cht_PowerUpFunc(plyr, pw_allmap, "Computer area map given");
			break;

		case 'r':
			if (cht_GiveFunc(plyr, 0, 0, true, 0))
				P_SetMessage(plyr, "Full armor given");
			break;

		case 's':
			cht_PowerUpFunc(plyr, pw_ironfeet, "Radiation shielding suit given");
			break;

		case 'v':
			cht_PowerUpFunc(plyr, pw_invisibility, "You are suddenly almost invisible!");
			break;

		case 'w':
			if (cht_GiveFunc(plyr, true, 0, 0, 0))
				P_SetMessage(plyr, "Weapons given");
			break;

		default:
			Con_Printf( "What do you mean, '%c'?\n", buf[i]);
		}
	}
	return true;
}

int CCmdCheatMassacre(int argc, char **argv)
{
	if (con_freezegame)
	{
		Con_Printf( "Game is no longer suspended\n");
		con_freezegame = false;
	}

	// v220-31 If using massacre cheat then don't drop weapons
	ma_inf.noItemDrop = true;
	Con_Printf( "%i monsters killed. NOTE: With cheat no items are dropped\n", P_Massacre(false));
	ma_inf.noItemDrop = false;
	return true;
}

// v220-10 (for map development only)
int CCmdCheatKill666(int argc, char **argv)
{
#include "r3d_mapinfo.h"

	if (nomonsters)
	{
		int		R3Dflags = mapinfo->R3Dflags;
		line_t	junk;
		int secnum = -1;

		junk.tag = 666;	// default
		junk.lnflags2 = 0;

		if (R3Dflags & specialaction_blazedoor)
			EV_DoDoor(&junk, blazeOpen);
		else
		if (R3Dflags & specialaction_opendoor)
			EV_DoDoor(&junk, openNormal);
		else
		if (R3Dflags & specialaction_lowerfloor)
			EV_DoFloor(&junk, lowerFloorToLowest);
		else
		{
			EV_DoFloor(&junk, lowerFloorToLowest);
			EV_DoDoor(&junk, openNormal);
		}

		junk.tag = 667;
		EV_DoFloor(&junk, raiseToTexture);
	}
	return true;
}

// Edit stuff
static int warpcoordx;
static int warpcoordy;
static int commandset;

void CheckEditWarc()
{
	if (commandset)
	{
		ED_MovePlayer(warpcoordx, warpcoordy);
		commandset = false;
	}
}
	    
int CCmdEditWarc(int argc, char **argv)
{
	int	 x, y;

	if (commandset || con_freezegame || (!R3D_edit && !ge_inf.devmap) || argc != 3) return false;
	
	x = atoi(argv[1]);
	y = atoi(argv[2]);

	if (x >  32767) x =  32767;
	if (x < -32768) x = -32768;
	if (y >  32767) y =  32767;
	if (y < -32768) x = -32768;

	x <<= 16, y <<= 16;

	warpcoordx = x, warpcoordy = y;
	commandset = true;
	CheckEditWarc();

	return true;
}

int CCmdEditWarl(int argc, char **argv)
{
	if (commandset || con_freezegame || (!R3D_edit && !ge_inf.devmap) || argc != 2) return false;
	
	int lnum = atoi(argv[1]);

	if (lnum < 0 || lnum >= numlines)
	{
		P_SetMessage(mbf_players, "WARL: line num out of range");
		return false;
	}

	warpcoordx = meanV(lines[lnum].v1->x, lines[lnum].v2->x);
	warpcoordy = meanV(lines[lnum].v1->y, lines[lnum].v2->y);

	commandset = true;
	CheckEditWarc();

	return true;
}

int CCmdEditWarss(int argc, char **argv)
{
	subsector_t *ss;
	int	 ssnum;

	if (commandset || con_freezegame || (!R3D_edit && !ge_inf.devmap) || argc != 2) return false;
	
	ssnum = atoi(argv[1]);

	if (ssnum < 0 || ssnum >= numsubsectors)
	{
		P_SetMessage(mbf_players, "WARSS: subsector out of range");
		return false;
	}

	ss = &subsectors[ssnum];

	warpcoordx = (int)ss->midpoint.x*FRACUNIT;
	warpcoordy = (int)ss->midpoint.y*FRACUNIT;
	commandset = true;
	CheckEditWarc();

	return true;
}

int CCmdEditWars(int argc, char **argv)
{
	subsector_t *ss, *sstemp = NULL, *ssuse = NULL;
	int	 i, k, snum;

	if (commandset || con_freezegame || (!R3D_edit && !ge_inf.devmap) || argc != 2) return false;
	
	snum = atoi(argv[1]);

	if (snum < 0 || snum >= numsectors)
	{
		P_SetMessage(mbf_players, "WARS: sector out of range");
		return false;
	}

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->sector - sectors == snum)
		{
			seg_t *seg = &segs[ss->firstline];

			if (!sstemp) sstemp = ss;

			for (k = 0; k < ss->linecount; k++)
			{
				if (   seg->linedef
					&& !(seg->linedef->spflags & SPF_CONTROLLINE))
				{
					ssuse = ss; break;
				}
			}
		}
	}

	if (!ssuse) ssuse = sstemp;	// default if all lines are control lines
	if (!ssuse && !sstemp) P_SetMessage(mbf_players, "WARSS: subsector not found");

	warpcoordx = (int)(ssuse->midpoint.x*FRACUNIT);
	warpcoordy = (int)(ssuse->midpoint.y*FRACUNIT);
	commandset = true;
	CheckEditWarc();

	return true;
}
