//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: con_action.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  The Action Manager
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_console.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// Con_ClearActions
//===========================================================================
void Con_ClearActions(void)
{
	action_t *act;
	for(act=actions; act->name[0]; act++) act->on = false;
}

//===========================================================================
// Con_ActionCommand
//	The command begins with a '+' or a '-'.
//	Returns true if the action was changed successfully.
//	If has_prefix is false, the state of the action is negated.
//===========================================================================
int Con_ActionCommand(char *cmd, RDBOOL has_prefix)
{
	char		prefix = cmd[0];
	int			v1, v2;
	char		name8[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	action_t	*act;

	if(actions == NULL) return false;
		
	strncpy(name8, cmd + (has_prefix? 1 : 0), 8);
	v1 = *(int*) name8;
	v2 = *(int*) &name8[4];

	// Try to find a matching action by searching through the list.
	for(act = actions; act->name[0]; act++)
	{
		if(v1 == *(int*) act->name && v2 == *(int*) &act->name[4])
		{
			// This is a match!
			if(has_prefix)
				act->on = prefix=='+'? true : false;
			else
				act->on = !act->on;
			return true;
		}
	}
	return false;
}

int CCmdListActs(int argc, char **argv)
{
	action_t	*act;

	Con_Message("Action commands registered by the game DLL:\n");
	for(act = actions; act->name[0]; act++)
		Con_Message("  %s\n", act->name);
	return true;
}

