//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 Randy Heit, G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dehmain.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 33 (v220-33) 25/Feb/2018 gmj
//
// Based on: Doomsday v1.7.8 dpDehRead.dll code which in turn is based
//           on ZDoom's DEH reader as credited in the Doomsday source.
//
// DESCRIPTION: Dehacked reader for Risen3D
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#pragma optimize("g",off)

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <stddef.h>
#include <direct.h>

#include "de_base.h"
#include "de_system.h"
#include "de_console.h"
#include "de_misc.h"

#include "doom\info.h"	// b211-0
#include "r3d_compat.h"
#include "AcFnLink.h"
#include "r3d_edit.h"
#include "r3d_mapinfo.h"
#include "p_data.h"
#pragma optimize("g", off)

// MACROS ------------------------------------------------------------------

// to report unsupported strings (int sized if more flags needed)
#define	NSR_OB_STR	0x0001
#define	NSR_ST5_STR	0x0002
#define	NSR_PSNDSTR	0x0004


#define OFF_INT		0x01
#define OFF_STATE	0x02
#define OFF_SOUND	0x04
#define OFF_FLOAT	0x08
#define OFF_SPRITE	0x10

#define DH_CODEPSIZE	 448

#define CHECKKEY(a,b)		if (!_stricmp (Line1, (a))) (b) = atoi(Line2);

#define myoffsetof(type,identifier) (((size_t)&((type*)0)->identifier))

// TYPES -------------------------------------------------------------------

typedef struct Key {
	char		*name;
	ptrdiff_t	offset;
	int			type;
} Key_t;

// v220-7 translation from standard strings to changed strings used in the deds
//        Note: PD_REDSK, PD_BLUESK, PD_YELLOWSK in the deds are reserved for
//              use with scripted maps so don't need translation.
typedef struct {
	char *name;
	int	 idx;
} nsub_t;

nsub_t nsub[] = {

	{"GOTREDSKUL",		TXT_GOTREDSKULL },	// typo in original ded
	{"PD_BLUEC",		TXT_PD_BLUEKCA	},	// changed names->
	{"PD_REDC",			TXT_PD_REDKCA	},
	{"PD_YELLOWC",		TXT_PD_YELLOWKCA},
	{"PD_BLUES",		TXT_PD_BLUESKA	},
	{"PD_REDS",			TXT_PD_REDSKA	},
	{"PD_YELLOWS",		TXT_PD_YELLOWSKA},
	{"PD_ANY",			TXT_PD_ANYKEY	},
	{"PD_ALL3",			TXT_PD_THREEKEYS},
	{"PD_ALL6",			TXT_PD_SIXKEYS	},
	{NULL,				0				}
};


// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

char *Def_ModifyTextDef(int texn, char *str);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static int PatchThing (int);
static int PatchSound (int);
static int PatchFrame (int);
static int PatchSprite (int);
static int PatchAmmo (int);
static int PatchWeapon (int);
static int PatchPointer (int);
static int PatchCheats (int);
static int PatchMisc (int);
static int PatchText (int);
static int PatchStrings (int);
static int PatchPars (int);
static int PatchCodePtrs (int);
static int DoInclude (int);
static void ApplyDEH(char *base, int length, int from);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL levelExMxExists;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

uint		miscDEHval[NUMMISCDEHVALS];
uint		flagsDEH[DH_NUMSPRITES];
byte		DEHCheck[DH_CHECKNUM];	// R3D rev95 
ushort		DEHoldTH[DH_CHECKNUM];	// R3D rev95 
int			DEHchange = 0;
RDBOOL		hasDEH = false;
RDBOOL		isarctic;
RDBOOL		dehdebug;

char *SoundMap[] = {
	"None",		// 000
	"pistol",	// 001
	"shotgn",	// 002
	"sgcock",	// 003
	"dshtgn",	// 004
	"dbopn",	// 005
	"dbcls",	// 006
	"dbload",	// 007
	"plasma",	// 008
	"bfg",		// 009
	"sawup",	// 010
	"sawidl",	// 011
	"sawful",	// 012
	"sawhit",	// 013
	"rlaunc",	// 014
	"rxplod",	// 015
	"firsht",	// 016
	"firxpl",	// 017
	"pstart",	// 018
	"pstop",	// 019
	"doropn",	// 020
	"dorcls",	// 021
	"stnmov",	// 022
	"swtchn",	// 023
	"swtchx",	// 024
	"plpain",	// 025
	"dmpain",	// 026
	"popain",	// 027
	"vipain",	// 028
	"mnpain",	// 029
	"pepain",	// 030
	"slop",		// 031
	"itemup",	// 032
	"wpnup",	// 033
	"oof",		// 034
	"telept",	// 035
	"posit1",	// 036
	"posit2",	// 037
	"posit3",	// 038
	"bgsit1",	// 039
	"bgsit2",	// 040
	"sgtsit",	// 041
	"cacsit",	// 042
	"brssit",	// 043
	"cybsit",	// 044
	"spisit",	// 045
	"bspsit",	// 046
	"kntsit",	// 047
	"vilsit",	// 048
	"mansit",	// 049
	"pesit",	// 050
	"sklatk",	// 051
	"sgtatk",	// 052
	"skepch",	// 053
	"vilatk",	// 054
	"claw",		// 055
	"skeswg",	// 056
	"pldeth",	// 057
	"pdiehi",	// 058
	"podth1",	// 059
	"podth2",	// 060
	"podth3",	// 061
	"bgdth1",	// 062
	"bgdth2",	// 063
	"sgtdth",	// 064
	"cacdth",	// 065
	"skldth",	// 066
	"brsdth",	// 067
	"cybdth",	// 068
	"spidth",	// 069
	"bspdth",	// 070
	"vildth",	// 071
	"kntdth",	// 072
	"pedth",	// 073
	"skedth",	// 074
	"posact",	// 075
	"bgact",	// 076
	"dmact",	// 077
	"bspact",	// 078
	"bspwlk",	// 079
	"vilact",	// 080
	"noway",	// 081
	"barexp",	// 082
	"punch",	// 083
	"hoof",		// 084
	"metal",	// 085
	"chgun",	// 086
	"tink",		// 087
	"bdopn",	// 088
	"bdcls",	// 089
	"itmbk",	// 090
	"flame",	// 091
	"flamst",	// 092
	"getpow",	// 093
	"bospit",	// 094
	"boscub",	// 095
	"bossit",	// 096
	"bospn",	// 097
	"bosdth",	// 098
	"manatk",	// 099
	"mandth",	// 100
	"sssit",	// 101
	"ssdth",	// 102
	"keenpn",	// 103
	"keendt",	// 104
	"skeact",	// 105
	"skesit",	// 106
	"skeatk",	// 107
	"radio",	// 108
	"dgsit",	// 109
	"dgatk1",	// 110
	"dgpain",	// 111
	"dgdth",	// 112
	"dgact",	// 113
};


// PRIVATE DATA DEFINITIONS ------------------------------------------------

// Original data will be stored here before it gets replaced by a patch.
static char	OrgSprNames  [DH_NUMSPRITES][ 5];
static char	OrgActionPtrs[DH_NUMSTATES ][40];

static char *unknown_str = "DEH UNKNOWN : at line %i Unknown key %s encountered in %s %d.\n";
static char	*PatchFile, *PatchPt;
static char	*Line1, *Line2;

static int	dversion, pversion;
static int	linecounter = 0;
static int	nsreported = 0;

static BOOL	BackedUpData = false;
static BOOL	including, includenotext;
static BOOL changespritename;
static BOOL	packwarn = false;

// From DeHackEd source.
static int toff[] = {129044, 129044, 129044, 129284, 129380};

// A conversion array to convert from the 448 code pointers
// to the 966 Frames that exist.
// Again taken from the DeHackEd source.
static short codepconv[DH_CODEPSIZE] = {			// DH_CODEPSIZE = 448
   1,   2,   3,   4,   6,   9,  10,  11,  12,  14,
  16,  17,  18,  19,  20,  22,  29,  30,  31,  32,
  33,  34,  36,  38,  39,  41,  43,	 44,  47,  48,
  49,  50,  51,  52,  53,  54,  55,	 56,  57,  58,
  59,  60,  61,  62,  63,  65,  66,	 67,  68,  69,
  70,  71,  72,  73,  74,  75,  76,	 77,  78,  79,
  80,  81,  82,  83,  84,  85,  86,	 87,  88,  89,
 119, 127, 157, 159, 160, 166, 167,	174, 175, 176,
 177, 178, 179, 180, 181, 182, 183,	184, 185, 188,
 190, 191, 195, 196, 207, 208, 209,	210, 211, 212,
 213, 214, 215, 216, 217, 218, 221,	223, 224, 228,
 229, 241, 242, 243, 244, 245, 246,	247, 248, 249,
 250, 251, 252, 253, 254, 255, 256, 257, 258, 259,
 260, 261, 262, 263, 264, 270, 272,	273, 281, 282,
 283, 284, 285, 286, 287, 288, 289,	290, 291, 292,
 293, 294, 295, 296, 297, 298, 299,	300, 301, 302,
 303, 304, 305, 306, 307, 308, 309,	310, 316, 317,
 321, 322, 323, 324, 325, 326, 327,	328, 329, 330,
 331, 332, 333, 334, 335, 336, 337,	338, 339, 340,
 341, 342, 344, 347, 348, 362, 363,	364, 365, 366,
 367, 368, 369, 370, 371, 372, 373, 374, 375, 376,
 377, 378, 379, 380, 381, 382, 383,	384, 385, 387,
 389, 390, 397, 406, 407, 408, 409,	410, 411, 412,
 413, 414, 415, 416, 417, 418, 419,	421, 423, 424,
 430, 431, 442, 443, 444, 445, 446,	447, 448, 449,
 450, 451, 452, 453, 454, 456, 458,	460, 463, 465,
 475, 476, 477, 478, 479, 480, 481,	482, 483, 484,
 485, 486, 487, 489, 491, 493, 502,	503, 504, 505,
 506, 508, 511, 514, 527, 528, 529,	530, 531, 532,
 533, 534, 535, 536, 537, 538, 539,	541, 543, 545,
 548, 556, 557, 558, 559, 560, 561,	562, 563, 564,
 565, 566, 567, 568, 570, 572, 574,	585, 586, 587,
 588, 589, 590, 594, 596, 598, 601,	602, 603, 604,
 605, 606, 607, 608, 609, 610, 611,	612, 613, 614,
 615, 616, 617, 618, 620, 621, 622,	631, 632, 633,
 635, 636, 637, 638, 639, 640, 641,	642, 643, 644,
 645, 646, 647, 648, 650, 652, 653,	654, 659, 674,
 675, 676, 677, 678, 679, 680, 681,	682, 683, 684,
 685, 686, 687, 688, 689, 690, 692,	696, 700, 701,
 702, 703, 704, 705, 706, 707, 708,	709, 710, 711,
 713, 715, 718, 726, 727, 728, 729,	730, 731, 732,
 733, 734, 735, 736, 737, 738, 739,	740, 741, 743,
 745, 746, 750, 751, 766, 774, 777,	779, 780, 783,
 784, 785, 786, 787, 788, 789, 790,	791, 792, 793,
 794, 795, 796, 797, 798, 801, 809,	811,
};

static const struct {
	char *name;
	int (*func)(int);
} Modes[] = {
	// These appear in .deh and .bex files
	{ "Thing",		PatchThing },
	{ "Sound",		PatchSound },
	{ "Frame",		PatchFrame },
	{ "Sprite",		PatchSprite },
	{ "Ammo",		PatchAmmo },
	{ "Weapon",		PatchWeapon },
	{ "Pointer",	PatchPointer },
	{ "Cheat",		PatchCheats },
	{ "Misc",		PatchMisc },
	{ "Text",		PatchText },
	{ "include",	DoInclude },
	{ "[STRINGS]",	PatchStrings },
	{ "[PARS]",		PatchPars },
	{ "[CODEPTR]",	PatchCodePtrs },
	{ NULL, },
};

// CODE --------------------------------------------------------------------

//====================================================================
//
// ADDITIONAL DEH routines
//
//  Code for Array to improve DEH support.
//  Script debugging help is given with
//  the command line switch -dehdebug.
//====================================================================

static char *miscSTRval(int index)
{
	static char buff[20];
	int i = miscDEHval[index];

	if (i == 0x80000000)
		strcpy(buff, "nochange");
	else
		itoa(i, buff, 10);

	return buff;
}

// For DEH script debugging
// v220-3 make 0x80000000 value more meaningful
static void PrintDEHdebug()
{
	pr_OFinsert ("BFGCells %s\n"		, miscSTRval(BFGCells));
	pr_OFinsert ("BlueAC %s\n"			, miscSTRval(BlueAC));
	pr_OFinsert ("FAAC %s\n"			, miscSTRval(FAAC));
	pr_OFinsert ("FAArmor %s\n"			, miscSTRval(FAArmor));
	pr_OFinsert ("GodHealth %s\n"		, miscSTRval(GodHealth));
	pr_OFinsert ("GreenAC %s\n"			, miscSTRval(GreenAC));
	pr_OFinsert ("Infight %s\n"			, miscSTRval(Infight));
	pr_OFinsert ("KFAAC %s\n"			, miscSTRval(KFAAC));
	pr_OFinsert ("KFAArmor %s\n"		, miscSTRval(KFAArmor));
	pr_OFinsert ("MaxArmor %s\n"		, miscSTRval(MaxArmor));
	pr_OFinsert ("MaxHealth %s\n"		, miscSTRval(MaxHealth));
	pr_OFinsert ("MaxSoulsphere %s\n"	, miscSTRval(MaxSoulsphere));
	pr_OFinsert ("MegasphereHealth %s\n", miscSTRval(MegasphereHealth));
	pr_OFinsert ("SoulsphereHealth %s\n", miscSTRval(SoulsphereHealth));
	pr_OFinsert ("StartBullets %s\n"	, miscSTRval(StartBullets));
	pr_OFinsert ("StartHealth %s\n"		, miscSTRval(StartHealth));
}

//====================================================================
typedef struct
{
	char oldSPR[6];
	char newSPR[6];
} sprArray_t;

static int			SPRarraycount = 0;
static int			maxSPRarraycount = 0;
static sprArray_t	*sprArray = NULL;

int CheckDEHThingSPR(char *name)
{
	sprArray_t	*spra;
	int			i;

	if (!hasDEH || !sprArray || !name[4]) return false;

	for (i = 0, spra = sprArray; i < SPRarraycount; i++, spra++)
	{
		if (!strnicmp(name, spra->oldSPR, 4))
		{
			// For DEH sprite name change rejection
#ifdef _DEBUG
			if (dehdebug) pr_OFinsert("REJECTED SPR %8.8s\n", name);
#endif
			return true;
		}
	}
	return false;
}

void ChangeDEHThingSPR(char *name, int num)
{
	sprArray_t	*spra;
	int			i;

	if (   num >= DH_NUMSTATES
		|| !sprArray
		|| strlen(name) < 4)
	{
		return;
	}

	for (i = 0, spra = sprArray; i < SPRarraycount; i++, spra++)
	{
		ded_t *ded = &defs;

		if (!strnicmp(name, spra->oldSPR, 4))
		{
			pr_OFinsert("SPRITE ID %s changed to %s num %i\n", name, spra->newSPR, num);

			strncpy(name, spra->newSPR, 4);
			break;
		}
		break;
	}
}

static void BackupData (void)
{
	int i;

	if (!BackedUpData)
	{
		BackedUpData = true;

		for (i = 0; i < DH_NUMSPRITES; i++)
		{
			// Defensively code in case of defs error
			strncpy(OrgSprNames[i], defs.sprites[i].id, 4);
			OrgSprNames[i][4] = 0;
		}

		int lim = MIN_OF(DH_NUMSTATES, defs.count.states.num);

		for (i = 0; i < lim ; i++)
			strcpy(OrgActionPtrs[i], defs.states[i].action);
	}
}

char		com_token[8192];
BOOL		com_eof;

/*
==============
COM_Parse

Parse a token out of a string
From ZDoom cmdlib.cpp
==============
*/
static char *COM_Parse (char *data)
{
	int		c;
	int		len;
	
	len = 0;
	com_token[0] = 0;
	
	if (!data)
		return NULL;
		
// skip whitespace
skipwhite:
	while ( (c = *data) <= ' ')
	{
		if (c == 0)
		{
			com_eof = true;
			return NULL;			// end of file;
		}
		data++;
	}
	
// skip // comments
	if (c=='/' && data[1] == '/')
	{
		while (*data && *data != '\n')
			data++;
		goto skipwhite;
	}
	
// handle quoted strings specially
	if (c == '\"')
	{
		data++;
		do
		{
			c = *data++;
			if (c=='\"')
			{
				com_token[len] = 0;
				return data;
			}
			com_token[len] = c;
			len++;
		} while (1);
	}

// parse single characters
	if (c=='{' || c=='}' || c==')'|| c=='(' || c=='\'' || c==':' || /*[RH]*/c=='=' || c=='|')
	{
		com_token[len] = c;
		len++;
		com_token[len] = 0;
		return data+1;
	}

// parse a regular word
	do
	{
		com_token[len] = c;
		data++;
		len++;
		c = *data;
		if (c=='{' || c=='}' || c==')' || c=='(' || c=='\'' || c==':' || c=='=')
			break;
	} while (c>32);
	
	com_token[len] = 0;
	return data;
}

static RDBOOL IsNum(char *str)
{
	char *end;
	int val = strtol(str, &end, 0);	// sets end
	if(*end && !isspace(*end)) return false;
	return true;
}

static BOOL HandleKey (const struct Key *keys, void *structure, const char *key, int value)
{
	while(keys->name && _stricmp (keys->name, key))
		keys++;

	if (keys->name)
	{
		int		off = keys->offset;
		void	*ptr = (void*) (((byte*)structure) + off);
		int		type = keys->type;

		// Apply value.
		if (type == OFF_STATE)
		{
			if (dehdebug)
				pr_OFinsert("[%s] OFF_STATE val %i deh id %s ptr %s\n", keys->name, value, defs.states[value].id, (char*)ptr);

			strcpy((char*)ptr, defs.states[value].id);

			if (!ge_inf.amwds) DEHCheck[value] |= DEHF_STATE;	// v220-4
		}
		else if(type == OFF_SPRITE)
		{
			if (value >= 0 && value < DH_NUMSPRITES)
			{
				if (dehdebug)
					pr_OFinsert("OFF_SPRITE val %i deh id %s\n", value, defs.states[value].id);

				strcpy((char*)ptr, defs.sprites[value].id);
				DEHCheck[value] |= DEHF_SPRITE;
			}
			else
				pr_OFinsert("DEH HandleKey line %i: OFF_SPRITE %i out of range\n", linecounter, value);
		}
		else if (type == OFF_SOUND)
		{
			if (value >= 0 && value < defs.count.sounds.num)
			{
				if (dehdebug)
					pr_OFinsert("[%s] OFF_SOUND val %i defid %s\n", keys->name, value, defs.sounds[value].id);

				strcpy((char*)ptr, defs.sounds[value].id);
			}
			else
				pr_OFinsert("[%s] DEH HandleKey line %i: OFF_SOUND %i out of range\n", keys->name, linecounter, value);
		}
		else if (type == OFF_FLOAT)
		{
			if (value < 0x2000)
				*(float*) ptr = (float) value;// / (float) 0x10000;
			else
				*(float*) ptr = value / (float) 0x10000;

			if (dehdebug)
				pr_OFinsert("[%s] OFF_FLOAT ptr %f\n", keys->name, *(float*) ptr);
		}
		else if (type == OFF_INT)
		{
			*(int*)ptr = value;

			if (dehdebug)
				pr_OFinsert("[%s] OFF_INT val %i\n", keys->name, value);
		}
		return false;
	}

	return true;
}

static BOOL ReadChars (char **stuff, int size)
{
	char *str = *stuff;

	if (!size) {
		*str = 0;
		return true;
	}

	do {
		// Ignore carriage returns
		if (*PatchPt != '\r')
			*str++ = *PatchPt;
		else
			size++;


		if (*PatchPt == '\n')
			linecounter++;

		if (*PatchPt)
			PatchPt++;
		else
			break;

	} while (--size);

	*str = 0;
	return true;
}

static char *skipwhite (char *str)
{
	if (str)
	{
		while (*str && isspace(*str))
			str++;
	}
	return str;
}

static void stripwhite (char *str)
{
	char *end = str + strlen(str) - 1;

	while (end >= str && isspace(*end))
		end--;

	end[1] = '\0';
}

static char *igets (void)
{
	char *line;

	if (*PatchPt == '\0') return NULL;	// EOF

	line = PatchPt;

	while (*PatchPt != '\n' && *PatchPt != '\0')
		PatchPt++;

	if (*PatchPt == '\n')
	{
		*PatchPt++ = 0;
		linecounter++;
	}

	return line;
}

static int GetLine (void)
{
	char *line, *line2;

	do {
		while ( (line = igets ()) )
		{
			if (line[0] != '#')	break;	// Skip comment lines
		}
				
		if (!line) return 0;

		Line1 = skipwhite (line);

	} while (Line1 && *Line1 == 0);	// Loop until we get a line with
									// more than just whitespace.
	line = strchr (Line1, '=');

	if (line) {					// We have an '=' in the input line
		line2 = line;
		while (--line2 >= Line1)
			if (*line2 > ' ')
				break;

		if (line2 < Line1)
			return 0;			// Nothing before '='

		*(line2 + 1) = 0;

		line++;
		while (*line && *line <= ' ')
			line++;

		if (*line == 0)
			return 0;			// Nothing after '='

		Line2 = line;

		return 1;
	} else {					// No '=' in input line
		line = Line1 + 1;
		while (*line > ' ')
			line++;				// Get beyond first word

		*line++ = 0;
		while (*line && *line <= ' ')
			line++;				// Skip white space

		//.bex files don't have this restriction
		//if (*line == 0)
		//	return 0;			// No second word

		Line2 = line;

		return 2;
	}
}

static int PatchNothing()
{
	int result;

	while ((result = GetLine ()) == 1);

	return result;
}

static int HandleMode(const char *mode, int num)
{
	int i = 0;

	while (Modes[i].name && _stricmp (Modes[i].name, mode))
		i++;

	if (Modes[i].name)
		return Modes[i].func (num);

	// Handle unknown or unimplemented data
	pr_OFinsert("DEH INFO  : at line %i chunk \"%s\" is not supported\n", linecounter, mode);

	return PatchNothing();
}

// v220-9 some tidying up done
static int PatchThing (int thingy)
{
	static const struct Key keys[] = {
		{ "Initial frame",		myoffsetof(ded_mobj_t,spawnstate),	OFF_STATE	},
		{ "First moving frame",	myoffsetof(ded_mobj_t,seestate),	OFF_STATE	},
		{ "Injury frame",		myoffsetof(ded_mobj_t,painstate),	OFF_STATE	},
		{ "Close attack frame",	myoffsetof(ded_mobj_t,meleestate),	OFF_STATE	},
		{ "Far attack frame",	myoffsetof(ded_mobj_t,missilestate),OFF_STATE	},
		{ "Death frame",		myoffsetof(ded_mobj_t,deathstate),	OFF_STATE	},
		{ "Exploding frame",	myoffsetof(ded_mobj_t,xdeathstate),	OFF_STATE	},
		{ "Respawn frame",		myoffsetof(ded_mobj_t,raisestate),	OFF_STATE	},

		{ "Alert sound",		myoffsetof(ded_mobj_t,seesound),	OFF_SOUND	},
		{ "Attack sound",		myoffsetof(ded_mobj_t,attacksound),	OFF_SOUND	},
		{ "Pain sound",			myoffsetof(ded_mobj_t,painsound),	OFF_SOUND	},
		{ "Death sound",		myoffsetof(ded_mobj_t,deathsound),	OFF_SOUND	},
		{ "Action sound",		myoffsetof(ded_mobj_t,activesound),	OFF_SOUND	},

		{ "Speed",				myoffsetof(ded_mobj_t,speed),		OFF_FLOAT	},
		{ "Width",				myoffsetof(ded_mobj_t,radius),		OFF_FLOAT	},
		{ "Height",				myoffsetof(ded_mobj_t,height),		OFF_FLOAT	},

		{ "ID #",				myoffsetof(ded_mobj_t,doomednum),	OFF_INT		},
		{ "Reaction time",		myoffsetof(ded_mobj_t,reactiontime),OFF_INT		},
		{ "Pain chance",		myoffsetof(ded_mobj_t,painchance),	OFF_INT		},
		{ "Hit points",			myoffsetof(ded_mobj_t,spawnhealth),	OFF_INT		},
		{ "Mass",				myoffsetof(ded_mobj_t,mass),		OFF_INT		},
		{ "Missile damage",		myoffsetof(ded_mobj_t,damage),		OFF_INT		},
		//{ "Translucency",		myoffsetof(ded_mobj_t,translucency),OFF_INT		},
		{ NULL, }
	};

	// flags can be specified by name (a .bex extension):
	static const struct {
		short bit;
		short whichflags;
		const char *name;
	} bitnames[] = {
		{ 0, 0, "SPECIAL"},
		{ 1, 0, "SOLID"},
		{ 2, 0, "SHOOTABLE"},
		{ 3, 0, "NOSECTOR"},
		{ 4, 0, "NOBLOCKMAP"},
		{ 5, 0, "AMBUSH"},
		{ 6, 0, "JUSTHIT"},
		{ 7, 0, "JUSTATTACKED"},
		{ 8, 0, "SPAWNCEILING"},
		{ 9, 0, "NOGRAVITY"},
		{10, 0, "DROPOFF"},
		{11, 0, "PICKUP"},
		{12, 0, "NOCLIP"},
		{14, 0, "FLOAT"},
		{15, 0, "TELEPORT"},
		{16, 0, "MISSILE"},
		{17, 0, "DROPPED"},
		{18, 0, "SHADOW"},
		{19, 0, "NOBLOOD"},
		{20, 0, "CORPSE"},
		{21, 0, "INFLOAT"},
		{22, 0, "COUNTKILL"},
		{23, 0, "COUNTITEM"},
		{24, 0, "SKULLFLY"},
		{25, 0, "NOTDMATCH"},

		// Only BOOM types supported
		{26, 0, "TRANSLATION"},		
		{27, 0, "UNUSED1"},			// BOOM compatibility
		{28, 0, "TOUCHY"},			// BOOM compatibility
		{29, 0, "BOUNCES"},			// BOOM compatibility
		{30, 0, "FRIEND"},			// BOOM compatibility
		{31, 0, "TRANSLUCENT"},		// BOOM compatibility?

		// ZDoom types ignored
		{-1, 0, "TRANSLATION1"},
		{-1, 0, "TRANSLATION2"},
		{-1, 0, "STEALTH"},

		// Not directly supported - treat as translucent
		{31, 0, "TRANSLUC50"},
		{31, 0, "TRANSLUC25"},
		{31, 0, "TRANSLUC75"},

		// Names for flags2
		{ 0, 1, "LOGRAV"},
		{ 1, 1, "WINDTHRUST"},
		{ 2, 1, "FLOORBOUNCE"},
		{ 3, 1, "BLASTED"},
		{ 4, 1, "FLY"},
		{ 5, 1, "FLOORCLIP"},
		{ 6, 1, "SPAWNFLOAT"},
		{ 7, 1, "NOTELEPORT"},
		{ 8, 1, "RIP"},
		{ 9, 1, "PUSHABLE"},
		{10, 1, "CANSLIDE"},			// Avoid conflict with SLIDE from BOOM
		{11, 1, "ONMOBJ"},
		{12, 1, "PASSMOBJ"},
		{13, 1, "CANNOTPUSH"},
		{14, 1, "DROPPED"},
		{15, 1, "BOSS"},
		{16, 1, "FIREDAMAGE"},
		{17, 1, "NODMGTHRUST"},
		{18, 1, "TELESTOMP"},
		{19, 1, "FLOATBOB"},
		{20, 1, "DONTDRAW"},
		{21, 1, "IMPACT"},
		{22, 1, "PUSHWALL"},
		{23, 1, "MCROSS"},
		{24, 1, "PCROSS"},
		{25, 1, "CANTLEAVEFLOORPIC"},
		{26, 1, "NONSHOOTABLE"},
		{27, 1, "INVULNERABLE"},
		{28, 1, "DORMANT"},
		{29, 1, "ICEDAMAGE"},
		{30, 1, "SEEKERMISSILE"},
		{31, 1, "REFLECTIVE"}
	};

	ded_mobj_t *info;

	int		result, thingNum = thingy;

	//---------------------------------------------

	if (thingNum < 1 || thingNum >= DH_NUMOBJECTS)
	{
		pr_OFinsert("DEH ERROR : at line %i Thing %d out of range.\n",
			linecounter, thingNum);
		while ((result = GetLine ()) == 1);
		return result;
	}

	thingNum--;

	DEHchange |= DEHF_THING;	// Thing change detected

	DEHCheck[thingNum] |= DEHF_THING;	// rev95
	hasDEHactors = true;
	info = defs.mobjs + thingNum;

	if (info->weapon_offset) info->weapon_offset = 0;	// v220-4 clear any value if dehacked

	if(dehdebug) pr_OFinsert("Thing %d\n", thingNum + 1);

	while ((result = GetLine ()) == 1) 
	{
		size_t sndmap = atoi (Line2);

		if (sndmap >= sizeof(SoundMap))
			sndmap = 0;

		if (HandleKey (keys, info, Line1, atoi (Line2))) 
		{
			if (!_stricmp (Line1, "Bits"))
			{
				int value = 0, value2 = 0;
				BOOL vchanged = false, v2changed = false;
				char *strval;

				for (strval = Line2; (strval = strtok (strval, ",+| \t\f\r")); strval = NULL)
				{
					size_t iy;

					if (IsNum (strval))
					{
						// Force the top 4 bits to 0 so that the user is forced
						// to use the mnemonics to change them. ZDoom

						// b211-0 flags allowed - eden.wad, written in may 2007,
						// uses bits, not the mnemonics. If flags end up being set at
						// 0xFFFFFFFF then R3D knows that they have not been changed
						value |= (atoi(strval) & 0xffffffff);
						vchanged = true;
					}
					else
					{
						for (iy = 0; iy < sizeof(bitnames)/sizeof(bitnames[0]); iy++)
						{
							if (!_stricmp (strval, bitnames[iy].name))
							{
								if (bitnames[iy].bit == -1)
								{
									pr_OFinsert("DEH INF) : at line %i %s is not supported.\n", linecounter, strval);
								}
								else
								{
									if (bitnames[iy].whichflags)
									{
										v2changed = true;
										if (bitnames[iy].bit & 0xff00)
											value2 |= 1 << (bitnames[iy].bit >> 8);
										value2 |= 1 << (bitnames[iy].bit & 0xff);
									}
									else
									{
										vchanged = true;
										if (bitnames[iy].bit & 0xff00)
											value |= 1 << (bitnames[iy].bit >> 8);
										value |= 1 << (bitnames[iy].bit & 0xff);
									}
								}

								break;
							}
						}

						if (iy >= sizeof(bitnames)/sizeof(bitnames[0]))
							pr_OFinsert("DEH ERROR : at line %i Unknown bit mnemonic %s\n",
							linecounter, strval);
					}
				}

				if (vchanged) flagsDEH[thingNum] = value;
				if (v2changed) pr_OFinsert("DEH INFO : at line %i flags2 are not supported.\n", linecounter);

				{
					int dmednum = (defs.mobjs + thingNum)->doomednum;

					DEHoldTH[thingNum] = (ushort)dmednum;

					if(dehdebug)
						pr_OFinsert("Bits [vchanged %i] for thingNum %4i dmednum %5i: %10i (0x%.8X)\n",vchanged,thingNum+1, dmednum, value, value);
				}
			}
			else
				pr_OFinsert(unknown_str, linecounter, Line1, "Thing", thingNum + 1);
		} 
	}

	return result;
}

static int PatchSound (int soundNum)
{
	nsreported |= NSR_PSNDSTR;
	return PatchNothing();
/*
	sfxinfo_t *info, dummy;
	int offset = 0;
	int result;
	if (soundNum >= 1 && soundNum <= NUMSFX) {
		info = &S_sfx[soundNum];
	} else {
		info = &dummy;
		Printf ("Sound %d out of range.\n");
	}

	while ((result = GetLine ()) == 1) {
		if (!_stricmp  ("Offset", Line1))
			offset = atoi (Line2);
		else CHECKKEY ("Zero/One",			info->singularity)
		else CHECKKEY ("Value",				info->priority)
		else CHECKKEY ("Zero 1",			info->link)
		else CHECKKEY ("Neg. One 1",		info->pitch)
		else CHECKKEY ("Neg. One 2",		info->volume)
		else CHECKKEY ("Zero 2",			info->data)
		else CHECKKEY ("Zero 3",			info->usefulness)
		else CHECKKEY ("Zero 4",			info->lumpnum)
		else Printf (unknown_str, linecounter, Line1, "Sound", soundNum);
	}

	if (offset) {
		// Calculate offset from start of sound names
		offset -= toff[dversion] + 21076;

		if (offset <= 64)			// pistol .. bfg
			offset >>= 3;
		else if (offset <= 260)		// sawup .. oof
			offset = (offset + 4) >> 3;
		else						// telept .. skeatk
			offset = (offset + 8) >> 3;

		if (offset >= 0 && offset < NUMSFX) {
			S_sfx[soundNum].name = OrgSfxNames[offset + 1];
		} else {
			Printf ("Sound name %d out of range.\n", offset + 1);
		}
	}

	return result;*/
}

static int PatchFrame (int frameNum)
{
	static struct Key keys[] = {
		{ "Sprite number",		myoffsetof(ded_state_t, sprite),	OFF_SPRITE	},
		{ "Sprite subnumber",	myoffsetof(ded_state_t, frame),		OFF_INT		},
		{ "Duration",			myoffsetof(ded_state_t, tics),		OFF_INT		},
		{ "Next frame",			myoffsetof(ded_state_t, nextstate),	OFF_STATE	},

		{ "Unknown 1",			0,									OFF_INT		},
		{ "Unknown 2",			0,									OFF_INT		},
		{ NULL, }
	};
	int result;
	ded_state_t *info;

	// C doesn't allow non-constant initializers.
	keys[4].offset = myoffsetof(ded_state_t, misct[0]);
	keys[5].offset = myoffsetof(ded_state_t, misct[1]);

	if (frameNum >= 0 && frameNum < DH_NUMSTATES) 
	{
		info = defs.states + frameNum;
		DEHchange |= DEHF_STATE;	// State change detected
		if (dehdebug)
			pr_OFinsert("Frame %i %s: ", frameNum, info->sprite.id);
	} 
	else 
	{
		pr_OFinsert("DEH ERROR : at line %i Frame %d out of range. limit %i\n",
			linecounter, frameNum, DH_NUMSTATES);
		while ((result = GetLine ()) == 1);
		return result;
	}

	while ((result = GetLine ()) == 1)
	{
		if (HandleKey (keys, info, Line1, atoi (Line2)))
		{
			pr_OFinsert (unknown_str, linecounter, Line1, "Frame", frameNum);
		}
	}

	return result;
}

static int PatchSprite (int sprNum)
{
	int result;
	int offset = 0;

	if (sprNum >= 0 && sprNum < DH_NUMSPRITES) 
	{
		// Don't patch the pistol if in edit mode
		if (R3D_edit && (sprNum == SPR_PISG || sprNum == SPR_PISF))
		{
			return PatchNothing();
		}

		if(dehdebug) pr_OFinsert ("Sprite %d\n", sprNum);
	} 
	else 
	{
		pr_OFinsert ("DEH ERROR : at line %i Sprite %d out of range\n",
			linecounter, sprNum);
		sprNum = -1;
	}

	while ((result = GetLine ()) == 1)
	{
		if (!_stricmp ("Offset", Line1))
			offset = atoi (Line2);
		else 
			pr_OFinsert (unknown_str, linecounter, Line1, "Sprite", sprNum);
	}

	if (offset > 0 && sprNum != -1)
	{
		// Calculate offset from beginning of sprite names.
		int calcoffset = (offset - toff[dversion] - 22044) / 8;

		if (calcoffset >= 0 && calcoffset < DH_NUMSPRITES) 
		{
			if(dehdebug) pr_OFinsert("Sprite num %i name %4.4s replaced with %4.4s\n",
				sprNum, defs.sprites[sprNum].id, OrgSprNames[calcoffset]);
			strncpy(defs.sprites[sprNum].id, OrgSprNames[offset], 4);
		} 
		else 
		{
			pr_OFinsert("DEH ERROR : At line %i Sprite num %i name %4.4s offset %d out of range.\n",
				linecounter, sprNum, OrgSprNames[sprNum], calcoffset);
		}
	}

	return result;
}

//===========================================================================
// DD_Realloc
//===========================================================================
static void *DD_AllocNew(void *ptr, int newsize)
{
	ded_count_t cnt;
	
	cnt.max = 0;
	cnt.num = newsize;
	DED_NewEntries(&ptr, &cnt, 1, 0);
	return ptr;
}

static void SetValueStr(const char *path, const char *id, char *str)
{
	int i;
	char realid[300];

	sprintf(realid, "%s|%s", path, id);
	
	for(i=0; i<defs.count.values.num; i++)
	{
		if(!_stricmp(defs.values[i].id, realid))
		{
			defs.values[i].text = (char*) DD_AllocNew(defs.values[i].text,
				strlen(str) + 1);
			strcpy(defs.values[i].text, str);
			return;
		}
	}
	
	// Not found, create a new Value.
	i = DED_AddValue(&defs, realid);

	// Must allocate memory using DD_AllocNew.
	defs.values[i].text = 0;
	defs.values[i].text = (char*) DD_AllocNew(defs.values[i].text, strlen(str) + 1);
	strcpy(defs.values[i].text, str);
}

static void SetValueInt(const char *path, const char *id, int val)
{
	char buf[20];
	
	itoa(val, buf, 10);
	SetValueStr(path, id, buf);
}

static int PatchAmmo (int ammoNum)
{
	int result, max, per;
	char *ammostr[6] = { "Clip", "Shell", "Cell", "Misl" };
	char *theAmmo = NULL;

	if (R3D_edit)
	{
		return PatchNothing();
	}

	if(ammoNum >= 0 && ammoNum < 4)
	{
		theAmmo = ammostr[ammoNum];
	}
	else
		pr_OFinsert("DEH ERROR : at line %i Ammo %d out of range.\n",
			linecounter, ammoNum);

	while ((result = GetLine ()) == 1) 
	{
		max = per = -1;
		CHECKKEY("Max ammo", max)
		else CHECKKEY("Per ammo", per)
		else pr_OFinsert(unknown_str, linecounter, Line1, "Ammo", ammoNum);
		if(!theAmmo) continue;

		if(max != -1)
		{
			SetValueInt("Player|Max ammo", theAmmo, max);
			if(dehdebug) pr_OFinsert("Ammo %d - max %d\n", ammoNum, max);
		}
		if(per != -1)
		{
			SetValueInt("Player|Clip ammo", theAmmo, per);
			if(dehdebug) pr_OFinsert("Ammo %d - per %d\n", ammoNum, per);
		}
	}

	return result;
}

static int PatchWeapon (int weapNum)
{
	char *ammotypes[] = { "clip", "shell", "cell", "misl", "-", "noammo", 0 };	
	char buf[80];
	int val;
	int result;

	if (R3D_edit)
	{
		return PatchNothing();
	}

	if(weapNum >= 0)
	{
		if(dehdebug) pr_OFinsert("Weapon %d\n", weapNum);
		sprintf(buf, "Weapon Info|%d", weapNum);
	}
	else
	{
		pr_OFinsert("DEH ERROR : at line %i Weapon %d out of range.\n",
			linecounter, weapNum);
		return PatchNothing();
	}

	while ((result = GetLine ()) == 1)
	{
		val = atoi(Line2);
			
		if(!_stricmp(Line1, "Ammo type"))
			SetValueStr(buf, "Type", ammotypes[val]);
		else if(!_stricmp(Line1, "Deselect frame"))
			SetValueStr(buf, "Up", defs.states[val].id);
		else if(!_stricmp(Line1, "Select frame"))
			SetValueStr(buf, "Down", defs.states[val].id);
		else if(!_stricmp(Line1, "Bobbing frame"))
			SetValueStr(buf, "Ready", defs.states[val].id);
		else if(!_stricmp(Line1, "Shooting frame"))
			SetValueStr(buf, "Atk", defs.states[val].id);
		else if(!_stricmp(Line1, "Firing frame"))
			SetValueStr(buf, "Flash", defs.states[val].id);
		else
			pr_OFinsert(unknown_str, linecounter, Line1, "Weapon", weapNum);
	}
	return result;
}

static int PatchPointer (int ptrNum)
{
	int result;

	if (R3D_edit)
	{
		return PatchNothing();
	}

	if (ptrNum < 0 || ptrNum >= DH_CODEPSIZE) 
	{
		pr_OFinsert ("DEH ERROR : at line %i Pointer %d out of range.\n",
			linecounter, ptrNum);
		ptrNum = -1;
	}

	while ((result = GetLine ()) == 1) 
	{
		if ((ptrNum != -1) && (!_stricmp (Line1, "Codep Frame")))
		{
			int nptr = atoi(Line2);

			//states[codepconv[ptrNum]].action = OrgActionPtrs[atoi (Line2)];
			if (nptr >= 0 && nptr < DH_NUMSTATES)	// Insurance
			{
				if (!ge_inf.amwds) DEHCheck[nptr] |= DEHF_STATE;	// v220-4

				strcpy(defs.states[codepconv[ptrNum]].action, 
					OrgActionPtrs[nptr]);

				if (dehdebug)
					pr_OFinsert("Pointer %i conversion %i : %s\n", nptr, codepconv[ptrNum], OrgActionPtrs[nptr]);
			}
			else 
				pr_OFinsert("DEH ERROR : at line %i Codep Frame number %i out of range.\n",
					linecounter, nptr);
		}
		else 
			pr_OFinsert(unknown_str, linecounter, Line1, "Pointer", ptrNum);
	}
	return result;
}

static int PatchCheats (int dummy)
{
/*	static const struct {
		char *name;
		byte *cheatseq;
		BOOL needsval;
	} keys[] = {
		{ "Change music",		cheat_mus_seq,				 true },
		{ "Chainsaw",			cheat_choppers_seq,			 false },
		{ "God mode",			cheat_god_seq,				 false },
		{ "Ammo & Keys",		cheat_ammo_seq,				 false },
		{ "Ammo",				cheat_ammonokey_seq,		 false },
		{ "No Clipping 1",		cheat_noclip_seq,			 false },
		{ "No Clipping 2",		cheat_commercial_noclip_seq, false },
		{ "Invincibility",		cheat_powerup_seq[0],		 false },
		{ "Berserk",			cheat_powerup_seq[1],		 false },
		{ "Invisibility",		cheat_powerup_seq[2],		 false },
		{ "Radiation Suit",		cheat_powerup_seq[3],		 false },
		{ "Auto-map",			cheat_powerup_seq[4],		 false },
		{ "Lite-Amp Goggles",	cheat_powerup_seq[5],		 false },
		{ "BEHOLD menu",		cheat_powerup_seq[6],		 false },
		{ "Level Warp",			cheat_clev_seq,				 true },
		{ "Player Position",	cheat_mypos_seq,			 false },
		{ "Map cheat",			cheat_amap_seq,				 false },
		{ NULL, }
	};
	int result;

	DPrintf ("Cheats\n");

	while ((result = GetLine ()) == 1) {
		int i = 0;
		while (keys[i].name && _stricmp (keys[i].name, Line1))
			i++;

		if (!keys[i].name)
			Printf (PRINT_HIGH, "Unknown cheat %s.\n", Line2);
		else
			ChangeCheat (Line2, keys[i].cheatseq, keys[i].needsval);
	}
	return result;*/

	pr_OFinsert("DEH INFO  : at line %i Cheat patches are not supported.\n", linecounter);
	return PatchNothing();
}

static int PatchMisc (int dummy)
{
	// The defaults shown in red will be loaded from the ded file.
	// Initially all the miscDEHvals are set to 0x80000000.
	// If any are changed here then they will override the ded values.
	static const struct Key keys[] = {
		{ "Initial Health",		StartHealth		 }, //100
		{ "Initial Bullets",	StartBullets	 }, // 50
		{ "Max Health",			MaxHealth		 }, //100
		{ "Max Armor",			MaxArmor		 }, //200
		{ "Basic Armor",		GreenAC			 }, //  1	Currently not used
		{ "Green Armor Class",	GreenAC			 },	//		Currently not used
		{ "Mega Armor",			BlueAC			 }, //  2	Currently not used
		{ "Blue Armor Class",	BlueAC			 },	//		Currently not used
		{ "Max Soulsphere",		MaxSoulsphere	 }, //200
		{ "Soulsphere Health",	SoulsphereHealth }, //100
		{ "Megasphere Health",	MegasphereHealth }, //200
		{ "God Mode Health",	GodHealth		 }, //100
		{ "IDFA Armor",			FAArmor			 }, //200	Currently not used
		{ "IDFA Armor Class",	FAAC			 }, //  2	Currently not used
		{ "IDKFA Armor",		KFAArmor		 }, //200
		{ "IDKFA Armor Class",	KFAAC			 }, //  2	Currently not used
		{ "BFG Cells/Shot",		BFGCells		 }, // 40
		{ "Monsters Infight",	Infight			 }, //  0
		{ NULL, }
	};

	int		i, result;

	while ((result = GetLine()) == 1)
	{
		int found = false;

		for (i = 0; keys[i].name; i++)
		{
			if (!_stricmp(keys[i].name, Line1))
			{
				miscDEHval[keys[i].offset] = atoi(Line2);
				found = true;
				break;
			}
		}
		if (!found)
			pr_OFinsert("DEH ERROR : at line %i Unknown miscellaneous info %s.\n",
				linecounter, Line1);
	}

	return result;
}

static int PatchPars (int dummy)
{
	char *space, mapname[8], *moredata;
	int result, par, i;

	if(dehdebug) pr_OFinsert ("[Pars]\n");

	while ( (result = GetLine()) ) 
	{
		// Argh! .bex doesn't follow the same rules as .deh
		if (result == 1)
		{
			pr_OFinsert ("DEH ERROR : at line %i Unknown key in [PARS] section: %s\n",
				linecounter, Line1);
			continue;
		}
		if (_stricmp ("par", Line1))
			return result;

		space = strchr (Line2, ' ');

		if (!space) 
		{
			pr_OFinsert ("DEH ERROR : at line %i Need data after par.\n", linecounter);
			continue;
		}

		*space++ = '\0';

		while (*space && isspace(*space))
			space++;

		moredata = strchr (space, ' ');

		if (moredata) 
		{
			// At least 3 items on this line, must be E?M? format
			sprintf (mapname, "E%cM%c", *Line2, *space);
			par = atoi (moredata + 1);
		} 
		else 
		{
			// Only 2 items, must be MAP?? format
			sprintf (mapname, "MAP%02d", atoi(Line2) % 100);
			par = atoi (space);
		}

		for(i = 0; i < defs.count.mapinfo.num; i++)
		{
			if (!_stricmp(defs.mapinfo[i].id, mapname))
			{
				ded_mapinfo_t *info = defs.mapinfo + i;

				info->partime = par;
				if(dehdebug) pr_OFinsert("At line %i : Par for %s changed to %d\n", linecounter, mapname, par);
				break;
			}
		}
	}
	return result;
}

static int PatchCodePtrs (int dummy)
{
	int result;

	if (R3D_edit)
	{
		return PatchNothing();
	}

	while ((result = GetLine()) == 1)
	{
		if (!strnicmp ("Frame", Line1, 5) && isspace(Line1[5])) 
		{
			int frame = atoi (Line1 + 5);

			if (frame < 0 || frame >= DH_NUMSTATES) 
			{
				pr_OFinsert("DEH ERROR : at line %i Frame %d out of range. limit %i\n",
					linecounter, frame, DH_NUMSTATES);
			} 
			else 
			{
				int i = 0;
				char *data = COM_Parse (Line2);
				char buff[32];

				DEHCheck[frame] |= DEHF_STATE;

				strcpy(buff, "A_");

				if ((com_token[0] == 'A' || com_token[0] == 'a') && com_token[1] == '_')
					data = com_token + 2;
				else
					data = com_token;

				strcat(buff, com_token);

				for (i = 0; i < nmEnd; i++)
				{
					if (!_stricmp(actionlinks[i].name, buff))
					{
						strcpy(defs.states[frame].action, buff);
						if (dehdebug) pr_OFinsert("CodePtr %i %s:%s\n", frame, actionlinks[i].name, com_token);
						break;
					}
				}
				if (!actionlinks[i].name)
					pr_OFinsert("DEH ERROR: Unknown CodePtr action %s\n", buff);
			}
		}
	}
	return result;
}

//===========================================================================
// ReplaceInString
//===========================================================================
void ReplaceInString(char *str, char *occurance, char *replacewith)
{
	if (strlen(str))
	{
		char *buf = (char*)M_Calloc(strlen(str) * 2, 205);
		char *out = buf, *in = str;
		int oclen = strlen(occurance), replen = strlen(replacewith);
		
		for(; *in; in++)
		{
			if (*in == 0xd) continue;

			if (!strncmp(in, occurance, oclen))
			{
				strcat(out, replacewith);
				out += replen;
				in += oclen-1;
			}
			else *out++ = *in;
		}
		strcpy(str, buf);
		free(buf);
	}
	else
		Con_ErrorD("ReplaceInString: sent empty string");
}

// v220-7 dehacked compatibility
static int PatchText(int oldSize)
{
	int		newSize;
	char	*oldStr;
	char	*newStr;
	char	*temp;
	BOOL	good;
	int		result;
	int		i, k, l1, l2;
	uint	n;
	char	buf[30];

	temp = COM_Parse(Line2);		// Skip old size, since we already have it

	if (!COM_Parse(temp)) 
	{
		pr_OFinsert ("DEH ERROR : at line %i Text chunk is missing size of new string.\n",
			linecounter);
		return 0;
	}
	newSize = atoi (com_token);

	oldStr = (char*)calloc(1, oldSize + 1);
	newStr = (char*)calloc(1, newSize + 1);

	if (!oldStr || !newStr) 
	{
		pr_OFinsert ("Out of memory.\n");
		goto donewithtext;
	}

	good  = ReadChars (&oldStr, oldSize);
	good += ReadChars (&newStr, newSize);

	if (!good) 
	{
		free(newStr);
		free(oldStr);
		pr_OFinsert ("Unexpected end-of-file.\n");
		return 0;
	}

	if (includenotext) 
	{
		pr_OFinsert ("Skipping text chunk in included patch.\n");
		goto donewithtext;
	}

	good = false;

	if (!dehdebug && strlen(newStr) > 200 && strstr(newStr, "HacX, do NOT"))
	{
		pr_OFinsert("\n%s\n", newStr);
	}

	// v220-4 Search through flat names. Only doom/doom2 flats can be substituted
	if (   strlen(oldStr) == 7
		&& (n = strlen(newStr)) > 4 && n < 9
		&& (l1 = W_CheckNumForName(oldStr)) >= 0
		&& (l2 = W_CheckNumForName(newStr)) >= 0
		&& DoomLumpIdentify(l1, NULL, NULL, DLI_GRAPHIC) == LMP_FLAT
		&& DoomLumpIdentify(l2, NULL, NULL, DLI_GRAPHIC) == LMP_FLAT)
	{
		char id[6];

		//  Check background flats (MFLR8_x)
		if (   !_stricmp("MFLR8_1", oldStr)
			|| !_stricmp("MFLR8_2", oldStr)
			|| !_stricmp("MFLR8_3", oldStr)
			|| !_stricmp("MFLR8_4", oldStr))
		{
			// Replace all maps' bexbgflat in episode
			for (i = 1; i < 10; i++)
			{
				sprintf(id, "E%cM%i", oldStr[6], i);

				for (k = 0; k < defs.count.mapinfo.num; k++)
				{
					if (!_stricmp(defs.mapinfo[k].id, id))
					{
						if(dehdebug) pr_OFinsert ("USE deh BACKGROUND FLAT %s in place of %s\n", newStr, oldStr);
						strcpy(defs.mapinfo[k].bexbgflat, newStr);
						break;
					}
				}
			}
			good = true;
		}
		else
		{
			// v220-13 support Doom2 .deh background flats
			id[0] = 0;

			// Check Doom2 flats
			if (!stricmp(oldStr, "SLIME16")) strcpy(id, "MAP07");
			else
			if (!stricmp(oldStr, "RROCK14")) strcpy(id, "MAP12");
			else
			if (!stricmp(oldStr, "RROCK07")) strcpy(id, "MAP21");
			else
			if (!stricmp(oldStr, "RROCK17")) strcpy(id, "MAP30");
			else
			if (!stricmp(oldStr, "RROCK13")) strcpy(id, "MAP31");
			else
			if (!stricmp(oldStr, "RROCK19")) strcpy(id, "MAP32");

			if (id[0])
			{
				for (i = 0; i < defs.count.mapinfo.num; i++)
				{
					if (!_stricmp(defs.mapinfo[i].id, id))
					{
						if(dehdebug) pr_OFinsert ("USE deh BACKGROUND FLAT %s in place of %s\n", newStr, oldStr);
						strcpy(defs.mapinfo[i].bexbgflat, newStr);
						good = true;
					}
				}
			}
		}
	}

	if (good) goto donewithtext;

	// Search through sprite names
	if (strlen(oldStr) == 4 && strlen(newStr) == 4)
	{
		for (i = 0; i < DH_NUMSPRITES; i++)
		{
			if (!strcmp(defs.sprites[i].id, oldStr))
			{
				strncpy(defs.sprites[i].id, newStr, 4);
				changespritename = true;

				if (SPRarraycount >= maxSPRarraycount)
				{
					maxSPRarraycount += 16;
					sprArray = (sprArray_t*)W_Realloc(sprArray, sizeof(sprArray_t)*maxSPRarraycount, STATICID, 204);
				}

				if(dehdebug) pr_OFinsert ("REPLACE SPRITE %s WITH %s\n", oldStr, newStr);
				strcpy(sprArray[SPRarraycount].oldSPR, oldStr);
				strcpy(sprArray[SPRarraycount].newSPR, newStr);
				SPRarraycount++;

				good = true;
			}
		}
	}

	if (good) goto donewithtext;

	// Search music names.
	if(oldSize <= 6)
	{
		sprintf(buf, "d_%s", oldStr);
		for(i=0; i<defs.count.music.num; i++)
		{
			if(!_stricmp(defs.music[i].lumpname, buf))
			{
				if(dehdebug) pr_OFinsert ("REPLACE MUSIC %s WITH %s\n", oldStr, newStr);
				sprintf(defs.music[i].lumpname, "D_%s", newStr);
				_strupr(defs.music[i].lumpname); // looks nicer...
				good = true;
			}
		}
	}

	if (good) goto donewithtext;

	if (strlen(oldStr) && strlen(newStr))
	{
	/*	char	tmp[4096];

		// v220-5 format string to match those in the defs
		strcpy(tmp, oldStr);
		DeEscapeString(tmp);*/

		// b220-0 Search through the base texts
		// v220-7 use NUMTEXT
		for (i = 0; i < NUMTEXT; i++) 
		{
			// .. just to make sure the defs are valid
			if (!defs.text[i].text) Con_Error("DEF ERROR: Undefined text %i\n", i);
			if (!_stricmp(defs.text[i].text, oldStr))
			{
				Def_ModifyTextDef(i, newStr);
				if(dehdebug)
				{
					int a = 0, b = 0;
					if (strlen(oldStr) > 36) { oldStr[36] = 0; a = 1; }
					if (strlen(newStr) > 36) { newStr[36] = 0; b = 1; }
					pr_OFinsert ("REPLACE STR \"%s%s\" WITH \"%s%s\"\n", oldStr, a ? "..." : "", newStr, b ? "..." : "");
				}
				good = true;
				break;
			}
		}
	}

	if (good) goto donewithtext;

	if (strlen(oldStr) > 280 && !strncmp(oldStr, "============", 12))	// Arbitrary cmp length
	{
		extern char	*headerDEHinsert;

		headerDEHinsert = (char*)malloc(strlen(newStr) + 4);
		sprintf(headerDEHinsert, "%s\n", newStr);
	}

	// Search map names
	for (i = 0; i < defs.count.mapinfo.num; i++)
	{
		if (!_stricmp(oldStr, defs.mapinfo[i].mapname))
		{
			int k = strlen(newStr) - 1;

			// v220-33 the new string will (or could) have following spaces
			for (; k > 0; --k) { if (newStr[k] == ' ') newStr[k] = 0; else break; }
		
			if(dehdebug) pr_OFinsert ("MAPNAME %s WITH %s\n", oldStr, newStr);
			strncpy(defs.mapinfo[i].mapname, newStr, DED_STRINGID_LEN);
			defs.mapinfo[i].mapname[DED_STRINGID_LEN] = 0;
			good = true;
			break;
		}
	}

	if (good) goto donewithtext;

	if (dehdebug)
	{
		char *ptr;
		
		l1 = strlen(oldStr);
		if (l1 && (ptr = strchr(oldStr, (int)'\n')) ) *ptr = 0;

		if (good)
		{
			if ((l2 = strlen(newStr)) && (ptr = strchr(newStr, (int)'\n')) ) *ptr = 0;
			pr_OFinsert ("REPLACE \"%.32s%s\" WITH \"%.32s%s\"\n",
				oldStr, l1 > 32 ? "..." : "",
				newStr, l2 > 32 ? "..." : "");
		}
		else
		{
			pr_OFinsert ("DEH INFO : at line %i Unmatched string \"%.32s%s\"\n", linecounter + 1, oldStr, l1 > 32 ? "..." : "");
		}
	}

donewithtext:
	if (newStr)	free(newStr);
	if (oldStr) free(oldStr);

	// Fetch next identifier for main loop
	while ((result = GetLine ()) == 1);

	return result;
}

static int PatchStrings (int dummy)
{
	size_t	maxstrlen = 128;
	char	*holdstring;
	int		result;

	holdstring = (char *)M_Malloc (maxstrlen, 0);

	while ((result = GetLine()) == 1)
	{
		RDBOOL	valid = false;
		char	id[8];
		int		i, n;

		*holdstring = '\0';

		do
		{
			while (maxstrlen < strlen (holdstring) + strlen (Line2))
			{
				maxstrlen += 128;
				holdstring = (char *)M_Realloc (holdstring, maxstrlen, 30);
			}

			strcat (holdstring, skipwhite (Line2));
			stripwhite (holdstring);

			if (holdstring[strlen(holdstring)-1] == '\\')
			{
				holdstring[strlen(holdstring)-1] = '\0';
				Line2 = igets ();
			} else
				Line2 = NULL;
		}
		while (Line2 && *Line2);

		// v220-33 support BGFLATE1 etc. (probably a complete waste of time)
		if (   !strncmp(Line1, "BGFLATE", 7)
			&& !Line1[8] && Line1[7] >= '1' && Line1[7] <= '9')
		{
			if (  (n = W_CheckNumForName(holdstring)) >= 0
				&& DoomLumpIdentify(n, NULL, NULL, DLI_GRAPHIC) == LMP_FLAT)
			{
				char id[8];
				int  k;

				// Replace all maps' bexbgflat in episode
				for (i = 1; i < 10; i++)
				{
					sprintf(id, "E%cM%i", Line1[7], i);

					for (k = 0; k < defs.count.mapinfo.num; k++)
					{
						if (!_stricmp(defs.mapinfo[k].id, id))
						{
							strcpy(defs.mapinfo[i].bexbgflat, holdstring);
							valid = true;
							break;
						}
					}
				}
			}
		
			if (dehdebug)
			{
				if (valid)
					pr_OFinsert ("USE deh BACKGROUND FLAT %s for E%c\n", holdstring, Line1[7]);
				else
				{
					pr_OFinsert("At line %i flat \"%s\" not found or invalid\n", linecounter, holdstring);
					valid = true;	// prevent error message at end being printed
				}
			}
		}
		else
		if (strlen(Line1) == 10 && !strnicmp(Line1, "HUSTR_E", 7))
		{
			if (gamemission == doom)
			{
				strncpy(id, &Line1[6], 4);
				id[4] = 0;

				for (i = 0; i < defs.count.mapinfo.num; i++)
				{
					if (!_stricmp(defs.mapinfo[i].id, id))
					{
						if (dehdebug) pr_OFinsert("MAPNAME: \"%s\" \"%s\"\n", Line1, holdstring);
						strncpy(defs.mapinfo[i].mapname, holdstring, DED_STRINGID_LEN);
						defs.mapinfo[i].mapname[DED_STRINGID_LEN] = 0;
						valid = true;
						break;
					}
				}
			}
			else
			{
				if (dehdebug) pr_OFinsert("%9s does not apply\n", Line1);
				if (!packwarn)
				{
					pr_OFdirect("CHECK THE CORRECT IWAD IS BEING USED\n");
					packwarn = true;
				}
				continue;
			}
		}
		else
		if (   (!strnicmp(Line1    , "HUSTR_", 6) && Line1[6])
			|| (!strnicmp(&Line1[1], "HUSTR_", 6) && Line1[7]) )
		{
			char	*p = Line1;
			int		type = (*p == 'p' || *p == 'P') + 2*(*p == 't' || *p == 'T');

			if (   (type == 0 && gamemission == doom2    )
				|| (type == 1 && gamemission == pack_plut)
				|| (type == 2 && gamemission == pack_tnt ) )
			{
				n = type ? atoi(&Line1[7]) : atoi(&Line1[6]);

				if (n > 0 && n < 99)
				{
					sprintf(id, "MAP%02i", n);

					for (i = 0; i < defs.count.mapinfo.num; i++)
					{
						if (!_stricmp(defs.mapinfo[i].id, id))
						{
							if (dehdebug) pr_OFinsert("MAPNAME: \"%s\" \"%s\"\n", Line1, holdstring);
							strncpy(defs.mapinfo[i].mapname, holdstring, DED_STRINGID_LEN);
							defs.mapinfo[i].mapname[DED_STRINGID_LEN] = 0;
							valid = true;
							break;
						}
					}
				}
			}
			else
			{
				if (dehdebug) pr_OFinsert("%9s does not apply\n", Line1);
				if (!packwarn)
				{
					pr_OFdirect("CHECK THE CORRECT IWAD IS BEING USED\n");
					packwarn = true;
				}
				continue;
			}

			if (!valid && !packwarn)
			{
				pr_OFdirect("CHECK THE CORRECT IWAD IS BEING USED\n");
				packwarn = true;
			}
		}
		else if (strlen(Line1) == 8
			&& !strnicmp(Line1, "BGFLAT", 6)
			&& (n = strlen(holdstring)) > 0 && n <= 8
			&& (n = atoi(&Line1[6])) > 0)
		{
			// v220-13 make compatible with defs
			switch(n)
			{
				case  6: n =  7; break;
				case 11: n = 12; break;
				case 15: n = 31; break;
				case 20: n = 21; break;
				case 31: n = 32; break;
			}

			sprintf(id, "MAP%02i", n);

			for (i = 0; i < defs.count.mapinfo.num; i++)
			{
				if (!_stricmp(defs.mapinfo[i].id, id))
				{
					if (W_CheckNumForName(holdstring) >= 0)
					{
						if(dehdebug) pr_OFinsert ("USE bex BACKGROUND FLAT %s for MAP%02i (def map%02i)\n", holdstring, atoi(&Line1[6]), n);
						strcpy(defs.mapinfo[i].bexbgflat, holdstring);
					}
					else
						pr_OFinsert("DEH error: flat \"%s\" does not exist\n", holdstring);

					valid = true;
					break;
				}
			}
		}
		else		// v220-4 Search the normal strings
		{
			// v220-7 first check the translated strings
			for (i = 0; nsub[i].name; i++)
			{
				if (!strcmp(nsub[i].name, Line1))
				{
					// Update the string.
					Def_ModifyTextDef(nsub[i].idx, holdstring);
					if (dehdebug) pr_OFinsert("[2] Def_ModifyTextDef ref %i: %s\n", i, holdstring);
					valid = true;
					break;
				}
			}

			if (!valid)
			{
				for (i = 0; i < defs.count.text.num; i++)
				{
					if (!strcmp(defs.text[i].id, Line1))
					{
						// Update the string.
						Def_ModifyTextDef(i, holdstring);
						valid = true;
						break;
					}
				}
			}
			
			if (dehdebug && valid)
			{
				int l2 = (int)strlen(holdstring);

				pr_OFinsert ("Def_ModifyTextDef %s - <%.32s%s>\n",
					Line1,
					holdstring, l2 > 32 ? "..." : "");
			}

		}
		
		if (!valid) pr_OFinsert ("At line %i Unsupported string: %s %s\n", linecounter, Line1, holdstring);
	}

	M_Free(holdstring);

	// v220-3 make sure next line is processed
	if (result == 2) return 2;	// process line

	return PatchNothing();
}

//===========================================================================
// DoInclude
//===========================================================================
static int DoInclude (int dummy)
{
	DFILE	*file;
	char	*savepatchfile, *savepatchpt;
	char	*data;
	char	*base;
	int		savedversion, savepversion;
	int		c, len;

	// ---------------------

	if (including) 
	{
		pr_OFinsert ("Sorry, can't nest includes\n");
		goto endinclude;
	}

	data = COM_Parse (Line2);

	if (!_stricmp (com_token, "notext")) 
	{
		includenotext = true;
		data = COM_Parse (data);
	}

	if (!com_token[0]) 
	{
		includenotext = false;
		pr_OFinsert ("Include directive is missing filename\n");
		goto endinclude;
	}

	if(dehdebug) pr_OFinsert("Including %s\n", com_token);
	savepatchfile = PatchFile;
	savepatchpt = PatchPt;
	savedversion = dversion;
	savepversion = pversion;
	including = true;

	if( (file = F_OpenFile(com_token, "rb")) == NULL)
	{
		pr_OFinsert("DEH ERROR : file \"%s\" does not exist.\n", com_token);
		goto endinclude;
	}

	F_Seek(file, 0, SEEK_END);
	len = F_Tell(file);
	F_Seek(file, 0, SEEK_SET);

	c = len & 1 ? 3 : 4;
	base = (char*)M_Calloc(len + c, 206);

	F_Read(base, len, file);
	F_Close(file);

	ApplyDEH(base, len, 0);
	M_Free(base);

	if(dehdebug) pr_OFdirect ("Done with include\n");
	PatchFile = savepatchfile;
	PatchPt = savepatchpt;
	dversion = savedversion;
	pversion = savepversion;

endinclude:
	including = false;
	includenotext = false;
	return GetLine();
}

// v220-33
// Having struggled with god knows what issues it was decided
// it was better to reformat whatever's been sent rather than
// trying to cope with all of it within the reader itself.

// Table for detecting original Doom2 mapnames
char *DoomMapnames[] = {
	"level 1: entryway",
	"level 2: underhalls",
	"level 3: the gantlet",
	"level 4: the focus",
	"level 5: the waste tunnels",
	"level 6: the crusher",
	"level 7: dead simple",
	"level 8: tricks and traps",
	"level 9: the pit",
	"level 10: refueling base",
	"level 11: 'o' of destruction!",
	"level 12: the factory",
	"level 13: downtown",
	"level 14: the inmost dens",
	"level 15: industrial zone",
	"level 16: suburbs",
	"level 17: tenements",
	"level 18: the courtyard",
	"level 19: the citadel",
	"level 20: gotcha!",
	"level 21: nirvana",
	"level 22: the catacombs",
	"level 23: barrels o' fun",
	"level 24: the chasm",
	"level 25: bloodfalls",
	"level 26: the abandoned mines",
	"level 27: monster condo",
	"level 28: the spirit world",
	"level 29: the living end",
	"level 30: icon of sin",
	"level 31: wolfenstein",
	"level 32: grosse",
	""
};

// Table for detecting original Plutonia mapnames
char *PlutMapnames[] = {
	"level 1: congo",
	"level 2: well of souls",
	"level 3: aztec",
	"level 4: caged",
	"level 5: ghost town",
	"level 6: baron's lair",
	"level 7: caughtyard",
	"level 8: realm",
	"level 9: abattoire",
	"level 10: onslaught",
	"level 11: hunted",
	"level 12: speed",
	"level 13: the crypt",
	"level 14: genesis",
	"level 15: the twilight",
	"level 16: the omen",
	"level 17: compound",
	"level 18: neurosphere",
	"level 19: nme",
	"level 20: the death domain",
	"level 21: slayer",
	"level 22: impossible mission",
	"level 23: tombstone",
	"level 24: the final frontier",
	"level 25: the temple of darkness",
	"level 26: bunker",
	"level 27: anti-christ",
	"level 28: the sewers",
	"level 29: odyssey of noises",
	"level 30: the gateway of hell",
	"level 31: cyberden",
	"level 32: go 2 it",
	"level 1: system control",
	""
};

static int compMapname(char *cmp, int *addnum)
{
	int i = 0;

	do {
		if (!stricmp(DoomMapnames[i], cmp)) return i + 1;
	} while (DoomMapnames[i++] != "");

	i = 0; *addnum = 1;

	do {
		if (!stricmp(PlutMapnames[i], cmp)) return i + 1;
	} while (PlutMapnames[i++] != "");

	return 0;
}

static char* reformatPatch(char *base, int len, int *newlen)
{
	// Get lines from memory
	char	*newbase;
	char	*tokenptr;
	char	token[8192];
	char	toke2[128];
	RDBOOL	addstrings = false;
	RDBOOL	wasaddstrings = false;
	int		wadstrings = levelExMxExists ? 0x10001 : 0;
	int		offset = 0;
	int		linenum = 1;
	int		linelen;
	int		tplinelen;
	int		newlength = 0;
	int		num1, num2;
	int		i, c;
	int		chkComp = ge_inf.switch_flags & (SWF_AP|SWF_FP);

	if (dehdebug)
	{
		pr_OFdirect("\n==================================\n"
			"The following printout shows the file used by the DEH reader after\n"
			"being reformatted by Risen3D.  If necessary this should be used to\n"
			"reference errors by line number before then modifying any original\n"
			"file. Note that some unsupported or commented strings or ones that\n"
			"do not apply with the iwad selected may not be included.\n\n");
	}

	nsreported = 0;

	// get rid of any zeros that are not at the end of the file
	// e.g. with hunt.deh shipped with the hunt.wad
	for (i = 0; i < len - 2; i++) { if (!base[i]) base[i] = ' '; }

	c = len & 1 ? 3 : 4;
	newbase = (char*)M_Calloc(len + c, 207);

	while (base[offset] && offset < len)
	{
		linelen = num1 = num2 = 0;

		while (offset < len && linelen < 8192)
		{
			if (base[offset] == 0xD && base[offset + 1] == 0xA)
			{
				offset += 2; break;
			}
			else if (base[offset] == 0xA)
			{
				offset++; break;
			}

			token[linelen++] = base[offset++];
		}

		token[linelen] = 0;
		tplinelen = linelen;
		tokenptr = token;

		// Top and tail white space
		while (*tokenptr == ' ') tokenptr++, tplinelen--;
		if (!tplinelen) continue;
		if (*tokenptr == '#') continue;	// In DEHACKED files # only acts at the start to mark any following characters as a comment
										// It can also be used within a line e.g. ID # = 36 (Who dreamt this up?)

		while (tplinelen > 0 && tokenptr[tplinelen] == ' ') tokenptr[tplinelen--] = 0;

		if (!tplinelen) continue;
		if (chkComp)
		{
			if (strstrCI(tokenptr, "FRAME") || strstrCI(tokenptr, "SPRITE"))
				Con_Error("DEHACKED detected with frame and/or sprite changes.\n"
				"This prevents support for Fullpack or Alternative weapons.\n\n"
				"Please untick these options before restarting");
		}
		if (!strncmp(tokenptr, "OB_", 3)) { nsreported |= NSR_OB_STR; continue; }
		if (!strncmp(tokenptr, "STARTUP5", 8)) { nsreported |= NSR_ST5_STR; continue; }
		if (!strncmp(tokenptr, "HUSTR_PLR", 9)) continue;

		if (tplinelen > 9)
		{
			// check for mapname strings
			if (!strncmp(tokenptr, "HUSTR_E", 7)) wadstrings |= 0x1;
			else
			if (!strncmp(tokenptr, "HUSTR_" , 6)) wadstrings |= 0x2;
			else
			if (!strncmp(tokenptr, "PHUSTR_", 7)) wadstrings |= 0x4;
			else
			if (!strncmp(tokenptr, "THUSTR_", 7)) wadstrings |= 0x8;

			if (gamemission == doom      && ((strncmp(tokenptr, "HUSTR_E", 7) && !strncmp(tokenptr, "HUSTR_" , 6)) || !strncmp(tokenptr, "PHUSTR_", 7) || !strncmp(tokenptr, "THUSTR_", 7)) ) { wadstrings |= 0x10000; continue; } 
			if (gamemission == doom2     && (!strncmp(tokenptr, "HUSTR_E", 7)  || !strncmp(tokenptr, "PHUSTR_", 7) || !strncmp(tokenptr, "THUSTR_", 7)) ) { wadstrings |= 0x20000; continue; }
			if (gamemission == pack_plut && (!strncmp(tokenptr, "HUSTR_" , 6)  || !strncmp(tokenptr, "HUSTR_E", 7) || !strncmp(tokenptr, "THUSTR_", 7)) ) { wadstrings |= 0x40000; continue; }
			if (gamemission == pack_tnt  && (!strncmp(tokenptr, "HUSTR_" , 6)  || !strncmp(tokenptr, "HUSTR_E", 7) || !strncmp(tokenptr, "PHUSTR_", 7)) ) { wadstrings |= 0x80000; continue; }
		}

		// conversion for old DEHACKED using original map names
		if (levelExMxExists) goto wadOK;

		if (tplinelen > 7 && !strnicmp(tokenptr, "text ", 5)) // get text n1 n2
		{
			// get first number
			for (i = 5; i < tplinelen; i++) { if (tokenptr[i] != ' ') break; }

			num1 = atoi(&tokenptr[i]);
			// get second number
			for (; i < linelen; i++) { if (tokenptr[i] == ' ') break; }
			for (; i < linelen; i++) { if (tokenptr[i] != ' ') break; }
			num2 = atoi(&tokenptr[i]);

			if (num1 >= 16 && num2 >= 16) { addstrings = false; goto wadOK; }

			int		off = offset, llen = 0;

			// get the next line
			while (off < len && llen < 128)
			{
				if (base[off] == 0xD && base[off + 1] == 0xA)
				{
					off += 2; break;
				}
				else if (base[off] == 0xA)
				{
					off++; break;
				}

				if (llen >= 8192) { Con_Error("DEHACKED string exceeds max length allowed of 8192 chars"); }
				toke2[llen++] = base[off++];
			}

			toke2[llen] = 0;

			{
				static char *addname[3] = {"", "P", "T" };
				char mname[128];
				int addnum = 0;
				strncpy(mname, toke2, num1); mname[num1] = 0;

				if ( !(c = compMapname(mname, &addnum)) ) { addstrings = false; goto wadOK; }

				strncpy(mname, toke2 + num1, num2); mname[num2] = 0;
				// strip following spaces off mname - they are completely unnecessary
				for (i = num2 - 1; i > 0; --i) { if (mname[i] == ' ') mname[i] = 0; else break; }

				// Add [STRINGS]
				if (!addstrings)
				{
					addstrings = true;
					sprintf(token, "[STRINGS]");
					linelen = (int)strlen(token);
					if (newlength + linelen + 1 > len) goto report_ovf;
					if (dehdebug) pr_OFinsert("%3i) %s\n", linenum, token);
					sprintf(newbase + newlength, "%s\n", token);
					newlength += (linelen + 1);
					linenum++;
				}

				sprintf(token, "%sHUSTR_%d = %s", addname[addnum], c, mname);
				linelen = (int)strlen(token);
				if (newlength + linelen + 1 > len) goto report_ovf;
				if (dehdebug) pr_OFinsert("%3i) %s\n", linenum, token);
				sprintf(newbase + newlength, "%s\n", token);
				newlength += (linelen + 1);
				linenum++;
				wadstrings |= addnum == 1 ? 0x40004: addnum == 2 ? 0x80008: 0x20002;
			}

			offset = off;
			continue;
		}
		else
		{
			if (addstrings)
			{
				if (!(*token))
				{
					wasaddstrings = true;
					continue;
				}
			}
		}

		if (!strnicmp(token, "IWAD", 4))
		{
			char *iwadneeded = "";
			if (strstrCI(token, "DOOM2.WAD")) { iwadneeded = "DOOM2"; if (gamemission == doom2) goto wadOK; }
			if (strstrCI(token, "PLUTONIA.WAD")) { iwadneeded = "PLUTONIA"; if (gamemission == pack_plut) goto wadOK; }
			if (strstrCI(token, "TNT.WAD")) { iwadneeded = "TNT"; if (gamemission == pack_tnt) goto wadOK; }
			if (strstrCI(token, "DOOM1.WAD")) { iwadneeded = "DOOM"; if (gamemission == doom) goto wadOK; }
			if (*iwadneeded == '\0') goto wadOK;
			M_Free(base);
			M_Free(newbase);
			Con_Error("Wrong IWAD selected - use %s\n", iwadneeded);
		}
wadOK:;

		if (num1 >= 16 && num2 >= 16)
		{
			int		off = offset, llen = 0, pos = 0;
			char	tmpstr[2046];
			
			tmpstr[pos] = 0;
			sprintf(tmpstr, "Text %i %i\n", num1, num2);
			if (dehdebug) pr_OFinsert("%3i) %s", linenum++, tmpstr);
			sprintf(newbase + newlength, "%s", tmpstr);
			newlength += strlen(tmpstr);
			tmpstr[pos] = 0;

			// get the next line
			while (off < len && llen < (num1 + num2))
			{
				if (base[off] == 0xD && base[off + 1] == 0xA)
					off++;

				tmpstr[pos++] = base[off];

				if (base[off] == 0xA)
				{
					tmpstr[pos] = 0;
					if (dehdebug) pr_OFinsert("%3i) %s", linenum++, tmpstr);
					sprintf(newbase + newlength, "%s", tmpstr);
					newlength += strlen(tmpstr);
					pos = 0;
					tmpstr[pos] = 0;
				}

				token[llen++] = base[off++];
			}

			if (pos)
			{
				tmpstr[pos] = 0;
				if (dehdebug) pr_OFinsert("%3i) %s\n", linenum++, tmpstr);
				sprintf(newbase + newlength, "%s\n", tmpstr);
				newlength += strlen(tmpstr) + 1;
			}

		//	if (dehdebug) pr_OFinsert("%3i) Text %i %i %s\n", linenum++, num1, num2, token);
		//	newlength += llen;
		//	sprintf(newbase + newlength, "Text %i %i\n%s\n", num1, num2, token);
			offset = off;
			continue;
		}

		if (newlength + 1 > len)
		{
report_ovf:;
		   pr_OFinsert("DEH REFORMAT file length needed of %i exceeds that allocated of %i\n"
				"Please report this together with a copy of Risen3D.out.\n",
				newlength + linelen + 1, len);
			break;
		}

		if (wasaddstrings)
		{
			wasaddstrings = false;
			// Add a newline

			if (dehdebug) pr_OFinsert("%3i)\n", linenum);
			sprintf(newbase + newlength, "\n");
			newlength += 1;
			linenum++;
		}

		if (dehdebug) pr_OFinsert("%3i) %s\n", linenum, token);

		sprintf(newbase + newlength, "%s\n", token);
		newlength += (linelen + 1);
		linenum++;
	}

	if (wadstrings & 0xF0000 && wadstrings & 0xF)
	{
		char *iwadptr = NULL;

		int  type = wadstrings & 0xF;
		if (type == 0x1 && gamemission != doom     ) iwadptr = "DOOM"    ; else
		if (type == 0x2 && gamemission != doom2    ) iwadptr = "DOOM2"   ; else
		if (type == 0x4 && gamemission != pack_plut) iwadptr = "PLUTONIA"; else
		if (type == 0x8 && gamemission != pack_tnt ) iwadptr = "TNT";

		if (iwadptr)
		{
			sprintf(toke2,	"It is possible that the wrong IWAD has been selected%\n"
							"as DEHACKED contains strings for another IWAD.\n\n"
							"The one most likely to be needed is %s.WAD\n\n", iwadptr);

			if (!Sys_CriticalMessage(toke2, true))
			{
				M_Free(base);
				M_Free(newbase);
				Sys_Quit(false);
			}
		}
	}

	if (dehdebug)
	{
		pr_OFinsert("\nEnd of reformatted file oldlen %i newlen %i ========\n\n", len, newlength);
	}

	*newlen = newlength;

	return newbase;
}

//===========================================================================
// ApplyDEH
//===========================================================================
static void ApplyDEH(char *base, int length, int from)
{
	int cont, newlen;

	char *dehpatch = reformatPatch(base, length, &newlen);

	length = newlen;

	pr_OFdirect("\nApply DEH File ============  type: ");
	switch (from)
	{
	case 0: pr_OFdirect("  Nested include file\n"); break;
	case 1: pr_OFdirect("  Wad lump\n"); break;
	case 2: pr_OFdirect("  External File\n"); break;
	}

	BackupData ();

	PatchFile = base;
	dversion  = pversion = -1;
	PatchPt = dehpatch;
	Line1 = Line2 = NULL;
	linecounter = 0;

	if ( !(cont = GetLine()) )
	{
		pr_OFinsert("Invalid DEHACKED sent.\n");
		goto enddeh;
	}

	if (Line1 && Line2 && !stricmp("Patch", Line1) && !stricmp("File for DeHackEd v3.0", Line2) ) 
	{
		while ((cont = GetLine()) == 1)
		{
			CHECKKEY ("Doom version", dversion)
				else CHECKKEY ("Patch format", pversion)
		}
		if (!cont || dversion == -1 || pversion == -1) 
		{
			pr_OFinsert("This is not a DeHackEd patch file!\n");
			goto enddeh;
		}
	} 
	else 
	{
		pr_OFinsert("Patch does not have DeHackEd signature. Assuming .bex\n");
		dversion = 19;
		pversion = 6;
	}
		
	hasDEH = true;

	if (pversion != 6) {
		pr_OFinsert("DeHackEd patch version is %d.\nUnexpected results may occur.\n", pversion);
	}
		
	if (dversion == 16)
		dversion = 0;
	else if (dversion == 17)
		dversion = 2;
	else if (dversion == 19)
		dversion = 3;
	else if (dversion == 20)
		dversion = 1;
	else if (dversion == 21)
		dversion = 4;
	else
		dversion = 3;

	pr_OFinsert("Patch created with dversion %i\n", dversion);
		
	do
	{
		if (cont == 1)
		{
			pr_OFinsert ("Key %s encountered out of context\n", Line1);
			cont = 0;
		}
		else if (cont == 2)
		{
			cont = HandleMode (Line1, atoi (Line2));
		}
	} while (cont);

enddeh:;
	M_Free(dehpatch);

	if (nsreported & NSR_OB_STR ) pr_OFdirect("DEH INFO: ZDoom obituary texts are not supported.\n");
	if (nsreported & NSR_ST5_STR) pr_OFdirect("DEH INFO: STARTUP5 strings are not supported.\n");
	if (nsreported & NSR_PSNDSTR) pr_OFdirect("DEH INFO: patch sounds ignored.\n");
	
	pr_OFinsert("\n======== End DEH\n\n");
}

//===========================================================================
// ReadDehackedLump
//	Reads and applies the given lump as a DEH patch.
//===========================================================================
static void ReadDehackedLump(int lumpnum)
{
	int		len;
	char	*lump, *ptr;
	int		a = ArgExists("-dogs"), b = ArgExists("-Bskull");

	if (a || b)
		Con_Error("This wad has been DEHACKED.\nPlease remove %s from the global command line\n"
				  "as not doing so could adversely affect dehacking.\n",
				  a && !b ? "-dogs" : !a && b ? "-Bskull" : "-dogs and -Bskull");

	lump = ptr = (char*)M_CacheLumpNum(lumpnum, (uint*)&len, false);

	if (ptr)
	{
		ApplyDEH(lump, len, 1);

		M_Free(ptr);
	}
}

//===========================================================================
// ReadDehacked
//	Reads and applies the given Dehacked patch file.
//===========================================================================
static void ReadDehacked(char *filename)
{
	DFILE	*f;
	int		c, len;
	char	*base;

	if (!(f = F_OpenFile(filename, "rb")))
	{
		pr_OFinsert("PROBLEM: Dehacked file %s could not be opened\n", filename);
		return;
	}

	F_Seek(f, 0, SEEK_END);
	len = F_Tell(f);
	F_Seek(f, 0, SEEK_SET);

	if (len > 16)	// Not just a label with nothing much in it
	{
		if (R3D_ScriptedMap)
		{
			F_Close(f);
			Con_Error("DEHACKED in a Risen3D scripted map?\nSacrilege: forget it.");
		}

		// Allocate enough memory and read it. Also ensures there is a terminating 0
		c = len & 1 ? 3 : 4;
		base = (char*)M_Calloc(len + c, 207);

		F_Read(base, len, f);		// Load the file into memory
		F_Close(f);

		ApplyDEH(base, len, 2);	
		M_Free(base);
	}
	else
	{
		F_Close(f);
	}

	if (strstr(filename, "arctic1.deh") != NULL) isarctic |= 1;
	if (strstr(filename, "arctic2.deh") != NULL) isarctic |= 2;
}

//===========================================================================
// Defs_DEH
//	This will be called after the engine has loaded all definitions but
//	before the data they contain has been initialized.
//===========================================================================
RDBOOL Defs_DEH(RDBOOL check_exists)
{
	static int def_arrays_inited = false;
	char	temp[MAXPATHNAMESIZE];
	char	*fn;
	RDBOOL	allextdehs;	
	int		i;

	if (!def_arrays_inited)
	{
		// Initialise the various arrays
		memset(flagsDEH  , 0xFF, DH_NUMSPRITES * sizeof(uint));
		memset(DEHoldTH  ,  0, DH_CHECKNUM * sizeof(ushort));
		memset(DEHCheck  ,  0, DH_CHECKNUM);

		for (i = 0; i < NUMMISCDEHVALS; i++) miscDEHval[i] = 0x80000000;

		def_arrays_inited = true;
	}

	if (ArgExists("-nodeh")) return false;

	dehdebug = ArgExists("-dehdebug");
	allextdehs = ArgExists("-allextdehs");

	changespritename = false;

	if (!allextdehs)
	{
		for(i = numlumps - 1; i >= 0; i--)
		{
			if(!strnicmp(W_LumpName(i), "DEHACKED", 8))
			{
				if (check_exists) return true;

				ReadDehackedLump(i);
				// We'll only continue this if the -alldehs option is given.
				if(!ArgExists("-alldehs")) break;
			}
		}
	}

	// How about the -deh option?
	if (ArgCheckWith("-deh", 1))
	{
		if (!check_exists && hasDEH && !allextdehs)
		{
			pr_OFinsert("WARNING: -deh switch ignored. DEHACKED lump found in pwad.\n"
						"  To override add the -allextdehs switch - this also inhibits pwad DH lumps.\n"
						"  If you also need the pwad DH lump use -dumplumps \"DEHACKED\" to extract it.\n"
						"  NOTE: loading the same deh lump more than once will cause problems.\n");
		}
		else
		{
			while((fn = ArgNext()) != NULL && fn[0] != '-')
			{
				// $implicit; Check for user child game directory
				// Check path not explicitly sent
				if (r3dUserDir.path[0] &&	// b211-0 Check if user directory set
					!strrchr(fn, '\\') &&
					!strrchr(fn, '/'))
				{
					sprintf(temp, "%s%s", r3dUserDir.path, fn);	// b211-0
					if ( (_access( temp, 4 )) == -1)
					{
						if (!check_exists)
							pr_OFinsert("UDIR: file not found\n"
										"Entry %s ignored\n", temp);
						continue;
					}
				}
				else
				{
					M_TranslatePath(fn, temp);
					if(!M_FileExists(temp))
						continue;
				}

				if (check_exists) return true;

				ReadDehacked(temp);
			}
		}
	}
	else if (r3dUserDir.path[0] && (!hasDEH || allextdehs))	// b211-0
	{
		long	hFile;
		struct	_finddata_t g_file;
		int		count = 0;
		RDBOOL	found = false;

		strcpy(temp, r3dUserDir.path);		// b211-0

		if (_chdir(temp))
			Con_Error("Unable to locate the directory: %s", temp);

		// b211-0 Revised
		// deh files are searched for first.  If none are
		// found then bex files are searched for.  This
		// search is split in case both a deh and bex file
		// has been included.

		// Find .deh files in current directory
		if ( (hFile = _findfirst( "*.DEH", &g_file )) != -1L )
		{
			do
			{
				if (Dir_ChExt(g_file.name, ".DEH"))
				{
					found = true;

					if (!check_exists)
						ReadDehacked(g_file.name);
				}
			}
			while( _findnext( hFile, &g_file ) == 0 );
			
			_findclose( hFile );
		}

		if (!found)
		{
			// Check for .bex files in current directory
			if ( (hFile = _findfirst( "*.BEX", &g_file )) != -1L )
			{
				do
				{
					if (Dir_ChExt(g_file.name, ".BEX"))
					{
						found = true;

						if (!check_exists)
							ReadDehacked(g_file.name);
					}
				}
				while( _findnext( hFile, &g_file ) == 0 );
				
				_findclose( hFile );
			}
		}
		_chdir(ddRuntimeDir.path);

		if (check_exists) return found;
	}

	if (check_exists) return false;

	// v220-13 Support single DEH that can be applied globally
	//         These are normally used with the IWADS
	if (!hasDEH && ArgCheckWith("-gDEH", 1))
	{
		char *arg = ArgNext();

		if (   arg
			&& *arg != '-'
			&& (Dir_ChExt(arg, ".DEH") || Dir_ChExt(arg, ".BEX")) )
		{
			char	fname[512];

			sprintf(fname, "%s%s", ddRuntimeDir.path, arg);
			ReadDehacked(fname);
		}
	}

#ifdef _DEBUG
	if (hasDEH && dehdebug) PrintDEHdebug();
#endif

	return changespritename;
}

#pragma optimize("", on)