//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_draw.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:   High level DGL drawing functions.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "de_base.h"
#include "de_graphics.h"
#include "de_refresh.h"
#include "de_render.h"
#include "cvars.h"

#include "r3d_OGLfunc.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		wimap_pic;
extern RDBOOL	wimap_el;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int fakefilter;			  // The under fake ceiling filter (0 = none).
int curfilter;			  // The current filter (0 = none).

// PRIVATE DATA DEFINITIONS ------------------------------------------------

RDBOOL dontUseXOffset = false; // hack to prevent x offset being used

// CODE --------------------------------------------------------------------

void subfn_DrawPatchQuads(float w, float h, float x, float y)
{
	if (!w || !h) return;

	glBegin(GL_QUADS);
	glTexCoord2i(0, 0); glVertex2f(x,   y);
	glTexCoord2i(1, 0); glVertex2f(x+w, y);
	glTexCoord2i(1, 1); glVertex2f(x+w, y+h);
	glTexCoord2i(0, 1); glVertex2f(x,   y+h);
	glEnd();
}

void GL_DrawBackground()
{
	if (wimap_pic >= 0)
		GL_DrawPagePatch(wimap_pic);
	else
		GL_DrawPagePatch(titlepic_num);
}

// Used only to draw automap markpoints
void GL_DrawFPatch_CS(float x, float y, float w, float h, int lnum)
{
	if (!GL_SetPatch(lnum, 0)) return;
	subfn_DrawPatchQuads(w, h, x, y);
}

//===========================================================================
// GL_DrawPatch_CS
//	Drawing with the Current State.
//  v220-8 nearest switch added
//===========================================================================
void GL_DrawPatch_CS(float x, float y, int lnum, RDBOOL nearest)
{
	// Set the texture.
	if (!GL_SetPatch(lnum, nearest)) return;

	if (!dontUseXOffset) x += (float)((short)lumptexinfo[lnum].offx);
	y += (float)((short)lumptexinfo[lnum].offy);

	subfn_DrawPatchQuads(
		(float)lumptexinfo[lnum].origw,
		(float)lumptexinfo[lnum].origh,
		x,
		y);
}

void GL_DrawMapnamePatch(int x, int y, int lnum, float alpha)
{
	if(lnum < 0) return;

	glColor4f(1.0f, 1.0f, 1.0f, alpha);
	GL_DrawPatch_CS((float)x, (float)y, lnum, 0);
}

//===========================================================================
// GL_DrawPagePatch	Deals with TITLEPIC/CREDITS
//===========================================================================
RDBOOL GL_DrawPagePatch(int lnum)
{
	if (!GL_SetPatch(lnum, 0))
	{
		GL_SetNoTexture();
		return false;
	}

	glColor3f(1.0f, 1.0f, 1.0f);
	subfn_DrawPatchQuads(SCREENWIDTH, SCREENHEIGHT, 0, 0);

	return true;
}

void GL_DrawPatchLitAlpha(int x, int y, float light, float alpha, int lumpnum)
{
	if(lumpnum < 0 || !gl_inited) return;
	glColor4f(light, light, light, alpha);
	GL_DrawPatch_CS((float)x, (float)y, lumpnum, 0);
}

// v220-8 Draws CS patch with nearest switch set true
void GL_DrawPatchNr(int x, int y, int lumpnum)
{
	if(lumpnum < 0 || !gl_inited) return;
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	GL_DrawPatch_CS((float)x, (float)y, lumpnum, 1);
}

void GL_DrawPatch(int x, int y, int lumpnum)
{
	GL_DrawPatchLitAlpha(x, y, 1.0f, 1.0f, lumpnum);
}

void GL_DrawRect(float x, float y, float w, float h, float r, float g, float b, float a)
{
	glColor4f(r, g, b, a);
	subfn_DrawPatchQuads(w, h, x, y);
}

void GL_DrawScreenFilter(float r, float g, float b, float a)
{
	glDisable(GL_TEXTURE_2D);
	glColor4f(r, g, b, a);
	glBegin(GL_QUADS);
	glTexCoord2i(0, 0); glVertex2i(0, 0);
	glTexCoord2i(1, 0); glVertex2i(SCREENWIDTH, 0);
	glTexCoord2i(1, 1); glVertex2i(SCREENWIDTH, SCREENHEIGHT);
	glTexCoord2i(0, 1); glVertex2i(0, SCREENHEIGHT);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void GL_DrawRectTiled(float x, float y, float w, float h, float tw, float th)
{
	float dw = w/tw, dh = h/th;

	glBegin(GL_QUADS);
	glTexCoord2f( 0,  0); glVertex2f(x,   y);
	glTexCoord2f(dw,  0); glVertex2f(x+w, y);
	glTexCoord2f(dw, dh); glVertex2f(x+w, y+h);
	glTexCoord2f( 0, dh); glVertex2f(x,   y+h);
	glEnd();
}

// Draws the filter if any is set
void GL_DrawFilter()
{
	int filter = curfilter ? curfilter : fakefilter;

	if (ma_inf.no_light[0] > 1) filter = 0x80800000;

	if (filter)
	{
		uint height = (cfg.sbarscale == 5 ? 168 : 200);

		glDisable(GL_TEXTURE_2D);
		glColor4ub(
			(GLubyte)((filter>>16)&0xff),	// r
			(GLubyte)((filter>> 8)&0xff),	// g
			(GLubyte)( filter     &0xff),	// b
			(GLubyte)((filter>>24)&0xff));	// a

		glBegin(GL_QUADS);
		glVertex2i(0  , 0);
		glVertex2i(320, 0);
		glVertex2i(320, height);
		glVertex2i(0  , height);
		glEnd();

		glEnable(GL_TEXTURE_2D);
	}
}

int GL_DrawPSprite(float x, float y, int flip, int splnum, RDBOOL ishudspr, RDBOOL shiftleft)
{
	spritelump_t *slump = spritelumps + splnum;
	int		w, h;
	int		w2, h2;
	int		flags = 0;
	float	wd, s, t, *tc;
	float	ttc[2];

	if (flip) flip = 1; // Make sure it's zero or one.

	// Set the defaults
	w = slump->width;
	h = slump->height;
	tc = slump->tc;

	if (ishudspr)
	{
		int		hsindex;

		GL_SetSprite(splnum, &hsindex, 2);

		if (hsindex)
		{
			int		whs, hhs, flags;

			if (GL_GetHSsize(hsindex, &whs, &hhs, &flags, ttc))
			{
				if (flags & HS_IGNORE) return 0;

				tc = ttc;
				w = whs; h = hhs;
			}
		}
	}
	else
	{
		GL_SetSprite(splnum, NULL, 5);
	}

	// To remove possible edge artefacts move the corner up/left slightly
	w2 = CeilPow2(w);
	h2 = CeilPow2(h);
	s = tc[VX] - 0.4f/w2;
	t = tc[VY] - 0.4f/h2;

	wd = (float)w;

	if (ishudspr)
	{
		y -= h;
		if (shiftleft) x -= w;
	}

	glBegin(GL_QUADS);
	glTexCoord2f( flip * s, 0); glVertex2f(x, y);
	glTexCoord2f(!flip * s, 0); glVertex2f(x + wd, y);
	glTexCoord2f(!flip * s, t); glVertex2f(x + wd, y + h);
	glTexCoord2f( flip * s, t); glVertex2f(x, y + h);
	glEnd();

	return ishudspr ? MIN_OF(w, 40) : 0;
}
