//-----------------------------------------------------------------------------
//
// Copyright (C) 2007 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_NewFunc.c
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: New functions added to release b211-0
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "doom\p_local.h"

//-----------------------------------------------------------------------------

//================================================================
// NF_UnclosedSecGroupsContained
//  b211-0 rewritten
// Deal with nested sectors creating false floors/ceilings
//================================================================

typedef struct {
	uint	area;
	int		secn;
	short	mark;
} area_t;

static int __cdecl AreaSorter(const void *e1, const void *e2)
{
	uint lc1 = ((area_t*)e1)->area;
	uint lc2 = ((area_t*)e2)->area;
	if (lc1 > lc2) return -1;
	if (lc1 < lc2) return  1;
	return 0;
}

static fixed_t refHeight(sector_t *sc, fixed_t fh)
{
	int i, ht1, ht2;

	for (i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if (   ld->ml_flags & ML_TWOSIDED
			&& (ht1 = ld->frontsector->floorheight) != (ht2 = ld->backsector->floorheight))
		{
			fixed_t retfh = MAX_OF(ht1, ht2);

			if (retfh > fh) return retfh;
		}
	}
	return fh;
}

static void markOffLines(subsector_t *ss)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i, txn;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		line_t *li = sg->linedef;

		if (!li || li->validcount == validcount)
			continue;

		li->validcount = validcount;

		SetLFG(li, LFG_NEWSET|LFG_NOARBITER, 47001);

		if (!secdone[li->frontsector - sectors])
			Setsecdone(li->frontsector, 47002);

		if (!(li->ml_flags & ML_TWOSIDED))
			continue;

		txn = origTEX[li - lines];

		if (li->frontsector == li->backsector)
		{
			if (txn & 0x22) Settype(li, R3D_03, 47003);
			else Settype(li, R3D_01, 47004);
			continue;
		}

		if (  (txn == 0x02 || txn == 0x20)
			&& li->frontsector->floorheight == li->backsector->floorheight
			&& li->frontsector->ceilingheight == li->backsector->ceilingheight)
		{
			// In barracks.wad the heights are incorrect leading the author to
			// set the ML_DONTPEGBOTTOM flag. It does not look good even with
			// software rendering and looks as it should (wrong) with OGL.
			// Setting <of> will cause the texure to be tiled to fit.
			if (li->ml_flags & ML_DONTPEGBOTTOM)
			{
				side_t *sd = txn == 0x02 ? &sides[li->sidenum[0]] : &sides[li->sidenum[1]];
				fixed_t ro = sd->rowoffset;
				fixed_t fh = refHeight(sd->sector, sd->sector->floorheight);
				fixed_t th = texArray[sd->midtexture]->height;
				fixed_t of = fh - ro - th*FRACUNIT - sd->sector->floorheight;

				if (of)
				{
					sd->rowoffadj = of;
					Settype(li, R3D_21, 47004);
				}
				else
				{
					Settype(li, R3D_03, 47004);
				}
			}
			else
			{
				Settype(li, R3D_03, 47004);
			}
		}

		SetLFG(li, LFG_NEWSET|LFG_FIXED, 47005);

		if (!secdone[li->backsector - sectors])
			Setsecdone(li->backsector, 47006);
	}
}

static void setFalseFloors(area_t *area, int count)
{
	subsector_t *ss;
	float		*outer = NULL;
	float		*inner = NULL;
	int			hi = DDMAXINT;
	int			lo = DDMININT;
	int			sechi = -1, seclo = -1;
	int			c, i, ni, no, p1, p2, p3;
	int			cpic;

	// The intermediate high and low sectors must now be established
	for (no = -1, i = 0; i < count; i++)
	{
		if (area[i].mark == 1)
		{
			int sn = area[i].secn;

			if (!outer) { outer = sectors[sn].sibounds; no = sn; }

			if ((c = sectors[sn].floorheight  ) > lo) { seclo = sn; lo = c; p1 = i; }
			if ((c = sectors[sn].ceilingheight) < hi) { sechi = sn; hi = c; p2 = i; }
		}
	}

	if (no < 0 || seclo < 0 || sechi < 0) return;	// insurance

	for (ni = -1, i = 0; i < count; i++)
	{
		if (area[i].mark == 1
			&& area[i].area < area[p1].area
			&& area[i].area < area[p2].area)
		{
			ni = area[i].secn;
			inner = sectors[ni].sibounds;
			p3 = i; break;
		}
	}

	if (ni < 0) return;

	// To be sure seclo, sechi and ni should each have a twin

	for (c = 3, i = 0; i < count; i++)
	{
		if (area[i].mark == 1)
		{
			if (i != p1 && area[i].area == area[p1].area) c--;
			if (i != p2 && area[i].area == area[p2].area) c--;
			if (i != p3 && area[i].area == area[p3].area) c--;
		}
	}

	if (c) return;	// unpaired sectors

	cpic = sectors[seclo].ceilingpic;

	// Mark the sectors in general as having been done
	for (i = 0; i < count; i++)
	{
		if (area[i].mark == 1) Setsecdone(&sectors[area[i].secn], 47007);
	}

	validcount++;

	// Not a model of efficiency but then hardly ever called.
	// Any optimisation would not be noticeable
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (   inner
			&& ss->midpoint.x > inner[BOXLEFT ]
			&& ss->midpoint.x < inner[BOXRIGHT]
			&& ss->midpoint.y > inner[BOXBASE ]
			&& ss->midpoint.y < inner[BOXTOP  ])
		{
			markOffLines(ss);
		}
		else
		if (   ss->midpoint.x > outer[BOXLEFT ]
			&& ss->midpoint.x < outer[BOXRIGHT]
			&& ss->midpoint.y > outer[BOXBASE ]
			&& ss->midpoint.y < outer[BOXTOP  ])
		{
			if (ss->sector->ceilingpic == cpic
				&& ss->sector->ceilingheight < sectors[seclo].ceilingheight)
			{
				Setr3dsscs(ss, seclo, 47008);
			}

			Setr3dsslcsec(ss, sechi, 47008);
			Setr3dsslfsec(ss, seclo, 47008);

			markOffLines(ss);
		}
	}
}

static RDBOOL secSplit(sector_t *s)
{
	ssline_t	*sl;
	int			i;
	int			refgrp = -1;

	for (i = 0; i < s->linecount; i++)
	{
		if (!(s->lines[i]->spflags & SPF_CONTROLLINE))
		{
			sl = &sslines[s->lines[i] - lines];

			if (sl->num)
			{
				refgrp = subsectors[sl->list[0]].viewgroup;
				break;
			}
		}
	}

	if (refgrp < 0) return true;

	for (i = 0; i < s->linecount; i++)
	{
		if (!(s->lines[i]->spflags & SPF_CONTROLLINE))
		{
			sl = &sslines[s->lines[i] - lines];

			if (sl->num)
			{
				if (subsectors[sl->list[0]].viewgroup != refgrp)
					return true;
			}
		}
	}

	return false;
}

void NF_UnclosedSecGroupsContained()
{
	sector_t *sec;
	int		 i, k, count = 0;

	SET_countsec(0);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& SecCtrl(i)
			&& !secSplit(sec))
		{
			countsec[count++] = i;
		}
	}

	if (count > 4)	// Two pairs plus main minimum
	{
		area_t	*area = (area_t*)M_Calloc(count*sizeof(area_t), 256);

		// Transfer the sectors stored in countsec
		for (i = 0; i < count; i++)
		{
			area[i].area = (uint)M_AreaFPBox(sectors[countsec[i]].sibounds);
			area[i].secn = countsec[i];
		}

		// Sort by area
		qsort(area, count, sizeof(area_t), AreaSorter);

		for (i = 0; i < count; i++)
		{
			int		c;
			float	*box;

			for (c = -1, k = 0; k < count; k++)
			{
				if (!area[k].mark) { area[k].mark = 1; c = k; break; }
			}

			if (c < 0) break;

			box = sectors[area[c].secn].sibounds;

			// Find what fits in where
			for (c = k = 0; k < count; k++)
			{
				if (!area[k].mark
					&& M_FitFPBox(sectors[area[k].secn].sibounds, box))
				{
					area[k].mark = 1;

					if ((sectors[area[k].secn].specFlags & 0xFFFFC) == 0xC)	// v220-9 changed from 0xFFFFF
					{
						c++;
					}
				}
			}

			// Set the subsectors
			// Must at least be a single sector and two pairs
			if (c > 4) setFalseFloors(area, count);

			// Mark off what was set
			for (k = 0; k < count; k++)
			{
				// Mark off as completed
				if (area[k].mark == 1) area[k].mark = 2;
			}
		}

		M_Free(area);
	}
}


// v220-13 Similar construction to that detected in NF_UnclosedSecGroupsContained()
// but where only a false floor needs to be set.
void NF_UnclosedSecGroupsContained2()
{
	typedef struct {
		sector_t *s;
		int		 lc;
		int		 used;
		int		 box[4];

	} marksec_t;

	marksec_t	*mrk;
	sector_t	*sec, *osec, *chk;
	line_t		*ld;
	int			c, i, k, n, count = 0;

	if (R3D_maptype) return;

	SET_countsec(0);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->heightsec < 0
			&& sec->stptr < 0
			&& (sec->specFlags & 0xFFFFC) == 0xC
			&& SecCtrl(i)
			&& !secSplit(sec))
		{
			countsec[count++] = i;
		}
	}

	if (count < 2) return;

	mrk = (marksec_t*)calloc(sizeof(marksec_t)*count, 1);

	// Find a paired sector that shares the same lines
	for (c = i = 0; i < count; i++)
	{
		osec = NULL;

		if (mrk[i].used) continue;

		M_ClearBox(mrk[i].box);

		sec = &sectors[countsec[i]];

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (!(ld->spflags & SPF_CONTROLLINE))
			{
				if (lineflags[ld - lines]) {mrk[i].used = -1; break; }

				chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (!chk || chk == sec) {mrk[i].used = -1; break; }
				if (!osec) osec = chk;
				else if (osec != chk) {mrk[i].used = -1; break; }
				mrk[i].lc++;
			}
			
			M_AddLineToBox(mrk[i].box, ld);
		}

		if (  !mrk[i].used
			&& mrk[i].lc > 2
			&& chk
			&& mrk[i].box[BOXTOP  ] - mrk[i].box[BOXBASE] > 16*FRACUNIT
			&& mrk[i].box[BOXRIGHT] - mrk[i].box[BOXLEFT] > 16*FRACUNIT)
		{
			mrk[i].s = chk; c++;
		}
		else
		{
			mrk[i].used = -1;
		}
	}

	if (c < 2) { free(mrk); return; }

	for (i = 0; i < count; i++)
	{
		if (!mrk[i].used)
		{
			for (k = 0; k < count; k++)
			{
				if (i != k
					&& !mrk[k].used
					&& mrk[k].lc == mrk[i].lc
					&& mrk[k].s - sectors == countsec[i]
					&& mrk[k].s->floorheight == sectors[countsec[i]].floorheight
					&& mrk[k].s->ceilingheight == sectors[countsec[i]].ceilingheight)
				{
					// We,ve pulled out two sectors that reference each other and share the same lines
					// Now get the subsectors inside the area enclosed by the lines

					fixed_t	loceil = mrk[k].s->floorheight;	// The floor will be above the ceiling
					fixed_t	*box = mrk[k].box;

					mrk[k].used = 1;
					osec = NULL;
					
					for (hitssecptr = n = 0; n < numsubsectors; n++)
					{
						if (M_FitBox(subsectors[n].ssbox, box))
						{
							hitssec[hitssecptr++] = n;

							if (   subsectors[n].sector->ceilingheight < loceil
								&& !(subsectors[n].sector->specFlags & 0xF000C))
							{
								osec = subsectors[n].sector;
								loceil = osec->ceilingheight;
							}
						}
					}

					if (   osec
						&& osec->ceilingheight < mrk[k].s->floorheight)
					{
						int p;

						CLR_hitline();

						for (n = 0; n < hitssecptr; n++)
						{
							subsector_t *ss = &subsectors[hitssec[n]];
							seg_t		*sg = &segs[ss->firstline];

							for (p = 0; p < ss->linecount; p++, sg++)
							{
								if ( (ld = sg->linedef) ) hitline[ld - lines] = 1;
							}

							if (ss->sector->specFlags & 0x8)
							{
								ss->lfsec = mrk[k].s - sectors;
							//	ss->lcsec = osec - sectors;
							}
						}

						for (n = 0; n < numlines; n++)
						{
							if (hitline[n])
								SetLFG(&lines[n], LFG_NEWSET|LFG_FIXED, 47072);
						}

						Setsecdone(mrk[i].s, 47072);
						Setsecdone(mrk[k].s, 47072);
					}
					break;
				}
			}
		}

		mrk[i].used = 1;
	}
	
	free(mrk);
}

//================================================================

//	NF_CheckLinesForInheritance()
//
// b211-0 Added

typedef struct {
	line_t	**line;
	short	*mark;
	int		count;
} ringlist_t;

static ringlist_t	rl;
static int			limit_check;

static int checkRinglist(int lptr, int *bbox, int condition, int num)
{
	vertex_t	*v1 = rl.line[lptr]->v1;
	RDBOOL		isclosed = false, repeat = true;
	int			c, i;

	if (   lptr < 0
		|| lptr >= rl.count
		|| rl.count > limit_check)
	{
		Con_Error("checkRinglist [%i]: bad parameter [%i %i %i]", num, lptr, rl.count, limit_check);
	}

	rl.mark[lptr] = 2;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < rl.count; i++)
		{
			if (rl.mark[i] != 1) continue;

			if (condition == 1)
			{
				if (!rl.line[i]->backsector) continue;
				if (origTEX[rl.line[i] - lines] & 0x11) continue;
				if (!(rl.line[i]->mtflags & R3D_MISLOTEX) ) continue;
			}

			if (rl.line[i]->v1 == v1)
			{
				v1 = rl.line[i]->v2; rl.mark[i] = 2; repeat = true;
			}
			else if (rl.line[i]->v2 == v1)
			{
				v1 = rl.line[i]->v1; rl.mark[i] = 2; repeat = true;
			}
		}
	}

	isclosed = (v1 == rl.line[lptr]->v2);

	if (bbox && isclosed) M_ClearBox(bbox);
	
	for (c = i = 0; i < rl.count; i++)
	{
		if (rl.mark[i] == 2)
		{
			if (isclosed)
			{
				rl.mark[i] = 4; c++;
				if (bbox) M_AddLineToBox(bbox, rl.line[i]);
			}
			else
			{
				rl.mark[i] = 1;
			}
		}
	}

	return c;
}

static void chkLinesConnected(line_t *ld, int secn)
{
	sector_t	*s1 = ld->frontsector;
	sector_t	*s2 = ld->backsector;
	int			i, lptr = -1;

	if (s1 == s2 || !s2 || s1->linecount != s2->linecount) return;

	// Check all lines in s1 are shared by s2
	for (i = 0; i < s1->linecount; i++)
	{
		line_t		*li = s1->lines[i];
		sector_t	*s = li->frontsector == s1 ?
							li->backsector : li->frontsector;

		if (s != s2) return;

		if (li == ld) lptr = i;

		if (subfn_FindLineCrossing(s->lines[i], LC_NORMAL) != NULL) return;
	}

	if (lptr < 0) return;	// code fault insurance

	rl.line  = s1->lines;
	rl.mark  = (short*)M_Calloc(s1->linecount*sizeof(short), 257);
	rl.count = limit_check = s1->linecount;
	for (i = 0; i < rl.count; i++) rl.mark[i] = 1;

	if (checkRinglist(lptr, NULL, 0, 1))
	{
		for (i = 0; i < rl.count; i++)
		{
			if (rl.mark[i] == 4)
			{
				Setspec(rl.line[i], secn, 47011);
			}
		}
	}

	M_Free(rl.mark);
}

static void checkUnsetAnimSecs(sector_t *sec)
{
	subsector_t *ss;
	line_t		*ld, *li;
	seg_t		*sg;
	int			i, k, m;

	ss_validcount++;
	CLR_hitline();		// Can't use validcount (used by intersect in line crossing search)

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->frontsector == ld->backsector)
		{
			ssline_t *sl = &sslines[ld - lines];
			int		 vt2 = Vtxs2(ld);

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount)
					continue;

				ss->ssvalidcount = ss_validcount;

				for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
				{
					if (!(li = sg->linedef) || hitline[li - lines])
						continue;

					hitline[li - lines] = true;

					if (   vt2
						&& lnsec[li - lines] < 0
						&& li->ml_flags & ML_TWOSIDED
						&& li->frontsector != li->backsector
						&& li->frontsector != sec
						&& li->backsector != sec
						&& li->lnflags2 & LNF2_OWNEQUALTOTWO
						&& li->frontsector->floorheight == li->backsector->floorheight
						&& li->frontsector->floorheight > sec->floorheight
						&& li->frontsector->floorpic == sec->floorpic
						&& li->backsector->floorpic == sec->floorpic
						&& !subfn_FindLineCrossing(ld, LC_NORMAL|LC_NOVTXONLINE))
					{
						int x = (int)(ss->midpoint.x * FRACUNIT);
						int y = (int)(ss->midpoint.y * FRACUNIT);
						// Has to be done this way as one can't depend on the
						// seg frontsector assignments.
						int	side = P_PointOnLineSide(x, y, ld);
						sector_t *chk = side ? ld->backsector : ld->frontsector;

						if (chk == sec)
						{
							Setspec(li, sec - sectors, 47012);
							// Check for other lines connected to this
							chkLinesConnected(li, sec - sectors);
						}
					}
				}
			}
		}
	}
}

void NF_CheckLinesForInheritance()
{
	sector_t	*sec;
	int			c, i, k;

	for (c = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& IsLiquidFloorFlat(sec)
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& (sec->specFlags & 0xC) != 0xC)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   ld->frontsector == ld->backsector
					&& lnsec[ld - lines] < 0
					&& !Vtxs2(ld)
					&& vertexowners[ld->v1 - vertexes].num < 3
					&& vertexowners[ld->v2 - vertexes].num < 3
					&& !subfn_FindLineCrossing(ld, LC_NORMAL|LC_NOVTXONLINE))
				{
					checkUnsetAnimSecs(sec);
					break;
				}
			}
		}
	}
}

//================================================================
//  NF_ContainerRings()
// b211-0
// Look for rings with 1's and missing tex lines
// Need to get these out of the way
static RDBOOL singleSidedLine(sector_t *sec)
{
	RDBOOL	line1s = false;
	int		i;

	rl.count = 0;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		rl.mark[rl.count  ] = 1;
		rl.line[rl.count++] = ld;

		if (!ld->backsector)
			line1s = true;
	}

	return line1s && checkRinglist(0, NULL, 0, 2) == sec->linecount;
}

void NF_ContainerRings()
{
	sector_t	*sec;
	line_t		*ld;
	int			i;

	rl.line = (line_t**)M_Malloc(maxsecsz * sizeof(line_t*), 34);
	rl.mark = (short*)M_Malloc(maxsecsz * sizeof(short  ), 35);
	limit_check = maxsecsz;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		sector_t	*osec = NULL;
		int			k, m, cnt2s = 0, lptr = -1;
		int			secnum;
		RDBOOL		movceil, valid;

		if ( (sec->specFlags & 0x1C) != 0x1C
			|| sec->stptr >= 0)
		{
			continue;
		}

		if (   SecCtrl(i)
			&& (movceil = (sec->specFlags & 0xF0000) == 0x20000)
			&& singleSidedLine(sec)								// special case
			&& (secnum = LS_GetContainingSecnum(sec, 1)) >= 0
			&& !((osec = &sectors[secnum])->specFlags & SF_MOVSECTOR) )
		{
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED
					&& origTEX[ld - lines] & 0x22)
				{
					Settype(ld, R3D_03, 47013);
				}
				SetLFG(ld, LFG_NEWSET|LFG_FIXED, 47014);
			}

			Setr3dfh(sec, osec, 47014);
			Setsecdone(sec, 47014);

			if (sec->floorheight > osec->floorheight)
				SetNOocclude(sec, 47014);

			continue;
		}
		
		rl.count = 0;
		valid = true;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (ld->ml_flags & ML_TWOSIDED)
			{
				RDBOOL span = Spans1sLines(ld);

				if (span) valid = false;

				if (   ld->frontsector != ld->backsector
					&& !(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
					&& !(origTEX[ld - lines] & 0x55))
				{
					int vn1 = ld->v1 - vertexes;
					int vn2 = ld->v1 - vertexes;

					if (   vtxarray[vn1] == 3
						&& vtxarray[vn2] == 3
						&& vtoflags[vn1] == 1
						&& vtoflags[vn2] == 1)
					{
						int fdiff = ld->frontsector->floorheight - ld->backsector->floorheight;

						if (abs(fdiff) <= 8*FRACUNIT)
						{
							int cdiff = ld->frontsector->ceilingheight - ld->backsector->ceilingheight;

							// $v62 added test to prevent SetLFG as prevented sectors
							// like mm map 22 sec 356
							// being processed in NF_Check1sLinesBoundedLiquidFloors
							// i.e. it failed in this routine but was then prevented from
							// working in the correct routine further on.
							if (   IsLiquidFloorFlat(ld->frontsector)
								&& ld->frontsector->floorpic == ld->backsector->floorpic)
							{
								// These are dealt with later in NF_Check1sLinesBoundedLiquidFloors
								continue;
							}

							// v220-8 add test for R3D_11
							if (origTEX[ld - lines] == 0x20 && ld->frontsector->heightsec >= 0 && ld->backsector->heightsec < 0)
								Settype(ld, R3D_11, 47015);
							else
							if (origTEX[ld - lines] == 0x02 && ld->frontsector->heightsec < 0 && ld->backsector->heightsec >= 0)
								Settype(ld, R3D_11, 47015);
							else
								Settype(ld, R3D_00, 47015);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 47015);
							continue;
						}

						if (fdiff || ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
						{
							if (!osec)
							{
								osec = ld->frontsector == sec ?
									ld->backsector : ld->frontsector;
							}

							lptr = rl.count;
							rl.mark[rl.count  ] = 1;
							rl.line[rl.count++] = ld; cnt2s++;
						}
					}
				}
			}
			else
			{
				rl.mark[rl.count  ] = 1;
				rl.line[rl.count++] = ld;
			}
		}

		if (osec && osec->specFlags & SF_MOVSECTOR) continue;

		if (valid && cnt2s == 2 && rl.count == 4)
		{
			int bbox[4];

			if (checkRinglist(lptr, bbox, 0, 3))
			{
				int count = 0;

				for (k = 0; k < rl.count; k++)
				{
					if (rl.mark[k] == 4) linearray[count++] = rl.line[k] - lines;
				}

				if (LS_MarkSubsecsInLoop(sec, count, NULL, bbox, MSIA_NONE, NULL))
				{
					validcount++;

					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						SetOverride(ss, osec - sectors, 47016);

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ((ld = sg->linedef) && ld->validcount != validcount)
							{
								ld->validcount = validcount;

								if (ld->ml_flags & ML_TWOSIDED
									&& origTEX[ld - lines] & 0x22)
								{
									Settype(ld, R3D_03, 47016);
								}

								SetLFG(ld, LFG_NEWSET|LFG_FIXED, 47016);
							}
						}
					}
				}
			}
		}
	}

	M_Free(rl.mark);
	M_Free(rl.line);
}

//=============================================================================
//  NF_CheckInheritance()
// b211-0
// Check line inheritance

static RDBOOL validateSec(sector_t *sec)
{
	int		i, k, m;

	ss_validcount++;

	for (i = 0; i < rl.count; i++)
	{
		line_t		*ld, *li = rl.line[i];
		ssline_t	*sl = &sslines[li - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			seg_t		*sg;

			if (ss->ssvalidcount != ss_validcount)
			{
				int x = (int)(ss->midpoint.x * FRACUNIT);
				int y = (int)(ss->midpoint.y * FRACUNIT);
				// Has to be done this way as one can't depend on the
				// seg frontsector assignments.
				sector_t *s = P_PointOnLineSide(x, y, li) ? li->backsector : li->frontsector;

				ss->ssvalidcount = ss_validcount;

				if (s != sec) break;
				
				for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
				{
					if ((ld = sg->linedef) && sg->backsector)
					{
						if (lxarray[ld - lines] & 2) return false;
						if (sg->frontsector->stptr >= 0) return false;
						if (    sg->frontsector == sg->backsector) return true;
						if (    sg->frontsector != sec
							&& (sg->frontsector->specFlags & 0xC) == 0xC) return true;
					}
				}
			}
		}
	}
	return false;
}

static int InspectSsec(sector_t *sec, int ssecidx)
{
	subsector_t *ss = &subsectors[ssecidx];
	seg_t		*sg = &segs[ss->firstline];
	line_t		*ld;
	int			i, k, ret = 0;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if ( (ld = sg->linedef) && ld->ml_flags & ML_TWOSIDED)
		{
			sector_t *fs = ld->frontsector;
			sector_t *bs = ld->backsector;

			if (lineflags[ld - lines] & (LFG_EXCLUDE|LFG_3DTYPE)) return 0;

			if (   fs->specFlags & SF_MOVSECTOR
				|| bs->specFlags & SF_MOVSECTOR)
			{
				continue;
			}

			if (   (fs->specFlags & 0xC) == 0xC
				&& (bs->specFlags & 0xC) == 0xC)
			{
				ret = 1; continue;
			}

			if (fs->floorheight != bs->floorheight
				&& !(fs->floorpic == bs->floorpic && IsLiquidFloorFlat(fs))
				&& !(origTEX[ld - lines] & 0x11))
			{
				return 0;
			}
			
			if (fs->ceilingheight != bs->ceilingheight
				&& !(origTEX[ld - lines] & 0x44))
			{
				return 0;
			}
			
			if (fs == bs)
			{
				if (fs != sec && bs != sec) ret = 1;
			}
			else if (!origTEX[ld - lines]
				&& fs->floorheight   == bs->floorheight
				&& fs->ceilingheight == bs->ceilingheight)
			{
				vertexowner_t *v1 = &vertexowners[ld->v1 - vertexes];
				vertexowner_t *v2 = &vertexowners[ld->v2 - vertexes];

				for (k = 0; k < v1->num; k++)
				{
					if ((sectors[v1->list[k]].specFlags & 0xC) != 0xC) return 0;
					ret = 1;
				}
				for (k = 0; k < v2->num; k++)
				{
					if ((sectors[v2->list[k]].specFlags & 0xC) != 0xC) return 0;
					ret = 1;
				}
			}
		}
	}
	return ret;
}

static RDBOOL secLowMissing(int snum)
{
	sector_t *sec = &sectors[snum];
	int		 i;

	if (!(sec->specFlags & SF_NOLOTEX) || sec->specFlags & SF_SSLINETRUE) return false;

	for (i = 0; i < sec->linecount; i++)
	{
		if ( !(sec->lines[i]->spflags & SPF_CONTROLLINE)
			&& sec->lines[i]->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2)
			&& Vtfl3(sec->lines[i]))
		{
			return true;
		}
	}
	return false;
}

static RDBOOL qualifyingSSEC(subsector_t *ss, short fpic)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	if ((ss->sector->specFlags & 0xC) == 0xC) return true;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->frontsector)
		{
			if (   sg->frontsector != sg->backsector
				&& sg->frontsector->floorpic != fpic)
			{
				return false;
			}
		}
	}
	return true;
}

// v220-5 reject a loop line if the subsector referenced by
//        it has a valid textured line unless all ssec sectors
//        referenced by the line are the same..
static RDBOOL lineValid(int	lnum)
{
	ssline_t *sl = &sslines[lnum];
	sector_t *s = NULL, *s2 = NULL;
	line_t	 *ld; 
	int		 i, k;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (!s) s = ss->sector;
		else
		if (s && s != ss->sector) goto do_checks;
	}
	return true;

do_checks:;
	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && sg->backsector)
			{
				if (sg->backsector->specFlags & SF_MOVFLOOR)
				{
					return false;
				}

				if (  !(origTEX[ld - lines] & 0x44)
					&& (sg->frontsector->ceilingpic == skyflatnum)^(ld->backsector->ceilingpic == skyflatnum))
				{
					return true;
				}
				if (  !(origTEX[ld - lines] & 0x11)
					&& (sg->frontsector->floorpic == skyflatnum)^(ld->backsector->floorpic == skyflatnum))
				{
					return true;
				}

				if (   sg->frontsector == lines[lnum].frontsector
					|| sg->frontsector == lines[lnum].backsector)
				{
					s2 = sg->frontsector;
				}
			}
		}

		sg = &segs[ss->firstline];


		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (  (ld = sg->linedef)
				&& ld->ml_flags & ML_TWOSIDED)
			{
				if (sg->frontsector == s2)
				{
					if (ld->frontsector->ceilingheight != ld->backsector->ceilingheight
						&& origTEX[ld - lines] & 0x44)
					{
							return false;
					}
					if (ld->frontsector->floorheight != ld->backsector->floorheight
						&& origTEX[ld - lines] & 0x11)
					{
						return false;
					}
				}
			}
		}
	}
	return true;
}

static RDBOOL same_type;	// set true only if two different sectors match

// v220-5
static RDBOOL sameLineSecType(line_t *li)
{
	if (li->frontsector == li->backsector)
	{
		same_type = false;
		return true;
	}

	same_type = subfn_EquivalentSec(li->frontsector, li->backsector, true);

	return same_type;
}

// v220-10
void NF_CheckDualSectorTypeLineRings()
{
	sector_t	*s;
	line_t		*ld;
	int			c, i, k, m;
	int			rc;
	int			box[4];

	CLR_hitline();

	for (c = i = 0; i < numsectors; i++)
	{
		if ((sectors[i].specFlags & 0xC) != 0xC) continue;

		for (k = 0; k < sectors[i].linecount; k++)
		{
			ld = sectors[i].lines[k];
			m = ld - lines;

			if (   ld->ml_flags & ML_TWOSIDED
				&& ld->frontsector != ld->backsector
				&& !lineflags[m]
				&& !hitline[m]
				&& !(origTEX[m] & 0x22)
				&& !lxarray[m]
				&& (ld->frontsector->specFlags & 0x3000C) == 0xC
				&& Vtxs2(ld)
				&& sameLineSecType(ld))
			{
				hitline[m] = 1; c++;
			}
		}
	}

	if (c < 3) return;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   hitline[i] == 1
			&& (s = ld->frontsector)
			&& subfn_CheckForRing(hitline, s, i, &rc, box, NULL, NULL))
		{
			sector_t	*chk;
			sector_t	*ref;
			seg_t		*sg;
			RDBOOL		valid = false;
			int			snum;
			int			newsec = -1;
			int			inner;
			int			a = 0, p;

			inner = LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_TESTAREA);
			
			// set chk to the outer sector
			chk = inner ? ld->backsector : s;
			// set ref to the inner sector
			ref = inner ? s : ld->backsector;

			if (chk != s)
			{
				for (k = 0; k < chk->linecount; k++)
				{
					int lnum = chk->lines[k] - lines;

					if (hitline[lnum] == 4) linearray[a++] = lnum;
				}

				if (a != rc) goto end_check;
			}

			ss_validcount++;
			hitssecptr = 0;

			for (k = 0; k < a; k++)
			{
				ssline_t *sl = &sslines[linearray[k]];

				for (m = 0; m < sl->num; m++)
				{
					subsector_t *ss = &subsectors[sl->list[m]];

					if (ss->ssvalidcount == ss_validcount) continue;

					ss->ssvalidcount = ss_validcount;

					if (ss->sector == ref) continue;

					hitssec[hitssecptr++] = sl->list[m];

					sg = &segs[ss->firstline];

					for (p = 0; p < ss->linecount; p++, sg++)
					{
						if (!sg->frontsector || sg->frontsector == s || sg->frontsector == chk) continue;

						snum = sg->frontsector - sectors;

						if (newsec < 0)
						{
							if (sectors[snum].specFlags & SF_MOVSECTOR)
								goto end_check;

							if ((sectors[snum].specFlags & 0x40C) == 0xC)
								goto end_check;

							newsec = snum;
						}
						else if (newsec != snum)
						{
							goto end_check;
						}
					}
				}
			}

			if (newsec >= 0)
			{
				int ssecnum = hitssecptr;
				int *storessec = (int*)malloc(ssecnum*sizeof(int));

				memcpy(storessec, hitssec, ssecnum*sizeof(int));

				// set chk to the inner sector
				chk = inner ? s : ld->backsector;

				if (   LS_MarkSubsecsInAreaGeneral(chk, hitline, 0, NULL, box, -1, MSIA_NONE)
					&& LS_OnlySSLinesInSSECarray(NULL, hitline))
				{
					// Set the inner and outer subsectors
					for (k = 0; k < ssecnum   ; k++) SetOverride(&subsectors[storessec[k]], newsec, 47017);
					for (k = 0; k < hitssecptr; k++) SetOverride(&subsectors[hitssec[k]]  , newsec, 47017);

					valid = true;
				}

				M_Free(storessec);
			}
end_check:;

			for (k = 0; k < s->linecount; k++)
			{
				line_t *li = s->lines[k];

				if (valid) SetLFG(li, LFG_NEWSET|LFG_FIXED, 47017);

				if (hitline[li - lines] == 4) hitline[li - lines] = 0;
			}

			if (valid) Setsecdone(s, 47017);
			if (valid) Setsecdone(chk, 47017);
		}
		else
		{
			hitline[i] = 0;
		}
	}
}

static void chkSameHeightLineInheritance(subsector_t *ss, int floorsec, int ceilsec)
{
	subsector_t	*so;
	seg_t		*sg;
	int			i, k;

	if (floorsec < 0 && ceilsec < 0) return;

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		line_t *ld = sg->linedef;

		if (ld && ld->backsector)
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				so = &subsectors[sl->list[k]];

				if (   floorsec >= 0
					&& ld->frontsector->floorheight == ld->backsector->floorheight
					&& so->floorsec != floorsec
					&& (so->sector->specFlags & 0x10004) == 0x4
					&& so->sector->floorpic == ss->sector->floorpic
					&& so->sector->floorheight < sectors[floorsec].floorheight)
				{
					Setr3dssfs(so, floorsec, 47073);
				}

				if (   ceilsec >= 0
					&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight
					&& so->ceilsec != ceilsec
					&& (so->sector->specFlags & 0x20008) == 0x8
					&& so->sector->ceilingpic == ss->sector->ceilingpic
					&& so->sector->ceilingheight > sectors[ceilsec].ceilingheight)
				{
					Setr3dssfs(so, ceilsec, 47074);
				}
			}
		}
	}
}

static RDBOOL subfn_prelimCheck(sector_t *sc, int *secnum, int *count)
{
	ssline_t	*sl;
	sector_t	*os;
	line_t		*ld;
	int			i, k, m;
	int			lnum;

	ss_validcount++;
	hitssecptr = 0;

	if (count) validcount++;

	for (i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		lnum = ld - lines;

		if (hitline[lnum] != 4) continue;

		sl = &sslines[lnum];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				seg_t	*sg = &segs[ss->firstline];

				ss->ssvalidcount = ss_validcount;

				hitssec[hitssecptr++] = sl->list[k];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg->linedef && sg->frontsector)
					{
						if (count)
						{
							if (sg->linedef->validcount == validcount) continue;
							sg->linedef->validcount = validcount;
							linearray[(*count)++] = sg->linedef - lines;
						}
						else if (sg->frontsector != ld->frontsector)
						{
							if (!os)
							{
								os = sg->frontsector;

								if (  (os->specFlags & 0x3040C) != 0xC	// v220-20 was 0x4
									|| os->floorheight > sc->floorheight)
								{
									return false;
								}
							}
							else if (os != sg->frontsector) return false;
						}
					}
				}
			}
		}
	}

	if (count) return true;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->backsector
				&& sg->backsector->specFlags & 0x400)
			{
				*secnum = sg->backsector - sectors;
				return true;
			}
		}
	}

	return false;
}

static void subfn_CheckRingInheritance(sector_t *sc)
{
	line_t		*ld;
	int			c, i, k, m;
	int			lnum, rc;
	int			box[4], chksec[1];
	int			losec, hisec, ref;
	// ---------------------------


	ss_validcount++;

	memcpy(excludeline, hitline, numlines);

	// v220-18 added (deal with simple cases first)
	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		lnum = ld - lines;

		if (hitline[lnum] != 1) continue;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
		{
			if ((sc->specFlags & 0x3040C) == 0xC && subfn_prelimCheck(sc, &hisec, NULL))
			{
				int count = 0;

				subfn_prelimCheck(sc, NULL, &count);

				if (sectors[hisec].floorheight - sc->floorheight > 8*FRACUNIT)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						SetOverride(&subsectors[hitssec[k]], hisec, 47075);
					}
				}

				for (k = 0; k < count; k++)
				{
					if (!origTEX[linearray[k]])
					{
						Settype(&lines[linearray[k]], R3D_01, 47075);
					}

					if (hitline[linearray[k]]) hitline[linearray[k]] = excludeline[linearray[k]] = 0;

					SetLFG(&lines[linearray[k]], LFG_NEWSET|LFG_FIXED, 47075);
				}
			}
		}
	}

	memcpy(hitline, excludeline, numlines);

	ss_validcount++;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		lnum = ld - lines;

		if (hitline[lnum] != 1) continue;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
		{
			int			lowestsec = -1;
			RDBOOL		valid;

			XS_GetBoundsFromLine(ld, &losec, &hisec, &ref);

			if (   losec == hisec
				&& losec >= 0
				&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
				&& sectors[losec].floorheight == sc->floorheight
				&& sectors[losec].ceilingheight == sc->ceilingheight)
			{
				continue;
			}
			
			valid = hisec >= 0 && losec >= 0 && ref >= 0;

			if (valid)
			{
				ssline_t *sl = &sslines[ref];

				lowestsec =
					sectors[hisec].floorheight < sectors[losec].floorheight ?
														hisec : losec;

				// v220-5 check if there's a line with the same ceiling height
				// and a missing low tex. The missing side will be the hisec required.
				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];
					line_t		*li;
					seg_t		*sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (  (li = sg->linedef)
							&& li->backsector
							&& li->backsector != li->frontsector
							&& !(origTEX[li - lines] & 0x11)
							&& li->backsector->floorheight != li->frontsector->floorheight
							&& li->backsector->ceilingheight == li->frontsector->ceilingheight)
						{
							if (li->backsector->floorheight > li->frontsector->floorheight)
							{
								if ((li->frontsector->specFlags & 0xC) == 0xC)
								{
									hisec = li->frontsector - sectors;
									goto hisec_exit;
								}
							}
							else
							{
								if ((li->backsector->specFlags & 0xC) == 0xC)
								{
									hisec = li->backsector - sectors;
									goto hisec_exit;
								}
							}
						}
					}
				}
hisec_exit:;
			}
			else
			{
				hitline[lnum] = 0; continue;
			}

			for (c = k = 0; k < sc->linecount; k++)
			{
				int lnum = sc->lines[k] - lines;

				if (hitline[lnum] == 4)
				{
					linearray[c++] = lnum;

					if (valid)
					{
						ssline_t *sl = &sslines[lnum];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->ssflags & SSF_NOCHANGE) continue;

							if (ss->ssvalidcount == ss_validcount) continue;

							ss->ssvalidcount = ss_validcount;

							if (linetable[ref] & 16)
							{
								int x = (int)(ss->midpoint.x * FRACUNIT);
								int y = (int)(ss->midpoint.y * FRACUNIT);
								int side = P_PointOnLineSide(x, y, ld);
								sector_t *chk = side ? ld->backsector : ld->frontsector;

								if (ss->sector != chk) ChangeSSECsec(ss, chk, 47022);
							}

							Setr3dsscs(ss, hisec, 47023);

							ss->ssflags |= SSF_NOCHANGECS;

							if (secLowMissing(losec))
								Setr3dssfs(ss, lowestsec, 47024);
							else
								Setr3dssfs(ss, losec, 47025);
						}
					}
					
					hitline[lnum] = 8;

					if (!origTEX[lnum])
					{
						Settype(sc->lines[k], R3D_01, 47026);
					}

					SetLFG(sc->lines[k], LFG_NEWSET|LFG_FIXED, 47027);
				}
			}

			if (valid)	// check subsecs inside
			{
				chksec[0] = sc - sectors;
	
				if (  (LS_MarkSubsecsInMultiSecLoop(0, chksec, rc, box, NULL, true, MSIA_NONE)
					|| LS_MarkSubsecsInMultiSecLoop(0, chksec, rc, box, NULL, true, MSIA_REVERSED))
					&& hitssecptr)
				{
					subsector_t	**ssecmem = (subsector_t**)M_Malloc(hitssecptr*sizeof(subsector_t*), 36);
					subsector_t *ss;
					seg_t		*sg;
					RDBOOL		found;
					byte		txn;
					int			side;
					int			subsecmemcnt = 0;
					int			flag = 0;

					// 1) The subsectors must not have a 1s line.
					// 2) All the subsecs inside must own sectors with no upper or lower
					//    textures or, where there are upper or lower textures, that the
					//    line's tex sector is the same as the replacememt sector.

					// v220-18 modified
					for (k = 0; k < hitssecptr; k++)
					{
						ss = &subsectors[hitssec[k]];

						if ( (ss->sector->specFlags & 0x400) ) { subsecmemcnt = 0; break; }
						if (!(ss->sector->specFlags & 0x4)   ) { flag |= 1; }
						if (!(ss->sector->specFlags & 0x8)   ) { flag |= 2; }

						// v220-21 Check for a seg having a lower backsector floor with a lo tex

						for (found = false, m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
						{
							if ( !(ld = sg->linedef) ) continue;
							if (!sg->backsector) continue;	// insurance
							if (sg->frontsector->floorheight <= sg->backsector->floorheight) continue;
							if ( !(txn = (origTEX[ld - lines] & 0x11)) ) continue;

							side = ld->frontsector == sg->frontsector ? 0 : 1;

							if ((side && txn == 0x01) || (!side && txn == 0x10))
							{
								found = true; break;
							}
						}

						if (!found) ssecmem[subsecmemcnt++] = ss;
					}

					if (subsecmemcnt && flag)
					{
						for (k = 0; k < hitssecptr; k++)
						{
							ss = &subsectors[hitssec[k]];
							sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ( !(ld = sg->linedef) ) continue;
								if ( !( (txn = origTEX[ld - lines]) & 0x55) ) continue;

								if (flag & 1)
								{
									if (txn & 0x01 && ld->frontsector - sectors != losec)
									{
										subsecmemcnt = 0;
										break;
									}
									if (txn & 0x10 && ld->backsector - sectors != losec)
									{
										subsecmemcnt = 0;
										break;
									}
								}

								if (flag & 2)
								{
									if (txn & 0x04 && ld->frontsector - sectors != hisec)
									{
										subsecmemcnt = 0;
										break;
									}
									if (txn & 0x40 && ld->backsector - sectors != hisec)
									{
										subsecmemcnt = 0;
										break;
									}
								}
							}
							if (!subsecmemcnt) break;
						}
					}

					for (k = 0; k < subsecmemcnt; k++)
					{
						if (ssecmem[k]->sector->heightsec < 0)	// v220-35 added $hschange
						{
							Setr3dssfs(ssecmem[k], losec, 47028);
							Setr3dsscs(ssecmem[k], hisec, 47028);
						}
					}

					for (k = 0; k < subsecmemcnt; k++)
					{
						chkSameHeightLineInheritance(ssecmem[k], losec, hisec);
					}

					M_Free(ssecmem);
				}
			}
		}
	}
}

// b211-0 v220-18
void NF_CheckRingInheritance()
{
	line_t	*ld;
	int		c, i;

	CLR_hitline();
	SET_countsec(0);

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   !lineflags[i]
			&& ld->frontsector == ld->backsector
			&& ld->frontsector->heightsec < 0
			&& !lxarray[i]
			&& Vtxs2(ld)
			&& lineValid(i))
		{
			countsec[ld->frontsector - sectors]++;
			hitline[i] = 1;
			c++;
		}
	}

	if (c < 3) return;

	for (i = 0; i < numsectors; i++)
	{
		if (countsec[i] > 2) subfn_CheckRingInheritance(&sectors[i]);
	}
}

//---------------------------------------------------

static sector_t *getOwningSector(sector_t *s, int *cl)
{
	sector_t	*osec = NULL;
	line_t		*ld;
	int			c, i, k, m;

	hitssecptr = 0;
	ss_validcount++;
	validcount++;

	for (c = i = 0; i < s->linecount; i++)
	{
		int lnum = s->lines[i] - lines;

		if (hitline[lnum] == 4)
		{
			ssline_t *sl = &sslines[lnum];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					seg_t		*sg = &segs[ss->firstline];

					ss->ssvalidcount = ss_validcount;

					hitssec[hitssecptr++] = sl->list[k];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if ( (ld = sg->linedef) && ld->validcount != validcount)
						{
							if (ld->frontsector != ld->backsector)
							{
								return NULL;
							}

							ld->validcount = validcount;

							if (hitline[ld - lines] != 4)
							{
								if (   sg->frontsector == s
									|| !(origTEX[ld - lines] & 0x22)
									||  (origTEX[ld - lines] & 0x55))
								{
									return NULL;
								}

								if (!osec) osec = sg->frontsector;
								else if (osec != sg->frontsector) return NULL;

							}

							linearray[c++] = ld - lines;
						}
					}
				}
			}
		}
	}

	*cl = c;
	return osec;
}

// Look for same sector sided textureless line loop and check
// if it is solely referenced by other same sector sided lines
// only having mid textures.
void NF_CheckSharedVertexRing()
{
	line_t	*ld;
	int		c, i;

	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& ld->frontsector->heightsec < 0
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& !lxarray[i]
			&& !lineflags[i]
			&& !origTEX[i])
		{
			hitline[i] = 1; c++;
		}
	}

	if (c < 3) return;

	while(1)
	{
		sector_t	*s, *osec;
		int			rc, box[4];
		int			lnum = -1;
		int			cl, k;
		//-------------------

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (hitline[i] == 1) { lnum = i; s = lines[i].frontsector; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			if ( (osec = getOwningSector(s, &cl)) )
			{
				for (k = 0; k < hitssecptr; k++)
					SetOverride(&subsectors[hitssec[k]], osec - sectors, 48051);

				for (k = 0; k < cl; k++)
				{
					ld = &lines[linearray[k]];

					if (origTEX[linearray[k]])
						Settype(ld, (byte)(ld->backsector ? R3D_03: R3D_00), 48052);
					else
						Settype(ld, R3D_3F, 48053);

					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 48054);
				}
			}

			for (k = 0; k < s->linecount; k++)
			{
				lnum = s->lines[k] - lines;

				if (hitline[lnum] == 4) hitline[lnum] = 0;
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

void NF_CheckInheritance()
{
	sector_t	*sec;
	int			i, k, m, bbox[4];
	byte		*excludessecs = (byte*)W_Calloc(numsubsectors, TEMPID, 0);
	int			*listssecs = (int*)W_Malloc(numsubsectors * sizeof(int), TEMPID, 4);

	rl.line = (line_t**)W_Malloc(maxsecsz * sizeof(line_t*), TEMPID, 5);
	rl.mark = (short*)W_Malloc(maxsecsz * sizeof(short), TEMPID, 6);
	limit_check = maxsecsz;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
		//	&& (sec->specFlags & 0xC) != 0xC
		// v220-17 replace with;
			&& (!(sec->specFlags & 0xC) || sec->specFlags & 0x400)	// != 0xC
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& !SecCtrl(i)
			&& IsLiquidFloorFlat(sec)
			&& sec->stptr < 0)
		{
			int sscount = 0;

			rl.count = 0;

			memset(rl.mark, 0, sec->linecount * sizeof(short));

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED)
				{
					if (   ld->frontsector == ld->backsector
						|| lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
						continue;
				}

				rl.mark[rl.count  ] = 1;
				rl.line[rl.count++] = ld;
			}

			if (rl.count < 3 || !validateSec(sec)) continue;

			for (k = 0; k < rl.count; k++)
			{
				if (rl.mark[k] == 1 && checkRinglist(k, bbox, 0, 4))
				{
					int		count, flag = MSIA_NONE;
					RDBOOL	valid = true;

					for (count = 0, m = 0; m < rl.count; m++)
					{
						if (rl.mark[m] == 4)
						{
							linearray[count++] = rl.line[m] - lines;
						}
					}

					if (!LS_MarkSubsecsInLoop(sec, count, NULL, bbox, MSIA_TESTAREA, NULL))
						flag = MSIA_REVERSED;

					if (flag == MSIA_NONE
						&& rl.line[k]->ml_flags & ML_TWOSIDED)
					{
						int side = rl.line[k]->frontsector == sec ? 0 : 1;
						int fdiff = rl.line[k]->frontsector->floorheight
							- rl.line[k]->backsector->floorheight;

						if ((fdiff > 0 && !side) || (fdiff < 0 && side))
						{
							valid = false;
						}
					}

					// Restore the array
					for (count = 0, m = 0; m < rl.count; m++)
					{
						if (rl.mark[m] == 4)
						{
							rl.mark[m] = 8;	// Can now mark line as done
							linearray[count++] = rl.line[m] - lines;
						}
					}

					// This should not fail
					if (LS_MarkSubsecsInLoop(sec, count, NULL, bbox, flag, NULL))
					{
						for (m = 0; m < hitssecptr; m++)
						{
							int type = InspectSsec(sec, hitssec[m]);

							if (!valid || flag == MSIA_REVERSED)
							{
								excludessecs[hitssec[m]] = true;
							}
							else if (type == 1)
								listssecs[sscount++] = hitssec[m];
						}
					}
				}
			}

			if (sscount)
			{
				short	fpic = IsLiquidFloorFlat(sec) ? sec->floorpic : -1;

				for (k = 0; k < sscount; k++)
				{
					int ssecidx = listssecs[k];

					if ((uint)ssecidx >= (uint)numsubsectors)
						Con_ErrorD("NF_CheckInheritance: idx error");

					if (!excludessecs[ssecidx])
					{
						subsector_t *ss = &subsectors[ssecidx];

						if (qualifyingSSEC(ss, fpic))
						{
							seg_t *sg = &segs[ss->firstline];

							if (sec->heightsec >= 0)
							{
								SetOverride(ss, i, 47029);
							}
							else
							{
								if (sec->floorheight < ss->sector->floorheight
									|| ss->sector->specFlags & SF_NOLOTEX)
									Setr3dssfs(ss, i, 47030); 
								if (sec->ceilingheight != ss->sector->ceilingheight
									|| ss->sector->specFlags & SF_NOHITEX)
									Setr3dsscs(ss, i, 47031);
							}

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								line_t *ld = sg->linedef;

								if (ld && !(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)))
								{
									if (ld->frontsector == ld->backsector)
									{
										int txn = origTEX[ld - lines];

										if (!txn)
											Settype(ld, R3D_01, 47032);
										else if (txn & 0x22)
											Settype(ld, R3D_03, 47033);
										else
											Settype(ld, R3D_00, 47034);

										SetLFG(ld, LFG_NEWSET|LFG_FIXED, 47035);
									}
								}
							}
						}
					}
				}
			//	Setsecdone(sec, 47035);
			}
		}
	}

	W_FreeALL(TEMPID);
}

// ================ NF_CheckInheritance2
// v220-4 checks for closed textureless support sectors

static RDBOOL checkAllSameSec(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (   ld->frontsector != ld->backsector
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOONE)
			|| !Vtxs2(ld)
			|| lxarray[ld - lines])
		{
			return false;
		}
	}
	return true;
}

static int getCommonSector(sector_t *sec, RDBOOL *allss, RDBOOL type)
{
	subsector_t	*ss;
	ssline_t	*sl;
	seg_t		*sg;
	line_t		*ld;
	int			secnum = -1;

	int			i, k, m;

	if (allss) *allss = true;
	ss_validcount++;

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines, snum;

		sl = &sslines[lnum];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;

				sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (type)
					{
						if ( (ld = sg->linedef) && (snum = lnsec[ld - lines]) >= 0)
						{
							if (secnum < 0) secnum = snum;
							else if (secnum != snum) return -1;
						}
					}
					else
					{
						if (sg->frontsector != sg->backsector)
						{
							if (allss) *allss = false;
							if (secnum < 0) secnum = sg->frontsector - sectors;
							else if (secnum != sg->frontsector - sectors) return -1;
						}
					}
				}
			}
		}
	}

	return secnum;
}
		
static int getCeilingSec(sector_t *sc)
{
	sector_t	*os;
	line_t		*ld;
	int			i;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->mtflags & R3D_MISHITEX)
		{
			os = ld->frontsector == sc ? ld->backsector : ld->frontsector;

			if (os->ceilingheight < sc->ceilingheight) return os - sectors;
		}
	}

	return -1;
}

void NF_CheckInheritance2()
{
	sector_t	*sec;
	int			i, k, m;
	int			count, secnum;
	RDBOOL		isset = false;
	RDBOOL		allss;

	for (count = i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& sec->linecount > 2
			&& (sec->specFlags & 0x3000C) == 0xC
			&& checkAllSameSec(sec))
		{
			secnum = getCommonSector(sec, &allss, false);

			if (secnum >= 0)
			{
				int f = sectors[secnum].specFlags;
				int fsec = (!(f & SF_NOLOTEX) || f & SF_SSLINETRUE) ? secnum : -1;
				int csec = (!(f & SF_NOHITEX) || f & SF_SSLINETRUE) ? secnum : -1;

				if (fsec < 0) continue;	// don't bother

				// v220-20 see if correct csec can be divined
				if (csec < 0)
				{
					csec = getCeilingSec(&sectors[fsec]);
				}

				for (k = 0; k < sec->linecount; k++)
				{
					ssline_t *sl = &sslines[sec->lines[k] - lines];

					lnsec[sec->lines[k] - lines] = secnum;

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->floorsec < 0 && ss->ceilsec < 0)
						{
							Setr3dssfs(ss, secnum, 47036);
							if (csec >= 0) Setr3dsscs(ss, csec, 47036);
							isset = true;
						}
					}
				}
			}
			else if (allss)
			{
				countsec[count++] = i;
			}
		}
	}

	if (isset)
	{
		for (i = 0; i < count; i++)
		{
			sec = &sectors[countsec[i]];

			if ( (secnum = getCommonSector(sec, NULL, true)) >= 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					ssline_t *sl = &sslines[sec->lines[k] - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->floorsec < 0 && ss->ceilsec < 0)
						{
							SetOverride(ss, secnum, 47080);
							isset = true;
						}
					}
				}
			}
		}
	}
}

//=======================================================================
// NF_ContainRingsFINAL()
//

void NF_ContainRingsFINAL()
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->linecount > 2
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& sec->stptr < 0
			&& sec->specFlags & SF_NOHITEX)
		{
			line_t	*lb = NULL;
			int		k, rc, bbox[4];
			RDBOOL	valid = true;

			CLR_vtxcheck();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t	 *ld = sec->lines[k];
				sector_t *s;
				int		 side;

				if (ld->spflags & SPF_CONTROLLINE)
				{
					valid = false; break;
				}

				if ( !(ld->ml_flags & ML_TWOSIDED)
					|| lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)
					|| ld->frontsector == ld->backsector)
				{
					valid = false; break;
				}
				
				side = ld->frontsector == sec ? 0 : 1;

				s = side ? ld->frontsector : ld->backsector;

				if (sec->floorheight < s->floorheight)
				{
					if (lb)
					{ valid = false; break; }
					else
						lb = ld;
				}

				if (sec->floorheight > s->floorheight)
				{
					if (   ( side && !(origTEX[ld - lines] & 0x01))
						|| (!side && !(origTEX[ld - lines] & 0x10)) )
					{
						if (lb)
						{ valid = false; break; }
						else
							lb = ld;
					}
				}

				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
			}

			if (!valid) continue;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t	 *ld = sec->lines[k];

				if (   vtxcheck[ld->v1 - vertexes] != 2
					|| vtxcheck[ld->v2 - vertexes] != 2)
				{
					valid = false; break;
				}
				hitline[ld - lines] = 1;
			}

			if (   valid
				&& subfn_CheckForRing(hitline, sec, sec->lines[0] - lines, &rc, bbox, NULL, NULL)
				&& rc == sec->linecount
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE)
				&& LS_SetSSINF(sec, hitline, NULL, NULL, false, false))
			{
				int		m;

				if (   ssinf.othersslinecnt
					&& !ssinf.othersslinechk
					&& ssinf.anysameseclines == ssinf.anylines)
				{
					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];

						// Test added for SetOverride
						if ((ss->sector->specFlags & 0x3000C) == 0x1000C
							&& ss->sector->ceilingpic != skyflatnum)
							SetOverride(ss, i, 47081);
						else
							Setr3dssfs(ss, i, 47081);

						LS_SetLines(sec, ss, false, false, false, 47037);
					}

					Setsecdone(sec, 47038);

					// Check if logical sector needs changing
					if (ssinf.othersslinecnt == ssinf.anylines
						&& !ssinf.othersslinetex)
					{
						rl.line = (line_t**)M_Malloc(ssinf.othersslinecnt * sizeof(line_t*), 37);
						rl.mark = (short*)M_Malloc(ssinf.othersslinecnt * sizeof(short), 38);
						limit_check = ssinf.othersslinecnt;

						rl.count = 0;

						for (m = 0; m < ssinf.othersslinecnt; m++)
						{
							rl.line[m] = &lines[ssinf_otherlineptr[m]];
							rl.mark[m] = 1;
							rl.count++;
						}

						if (checkRinglist(0, NULL, 0, 5))
						{
							sector_t *logical = rl.line[0]->frontsector;

							for (m = 0; m < hitssecptr; m++)
							{
								subsector_t *ss = &subsectors[hitssec[m]];

								if (ss->sector != logical)
									ChangeSSECsec(ss, logical, 47039);
							}
						}

						M_Free(rl.line);
						M_Free(rl.mark);
					}
				}
			}
		}
	}
}

// Find rings containing same sided lines
void NF_ContainRingsCTRL()
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& (sec->specFlags & 0xC) != 0xC
			&& SecCtrl(i)
			&& sec->stptr < 0)
		{
			int		c, k, m;
			int		bbox[4];
			RDBOOL	valid = true, lfgmarked = false;

			CLR_vtxcheck();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
				{
					valid = false; break;
				}

				if (  !(ld->ml_flags & ML_TWOSIDED)
					|| ld->frontsector == ld->backsector)
				{
					valid = false; break;
				}

				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
			}

			if (!valid) continue;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (lineflags[ld-lines] & (LFG_NEWSET|LFG_3DTYPE))
					lfgmarked = true;

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (   vtxcheck[ld->v1 - vertexes] != 2
					|| vtxcheck[ld->v2 - vertexes] != 2)
				{
					valid = false; break;
				}

				hitline[ld - lines] = 1;
			}

			if (!valid) continue;

			validcount++;

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   hitline[ld - lines] == 1
					&& subfn_CheckForRing(hitline, sec, ld - lines, NULL, bbox, NULL, NULL))
				{
					if (   LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE)
						&& LS_SetSSINF(sec, hitline, NULL, NULL, false, false))
					{
						if (ssinf.anylines && ssinf.anylines == ssinf.anysameseclines)
						{
							for (m = 0; m < hitssecptr; m++)
							{
								subsector_t *ss = &subsectors[hitssec[m]];
								
								SetOverride(ss, i, 47040);
								LS_SetLines(sec, ss, false, false, false, 47041);
							}
							c++;
						}
					}
					for (m = 0; m < sec->linecount; m++)
					{
						ld = sec->lines[m];

						if (hitline[ld - lines] == 4)
							hitline[ld - lines] = 8;
					}
				}
			}

			if (c)
				Setsecdone(sec, 47042);
		}
	}
}

//-----------------------------------------

// v220-18 Look for simple sectors with lo missing
//         that could have the wrong floorsec set
static void subfn_CheckFloorSecWithMissingReset(sector_t *sc)
{
	sector_t	*ch, *os;
	ssline_t	*sl;
	line_t		*ld;
	int			box[4];
	int			i, rc;

	for (i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		if (ld->specType != R3D_00) return;
		if (ld->mtflags & R3D_MISHITEX) return;
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
		if (!ch) return;
		if (origTEX[ld - lines] & 0x11) return;
		if (ch->floorheight - sc->floorheight <= 8*FRACUNIT) return;
		if (!os)
		{
			if (ch->specFlags & SF_MOVFLOOR) return;
			os = ch;
		}
		else if (os != ch
			&& (ch->specFlags & SF_MOVFLOOR
			|| os->floorpic != ch->floorpic
			|| os->floorheight != ch->floorheight
			|| os->floorlightsec != ch->floorlightsec) )
		{
			return;
		}
	}

	if (!os) return;

	// Find a subsector referencing all the sector's lines
	sl = &sslines[sc->lines[0] - lines];

	CLR_hitline();

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		hitline[ld - lines] = 1;
	}

	if (subfn_CheckForRing(hitline, sc, sc->lines[0] - lines, &rc, box, NULL, NULL) && rc == sc->linecount)
	{
		// v220-22 Getting ssecs from the lines can fail, unfortunately, so
		//         they must be obtained from the area bounded by the ring.
		if (LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE))
		{
			for (i = 0; i < hitssecptr; i++)
			{
				subsector_t	*ss = &subsectors[hitssec[i]];
				Setr3dsslfsec(ss, os - sectors, 47076);
				Setsecdone(sc, 47076);
			}
		}
	}
}

static void setContainedSSECS(int p, int c)
{
	sector_t *s;
	int		 cs = countsec[0];
	int		 i, k, m;

	if (countsec[p] < 0) return;

	s = &sectors[countsec[p]];

	for (i = 0; i < s->linecount; i++)
	{
		line_t	 *li, *ld  = s->lines[i];
		ssline_t *sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			RDBOOL		setlf;

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			setlf = ld->frontsector == ld->backsector;

			if (!setlf)
			{
				for (m = 1; m < c; m++)
				{
					if (ss->sector - sectors == countsec[m])
					{
						setlf = true; break;
					}
				}
			}

			if (setlf)
			{
				seg_t *sg = &segs[ss->firstline];

				// v220-10 $light replaced
			//	ss->sector->floorlightsec = ss->sector->ceilinglightsec = cs;
				if (ss->floorsec >= 0) Setr3dssfs(ss, -1, 47046);
				if (ss->ceilsec  >= 0) Setr3dsscs(ss, -1, 47046);
				SetFloorlightsec(ss, cs, 47046);

				if (ss->sector->specFlags & SF_MOVFLOOR)
				{
					Setr3dssfs(ss, cs, 47046);
					Setr3dsslfsec(ss, ss->sector - sectors, 47046);
				}
				else
					Setr3dsslfsec(ss, cs, 47046);

				if (cs < 0 || sectors[cs].ceilingpic != skyflatnum)
					Setr3dsslcsec(ss, cs, 47047);
			
				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (!(li = sg->linedef) || li->validcount == validcount)
						continue;

					li->validcount = validcount;

					if (!(li->ml_flags & ML_TWOSIDED)) continue;

					if (origTEX[li - lines] == 0x22)
						Forcetype(li, R3D_03, 47048);
					else
					{
						Forcetype(li, R3D_00, 47048);
					}

					Clr_allmtflags(li, 47050);
					SetLFG(li, LFG_FIXED|LFG_NEWSET|LFG_EXCLUDE, 47048);
				}
			}
		}
	}
	Setsecdone3(s, 49048);
}

static RDBOOL secContained(int idx, int *cbox)
{
	sector_t *s = &sectors[idx];
	int		 i, box[4];

	M_ClearBox(box);

	for (i = 0; i < s->linecount; i++)
	{
		if (s->lines[i]->spflags & SPF_CONTROLLINE)
			continue;

		M_AddLineToBox(box, s->lines[i]);
	}

	return M_FitBox(box, cbox);
}

// b211-0  This depends on control lines having been set
void NF_EnclosedFalseFloors()
{
	sector_t *sec;
	int		 i, k;
	int		 mask1 = SF_SECCLOSED|SF_NOHITEX;
	int		 mask2 = mask1|SF_MOVSECTOR;
	int		 seccnt = 0;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->ceilingpic == skyflatnum
			&& !SecCtrl(i)
			&& sec->stptr < 0
			&& (sec->specFlags & mask2) == mask1
			&& sec->linecount > 5)
		{
			int		rc, box[3][4];
			int		c0 = 0, c1 = 0, miss = 0;
			RDBOOL	valid = true;

			CLR_hitline();

			// Check if this splits into more than one ring
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED
					&& (origTEX[ld - lines] & 0x22
					|| ld->frontsector->floorheight == ld->backsector->floorheight))
				{
					valid = false; break;
				}

				// There must be at least 2 missing lower textures with
				// a height difference of > 8
				if (   ld->ml_flags & ML_TWOSIDED
					&& abs(ld->frontsector->floorheight - ld->backsector->floorheight) > 8*FRACUNIT
					&& !(origTEX[ld - lines] & 0x11) )
				{
					miss++;
				}

				hitline[ld - lines] = 1;
			}

			if (!valid || miss < 2) continue;

			valid = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (hitline[ld - lines] == 1
					&& subfn_CheckForRing(hitline, sec, ld - lines, &rc, box[c0], NULL, NULL))
				{
					c0++; c1 += rc;
					if (c0 > 1) break;
				}
			}

			if (   c0 == 2
				&& c1 == sec->linecount
				&& !M_FitBox(box[0], box[1])
				&& !M_FitBox(box[1], box[0]))
			{
				// Get all sectors with no high textures between the box limits
				// NOTE: the lines must be wholly inside the box.
				SET_countsec(0);
				c0 = 0;
				countsec[c0++] = i;	// put the reference sector at the start

				if (box[1][BOXRIGHT] > box[0][BOXRIGHT]) box[0][BOXRIGHT] = box[1][BOXRIGHT];
				if (box[1][BOXLEFT ] < box[0][BOXLEFT ]) box[0][BOXLEFT ] = box[1][BOXLEFT ];
				if (box[1][BOXTOP  ] > box[0][BOXTOP  ]) box[0][BOXTOP  ] = box[1][BOXTOP  ];
				if (box[1][BOXBASE ] < box[0][BOXBASE ]) box[0][BOXBASE ] = box[1][BOXBASE ];

				LS_SectorsFromLinesInBox(box[0], countsec, &c0);

				if (c0 > 4)
				{
					int oxc, reject;

					// All enclosed floor heights must be lower than the test sector
					// All enclosed ceil heights must be the same as sec unless it's
					// an unclosed textureless sector.
					// At least two 0xC secs must be enclosed
					for (oxc = reject = 0, k = 1; k < c0; k++)
					{
						int spf = sectors[countsec[k]].specFlags & 0x1C;
						int ctd = secContained(countsec[k], box[0]);

						if (!ctd && spf & SF_NOHITEX)
						{
							if (i==268)Con_Error("FAILED[1] sec %i\n", countsec[k]);
							oxc = 0; break;
						}

						if (!(spf & SF_NOHITEX))
						{
							if (ctd)
							{
								if (spf = 0x14) continue;
							}
							// Only allow one uncompliant sector that
							// extends beyond the box[0] range.
							else if (spf == 0x10 && !reject++)
							{
								countsec[k] = -1; continue;
							}

							oxc = 0; break;
						}
						
						if (sectors[countsec[k]].floorheight >= sec->floorheight)
						{
							oxc = 0; break;
						}

						if ( (spf & 0xC) == 0xC) oxc++;

						if (sectors[countsec[k]].ceilingheight != sec->ceilingheight
							&& spf != 0xC)
						{
							oxc = 0; break;
						}
					}

					if (oxc > 1)
					{
						ss_validcount++;
						validcount++;

						for (k = 1; k < c0; k++)
							setContainedSSECS(k, c0);

						Setsecdone3(sec, 49048);
					}
				}
			}
		}
	}

	// v220-18 function added
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->linecount < 7
			&& sec->linecount > 2
			&& sec->stptr < 0
			&& (sec->specFlags & 0x3000C) == 0x4
			&& !SecCtrl(i))
		{
			subfn_CheckFloorSecWithMissingReset(sec);
		}
	}
}

//-----------------------

static RDBOOL nfLinesInside()
{
	int i, k;

	CLR_hitline();

	for (i = 0; i < rl.count; i++)
	{
		if (rl.mark[i] == 4)
			hitline[rl.line[i] - lines] = true;
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *li = sg->linedef;

			if (li && !hitline[li - lines])
				return true;
		}
	}
	return false;
}

RDBOOL NF_CheckLoop(sector_t *sec, int fnnum)
{
	int		i, count, bbox[4];
	RDBOOL	ret = false;

	rl.line = sec->lines;
	rl.count = sec->linecount;
	rl.mark  = (short*)M_Malloc(rl.count * sizeof(short), 42);
	limit_check = sec->linecount;

	for (i = 0; i < rl.count; i++)
	{
		if (lineflags[rl.line[i] - lines] & (LFG_NEWSET|LFG_3DTYPE))
		{
			M_Free(rl.mark); return true;
		}
		rl.mark[i] = 1;
	}

	if (checkRinglist(0, bbox, 0, 9) == rl.count)
	{
		for (count = i = 0; i < rl.count; i++)
		{
			if (rl.mark[i] == 4) linearray[count++] = rl.line[i] - lines;
		}

		if (LS_MarkSubsecsInLoop(sec, count, NULL, bbox, MSIA_NONE, NULL))
		{
			line_t		*ref = sec->lines[0];
			sector_t	*s = ref->frontsector == sec ?
							 ref->backsector : ref->frontsector;

			switch(fnnum)
			{
			case 0:
				if (!nfLinesInside())
				{
					for (i = 0; i < hitssecptr; i++)
					{
						subsector_t *ss = &subsectors[hitssec[i]];

						Setr3dssfs(ss, s - sectors, 47058);
					}
					ret = true;
				}
				break;
			default:
				break;
			}
		}
	}

	M_Free(rl.mark);

	return ret;
}

//================================================================


// For external calls - the global linearray is used
// The function looks for a ring using ld as the starting line.
// If one is found the line numbers are transferred to, if set,
// lnarray which must, therefore, either be NULL or large enough
// to store the lines found. The count is returned.
int nf_lineList(line_t *ld, sector_t *sec, int *bbox, int *lnarray, int condition)
{
	int c = 0, i, lnum = -1;

	rl.line  = sec->lines;
	rl.count = sec->linecount;
	rl.mark  = (short*)M_Calloc(rl.count * sizeof(short), 260);
	limit_check = sec->linecount;

	for (i = 0; i < rl.count; i++)
	{
		if (rl.line[i]->spflags & SPF_CONTROLLINE) continue;
		if (rl.line[i] == ld) lnum = i;
		rl.mark[i] = 1;
	}

	if (lnum >= 0 && checkRinglist(lnum, bbox, condition, 10))
	{
		for (i = 0; i < rl.count; i++)
		{
			if (rl.mark[i] == 4)
			{
				if (lnarray)
				{
					lnarray[c++] = rl.line[i] - lines;
				}
				else
				{
					c++;
				}
			}
		}
	}

	M_Free(rl.mark);

	return c;
}

// For external calls - the lines to be tested should be
// in the global linearray.
// The function looks for a ring with the lines sent.
// If one is found the count is returned.
int nf_chkLineListRing(int count, int *bbox, int lptr, byte *linemem)
{
	int c = 0, i;

	rl.count = count;
	rl.mark  = (short*)M_Calloc(count * sizeof(short), 261);
	rl.line  = (line_t**)M_Calloc(count * sizeof(line_t*), 262);
	limit_check = count;

	for (i = 0; i < rl.count; i++)
	{
		rl.line[i] = &lines[linearray[i]];
		rl.mark[i] = 1;
	}

	c = checkRinglist(lptr, bbox, 0, 11);

	if (linemem && c)
	{
		for (i = 0; i < rl.count; i++)
		{
			if (rl.mark[i] == 4)
				linemem[linearray[i]] |= 4;
		}
	}

	M_Free(rl.mark);
	M_Free(rl.line);

	return c;
}

RDBOOL nf_CheckBoxRing(sector_t *s, int *segbox, line_t *ld)
{
	int i, lptr = -1, bbox[4];

	for (i = 0; i < s->linecount; i++)
	{
		linearray[i] = s->lines[i] - lines;
		if (lptr < 0 && s->lines[i] == ld) lptr = i;
	}

	if ( (nf_chkLineListRing(s->linecount, bbox, lptr, NULL)) > 0)
		return M_FitBox(segbox, bbox);

	return false;
}

static int	*linememory;
static int	linememcnt;

static RDBOOL subfn_chkSSECSinBox(int floortex, int ht, int *tex)
{
	int i, k, lowht = ht;

	validcount++;

	*tex = 0;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (   ss->sector->floorpic != floortex
			|| ss->sector->floorheight > ht)
		{
			return false;
		}

		if (ss->sector->floorheight < lowht) lowht = ss->sector->floorheight;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *li = sg->linedef;

			if (li && li->validcount != validcount)
			{
				int lnum = li - lines;
				
				if (li->ml_flags & ML_TWOSIDED)
				{
					(*tex) |= origTEX[lnum];

					if (   *tex & 0x11
						|| li->frontsector->specFlags & SF_MOVFLOOR
						|| li->backsector->specFlags  & SF_MOVFLOOR
						|| li->frontsector->stptr >= 0
						|| li->backsector->stptr  >= 0
						|| li->frontsector->heightsec >= 0
						|| li->frontsector->heightsec >= 0
						|| li->specType >= R3D_40)
					{
						return false;
					}
				}

				li->validcount = validcount;
				
				linememory[linememcnt++] = lnum;
			}
		}
	}
	return lowht < ht;
}

void NF_Check1sLinesBoundedLiquidFloors()
{
	sector_t	*sec;
	line_t		*ld;
	int			c, i, k, m, tex, bbox[4];

	linememory = (int*)M_Calloc(numlines * sizeof(int), 263);
	
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   secdone[i]
			|| sec->specFlags != 0x0000849C
			|| !IsLiquidFloorFlat(sec)
			|| sec->stptr >= 0
			|| sec->heightsec >= 0)
		{
			continue;
		}

		CLR_hitline();

		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (lineflags[ld - lines]) continue;

			if (!(ld->ml_flags & ML_TWOSIDED) || Spans1sLines(ld))
			{
				hitline[ld - lines] = 1; c++;
			}
		}

		if (c < 3) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			int lnum = sec->lines[k] - lines;

			if (hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, ld->frontsector, lnum, NULL, bbox, NULL, NULL)
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE))
			{
				RDBOOL valid = false;

				linememcnt = 0;

				if (subfn_chkSSECSinBox(sec->floorpic, sec->floorheight, &tex))
				{
					valid = true;

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];

						if (ss->sector->floorheight < sec->floorheight)
						{
							Setr3dssfs(ss, i, 47059);
							if (!(tex & 0x44)) Setr3dsscs(ss, i, 47059);
						}
					}

					for (m = 0; m < linememcnt; m++)
					{
						SetLFG(&lines[linememory[m]], LFG_NEWSET|LFG_EXCLUDE, 47060);
					}
				}

				for (m = 0; m < sec->linecount; m++)
				{
					lnum = sec->lines[m] - lines;

					if (hitline[lnum] == 4) hitline[lnum] = 8;
				}
			}
		}
	}

	M_Free(linememory);
}

static void setSSLINE(sector_t *sec)
{
	line_t	*ld;
	int		i, k;

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (!(ld->spflags & SPF_CONTROLLINE))
			hitline[ld - lines] = 1;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (   hitline[ld - lines] == 1
			&& subfn_CheckForRing(hitline, sec, ld - lines, NULL, NULL, NULL, NULL))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (hitline[ld - lines] == 4)
				{
					if (!(ld->ml_flags & ML_TWOSIDED))
					{
						sec->specFlags |= SF_SSLINETRUE;
						return;
					}
					hitline[ld - lines] = 8;
				}
			}
		}
	}
}

// Check if missing low tex line loops include a single sided line
void NS_CheckSSLineFlag()
{
	sector_t	*sec;
	int			i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if ( (sec->specFlags & 0x14) == 0x14)
		{
			// Check for single sided lines in the loop
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!(ld->spflags & SPF_CONTROLLINE) && Vtxs2(ld) && Vtfl3(ld))
				{
					setSSLINE(sec);
					break;
				}
			}
		}
	}
}

static int chkseccnt;

static void setRingSEC(int snum, int *chksec)
{
	int		i;

	for (i = 0; i < chkseccnt; i++)
	{
		if (chksec[i] == snum) return;
	}

	chksec[chkseccnt++] = snum;
}

static int sidelistcnt;

// Used to get subsectors in a box where some or all
// of the lines have same sector sided lines.
// Lines defining the box must be in linearray
// The subsectors must not contain lines referencing a
// different sector else false will be returned.
// True returns with the hitssec array set
RDBOOL nf_boxAreaSameSec(sector_t *sec, int *cs, int *box, int count)// v220-7 add cs
{
	sector_t	*osec = NULL;
	int		i, k, m;

	sidelistcnt = 0;

	// NOTE: ss_validcount cannot be used as a single subsector
	//       may contain more than one of the bounding lines.
	for (i = 0; i < count; i++)
	{
		line_t	 *ld = &lines[linearray[i]];
		ssline_t *sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			seg_t		*sg = &segs[ss->firstline];
			RDBOOL		valid = true;

			// all the frontsectors must be the sector being checked
			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (sg->linedef && sg->frontsector != sec)
				{
					valid = false;

					// v220-7 same sector ring inside same sector ring
					if (cs)
					{
						if (sg->linedef->frontsector != sg->linedef->backsector)
							return false;
						if (!osec) osec = sg->frontsector;
						else if (osec != sg->frontsector) return false;
						continue;
					}
					break;
				}
			}

			if (valid)
			{
				int x = (int)(ss->midpoint.x * FRACUNIT);
				int y = (int)(ss->midpoint.y * FRACUNIT);
				int side = P_PointOnLineSide(x, y, ld);

				sidelist[sidelistcnt++] = ld->sidenum[side];
				break;
			}
		}
	}

	if (cs) *cs = osec ? osec - sectors : -1;

	return sidelistcnt == count
		&& LS_MarkSubsecsInMultiSecLoop(0, NULL, sidelistcnt, box, NULL, true, MSIA_MSREVERSED);
}

static void setSideList(line_t *li, int *box)
{
	ssline_t *sl = &sslines[li - lines];
	line_t	 *ld;
	int		 i, k, m;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			seg_t		*sg = &segs[ss->firstline];
			RDBOOL		found = false;
			
			ss->ssvalidcount = ss_validcount;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) && hitline[ld - lines])
				{
					RDBOOL	found = false;

					for (m = 0; m < sidelistcnt; m++)
					{
						if (linearray[m] == ld - lines)
						{
							found = true;
							break;
						}
					}

					if (!found)
					{
						int side = P_PointOnLineSide(
							(int)(ss->midpoint.x*FRACUNIT),
							(int)(ss->midpoint.y*FRACUNIT), ld);

						int sdnum = &sides[ld->sidenum[side]] - sides;

						if ((unsigned)sidelistcnt >= maxsecsz*2)
							Con_ErrorD("setSideList: limit exceeded");

						M_AddLineToBox(box, ld);
						linearray[sidelistcnt] = ld - lines;
						sidelist[sidelistcnt++] = sdnum;
					}
				}
			}
		}
	}
}

// Routine to search for a loops using midtex lines with a missing lower tex.
void NF_SpecialFX_MidtexRing()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			a, b, c, i, k, rc, box[4];
	byte		*looparray;
	int			*newsecmark;
	RDBOOL		valid;

	int			*secmark = (int*)M_Malloc(numlines*sizeof(int), 44);

	//------------------------------------

	CLR_hitline();

	// Find and mark any lines of interest
	for (a = b = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->ml_flags & ML_DONTPEGBOTTOM
			&& !lineflags[i]
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& !(origTEX[i] & 0x11)
			&& origTEX[i] & 0x22
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& !(bs->specFlags & SF_MOVSECTOR)
			&& abs(fs->floorheight - bs->floorheight) > 4*FRACUNIT
			&& !((origTEX[i] & 0x22) == 0x02
			&& bs->floorheight < fs->floorheight)
			&& !((origTEX[i] & 0x22) == 0x20
			&& bs->floorheight > fs->floorheight)
			&& !(!(origTEX[i] & 0x44) && fs->ceilingheight != bs->ceilingheight)
			&& vtxarray[ld->v1 - vertexes] >= 2
			&& vtxarray[ld->v2 - vertexes] >= 2
			&& fs->heightsec < 0
			&& bs->heightsec < 0
			&& fs->stptr < 0
			&& bs->stptr < 0)
		{
			int maxf = MAX_OF(fs->floorheight  , bs->floorheight) >> FRACBITS;
			int side = fs->floorheight == maxf ? 0 : 1;
			side_t *sd = &sides[ld->sidenum[side^1]];

			if (sd->midtexture && sd->rowoffset < 0
				&& !IsLiquidFloorFlat(fs)
				&& !IsLiquidFloorFlat(bs))
			{
				secmark[i] = sd->sector - sectors;
				hitline[i] = 2; a++;
			}
			else
				hitline[i] = 1; b++;
		}
	}

	if ((a + b) < 3)
	{
		M_Free(secmark);
		return;
	}

	// PHASE ONE
	// Look for textureless line loops.  If found then scan the subsectors for
	// an adjoining missing low tex line.  If these are found then check if they
	// also form a loop and then set the necessary subsec rendering pointers,

	looparray = (byte*)M_Calloc(numlines, 264);

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   !lineflags[i]
			&& ld->lnflags2 & LNF2_OWNEQUALTOONE
			&& !lxarray[i]
			&& (fs = ld->frontsector) == ld->backsector	// must be
			&& Vtxs2(ld)
			&& !origTEX[i]
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& fs->heightsec < 0)
		{
			looparray[i] = true; c++;
		}
	}

	if (c > 2)
	{
		sector_t *s;

		for (i = 0; i < numlines; i++)
		{
			if (looparray[i] == 1
				&& subfn_CheckForRing(looparray, (s = lines[i].frontsector), i, &rc, NULL, NULL, NULL)  )
			{
				M_ClearBox(box);

				if ((uint)rc < maxsecsz) // should be
				{
					sidelistcnt = 0;

					ss_validcount++;

					for (k = 0; k < s->linecount; k++)
					{
						ld = s->lines[k];

						if (looparray[ld - lines] == 4)
						{
							setSideList(ld, box);
						}
					}

					valid = false;
				
					if (sidelistcnt > 2)
					{
						int box2[4];

						M_ClearBox(box2);

						rl.count = sidelistcnt;
						rl.line = (line_t**)M_Malloc(sidelistcnt * sizeof(line_t*), 45);
						rl.mark = (short*)M_Malloc(sidelistcnt * sizeof(short), 46);
						limit_check = sidelistcnt;

						for (k = 0; k < sidelistcnt; k++)
						{
							rl.mark[k] = 1;
							rl.line[k] = &lines[linearray[k]];
						}

						valid = checkRinglist(0, box2, 0, 14) == sidelistcnt
							&& M_FitBox(box, box2)
							&& LS_MarkSubsecsInMultiSecLoop(0, NULL, sidelistcnt, box, NULL, true, MSIA_MSREVERSED);

						M_Free(rl.line);
						M_Free(rl.mark);
					}

					if (valid)
					{
						sector_t *low = NULL, *high = NULL;

						for (k = 0; k < sidelistcnt; k++)
						{
							sector_t *chk = sides[sidelist[k]].sector;

							ld = &lines[linearray[k]];

							Settype(ld, R3D_2D, 47061);
							ld->lnflags2 |= LNF2_NORENDLOW;

							SetLFG(ld, LFG_NEWSET, 47061);

							if (chk != lines[i].frontsector)
							{
								if (!low) low = chk;
								else if (chk->floorheight < low->floorheight) low = chk;
							}

							if (chk != lines[i].frontsector)
							{
								if (!high) high = chk;
								else if (chk->floorheight > high->floorheight) high = chk;
							}
						}

						if (low )
						{
							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];

								Setr3dsslfsec(ss, low - sectors, 47062);
								Setr3dssfs(ss, lines[i].frontsector - sectors, 47063);
								ss->ssflags |= SSF_NOCHANGEFS;
							}
						}
					}

					for (k = 0; k < s->linecount; k++)
					{
						ld = s->lines[k];

						if (looparray[ld - lines] == 4)
						{
							looparray[ld - lines] = 8;

							if (valid)
							{
								Settype(ld, R3D_01, 47064);
								SetLFG(ld, LFG_NEWSET, 47064);
								Setsecdone(ld->frontsector, 47065);
							}
						}
					}
				}
			}
		}
	}

	M_Free(looparray);

	// PHASE TWO
	// Look for rings of midtex lines with missing lower textures
	// that are not associated with textureless line rings.

	rl.count = 0;

	if (a < 3)
	{
		M_Free(secmark);
		return;
	}

	newsecmark = (int*)M_Malloc(a * sizeof(int), 47);

	for (i = 0; i < numlines; i++)
	{
		if (hitline[i] == 2 && !lineflags[i])
		{
			if (rl.count == a) Con_Error("A failed");
			newsecmark[rl.count] = secmark[i];
			linearray[rl.count++] = i;
		}
	}

	if (rl.count > 2)
	{
		int loopcnt;
		int *chksec = (int*)M_Malloc(rl.count * sizeof(int), 48);

		M_Free(secmark);
		secmark = newsecmark;

		c = rl.count;
		rl.count = 0;

		rl.line = (line_t**)M_Malloc(c * sizeof(line_t*), 49);
		rl.mark = (short*)M_Calloc(c * sizeof(short), 265);
		limit_check = c;

		for (i = 0; i < c; i++)
		{
			rl.mark[rl.count  ] = 1;
			rl.line[rl.count++] = &lines[linearray[i]];
		}

		for (i = 0; i < c; i++)
		{
			if (rl.mark[i] == 1
				&& (rc = checkRinglist(i, box, 0, 15)) > 2 )
			{
				int		type = 0;

				chkseccnt = loopcnt = 0;

				for (k = 0; k < c; k++)
				{
					if (rl.mark[k] == 4)
					{
						setRingSEC(secmark[k], chksec);
						linearray[loopcnt++] = rl.line[k] - lines;
					}
				}

				valid = LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, loopcnt, box, NULL, false, MSIA_MSREVERSED);

				if (!valid)
				{
					sector_t	*os = NULL;
					int			m;

					valid = LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, loopcnt, box, NULL, false, MSIA_NONE);

					if (valid)
					{
						int ht = DDMININT;

						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];

							if (ss->sector - sectors == chksec[0])
							{
								seg_t		*sg = &segs[ss->firstline];

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if (sg->linedef)
									{
										sector_t *ck = sg->frontsector - sectors == chksec[0] ?
											sg->backsector : sg->frontsector;

										if (  !ck
											|| origTEX[sg->linedef - lines] & 0x55
											|| ck->specFlags & SF_MOVSECTOR)
										{
											valid = false; break;
										}

										if (ht == DDMININT)
										{
											os = ck;
											ht = ck->floorheight;

											if (ht >= sectors[chksec[0]].floorheight)
											{
												valid = false; break;
											}
										}
										else if (ht != ck->floorheight)
										{
											valid = false; break;
										}
									}
									if (!valid) break;
								}
							}
						}

						if (valid && os)
						{
							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];
								seg_t		*sg = &segs[ss->firstline];

								if (ss->sector - sectors == chksec[0])
								{
									Setr3dsslfsec(ss, os - sectors, 47066);
								}

							}

							for (k = 0; k < c; k++)
							{
								if (rl.mark[k] == 4)
								{
									Settype(rl.line[k], R3D_1D, 47066);
									SetLFG(rl.line[k], LFG_NEWSET, 47066);
									rl.line[k]->lnflags2 |= LNF2_NORENDLOW;
								}
							}
						}
					}
					valid = false;
				}

				if (valid)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];

						if (!(ss->sector->specFlags & SF_NOLOTEX))
						{
							valid = false; break;
						}
					}

					if (valid)
					{
						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];

							if (type)
							{
								Setr3dsslfsec(ss, chksec[0], 47066);
							}
							else
							{
								if (ss->sector->floorheight < sectors[chksec[0]].floorheight)
									Setr3dsslfsec(ss, ss->sector - sectors, 47066);
								Setr3dssfs(ss, chksec[0], 47067);
							}
						}
					}
				}

				for (k = 0; k < c; k++)
				{
					if (rl.mark[k] == 4)
					{
						rl.mark[k] = 8;

						if (valid)
						{
							ld = rl.line[k];

							Settype(ld, R3D_1D, 47069);
							ld->lnflags2 |= LNF2_NORENDLOW;

							SetLFG(ld, LFG_NEWSET, 47069);
							Setsecdone(ld->frontsector, 47070);
							Setsecdone(ld->backsector , 47071);
						}
					}
				}
			}
		}

		M_Free(rl.line);
		M_Free(rl.mark);
		M_Free(chksec);
	}
	M_Free(secmark);
}

// ===============================================================

// v220-16
static void subfn_CheckForSubset(sector_t *sc)
{
	sector_t	*ch, *rf = NULL;
	line_t		*ld;
	int			fht = sc->floorheight;
	int			i, c, d;

	CLR_hitline();

	for (c = d = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (lineflags[ld - lines]) return;
		if (!(ld->ml_flags & ML_TWOSIDED)) continue;
		if (ld->frontsector == ld->backsector) { d++; continue; }
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
		if (!rf)
		{
			if (ch->floorheight >= fht) return;
			if (ch->specFlags & SF_MOVSECTOR) return;

			rf = ch;
		}
		else
		{
			if (rf != ch) return;
		}
		hitline[ld - lines] = true; c++;
	}

	if (c < 3 || !d) return;

	while(1)
	{
		int	rc, box[4];
		int	lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
			&& rc > 2)
		{	
			if (   LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(sc, hitline))
			{
				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (ss->sector != sc) SetOverride(ss, sc - sectors, 47077);

					ss->ssflags |= SSF_NOTRACEN;	// Likely to be a subset of a larger set
				}
			}

			for (i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];
				if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

void NF_CheckForSubset()
{
	int			i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i] 
			&& (sectors[i].specFlags & 0xFFFFC) == 0x818
			&& sectors[i].heightsec < 0
			&& sectors[i].stptr < 0)
		{
			subfn_CheckForSubset(&sectors[i]);
		}
	}
}