//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2011 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_OGL_init.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  cMode handling added
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
//
// Based on: Doomsday v1.7.8 drOpenGL.dll code
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include  <io.h>
#include  <stdio.h>
#include  <math.h>
#include  <stdlib.h>

#include "m_winmem.h"
#include "m_misc.h"

#include "dd_main.h"
#include "dd_types.h"
#include "gl_main.h"
#include "gl_tex.h"

#include "r3d_OGLfunc.h"
#include "m_args.h"
#include "dglib.h"

#include "r3d_edit.h"
#include "r3d_compat.h"
#include "con_console.h"
#include "sys_system.h"


// MACROS ------------------------------------------------------------------

#define WINDOW_STYLE	(WS_OVERLAPPED | WS_VISIBLE | WS_CAPTION | WS_THICKFRAME | WS_CLIPCHILDREN | WS_CLIPSIBLINGS)
#define FULLSCR_STYLE	(WS_POPUP | WS_VISIBLE | WS_CLIPSIBLINGS | SS_OWNERDRAW)

// TYPES -------------------------------------------------------------------

// "wiggle" callbacks for vsync
typedef BOOL (WINAPI * PFNWGLSWAPINTERVALEXTPROC   ) (int interval);
typedef int  (WINAPI * PFNWGLGETSWAPINTERVALEXTPROC) (void);

static PFNWGLSWAPINTERVALEXTPROC vsptr = NULL;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// FUNCTION PROTOTYPES -----------------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	resolution_set;
extern RDBOOL	ScreenError;
extern uint		current_win[4][2];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		anisotropic;
int		pspr_xoffset;
int		maxTexSize;
int		scrwddiv120;
int		scrwddiv240;
float	scrwddiv320;
RDBOOL	valid_screen_ratio;
char	Hz_str[3][16] = { "unknown Hz", "", "" };	// v220-3

glext_t glx;

// The State.
float			maxAniso = 1.0f;

currentmode_t	cMode;

int		current_w;
int		current_h;
int		aspect_offset_x;
int		aspect_offset_y;

uint swp_flags = SWP_DRAWFRAME | SWP_NOCOPYBITS | SWP_NOZORDER;

float	occAngle;	// v220-9
RDBOOL	vsyncSupported = false;	// v220-13	made global

GLfloat	linesizes[2];
//GLfloat	linegranularity;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// R3D
static RDBOOL	use43 = false;
static int		origvidVsync = -1;			// force update

// v220-3 new screen handling constants
static RECT		current_rect;

// CODE --------------------------------------------------------------------

// THE ROUTINES ------------------------------------------------------------

//===========================================================================
// R3D
//===========================================================================
void GL_CheckVSYNC()
{
	if (vsptr) vsptr(vidVsync);
}

//===========================================================================
// query
//  It takes a bit of care to be fool-proof about parsing the
//  OpenGL extensions string. Don't be fooled by sub-strings, etc.
//===========================================================================
static void query(const char *name, int *var)
{
	const GLubyte	*start = glGetString(GL_EXTENSIONS);
	GLubyte			*where;
	GLubyte			*terminator;
	uint			len = strlen(name);

	*var = false;	// until proven otherwise

	if (start)
	{
		for (; ; start = terminator) 
		{
			if( !(where = (GLubyte*) strstr((const char*) start, name)) )
				break;

			terminator = where + len;

			if (   (where == start     || *(where - 1) == ' ')
				&& (*terminator == ' ' || *terminator == '\0') )
			{
				*var = true; break;
			}
		}
	}

	glGetError();	// Clear any glGetString error
}

// v220-4
char *geErrStr(uint err)
{
	switch(err)
	{
	case GL_INVALID_ENUM     : return "INVALID_ENUM";
	case GL_INVALID_VALUE    : return "INVALID_VALUE";
	case GL_INVALID_OPERATION: return "INVALID_OPERATION";
	case GL_STACK_OVERFLOW   : return "STACK_OVERFLOW";
	case GL_STACK_UNDERFLOW  : return "STACK_UNDERFLOW";
	case GL_OUT_OF_MEMORY    : return "OUT_OF_MEMORY";
	case GL_TABLE_TOO_LARGE  : return "TABLE_TOO_LARGE";
	default: return "unknown";
	}
}

RDBOOL GL_CheckError(char *id)
{
	uint	err;
	RDBOOL	gl_err = false;

	// v220-4 Check for errors at the end of each frame
	while ( (err = glGetError()) != GL_NO_ERROR)
	{
		pr_OFinsert("OGL %s error %s\n", id, geErrStr(err));
		gl_err = true;
	}

	return gl_err;
}

//===========================================================================
// v220-3 modified screen handling
//  a) finds all supported x y resolutions for user system
//  b) generates lists used in console
//  c) validates entries
//  d) auto sets FOV for screen with a user selectable base fov
//
//===========================================================================

// v220-6 modified
static uint getResolutionMatch(uint w, uint h, uint f, uint bpp, RDBOOL findfreq)
{
	DEVMODE	testMode;
	uint	freq = f;
	int		i;

	// Let's enumerate all possible modes to find the most suitable one.
	for (i = 0; ; i++)
	{
		memset(&testMode, 0, sizeof(testMode));
		testMode.dmSize = sizeof(testMode);

		if (!EnumDisplaySettings(NULL, i, &testMode)) break;

		if (   testMode.dmPelsWidth  != w
			|| testMode.dmPelsHeight != h
			|| testMode.dmBitsPerPel != bpp)
		{
			continue;
		}
		// v220-6 If no match find the highest registered frequency
		//        below the current desktop frequency
		if (findfreq)
		{
			if (testMode.dmDisplayFrequency < (uint)cMode.w_initialfreq
			 && testMode.dmDisplayFrequency > freq)
			{
				freq = testMode.dmDisplayFrequency;
			}
		}
		else
		{
			if (freq == testMode.dmDisplayFrequency) break;
		}
	}

	return freq;
}

void setHZvalues(int Hz_val, int num)
{
	// 0 or 1 means the frequency setting are not available
	if (Hz_val > 1)
	{
		if (num & UIHZ_DESKTOP) sprintf(Hz_str[0], "%iHz", Hz_val);
		if (num & UIHZ_DEFAULT) sprintf(Hz_str[1], "%iHz", Hz_val);
		if (num & UIHZ_CURRENT) sprintf(Hz_str[2], "%iHz", Hz_val);
	}
}

uint checkEnumerated()
{
	uint f = getResolutionMatch(
				cMode.max_width,
				cMode.max_height,
				cMode.freq ? cMode.freq : cMode.w_initialfreq,
				cMode.bpp,
				false );

	if (!f)
	{
		// v220-6 If a monitor's been registered as generic and the user has wound
		// the fps value beyond the generic failsafe then we have to get inventive.
		f = getResolutionMatch(
			cMode.max_width,
			cMode.max_height,
			0,
			cMode.w_initialbpp,
			true );

		cMode.w_initialfreq = f;
		cMode.freq = f;
	}

	if (!cMode.freq) cMode.freq = f;

	return f;
}

RDBOOL compareFULLSCREEN()
{
	return current_win[0][0] == current_win[0][1]
		&& current_win[1][0] == current_win[1][1]
		&& current_win[2][0] == current_win[2][1]
		&& current_win[3][0] == current_win[3][1];
}

static void changeDeskTopSize(uint width, uint height, uint freq)
{
	if (getResolutionMatch(width, height, cMode.freq, cMode.bpp, false))
	{
		DEVMODE newMode;
		
		memset(&newMode, 0, sizeof(newMode));
		newMode.dmSize = sizeof(newMode);
		newMode.dmBitsPerPel = cMode.bpp;
		newMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
		newMode.dmPelsWidth  = width;
		newMode.dmPelsHeight = height;
		newMode.dmDisplayFrequency = (uint)cMode.freq;
		newMode.dmFields |= DM_DISPLAYFREQUENCY;
		current_win[0][1] = newMode.dmPelsWidth;
		current_win[1][1] = newMode.dmPelsHeight;
		current_win[2][1] = newMode.dmDisplayFrequency;
		current_win[3][1] = newMode.dmBitsPerPel;

		setHZvalues(cMode.freq, UIHZ_CURRENT);

		if (ChangeDisplaySettings(&newMode, CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL)
			return;
	}

	ScreenError = true;

	Con_ErrorD(NULL);
}

void getDesktopParms(int *width, int *height, int *freq, int *bpp)
{
	HWND	desktopwin = GetDesktopWindow();
	HDC		desktopdc  = GetDC(desktopwin);

	if (width ) *width  = GetDeviceCaps(desktopdc, HORZRES);
	if (height) *height = GetDeviceCaps(desktopdc, VERTRES);
	if (freq  ) *freq   = GetDeviceCaps(desktopdc, VREFRESH);
	if (bpp   ) *bpp    = GetDeviceCaps(desktopdc, BITSPIXEL);

	ReleaseDC(desktopwin, desktopdc);
}

char *getHZstr(int Hz_val)
{
	static char hZstr[16];

	if (Hz_val > 1)
		sprintf(hZstr, "%iHz", Hz_val);
	else
		hZstr[0] = 0;

	return hZstr;
}

// v220-5 Revised to allow fullscreen 4:3 aspect ratio with a widescreen monitor
//        or a 5:4 monitor
void GL_SetRatioAndFov()
{
	float calc_yfov;
	float current_aspect;

	aspect_offset_x = aspect_offset_y = 0;
	current_w = cMode.width;
	current_h = cMode.height;
	cMode.scr_aspect = (float)scr_max_width / (float)scr_max_height;
	valid_screen_ratio = cMode.fullscreen
				&& screen_ratio != (4.0f/3.0f)
				&& (unsigned)cMode.width == scr_max_width
				&& (unsigned)cMode.height == scr_max_height;

	if (cfg.use43 && valid_screen_ratio)
	{
		if (screen_ratio > (4.0f/3.0f))
		{
			current_w = (int)((float)cMode.height*(4.0f/3.0f) * ((float)cMode.width/(float)scr_max_width));
			aspect_offset_x = (cMode.width - current_w)/2;
			cMode.scr_aspect *= (float)current_w/(float)scr_max_width;
		}
		else
		{
			current_h = (int)((float)cMode.width*(3.0f/4.0f) * ((float)cMode.height/(float)scr_max_height));
			aspect_offset_y = (cMode.height - current_h)/2;
			cMode.scr_aspect *= (float)scr_max_height/(float)current_h;
		}
	}

	if (cfg.use43)
		ge_inf.clear_screen = 3;
	else if (cMode.fullscreen)
		ge_inf.clear_screen = 2;

	// Update the ui variables.
	scrwddiv120	 = (int)((float)current_w / 120.0f);
	scrwddiv240	 = (int)((float)current_w / 240.0f);
	scrwddiv320	 = (float)current_w / 640.0f;	// Only used for control panel text shadow

	// A base FOV of 90deg will give a calc_yfov of 67.5 with a 3/4 ratio
	calc_yfov = aspect_offset_x ? (float)init_FOV : (float)init_FOV * (3.0f/4.0f); // default
	current_aspect = (float)current_w/(float)current_h;

	// v220-4 ensure FOV is reset
	if (cMode.fullscreen)
	{
		cMode.weaponFOV = cMode.scr_aspect * 90.0f * (3.0f/4.0f);
		setFOV = aspect_offset_x ? (float)init_FOV : init_FOV * screen_ratio * (3.0f/4.0f);
	}
	else
	{
		setFOV = calc_yfov * current_aspect;
		cMode.weaponFOV = current_aspect * 90.0f * (3.0f/4.0f);
	}

	glViewport(aspect_offset_x, aspect_offset_y, current_w, current_h);	// v220-5
	ZoomReset(true);

	// Used with glOrtho to get correct sprite weapon sizing
	pspr_xoffset = 160 - (int)(cMode.weaponFOV * 160.0f / 90.0f);

	Con_Printf("Resolution   : %d x %d  %dHz  %dbit (%s)\nBase 4:3 FOV : %i deg",
		current_w,
		current_h,
		cMode.freq,
		cMode.bpp,
		cMode.fullscreen ? "fullscreen" : "window",
		init_FOV);

	if (init_FOV != (int)setFOV)
		Con_Printf(" [adjusted to %i deg to suit current aspect ratio]", (int)setFOV);

	occAngle = .2f/30.f * init_FOV;

	Con_Printf("\n");
}

typedef struct {
	// parms copied
	int		width;
	int		height;
	int		freq;
	int		xpos;
	int		ypos;
	int		bpp;
	int		fullscreen;

	//----------------
	RDBOOL	custom;

} lastvalid_t;

static lastvalid_t	lvscr;

static void subfn_cModeResolution(int width, int height, RDBOOL centre)
{
	static RDBOOL firsttime = true;
	static int mouse_suppressed = false;

	int		cwidth = width, cheight = height;
	int		dxsc, dysc, caption, border;
	int		xoff, yoff;
	RDBOOL	set_custom = false;

	if (firsttime)
	{
		RDBOOL	argset = false;

		memset(&lvscr, 0, sizeof(lastvalid_t));

		caption = GetSystemMetrics(SM_CYCAPTION);		// get the default caption size (normally 26)
		border  = GetSystemMetrics(SM_CXFIXEDFRAME);	// get the default thick border size (normally 3)
		
		cMode.szCaption = caption;
		cMode.szBorder  = border;

		modeNlist[0].freq = cMode.w_initialfreq;
		setHZvalues(cMode.w_initialfreq, UIHZ_DESKTOP);

		if (cMode.xpos || cMode.ypos)
		{
			cMode.fullscreen = false;
			centre = false;
		}

		firsttime = false;
	}
	else
	{
		caption = cMode.szCaption;
		border  = cMode.szBorder;
	}

	if (cMode.fullscreen)
	{
		// Set the correct window style and size.
		changeDeskTopSize(cwidth, cheight, cMode.freq);
		SetWindowLong(hWndMain, GWL_STYLE, FULLSCR_STYLE);
		SetWindowPos(hWndMain, 0, 0, 0, cwidth, cheight, swp_flags);
	}
	else
	{
		int		inwidth = width, inheight = height;
		int		less_x, less_y;
		
		//---------------------------------

		// Use the desktop settings
		dxsc = cMode.max_width;
		dysc = cMode.max_height;

		cMode.freq = cMode.w_initialfreq;
		changeDeskTopSize(dxsc, dysc, cMode.freq);

		caption = cMode.szCaption;		// default caption size (normally 26)
		border  = cMode.szBorder;		// default thick border size (normally 3)

		less_x = 2*border;
		less_y = 2*border + caption;

		// Restrict minimum window size
		if (width  < MIN_SCR_WIDTH  - less_x) width  = MIN_SCR_WIDTH  - less_x;
		if (height < MIN_SCR_HEIGHT - less_y) height = MIN_SCR_HEIGHT - less_y;

		// v220-4 don't allow an aspect ratio of greater than 2.3f
		if ((float)width/(float)height > 2.3f)
		{
			while ((float)width/(float)height > 2.3f) width--;
		}

		// v220-4 don't allow an aspect ratio of less than 1.2f
		if ((float)width/(float)height < 1.2f)
		{		
			while ((float)width/(float)height < 1.2f) width++;
		}

		if (width  + less_x > dxsc) width  = dxsc - less_x;
		if (height + less_y > dysc) height = dysc - less_y;

		// Check the default values were sensible
		if (firsttime && (inwidth != width || inheight != height))
		{
			cMode.w_setdefaults = true;
		}

		// Set the client area values
		cwidth = width, cheight = height;

		// Adjust the overall values to accomodate border and caption
		width  += 2*border;
		height += (2*border + caption);

		if (centre)
		{
			// default is centered
			// absolute size of screen does not account for bottom task bar
			xoff = ( dxsc - width ) / 2;
			yoff = ( dysc - height) / 2;
			cMode.xpos = cMode.ypos = 0;
		}
		else
		{
			xoff = cMode.xpos;
			yoff = cMode.ypos;
			set_custom = true;
		}

		// Check values requested are valid
		if (   xoff + width  >= dxsc
			|| yoff + height >= dysc
			|| xoff < 0
			|| yoff < 0)
		{
			if (lvscr.width && !firsttime)
			{
				memcpy(&cMode.width, &lvscr, SZWDATA);
				width  = cwidth  = cMode.width;
				height = cheight = cMode.height;
				xoff   = cMode.xpos;
				yoff   = cMode.ypos;
				set_custom = lvscr.custom;

				if (!cMode.fullscreen && !xoff && !yoff)
				{
					xoff = ( dxsc - width ) / 2;
					yoff = ( dysc - height - caption) / 2;
				}

				cMode.w_setdefaults = false;
			}
			else
			{
				xoff   = ( dxsc - width ) / 2;
				yoff   = ( dysc - height - caption) / 2;
				set_custom = false;

				cMode.w_setdefaults = true;
			}
		}

		if (   xoff + width >= cMode.max_width
			|| yoff + height + 20 >= cMode.max_height)
		{
			cMode.fullscreen = true;
			set_custom = false;
			width = cMode.max_width, height = cMode.max_height;

			// Set the correct window style and size.
			SetWindowLong(hWndMain, GWL_STYLE, FULLSCR_STYLE);
			SetWindowPos(hWndMain, 0, 0, 0, width, height, swp_flags);
		}
		else
		{
			SetWindowLong(hWndMain, GWL_STYLE, WINDOW_STYLE);	// set style
			SetWindowPos(hWndMain, 0, xoff, yoff,  width, height, swp_flags);
		}
	}

	if (ge_inf.nomouse)
	{
		if (cMode.fullscreen)
		{
			mouse_suppressed = true;
			ShowCursor(FALSE);
		}
		else if (mouse_suppressed)
		{
			mouse_suppressed = false;
			ShowCursor(TRUE);
		}
	}
	
	ShowWindow(hWndMain, SW_NORMAL);
	GetClientRect(hWndMain, &current_rect);

	cMode.width  = cwidth;
	cMode.height = cheight;

	// Model player sprite weapon scaling uses a fixed FOV of 90 degrees
	cMode.scalex = (float)scr_max_width  / (float)cwidth;
	cMode.scaley = (float)scr_max_height / (float)cheight;

	if (set_custom)
	{
		modeNlist[0].width  = cwidth;
		modeNlist[0].height = cheight;
		modeNlist[0].xpos   = xoff;
		modeNlist[0].ypos   = yoff;
	}
	else
	{
		cMode.xpos = cMode.ypos = 0;
	}

	if (cMode.w_setdefaults)
	{
		cMode.w_setdefaults = false;

		memcpy(&cMode.d_width, &cMode.width, SZWDATA);
		setHZvalues(cMode.freq, UIHZ_DEFAULT);
	}

	memcpy(&lvscr, &cMode.width, SZWDATA);
	lvscr.custom = set_custom;

	// Check for a monitor change
	if (cMode.fullscreen)
	{
		static int firsttime = true;

		if (firsttime)
		{
			changeDeskTopSize(cMode.width, cMode.height, cMode.freq);
			firsttime = false;
		}
	}

	resolution_set = true;
}

RDBOOL GL_CheckWindow(int *width, int *height)
{
	RECT	new_rect;
	int		w, h, xc, yc, rw = 2*cMode.szBorder;

	if (!modeNlist) return false;
	
	GetWindowRect(hWndMain, &new_rect);

	cMode.xpos = modeNlist[0].xpos = new_rect.left;
	cMode.ypos = modeNlist[0].ypos = new_rect.top;

	w = new_rect.right - new_rect.left - rw;
	h = new_rect.bottom - new_rect.top - 2*cMode.szBorder - cMode.szCaption;

	xc = cMode.xpos, yc = cMode.ypos;

	// Check values requested are valid
	if (xc + w > cMode.max_width + 2*cMode.szBorder)
		cMode.xpos = cMode.max_width - w - 2*cMode.szBorder;

	if (yc + h > cMode.max_height + 2*cMode.szBorder)
		cMode.ypos = cMode.max_height - h - 2*cMode.szBorder;

	if (xc < 0) cMode.xpos = 0;
	if (yc < 0) cMode.ypos = 0;

	if (xc != cMode.xpos || yc != cMode.ypos)
	{
		modeNlist[0].xpos = cMode.xpos;
		modeNlist[0].ypos = cMode.ypos;

		SetWindowPos(hWndMain, 0, cMode.xpos, cMode.ypos,
			new_rect.right - new_rect.left,
			new_rect.bottom - new_rect.top, swp_flags);
	}

	if (   w == cMode.width
		&& h == cMode.height)
	{
		modeNlist[0].width  = w;
		modeNlist[0].height = h;
		return false;
	}

	*width = w; *height = h;
	return true;
}

// ======================== end v220-3 screen handling extensions

// v220-3 Text changed
void HardwareNoGood(char *msg)
{
	Con_Error("OUTDATED HARDWARE: %s\n"
		"Regrettably your graphics card cannot support Risen3D.\n", msg);
}

// v220-4 Use a truetype font for screen messages

PFNGLWINDOWPOS2IPROC	pglWindowPos2i;
GLint					ttFontList;

//  R3D If successful this will be used by DR_SCP_Drawer() and in uimain
void InitTTFont(int height)
{
	// Initialise the font
	HFONT		hFont;
	LOGFONT		logfont;

	//----------------------

	GL_DeleteFontLists();
	if (ttFontList || !ge_inf.usettfont || !pglWindowPos2i) return;

	logfont.lfHeight = height/32;			// maintain at 32 rows
	logfont.lfWidth = 0;
	logfont.lfEscapement = 0;
	logfont.lfOrientation = 0;
	logfont.lfWeight = FW_SEMIBOLD;
	logfont.lfItalic = FALSE;
	logfont.lfUnderline = FALSE;
	logfont.lfStrikeOut = FALSE;
	logfont.lfCharSet = ANSI_CHARSET;
	logfont.lfOutPrecision = OUT_TT_ONLY_PRECIS;
	logfont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	logfont.lfQuality = DEFAULT_QUALITY;
	logfont.lfPitchAndFamily = DEFAULT_PITCH;
	strcpy(logfont.lfFaceName,"Arial");		// all systems should have this!

	// Create the font and display list

	if ( (hFont = CreateFontIndirect(&logfont)) )
	{
		SelectObject(mainHdc, hFont);

		// create list for glyphs 0 through 128
		ttFontList = glGenLists(128);

		if (glGetError() == GL_NO_ERROR)
			wglUseFontBitmaps(mainHdc, 0, 128, ttFontList);
		else
			ttFontList = 0;

		DeleteObject(hFont);	// No longer needed
	}

	if (!ttFontList)
		pr_OFinsert("Required Windows font \"%s\" is not installed\n",
		logfont.lfFaceName);
}

//===========================================================================
//
//===========================================================================
void GL_PostInit(int width, int height, RDBOOL centre, RDBOOL set_fov)
{
	curtex = 0;

	// v220-3 prepare the container values for the required screen size etc.
	subfn_cModeResolution(width, height, centre);
	InitTTFont(cMode.height);

	if (set_fov) GL_SetRatioAndFov();
}

//===========================================================================
//
//===========================================================================
void GL_MainInit()
{
	GLfloat fogcol[4] = { .54f, .54f, .54f, 1.0f };
	GLenum	flag = GL_FASTEST;
	double	version;

	PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),
			1,
			PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
			PFD_TYPE_RGBA,
			0, // cColorBits
			0, 0, 0, 0, 0, 0,
			0,
			0,
			0,
			0, 0, 0, 0,
			32,	// z depth
			0,
			0,
			PFD_MAIN_PLANE,
			0,
			0, 0, 0
	};

	HGLRC	hglrc;
	int		res, pixForm;

	//------------------------

	pfd.cColorBits = cMode.bpp;

	// Set the pixel format for the device context
	if (!(pixForm = ChoosePixelFormat(mainHdc, &pfd)))
	{
		res = GetLastError();
		Con_Error("drOpenGL.Init: Choosing of pixel format failed. Error %d", res);
	}

	SetPixelFormat(mainHdc, pixForm, &pfd);

	// Make sure that the driver is hardware-accelerated.
	DescribePixelFormat(mainHdc, pixForm, pfd.nSize, &pfd);

	if (pfd.dwFlags & PFD_GENERIC_FORMAT || !(pfd.dwFlags & PFD_SUPPORT_OPENGL))
	{
		Con_ErrorD( "HARDWARE PROBLEM: graphics card does not support OpenGL\n"
					"or graphics card drivers not installed correctly");
	}

	// Create the OpenGL rendering context.
	if (   !(hglrc = wglCreateContext(mainHdc))
		|| !wglMakeCurrent(mainHdc, hglrc))
	{
		Con_ErrorD("OpenGL init failed");
	}

	gl_inited = true;

	// OGL initialised - do some basic queries

	query("GL_EXT_texture_filter_anisotropic", &glx.extAniso);

	if (glx.extAniso)
	{
		glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxAniso);
	}

	// Check the maximum texture size.
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexSize);

	if (maxTexSize < 1024 || glGetError() != GL_NO_ERROR)
		HardwareNoGood("Max texture size supported less than 1024");

	version = strtod((const char *)glGetString(GL_VERSION), NULL);

	if (version < 1.3) HardwareNoGood("gl card driver version < 1.3");

	// Check for vsync support
	query("WGL_EXT_swap_control", &glx.extSwapinterval);

	if (glx.extSwapinterval)
	{
		vsptr = (PFNWGLSWAPINTERVALEXTPROC)wglGetProcAddress("wglSwapIntervalEXT");
		vsyncSupported = vsptr != NULL;
	}

	glGetError();
	glGetFloatv(GL_LINE_WIDTH_RANGE, linesizes);	// used in the automap
	if (glGetError() != GL_NO_ERROR) linesizes[0] = linesizes[1] = 1.f;	// crap card insurance

	pglWindowPos2i = (PFNGLWINDOWPOS2IPROC)wglGetProcAddress("glWindowPos2i");

	// v220-4 Use truetype font but only if no palette is needed
	if (pfd.dwFlags & PFD_NEED_PALETTE) ge_inf.usettfont = false;

	// Initialize the matrices
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 320, 200, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();

	// Set up conditions
	glClearDepth(1);
	glDepthFunc(GL_LESS);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// Automatic mipmap generation.
	query("GL_SGIS_generate_mipmap", &glx.extGenMip);

	if (glx.extGenMip)
	{
		flag = GL_NICEST;
		glHint(GL_GENERATE_MIPMAP_HINT_SGIS, GL_NICEST);
	}

	// Set hints
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, flag);
	glHint(GL_POLYGON_SMOOTH_HINT, flag);
	glHint(GL_POINT_SMOOTH_HINT, flag);
	glHint(GL_LINE_SMOOTH_HINT, flag);
	glHint(GL_FOG_HINT, flag);
	
	// Set default fog
	glDisable(GL_FOG);
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogi(GL_FOG_END, 2100);
	glFogfv(GL_FOG_COLOR, fogcol);
	whitefog = false;

	origvidVsync = -1;
}

void GL_Info()
{
	const GLubyte *ptr;

	pr_OFrepeat("\nOpenGL information ", "=", 50, 1);

	if (ArgExists("-verbose"))
	{
		char *token  = (char*)glGetString(GL_EXTENSIONS);
		char *extbuf = (char*)M_Malloc(strlen(token) + 1, 50);
		strcpy(extbuf, token);
		
		// Show the list of GL extensions.
		token = strtok(extbuf, " ");
		while(token)
		{
			// Two on one line, clamp to 32 characters.
			pr_OFinsert("  %-32.32s", token);
			if(token = strtok(NULL, " "))
				pr_OFinsert(" %-32.32s", token);
			pr_OFdirect("\n");
			token = strtok(NULL, " ");
		}
		M_Free(extbuf);
	}

	if ( (ptr = glGetString(GL_VENDOR  )) ) pr_OFinsert("  Vendor           : %s\n", ptr);
	if ( (ptr = glGetString(GL_RENDERER)) )
	{
		pr_OFinsert("  Renderer         : %s\n", ptr);
		if (strstr((char*)ptr, "GTX 1080")) cfg.glflush = true;
	}
	if ( (ptr = glGetString(GL_VERSION )) ) pr_OFinsert("  Version          : %s\n", ptr);

	glGetError();	// v220-4 In case one of the enums was not supported.

	pr_OFinsert("  Max texture size : %4i\n", maxTexSize);
	pr_OFinsert("  Vid swap interval: %ssupported\n", vsyncSupported ? "" : "not ");

	// v220-10 revised
	if (glx.extAniso && maxAniso > 1.0f)
	{
		int val = anisotropic;

		if (anisotropic < 1)
			anisotropic = 4;	// default
		else if (anisotropic > (int)maxAniso)
			anisotropic = (int)maxAniso;

		pr_OFdirect("  Anisotropic texture filtering enabled.\n");
		pr_OFinsert("  Anisotropic %.0f levels available, level %i requested, level %i set\n", maxAniso, val, anisotropic);
	}
	else
	{
		pr_OFdirect("  Anisotropic texture filtering is not supported\n");
		anisotropic = 1;
	}

	pr_OFrepeat("", "=", 69, 2);
}
