//-----------------------------------------------------------------------------
//
// Copyright (C) 2006-2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_actionScript.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 12/Jan/2013 gmj
// Revision 2.2.0 build 19 (v220-19) 21/Jan/2013 gmj
//		SETDEFAULTSKYREVERB added plus two pass script reading
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 34 (v220-34) 12/Nov/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: reads in an action script
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <eax.h>

#include "doom/d_englsh.h"
#include "doom/p_local.h"
#include "s_sound.h"
#include "r3d_MAINspec.h"
#include "r3d_edit.h"
#include "r3d_mapinfo.h"
#include "con_bar.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "m_winmem.h"
#include "rend_dyn.h"
#include "rend_model.h"
#include "sys_system.h"
#include "sys_timer.h"

// MACROS ------------------------------------------------------------------

#define	MAXTEXTCOUNT	300

#define MAXDROPNUM		100
#define MAXDROPITEM		  4

typedef enum {
	TXT79_MESSAGE,
	TXT79_LOCKED,
	TXT79_U79MSG
} txt79message_e;

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	Def_SetDMOnum(int num);
void	Def_ResetMobjinfo(void);
int		R_PrecacheSkinsForMoType(int type);
int		Def_SetScriptedNoInfight(char *id);
int		Def_GetMobjNum(char *id);
int		Def_GetThingHeight(char *id);
short	identifySpecial(line_t *line);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		skyboxname[34];
extern int		linkedcard[NUMCARDS][2];
extern byte		*moTypeArray;
extern int		pistoldamage;
extern int		plasmadamage;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

scriptMessage_t	scpMessage;

redirect_t	redir77[NUMSPOTS];
deathmon_t	deathmonitor[MAXDM_NUM];
fixed_t		*amodelintensity;
int			scp_keyscale[6];
byte		*linkSectors;
char		*music_list[20];	// 0x1FCC - 0x1FB8
int			scripted_hudscale = 14;	// Default width
byte		aggression_array[1000];
int			scpPickupValue[SPPICKUPLIMIT];	// v220-21 increased from 8 to 10
short		*listarray[NUMSPOTS];

// v220-31 used to check for things to be spawned
int			*preSpawnCache;
int			psc_num = 0;
static int	psc_maxnum = 0;

soundthing_t *sthing;

// v220-19 push sounds
#define MAXPUSHSOUNDS	8

typedef struct {
	int	dmednum;
	int	sndid;
} pushsound_t;

pushsound_t pushsounds[MAXPUSHSOUNDS];
int numpushsounds;

// v220-22
int		plrviewheight;
int		plrfiringheight;	// used with missiles

// v220-35
int		lockedsound;

// v220-35
ambiscroll_t scrollAMB[MAXAMBIENT];
float	evenLightFactor = 0.3f;
mobj_t	*vdoormo[MAXVDOORNUM][2];
int		vdoornum;
int		radx_rckt;	// change damage with rocket
int		radx_gren;	// change damage with grenade

// Used with scripted 67nn/77nn types
char	*linetext[MAXTEXTCOUNT];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static RDBOOL	debug_scp;
static int		linetextcount;
static int		mobjinfochanged = false;

typedef struct {
	int	item[4];
} dropitem_t;

dropitem_t *dropitems;

static 	int	FCOLIM_idx78xx;	// v220-4

char	*msga8001[10];
char	*msgb8001[10];

// CODE --------------------------------------------------------------------

// v220-35
// Used to determine whether a 7800 blocktextmsg
// should only be printed if on a use line special
static void set7800usetrigonlyflag(int n78)
{
	if (th7800[n78].refline78 >= 0)
	{
		line_t	*ld = &lines[th7800[n78].refline78];
		short	special = ld->special;
		int		n;

		if (   ld->special > 0					// detect special is set
			&& ld->tag							// tag must be set
			// Only use types allowed
			&& ((n = identifySpecial(ld)) == 0	// doom use triggers
			|| n == 5))							// scripted use triggers
		{
			th7800[n78].setflags |= T74_TRIGLINEONLY;
		}
	}
}

// v220-31 check types to be spawned for precacheing
static void addTypeNum(int dmednum)
{
	for (int i = 0; i < psc_num; i++)
	{
		if (preSpawnCache[i] == dmednum) return;
	}

	if (psc_num >= psc_maxnum)
	{
		psc_maxnum += 32;
		preSpawnCache = (int*)W_Realloc(preSpawnCache, psc_maxnum*sizeof(int), STATICID, 0);
	}

	preSpawnCache[psc_num++] = dmednum;
}

static int lineNum(mobj_t *mo)
{
	sector_t	*sec = mo->subsector->sector;
	int			i;

	for (i = 0; i < sec->linecount; i++)
	{
		if (PointIsOnLine(mo->x, mo->y, sec->lines[i], 1))
			return sec->lines[i] - lines;
	}
	return -1;
}

static short getMobjType(int num)
{
	int	t;

	if ( (t = P_GetTypeFromDoomednum(num)) == stats.num_mobjs)	// b211-0
	{
		SR_SCPErr("DOOMED NUMBER [%i] invalid", num);
	}
	
	Def_SetDMOnum(num);

	return (short)t;
}

static int getSPOTnum7400()
{
	int num = SR_NextNum() - 7400;

	if (num < 0 || num >= NUM7400SPOTS)
		SR_SCPErr("LINK THING not type 74nn");

	return num;
}

static int getSPOTnum7500()
{
	int num = SR_NextNum() - 7500;

	if (num < 0 || num >= NUM7400SPOTS)
		SR_SCPErr("LINK THING not type 75nn");

	return num;
}

static int getSPOTnum7700()
{
	int getnum = SR_NextNum();
	int num, pairnum;

	// v220-35
	if (getnum >= 7700 && getnum <= 7799) { num = getnum - 7700; pairnum = 7800 + num; }
	else
	if (getnum >= 6700 && getnum <= 6799) { num = getnum - 6700 + 100; pairnum = 6800 + num - 100; }
	else
	if (getnum >= 5700 && getnum <= 5799) { num = getnum - 5700 + 200; pairnum = 5800 + num - 200; }
	else
	if (getnum >= 4700 && getnum <= 4799) { num = getnum - 4700 + 300; pairnum = 4800 + num - 300; }
	else
		SR_SCPErr("LINK THING %i: not type 47nn or 57nn or 67nn or 77nn", getnum);

	if (!(th7900[num].setflags & T_SET))
		SR_SCPErr("LINK THING %i has no paired SCRIPT THING %i", getnum, pairnum);

	if (!th7700[num].spawn_count)
		SR_SCPErr("LINK THING %i not in map", getnum);

	return num;
}

static int getSPOTnum7800(RDBOOL test)
{
	int getnum = SR_NextNum();
	int num;

	// v220-35
	if (getnum >= 7800 && getnum <= 7899) num = getnum - 7800;
	else
	if (getnum >= 6800 && getnum <= 6899) num = getnum - 6800 + 100;
	else
	if (getnum >= 5800 && getnum <= 5899) num = getnum - 5800 + 200;
	else
	if (getnum >= 4800 && getnum <= 4899) num = getnum - 4800 + 300;
	else
		SR_SCPErr("SPAWN THING %i: not type 48nn or 58nn or 68nn or 78nn", getnum);

	if ( !(th7800[num].setflags & T_SET) )
		SR_SCPErr("SPAWN THING %i not found in map", getnum);

	if (test && (th7800[num].setflags & T_MULTISET) )
		SR_SCPErr("SPAWN THING %i has more than one map instance", getnum);

	return num;
}

static int getSPOTnum7900(RDBOOL fetch)
{
	int getnum = (fetch ? SR_NextNum() : SR_NumForString(NULL));
	int num;

	// v220-35
	if (getnum >= 7900 && getnum <= 7999) num = getnum - 7900;
	else
	if (getnum >= 6900 && getnum <= 6999) num = getnum - 6900 + 100;
	else
	if (getnum >= 5900 && getnum <= 5999) num = getnum - 5900 + 200;
	else
	if (getnum >= 4900 && getnum <= 4999) num = getnum - 4900 + 300;
	else
		SR_SCPErr("SCRIPT THING not type 49nn or 59nn or 69nn or 79nn");

	if (!(th7900[num].setflags & T_SET))
		SR_SCPErr("SCRIPT THING %i not found in map", getnum);

	if (th7900[num].setflags & T_MULTISET)
		SR_SCPErr("SCRIPT THING %i has more than one map instance", getnum);

	return num;
}

static int setRange(int num)
{
	if (num <=    0) return 1;	// ensures things first get going
	if (num > 16000) return 16000;

	return num;
}

static void setSpawn(th7900_t *t79, int spawnnum)
{
	th7800_t *t78 = &th7800[spawnnum];
	int		 sdnum = t79->spawndef_num;
	char	 *nextstr;

	if (sdnum > 9)
		SR_SCPErr("Max of 10 spawn definitions exceeded");

	t79->spawn78nn[sdnum] = spawnnum;

	nextstr = SR_NextSTR(true);

	if (strcmp(nextstr, "{")) SR_SCPErr("missing {");

	nextstr = SR_NextSTR(true);

	while (strcmp(nextstr, "}"))
	{
		int count = t78->spawn_count;

		if (count > 9)
			SR_SCPErr("Max of 10 spawned things exceeded");

		int doomednum = SR_NumForString(NULL);

		if (doomednum >= 0) addTypeNum(doomednum);

		// [0] Get doomednum and store translated type
		t78->spawn[count][0] = getMobjType(doomednum);


		// [1] Get z-offset
		t78->spawn[count][1] = setRange(SR_NextNum());

		// [2] Get delay
		t78->spawn[count][2] = setRange(SR_NextNum());

		// [3] Get fog
		t78->spawn[count][3] = SR_NextBOOL();

		t78->spawn_count++;

		nextstr = SR_NextSTR(true);
	}

	if (strcmp(nextstr, "}")) SR_SCPErr("missing }");

	t79->spawndef_num++;
}

static void setKeys(th7900_t *t79)
{
	char	 *nextstr = SR_NextSTR(true);

	if (strcmp(nextstr, "{")) SR_SCPErr("missing {");

	nextstr = SR_NextSTR(true);

	while (strcmp(nextstr, "}"))
	{
		RDBOOL m;

		if (strlen(nextstr) != 3 || (nextstr[2] != '0' && nextstr[2] != '1'))
			SR_SCPErr("Invalid key identifier \"%s\"", nextstr);

		m = nextstr[2] == '1';	// monster inhibit

		if      (!strnicmp(nextstr, "CB", 2)) { t79->t79keys |= T79_PK_BCARD ; if (m) t79->t79keys |= T79_MK_BCARD ; }
		else if (!strnicmp(nextstr, "CR", 2)) { t79->t79keys |= T79_PK_RCARD ; if (m) t79->t79keys |= T79_MK_RCARD ; }
		else if (!strnicmp(nextstr, "CY", 2)) { t79->t79keys |= T79_PK_YCARD ; if (m) t79->t79keys |= T79_MK_YCARD ; }
		else if (!strnicmp(nextstr, "SB", 2)) { t79->t79keys |= T79_PK_BSKULL; if (m) t79->t79keys |= T79_MK_BSKULL; }
		else if (!strnicmp(nextstr, "SR", 2)) { t79->t79keys |= T79_PK_RSKULL; if (m) t79->t79keys |= T79_MK_RSKULL; }
		else if (!strnicmp(nextstr, "SY", 2)) { t79->t79keys |= T79_PK_YSKULL; if (m) t79->t79keys |= T79_MK_YSKULL; }
		else SR_SCPErr("Unknown key identifier \"%s\"", nextstr);

		nextstr = SR_NextSTR(true);
	}

	if (strcmp(nextstr, "}")) SR_SCPErr("missing }");
}

static void setTaggedLineTrigger(int num)
{
	th7700_t *t77 = &th7700[num];
	int		 i, found = false;
	int		 nextnum = SR_NextNum();

	if (t77->direccount > 9)
		SR_SCPErr("SETTAGGED : Max tagged line defs exceeds limit of 10");

	for (i = 0; i < t77->spawn_count; i++)
	{
		t77info_t *info77 = &t77->info[i];

		if (info77->options & MTF_AMBUSH)
		{
			line_t *ld = &lines[t77->info[i].t77lnum];

			if (ld->tag && ld->special == 8107)
			{
				t77->direcset[t77->direccount  ][0] = i;
				t77->direcset[t77->direccount++][1] = nextnum;
				found = true;
			}
		}
	}
	if (!found)
		SR_SCPErr("SETTAGGED : Tagged line with special 8107 not found");
}

static void setTrigger(int num, int redirect)
{
	redirect_t	*rd = &redir77[num];
	th7700_t	*t77 = &th7700[redirect];
	int			i, nextnum, found = false;

	if (num == redirect) SR_SCPErr("Cannot set self");

	// Store the line special
	nextnum = SR_NextNum();

	for (i = 0; i < t77->spawn_count; i++)
	{
		t77info_t *info77 = &t77->info[i];

		if (info77->options & MTF_AMBUSH)
		{
			line_t	*li = &lines[info77->t77lnum];

			if (li->tag) continue;

			if (li->special && li->special != 8106)
				SR_SCPErr("Deaf %i has untagged line %i but its special is set to %i (only 8106 or 0 is allowed)",
							get77N(num), li - lines, li->special);

			rd->lnum = (int*)W_ReallocCLR(rd->lnum, sizeof(int  ) * (rd->count + 1), LEVELID, 6);
			rd->spec = (short*)W_ReallocCLR(rd->spec, sizeof(short) * (rd->count + 1), LEVELID, 7);

			rd->lnum[rd->count] = info77->t77lnum;
			rd->spec[rd->count] = nextnum & 0xFFFF;

			if (debug_scp)
				pr_OFinsert("SETTRIGGER %i trigline %5i type %3i\n", get77N(num), rd->lnum[rd->count], rd->spec[rd->count]);
			
			rd->count++;
			found = true;
		}
	}

	if (!found)
		SR_SCPErr("SETTRIGGER : found no untagged line with special unset");
}

static int getText()
{
	char	tmp[256];
	int		len, i;
	char	*nextstr = SR_NextSTR(true);

	if ( !(len = strlen(nextstr)) ) return -1;

	if (len > 255) len = 255;

	strncpy(tmp, nextstr, len);
	tmp[len] = 0;

	// check if already stored (could be repeated for locked doors)
	for (i = 0; i < linetextcount; i++)
	{
		if (!strcmp(linetext[i], tmp)) return i;
	}

	if (linetextcount < MAXTEXTCOUNT)	// most probably is
	{
		int num = linetextcount;

		linetext[linetextcount++] = (char*)W_Calloc(len + 1, LEVELID, 8);

		strcpy(linetext[num], tmp);

		return num;
	}

	SR_SCPErr("Too many text messages. Max allowed is %i\n", MAXTEXTCOUNT);

	return 0;	// Happy compiler :)
}

static void setText(th7900_t *t79, RDBOOL repeat, txt79message_e type)
{
	int num = getText();

	switch (type)
	{
	case TXT79_MESSAGE:
		if (t79->textmsg) SR_SCPErr("TEXT; textmsg already set");
		t79->textmsg = linetext[num];
		t79->t79flags |= T79_MESSAGESET;
		if (repeat) t79->t79flags |= T79_MESSAGEMANY;
		break;
	case TXT79_LOCKED:
		if (t79->lockmsg) SR_SCPErr("TEXT; lockmsg already set");
		t79->lockmsg = linetext[num];
		break;
	case TXT79_U79MSG:
		if (t79->u79msg) SR_SCPErr("TEXT; u79msg already set");
		t79->u79msg = linetext[num];
		break;
	}
}

static void scriptParse7900(int n79)
{
	th7900_t *t79 = &th7900[n79];
	char	 *nextstr = SR_NextSTR(true);
	int		 n78, nextnum, matched;

	if (strcmp(nextstr, "{")) SR_SCPErr("missing {");

	while (1)
	{
		nextstr = SR_NextSTR(true);
		matched = 0;

// CHANGETYPE
		if (!_stricmp(nextstr, "CHANGETYPE"))
		{
			mobjinfo_t	*info;
			int i, n = t79->replacemocount;

			if (n > 9)
				SR_SCPErr("CHANGETYPE [entry %i] exceeds max of 10", get79N(n79));

			t79->replacemo[n][0] = SR_NextNum();	// current doomednum
			t79->replacemo[n][1] = SR_NextNum();	// new doomednum
			t79->replacemo[n][2] = SR_NextBOOL();	// set as friend

			// Check types sent
			for (i = 0; i < 2; i++)
			{
				int type = P_GetTypeFromDoomednum(t79->replacemo[n][i]);

				if (type == stats.num_mobjs)
					SR_SCPErr("Invalid parm %i unrecognised doomednum %i)", i, t79->replacemo[n][i]);
				
				if ( !((info = &mobjinfo[type])->flags & MF_COUNTKILL) )
					SR_SCPErr("Invalid parm %i doomednum %i (must have countkill flag set)", i, t79->replacemo[n][i]);
				
				t79->replacemo[n][i] = type;
			}

			if (t79->replacemo[n][2])
				ge_inf.isfriendly = true;	// Ensure console player gets set

			mobjinfochanged = true;

			t79->replacemocount++;
		}
// KILL
		else if (!_stricmp(nextstr, "KILL"))
		{
			if (t79->killcount > 9)
				SR_SCPErr("Killcount with THING %i exceeds max of 10", get78N(n79));
			
			n79 = getSPOTnum7900(true);
			t79->killthing[t79->killcount][0] = n79;
			t79->killthing[t79->killcount][1] = setRange(SR_NextNum());	// timeout
			t79->killthing[t79->killcount++][2] = SR_NextBOOL();
		}
// KILLAMBIENT
		else if (!_stricmp(nextstr, "KILLAMBIENT"))
		{
			nextnum = SR_NextNum();

			nextnum -= 14001;

			if (nextnum < 0 || nextnum >= MAXAMBIENT)
			{
				SR_SCPErr("KILLAMBIENT number %i not in range 14001 to 14064", nextnum);
			}

			if (t79->killambient >= 0)
			{
				SR_SCPErr("KILLAMBIENT already set");
			}

			t79->killambient = nextnum;
		}
// NULLBLOCKLINE
		else if (!_stricmp(nextstr, "NULLBLOCKLINE"))
		{
			n78 = getSPOTnum7800(true);	// spawn spot

			if (t79->thingref78 >= 0)
				SR_SCPErr("link to %i already set", get78N(n78));

			t79->thingref78 = n78;
			t79->t79flags |= T79_NULLTRIGGER;
			t79->t79flags &= ~(T79_SETTRIGGER|T79_SETTOGGLE);
		}
// SETBLOCKLINE
		else if (!_stricmp(nextstr, "SETBLOCKLINE"))
		{
			n78 = getSPOTnum7800(true);	// spawn spot

			if (t79->thingref78 >= 0)
				SR_SCPErr("link to %i already set", get78N(n78));

			if (th7800[n78].refline78 < 0)
				SR_SCPErr("%i is not on a line", get78N(n78));

			t79->thingref78 = n78;
			t79->t79flags |= T79_SETTRIGGER;
			t79->t79flags &= ~(T79_NULLTRIGGER|T79_SETTOGGLE);
		}
// TOGGLEBLOCKLINE 
		else if (!_stricmp(nextstr, "TOGGLEBLOCKLINE"))
		{
			n78 = getSPOTnum7800(true);	// spawn spot

			if (t79->thingref78 >= 0)
				SR_SCPErr("link to %i already set", get78N(n78));

			t79->thingref78 = n78;
			t79->t79flags |= T79_SETTOGGLE;
			t79->t79flags &= ~(T79_NULLTRIGGER|T79_SETTRIGGER);
		}
// SETSECBLOCK
		else if (!_stricmp(nextstr, "SETSECBLOCK"))
		{
			n78 = getSPOTnum7800(true);	// spawn spot

			if (t79->thingref78 < 0)
				SR_SCPErr("SPOT %i is unreferenced", get78N(n78));

			if (!(th7800[n78].blocklines[0]))
				SR_SCPErr("SPOT %i block lines unset", get78N(n78));

			t79->t79flags |= T79_SETTRIGGER;
			t79->t79flags &= ~(T79_NULLTRIGGER|T79_SETTOGGLE);
		}
// UNSETSECBLOCK
		else if (!_stricmp(nextstr, "UNSETSECBLOCK"))
		{
			int	n78 = getSPOTnum7800(true);	// spawn spot

			if (t79->thingref78 < 0)
				SR_SCPErr("SPOT %i is unreferenced", get78N(n78));

			if (!(th7800[n78].blocklines[0]))
				SR_SCPErr("SPOT %i block lines unset", get78N(n78));

			t79->t79flags |= T79_NULLTRIGGER;
			t79->t79flags &= ~(T79_SETTRIGGER|T79_SETTOGGLE);
		}
// RESURRECT
		else if (!_stricmp(nextstr, "RESURRECT"))
		{
			t79->resurrect = setRange(SR_NextNum()) + 10;
		}
// TRANSFERTRIGGER
		else if (!_stricmp(nextstr, "TRANSFERTRIGGER"))
		{
		//	th7900_t *t79 = &th7900[t79];	// $change
			int		 n = getSPOTnum7400();

			if (t79->t79flags & T79_ACTIVATE)
				SR_SCPErr("SPOT NUM %i has been used by an ACTIVATE command", get79N(n79));

			if (t79->t79flags & T79_TRANSFER)
				SR_SCPErr("SPOT NUM %i has already had TRANSFERTRIGGER set", get79N(n79));

			if (t79->t79flags & T79_SWAPTRIGGER)
				SR_SCPErr("SPOT NUM %i TRANSFERTRIGGER cannot be used with SWAPTRIGGER", get79N(n79));

			if (!th7400[n].refline)
				SR_SCPErr("TRANSFERTRIGGER : No THING 74%i was found", n);

			if (!th7400[n].ptrnum)
				SR_SCPErr("TRANSFERTRIGGER : No THING 75%i was found with THING 74%i", n, n);

			th7400[n].flags = SR_NextNum();
			th7900[n79].activate = n;
			th7900[n79].t79flags |= T79_TRANSFER;
		}
// ACTIVATE
// ACTIVATEDELAY
		else if ( (matched = !_stricmp(nextstr, "ACTIVATEDELAY")) || !_stricmp(nextstr, "ACTIVATE") )
		{
		//	th7900_t *t79n = &th7900[num];
			int		 n = getSPOTnum7400();

			if (t79->t79flags & (T79_CHECKACTIVATE|T79_ACTIVATE))
				SR_SCPErr("SPOT NUM %i has already had ACTIVATE|ACTIVATEDELAY|CHECKU79ACTIVATE type set", get79N(n79));

			if (t79->t79flags & T79_TRANSFER)
				SR_SCPErr("SPOT NUM %i has been used by a TRANSFERTRIGGER command", get79N(n79));

			if (!th7400[n].refline || !th7400[n].refline->special)
				SR_SCPErr("ACTIVATE THING 74%.2i is not on a valid trigger line", n);

			t79->activate = n;
			t79->t79flags |= T79_ACTIVATE;

			if (matched)
			{
				if ( !(th7400[n].refline->frontsector->specFlags & SF_MOVSECTOR)
					|| th7400[n].refline->tag != th7400[n].refline->frontsector->tag)
				{
					SR_SCPErr("ACTIVATEDELAY THING 74%i is not on a valid trigger line", n);
				}
				
				th7400[n].delay = setRange(SR_NextNum());
			}
		}
// CHECKU79ACTIVATE
		else if (!_stricmp(nextstr, "CHECKU79ACTIVATE"))
		{
			int listnum = SR_NextNum();

			if (listnum < 0 || listnum >= NUMSPOTS || !listarray[listnum])
				SR_SCPErr("Invalid listnum or list not set");

			int n = getSPOTnum7400();

			if (t79->t79flags & (T79_CHECKACTIVATE|T79_ACTIVATE))
				SR_SCPErr("SPOT NUM %i has already had ACTIVATE|ACTIVATEDELAY|CHECKU79ACTIVATE type set", get79N(n79));

			if (t79->t79flags & T79_TRANSFER)
				SR_SCPErr("SPOT NUM %i has been used by a TRANSFERTRIGGER command", get79N(n79));

			if (!th7400[n].refline || !th7400[n].refline->special)
				SR_SCPErr("ACTIVATE THING 74%i is not on a valid trigger line", n);

			t79->list79num = listnum;
			t79->activate = n;
			t79->t79flags |= T79_CHECKACTIVATE;

			nextstr = SR_NextSTR(true);	// sound lumpname

			if (stricmp(nextstr, "NONE") && !t79->u79sndid && nextstr[0])
			{
				if (!(t79->u79sndid = R3D_AddNewSound(nextstr)) )
					SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);
			}

			setText(t79, false, TXT79_U79MSG);
		}
// CHECKU79LIST
		else if (!_stricmp(nextstr, "CHECKU79LIST"))
		{
			int listnum = SR_NextNum();

			if (listnum < 0 || listnum >= NUMSPOTS || !listarray[listnum])
				SR_SCPErr("Invalid listnum or list not set");

			nextstr = SR_NextSTR(true);	// sound lumpname

			if (stricmp(nextstr, "NONE") && !t79->u79sndid && nextstr[0])
			{
				if (!(t79->u79sndid = R3D_AddNewSound(nextstr)) )
					SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);
			}

			setText(t79, false, TXT79_U79MSG);
			t79->list79num = listnum;
			t79->t79flags |= T79_CHECKLIST;
		}
// ONLOCKED
		else if (!_stricmp(nextstr, "ONLOCKED"))
		{
			nextstr = SR_NextSTR(true);	// sound lumpname

			if (_stricmp(nextstr, "NONE"))
			{
				if (t79->locksndid > 0)
					SR_SCPErr("SCRIPT SPOT locked sound already set", nextstr);

				if (!(t79->locksndid = R3D_AddNewSound(nextstr)) )
					SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);
			}

			setText(t79, false, TXT79_LOCKED);
		}
// TEXT
		else if (!_stricmp(nextstr, "TEXT"))
		{
			setText(t79, SR_NextBOOL(), TXT79_MESSAGE);
		}
// CHANGETEXT
		else if (!_stricmp(nextstr, "CHANGETEXT"))
		{
			n79 = getSPOTnum7900(true);	// Pointer to SPOT with text already set

			if (&th7900[n79] == t79)
				SR_SCPErr("Cannot change self");
			if ( !(th7900[n79].t79flags & T79_MESSAGESET) )
				SR_SCPErr("Redirect only possible where TEXT keyword was used with %i", get79N(n79));

			int newnum = getText();

			if (newnum >= 0)
			{
				t79->changemsgnum = newnum + 1;
				t79->witht79num = n79;
			}
		}
// UNBLOCK
		else if (!_stricmp(nextstr, "UNBLOCK"))
		{
			t79->delay = setRange(SR_NextNum());
		}
// REMOVE
		else if (!_stricmp(nextstr, "REMOVE"))
		{
			t79->t79flags |= T79_REMOVE;
		}
// REMOVEHUDKEY
		else if (!_stricmp(nextstr, "REMOVEHUDKEY"))
		{
			t79->t79keys |= T79_REMOVEHUDKEY;
		}
// SPAWN
		else if (!_stricmp(nextstr, "SPAWN"))
		{
			setSpawn(t79, getSPOTnum7800(true));
		}
// SPAWNONCE
		else if (!_stricmp(nextstr, "SPAWNONCE"))
		{
			setSpawn(t79, getSPOTnum7800(true));
			t79->t79flags |= T79_SPAWNONCE;
		}
// IFKEYS
		else if (!_stricmp(nextstr, "IFKEYS"))
		{
			setKeys(t79);
		}
// RESETMODEL
		else if (!_stricmp(nextstr, "RESETMODEL"))
		{
			n79 = getSPOTnum7900(true);	// $check

			th7900[n79].t79flags |= T79_RESET;

			th7900[n79].resetdelay = setRange(SR_NextNum());
		}
// REPLACE
		else if (!_stricmp(nextstr, "REPLACE"))
		{
			t79->replacespot = getSPOTnum7900(true);
			t79->replacetype[0] = getMobjType(SR_NextNum());
		}
// TOGGLEREPLACE
		else if (!_stricmp(nextstr, "TOGGLEREPLACE"))
		{
			t79->replacespot = getSPOTnum7900(true);
			t79->replacetype[0] = getMobjType(SR_NextNum());
			t79->replacetype[1] = getMobjType(SR_NextNum());
			t79->t79flags |= T79_TRIGRESURRECT;
		}
// SETWAIT
		else if (!_stricmp(nextstr, "SETWAIT"))
		{
			int	n, t77 = getSPOTnum7700();
			th7900_t *t79n = &th7900[t77];

			if (t79n->t79flags & T79_VDOORWAIT)
				SR_SCPErr("SCRIPT SPOT SETWAIT already set");

			t79n->t79flags |= T79_VDOORWAIT;
			t79n->vdoorsec = t79->mo->subsector->sector - sectors;

			if (t79n->vdoorsec < 0 || t79n->vdoorsec >= numsectors)
				SR_SCPErr("Program error [please report]: vdoorsec %i", t79n->vdoorsec);

			if ((n = SR_NextNum()) < -32768 || n > 32767)
				SR_SCPErr("sec height out of range %i", n);

			t79n->vdoorsecheight = n*FRACUNIT;

			if ( (n = getText()) >= 0)
				t79n->vdoorwaitmsg = linetext[n];
		}
// SETTAGGED
		else if (!_stricmp(nextstr, "SETTAGGED"))
		{
			setTaggedLineTrigger(n79);
		}
// SETTRIGGER
		else if (!_stricmp(nextstr, "SETTRIGGER"))
		{
			setTrigger(n79, getSPOTnum7900(true));
		}
// PLAYSOUND
		else if (!_stricmp(nextstr, "PLAYSOUND"))
		{
			nextstr = SR_NextSTR(true);	// sound lumpname

			if (t79->soundid > 0)
				SR_SCPErr("SCRIPT SPOT sound already set");

			if ( !(t79->soundid = R3D_AddNewSound(nextstr)) )
				SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);

			if (SR_NextBOOL()) t79->soundid |= HSF_REPEAT;	// loop ?

			t79->volume = SR_NextFloat();

			if (t79->volume <   0 ) t79->volume = 0;
			if (t79->volume > 1.0f) t79->volume = 1.0f;

			if (!SR_NextBOOL()) t79->soundid |= HSF_NO_ATTENUATION;	// attenuate?
		}
// STOPSOUND
		else if (!_stricmp(nextstr, "STOPSOUND"))
		{
			int stopnum = getSPOTnum7900(true);

			t79->stopnum = stopnum + 1;
		}
// DISABLE
		else if (!_stricmp(nextstr, "DISABLE"))
		{
			int dis = getSPOTnum7800(false);
			int val = 1 << (dis % 32);
			int idx = dis / 32;

			if (T78_disable[n79][idx] & val)
				SR_SCPErr("ACTION SPOT already set");

			T78_disable[n79][idx] |= val;
		}
// ENABLE
		else if (!_stricmp(nextstr, "ENABLE"))
		{
			int dis = getSPOTnum7800(false);
			int val = 1 << (dis % 32);
			int idx = dis / 32;

			if (T78_disable[n79][idx] & val)
				SR_SCPErr("ACTION SPOT already set");

			th7800[dis].setflags |= T_ENABLE;
			T78_disable[n79][idx] |= val;
		}
// KILLMOBJ
		else if (!_stricmp(nextstr, "KILLMOBJ"))
		{
			redirect_t	*rd = &redir77[n79];

			if (strcmp(SR_NextSTR(true), "{"))
				SR_SCPErr("missing {");

			while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
			{
				char	*stopstring;
				int		num = strtol(nextstr, &stopstring, 10);

				if (*stopstring != 0)
					SR_SCPErr("Invalid numeric string \"%s\"", nextstr);

				rd->monum = (int*)W_ReallocCLR(rd->monum, sizeof(int) * (rd->mocount + 1), LEVELID, 9);

				rd->monum[rd->mocount++] = num;
			}
		}
		else if (!_stricmp(nextstr, "}"))
		{
			return;
		}
		else
		{
			SR_SCPErr("Unrecognised command %s or missing }", nextstr);
		}
	}
}

static void checkActiveLinks()
{
	int			i, k, n;
	RDBOOL		error = false;

	for (i = 0; i < NUMSPOTS; i++)
	{
		th7900_t *t79 = &th7900[i];
		th7700_t *t77 = &th7700[i];

		t79->activesec = -1;
		t79->killambient = -1;

		if ( (n = t77->spawn_count) > 0)
		{
			if ( !(t79->setflags & T_SET) )
			{
				PrintProgInfo("TYPE %i has no matching SCRIPT type", get77N(i));
				continue;
			}

			for (k = 0; k < n; k++)
			{
				line_t	*ld = &lines[t77->info[k].t77lnum];
				int		sector = -1;

				if (!ld->tag || ld->special == 8107) continue;	// 0x1FAB

				if (ld->special >= 0x1F80 && ld->special < 0x1FB0)
				{
					error = true;
					PrintProgInfo("LINKLINE %i's tagged line %i cannot use a script special",
																	get77N(i), ld - lines);
				}

				if (!t79->tag)
				{
					for ( ; (sector = P_FindSectorFromTag(ld->tag, sector)) >= 0 ; )
					{
						break;						// Only need to track one sector
					}

					if (sector >= 0)
					{
						// Check if this is a mov sector
						if (sectors[sector].specFlags & SF_MOVSECTOR)
							t79->t79flags |= T79_MOVSEC;

						t79->activesec = sector;
					}
					else
					{
						error = true;
						PrintProgInfo("THING %i's line %i tag %i has no matching sector tag",
											get77N(i), ld - lines, ld->tag);
					}

					t79->tag = ld->tag;
				}
				else if (ld->tag != t79->tag)
				{
					error = true;
					PrintProgInfo("%i is linked to multiple active sectors\n", get79N(i));
				}

				t77->info[k].t77special = ld->special;

				if (debug_scp)
					pr_OFinsert("TAGGED MOVSEC %i: LINE %5i sector %5i\n", get79N(i), ld-lines, t79->activesec);

				if ( !(t77->info[k].options & MTF_HARD) )
				{
					ld->special = 8064;
					ld->tag = 0;
				}
			}
		}
	}

	if (error) Sys_Quit(false);
}

static int checkSPACE(char *str, char *err)
{
	int len;

	while(*str && isspace(*str)) str++;

	if ( (len = strlen(str)) )
	{
		char *ptr = &str[len];

		while(--ptr > str && isspace(*ptr)) *ptr = 0;

		if ( (len = strlen(str)) ) return len;
	}

	SR_SCPErr(err); return 0;
}

static int getEAXtype(char *str)
{
//	if (!_stricmp(str, "MOUNTAINS"	)) return EAX_ENVIRONMENT_MOUNTAINS;	has no effect

	if (!_stricmp(str, "GENERIC"	)) return EAX_ENVIRONMENT_GENERIC;
	if (!_stricmp(str, "PADDEDCELL"	)) return EAX_ENVIRONMENT_PADDEDCELL;
	if (!_stricmp(str, "ROOM"		)) return EAX_ENVIRONMENT_ROOM;
	if (!_stricmp(str, "BATHROOM"	)) return EAX_ENVIRONMENT_BATHROOM;
	if (!_stricmp(str, "LIVINGROOM"	)) return EAX_ENVIRONMENT_LIVINGROOM;
	if (!_stricmp(str, "STONEROOM"	)) return EAX_ENVIRONMENT_STONEROOM;
	if (!_stricmp(str, "AUDITORIUM"	)) return EAX_ENVIRONMENT_AUDITORIUM;
	if (!_stricmp(str, "CONCERTHALL")) return EAX_ENVIRONMENT_CONCERTHALL;
	if (!_stricmp(str, "CAVE"		)) return EAX_ENVIRONMENT_CAVE;
	if (!_stricmp(str, "ARENA"		)) return EAX_ENVIRONMENT_ARENA;
	if (!_stricmp(str, "HANGAR"		)) return EAX_ENVIRONMENT_HANGAR;
	if (!_stricmp(str, "HALLWAY"	)) return EAX_ENVIRONMENT_HALLWAY;
	if (!_stricmp(str, "ALLEY"		)) return EAX_ENVIRONMENT_ALLEY;
	if (!_stricmp(str, "FOREST"		)) return EAX_ENVIRONMENT_FOREST;
	if (!_stricmp(str, "CITY"		)) return EAX_ENVIRONMENT_CITY;
	if (!_stricmp(str, "QUARRY"		)) return EAX_ENVIRONMENT_QUARRY;
	if (!_stricmp(str, "PLAIN"		)) return EAX_ENVIRONMENT_PLAIN;
	if (!_stricmp(str, "PARKINGLOT"	)) return EAX_ENVIRONMENT_PARKINGLOT;
	if (!_stricmp(str, "SEWERPIPE"	)) return EAX_ENVIRONMENT_SEWERPIPE;
	if (!_stricmp(str, "UNDERWATER"	)) return EAX_ENVIRONMENT_UNDERWATER;
	if (!_stricmp(str, "DRUGGED"	)) return EAX_ENVIRONMENT_DRUGGED;
	if (!_stricmp(str, "DIZZY"		)) return EAX_ENVIRONMENT_DIZZY;
	if (!_stricmp(str, "PSYCHOTIC"	)) return EAX_ENVIRONMENT_PSYCHOTIC;
	if (!_stricmp(str, "CARPETEDHALLWAY")) return EAX_ENVIRONMENT_CARPETEDHALLWAY;
	if (!_stricmp(str, "STONECORRIDOR"	)) return EAX_ENVIRONMENT_STONECORRIDOR;
	return -1;
}

static void reverbRangeError(char *type, float val)
{
	SR_SCPErr("SETREVERB: %s outside the allowed range %.2f", type, val);
}

static int getSoundData(float *data)
{
	RDBOOL	scperror = false;
	char	*eaxtype = SR_NextSTR(true);
	int		eaxnum = getEAXtype(eaxtype);
	float	val;

	if (eaxnum < 0)
		SR_SCPErr("reverb EAX type %s not recognised", eaxtype);

	data[SRD_PRESET] = (float)eaxnum;
	data[SRD_VOLUME] = SR_NextFloat();

	if (data[SRD_VOLUME] < 0 || data[SRD_VOLUME] > 1.0f)
		SR_SCPErr("SETREVERB: volume not in range 0 to 1");

	SR_NextSTR(true);

	val = SR_FloatForString(&scperror);	// Returns 0 if error

	if (scperror)
	{
		// Set defaults
		data[SRD_DECAY  ] = EAXLISTENER_DEFAULTDECAYTIME;
		data[SRD_DAMPING] = 1.0f;
		SR_KeepString();
	}
	else
	{
		data[SRD_DECAY] = val;

		if (data[SRD_DECAY] < EAXLISTENER_MINDECAYTIME)
			reverbRangeError("min decay", EAXLISTENER_MINDECAYTIME);
		if (data[SRD_DECAY] > EAXLISTENER_MAXDECAYTIME)
			reverbRangeError("max decay", EAXLISTENER_MAXDECAYTIME);

		data[SRD_DAMPING] = SR_NextFloat();

		if (data[SRD_DAMPING] < EAXLISTENER_MINROOMROLLOFFFACTOR)
			reverbRangeError("min damping", EAXLISTENER_MINROOMROLLOFFFACTOR);
		if (data[SRD_DAMPING] > EAXLISTENER_MAXROOMROLLOFFFACTOR)
			reverbRangeError("max damping", EAXLISTENER_MAXROOMROLLOFFFACTOR);
	}
	return scperror;
}

static int *traversed;

static void AS_RecurseSectors(sector_t *sec, int soundblocks)
{
	int i;
	
	// wake up all monsters in this sector
	if (sec->validcount == validcount && traversed[sec - sectors] <= soundblocks + 1)
		return;             // already flooded
	
	sec->validcount = validcount;
	traversed[sec - sectors] = soundblocks + 1;
	
	for (i = 0; i < sec->linecount; i++)
	{
		sector_t	*other;
		line_t		*ld = sec->lines[i];
		
		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| ld->specType >= R3D_40
			|| (MIN_OF(ld->frontsector->ceilingheight, ld->backsector->ceilingheight)
			  - MAX_OF(ld->frontsector->floorheight, ld->backsector->floorheight)) <= 0) // closed door
		{
			continue;
		}
		
		other = sides[ld->sidenum[sides[ld->sidenum[0]].sector == sec]].sector;
		
		if (!(ld->ml_flags & ML_SOUNDBLOCK) && !(ld->lnflags2 & LNF2_OCCLUDE))
			AS_RecurseSectors(other, soundblocks);
		else if (!soundblocks)
			AS_RecurseSectors(other, 1);
	}
}

void AS_CheckLinkedSeeSectors()
{
	RDBOOL	show_list = ArgExists("-show_list");
	int		c, i, k, group = 0;

	if (!linkSectors) return;
	
	traversed = (int*)M_Malloc(numsectors*sizeof(int), 0);

	for (i = 0; i < numsectors; i++)
	{
		if (linkSectors[i])
		{
			validcount++;
			memset(traversed, 0, numsectors*sizeof(int));

			AS_RecurseSectors(&sectors[i], 0);

			if (show_list)
				pr_OFinsert("\nSectors linked by T15600 in sector %i;\n", i);

			group++;

			for (c = k = 0; k < numsectors; k++)
			{
				if (traversed[k])
				{
					int set = sectors[k].sightgroup;

					sectors[k].sightgroup = group;

					if (show_list)
					{
						pr_OFinsert(" %5i", k);
						if (!(++c % 10)) pr_OFdirect("\n");
					}

					if (set)
						Con_Error("Sight Things 15600 overlap sectors");
				}
			}

			if (show_list)
			{
				pr_OFdirect("\n");
				if (c % 10) pr_OFdirect("\n");
			}
		}
	}

	M_Free(traversed);
	W_Free(linkSectors, LEVELID, 10);
}

static mobj_t *getMobjFromNextNum()
{
	thinker_t	*th;
	mobj_t		*iter;
	int			mnum = SR_NextNum();

	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		if (   th->function == (void (__cdecl *)(void))P_MobjThinker
			&& MAPNUM( (iter = (mobj_t*)th) ) == mnum)
		{
			return iter;
		}
	}
	
	SR_SCPErr("T%i not in map", mnum);
				
	return NULL;	// happy compiler
}

// v220-35 check for a vdoor linked mobj
static void CheckVdoorLinkedMobj(mobj_t *mo)
{
	thinker_t	*th;
	mobj_t		*iter;

	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		if (   th->function == (void (__cdecl *)(void))P_MobjThinker
			&& (iter = (mobj_t*)th) != mo
			&& iter->x == mo->x
			&& iter->y == mo->y
			&& iter->flags3 & MF3_AMODEL)
		{
			if (vdoormo[vdoornum][0]) continue;

			if (vdoornum >= MAXVDOORNUM)
				SR_SCPErr("VDOOR Map error: vdoors with linked mobj exceeds max allowed of %i", MAXVDOORNUM);

			vdoormo[vdoornum  ][0] = mo;
			vdoormo[vdoornum++][1] = iter;
			return;		// only one allowed
		}
	}
}

void AS_ResetMobjInfo()
{
	if (mobjinfochanged)
	{
		Def_ResetMobjinfo();
		mobjinfochanged = false;
	}
}

void Init_ActionScript(char *mapid)
{
	ded_mapinfo_t *mt = NULL;
	RDBOOL	inMap = false;
	RDBOOL	hudscaleset = false;
	RDBOOL	maxradset = false;
	char	*nextstr;
	byte	*secreverbset;
	int		skyreverbflags = 0;
	int		i, n, nextnum;
	int		n78, n79;
	// soundthings
	int		numsoundthings = 0;
	int		maxnumsoundthings = 0;

	//-------------------------------------

	// ptrs are LEVELID so will have been reclaimed by this point 
	sthing = NULL;
	evenLightFactor = 0.4f;

	memset(listarray, 0, sizeof(listarray));

	for (i = 0; i < NUMSPOTS; i++)
	{
		th7900[i].list79num = -1;
		th7900[i].thingref78 = -1;
	}

	// Clear any scripted stuff
	memset(&scpMessage,  0, sizeof(scriptMessage_t));
	memset(&music_list,  0, sizeof(music_list));
	memset(linkedcard , -1, sizeof(int) * 2 * NUMCARDS);
	memset(aggression_array, 0, 1000);
	memset(scp_keyscale, 0xff, sizeof(scp_keyscale));
	memset(scrollAMB, 0xff, sizeof(ambiscroll_t) * MAXAMBIENT);
	numpushsounds = 0;
	radx_rckt = 0;
	radx_gren = 0;

	for (i = 0; i < SPPICKUPLIMIT; i++) scpPickupValue[i] = 10;
	for (i = 0; i < 10; i++) msga8001[i] = msgb8001[i] = NULL;

	if (!ge_inf.noskyreset) skyboxname[0] = 0;

	dropitems      = NULL;
	linetextcount  = 0;
	set_intensity  = 0;
	FCOLIM_idx78xx = -1;
	plrviewheight = PLR_VIEWHT;	// default. Might be changed if scripted
	plrfiringheight	= 32 * FRACUNIT;

	if (!amodelintensity)
		amodelintensity = (fixed_t*)W_Calloc(7700 * sizeof(fixed_t), STATICID, 0);
	else
		memset(amodelintensity, 0, 7700 * sizeof(fixed_t));

	// v220-35
	vdoornum = 0;
	memset(vdoormo, 0, MAXVDOORNUM * 2 * sizeof(mobj_t*));

	if (  !R3D_ScriptedMap
		|| R3D_edit
		|| !CFX
		|| !W_OpenLatestScriptLump("SCP_DEFN", mapid))
	{
		R3D_ScriptedMap = false;
		return;
	}

	// Reverb settings are allowed with all doom iwads
	R3D_ScriptedMap = false;
	
	dropitems = (dropitem_t*)W_Calloc(MAXDROPNUM * sizeof(dropitem_t), LEVELID, 11);

	secreverbset = (byte*)W_Calloc(numsectors, STATICID, 0);

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (inMap || !_stricmp(nextstr, mapid))
		{
			if (!inMap)
			{
				mt = Def_GetMapInfo(nextstr);
				if (!mt) Con_Error("MAPINFO ERROR with id %s", nextstr);

				pr_OFinsert("SCP_DEFN   : Reading script for %s\n", mapid);

				if ( (debug_scp = ArgExists("-debug_scp")) )
					pr_OFdirect("\n\n==== SCRIPT Settings check ====\n\n");

				inMap = true;
				ma_inf.useSectorLight = ma_inf.r3dskyoffsets = true;	// v220-31 default settings for scripted maps
				if (cfg.oldbarscale < 0) cfg.oldbarscale = cfg.sbarscale;
				cfg.sbarscale = 1;
				continue;
			}

// SETDEFAULTSKYREVERB
			if (!_stricmp(nextstr, "SETDEFAULTSKYREVERB"))
			{
				float	 data[NUM_REVERB_DATA];

				if (skyreverbflags & 1)
					SR_SCPErr("SETDEFAULTSKYREVERB has already been set");
				if (skyreverbflags & 2)
					SR_SCPErr("SETDEFAULTSKYREVERB must precede all SETREVERB commands");
				if (skyreverbflags & 4)
					SR_SCPErr("SETDEFAULTSKYREVERB must precede all SETFLATSREVERB commands");

				skyreverbflags |= 1;

				getSoundData(data);

				for (i = 0; i < numsectors; i++)
				{
					if (sectors[i].ceilingpic == skyflatnum)
					{
						memcpy(&sectors[i].reverb, data, sizeof(sectors[i].reverb));
					}
				}
			}
// SETREVERB
			else if (!_stricmp(nextstr, "SETREVERB"))
			{
				reverb_t *rev;
				int		 i, n = SR_NextNum();
				float	 data[NUM_REVERB_DATA];
				
				if (n < 20000 || n > 20127)
					SR_SCPErr("reverb thing number %i not in range 20000 to 20127", n);

				skyreverbflags |= 2;

				rev = &reverbThing[n - 20000];

				if (rev->num != n)
					SR_SCPErr("reverb thing number %i not in map", n);

				getSoundData(data);

				for (i = 0; i < rev->seccount; i++)
				{
					int secn = rev->secnum[i];

					if (secreverbset[secn] == 2)
						SR_SCPErr("SECTOR %i reverb already set.\nThe SETREVERB keyword must precede all other reverb keywords\n", secn);

					// 3 cannot be overwritten but can overwrite all other settings
					secreverbset[secn] = 3;

					memcpy(&sectors[secn].reverb, data, sizeof(sectors[secn].reverb));
				}
			}
// SETFLATSREVERB	check for order
			else if (!_stricmp(nextstr, "SETFLATSREVERB"))
			{
				skyreverbflags |= 4;
			}
			else if (!_stricmp(nextstr, "END") || M_ValidateID(nextstr))
			{
				break;
			}
		}
	}

	SR_EndScript();
	W_OpenLatestScriptLump("SCP_DEFN", mapid);
	inMap = false;

	// Note: MAXTXTDELAY to PUSH_SOUND are common to all maps in a script
	//       and should  be scripted prior to any map specific stuff.
	while( (nextstr = SR_NextSTR(false)) )
	{
// MAXTXTDELAY
		if (!_stricmp(nextstr, "MAXTXTDELAY"))
		{
			extern int maxtxtdelay;
			maxtxtdelay = SR_NextNum();

			if (maxtxtdelay < 90 || maxtxtdelay > 180)
				SR_SCPErr("%i must be in the range 90 to 180", maxtxtdelay);
		}
// RADTIME
		else if (!_stricmp(nextstr, "RADTIME"))
		{
			int n = SR_NextNum();

			if (n < 10 || n > 90)
				SR_SCPErr("%i must be a percentage in the range 10 to 90", n);

			ma_inf.radtime = (float)(n)/100.f;
		}
// RADX_RCKT
		else if (!_stricmp(nextstr, "RADX_RCKT"))
		{
			radx_rckt = SR_NextNum();

			if (radx_rckt < 1 || radx_rckt > 10)
				SR_SCPErr("%i must be in the range 2 to 10", radx_rckt);
		}
// RADX_GREN
		else if (!_stricmp(nextstr, "RADX_GREN"))
		{
			radx_gren = SR_NextNum();

			if (radx_gren < 1 || radx_gren > 10)
				SR_SCPErr("%i must be in the range 2 to 10", radx_gren);
		}
// HUDSCALE
		else if (!_stricmp(nextstr, "HUDSCALE"))
		{
			int n = SR_NextNum();

			if (hudscaleset)
			{
				char *errmsg = "HUDSCALE has already been set";

				if (ge_inf.devmap) 
					SR_SCPErr(errmsg);
				else
					SR_SCPWarn(errmsg);
			}
			else
			{
				if (n < 10) n = 10;
				if (n > 40) n = 40;

				// This works in conjunction with cfg.hudScale
				scripted_hudscale = n;
				hudscaleset = true;
			}
		}
// HUDSCALEKEY	v220-13
		else if (!_stricmp(nextstr, "HUDSCALEKEY"))
		{
			int	idx = -1;

			nextstr = SR_NextSTR(false);

			if (!stricmp(nextstr, "RKEY")) idx = 0;
			if (!stricmp(nextstr, "YKEY")) idx = 1;
			if (!stricmp(nextstr, "BKEY")) idx = 2;
			if (!stricmp(nextstr, "RSKU")) idx = 3;
			if (!stricmp(nextstr, "YSKU")) idx = 4;
			if (!stricmp(nextstr, "BSKU")) idx = 5;

			if (idx >= 0)
			{
				int n = SR_NextNum();

				if (n < 0 || n > 60)
					SR_SCPErr("HUDSCALEKEY: scale value must be in range 0 to 60");

				scp_keyscale[idx] = n;
			}
			else
				SR_SCPErr("HUDSCALEKEY: %s not recognised", nextstr);
		}
// PUSH_SOUND	v220-21 bug fix
		else if (!_stricmp(nextstr, "PUSH_SOUND"))
		{
			int		sndid, snum, type;
			int		n = SR_NextNum();

			if (numpushsounds >= MAXPUSHSOUNDS)
				SR_SCPErr("PUSH_SOUND: max number of %i exceeded", MAXPUSHSOUNDS);
		
			nextstr = SR_NextSTR(true);

			if (   !nextstr
				|| !nextstr[0]
				|| (snum = W_CheckPWADNumForName(nextstr)) < 0
				|| ((type = DoomLumpIdentify(snum, NULL, NULL, DLI_SOUND)) != LMP_SND && type != LMP_WAV)
				|| !(sndid = R3D_AddNewSound(nextstr)) )
			{
				SR_SCPErr("PUSH_SOUND %s not found or not a sound file", nextstr);
			}

			for (snum = 0; snum < numpushsounds; snum++)
			{
				if (pushsounds[snum].sndid == sndid && pushsounds[snum].dmednum == n)
				{
					SR_SCPErr("PUSH_SOUND: dmednum %i %s already set", n, nextstr);
				}
			}

			pushsounds[numpushsounds].sndid = sndid;
			pushsounds[numpushsounds++].dmednum = n;
		}
// Restart inmap check
		else if (inMap || !_stricmp(nextstr, mapid))
		{
			if (!inMap)
			{
				mt = Def_GetMapInfo(nextstr);

				checkActiveLinks();

				R3D_ScriptedMap = true;

				memset(moTypeArray, 0, stats.num_mobjs);	// v220-13
				modelLights = 6;	// default for scripted map (can be changed in script)
				maxLightDistance = 4096.0f;	// scripted map default
				cfg.jumpPower = 8.f;
				useDynLights = true;
				water_trans = true;
				cfg.underwaterDamage = true;	// v220-6 explicitly set
				cfg.fallingDamage = true;
				cfg.underwaterDamage = true;
				cfg.jumpEnabled = true;
				cfg.footsteps = true;
				cfg.weaponsounds = true;
				cfg.splashsprite = true;

				if (!(ma_inf.noscriptset & NSF_SCPLIGHT))
				{
					cfg.useTorchLight = false;
					cfg.setVtxLight = true;
					cfg.sineCompress = false;
				}

				// Set defaults
				mt->R3Dflags |= (R3Dmapnameset|R3Dhasmapinfo);

				if (!(mt->R3Dflags & R3Dcustomded))
				{
					memset(mt->mapname, 0, DED_STRINGID_LEN + 1);
					memset(mt->author , 0, DED_STRINGID_LEN + 1);
					strncpy(mt->mapname, analysename, 20);
				}

				inMap = true;
				continue;
			}

// SPAWNSTART / SPAWNVDOOR
			if (   !_stricmp(nextstr, "SPAWNSTART")
				|| !_stricmp(nextstr, "SPAWNVDOOR"))
			{
				modeldef_t	*mf = NULL;
				spawnspot_t sp;
				int			vdoor_command = !_stricmp(nextstr, "SPAWNVDOOR"), onum;

				n79 = getSPOTnum7900(true);

				if (th7900[n79].mo)
					SR_SCPErr("%i has already been used", get79N(n79));

				memcpy(&sp, &th7900[n79], sizeof(mapthing_t));
				
				onum = SR_NextNum();

				sp.type			= getMobjType(onum);
				sp.spawn_fog	= 0;
				sp.z_off		= 0;
				sp.z_nofall		= 0;
				sp.sspflags		= 0;
				sp.mo			= NULL;

				P_SpawnScriptThing(&sp, NULL, get79N(n79), false, true);

				R_CheckModelFor(sp.mo, &mf, NULL);

				if (!mf || !R_GetVisibleFrame(mf, 0, sp.mo->thinker.id, 0) )
				{
					SR_SCPErr("\nAMODEL %i specified with %i %s\n"
							  "Check models have been properly installed",
							  onum, get79N(n79),
							  mf ? "is not present" : "has no visible frame");
				}

				sp.mo->misc |= MJMISC_LIT3D;

				if (!vdoor_command)
				{
					if (!SR_NextBOOL())
					{
						if (sp.mo->info->deathstate)
						{
							state_t *st;

							sp.mo->health = 0;

							sp.mo->state  = st = states + sp.mo->info->deathstate;
							sp.mo->tics   = st->tics;
							sp.mo->sprite = (spritenum_t)st->sprite;
							sp.mo->frame  = st->frame;

							th7900[n79].t79flags |= T79_INITDEAD;
						}
						else
							SR_SCPWarn("Model has no deathstate - spawn dead ignored");
					}
				}

				if (sp.mo->flags2 & MF2_OCCLUDE)
				{
					int lnum = lineNum(sp.mo);

					if (lnum < 0)
					{
						SR_SCPErr("Spawned thing has occlude flag but %i is not on a line", get79N(n79));
					}

					if (lines[lnum].frontsector != lines[lnum].backsector
						|| origTEX[lnum])
					{
						SR_SCPErr("Invalid occlusion line %i", lnum);
					}

					sp.mo->lineocclude = lnum;

					if (sp.mo->health) lines[lnum].lnflags2 |= LNF2_OCCLUDE;
				}

				th7900[n79].mo = sp.mo;

				if (vdoor_command)
				{
					sector_t *s = sp.mo->subsector->sector;
					RDBOOL	 found = false;
					int		 i, secnum = s - sectors;

					if (s->ceilingheight <= s->floorheight)
						SR_SCPErr("VDOOR Map error: sector %i's ceiling height must be > than the floorheight", s - sectors);

					// Required rendering height of containing sector
					s->ceiloffset = SR_NextNum() * FRACUNIT;

					for (i = 0; i < s->linecount; i++)
					{
						s->lines[i]->specType = R3D_00;
						s->lines[i]->mtflags = 0;
						s->lines[i]->ml_flags &= ~(ML_IGNOREMISSINGHI|ML_IGNOREMISSINGLO);
					}

					th7900[n79].startheight = th7900[n79].mo->subsector->sector->ceilingheight;
				
					// Required rise height of containing sector
					th7900[n79].offset = SR_NextNum();
					
					// Rise time in tics
					th7900[n79].doortics = SR_NextNum();
					if (th7900[n79].doortics < 1) th7900[n79].doortics = 1;

					th7900[n79].t79flags |= T79_VDOOR;

					sp.mo->flags2 |= MF2_VDOOR;
					sp.mo->z = sp.mo->subsector->sector->ceilingheight;
					
					CheckVdoorLinkedMobj(sp.mo);	// v220-35

					// Make sure the SF_MOVCEIL flag is set
					sectors[secnum].specFlags |= SF_MOVCEIL;

					// Add the sector to the movsec list
					AddToMovsecList(secnum);
				}
			}
// VDOOR_PARMS
			else if (!_stricmp(nextstr, "VDOOR_PARMS"))
			{
				int	wait, fall, crush;
				int	xpos, ypos, scperror;

				n79 = getSPOTnum7900(true);

				if (!th7900[n79].mo || !(th7900[n79].t79flags & T79_VDOOR))
				{
					SR_SCPErr("No VDOOR has been set with %i", get79N(n79));
				}

				if (th7900[n79].wait)
					SR_SCPErr("Wait time has already been set with %i", get79N(n79));

				wait = SR_NextNum();
				if (wait <= 0) wait = 0;
				if (wait >  0xFFFF) wait = 0xFFFF;

				fall  = SR_NextNum();
				crush = SR_NextNum();

				if (wait)
				{
					if (fall <= 0) fall = 1;
					if (fall >  0xFFFF) fall = 0xFFFF;

					th7900[n79].wait = wait;
					th7900[n79].falltics = fall;
					if (crush > 0) th7900[n79].t79flags |= T79_VDOORCRUSH;
					if (crush < 0) th7900[n79].t79flags |= T79_VDOORREVERSE;
				}

				nextstr = SR_NextSTR(true);

				// xpos and ypos are not mandatory but
				// if xpos is set then so must ypos be
				xpos = SR_NumForString(&scperror);

				if (scperror)
				{
					SR_KeepString();
				}
				else
				{
					ypos = SR_NextNum();

					th7900[n79].mo->x = xpos * FRACUNIT;
					th7900[n79].mo->y = ypos * FRACUNIT;
				}
			}
// INITBLOCKLINE 
			else if (!_stricmp(nextstr, "INITBLOCKLINE"))
			{
				n78 = getSPOTnum7800(true);	// spawn spot

				if (th7800[n78].refline78 < 0)
					SR_SCPErr("SPOT NUM %i is not on a line", get78N(n78));

				if (th7800[n78].refmo)
					SR_SCPErr("SPOT NUM %i already has the blocking mobj set", get78N(n78));

				n79 = getSPOTnum7900(true);	// spot num that activates

				if (!th7900[n79].mo)
					SR_SCPErr("SPOT NUM %i has no mobj set. Check script commands order.", get79N(n79));

				th7800[n78].refmo = th7900[n79].mo;

				if (SR_NextNum())	// if !0 set blockline
					lines[th7800[n78].refline78].lnflags2 |= LNF2_BLOCKALL;
				else
					lines[th7800[n78].refline78].lnflags2 &= ~LNF2_BLOCKALL;

				n = SR_NextNum();	// sector damage

				if (th7800[n78].refmo->subsector->sector->special || (n != 0 && n != 5 && n != 7 && n != 16))
					SR_SCPErr("Number %i not 0,5,7 or 16 or S%i special already set", n, th7800[n78].refmo->subsector->sector - sectors);

				th7800[n78].secdamage = n;
				th7800[n78].refmo->subsector->sector->special = n;

				if (SR_NextNum())
					set7800usetrigonlyflag(n78);	// trigonly blockmessage

				if ( (n = getText()) >= 0) th7800[n78].blocktextmsg = linetext[n];
				if ( (n = getText()) >= 0) th7800[n78].unblocktextmsg = linetext[n];
			}
// INITBLOCKSECLINES  
			else if (!_stricmp(nextstr, "INITBLOCKSECLINES"))
			{
				int		n79 = getSPOTnum7900(true);	// script spot linked to
				int		n78 = getSPOTnum7800(true);	// spot num that activates
				int		i, n;

				if (th7800[n78].refline78 < 0)
					SR_SCPErr("SPOT %i is not on a line", get78N(n78));

				if (th7900[n79].thingref78 >= 0)
					SR_SCPErr("SPOT %i already references type %i", get79N(n79), get79N(th7900[n79].thingref78));

				th7900[n79].thingref78 = n78;

				if (th7800[n78].blocklines[0])
					SR_SCPErr("SPOT NUM %i block lines have already been set", get78N(n78));

				sector_t	*sec = lines[th7800[n78].refline78].backsector;

				if (!sec) SR_SCPErr("SPOT NUM %i is on a 1's line", get78N(n78));

				for (i = n = 0; n < sec->linecount; n++)
				{
					line_t *ld = sec->lines[n];

					if (ld->backsector && ld->backsector != ld->frontsector)
					{
						if (i > 3) break;
						th7800[n78].blocklines[i++] = ld;
					}
				}

				if (i < 2 || i > 4)
				{
					SR_SCPErr("SPOT NUM %i 2's lines linecount invalid. It should be between 2 and 4", get78N(n78));
				}

				th7800[n78].blocklines[i] = 0;	// terminator belt and braces.

				for (i = n = 0; n < sec->linecount; n++)
				{
					line_t *ld = sec->lines[n];

					if (ld->backsector && ld->backsector != ld->frontsector) th7800[n78].blocklines[i++] = ld;
				}

				th7800[n78].blocklines[i] = NULL;	// terminator

				n = SR_NextNum();

				if (n != 0 && n != 1)
					SR_SCPErr("ON/OFF number should be 0 or 1 not %i\n", n);

				if (n)
				{
					for (i = 0; n < sec->linecount; n++)
					{
						line_t *ld = sec->lines[n];

						if (ld->backsector) ld->lnflags2 |= LNF2_BLOCKALL;
					}
				}
			}
// SWAPTRIGGER
			else if (!_stricmp(nextstr, "SWAPTRIGGER"))
			{
				int		 num = getSPOTnum7900(true);	// spot num that activates
				int		 an = getSPOTnum7900(true);		// to get sector
				int		 n = getSPOTnum7400();
				int		 triga, trigb, scperror;
				sector_t *sc = th7900[an].mo->subsector->sector;
				int		 tag = sc->tag;

				if (th7900[num].t79flags & T79_TRANSFER)
					SR_SCPErr("SPOT NUM %i has already had T79_TRANSFER set", get79N(num));

				if (!th7400[n].refline)
					SR_SCPErr("TRANSFERTRIGGER : THING 74%i was not found or is not on a line", n);

				if (!(sc->specFlags & SF_MOVSECTOR))
					SR_SCPErr("SPOT NUM %i is not in a move sector", get79N(num));

				if (!tag)
					SR_SCPErr("Sector %i has no tag", sc - sectors);

				nextstr = SR_NextSTR(true);
				triga = SR_NumForString(&scperror);
				if (scperror || triga & 0xFFFF0000) SR_SCPErr("Invalid numeric entry)", nextstr);

				nextstr = SR_NextSTR(true);
				trigb = SR_NumForString(&scperror);
				if (scperror || trigb & 0xFFFF0000) SR_SCPErr("Invalid numeric entry)", nextstr);

				th7900[num].t79flags |= T79_SWAPTRIGGER;
				th7400[n].msect = sc - sectors;
				th7400[n].triga = triga;
				th7400[n].trigb = trigb;
			}
// VDOOR_SCATTER
			else if (!_stricmp(nextstr, "VDOOR_SCATTER"))
			{
				int	scatter;
				int	num = getSPOTnum7900(true);

				if (!th7900[num].mo || !(th7900[num].t79flags & T79_VDOOR))
				{
					SR_SCPErr("No VDOOR has been set with %i", get79N(num));
				}

				scatter = SR_NextNum();	// percentage bullet scatter

				// Check scatter value
				if (scatter > 100) scatter = 100;
				if (scatter <   0) scatter =   0;

				th7900[num].mo->subsector->sector->scatter = (short)scatter;
			}
// USED79LIST
			else if (!_stricmp(nextstr, "USED79LIST"))
			{
				int	array[NUMSPOTS];
				int	arraycount = 0;
				int	getnum, num, scperror;
				int listnum = SR_NextNum();

				if (listnum < 0 || listnum >= NUMSPOTS)
					SR_SCPErr("listnum must be in range 0 to %i", NUMSPOTS - 1);

				if (listarray[listnum])
					SR_SCPErr("listnum has already been used");

				nextstr = SR_NextSTR(true);

				if (_stricmp(nextstr, "{")) SR_SCPErr("missing {");

				while(1)
				{
					nextstr = SR_NextSTR(true);
					if (!_stricmp(nextstr, "}")) break;

					getnum = SR_NumForString(&scperror);
					if (scperror) SR_SCPErr("Invalid spot num %s (or is closing \"}\" missing?)", nextstr);

					// v220-35
					if (getnum >= 7900 && getnum <= 7999) num = getnum - 7900;
					else
					if (getnum >= 6900 && getnum <= 6999) num = getnum - 6900 + 100;
					else
					if (getnum >= 5900 && getnum <= 5999) num = getnum - 5900 + 200;
					else
					if (getnum >= 4900 && getnum <= 4999) num = getnum - 4900 + 300;
					else
						SR_SCPErr("SCRIPT THING not type 49nn or 59nn or 69nn or 79nn");

					array[arraycount++] = num;
				}

				if (arraycount)
				{
					listarray[listnum] = (short*)W_Malloc((arraycount + 1) * sizeof(short), LEVELID, 8);

					for (i = 0; i < arraycount; i++)
					{
						listarray[listnum][i] = array[i];
					}

					listarray[listnum][i] = -1;	// terminator
				}
			}
// SETTHINGFLAGS
			else if (!_stricmp(nextstr, "SETTHINGFLAGS"))
			{
				thinker_t	*th;
				short		mt_type = getMobjType(SR_NextNum());
				mobjinfo_t	*inf = &mobjinfo[mt_type];
				int			doomednum = mobjinfo[mt_type].doomednum;
				int			flags = (SR_NextNum() & SSPF_MASK) << 28;
				RDBOOL		friendly = (flags & MF3_FRIEND) != 0;
				RDBOOL		T_or_B = (flags & (MF3_TOUCHY|MF3_BOUNCES)) != 0;
				
				if (!friendly && !T_or_B)
					SR_SCPErr("Unrecognised flags or no flags set");

				if (doomednum >= 6700 && doomednum <= 7999)
					SR_SCPErr("FLAGS cannot be set with scripted architectural models or spawnspots");

				if (friendly && T_or_B)
					SR_SCPErr("FRIEND cannot be set with TOUCHY or BOUNCES");

				if (   friendly
					&& !(inf->flags & MF_COUNTKILL || mt_type == MT_SKULL))
					SR_SCPErr("FRIEND can only be set with countkill or MT_SKULL types");

				if (T_or_B
					&& (inf->flags & MF_COUNTKILL || mt_type == MT_SKULL))
					SR_SCPErr("TOUCHY or BOUNCE cannot be set with countkill or MT_SKULL types");

				mobjinfo[mt_type].flags3 |= flags;
	
				if (friendly)
					ge_inf.isfriendly = true;	// Ensure the player gets set

				// Check for any that may have already been spawned
				for (th = thinkercap.next; th != &thinkercap; th = th->next)
				{
					if (th->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						mobj_t *mo = (mobj_t*)th;

						if (   friendly
							&& mobjinfo[mo->type].flags3 & MF3_FRIEND
							&& !(mo->flags3 & MF3_FRIEND))
						{
							mo->flags3 |= flags;
							P_UpdateThinker(&mo->thinker);	// ensure on correct list

							if (mo->flags & MF_COUNTKILL)
							{
								totalkills--;
								totalfriends++;
							}
						}

						if (   T_or_B
							&& mobjinfo[mo->type].doomednum == doomednum)
						{
							mo->flags3 |= flags;
						}
					}
				}
			}
// SET80NN
			else if (!_stricmp(nextstr, "SET80NN"))
			{
				nextnum = SR_NextNum();

				if (nextnum < 1 || nextnum > 10)
					SR_SCPErr("Only Things type 8000 to 8010 can be set");

				nextstr = SR_NextSTR(true);
				msga8001[nextnum - 1] = (char*)W_Calloc(strlen(nextstr) + 1, LEVELID, 0);
				strcpy(msga8001[nextnum - 1], nextstr);
				nextstr = SR_NextSTR(true);
				msgb8001[nextnum - 1] = (char*)W_Calloc(strlen(nextstr) + 1, LEVELID, 0);
				strcpy(msgb8001[nextnum - 1], nextstr);
			}
// ONDEATH
			else if (!_stricmp(nextstr, "ONDEATH"))
			{
				scriptParse7900(getSPOTnum7900(true));
			}
// ONRESET
			else if (!_stricmp(nextstr, "ONRESET"))
			{
				short	type;
				int		resetsound;
				
				nextnum = SR_NextNum();

				if (nextnum < 7000 || nextnum >= 7700)
					SR_SCPErr("Sound can only be set with amodel type [7000 to 7699]");

				type = getMobjType(nextnum);

				if (mobjinfo[type].attacksound > 0)
					SR_SCPErr("Mobj type %i already has attacksound defined", type);

				nextstr = SR_NextSTR(true);

				if ( !(resetsound = R3D_AddNewSound(nextstr)) )
					SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);

				mobjinfo[type].attacksound = resetsound;	// Save in attacksound
				mobjinfochanged = true;
			}
// RESETMANUAL
			else if (!_stricmp(nextstr, "RESETMANUAL"))
			{
				int		num = getSPOTnum7900(true);
				th7900[num].t79flags |= T79_TRIGRESURRECT;
			}
// LINKKEY
			else if (!_stricmp(nextstr, "LINKKEY"))
			{
				int		num = getSPOTnum7900(true);
				int		n = -1;

				nextstr = SR_NextSTR(true);

				if      (!_stricmp(nextstr, "CB")) n = it_bluecard;
				else if (!_stricmp(nextstr, "CY")) n = it_yellowcard;
				else if (!_stricmp(nextstr, "CR")) n = it_redcard;
				else if (!_stricmp(nextstr, "SB")) n = it_blueskull;
				else if (!_stricmp(nextstr, "SY")) n = it_yellowskull;
				else if (!_stricmp(nextstr, "SR")) n = it_redskull;

				if (n == -1)
					SR_SCPErr("Unrecognised keyid \"%s\"", nextstr);

				if (linkedcard[n][0] >= 0)
					SR_SCPErr("Slot for keyid \"%s\" has already been set", nextstr);

				linkedcard[n][0] = num;						// spotnum
				linkedcard[n][1] = setRange(SR_NextNum());	// delay
			}
// LINKPICKUP
			else if (!_stricmp(nextstr, "LINKPICKUP"))
			{
				thinker_t	*th;
				mobj_t		*mo = NULL;
				int			num = getSPOTnum7900(true);

				if (!th7900[num].mo)
					SR_SCPErr("%i spawndummy unset", get79N(num));

				if (th7900[num].pickupmo)
					SR_SCPErr("%i has already been used", get79N(num));

				// Get the object it is linked to
				for (th = thinkercap.next; th != &thinkercap; th = th->next)
				{
					if (th->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						if ( (mo = (mobj_t*)th)->flags & MF_SPECIAL
							&& mo->x>>FRACBITS == th7900[num].x
							&& mo->y>>FRACBITS == th7900[num].y)
						{
							int n = mo->info->doomednum;

							//  Eliminate keys
							if (n == 5 || n == 6 || n == 13 || n == 38 || n == 39 || n == 40)
								SR_SCPErr("LINKPICKUP links to a key. LINKKEY should be used");

							mo->misc |= MJF_SCP_PICKUP;
							th7900[num].pickupmo = mo;
							break;
						}
					}
				}

				if (!th7900[num].pickupmo)
					SR_SCPErr("No pickup Thing was linked to %i's Thing\n", get79N(num));
				/*	SR_SCPErr("No pickup Thing was linked to %i's Thing\n"
							  "Please ensure you have installed the latest\n"
							  "version of this game from our site", get79N(num));*/

				th7900[num].delay = setRange(SR_NextNum());
			}
// PICKUPDEFAULTS	v220-15
			else if (!_stricmp(nextstr, "PICKUPDEFAULTS"))
			{
				for (i = 0; i < SPPICKUPLIMIT; i++)
				{
					scpPickupValue[i] = SR_NextNum();
				}
			}
// SPAWNDUMMY
			else if (!_stricmp(nextstr, "SPAWNDUMMY"))
			{
				spawnspot_t sp;
				int			num = getSPOTnum7900(true);

				if (th7900[num].mo)
					SR_SCPErr("%i has already been used", get79N(num));

				memcpy(&sp, &th7900[num], sizeof(mapthing_t));

				sp.type			= getMobjType(SR_NextNum());
				sp.spawn_fog	= 0;
				sp.z_off		= 0;
				sp.z_nofall		= 0;
				sp.sspflags		= 0;
				sp.mo			= NULL;

				P_SpawnScriptThing(&sp, NULL, get79N(num), false, true);

				sp.mo->flags2 |= MF2_NORENDER;

				if (sp.mo->flags2 & MF2_OCCLUDE)
				{
					int lnum = lineNum(sp.mo);

					if (lnum < 0)
					{
						SR_SCPErr("Spawned thing has occlude flag but %i is not on a line", get79N(num));
					}

					if (lines[lnum].frontsector != lines[lnum].backsector
						|| origTEX[lnum])
					{
						SR_SCPErr("Invalid occlusion line %i", lnum);
					}

					sp.mo->lineocclude = lnum;

					if (sp.mo->health) lines[lnum].lnflags2 |= LNF2_OCCLUDE;
				}

				th7900[num].mo = sp.mo;
			}
// DROPITEM
			else if (!_stricmp(nextstr, "DROPITEM"))
			{
				int			i, num, thingidx = SR_NextNum();
				dropitem_t	*di; 

				thingidx -= 6000;

				if (thingidx < 0 || thingidx >= MAXDROPNUM)
					SR_SCPErr("DROPITEM thingidx must be between 6000 and %i", 6000 + MAXDROPNUM - 1);

				di = &dropitems[thingidx];

				for (i = 0; i < MAXDROPITEM; i++)
				{
					if ( (num = SR_NextNum()) )
					{
						di->item[i] = getMobjType(num);
						Def_SetDMOnum(num);
					}
					else
						di->item[i] = 0;
				}
			}
// COTRIGGER
			else if (!_stricmp(nextstr, "COTRIGGER"))
			{
				int	conum, num = getSPOTnum7900(true);

				if (!th7700[num].spawn_count)
					SR_SCPErr("No link spots found with %i", get79N(num));

				if (th7900[num].cotrigger)
					SR_SCPErr("Cotrigger already set with %i", get79N(num));

				conum = getSPOTnum7900(true);

				if (th7700[conum].spawn_count)
					SR_SCPErr("Cotrigger %i must not have link spots", get79N(conum));

				th7900[num].cotrigger = conum + 1;
			}
// SOUNDTHING
			else if (!_stricmp(nextstr, "SOUNDTHING"))
			{
				soundthing_t *sth;
				int			 i, doomednum = SR_NextNum();
				short		 type = getMobjType(doomednum);

				for (i = 0; i < numsoundthings; i++)
				{
					if (sthing[i].type == type)
						SR_SCPErr("type %i with doomednum %i already set", type, doomednum);
				}

				if (numsoundthings >= maxnumsoundthings)
				{
					maxnumsoundthings += 16;
					sthing = (soundthing_t*)W_Realloc(sthing, maxnumsoundthings * sizeof(soundthing_t), LEVELID, 12);
				}

				sth = &sthing[numsoundthings];

				// check there's not a problem
				if (mobjinfo[type].painsound < 0)
					Con_Error("program error: type %i painsound %i", type, mobjinfo[type].painsound);

				sth->type = type;

				// This will be treated as a pointer to the sthings array
				sth->oldsoundid = mobjinfo[type].painsound;	// v220-33 remember origial sound
				mobjinfo[type].painsound = numsoundthings|HSF_SNDTHING;
				numsoundthings++;
				mobjinfochanged = true;

				nextstr = SR_NextSTR(true);

				if (!(sth->soundid = R3D_AddNewSound(nextstr)) )
					SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);

				sth->volume      = SR_NextFloat();
				sth->attenuation = SR_NextFloat();

				if (sth->volume <   0 ) sth->volume =   0 ;
				if (sth->volume > 1.0f) sth->volume = 1.0f;
				if (sth->attenuation > MAX_ATTN) sth->attenuation = MAX_ATTN;
				if (sth->attenuation <        0) sth->attenuation = 0;
			}
// SOUNDGLOBAL
			else if (!_stricmp(nextstr, "SOUNDGLOBAL"))
			{
				if (strcmp(SR_NextSTR(true), "{"))	// opening {
					SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					mobj_t	*mo;
					int		num = getSPOTnum7900(false);

					if ( (mo = th7900[num].mo) == NULL)
						SR_SCPErr("%i has no THING set", get79N(num));

					mo->flags3 |= MF3_GLOBALAMBIENT;
				}
			}
// AMBSCROLL
			else if (!_stricmp(nextstr, "AMBSCROLL"))
			{
				int found = 0;
				int n = SR_NextNum() - 1;

				if (n < 0 || n > 63)
						SR_SCPErr("AMBSCROLL: number must be in range 1 to 64");

				if (compat[c_nosound]) continue;

				thinker_t	*thinker, *next;
				thinker = thinkercap.next;

				while(thinker != &thinkercap)
				{
					next = thinker->next;

					if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						mobj_t *iter = (mobj_t *) thinker;

						if (iter->type == MT_SOUNDSOURCE && TIDNUM(iter) == n)
						{
							if (!found)
							{
								scrollAMB[n].ambnum = n;
								scrollAMB[n].count = 0;
								scrollAMB[n].secnum = iter->subsector->sector - sectors;
								scrollAMB[n].mo = iter;
								found++;
							}
							else
							{
								SR_SCPErr("AMBSCROLL: only 1 type 140%.2i can be present but more than 1 found", n + 1);
							}
						}
						else if (iter->type == MT_TRIGAMBIENT && TIDNUM(iter) == n)
						{
							SR_SCPErr("AMBSCROLL: cannot be shared with a 139%.2i type", n + 1);
						}

					}
					thinker = next;
				}

				if (!found)
					SR_SCPErr("AMBSCROLL: no 140%2i types found", n + 1);
				else if (!S_ExistsAmbient(n))
					SR_SCPErr("AMBSCROLL: no ambient exists for type 140%.2i", n + 1);
			}
// PLAYLISTSOUND
			else if (!_stricmp(nextstr, "PLAYLISTSOUND"))
			{
				ded_state_t *dst;
				fixed_t	lim;
				int		num = getSPOTnum7900(true);
				int		i, n = SR_NextNum();
				int		*list;

				if (n < 1 || n > 4)
					SR_SCPErr("number of sounds [%i] not in range 1 to 4", n);

				list = th7900[num].soundlist;

				if (list[0])
					SR_SCPErr("soundlist already set with %i", get79N(num));

				list[0] = n; // number of sounds in list

				while(n > 0)
				{
					nextstr = SR_NextSTR(true);	// sound lumpname

					if (!(list[n] = R3D_AddNewSound(nextstr)) )
						SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);
					n--;
				}

				nextstr = SR_NextSTR(true);	// state to start sound

				for (i = 0, dst = defs.states; i < count_states.num; i++, dst++)
				{
					if (!_stricmp(dst->id, nextstr))
					{
						list[5] = Def_GetStateNum(dst->id);
						break;
					}
				}

				if (i == count_states.num)
					SR_SCPErr("state \"%s\" not found", nextstr);

				list[6] = SR_NextNum();	// min delay
				list[7] = SR_NextNum();	// max delay

				if (list[6] < 35)
					SR_SCPErr("min delay is < 35 tics");
				if (list[6] > 1050 || list[7] > 1050)
					SR_SCPErr("%s delay is > 1050 tics", list[6] > 1050 ? "min" : "max");
				if (list[7] < 0) list[7] = 0;

				// volume
				list[8] = (fixed_t)(SR_NextFloat()*FRACUNIT);
				lim = FRACUNIT;

				if (list[8] <   0) list[8] = 0;
				if (list[8] > lim) list[8] = lim;

				// attenuation
				list[9] = (fixed_t)(SR_NextFloat()*FRACUNIT);
				lim = (fixed_t)(MAX_ATTN*FRACUNIT);

				if (list[9] <   0) list[9] = 0;
				if (list[9] > lim) list[9] = lim;
			}
// SETWATERVARS
			else if (!_stricmp(nextstr, "SETWATERVARS"))
			{
				int	 sn = SR_NextNum();				// index 6100 to 6199

				sn -= 6100;

				if (sn < 0 || sn > 99)
					SR_SCPErr("identifier not in range 6100 to 6199", sn);

				// Get the fake sector number set in P_SpawnMapThing
				if (   fakeidentifier[sn] < 0
					&& fakeidentifier[sn] >= numsectors)
				{
					SR_SCPErr("identifier %i not found or invalid", sn + 6100);
				}
				else
				{
					watervar_t *wv = (watervar_t*)W_Calloc(sizeof(watervar_t), LEVELID, 13);

					int		op = SR_NextNum();		// opacity
					float	dn = SR_NextFloat();	// fog density
					//--------------------------

					if (sectors[fakeidentifier[sn]].wvar)
						SR_SCPErr("type %i already set", sn + 6100);

					sectors[fakeidentifier[sn]].wvar = wv;
					wv->opacity =  op & 0xFF;

					if (dn != 0)
					{
						int r = SR_NextNum();		// R, G, B
						int g = SR_NextNum();
						int b = SR_NextNum();

						wv->fogrgba[CR] = (float)(r & 0xFF) / 255.0f;
						wv->fogrgba[CG] = (float)(g & 0xFF) / 255.0f;
						wv->fogrgba[CB] = (float)(b & 0xFF) / 255.0f;
						wv->fogrgba[CA] = 1.0f;
						wv->fog_density = dn;
					}

					nextstr = SR_NextSTR(true);

					if ( !_stricmp(nextstr, "MOD") )
					{
						// modulation vars
						float md = SR_NextFloat();		// amplitude

						if (md <   0.0f) md = 0;
						if (md > 128.0f) md = 128.0f;

						wv->modulate = (int)(md * FRACUNIT);

						wv->period = SR_NextNum();
						if (wv->period < 1) wv->period = 1;

						wv->phase = SR_NextNum();

						if (wv->phase <   0) wv->phase = 0;
						if (wv->phase > 359) wv->phase = 359;
					}
					else
					{
						SR_KeepString();
					}
				}
			}
// SETGENERALREVERB
			else if (!_stricmp(nextstr, "SETGENERALREVERB"))
			{
				float data[NUM_REVERB_DATA];
				int   i;

				getSoundData(data);

				for (i = 0; i < numsectors; i++)
				{
					if (  !(sectors[i].specFlags & SF_IGNORE)
						&& !secreverbset[i]
						&& sectors[i].ceilingpic != skyflatnum)
					{
						memcpy(sectors[i].reverb, data, sizeof(sectors[i].reverb));
					}
				}
			}
// SETFLATSREVERB
			else if (!_stricmp(nextstr, "SETFLATSREVERB"))
			{
				float	data[NUM_REVERB_DATA];
				int		ceilnum = -1, floornum = -1;
				int		i, test;

				//-----------------------------

				if ( strlen(nextstr = SR_NextSTR(true)) )
				{
					if (!strcmp(nextstr, "SKY"))
						ceilnum = skyflatnum;
					else if ( (ceilnum = W_CheckNumForName(nextstr)) < 0)
						SR_SCPErr("ceil flat %s not found", nextstr);
				}

				if ( strlen(nextstr = SR_NextSTR(true)) )
				{
					if (!strcmp(nextstr, "SKY"))
						floornum = skyflatnum;
					else if ( (floornum = W_CheckNumForName(nextstr)) < 0)
						SR_SCPErr("ceil flat %s not found", nextstr);
				}

				if (ceilnum < 0 && floornum < 0)
					SR_SCPErr("A floor and/or ceil must be set");

				getSoundData(data);

				test = floornum >= 0 && ceilnum >= 0;

				for (i = 0; i < numsectors; i++)
				{
					if (secreverbset[i] == 3) continue;	// Thing 20xxx override

					if (   sectors[i].ceilingpic == skyflatnum
						&& ceilnum != skyflatnum)
					{
						continue;
					}

					if (   ( test &&  sectors[i].floorpic == floornum && sectors[i].ceilingpic == ceilnum)
						|| (!test && (sectors[i].floorpic == floornum || sectors[i].ceilingpic == ceilnum)) )
					{
						if (secreverbset[i] == 2)
							SR_SCPErr("Sector %i has already been set with a SETFLATSREVERB defn", i);

						memcpy(sectors[i].reverb, data, sizeof(sectors[i].reverb));
						secreverbset[i] = 2;
					}
				}
			}
// SETFLATSREVERB_FAKE	v220-4
			else if (!_stricmp(nextstr, "SETFLATSREVERB_FAKE"))
			{
				float	data[NUM_REVERB_DATA];
				int		floornum = -1;
				int		i, k;
				int		*slst = (int*)W_Malloc(numsectors * sizeof(int), STATICID, 9);
				int		scnt = 0;

				//-----------------------------

				if ( strlen(nextstr = SR_NextSTR(true)) )
				{
					if (!strcmp(nextstr, "SKY"))
						floornum = skyflatnum;
					else if ( (floornum = W_CheckNumForName(nextstr)) < 0)
						SR_SCPErr("ceil flat %s not found", nextstr);
				}

				if (floornum < 0) SR_SCPErr("A floor must be set");

				getSoundData(data);

				// Find all sectors referencing a fake sector with the specified floor flat
				for (i = 0; i < numsectors; i++)
				{
					if (secreverbset[i] == 3) continue;	// Thing 20xxx override

					if (  (k = sectors[i].heightsec) >= 0
						&& sectors[k].floorpic == floornum)
					{
						slst[scnt++] = i;
					}
				}

				for (i = 0; i < scnt; i++)
				{
					k = slst[i];

					if (secreverbset[k] == 2)
						SR_SCPErr("Sector %i already set with a SETFLATSREVERB defn", k);

					memcpy(sectors[k].reverb, data, sizeof(sectors[k].reverb));
					secreverbset[k] = 2;
				}

				W_Free(slst, STATICID, 0);
			}
// SETREVERB
			else if (!_stricmp(nextstr, "SETREVERB") || !_stricmp(nextstr, "SETDEFAULTSKYREVERB"))
			{
				RDBOOL	scperror = false;
				float	val;

				if (!_stricmp(nextstr, "SETREVERB")) SR_NextSTR(true);

				SR_NextSTR(true);
				SR_NextSTR(true);

				val = SR_FloatForString(&scperror);

				if (scperror)
				{
					SR_KeepString();
				}
				else
				{
					SR_NextSTR(true);
					SR_NextSTR(true);
				}
			}
// DUMMYHIGHFLOORSEC
			else if (!_stricmp(nextstr, "DUMMYHIGHFLOORSEC"))
			{
				int		i, k;
				int		tag = SR_NextNum();
				int		snd = SR_NextBOOL();

				if (tag <= 0)
					SR_SCPErr("bad parameter %i", tag);

				for (i = 0; i < numsectors; i++)
				{
					RDBOOL	secset = false;

					if (sectors[i].tag == tag)
					{
						fixed_t	ht = sectors[i].floorheight;

						for (k = 0; k < sectors[i].linecount; k++)
						{
							sector_t *s = sectors[i].lines[k]->frontsector;
							int		 sn = -1;

							if (s->floorheight > ht)
							{
								sn = s - sectors;
							}
							else if ( (s = sectors[i].lines[k]->backsector)
								&& s->floorheight > ht)
							{
								sn = s - sectors;
							}

							if (sn >= 0)
							{
								sectors[i].R3Dfloorsec = sn;
								sectors[i].specFlags2 |= SF2_DUMMYFLOOR;
								if (!snd)
									sectors[i].specFlags2 |= SF2_NOMOVSOUND;
								ht = s->floorheight;
							}
						}
					}
				}
			}
// FLATSOFFSET
			else if (!_stricmp(nextstr, "FLATSOFFSET"))
			{
				RDBOOL	term = false;

				if (strcmp(SR_NextSTR(true), "{"))	// opening {
					SR_SCPErr("missing {");

				while(1)
				{
					if ( !(nextstr = SR_NextSTR(false)) ) break;
					if (!_stricmp(nextstr, "}")) { term = true; break; }

					{
					int sn = SR_NumForString(NULL);	// Sector number
					int cx = SR_NextNum();			// Ceiling x offset
					int cy = SR_NextNum();			// Ceiling y offset
					int fx = SR_NextNum();			// Floor x offset
					int fy = SR_NextNum();			// Floor y offset

					if (sn < 0 || sn > numsectors)
						SR_SCPErr("Invalid sector num %i", sn);

					if (cx < -63 || cx > 63 || cy < -63 || cy > 63 ||
						fx < -63 || fx > 63 || fy < -63 || fy > 63)
					{
						SR_SCPErr("x and y offsets must be in the range -63 to 63");
					}

					sectors[sn].ceiloffx  += cx * FRACUNIT;
					sectors[sn].ceiloffy  += cy * FRACUNIT;
					sectors[sn].flooroffx += fx * FRACUNIT;
					sectors[sn].flooroffy += fy * FRACUNIT;
					}
				}

				if (!term)	// closing }
					SR_SCPErr("missing }");
			}
// THINGOFFSET		v220-21
			else if (!_stricmp(nextstr, "THINGOFFSET"))
			{
				thinker_t	*th;
				mobj_t		*iter;
				int			doomednum = SR_NextNum();
				int			offset = SR_NextNum();
				int			c;

				if (offset < 0 || offset >= 32768)
					SR_SCPErr("offset value must be positive and less than 32768");

				offset <<= FRACBITS;

				for (c = 0, th = thinkercap.next ; th != &thinkercap ; th = th->next)
				{
					if (   th->function == (void (__cdecl *)(void))P_MobjThinker
						&& (iter = (mobj_t*)th)->info->doomednum == doomednum)
					{
						if (!(iter->flags & MF_NOGRAVITY))
							SR_SCPErr("offsets can only be applied to Things with the mf_nogravity flag set");

						iter->z += offset;

						if (iter->z + iter->height > iter->subsector->sector->ceilingheight)
							iter->z = iter->subsector->sector->ceilingheight - iter->height;

						c++;
					}
				}
				
				if (!c && !nomonsters)
					SR_SCPErr("No Things with doomednum %i found in map", doomednum);
			}
// SETLINE
			else if (!_stricmp(nextstr, "SETLINE"))
			{
				int			soundid = 0;
				int			spotnum = getSPOTnum7700();
				th7700_t	*t77 = &th7700[spotnum];
				int			i, num, n = t77->spawn_count, onmove;

				onmove = !SR_NextBOOL();

				nextstr = SR_NextSTR(true);

				if (   _stricmp(nextstr, "NONE")	// i.e. not "none"
					&& !(soundid = R3D_AddNewSound(nextstr)) )
				{
					SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);
				}

				if (soundid)	// Only values > 0 will be accepted (should always be the case)
				{
					for (i = 0; i < n; i++)
					{
						line_t	*ld = &lines[t77->info[i].t77lnum];

						// If only one 77xx linkline type exists then a line
						// sound is always played.  More than one means only
						// those with MTF_EASY|MTF_NORMAL set will play.
						if (n == 1 || t77->info[i].options & (MTF_EASY|MTF_NORMAL))
						{
							ld->linesound = soundid;
							if (!onmove) ld->lnflags2 |= LNF2_LINESOUND;
						}
					}
				}

				if ( (num = getText()) >= 0)
				{
					for (i = 0; i < n; i++)
					{
						if (t77->info[i].options & (MTF_EASY|MTF_NORMAL))
							lines[t77->info[i].t77lnum].ltext = linetext[num];
					}
				}

			//	t77->
			}
// SEETHRULINE
			else if (!_stricmp(nextstr, "SEETHRULINE"))
			{
				int			spotnum = getSPOTnum7700();
				th7700_t	*t77 = &th7700[spotnum];
				int			i, n = t77->spawn_count;

				for (i = 0; i < n; i++)
				{
					line_t	*ld = &lines[t77->info[i].t77lnum];

					ld->lnflags2 |= LNF2_ALLOWSIGHT;
				}
			}
// SURROGATETRIG
			else if (!_stricmp(nextstr, "SURROGATETRIG"))
			{
				int			spotnum = getSPOTnum7700();
				th7700_t	*t77 = &th7700[spotnum];
				sector_t	*sec;

				t77->info->t77dummyspecial = (short)SR_NextNum();

				if ( (i = t77->info->t77dummyspecial) != 1 && i != 117)
					SR_SCPErr("special %i not 1 or 117", t77->info->t77dummyspecial);

				t77->info->t77dummytag = (short)SR_NextNum();

				if (!t77->info->t77dummytag)
					SR_SCPErr("Tag 0 not allowed");

				for (i = 0; i < numsectors; i++)
				{
					if (sectors[i].tag == t77->info->t77dummytag)
					{
						sec = &sectors[i]; break;
					}
				}

				if (!sec)
					SR_SCPErr("Sector not found with tag %i", t77->info->t77dummytag);
		
				sec->specFlags |= SF_MOVCEIL;

				// Add the sector to the movsec list
				AddToMovsecList(sec - sectors);
			}
// SETZ / SETLZ
			else if (!_stricmp(nextstr, "SETZ" )
				||   !_stricmp(nextstr, "SETLZ") )
			{
				mobj_t	*mo;
				RDBOOL	setlz = !_stricmp(nextstr, "SETLZ");
				int		llevel, z, num = getSPOTnum7900(true);

				if ( (mo = th7900[num].mo) == NULL)
					SR_SCPErr("%i has no THING set", get79N(num));

				if (setlz)
				{
					llevel = SR_NextNum();			// Light level

					if (llevel <   0) llevel =   0;
					if (llevel > 255) llevel = 255;

					mo->scplightlevel = (short)llevel;
				}

				z = setRange(SR_NextNum());

				mo->z = mo->floorz = mo->subsector->sector->floorheight + z * FRACUNIT;
				if (SR_NextBOOL()) mo->dropoffz = mo->z;
			}
// INTERREACT
			else if (!_stricmp(nextstr, "INTERREACT"))
			{
				int   doomednum = SR_NextNum();
				short type = getMobjType(doomednum);

				if (doomednum < 7000 || doomednum >= 7700)
					SR_SCPErr("Doomednum %i is not an AMODEL", doomednum);

				mobjinfo[type].flags3 |= MF3_INTERREACT;
				mobjinfochanged = true;
			}
// DYNLIGHTNUM
			else if (!_stricmp(nextstr, "DYNLIGHTNUM"))
			{
				modelLights = SR_NextNum();

				if (modelLights <  4) modelLights =  4;
				if (modelLights > MAX_MODEL_LIGHTS) modelLights = MAX_MODEL_LIGHTS;
			}
// DYN7000
			else if (!_stricmp(nextstr, "DYN7000"))
			{
				int   doomednum = SR_NextNum();
				short type = getMobjType(doomednum);
				float intensity = SR_NextFloat();

				if (doomednum < 7000 || doomednum >= 7700)
					SR_SCPErr("Doomednum %i is not an AMODEL", doomednum);

				if (intensity <  0.05f) intensity = 0.05f;
				if (intensity >= 1.0f ) intensity = 0;	// Since 1 means no change and 0 is ignored

				if (intensity)
				{
					amodelintensity[doomednum - 7000] = (fixed_t)(intensity*FRACUNIT);
				}
			}
// DYNGLOBAL
			else if (!_stricmp(nextstr, "DYNGLOBAL"))
			{
				int spare = SR_NextNum();

				set_intensity = SR_NextFloat();

				if (set_intensity <  0   ) set_intensity = 0.05f;
				if (set_intensity >= 1.0f) set_intensity = 0;
			}
// DLMAXRAD
			else if (!_stricmp(nextstr, "DLMAXRAD"))
			{
				if (maxradset)
					SR_SCPErr("DLMAXRAD has already been set");

				nextnum = SR_NextNum();

				if (nextnum <  64) nextnum =  64;
				if (nextnum > 256) nextnum = 256;

				dlMaxRad   = nextnum;
				dlMaxRadft = nextnum << FRACBITS;
				maxradset = true;
			}
// MAPNAME
			else if (!_stricmp(nextstr, "MAPNAME"))
			{
				int len;

				if (!mt) Con_ErrorD("MAPINFO unset");

				nextstr = SR_NextSTR(true);

				len = checkSPACE(nextstr, "mapname not set");

				if (len > 20) len = 20;
				memset(mt->mapname, 0, DED_STRINGID_LEN + 1);
				memset(mt->author , 0, DED_STRINGID_LEN + 1);
				strncpy(mt->mapname, nextstr, len);

				nextstr = SR_NextSTR(true);

				len = checkSPACE(nextstr, "authorname not set");

				if (len > DED_STRINGID_LEN) len = DED_STRINGID_LEN;
				strncpy(mt->author, nextstr, len);
			}
// MAPFOG
			else if (!_stricmp(nextstr, "MAPFOG"))
			{
				ded_mapinfo_t *dmt = Def_GetMapInfo(mapid);

				nextstr = SR_NextSTR(true);

				strcpy(dmt->flags, nextstr);
				dmt->fog_start    = SR_NextFloat();
				dmt->fog_end      = SR_NextFloat();
				dmt->fog_density  = SR_NextFloat();
				dmt->fog_color[0] = (byte)SR_NextNum();
				dmt->fog_color[1] = (byte)SR_NextNum();
				dmt->fog_color[2] = (byte)SR_NextNum();
				dmt->fog_color[3] = 0xFF;
			}
// FCOLIMIT
			else if (!_stricmp(nextstr, "FCOLIMIT"))
			{
				if (FCOLIM_idx78xx >= 0)
					SR_SCPErr("FCOLIMIT has already been set with %i", get78N(FCOLIM_idx78xx));

				FCOLIM_idx78xx = getSPOTnum7800(false);
			}
// BERSERKBUG
			else if (!_stricmp(nextstr, "BERSERKBUG"))
			{
				cfg.noberserkBug = !SR_NextBOOL();	// v220-6 was cfg.berserkBug
			}
// FALLINGDAMAGE
			else if (!_stricmp(nextstr, "FALLINGDAMAGE"))
			{
				cfg.fallingDamage = SR_NextBOOL();
			}
// MUSICONCE
			else if (!_stricmp(nextstr, "MUSICONCE"))
			{
				ded_mapinfo_t *dmt = Def_GetMapInfo(mapid);

				if (dmt) dmt->R3Dflags |= R3Dmusiconce;
			}
// PRECACHE Note: no longer necessary with v220-33 on - left for script compatibility
			else if (!_stricmp(nextstr, "PRECACHE"))
			{
				if (strcmp(SR_NextSTR(true), "{"))
					SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					char	*stopstring;
					int		t, num = strtol(nextstr, &stopstring, 10);

					if (*stopstring != 0)
						SR_SCPErr("Invalid numeric string \"%s\"", nextstr);

					t = getMobjType(num);
					if (!R_PrecacheSkinsForMoType(t))
						SR_SCPErr("Invalid doomednum %i", t);
				}
			}
// OCCLUDEWITHSECTORS
			else if (!_stricmp(nextstr, "OCCLUDEWITHSECTORS"))
			{
				if (strcmp(SR_NextSTR(true), "{"))
					SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					char	*stopstring;
					int		t, num = strtol(nextstr, &stopstring, 10);

					if (*stopstring != 0)
						SR_SCPErr("Invalid numeric string \"%s\"", nextstr);

					if (num < 0 || num >= numsectors)
						SR_SCPErr("Invalid sector number %i", num);

					for (t = 0; t < sectors[num].linecount; t++)
						sectors[num].lines[t]->xtraflags |= XT_OCCLUDE;
				}
			}
// ENDLEVEL
			else if (!_stricmp(nextstr, "ENDLEVEL"))
			{
				int		len, num, startdelay, enddelay;
				RDBOOL	term = false;

				if (scpMessage.smflags & SM_ENDSET)
					SR_SCPErr("ENDLEVEL: already set");

				scpMessage.smflags |= SM_ENDSET;
				startdelay = setRange(SR_NextNum());
				num = SR_NextNum();
				enddelay = setRange(abs(num));
				if (enddelay < 5) enddelay = 5;	// Min of 5/35 second

				enddelay += startdelay;

				if (num < 0)
					scpMessage.smflags |= SM_DONTENDLEVEL;

				if (SR_NextBOOL()) scpMessage.smflags |= SM_MASSACRE;

				nextstr = SR_NextSTR(true);

				if ( (len = strlen(nextstr)) )
				{
					if (len > 255) len = 255;

					scpMessage.endlevelMessage = (char*)W_Calloc(len + 1, LEVELID, 14);
					scpMessage.endlevel_start_delay = startdelay;
					scpMessage.endlevel_delay_to_fade = enddelay;
					strncpy(scpMessage.endlevelMessage, nextstr, len);
				}

				nextstr = SR_NextSTR(true);

				if (_stricmp(nextstr, "{")) SR_SCPErr("Missing {");

				while(1)
				{
					if ( !(nextstr = SR_NextSTR(false)) ) break;
					if (!_stricmp(nextstr, "}")) { term = true; break; }

					num = atoi(nextstr);

					if      (num >= 7900 && num <= 7999) num -= 7900;
					else num = -1;

					if (num < 0)
						SR_SCPErr("num not in range 7900 to 7999");

					th7900[num].t79flags |= T79_ENDLEVELTYPE;
				}

				if (!term)
					SR_SCPErr("Missing }");
			}
// MAPTHING_ANGLE	v220-3
			else if (!_stricmp(nextstr, "MAPTHING_ANGLE"))
			{
				uint	an;
				int		angle;
				mobj_t	*mobj = getMobjFromNextNum();

				// Get and convert the angle. Allow for two decimal places
				angle = ((int)(SR_NextFloat() * 100)) % 36000;

				if (angle < 0) angle += 36000;

				an = (uint)(ANG45 * ((float)angle/4500.0f));

				mobj->angle = an;
				mobj->visangle = an >> 16;
				mobj->startangle = 0;
			}
// ROTATEM
			else if (!_stricmp(nextstr, "ROTATEM"))
			{
				modeldef_t	*mf = NULL;
				model_t	*mdl;
				mobj_t	*mobj = getMobjFromNextNum();

				if (mobj->rotateperiod)
					SR_SCPErr("ROTATEM: T%i has already been set", MAPNUM(mobj));

				R_CheckModelFor(mobj, &mf, NULL);

				if (   !mf
					|| !(mdl = modellist[mf->sub[0].model])
					|| mdl->info.numFrames != 1)
				{
					SR_SCPErr("Invalid model with T%i. Frames not equal to 1\n", MAPNUM(mobj));
				}

				mobj->rotateperiod = SR_NextFloat() * 1000.0f;

				if (mobj->rotateperiod <= 0)
					SR_SCPErr("ROTATEM: T%i rotate period must be > 0", MAPNUM(mobj));

				mobj->rotateprev = Sys_GetRealTime();
			}
// MAXLIGHTDISTANCE
			else if (!_stricmp(nextstr, "MAXLIGHTDISTANCE"))
			{
				maxLightDistance = SR_NextFloat();

				if (maxLightDistance < 1024.0f) maxLightDistance = 1024.0f;
				if (maxLightDistance > 8192.0f) maxLightDistance = 8192.0f;
			}
// SKYBOXTEXTURE	v220-6
			else if (!_stricmp(nextstr, "SKYBOXTEXTURE"))
			{
				nextstr = SR_NextSTR(true);

				if (strlen(nextstr) > 32)
					SR_SCPErr("SKYBOXTEXTURE: strlen exceeds max allowed of 32");

				strcpy(skyboxname, nextstr);
			}
// LINEMUSIC
			else if (!_stricmp(nextstr, "LINEMUSIC"))
			{
				char	temp[256];
				int		n = SR_NextNum();
				uint	len;

				if (n < 8120 || n > 8139)
					SR_SCPErr("LINEMUSIC: special must be in range 8120 to 8139");

				n -= 8120;

				if (music_list[n])
					SR_SCPErr("LINEMUSIC: Entry %i has already been set", n);

				nextstr = SR_NextSTR(true);

				sprintf(temp, "%sData\\custom.music\\%s", ddRuntimeDir.path, nextstr);

				if (   !(len = strlen(nextstr))
					|| nextstr[len - 1] == '\\'
					|| nextstr[len - 1] == '/'
					|| (_access(temp, 4)) == -1)
				{
					SR_SCPErr("LINEMUSIC: %s invalid filename or file not found", nextstr);
				}

				music_list[n] = (char*)W_Malloc(strlen(temp) + 1, LEVELID, 10);

				strcpy(music_list[n], temp);
			}
// CUSTOMPUNCHVARS
			else if (!_stricmp(nextstr, "CUSTOMPUNCHVARS"))
			{
				ge_inf.customhits = SR_NextNum();
				ge_inf.customdamage = SR_NextNum();

				if (ge_inf.customhits < 12) ge_inf.customhits = 12;
				if (ge_inf.customdamage < 250) ge_inf.customdamage = 250;
			}
// TEXOFFSET	v220-13
			else if (!_stricmp(nextstr, "TEXOFFSET"))
			{
				int	s = SR_NextNum();	// Get the sector number
				int	t = SR_NextNum();	// Get the type (0 floor, 1 ceiling, 2 floor and ceiling)
				int	x = SR_NextNum();
				int	y = SR_NextNum();

				if (s < 0 || s >= numsectors)
					SR_SCPErr("TEXOFFSET sector: %i invalid", s);
				
				if (t == 0 || t == 2)
				{
					sectors[s].flooroffx = x * FRACUNIT;
					sectors[s].flooroffy = y * FRACUNIT;
				}

				if (t == 1 || t == 2)
				{
					sectors[s].ceiloffx = x * FRACUNIT;
					sectors[s].ceiloffy = y * FRACUNIT;
				}
			}
// DAMAGEFACTOR_TO_PLAYER		v220-33
			else if (!_stricmp(nextstr, "DAMAGEFACTOR_TO_PLAYER"))
			{
				if (playerdamageidx >= MAXDAMAGENUM)
					SR_SCPErr("Maximum entries of %i exceeded", MAXDAMAGENUM);

				int num = Def_GetMobjNum( (nextstr = SR_NextSTR(false)) );

				if (num < 0) SR_SCPErr("Unrecognised def Thing ID %s", nextstr);

				for (int n = 0; n < playerdamageidx; n++)
				{
					if (playerdamagetype[n] == num)
						SR_SCPErr("ID %s has already been set with factor %.f", nextstr, playerdamagefactor[n]);
				}

				// Remember ID string
				float factor = SR_NextFloat();

				if (factor <= 0.f)
					SR_SCPErr("invalid factor %.f - must be > 0", factor);

				if (ge_inf.reportdamage)
					Con_Printf("DAMAGEFACTOR SET: %s %.f [THING type %i]\n", Def_IDfromType(num), factor, num);

				playerdamagetype[playerdamageidx] = num;
				playerdamagefactor[playerdamageidx++] = factor;
			}
// PISTOLHEADDAMAGE		v220-33
			else if (!_stricmp(nextstr, "PISTOLHEADDAMAGE"))
			{
				int num = Def_GetMobjNum( (nextstr = SR_NextSTR(false)) );

				if (num < 0) SR_SCPErr("Unrecognised def Thing ID %s", nextstr);

				for (int n = 0; n < enemyheadshotidx; n++)
				{
					if (enemyheadshottype[n][0] == num)
						SR_SCPErr("ID %s has already been set", nextstr);
				}

				int min = SR_NextNum();
				int max = SR_NextNum();

				if (ge_inf.reportdamage)
					Con_Printf("PISTOLHEADDAMAGE SET: %s min %i max %i\n", Def_IDfromType(num), min, max);

				enemyheadshottype[enemyheadshotidx][0] = num;
				enemyheadshottype[enemyheadshotidx][1] = min * FRACUNIT;
				enemyheadshottype[enemyheadshotidx++][2] = max * FRACUNIT;
			}
// JUMPKILL		v220-33
			else if (!_stricmp(nextstr, "JUMPKILL"))
			{
				int num = Def_GetMobjNum( (nextstr = SR_NextSTR(false)) );

				if (num < 0) SR_SCPErr("Unrecognised def Thing ID %s", nextstr);

				if (ge_inf.reportdamage)
					Con_Printf("JUMPKILL SET: %s\n", Def_IDfromType(num));

				cfg.jumpkill = num;
			}
// JUMPPOWER	v220-13
			else if (!_stricmp(nextstr, "JUMPPOWER"))
			{
				cfg.jumpPower = SR_NextFloat();

				if (cfg.jumpPower <  0.f) cfg.jumpPower = 0.f;
				if (cfg.jumpPower > 10.f) cfg.jumpPower = 10.f;
			}
// NOINFIGHT	v220-13/ v220-34 added NOINFIGHTALL
			else if (!_stricmp(nextstr, "NOINFIGHT") || !_stricmp(nextstr, "NOINFIGHTALL"))
			{
				if (!_stricmp(nextstr, "NOINFIGHTALL")) ma_inf.NOINFIGHTALL = true;

				if (strcmp(SR_NextSTR(true), "{"))	// opening {
					SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					int type = Def_SetScriptedNoInfight(nextstr);

					if (type > 0)
					{
						thinker_t *th;

						for (th = thinkercap.next; th != &thinkercap; th = th->next)
						{
							if ((void (__cdecl *)(void))th->function == (void (__cdecl *)(void))P_MobjThinker
								&& ((mobj_t *) th)->type == type)
							{
								if (ma_inf.NOINFIGHTALL)
									((mobj_t *) th)->ddflags |= DDF_SCPINFIGHT;
								else
									((mobj_t *) th)->ddflags |= DDMF_NOINFIGHT;
							}
						}
					}
					else
						SR_SCPErr("Unrecognised ID %s", nextstr);
				}
			}
// MAKE_AGGRESSIVE	v220-13
			else if (!_stricmp(nextstr, "MAKE_AGGRESSIVE"))
			{
				int		n = SR_NextNum();

				if (n < 6000 || n > 6099)
					SR_SCPErr("MAKE_AGGRESSIVE %i must be in range 6000 to 6099", n);

				// Method resorted to to allow backward compatibility
				// No other way of setting the flag comes to mind
				aggression_array[n - 6000] = true;

				// Set any that may already be spawned
				{
					thinker_t *th;

					for (th = thinkercap.next; th != &thinkercap; th = th->next)
					{
						if ((void (__cdecl *)(void))th->function == (void (__cdecl *)(void))P_MobjThinker
							&& ((mobj_t *) th)->info->doomednum == n)
						{
							((mobj_t *) th)->ddflags |= DDF_AGGRESSIVE;
						}
					}
				}
			}
// PLAYERVIEWHEIGHT	v220-22
			else if (!_stricmp(nextstr, "PLAYERVIEWHEIGHT"))
			{
				extern int defaultNearClip;
				int		n = SR_NextNum();
				int		h = Def_GetThingHeight("PLAYER");

				if (n < 40 || n >= h)
					SR_SCPErr("PLAYERVIEWHEIGHT %i must be in range 40 to %i", n, h - 1);

				if (h - n < 5) defaultNearClip = h - n;

				plrviewheight = n * FRACUNIT;
				plrfiringheight = plrviewheight - (9 * FixedDiv(plrviewheight, PLR_VIEWHT));
				ppt.viewplayer->viewheight = plrviewheight;
			}
// USEFLOORSECTORLIGHT v220-31
			else if (!_stricmp(nextstr, "USEFLOORSECTORLIGHT"))
			{
				ma_inf.useSectorLight = false;
			}
// SETEVENLIGHTMULTIPLIER
			else if (!_stricmp(nextstr, "SETEVENLIGHTMULTIPLIER"))
			{
				evenLightFactor = SR_NextFloat();

				if (evenLightFactor < 0.0f || evenLightFactor > 1.0f)
					SR_SCPErr("SETEVENLIGHTMULTIPLIER Multiplier %i not in range 0.0 to 1.0", evenLightFactor);
			}
// ADDLOCKEDSOUND
			else if (!_stricmp(nextstr, "ADDLOCKEDSOUND"))
			{
				nextstr = SR_NextSTR(true);
				lockedsound = R3D_AddNewSound(nextstr);
			}
// END
			else
			{
				if (!_stricmp(nextstr, "END") || M_ValidateID(nextstr))
					break;

				SR_SCPErr("Unknown keyword \"%s\"", nextstr);
			}
		}
	}

	SR_EndScript();

	W_Free(secreverbset, STATICID, 0);

	AS_CheckLinkedSeeSectors();

	ResetProgHeader();

	if (debug_scp)
	{
		int	rnum[4] = {7900, 6900, 5900, 4900 };
		int limit = NUMSPOTS / NUMRANGES;
		int r, c, n = 79;

		for (r = 0; r < NUMRANGES; r++)
		{
			for (i = r * limit, c = 0; i < (r + 1) * limit; i++)
			{
				if (!(th7900[i].setflags & T_SET)) c++;
			}

			pr_OFinsert("\n\n==== %i Free SCRIPT %i spots;", rnum[r], c);

			for (i = r * limit, c = 0; i < (r + 1) * limit; i++, c++)
			{
				if (!(c % 10)) pr_OFinsert("\n-- %i%.2i to %i%.2i:", n, c, n, c + 9);

				if ( !(th7900[i].setflags & T_SET) ) pr_OFinsert(" %.2i", c);
				else pr_OFdirect(" ..");
			}

			n -= 10;
		}

		pr_OFdirect("\n\n=======================\n");

		for (i = c = 0; i < NUMSPOTS; i++)
		{
			if (!(th7800[i].setflags & T_SET) ) c++;
		}

		pr_OFinsert("\n==== %i Free SPAWN spots (RS means Remain Static);\n"
					"     Of the three skill flags, if only NORMAL is set,\n"
					"     then anything spawned will neither fall nor rotate.\n"
					"     RS, if shown, indicates this.", c);

		for (i = 0; i < NUMSPOTS; i++)
		{
			if (!(i % 100)) pr_OFinsert("\n");

			if (!(i % 10))
			{
				if (i < 100)
					pr_OFinsert("\n-- 78%.2i to 78%.2i:", i, i + 9);
				else if (i < 200)
					pr_OFinsert("\n-- 68%.2i to 68%.2i:", i - 100, i - 100 + 9);
				else if (i < 300)
					pr_OFinsert("\n-- 58%.2i to 58%.2i:", i - 200, i - 200 + 9);
				else
					pr_OFinsert("\n-- 48%.2i to 48%.2i:", i - 300, i - 300 + 9);
			}

			if ( !(th7800[i].setflags & T_SET) )
				pr_OFinsert(" %.2i", i > 299 ? i - 300 : i > 199 ? i - 200 : i > 99 ? i - 100 : i);
			else if ( (th7800[i].options & (MTF_EASY|MTF_NORMAL|MTF_HARD)) == MTF_NORMAL)	// v220-10
				pr_OFdirect(" RS");
			else
				pr_OFdirect(" ..");
		}

		pr_OFdirect("\n\n=======================\n");

		for (i = c = 0; i < NUM7400SPOTS; i++)
		{
			if (!th7400[i].refline) c++;
		}

		pr_OFinsert("\n==== %i Free 7400 spots;\n", c);

		for (i = 0; i < NUM7400SPOTS; i++)
		{
			if (!(i % 10))
				pr_OFinsert("\n-- 74%.2i to 74%.2i:", i, i + 9);

			if (!th7400[i].refline)
				pr_OFinsert(" %.2i", i);
			else
				pr_OFdirect(" ..");
		}

		pr_OFdirect("\n\n=======================\n");

		for (i = c = 0; i < 100; i++)
		{
			if (fakeidentifier[i] < 0) c++;
		}

		pr_OFinsert("\n==== %i Free THINGS type 61xx;\n", c);

		for (i = 0; i < 100; i++)
		{
			if (!(i % 10))
				pr_OFinsert("\n-- 61%.2i to 61%.2i:", i, i + 9);

			if ( fakeidentifier[i] < 0)
				pr_OFinsert(" %.2i", i);
			else
				pr_OFdirect(" ..");
		}

		pr_OFinsert("\n\n=======================\n");
	}
}

int FetchDropItem(int thingidx)
{
	int dridx = thingidx - 6000;

	if (   R3D_ScriptedMap
		&& dropitems
		&& dridx >= 0
		&& dridx < MAXDROPNUM)
	{
		return dropitems[dridx].item[rand() % MAXDROPITEM];
	}
	return -1;
}

RDBOOL AS_CheckSCP_DEFN(char *mapid)
{
	if (W_OpenLatestScriptLump("SCP_DEFN", mapid))
	{
		SR_EndScript();
		return true;
	}

	return false;
}

// v220-4 Check if a spawn list exceeds the number of friends
//        carried over from a previous map.
void AS_SetFCO_Limit()
{
	if (FCOLIM_idx78xx >= 0)
	{
		th7800_t *t78 = &th7800[FCOLIM_idx78xx];
		int i, n = ma_inf.friendscarriedover;

		if (ge_inf.devmap) 	pr_OFinsert("friends carried over %i\n", n);

		if (t78->spawn_count >= n)
		{
			t78->spawn_count = n;

			for (i = 0; i < n; i++)
			{
				t78->spawn[i][0] = G_GetFriendType(i);
				t78->spawn[i][4] = true;
			}
		}
	}
}

int get79N(int num)
{
	if (num < 100) return 7900 + num;
	if (num < 200) return 6900 + num - 100;
	if (num < 300) return 5900 + num - 200;
	return 4900 + num - 200;
}

int get78N(int num)
{
	if (num < 100) return 7800 + num;
	if (num < 200) return 6800 + num - 100;
	if (num < 300) return 5800 + num - 200;
	return 4800 + num - 300;
}

int get77N(int num)
{
	if (num < 100) return 7700 + num;
	if (num < 200) return 6700 + num - 100;
	if (num < 300) return 5700 + num - 200;
	return 4700 + num - 300;
}

int AS_GetPushSound(int dmednum)
{
	int i;

	for (i = 0; i < numpushsounds; i++)
	{
		if (pushsounds[i].dmednum == dmednum)
			return pushsounds[i].sndid;
	}

	return 0;
}
