//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_chained.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//
// DESCRIPTION: analyses unclosed lines forming a chain
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_MAINspec.h"
#include "m_misc.h"

// CODE --------------------------------------------------------------------

// subfn_FindChainedLines finds a chain of samesided midtex lines
// and establishes the owner sector;

static int IsChain(sector_t *sec, int lnum)
{
	vertex_t	*v = lines[lnum].v2;
	int			i, lineptr = 0, repeat = true, valid = false;

	hitline[lnum] &= ~4;
	linearray[lineptr++] = lnum;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < sec->linecount; i++)
		{
			int lnum = sec->lines[i] - lines;
			int hl = hitline[lnum];

			if (hl & (2|4)
				&& lines[lnum].v1 == v)
			{
				hitline[lnum] &= ~(2|4);
				linearray[lineptr++] = lnum;
				if (hl & 4)
				{
					valid = true;
					break;
				}
				v = lines[lnum].v2;
				repeat = true;
			}
		}
	}
	return valid ? lineptr : 0;
}

static int getSec(int lnum, int type)
{
	ssline_t *sl = &sslines[lnum];
	sector_t *sec = lines[lnum].frontsector;
	int		 i, k;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef
				&& sg->frontsector != sec)
			{
				if (sg->backsector != sg->frontsector)
				{
					return sg->frontsector - sectors;
				}
				else if (type)
				{
					return getSec(sg->linedef - lines, false);
				}
			}
		}
	}
	return -1;
}

void LS_SetChainedLines()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k;

	SET_countsec(0);
	CLR_hitline();

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->frontsector == ld->backsector)
		{
			hitline[i] = 1;

			if (origTEX[ld - lines] & 0x22
				&& SecCtrl(ld->frontsector - sectors))
			{
				int v1 = vtxarray[ld->v1 - vertexes];
				int v2 = vtxarray[ld->v2 - vertexes];

				if (v1 == 2 && v2 == 2)
					hitline[i] |= 2;
				else if ((v1 == 1 && v2 == 2)
					  || (v1 == 2 && v2 == 1))
				{
					countsec[ld->frontsector - sectors]++;
					hitline[i] |= 4;
				}
			}
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (countsec[i] > 1)
		{
			int lineptr;

			for (k = 0; k < sec->linecount; k++)
			{
				int lnum = sec->lines[k] - lines;
				if (hitline[lnum] & 4
					&& vtxarray[lines[lnum].v1 - vertexes] == 1
					&& (lineptr = IsChain(sec, lnum)) > 1)
				{
					int m, secnum = getSec(linearray[0], true);

					if (secnum >= 0)
					{
						for (m = 0; m < lineptr; m++)
						{
							Setspec(&lines[linearray[m]], secnum, 27001);
							Settype(&lines[linearray[m]], R3D_3D, 27002);
							SetLFGSecSet(&lines[linearray[m]], 27003);
						}
					}
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////
//
// LS_CheckLinesR3D_01 - used for sector rendering inheritance.
//
// b220-0: Made partially redundant - complexity reduced.
//
// b211-0 Pass 2 removed.

static void chkVtxLines(line_t *li)
{
	if (   vtxarray[li->v1 - vertexes] > 2
		&& vertexowners[li->v1 - vertexes].num == 1)
	{
		sector_t *s = li->frontsector;
		int		 i;

		for (i = 0; i < s->linecount; i++)
		{
			line_t *ld = s->lines[i];

			if (LSS(ld)
				&& lnsec[ld - lines] < 0
				&& ld != li
				&& (ld->v1 == li->v1 || ld->v2 == li->v1))
			{
				Setspec(ld, lnsec[li - lines], 27004);
				LS_FindAdjacentSecnum(ld, 11);
			}
		}
	}
	if (   vtxarray[li->v2 - vertexes] > 2
		&& vertexowners[li->v2 - vertexes].num == 1)
	{
		sector_t *s = li->frontsector;
		int		 i;

		for (i = 0; i < s->linecount; i++)
		{
			line_t *ld = s->lines[i];

			if (   LSS(ld)
				&& lnsec[ld - lines] < 0
				&& ld != li
				&& (ld->v1 == li->v2 || ld->v2 == li->v2))
			{
				Setspec(ld, lnsec[li - lines], 27005);
				LS_FindAdjacentSecnum(ld, 11);
			}
		}
	}
}

void LS_CheckLinesR3D_01(void)	// Actually textureless lines
{
	sector_t	*sec, *fs, *bs;
	side_t		*sid1, *sid2;
	line_t		*ld;
	int			i, k, lnum, txn, specSec;

	////////////////////////////////////////////////////////////////////////
	//
	// Use LS_GetAdjacentLineSecnum to see if other lines in the same sub-
	// sector carry sectors that differ from the test line's frontsector
	// and backsector.  If so the sector pointer is put into specSec.
	//

	CLR_hitline();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secdone[i]
		|| sec->heightsec >= 0
		|| sec->stptr >= 0)
		{
			continue;
		}

		specSec = -1;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			lnum = ld - lines;

			if (   !(ld->ml_flags & ML_TWOSIDED)
				|| ld->spflags & SPF_CONTROLLINE
				|| lineflags[lnum] & (LFG_LINESSECSET|LFG_NEWSET|LFG_3DTYPE)	// b211-0
				|| (lxarray[lnum] && !(sec->specFlags & SF_MOVFLOOR))
				// Eliminate heightless sectors
				|| (fs  = ld->frontsector)->floorheight == fs->ceilingheight
				|| (bs  = ld->backsector )->floorheight == bs->ceilingheight)
			{
				continue;
			}

			txn = origTEX[lnum];

			// v220-4 Eliminate textureless lines with no floor/ceiling height differences
			//        Not doing this can cause skyfix problems as setting ss->ceilsec
			//        causes the renderer to ignore skyfix values
			if (   !txn
				&& fs->floorheight == bs->floorheight
				&& (fs->ceilingpic == skyflatnum && fs->ceilingpic == bs->ceilingpic))
			{
				continue;
			}

			hitline[lnum] = 1;

			if (   fs == bs
				&& txn & 0x11
				&& IsLiquidFloorFlat(fs))
				continue;

			if (SecONE(i) && SecMIDC(i))
				continue;

			sid1 = &sides[ld->sidenum[0]];
			sid2 = &sides[ld->sidenum[1]];

			if (sid1->texFlags == (T_LOW|T_HI) &&
				sid2->texFlags == (T_LOW|T_HI))
				continue;

			// b211-0 R3D_37 can be one of two different types
			//    but they are differentiated in the linetable
			if (ld->specType == R3D_37 && linetable[lnum] & 2)
			{
				int lo, hi;

				XS_GetBoundsFromLine(ld, &lo, &hi, NULL);

				if (lo >= 0 && lo == hi) Setspec(ld, lo, 27005);

				continue;
			}

			if (fs != sec && SecONE(fs - sectors) && SecMIDC(fs - sectors))
				continue;

			if (   ld->specType == R3D_01
				|| ld->specType == R3D_03
				|| ld->specType == R3D_05
				||
				(  fs == bs
				&& !(txn & 0x22)
				&& ld->specType != R3D_04
				&& sec->specFlags & SF_SECCLOSED
				&& !(sec->specFlags & SF_MOVSECTOR)) )
			{
				sector_t *sect;
				int		 secnum;

				sampleline = NULL;

				if ((secnum = LS_GetAdjacentLineSecnum(ld, 1)) < 0)
					continue;

				sect = &sectors[secnum];

				if (   ld->specType == R3D_01
					&&  sampleline
					&&  sampleline->ml_flags & ML_TWOSIDED
					&& !sampleline->mtflags)
				{
					continue;
				}

				if (ld->specType == R3D_03)
				{
					if (sect->floorheight < sec->floorheight ||
						 ld->frontsector->specFlags & SF_MOVFLOOR)
					{
						Setspec(ld, secnum, 27010);
					}
				}

				if (!Vtxs1(ld))
				{
					RDBOOL diffch = sect->ceilingheight < ld->frontsector->ceilingheight;
					RDBOOL novtx3 = vtxarray[ld->v1 - vertexes] < 3 &&
									vtxarray[ld->v2 - vertexes] < 3;

					sector_t *sect = &sectors[secnum];

					if (novtx3 || txn || !diffch)
					{
						if (txn == 0x55)	// Batman mainly
							Settype(ld, R3D_3D, 27012);
					}
					// rev34
					else if (!novtx3
						&& ld->frontsector->specFlags & SF_NOHITEX
						&& !txn
						&& diffch
						&& sect->ceilingpic == ld->frontsector->ceilingpic)
					{
						ssline_t *sl = &sslines[lnum];
						int		 m;

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->sector != sect && ss->ceilsec < 0)
								Setr3dsscs(ss, secnum, 27015);
						}

						Setr3dch(ld->frontsector, sect, 27015);
					}
				}
			}
		}
	}
}
