//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_closed.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  void LS_CheckClosedMissingHiLinked() removed
//  static void subfn_CheckVtxOwners() removed
//
// DESCRIPTION: analyses closed sectors
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

//-----------------------------------------------------------------------------

// Look for lines defining height of hanging mobj
void LS_CheckOpenUntexturedContained()
{
	sector_t *sec;
	int i, specSec;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (MSFC(i, 0x0000000C) &&
			sec->linecount > 4 &&
			SecCtrl(i) &&		// NOTE: has control lines..
			!(LS_ChkLineLoops(sec, true, false, true)) && // Has no loops
			!IsLiquidFloorFlat(sec))
		{
			int k, valid = true, count = 0;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (ld->spflags & SPF_CONTROLLINE)
					continue;
				if (ld->frontsector != ld->backsector)
				{
					valid = false;
					break;
				}
				count++;
			}
			if (!valid || !count) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;
					
				if ((specSec = LS_FindAdjacentSecnum(ld, 3)) >= 0)
					Setspec(ld, specSec, 1002);

				if (subfn_FindLineCrossing(ld, LC_NOVTXONLINE|LC_DIFFSECLINE) != NULL)
					Settype(ld, R3D_0C, 1002);
			}
		}
	}
}

static void subfn_HiClosedLoops(sector_t *sec)
{
	sector_t *s = NULL, *schk;
	line_t *ld;
	int i, specSec = -1, count = 0, mtflags = 0;

	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE)
			continue;

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			origTEX[ld - lines] ||
			ld->frontsector == ld->backsector)
			return;

		count++;

		if (!mtflags)
		{
			mtflags = ld->mtflags;
			if (mtflags & R3D_MISLOTEX ||
				!(mtflags & R3D_MISHITEX))
				return;
		}
		else if (mtflags != ld->mtflags)
			return;

		schk = mtflags & R3D_MISHTEX1 ? ld->backsector : ld->frontsector;

		if (schk->ceilingpic == skyflatnum ||
			schk->specFlags & SF_MOVSECTOR)
			return;

		if (!s) s = schk;
		else if (s != schk) return;

		vtxcheck[ld->v1 - vertexes]++;
		vtxcheck[ld->v2 - vertexes]++;
	}

	if (count < 3 || !s) return;

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE)
			continue;

		if (vtxcheck[ld->v1 - vertexes] != 2 ||
			vtxcheck[ld->v2 - vertexes] != 2)
			return;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (!(ld->spflags & SPF_CONTROLLINE))
		{
			Clr_allmtflags(ld, 1001);
			Settype(ld, R3D_3D, 1003);
			Setspec(ld, s - sectors, 1003);
		}
	}
	Setsecdone2(sec, 1002);
}

// Find miss hi loops identified by having
// a control sec and sky sec.
void LS_CheckMissingHiClosedLoops()
{
	sector_t *sec = sectors;
	int i;
	for ( i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			SecCtrl(i) &&
			sec->linecount > 5 &&	//MIN 3 CLOSED LINES - 3 CONTROL LINES
			!(sec->specFlags & SF_MOVSECTOR) &&
			sec->ceilingpic == skyflatnum &&
			sec->R3Dceilsec < 0 &&
			sec->R3Dfloorsec < 0 &&
			sec->R3Dfloorheight == DDMAXINT)
		{
			subfn_HiClosedLoops(sec);
		}
	}
}

static void subfn_LoClosedLoops(sector_t *sec)
{
	sector_t *s = NULL, *schk;
	line_t *ld;
	int i, specSec = -1, count = 0, mtflags = 0;

	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			origTEX[ld - lines] & 0x33 ||
			ChkLFGSecSet(ld) ||
			ld->frontsector == ld->backsector)
			return;

		count++;

		if (!mtflags)
		{
			mtflags = ld->mtflags;
			if (mtflags & R3D_MISHITEX ||
				!(mtflags & R3D_MISLOTEX))
				return;
		}

		schk = mtflags & R3D_MISLTEX1 ? ld->frontsector : ld->backsector;

		if (schk->floorpic == skyflatnum ||
			schk->specFlags & SF_MOVSECTOR)
			return;

		vtxcheck[ld->v1 - vertexes]++;
		vtxcheck[ld->v2 - vertexes]++;

		if (sides[ld->sidenum[0]].toptexture !=
			sides[ld->sidenum[1]].toptexture)
			continue;

		if (!s) s = schk;
		else if (s != schk) return;

	}

	if (count < 3 || !s || s == sec
		|| sec->floorheight > s->floorheight)	// test added v2.1.00 build_58
	{
		return;
	}

	CLR_conarray();

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (vtxcheck[ld->v1 - vertexes] != 2 ||
			vtxcheck[ld->v2 - vertexes] != 2)
		{
			return;
		}
		conarray[ld - lines] = 1;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (sides[ld->sidenum[0]].toptexture ==
			sides[ld->sidenum[1]].toptexture)
		{
			if (!origTEX[ld - lines]
				&& !Vtxs2(ld)
				&& ld->mtflags & R3D_MISHITEX)
			{
				Settype(ld, R3D_16, 1004);
			}
			else
				Settype(ld, R3D_3D, 1004);
			Clr_allmtflags(ld, 1002);
		}
	}
	LS_FillGenArea(sec, s - sectors, FGA_SSFLOORSEC, 1001);
	Setsecdone2(sec, 1003);
}

void LS_CheckMissingLoClosedLoops()
{
	sector_t *sec = sectors;
	int i;
	for ( i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			!SecCtrl(i) &&
			sec->linecount > 2 &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			sec->ceilingpic != skyflatnum &&
			sec->floorpic != skyflatnum &&
			sec->R3Dceilsec < 0 &&
			sec->R3Dfloorsec < 0)
		{
			subfn_LoClosedLoops(sec);
		}
	}
}
