//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_contained.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 34 (v220-34) 12/Nov/2018 gmj
//
// DESCRIPTION: analyses lines contained in closed loops
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "r_data.h"
#include "m_misc.h"
#include "p_maputil.h"

// b211-0
// Check for contained movfloor animsecs where sector has same sector lines
// and where the sector does not have a the more usual moving support sector
// using crossing lines

void LS_CheckMovAnimContained()
{
	sector_t *sec;
	int		 c, i, k, m;
	// mask screen
	int		 mc = SF_NOLOTEX|SF_NOHITEX|SF_SECCLOSED|SF_SSLINETRUE|SF_MOVFLOOR|SF_MOVCEIL|SF_MOVINVALID;
	// mask required
	int		 mr = SF_NOLOTEX|SF_NOHITEX|SF_SECCLOSED|SF_MOVFLOOR|SF_MOVINVALID;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& (sec->specFlags & mc) == mr
			&& IsLiquidFloorFlat(sec))
		{
			RDBOOL	 valid = true;

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t	 *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (   ld->frontsector != ld->backsector
					|| lxarray[ld - lines]
					|| lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
				{
					valid = false; break;
				}

				linearray[c++] = ld - lines;
			}

			if (!valid) continue;

			for (k = 0; k < c; k++)
			{
				ssline_t *sl = &sslines[linearray[k]];

				for (m = 0; m < sl->num; m++)
				{
					sector_t *s = subsectors[sl->list[m]].sector;
					
					if (s != sec && s->floorpic == sec->floorpic
						&& !(s->specFlags & SF_MOVSECTOR))
					{
						Setspec(&lines[linearray[k]], s - sectors, 2100);
					}
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////
//
// LS_FindLinkedAnimSectors
//

static void subfn_IterateSectors(sector_t *sec, int height)
{
	int i;
	line_t *ld;

	if (!countsec[sec - sectors] || height & FRACUNIT)
		return;

	Setr3dfh_byH(sec, height, true, 2001);

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (ld->ml_flags & ML_TWOSIDED &&
			!(ld->spflags & SPF_CONTROLLINE) &&
			IsLiquidFloorFlat(ld->frontsector) &&
			IsLiquidFloorFlat(ld->backsector) &&
			!(ld->backsector != ld->frontsector &&
			 (sides[ld->sidenum[0]].texFlags & T_MID ||
			  sides[ld->sidenum[1]].texFlags & T_MID)))
		{
			if (ld->backsector->R3Dfloorheight == DDMAXINT)
				subfn_IterateSectors(ld->backsector, height);
			if (ld->frontsector->R3Dfloorheight == DDMAXINT)
				subfn_IterateSectors(ld->frontsector, height);
		}
	}
}
// By scanning pre-marked lines looks for a continuous ring

static short checkPlanepic(int lnum)
{
	if (IsLiquidFloorFlat(lines[lnum].frontsector))
		return lines[lnum].frontsector->floorpic;
	if (lines[lnum].ml_flags & ML_TWOSIDED)
		return lines[lnum].backsector->floorpic;

	Con_Error("Program fault [1] line %i", lnum);
	return false;	//happy compiler
}

static int getFloorHeight(int lnum)
{
	if (IsLiquidFloorFlat(lines[lnum].frontsector))
		return lines[lnum].frontsector->floorheight;
	if (lines[lnum].ml_flags & ML_TWOSIDED)
		return lines[lnum].backsector->floorheight;

	return DDMININT;
}

static int CheckIsRing(int lnum)
{
	line_t	 *ld = &lines[lnum];
	vertex_t *v = ld->v1;
	int		 i, isclosed = false, repeat = true;
	short	 planepic = checkPlanepic(lnum);

	hitline[lnum] = 2;

	while (repeat)
	{
		repeat = false;
		for (i = 0; i < numlines; i++)
		{
			if (hitline[i] == 1)
			{
				if (checkPlanepic(i) == planepic)
				{
					if (lines[i].v1 == v)
					{
						v = lines[i].v2;
						hitline[i] = 2;
						repeat = true;
					}
					else if (lines[i].v2 == v)
					{
						v = lines[i].v1;
						hitline[i] = 2;
						repeat = true;
					}
				}
			}
			else if (i == lnum && v == ld->v2)
			{
				isclosed = true;
				break;
			}
		}
	}
	if (!isclosed)
	{
		for (i = 0; i < numlines; i++)
			if (hitline[i] == 2)
				hitline[i] = 1;
	}
	return isclosed;
}

static void setChainedSectors(int floorheight)
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (countsec[i] && sectors[i].floorheight != floorheight)
		{
			Setr3dfh_byH(&sectors[i], floorheight, false, 2002);
		}
	}
}

// Transfer highest floorheight to linked animflat sectors.
// Note that care must be taken not to get false settings.
void LS_FindLinkedAnimSectors()
{
	line_t	 *ld;
	int		 i, k, count = 0;

	CLR_hitline();
	CLR_vtxcheck();

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->spflags & SPF_CONTROLLINE ||
			Vtxs2(ld))	// No chained lines allowed
			continue;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			if (ld->frontsector != ld->backsector &&
				IsLiquidFloorFlat(ld->frontsector) ^
				IsLiquidFloorFlat(ld->backsector) &&
				!(ld->mtflags & R3D_MISLOTEX) &&
				origTEX[i] & 0x11 &&
				ld->frontsector->heightsec < 0 &&
				ld->backsector->heightsec < 0 &&
				vtxarray[ld->v1 - vertexes] > 1 &&
				vtxarray[ld->v2 - vertexes] > 1)
			{
				if (ld->frontsector->floorheight <
					ld->backsector->floorheight &&
					IsLiquidFloorFlat(ld->frontsector))
				{
					hitline[i] = 1;
				}
				else if (ld->frontsector->floorheight >
					ld->backsector->floorheight &&
					IsLiquidFloorFlat(ld->backsector))
				{
					hitline[i] = 1;
				}
			}
			else if (origTEX[i] & 0x11 &&
				(IsLiquidFloorFlat(ld->frontsector) ||
				 IsLiquidFloorFlat(ld->backsector)))
			{
				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
			}
		}
		else
		{
			if (IsLiquidFloorFlat(ld->frontsector) &&
				ld->frontsector->heightsec < 0)
				hitline[i] = 1;
		}
		if (hitline[i])
		{
			vtxcheck[ld->v1 - vertexes]++;
			vtxcheck[ld->v2 - vertexes]++;
			count++;
		}
	}
	if (!count) return;

	for (i = 0; i < numlines; i++)
	{
		if (hitline[i] == 1 && CheckIsRing(i))
		{
			int		chkht = DDMININT;
			short	planepic = checkPlanepic(i);
			RDBOOL	changeht = false;
			RDBOOL	vtxvalid = true;

			SET_countsec(0);

			for (k = 0; k < numlines; k++)
			{
				if (hitline[k] == 2)
				{
					int fht = getFloorHeight(k);

					if (lines[k].ml_flags & ML_TWOSIDED)
					{
						if (lines[k].frontsector->floorpic == planepic)
							countsec[lines[k].frontsector - sectors]++;
						else
							countsec[lines[k].backsector - sectors]++;
					}

					if (vtxcheck[lines[k].v1 - vertexes] != 2 ||
						vtxcheck[lines[k].v2 - vertexes] != 2)
					{
						vtxvalid = false;
						break;
					}

					if (chkht == DDMININT)
						chkht = fht;
					else if (fht != chkht)
					{
						changeht = true;
						if (fht > chkht)
							chkht = fht;
					}
				}
			}

			if (changeht && vtxvalid)
			{
				RDBOOL valid = true;
				// Check that a marked line occurred at least
				// twice in the same sector
				for (k = 0; k < numlines; k++)
				{
					if (hitline[k] == 2 &&
						!(lines[k].ml_flags & ML_TWOSIDED) &&
						!countsec[lines[k].frontsector - sectors])
					{
						valid = false;
						break;
					}
				}

				if (valid)
				{
					setChainedSectors(chkht);
				}
			}

			for (k = 0; k < numlines; k++)
				if (hitline[k] == 2) hitline[k] = 0;
		}
	}
}

//===========================================================================

static sector_t *nextHighestSector(sector_t *sc)
{
	sector_t		*os;
	vertexowner_t	*vo;
	line_t			*ld;
	int				i, k;

	for (i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->frontsector == ld->backsector)
		{
			vo = &vertexowners[ld->v1 - vertexes];
			if (vo->num == 1) vo = &vertexowners[ld->v2 - vertexes];
			if (vo->num == 1) continue;

			for (k = 0; k < vo->num; k++)
			{
				sector_t *s = &sectors[vo->list[k]];

				if (   s != sc
					&& (!(s->specFlags & SF_NOLOTEX) || (s->specFlags & SF_SSLINETRUE))
					&& s->floorpic == sc->floorpic
					&& s->floorheight > sc->floorheight)
				{
					if (!os || (s != os && s->floorheight < os->floorheight))
					{
						os = s;
					}
				}
			}
		}
	}

	return os;
}

// This catches AnimSecs that haven't been adjusted due to
// particular line conditions
// v220-10 rewritten
void LS_CheckLinkedAnimSecs()
{
	sector_t	*sec;
	int			i, k;

	// Set flag bit 1 for any animsec with line(s) having same sec each side
	// and line(s) with different sectors each side
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& IsLiquidFloorFlat(sec)
			&& (sec->specFlags & SF_NOLOTEX)
			&& !(sec->specFlags & (SF_MOVSECTOR|SF_SSLINETRUE|SF_SECCLOSED))
			&& sec->heightsec < 0)
		{
			sector_t *os;

			// The flag for the sector SF_SECCLOSED will be unset if it has same sector
			// sided lines. The lines must, however, form a loop so check if this is true.
			for (k = 0; k < sec->linecount; k++) linearray[k] = sec->lines[k] - lines;

			if (   nf_chkLineListRing(k, NULL, 0, NULL)
				&& (os = nextHighestSector(sec))
				&& !(os->specFlags & SF_MOVFLOOR))
			{
				Setr3dfh_byH(sec, os->floorheight, false, 2010);
				Setsecdone(sec, 2010);
			}
		}
	}
}

void LS_CheckMovCeilInAnim()	// Rev33
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			IsLiquidFloorFlat(sec) &&
			sec->specFlags & SF_MOVCEIL &&
			sec->specFlags & SF_LINESCLOSED &&
			sec->specFlags & SF_NOLOTEX &&
			!(sec->specFlags & (SF_MOVFLOOR|SF_MOVLOWER)))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->backsector &&
					ld->backsector != ld->frontsector)
				{
					sector_t *s = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;
					if (s->floorpic == sec->floorpic)
					{
						if (s->R3Dfloorheight < DDMAXINT)
						{
							Setr3dfh_byH(sec, s->R3Dfloorheight, false, 2012);
							Setsecdone2(sec, 2013);
							break;
						}
						else if (sec->floorheight < s->floorheight)
						{
							Setr3dfh(sec, s, 2014);
							Setsecdone2(sec, 2015);
							break;
						}
					}
				}
			}
		}
	}
}

//======================================================================
// LS_ChkContained808Csectors
//======================================================================

static sector_t *check808Ccontained(line_t *ld)
{
	subsector_t *ss;
	sector_t	*sec = NULL;
	int			i, k;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (countsec[ss->sector - sectors]
			&& ss->sector == ld->frontsector)
		{
			line_t	*lc;
			seg_t	*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ((lc = sg->linedef) != NULL &&
					lc->ml_flags & ML_TWOSIDED)
				{
					if (lc->frontsector != lc->backsector)
					{
						return NULL;
					}
					if (lc->frontsector != ld->frontsector)
					{
						if (LSS(lc)
							&& countsec[lc->frontsector - sectors])
						{
							sec = lc->frontsector;
						}
						else 
						{
							return NULL;
						}
					}
				}
			}
		}
	}
	return sec;
}

static void Set808COwner(subsector_t *ss, line_t *ld, line_t *lx)
{
	seg_t	*seg = &segs[ss->firstline];
	int		i, secn, found = true;
	int		grpnum = ss->viewgroup;

	for (i = 0; i < ss->linecount; i++, seg++)
	{
		if (   seg->linedef
			&& lnsec[seg->linedef - lines] >= 0)
		{
			found = false;
			break;
		}
	}

	if (!found) return;	// this subsec will be set in subfn_PRESET()

	seg = &segs[ss->firstline];
	for (i = 0; i < ss->linecount; i++, seg++)
	{
		if (seg->linedef &&
			seg->frontsector != ld->frontsector &&
			ss->sector != seg->frontsector)
		{
			SetOverride(ss, seg->frontsector - sectors, 2023);
			return;
		}
	}

	// If got here then ssec will only have segs with ld
	if ((secn = subfn_SsecFacingLine(ss, lx)) >= 0)
		SetOverride(ss, secn, 2024);
}

static void subfn_set808Clinessecs(line_t *ld, line_t *lx)
{
	ssline_t *sl = &sslines[ld - lines];
	int i;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (   ss->floorsec < 0
			&& ss->ceilsec < 0
			&& ss->sector == ld->frontsector)
		{
			Set808COwner(ss, ld, lx);
		}
	}
}

static RDBOOL subfn_Check808CCrossing(int secnum)
{
	sector_t *sec = &sectors[secnum];
	line_t	 *lc  = NULL, *lx;
	int		 i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (LSS(ld)
			&& origTEX[ld - lines] & 0x22
			&& Vtxs1(ld)
			&& lnsec[ld - lines] < 0)
		{
			if (!lc)
				lc = ld;
			else	// Only one line allowed
			{
				lc = NULL;
				break;
			}
		}
	}

	if (lc != NULL
		&& (lx = subfn_FindLineCrossing(lc, LC_NORMAL|LC_NOVTXONLINE|LC_DIFFSECLINE)) != NULL)
	{
		subfn_set808Clinessecs(lc, lx);
		return true;
	}
	return false;
}

void LS_ChkContained808Csectors()
{
	sector_t	*sec;
	line_t		*ld;
	int			i, secnum;
	int			count = 0;

	SET_countsec(0);

	for (i = 0; i < numsectors; i++)
	{
		if (MSFC(i, 0x0000808C)
			&& SecCtrl(i))
		{
			if (!subfn_Check808CCrossing(i))
			{
				countsec[i] = 1;
				count++;
			}
		}
	}

	if (count < 2)
	{
		return;
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (LSS(ld)
			&& Vtxs1(ld)
			&& countsec[ld->frontsector - sectors]
			&& origTEX[i] & 0x22
			&& lnsec[ld - lines] < 0
			&& subfn_FindLineCrossing(ld, LC_NOCONDITION) == NULL
			&& subfn_FindLineOnLine(ld, 0) == NULL
			&& (sec = check808Ccontained(ld)) != NULL
			&& (secnum = LS_GetContainingSecnum(sec, 1)) >= 0)
		{
			Setspec(ld, secnum, 2025);
			Setsecdone3(ld->frontsector, 2026);
		}
	}
}

static RDBOOL movLineInside(int *bbox, int linptr, int *linarray)
{
	int i, box[4];

	for (i = 0; i < linptr; i++)
	{
		memcpy(box, lines[linarray[i] & 0x7FFFFFFF].bbox, sizeof(box));

		if (box[BOXLEFT ] >= bbox[BOXLEFT ] &&
			box[BOXRIGHT] <= bbox[BOXRIGHT] &&
			box[BOXBASE ] >= bbox[BOXBASE ] &&
			box[BOXTOP  ] <= bbox[BOXTOP  ])
		{
			return true;
		}
	}
	return false;
}

// Look for bounded secs b211-0 - revised
void LS_ChkBoundedSector()
{
	sector_t	*sec;
	line_t		*ld;
	int			i, k, m, count;
	int			mask = SF_NOLOTEX|SF_MOVFLOOR;

	int			*linarray = (int*)M_Calloc(numlines * sizeof(int), 270);
	int			linptr = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   LSS(ld)
			&& !lineflags[i]
			&& !origTEX[i]
			&& ld->specType < R3D_40
			&& !LineVertexOwnsSec(ld, i)
			&& ld->frontsector->specFlags & SF_MOVFLOOR
			&& !(subfn_FindLineCrossing(ld, LC_DIFFSECSTATIC)))
		{
			linarray[linptr++] = i;
			linarray[i] |= 0x80000000;
		}
	}

	if (!linptr)
	{
		M_Free(linarray);
		return;	// nothing to do
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->specFlags & SF_SECCLOSED
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& (count = sec->linecount) > 3)
		{
			int ring_count, bbox[4];

			CLR_conarray();

			for (k = 0; k < count; k++)
			{
				line_t   *ld = sec->lines[k];
				sector_t *s;

				if (LSS(ld) || checkMissingTexLine(ld)) { count = 0; break; }

				// Test added $v93
				if (ld->ml_flags & ML_TWOSIDED)
				{
					s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if ((s->specFlags & mask) == mask) { count = 0; break; }
				}

				conarray[ld - lines] = 1;
			}

			if (   count
				&& subfn_CheckForRing(conarray, sec, sec->lines[0] - lines, &ring_count, bbox, NULL, NULL)
				&& ring_count == count
				&& movLineInside(bbox, linptr, linarray)	// Quick test
				&& LS_MarkSubsecsInAreaGeneral(sec, conarray, 0, NULL, bbox, -1, MSIA_NONE))
			{
				int snum = -1;

				count = 0;

				validcount++;

				CLR_ssecarray();

				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t		*sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						line_t *ld = sg->linedef;

						if (ld && ld->validcount != validcount)
						{
							ld->validcount = validcount;

							// Can't use conarray as used by LS_MarkSubsecsInArea()
							// hence setting the high bit of linarray;
							if (linarray[ld - lines] < 0)	// boundary
							{
								int chksnum = ld->frontsector - sectors;
								if (snum < 0)
									snum = chksnum;
								else if (snum >= 0 && snum != chksnum)
								{ snum = -1; goto endcheck; }

							}
							// lines inside must not have a missing tex
							else if (checkMissingTexLine(ld))
								{ snum = -1; goto endcheck; }
						}
					}
				}
endcheck:;
				if (snum >= 0)
				{
					// Check for any lines that need texture added (sfx line R3D_3A)
					// Note: setting the R3D_3A lines is a bit awkward
					// Non-compliant lines have to be scanned more than once
					validcount++;

					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (!sg->linedef) 	continue;

							SetLFG(sg->linedef, LFG_NEWSET, 2044);

							if (   origTEX[sg->linedef - lines] & 0x55
								&& sg->frontsector - sectors == snum
								&& sg->linedef->validcount != validcount
								&& sg->backsector
								&& sg->backsector != sg->frontsector)
							{
								sg->linedef->validcount = validcount;
								Settype(sg->linedef, R3D_3A, 2044);
							}
						}
						if (ss->sector - sectors == snum)
							SetOverride(ss, i, 2045);
					}
				}
			}
		}
	}
	M_Free(linarray);
}

// Rings are determined by their vertex count.  If a line has been added and
// stitched to one of these vertexes then it can cause a failure with later checks.
//
// This determines whether such a vertex exists and adjusts the vtxarray value accordingly.
//
void LS_CheckForRingLinePairs()
{
	line_t	*ld;
	int		c, i, k;

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& ld->lnflags2 & LNF2_OWNEQUALTOONE
			&& !lxarray[i]
			&& !vtoflags[ld->v1 - vertexes]
			&& !vtoflags[ld->v2 - vertexes]
			&& ((vtxarray[ld->v1 - vertexes] > 2
			&&   vtxarray[ld->v2 - vertexes] == 2)
			||  (vtxarray[ld->v1 - vertexes] == 2
			&&   vtxarray[ld->v2 - vertexes] > 2))
			&& subfn_FindLineConnected(ld, 26))
		{
			linearray[c++] = i;
		}
	}

	if (c)	// look for pairs
	{
		vertex_t *v;
		int		 p1, p2;

		for (i = 0; i < c; i++)
		{
			if (linearray[i] < 0) continue;

			p1 = linearray[i];

			if ( vtxarray[(v = lines[p1].v1) - vertexes] == 2)
				v = lines[p1].v2;

			for (k = i + 1; k < c; k++)
			{
				if ( (p2 = linearray[k]) < 0) continue;

				if (v == lines[p2].v1 || v == lines[p2].v2)
				{
					vtxarray[v - vertexes] = 2;
					linearray[i] = -1;
					linearray[k] = -1;
					break;
				}
			}
		}
	}
}

void LS_SetMissLowtexMovSectorsContained()
{
	sector_t *sec;
	int		 i, mask = SF_DUALFMOVSEC|SF_MOVFLOOR|SF_HASTEXTURE|SF_SECCLOSED;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& SecCtrl(i)
			&& IsLiquidFloorFlat(sec)
			&& (sec->specFlags & 0x40F3B01C) == mask)
		{
			int k, valid = false, count = 0, done = false;
			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (!(ld->ml_flags & ML_TWOSIDED)
					|| !Vtxs2(ld)
					|| ld->frontsector == ld->backsector
					|| origTEX[ld - lines] & 0x22)
				{
					count = 0;
					break;
				}
				if (Vtfl3(ld))
				{
					if ( !(origTEX[ld - lines] & 0x11)
						&& origTEX[ld - lines] & 0x44)
					{
						hitline[ld - lines] = 1;
						count++;
					}
					else
					{
						count = 0;
						break;
					}
				}
			}
			if (count < 3) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				int	m, n, bbox[4];

				line_t *ld = sec->lines[k];

				if (hitline[ld - lines] == 1
					&& subfn_CheckForRing(hitline,sec,ld-lines,NULL,bbox,NULL,NULL)
					&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1,MSIA_REVERSED))
				{
					int valid = true;

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];
						seg_t		*sg = &segs[ss->firstline];

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							line_t *lc = sg->linedef;

							if (!lc || hitline[lc - lines]) continue;

							if (lc->frontsector != lc->backsector
								&& origTEX[lc - lines] & 0x33)
							{
								valid = false;
								break;
							}
						}
						if (!valid) break;
					}
					if (!valid)
					{
						done = false;
						break;
					}

					for (m = 0; m < hitssecptr; m++)
						Setr3dssfs(&subsectors[hitssec[m]], i, 2046);

					done = true;
				}
			}
			if (done)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (hitline[ld - lines])
						SetLFGSecSet(ld, 2047);
				}
				Setsecdone3(sec, 2048);
			}
		}
	}
}

//==================================================================================
// LS_MarkLoopsWithContainedLoops - Looks for line loops that don't have
// a liquid floor flat that are contained in an area with a liquid floor
// flat. Currently the containing sector is only derived from an adjacent
// line having no backsector.
static int subsecHasNoBacksecLine(int lnum)
{
	ssline_t *sl = &sslines[lnum];
	sector_t *s = lines[lnum].frontsector;
	int		 i, k;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& !(sg->linedef->ml_flags & ML_TWOSIDED)
				&& sg->linedef->frontsector != s
				&& IsLiquidFloorFlat(sg->linedef->frontsector))
			{
				return sg->linedef->frontsector - sectors;
			}
		}
	}
	return -1;
}

void LS_MarkLoopsWithContainedLoops()
{
	line_t	*ld;
	int		i, k, bbox[4], ring_count, count = 0;
	int		secnum, lpptr = 0;

	SET_countsec(0);
	CLR_hitline();

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->frontsector == ld->backsector
			&& Vtxs2(ld)
			&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
			&& !IsLiquidFloorFlat(ld->frontsector)
			&& !subfn_FindLineCrossing(ld, LC_NORMAL))
		{
			countsec[ld->frontsector - sectors]++;
			hitline[i] = 1;
			count++;
		}
	}

	if (count < 3) return;

	for (i = 0; i < numsectors; i++)
	{
		if (countsec[i] > 2)
		{
			sector_t *sec = &sectors[i];

			for (k = 0; k < sec->linecount; k++)
			{
				int lnum = sec->lines[k] - lines;

				if (hitline[lnum] == 1
					&& (secnum = subsecHasNoBacksecLine(lnum)) >= 0
					&& subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL))
				{
					int		c = 0, m, n;

					for (m = 0; m < sec->linecount; m++)
					{
						n = sec->lines[m] - lines;

						if (hitline[n] == 4)
						{
							c++;
							hitline[n] = 8;
							Setspec(&lines[n], secnum, 2049);
							SetLFGSecSet(&lines[n], 2050);
						}
					}
					if (c != ring_count)
						Con_Error("Program fault [2] sec %i sl %i c %i rc %i",i,lnum,c,ring_count);

				}
			}
		}
	}
}

// =====================================================

// Find if the inner ring is contained in an outer ring with different sectors
// each side of the lines. This should establish that the sector should apply
// to the inner ring.
static RDBOOL ringOutside(sector_t *sc, int *box)
{
	line_t	*ld;
	int		rc, lnum = -1;
	int		c, i, k, m;

	CLR_conarray();
	hitssecptr = 0;
	validcount++;
	ss_validcount++;

	for (c = i = 0; i < sc->linecount; i++)
	{
		int n = sc->lines[i] - lines;

		if (hitline[n] == 4)
		{
			ssline_t *sl = &sslines[n];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount != ss_validcount)
				{
					seg_t	*sg = &segs[ss->firstline];
					RDBOOL	valid = true;

					ss->ssvalidcount = ss_validcount;

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (sg->frontsector && sg->frontsector != sc)
						{
							valid = false; break;
						}
					}

					if (valid)
					{
						sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (   (ld = sg->linedef)
								&& ld->validcount != validcount
								&& (ld->frontsector == sc || ld->backsector == sc)
								&& (ld->frontsector != sc || ld->backsector != sc))
							{
								lnum = ld - lines;
								conarray[lnum] = 1;
								ld->validcount = validcount;
								c++;
							}
						}
					}
					else
					{
						hitssec[hitssecptr++] = sl->list[k];
					}
				}
			}
		}
	}

	if (lnum < 0 || c < 3) return false;

	return (subfn_CheckForRing(conarray, sc, lnum, &rc, box, NULL, NULL) && rc == c);
}

static void subfn_SameSectorContained(sector_t *s)
{
	int		 c, i, k, rc, box[4], box2[4];

	CLR_hitline();

	for (c = i = 0; i < s->linecount; i++)
	{
		int n = s->lines[i] - lines;

		if (   !lineflags[n]
			&& !origTEX[n]
			&& lines[n].frontsector == lines[n].backsector)
		{
			hitline[n] = 1; c++;
		}
	}

	if (c < 3) return;

	while (1)
	{
		int lnum = -1;

		for (i = 0; i < s->linecount; i++)
		{
			int n = s->lines[i] - lines;

			if (hitline[n] == 1) { lnum = n; break; }
		}

		if (lnum < 0) break;

		if (   subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL)
			&& rc > 2
			&& ringOutside(s, box2)
			&& M_CompareBox(box, box2))
		{
			int		sd = 0;	// sidelist count

			// set up the side array
			// The logic used is that for one side of the line the ssec's sector will
			// be the same as the line's sector and the other a different sector. It
			// is the different one that will determine which line side is used. It
			// may seem like overkill but should work for re-entrant boxes.
			for (i = 0; i < hitssecptr; i++)
			{
				subsector_t *ss = &subsectors[hitssec[i]];
				seg_t		*sg = &segs[ss->firstline];

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					line_t *ld = sg->linedef;

					if (ld && ld->validcount != validcount)	// set in ringOutside
					{
						ld->validcount = validcount;

						if (hitline[ld - lines] == 4)
						{
							linearray[sd] = ld - lines;
							sidelist[sd++] = sg->sidedef - sides;
						}
					}
				}
			}

			if (sd == rc && LS_MarkSsecsUsingSidelist(rc, box, MSIA_NONE))
			{
				RDBOOL hi = true, lo = true;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];
					seg_t		*sg = &segs[ss->firstline];

					for (k = 0; k < ss->linecount; k++, sg++)
					{
						if (sg->sidedef)
						{
							if (sg->sidedef->toptexture   ) hi = false;
							if (sg->sidedef->bottomtexture) lo = false;
						}

						if (!hi && !lo) return;
					}
				}

				for (i = 0; i < s->linecount; i++)
				{
					line_t	*ld = s->lines[i];
					int		n = ld - lines;

					if (hitline[n] == 4)
					{
						hitline[n] = 0;
					}
				}

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (hi && ss->ceilsec < 0 && ss->sector->ceilingheight > s->ceilingheight)
						Setr3dsscs(ss, s - sectors, 2051);

					if (lo) Setr3dssfs(ss, s - sectors, 2052);
					
					subfn_setSSEC_LFG(2053, false);
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// v220-10 check for superfluous low textures
static RDBOOL subfn_checkIfInMovsec(sector_t *s)
{
	sector_t	*chk, *osec = NULL;
	line_t		*ld;
	int			sdone = 0;
	int			rc, box[4];
	int			c, i;

	CLR_hitline();

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| lineflags[ld - lines]
			|| origTEX[ld - lines]
			|| ld->frontsector == ld->backsector)
		{
			return false;
		}

		chk = ld->frontsector == s ? ld->backsector : ld->frontsector;

		if (!osec)
		{
			osec = chk;
			if ((chk->specFlags & 0xFFFFFC) != 0x118198) return false;
			if (chk->ceilingheight != s->ceilingheight) return false;
		}
		else if (osec != chk) return false;

		hitline[ld - lines] = 1; c++;
	}

	if (c < 3) return false;

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < s->linecount; i++)
		{
			ld = s->lines[i];

			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			RDBOOL isset = false;

			if (   LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE)
				&& LS_OnlySSLinesInSSECarray(s, hitline))
			{
				sdone = isset = 1;
			}

			for (i = 0; i < s->linecount; i++)
			{
				ld = s->lines[i];
				lnum = ld - lines;

				if (hitline[lnum] == 4)
				{
					hitline[lnum] = 0;

					if (isset)
					{
						Settype(ld, R3D_14, 2057);
						SetLFG(ld, LFG_NEWSET, 2057);
					}
				}
			}
		}
		else
			hitline[lnum] = 0;
	}

	if (sdone)
	{
		Setsecdone(s, 2058);
		return true;
	}

	return false;
}

// This depends on line type set from 46072
static void subfn_SameSectorContainingR3D_16ring(int secidx)
{
	sector_t	*s = &sectors[secidx], *chk, *osec = NULL;
	line_t		*ld;
	int			c, i;

	for (c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (   ld->backsector
			&& ld->specType == R3D_16
			&& !origTEX[ld - lines]
			&& ld->frontsector->floorheight == ld->backsector->floorheight
			&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
		{
			chk = ld->frontsector->ceilingheight > ld->backsector->ceilingheight ?
				ld->frontsector : ld->backsector;

			if (chk == s) return;

			linearray[c++] = ld - lines;
		}
	}

	if (c < 3) return;

	CLR_hitline();

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	while(1)
	{
		sector_t	*osec = NULL;
		int			lnum = -1, rc, box[4];

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1)
			{
				lnum = linearray[i]; break;
			}
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL))
		{
			RDBOOL	inner = LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_TESTAREA);
			RDBOOL	valid = SetAndCheckSSECSInLoop(s, NULL);

			if (valid)
			{
				for (i = 0; i < c; i++)
				{
					if (hitline[linearray[i]] == 4)
					{
						ld = &lines[linearray[i]];
						chk = ld->frontsector->ceilingheight > ld->backsector->ceilingheight ?
							ld->frontsector : ld->backsector;

						if (!osec)
						{
							if ( !(chk->specFlags & SF_SSLINETRUE)
								|| chk->specFlags & SF_MOVCEIL
								|| chk->ceilingpic == skyflatnum)
							{
								return;
							}

							osec = chk;
						}
						else if (osec != chk
							&& (osec->specFlags & SF_MOVCEIL
							|| osec->floorheight != chk->floorheight
							|| osec->ceilingheight != chk->ceilingheight
							|| osec->ceilingpic != chk->ceilingpic))
						{
							valid = false; break;
						}
					}
				}
			}

			if (valid)
			{
				if (!inner) osec = s;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];
					
					if (ss->sector != osec) Setr3dsscs(ss, osec - sectors, 2059);
				}
			}

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					hitline[linearray[i]] = 0;

					if (valid)
						Forcetype(&lines[linearray[i]], R3D_3F, 2060);
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}
}

// v220-10 Look for same sector sided line loops to see
//         if the internal floor or ceil should be rendered
void LS_SameSectorContained()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].stptr >= 0 || SecCtrl(i)) continue;

		if ( (sectors[i].specFlags & 0xFFFFFC) == 0x918198 )
		{
			continue;
		}

		if (  (sectors[i].specFlags & 0x7FFFFC) == 0x1C
			&& subfn_checkIfInMovsec(&sectors[i]))
		{
			continue;
		}

		// Check for blocking sector extending down from ceiling
		if (  (sectors[i].specFlags & 0xFFBEC) == 0xC
			&& sectors[i].ceilingpic != skyflatnum)
		{
			subfn_SameSectorContainingR3D_16ring(i);
		}
		else
		if (  !(sectors[i].specFlags & SF_MOVSECTOR)
			&& (sectors[i].specFlags & 0xC) != 0xC)
		{
			subfn_SameSectorContained(&sectors[i]);
		}
	}
}

// ============================================================

static int addToCountsec(int secn, int c)
{
	int i;

	for (i = 0; i < c; i++)
	{
		if (countsec[i] == secn) return c;
	}

	countsec[c++] = secn;
	return c;
}

static int findInCountsec(int secn, int c)
{
	int i;

	for (i = 0; i < c; i++)
	{
		if (countsec[i] == secn) return i;
	}

	return 0;
}

static int lineCheck(sector_t *sc)
{
	int c, i, k;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& sg->linedef->validcount != validcount)
			{
				sg->linedef->validcount = validcount;
				linearray[c++] = sg->linedef - lines;

				if (hitline[sg->linedef - lines] == 4) continue;

				if (!sg->linedef->backsector)
				{
					if (sg->linedef->frontsector != sc)
						return 0;
				}
				else if (sg->linedef->frontsector != sg->linedef->backsector)
				{
					return 0;
				}
			}
		}
	}

	return c;
}

// Check for same sided lines in a different sector
static RDBOOL subfn_sameSidedLineSectors(sector_t *sc, int lnum)
{
	line_t	*ld;
	int		rc, box[4];
	int		c, i, k;

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->frontsector != ld->backsector && !Vtxs1(ld))
		{
			hitline[ld - lines] = 1; c++;
		}
	}

	if (c < 3) return false;

	if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
		&& rc > 2
		&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
		&& (c = lineCheck(sc)) )
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];

			if (ss->sector != sc)
			{
				// v220-34 add additional test.
				if (   sc->specFlags & SF_MOVSECTOR
					|| ss->sector->specFlags & SF_MOVSECTOR
					|| ss->sector->floorheight != sc->floorheight
					|| ss->sector->floorpic != sc->floorpic
					|| ss->sector->ceilingpic != sc->ceilingpic
					|| ss->sector->ceilingheight != sc->ceilingheight)
				{
					SetOverride(ss, sc - sectors, 2061);
				}
				else
				{
					ss->ssflags &= ~SSF_BASESSECFIX;
					ChangeSSECsec(ss, sc, 2061);
					ss->ssflags |= SSF_BASESSECFIX;
				}
			}
		}

		for (k = 0; k < c; k++)
		{
			ld = &lines[linearray[k]];

			if (!origTEX[ld - lines]) Settype(ld, R3D_3F, 2061);
			else if (!(origTEX[ld - lines] & 0x55)) Settype(ld, R3D_03, 2061);

			SetLFG(ld, LFG_FIXED|LFG_NEWSET|LFG_FIXABS, 2061);
		}

		Setsecdone(sc, 2061);
		return true;
	}

	return false;
}

static int secReferencesSc(sector_t *sc, sector_t *chk)
{
	subsector_t	*ss;
	ssline_t	*sl;
	seg_t		*sg;
	line_t		*ld;
	int			a, i, k, m;

	ss_validcount++;

	for (i = 0; i < chk->linecount; i++)
	{
		ld = chk->lines[i];

		if (!ld->backsector) continue;

		if (   (ld->frontsector == sc && ld->backsector == chk)
			|| (ld->backsector == sc && ld->frontsector == chk) )
		{
			sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;

				ss->ssvalidcount = ss_validcount;
				sg = &segs[ss->firstline];

				for (a = m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg->frontsector == chk) a |= 1;
					if (sg->frontsector == sc && sg->backsector == sc) a |= 2;
					if (a == 3) return ld - lines;
				}
			}
		}
	}

	return -1;
}

static RDBOOL allSameSecLines(sector_t *sc)
{
	int i, c;

	for (c = i = 0; i < sc->linecount; i++)
	{
		if (sc->lines[i]->spflags & SPF_CONTROLLINE) continue;
		if (lineflags[sc->lines[i] - lines]) continue;
		if (sc->lines[i]->frontsector != sc->lines[i]->backsector) return false;
		c++;
	}

	return c > 1;
}

static sector_t *getOSfromSameSecLineSector(sector_t *sc)
{
	subsector_t	*ss;
	sector_t	*os = NULL;
	ssline_t	*sl;
	seg_t		*sg;
	int			i, k, m;

	ss_validcount++;
	hitssecptr = 0;

	for (i = 0; i < sc->linecount; i++)
	{
		if (sc->lines[i]->spflags & SPF_CONTROLLINE) continue;

		sl = &sslines[sc->lines[i] - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;
			sg = &segs[ss->firstline];

			if (ss->sector == sc)
				hitssec[hitssecptr++] = sl->list[k];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (sg->frontsector != sc)
				{
					if (sg->frontsector == sg->backsector) return NULL;
					if (!os) os = sg->frontsector;
					else if (os != sg->frontsector) return NULL;
				}

				if (sg->backsector != sc)
				{
					if (!os) os = sg->backsector;
					else if (os != sg->backsector) return NULL;
				}
			}
		}
	}

	return os;
}

// v220-16
static RDBOOL diffSecSSEC(sector_t *ref, int lnum)
{
	ssline_t *sl = &sslines[lnum];
	int		 i, k;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->frontsector != ref) return true;
		}
	}
	return false;
}

static int overrideSecSSEC(sector_t *ref, int lnum)
{
	subsector_t *ss;
	sector_t	*os;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			ssecset;
	int			i, k;

	ssecset = 0;
	ss_validcount++;
	hitssecptr = 0;

	// v220-18 reject test if any missing tex lines are found
	for (i = 0, sl = &sslines[lnum]; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			ss->ssvalidcount = ss_validcount;
			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef && sg->linedef->mtflags & R3D_MISTEXANY) return 0;
			}
		}
	}

	ss_validcount++;

	for (i = 0, sl = &sslines[lnum]; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			ss->ssvalidcount = ss_validcount;
			hitssec[hitssecptr++] = sl->list[i];
			os = NULL;
			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->frontsector != ref && sg->frontsector != sg->backsector)
				{
					if (!os) os = sg->frontsector;
					else if (os != sg->frontsector) { os = NULL; break; };
				}
			}

			if (os)
			{
				if (ss->sector != os)
					SetOverride(ss, os - sectors, 2063);
				ssecset++;
				Setsecdone(ref, 2063);
				Setsecdone(os, 2063);
			}
		}
	}

	if (ssecset)
	{
		validcount++;

		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];
			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) && ld->validcount != validcount)
				{
					ld->validcount = validcount;

					if (!origTEX[ld - lines]) Settype(ld, R3D_3F, 2063);
					else if (!(origTEX[ld - lines] & 0x55)) Settype(ld, R3D_03, 2063);

					SetLFG(ld, LFG_FIXED|LFG_NEWSET|LFG_FIXABS, 2063);
				}
			}
		}
	}

	return ssecset;
}

static sector_t *checkAllSameSectorLines(sector_t *sc)
{
	subsector_t	*ss;
	sector_t	*os;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k, m;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (ld->frontsector != ld->backsector) return NULL;
		c++;
	}

	if (c < 3) return NULL;

	ss_validcount++;
	hitssecptr = 0;
	os = NULL;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (lineflags[ld - lines]) return NULL;

		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;
			sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if (sg->frontsector && sg->frontsector != sc)
				{
					if (!os) {
						if (    sg->frontsector->specFlags & SF_MOVSECTOR
							|| (sg->frontsector->specFlags & 0x404) == 0x4)
						{
							return NULL;
						}
						os = sg->frontsector;
					}
					else if (os != sg->frontsector) return NULL;
				}

				if (sg->backsector && sg->backsector != sc)
				{
					if (!os) {
						if (    sg->backsector->specFlags & SF_MOVSECTOR
							|| (sg->backsector->specFlags & 0x404) == 0x4)
						{
							return NULL;
						}
						os = sg->backsector;
					}
					else if (os != sg->backsector) return NULL;
				}
			}

			if (ss->sector == sc) hitssec[hitssecptr++] = sl->list[k];
		}
	}

	return os;
}

// v220-16 Check for same sided lines in a different sector
void LS_SameSidedLineSectors()
{
	sector_t	*os, *sc;
	sector_t	*bs, *chk;
	line_t		*ld;
	int			a, b, c;
	int			i, k;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		int		fht = sc->floorheight;
		int		cht = sc->ceilingheight;

		if (   secdone[i]
			|| sc->heightsec >= 0
			|| sc->stptr >= 0)
		{
			continue;
		}
		
		if ((sc->specFlags & 0xFBFC) != 0x810)
		{
// 1)
			if (   (sc->specFlags & 0x3FFFC) == 0x1081C
				&& (os = checkAllSameSectorLines(sc)) )
			{
				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[i];

					if (ld->spflags & SPF_CONTROLLINE) continue;

					if (!origTEX[ld - lines]) Settype(ld, R3D_3F, 2064);
					else if (!(origTEX[ld - lines] & 0x55)) Settype(ld, R3D_03, 2064);

					SetLFG(ld, LFG_FIXED|LFG_NEWSET|LFG_FIXABS, 2064);
				}

				for (k = 0; k < hitssecptr; k++)
				{
					SetOverride(&subsectors[hitssec[k]], os - sectors, 2064);
				}

				Setsecdone(sc, 2064);
			}
			continue;
		}
// 2)
		for (a = b = c = k = 0; k < sc->linecount; k++)
		{
			ld = sc->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;
			if (lineflags[ld - lines]) { a = 0; break; }
			if (lxarray[ld - lines] & 4) { a = 0; break; }

			if (ld->frontsector == ld->backsector)
			{
				if (origTEX[ld - lines] & 0x55) { a = 0; break; }

				if (   ld->lnflags2 & LNF2_OWNEQUALTOONE
					&& vtxarray[ld->v1 - vertexes] > 2
					&& vtxarray[ld->v2 - vertexes] > 2)
				{
					a++;
				}
			}
			else if ( (bs = ld->backsector) )
			{
				chk = ld->frontsector == sc ? bs : ld->frontsector;

				if (   chk->floorheight   == fht
					&& chk->ceilingheight == cht)
				{
					b++;

					c = addToCountsec(chk - sectors, c);
				}
			}
		}

		if (a && c && !findInCountsec(i, c))
		{
			int lnum;

			for (k = 0; k < c; k++)
			{
				sector_t *os = &sectors[countsec[k]];

				if ( (lnum = secReferencesSc(sc, os)) >= 0 && subfn_sameSidedLineSectors(os, lnum))
				{
					// do nothing
				}
			}
		}
	}

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (   secdone[i]
			|| sc->heightsec >= 0
			|| sc->stptr >= 0
			|| (sc->specFlags & 0x3000C) != 0x1000C)
		{
			continue;
		}

		if (allSameSecLines(sc) && (os = getOSfromSameSecLineSector(sc)))
		{
			for (k = 0; k < hitssecptr; k++)
				SetOverride(&subsectors[hitssec[k]], os - sectors, 2062);

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (!origTEX[ld - lines]) Settype(ld, R3D_3F, 2062);
				else if (!(origTEX[ld - lines] & 0x55)) Settype(ld, R3D_03, 2062);

				SetLFG(ld, LFG_FIXED|LFG_NEWSET|LFG_FIXABS, 2062);
			}

			Setsecdone(sc, 2062);
		}
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& ld->frontsector->specFlags & SF_MOVFLOOR
			&& !(origTEX[i] & 0x55)
			&& !lineflags[i]
			&& ((vtxarray[ld->v1 - vertexes] == 1 && vertexowners[ld->v2 - vertexes].num > 1)
			||  (vtxarray[ld->v2 - vertexes] == 1 && vertexowners[ld->v1 - vertexes].num > 1))
			&& diffSecSSEC(ld->frontsector, i))
		{
			overrideSecSSEC(ld->frontsector, i);
		}
	}
}

