//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_edit.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// DESCRIPTION: used in edit mode
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <stdio.h>
#include <direct.h>

#include "gl_font.h"
#include "r3d_OGLfunc.h"
#include "r3d_edit.h"
#include "r3d_compat.h"
#include "dd_share.h"
#include "dd_main.h"
#include "dd_wad.h"
#include "ui_main.h"
#include "m_args.h"
#include "r_data.h"
#include "r_main.h"
#include "doom\mbf_data.h"
#include "con_console.h"
#include "sys_system.h"
#include "p_maputil.h"
#include "rend_particle.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// Used with evt.type
#define EV_LO		0x01
#define EV_MD		0x02
#define EV_HI		0x04
#define EV_MDH		0x08
#define EV_MDL		0x10
#define EV_INFILL	0x20
#define EV_WALLM	0x3F

#define EDITINC		FRACUNIT/48

// TYPES -------------------------------------------------------------------

static struct editv_t
{
	char	editstate;	// Set to 1 if in edit mode
	char	masked;		// Set to 1 if a masked texture
	char	r3d;		// Set to 1 if a 3D line
	char	esc;		// Set true if values to be reset
	char	mb2;		// Set if mouse button 2 is pressed when in editstate
	char	startm;		// For smooth and accelerative mouse adjustment

	int		use_g;		// Stores r_texglow
	int		use_l;		// Stores lc
	int		use_sbar;	// Stores cfg.sbarscale
	int		use_fstp;	// Stores cfg.footsteps

	// crosshair
	int		xhairType;
	int		xhairSize;
	byte	xh_col[4];	// Needs 3, use 4 to align

	int		mousex;		//			"
	int		mousey;		//			"
	int		xadj;		//			"
	int		yadj;		//			"
	int		type;		// Set to EV_ types
	int		snum;		// set to the side number
	int		tnum;		// set to line num or flat's sector num, or -1 if nothing selected
	int		offx;		// tex xoffset
	int		offy;		// tex yoffset
	int		val_low;	// low value in quad of intercepted texture (for debug)
	int		val_hi;		// hi  value in quad of intercepted texture (for debug)
	char	name[20];	// name of texture (8 + term + align)
	char	sbst[20];	// name of any texture that may have been substituted for a missing tex
	char	text[16];	// set to LINE num or FLAT sec
	char	textdet[18];	// HIGH, MID, LOW, INFILL, (FLOOR or CEIL TEX not supported)
} evt;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void delNodesBuild(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void ForceEditDefaultStatus(void);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static int P_ListChangedSides(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int	R3D_edit;
int	cpossubsec;
int	cpossec;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int xhairsaved  = false;
static int changemade  = false;
static int editvarsset = false;

// CODE --------------------------------------------------------------------

int Edit_Init()
{
	if ( (R3D_edit = (ArgExists("-texedit") || ArgExists("-txem"))) && !ge_inf.devmap)
	{
		if (!Sys_CriticalMessage(
			"\n\t\t\tIMPORTANT NOTICE\n\n"
			"CHECK THE WAD IS NOT OPEN IN ANOTHER APPLICATION, AN EDITOR FOR EXAMPLE.\n\n"
			"IF IT IS THEN THE FILE RISKS BEING CORRUPTED AND YOU *MUST* EXIT NOW AND\n"
			"CLOSE ANY AND ALL OTHER SUCH APPLICATIONS FIRST.\n", true))
		{
			return false;
		}
	}
	
	memset(&evt, 0, sizeof(evt));
	evt.tnum = -1;	// target - could be sector or line number
	strcpy(evt.text, "LINE ");
	return true;
}

void SaveXhair()
{
	if (!xhairsaved)
	{
		evt.xh_col[0] = cfg.xhairColor[0];
		evt.xh_col[1] = cfg.xhairColor[1];
		evt.xh_col[2] = cfg.xhairColor[2];
		evt.xhairType = cfg.xhairType;
		evt.xhairSize = cfg.xhairSize;
		xhairsaved = true;
	}
}

void RestoreXhair()
{
	if (xhairsaved)
	{
		cfg.xhairColor[0] = evt.xh_col[0];
		cfg.xhairColor[1] = evt.xh_col[1];
		cfg.xhairColor[2] = evt.xh_col[2];
		cfg.xhairType = evt.xhairType;
		cfg.xhairSize = evt.xhairSize;
		xhairsaved = false;
	}
}

void Edit_SetVars()
{
	if (!R3D_edit) return;

	editvarsset = true;

	ForceEditDefaultStatus();

	Con_Executef(false, "reveal 1");

	SaveXhair();

	cfg.xhairColor[0] = 255;
	cfg.xhairColor[1] = 255;
	cfg.xhairColor[2] = 255;
	cfg.xhairType = 1;
	cfg.xhairSize = 1;

	evt.use_g = r_texglow > 0;
	evt.use_l = r_ambient;
	evt.use_sbar = cfg.sbarscale;
	evt.use_fstp = cfg.footsteps;

	cfg.sbarscale = 1;
	cfg.footsteps = 0;

	cfg.useTorchLight = false;
	cfg.setVtxLight = false;
	cfg.sineCompress = false;

	r_texglow = false;
	r_ambient = evt.use_l;

	if (ArgCheckWith("-edit_ambient", 1))
	{
		int ll = atoi(ArgNext());
		if (ll <   0) ll =   0;
		if (ll > 255) ll = 255;
		r_ambient = ll;
	}
}

void Edit_ResetVars()
{
	if (!editvarsset) return;
	
	editvarsset = false;

	cfg.sineCompress = (ge_inf.flags & GEF_SINECOMPRESS) > 0;
	cfg.useTorchLight = (ge_inf.flags & GEF_TORCHLIGHT) > 0;
	cfg.setVtxLight = (ge_inf.flags & GEF_VTXLIGHT) > 0;

	r_texglow = evt.use_g > 0;
	r_ambient = evt.use_l;
	cfg.sbarscale = evt.use_sbar;
	cfg.footsteps = evt.use_fstp;

	RestoreXhair();

	modifyWAD(P_ListChangedSides());

	if (timecache[ITC_NODESBUILD] > 1.0f) delNodesBuild();
}

static int P_ListChangedSides()
{
	char	buff[512];

	FILE	*outfile;

	byte	*sidarray;
	side_t	*sd = sides;
	line_t	*ld;
	int		c, i;

	sprintf(buff, "%sedit.out", ddRuntimeDir.path);

	// Always clear edit.out to save any confusion
	if ( (outfile = fopen(buff, "w")) == NULL)
	{
		pr_OFdirect("ERROR: Failed to create edit.out file.\n");
		return 0;
	}

	if (!changemade)
	{
		fclose(outfile);
		return 0;
	}

	sidarray = (byte*)M_Calloc(numsides, 271);

	for (c = i = 0, sd = sides; i < numsides; i++, sd++)
	{
		int noffx = sd->rowoffset>>16;
		int noffy = sd->textureoffset>>16;
		int offx  = (int)sd->orig_rowoffset;
		int offy  = (int)sd->orig_textureoffset;

		if (offx != noffx || offy != noffy)
		{
			sidarray[i] = 1;
			if (offx != noffx) c++;
			if (offy != noffy) c++;
		}
	}

	if (c)
	{
		int snum;

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if ( (snum = ld->sidenum[0]) >= 0 && sidarray[snum])
			{
				fprintf(outfile ,"LINE %i side 0 offx %i offy %i [was %i %i]\n", i,
					sides[snum].textureoffset>>16,
					sides[snum].rowoffset>>16,
					sides[snum].orig_textureoffset,
					sides[snum].orig_rowoffset);
			}

			if ( (snum = ld->sidenum[1]) >= 0 && sidarray[snum])
			{
				fprintf(outfile ,"LINE %i side 1 offx %i offy %i [was %i %i]\n", i,
					sides[snum].textureoffset>>16,
					sides[snum].rowoffset>>16,
					sides[snum].orig_textureoffset,
					sides[snum].orig_rowoffset);
			}
		}
	}

	M_Free(sidarray);

	fflush(outfile);
	fclose(outfile);

	return ma_inf.iwadedit ? 0 : c;
}

// Round to integer.
static void subfn_SetTNUMvals()
{
	changemade = true;	// Registers the fact that we got as far as the map

	if (evt.tnum >= 0 && evt.type & EV_WALLM)
	{
		side_t *sd = &sides[evt.snum];
		sd->rowoffset     = (int)(FIX2FLT(evt.xadj))*FRACUNIT;
		sd->textureoffset = (int)(FIX2FLT(evt.yadj))*FRACUNIT;
		// Ensure values get updated by renderer
		sides[evt.snum].lsiflags &= ~LS_TEXPRESENT;
	}
}

// Set adjust start values
static void subfn_SetTNUMadjust()
{
	if (evt.tnum >= 0 && evt.type & EV_WALLM)
	{
		side_t *sd = &sides[evt.snum];
		evt.xadj = sd->rowoffset;
		evt.yadj = sd->textureoffset;
	}
}

static void subfn_ResetTNUMvals()
{
	if (evt.type & EV_WALLM)
	{
		side_t *sd = &sides[evt.snum];
		sd->rowoffset     = evt.offx;
		sd->textureoffset = evt.offy;
	}

	evt.editstate = 0;
	evt.tnum = -1;
}

static void subfn_SetEVtex(side_t *sd, short lsiflags, int flat, int lval, int hval, int type)
{
	int texn  = type == EV_LO ? sd->bottomtexture :
				type == EV_HI ? sd->toptexture :
								sd->midtexture;

	if (sd->sector->stptr >= 0 && !(lines[evt.tnum].lnflags2 & (LNF2_LOWTRIWALL|LNF2_HITRIWALL)) )
	{
		if (!texn) texn = sides[lines[evt.tnum].sidenum[0]].midtexture;
		type = EV_INFILL;
	}

	if (!texn)
	{

		if (flat)
			sprintf(evt.sbst, "Used FLAT %s", lumpinfo[flat].name);
		else
			strcpy(evt.sbst, "tex unknown");
	}
	else
	{
		strcpy(evt.sbst, evt.r3d ? "3D line" : "");

		if (texArray[texn]->oldname[0])
		{
			sprintf(evt.name, "TEX %s", texArray[texn]->oldname);
			
			if (R3D_linemap && texArray[texn]->txf_flags & TXF_3DTEX)
			{
				strcat(evt.sbst, " scaled");
			}
		}
		else
			sprintf(evt.name, "TEX %s", texArray[texn]->name);
	}
	evt.masked	= texn && texArray[texn]->txf_flags & TXF_MASKED;
	evt.type	= type;
	evt.val_low	= lval;
	evt.val_hi	= hval;
}

int Edit_Textures(line_t *li, int z, int from) 
{
	if (!ppt.inmap) { evt.tnum = -1; return 0; }

	evt.name[0] = 0;
	evt.type = 0;
	evt.offx = 0;
	evt.offy = 0;
	evt.tnum = li - lines;
	strcpy(evt.text, "LINE ");

	if (evt.tnum >= 0)
	{
		side_t	*sd;
		int	  snum = li->sidenum[P_PointOnLineSide(viewx, viewy, li)];
		short lsiflags;

		evt.snum = snum;
		evt.r3d  = li->specType >= R3D_40;

		if (evt.snum < 0)
		{
			evt.tnum = -1;
			return 0;
		}

		sd = &sides[snum];
		lsiflags = sd->lsiflags;

		if (lsiflags & LS_LOWTEX      && z >= sd->ll && z <= sd->lh)
			subfn_SetEVtex(sd, lsiflags, sd->flatl, sd->ll, sd->lh, EV_LO);
		else if (lsiflags & LS_MIDTEX && z >= sd->ml + sd->editml && z <= sd->mh - sd->editmh)
			subfn_SetEVtex(sd, lsiflags,         0, sd->ml, sd->mh, EV_MD);
		else if (lsiflags & LS_HITEX &&  z >= sd->hl && z <= sd->hh)
			subfn_SetEVtex(sd, lsiflags, sd->flath, sd->hl, sd->hh, EV_HI);
		else if (lsiflags & LS_MIDTEXH && z >= sd->sl0 && z <= sd->sh0)
			subfn_SetEVtex(sd, lsiflags,         0, sd->sl0, sd->sh0, EV_MDH);
		else if (lsiflags & LS_MIDTEXL && z >= sd->sl1 && z <= sd->sh1)
			subfn_SetEVtex(sd, lsiflags,         0, sd->sl1, sd->sh1, EV_MDL);

		if (evt.name[0])
		{
			evt.offx = sd->rowoffset;
			evt.offy = sd->textureoffset;
			evt.startm  = 1;
		//	evt.name[8] = 0;
			if (evt.type == EV_LO)
			{
				strcpy(evt.textdet, "LOW" );
				if (lsiflags & LS_LOWALPHA)
					strcat(evt.textdet, " MASKED" );
				else if (lsiflags & LS_LOWTRANS)
					strcat(evt.textdet, " TRANS" );
			}
			if (evt.type == EV_MD)
			{
				strcpy(evt.textdet, "MID" );
				if (lsiflags & LS_MIDALPHA)
					strcat(evt.textdet, " MASKED" );
				else if (lsiflags & LS_MIDTRANS)
					strcat(evt.textdet, " TRANS" );
			}
			if (evt.type == EV_HI)
			{
				strcpy(evt.textdet, "HIGH");
				if (lsiflags & LS_HIALPHA)
					strcat(evt.textdet, " MASKED" );
				else if (lsiflags & LS_HITRANS)	// here for completeness (rare)
					strcat(evt.textdet, " TRANS" );
			}
			if (evt.type == EV_MDH)
			{
				strcpy(evt.textdet, "HIGH SLOPE");
			}
			if (evt.type == EV_MDL)
			{
				strcpy(evt.textdet, "LOW SLOPE");
			}
			if (evt.type == EV_INFILL)
			{
				strcpy(evt.textdet, "SLOPE INFILL");
			}
		}
		else
			evt.tnum = -1;
	}
	return evt.tnum >= 0;
}

void Edit_SetMB2()
{
	if (R3D_edit && evt.tnum >= 0)
		evt.mb2 = 1 + evt.editstate;
}

int Edit_CheckTNUM()
{
	if (evt.tnum >= 0)
	{
		evt.tnum = -1;
		return true;
	}
	return false;
}

int Edit_CheckEditstate(event_t *ev)
{
	if (R3D_edit &&
		evt.editstate &&
		evt.tnum >= 0 &&
		ev->type == ev_keydown)
	{
		if (ev->data1 == DDKEY_ESCAPE)
		{
			evt.esc = true;
			return true;
		}
		if (ev->data1 == DDKEY_BACKSPACE)
		{
			if (evt.type & EV_WALLM)
			{
				side_t *sd = &sides[evt.snum];
				evt.offx = sd->orig_rowoffset << FRACBITS;
				evt.offy = sd->orig_textureoffset << FRACBITS;
				evt.esc = true;
				return true;
			}
		}
	}
	return false;
}

int Edit_CheckEditSelect(int fire, int *mousex, int *mousey)
{
	int ret = false;

	if (evt.tnum >= 0 && evt.mb2)
	{
		evt.editstate = evt.mb2 == 1 ? 1 : 0;
		if (!evt.editstate)
		{
			if (evt.mb2 == 2)
				subfn_ResetTNUMvals();
			else
				evt.tnum = -1;
			// Reset flags so positions will be updated in renderer
			if (evt.snum >= 0)
				sides[evt.snum].lsiflags &= ~LS_TEXPRESENT;
			*mousex = *mousey = 0;
		}
		evt.mb2 = 0; 
		ret = true;
	}

	if (evt.editstate)
	{
		if (evt.tnum < 0 || fire)
		{
			evt.editstate = 0;
			*mousex = *mousey = 0;
		}
		else if (evt.tnum >= 0)
		{
			int xc, yc;
			if (evt.startm)
			{
				evt.mousex = *mousex;
				evt.mousey = *mousey;
				evt.startm = 0;
				evt.esc    = 0;
				subfn_SetTNUMadjust();
			}
			xc = EDITINC*((*mousex) - evt.mousex);
			yc = EDITINC*((*mousey) - evt.mousey);

			if (evt.type & EV_WALLM)
			{
				// Note the mouse x axis adjusts the y offset and y the x offset
				evt.yadj -= xc;
				if (!(evt.r3d && evt.masked))
					evt.xadj += yc;
			}
			else
			{
				evt.xadj -= xc;
				evt.yadj -= yc;
			}
			subfn_SetTNUMvals();
			evt.mousex = *mousex;
			evt.mousey = *mousey;
			if (evt.esc)
			{
				subfn_ResetTNUMvals();
				evt.editstate = 0;
				*mousex = *mousey = 0;
			}
		}
		return true;
	}
	return ret;
}

//	Displays the edit window.
void Edit_DrawerEditWindow()
{
	side_t		*sd = NULL;
	sector_t	*sc = NULL;
	int			x, y = 30, w = cfg.amEditboxWidth, h;
	int			posx, posy, posz;
	char		buf[64];

	if (cpossec < 0) return;

	FR_SetFont(0);

	posx = viewx>>16;
	posy = viewy>>16;
	posz = viewz>>16;
	
	// R3D_edit checked in case being used for debugging in normal game mode
	if (R3D_edit && evt.tnum >= 0)
	{
		if (evt.type & EV_WALLM)
		{
			sd = &sides[evt.snum];
			sprintf(buf, "%s %i side %i",
				evt.text, evt.tnum, !(evt.snum == lines[evt.tnum].sidenum[0]));
		}
	}
	else
	{
		strcpy(buf, "No selection");
	}

	h = FR_TextHeight(buf);
	x = current_w - w - 10;

	// Go into screen projection mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, current_w, current_h, 0, -1, 1);
	
	UI_DrawRectEx((float)x, (float)y, w, h*6 + 24, 4, 
		UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW), .5f);
	UI_Gradient(x + 4, y + 4, w - 8, h - 8, 
		UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), .5f, .5f);
	UI_Color(UI_COL(UIC_TEXT));
	UI_TextOutEx(buf, x + 8, y + 12, false, true, UI_COL(UIC_TEXT), 1);

	if (evt.tnum >= 0)
	{
		sprintf(buf, "%s TEX", evt.textdet);
		UI_TextOutEx(buf, x + 8, y + h + 14, false, true, UI_COL(UIC_TEXT), 1);

		if (sd)
			sprintf(buf, "xoff %3i yoff %3i", sd->textureoffset>>16, sd->rowoffset>>16);

		UI_TextOutEx(buf, x + 8, y + h*2 + 16, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "%s", evt.name);
		UI_TextOutEx(buf, x + 8, y + h*3 + 18, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "%s", evt.sbst);
		UI_TextOutEx(buf, x + 8, y + h*4 + 20, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "EDIT MODE : %s", evt.editstate ? evt.r3d && evt.masked ? "X only" : "On" : "Off");
		UI_TextOutEx(buf, x + 8, y + h*5 + 22, false, true, UI_COL(UIC_TEXT), 1);
	}

	if (compat[c_camera])
	{
		int d = vidShowFPS ? h*6 + 26 : h*5 + 24;

		y += h*8;

		UI_DrawRectEx((float)x, (float)y, w, d+16, 4, 
				UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW), .5f);

		UI_Gradient(x + 4, y + 4, w - 8, h - 8, 
			UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), .5f, .5f);
		UI_Color(UI_COL(UIC_TEXT));
		sprintf(buf, "Camera position");
		UI_TextOutEx(buf, x + 8, y + 12, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "xy: %5i %5i", posx, posy);
		UI_TextOutEx(buf, x + 8, y + h + 14, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, " z: %5i", posz);
		UI_TextOutEx(buf, x + 8, y + h*2 + 16, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "sector : %i", cpossec);
		UI_TextOutEx(buf, x + 8, y + h*3 + 18, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "subsec : %i", cpossubsec);
		UI_TextOutEx(buf, x + 8, y + h*4 + 20, false, true, UI_COL(UIC_TEXT), 1);
		sprintf(buf, "angle : %.2f", dplookdir * 90.0f/110.0f);
		UI_TextOutEx(buf, x + 8, y + h*5 + 20, false, true, UI_COL(UIC_TEXT), 1);
		if (vidShowFPS)
		{
			sprintf(buf, "%3i FPS", stats.fpsnum);
			UI_TextOutEx(buf, x + 8, y + h*6 + 22, false, true, UI_COL(UIC_TEXT), 1);
		}
	}

	// Restore original matrix.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}


void P_CheckChangedSides()
{
	side_t	*sd = sides;
	int		c, i;

	for (c = i = 0, sd = sides; i < numsides; i++, sd++)
	{
		int noffx = sd->rowoffset>>16;
		int noffy = sd->textureoffset>>16;
		int offx  = (int)sd->orig_rowoffset;
		int offy  = (int)sd->orig_textureoffset;

		if (offx != noffx || offy != noffy)
		{
			if (offx != noffx) c++;
			if (offy != noffy) c++;
		}
	}

	if (c)
	{

	}
}
