//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_inherit.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//
// DESCRIPTION: checks for true inheritance with multiple choice
//
//-----------------------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"

//LS_chkInherit ------------------------------------------------------------

static int subfn_chkSecHasSSLine(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
		if (!(sec->lines[i]->ml_flags & ML_TWOSIDED))
			return true;
	return false;
}

static int subfn_CheckLinesForSector(subsector_t *ss, line_t *ld)
{
	seg_t	*seg = &segs[ss->firstline];
	int		i;

	for (i = 0; i < ss->linecount; i++, seg++)
	{
		line_t *li = seg->linedef;

		if (li != NULL
			&& LSS(li)
			&& li->frontsector != ld->frontsector
			&& (lnsec[li - lines] < 0 || lnsec[li - lines] == li->frontsector - sectors)
			&& !SecCtrl(li->frontsector - sectors)
			&& !LineHasMovSectors(li)
			&& li->frontsector->heightsec < 0
			&& subfn_chkSecHasSSLine(li->frontsector))
		{
			return li->frontsector - sectors;
		}
	}
	return -1;
}

static int subfn_chkInheritance(line_t *ld)
{
	int i, res;
	ssline_t *sl = &sslines[ld - lines];

	for (i = 0; i < sl->num; i++)
	{
		if ((res = subfn_CheckLinesForSector(&subsectors[sl->list[i]], ld)) >= 0)
			return res;
	}
	return -1;
}

void LS_chkInherit()
{
	sector_t *sec;
	int		 i, k, secnum;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& SecCtrl(i)
			&& !IsLiquidFloorFlat(sec)
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& sec->specFlags & SF_NOHITEX
			&& sec->specFlags & SF_NOLOTEX
			&& sec->heightsec < 0)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (LSS(ld)
					&& Vtxs1(ld)
					&& !(ld->spflags & SPF_CONTROLLINE)
					&& !ChkLFGSecSet(ld)
					&& (lnsec[ld - lines] < 0 || lnsec[ld - lines] == ld->frontsector - sectors)
					&& (secnum = subfn_chkInheritance(ld)) >= 0)
				{
					Setspec(ld, secnum, 41005);
					SetLFGSecSet(ld, 41006);
				}
			}
		}
	}
}

// LS_InheritanceMovFloorLoop ------------------------------------------------

// sc is the floor move sector
static void checkInheritanceLinesLoop(sector_t *sc)
{
	sector_t	*chk, *os = NULL;
	line_t		*ld;
	int			rc, box[4];
	int			c, i, lnum;

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (   origTEX[ld - lines] == 0x15
			&& ld->frontsector != ld->backsector
			&& ld->ml_flags & ML_TWOSIDED	// should be
			&& Vtxs2(ld))
		{
			chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

			if (!os)
			{
				os = chk;
				if (os->specFlags & SF_MOVSECTOR) return;
			}
			else if (os != chk)
			{
				return;
			}

			lnum = ld - lines;
			hitline[lnum] = 1; c++;
		}
	}

	if (c < 3) return;

	if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
		&& rc == c
		&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE))
	{
		subsector_t	*ss;
		seg_t		*sg;
		int			k;

		validcount++;

		for (c = i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];
			sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (  (ld = sg->linedef)
					&& ld->validcount != validcount)
				{
					ld->validcount = validcount;
					lnum = ld - lines;
					linearray[c++] = lnum;

					if (hitline[lnum]) continue;

					if (   ld->ml_flags & ML_TWOSIDED
						&& ld->frontsector->floorheight != ld->backsector->floorheight
						&& origTEX[lnum] & 0x11)
					{
						return;
					}
				}
			}
		}

		if (c <= rc) return;	// No lines inside

		// No lotex found so set floorsec of ssecs to sc
		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];

			if (ss->sector != sc)
			{
				Setr3dssfs(ss, sc - sectors, 41012);
				if (ss->sector->specFlags & 0x8 && ss->sector->ceilingheight > sc->ceilingheight)
					Setr3dsslcsec(ss, sc - sectors, 41012);
			}

			ss->ssflags |= SSF_NOCHANGE;
		}

		// Set the line types
		for (i = 0; i < c; i++)
		{
			lnum = linearray[i];

			ld = &lines[lnum];

			if (origTEX[lnum] & 0x22 && !(origTEX[lnum] & 0x44))
				Settype(ld, R3D_03, 41013);
			else
				Settype(ld, R3D_36, 41013);

			if (  !(origTEX[lnum] & 0x11)
				&& ld->ml_flags & ML_TWOSIDED
				&& ld->frontsector->floorheight != ld->backsector->floorheight)
			{
				SetMissinglo(ld, 31013);
			}

			if (ld->mtflags) Clr_allmtflags(ld, 41013);

			SetLFG(ld, LFG_NEWSET|LFG_FIXED, 41013);
		}

		Setsecdone(sc, 41013);
	}
}

static int checkOwner(sector_t *sc, int count)
{
	subsector_t *ss;
	seg_t		*sg;
	sector_t	*os = NULL;
	int			i, k, m, n;
	
	ss_validcount++;

	for (i = 0; i < count; i++)
	{
		n = linearray[i];

		if (hitline[n] == 4)
		{
			ssline_t *sl = &sslines[n];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;

				for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
				{
					if (!sg->frontsector || sg->frontsector == sc) continue;
					if (!os)
					{
						os = sg->frontsector;
						if ((os->specFlags & 0x40C) == 0xC) return -1;
					}
					else if (os != sg->frontsector) return -1;
				}
			}
		}
	}

	return os ? os - sectors : -1;
}

// v220-10
static void checkInheritanceNoTexLinesLoop(sector_t *sc)
{
	line_t	*ld;
	int		c, d, i;
	int		*lineowner;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   ld->frontsector != ld->backsector
			|| lineflags[ld - lines]
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOONE)
			|| origTEX[ld - lines]
			|| !Vtxs2(ld))
		{
			return;
		}

		linearray[c++] = ld - lines;
	}

	if (c < 3) return;

	CLR_hitline();
	lineowner = (int*)malloc(c * sizeof(int));
	memset(lineowner, -1, c * sizeof(int));

	for (i = 0; i < c; i++) hitline[linearray[i]] = 1;

	d = 0;

	while(1)
	{
		int		lown, lnum = -1;

		for (i = 0; i < c; i++)
		{
			if (hitline[linearray[i]] == 1)
			{
				lnum = linearray[i]; break;
			}
		}

		if (lnum < 0) break;

		if (subfn_CheckForRing(hitline, sc, lnum, NULL, NULL, NULL, NULL))
		{
			lown = checkOwner(sc, c);

			for (i = 0; i < c; i++)
			{
				if (hitline[linearray[i]] == 4)
				{
					hitline[linearray[i]] = 0;
					lineowner[i] = lown; d++;
				}
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}

	if (d)
	{
		int k, m, n;

		ss_validcount++;

		for (i = 0; i < c; i++)
		{
			if ( (n = lineowner[i]) >= 0)
			{
				ssline_t *sl = &sslines[linearray[i]];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];
					seg_t		*sg;

					if (ss->ssvalidcount == ss_validcount) continue;
					
					ss->ssvalidcount = ss_validcount;

					if (ss->sector - sectors != n)
						SetOverride(ss, n, 41014);

					for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
					{
						line_t *ld = sg->linedef;

						if (ld && !lineflags[ld - lines])
						{
							if (   ld->frontsector == sc
								&& ld->frontsector == ld->backsector)
							{
								Settype(ld, R3D_3F, 41014);
								SetLFG(ld, LFG_NEWSET|LFG_FIXED, 41014);
							}
						}
					}
				}
			}
		}

		if (d == c) Setsecdone(sc, 41014);
	}

	M_Free(lineowner);
}

// v220-10
// e.g. PAR-Lutz E1M6 lines in S397
// Note: this is critically depedant upon mov flag settings
void LS_InheritanceMovFloorLoop()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if ((sectors[i].specFlags & 0x3000C) == SF_MOVFLOOR)
		{
			checkInheritanceLinesLoop(&sectors[i]);
		}
		else if ((sectors[i].specFlags & 0x3000C) == (SF_MOVFLOOR|0xC))
		{
			checkInheritanceNoTexLinesLoop(&sectors[i]);
		}
	}
}

// Code end