//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2014 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_lcross.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
//	Removed LS_CheckCrossingSupportInFake and LS_CheckCeilWithCrossingLine
//
// DESCRIPTION: analyses lines crossing lines
//
//-----------------------------------------------------------------------------

#include "r3d_MAINspec.h"

void LS_SetTypeR3D_2ECrossing()
{
	line_t *ld = lines, *cross;
	int    i;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (!(ld->ml_flags & ML_TWOSIDED))
			continue;

		if (ld->specType == R3D_00 &&
			ld->tyflags & TYF_MIDTEXLINE &&
			sides[ld->sidenum[0]].rowoffset <= 0 &&
			sides[ld->sidenum[0]].rowoffset ==
			sides[ld->sidenum[1]].rowoffset)
		{
			Settype(ld, R3D_04, 5001);
		}
		else if (ld->specType == R3D_01 &&
			Vtxs1(ld) &&
			!ld->mtflags &&
			!sides[ld->sidenum[0]].texFlags  &&
			!sides[ld->sidenum[1]].texFlags  &&
			!sides[ld->sidenum[0]].rowoffset &&
			!sides[ld->sidenum[1]].rowoffset &&
			ld->frontsector->ceilingheight -
			ld->frontsector->floorheight >= 8*FRACUNIT &&
			(cross = subfn_FindLineCrossing(ld, LC_SAMESECLINE)) != NULL &&
			 cross->tyflags & TYF_MIDTEXLINE)
			Settype(ld, R3D_2E, 5002);
	}
}

// Mark all lines connected to lines crossing the test line
// Sets vtxflags of all lines connected to a seed line with
// the same front and back sector.
// The seeding lines vtxcheck flags must have been set prior to calling,
static void markLineVertexes(sector_t *s)
{
	RDBOOL	repeat = true;
	int		i;

	while(repeat)
	{
		repeat = false;

		for (i = 0; i < s->linecount; i++)
		{
			line_t *li = s->lines[i];

			if (li->frontsector == li->backsector)
			{
				RDBOOL c1 = (vtxcheck[li->v1 - vertexes]) > 0;
				RDBOOL c2 = (vtxcheck[li->v2 - vertexes]) > 0;
				
				if ((c1 && !c2) || (!c1 && c2))
				{
					vtxcheck[li->v1 - vertexes]++;
					vtxcheck[li->v2 - vertexes]++;
					repeat = true;
				}
			}
		}
	}
}

void LS_MiscLineCrossAndLineCoincident()
{
	sector_t	*s;
	line_t		*ld = lines, *cross, *ldc;
	int			i, lnarraycnt = 0;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (!(ld->ml_flags & ML_TWOSIDED)
			|| ChkLFGSecSet(ld)
			|| ld->specType >= R3D_40)
		{
			continue;
		}

		// Fix coincident line problem preventing tagged operation
		if (ld->tag && ld->special
			&& P_FindSectorFromTag(ld->tag, -1) == ld->backsector - sectors
			&& (ldc = subfn_FindCoincidentLine(ld, 0)) != NULL
			&& !ldc->tag && !ldc->special
			&& ldc->backsector == ld->backsector)
		{
			ldc->lnflags2 |= LNF2_TRIGCOINCIDENT;
		}

		if (ld->specType == R3D_01
			&& Vtxs2(ld)
			&& subfn_FindMultiLineCrossing(ld, LC_NOVTXONLINE) > 1)
		{
			Settype(ld, R3D_2E, 5003);
			Setspec(ld, multicrossline->frontsector - sectors, 5004);
		}

		else if (ld->specType == R3D_2E
			&& (cross = subfn_FindLineCrossing(ld, LC_NORMAL)) != NULL
			&& cross->specType == R3D_03)
		{
			Settype(ld, R3D_01, 5005);
			Setspec(ld, -1, 5006);
		}

		else if (ld->specType == R3D_00
			&& !ld->mtflags
			&& ld->tyflags & TYF_MIDTEXLINE
			&& sides[ld->sidenum[0]].rowoffset >= 0
			&& sides[ld->sidenum[1]].rowoffset >= 0)
		{
			Settype(ld, R3D_04, 5007);
		}

		if (ld->specType == R3D_00
			&& sides[ld->sidenum[0]].texFlags == T_ALL
			&& sides[ld->sidenum[1]].texFlags == T_MID
			&& (ldc = subfn_FindCoincidentLine(ld, 0)) != NULL
			&& (ldc->specType == R3D_04 || ldc->specType == R3D_06)
			&& LSS(ldc)
			&& ldc->frontsector != ld->frontsector
			&& ldc->frontsector != ld->backsector)
		{
			subfn_CheckNodesWithLine(ld->backsector, ldc->frontsector, ld);
			Setspec(ldc, -1, 5008);
			Setspec(ld , -1, 5009);
			Settype(ldc, R3D_04, 5010);
		}

		if (ld->specType == R3D_01 &&
			!origTEX[i] && Vtxs1(ld))
		{
			if ((cross = subfn_FindLineCrossing(ld, LC_NOT_SHARED|LC_NOVTXONLINE)) != NULL
				&& cross->frontsector != cross->backsector
				&& !(origTEX[cross - lines] & 0x22))
			{
				Settype(ld, R3D_2E, 5011);
			}
			else
			{
				if (!IsLiquidFloorFlat(ld->frontsector))
				{
					// E.g. COD MAP07 Sec 268 line 4960.
					// This should have been caught by the traceGen code?
					int secnum = LS_GetAdjacentLineSecnum(ld, 2);

					if (secnum >= 0 && IsLiquidFloorFlat(&sectors[secnum]))
					{
						Setspec(ld, secnum, 5012);
					}
					else if (cross
						&& LSS(cross)
						&& origTEX[cross - lines] == 0x22)
					{
						if (LSS(ld)
							&& SecCtrl(cross->frontsector - sectors)
							&& cross->frontsector->specFlags & SF_MOVSECTOR
							&& !SecCtrl(ld->frontsector - sectors)
							&& !(ld->frontsector->specFlags & SF_MOVSECTOR))
						{
							Setspec(cross, ld->frontsector - sectors, 5013);
						}
						else
						{
							Setspec(ld, cross->frontsector - sectors, 5014);
							if (ld->frontsector == ld->backsector)
								linearray[lnarraycnt++] = i;
						}
					}
				}
				Settype(ld, R3D_3F, 5015);
			}
		}
	}

	for (i = 0; i < lnarraycnt; i++)
	{
		int		c, k, specSec;

		// check crossing lines are set
		ld		= &lines[linearray[i]];
		s		= ld->frontsector;
		specSec	= lnsec[ld - lines];

		CLR_vtxcheck();

		for (c = k = 0; k < s->linecount; k++)
		{
			line_t *li = s->lines[k];

			if (li->frontsector == li->backsector
				&& !(lineflags[li - lines] & (LFG_NEWSET|LFG_3DTYPE))
				&& li - lines != linearray[i]
				&& (ld = subfn_FindLineCrossing(li, LC_SAMESECLINE|LC_OWNSECLINE)))
			{
				if (ld->frontsector - sectors == specSec)
				{
					Setspec(li, specSec, 5016);
					vtxcheck[li->v1 - vertexes]++;
					vtxcheck[li->v2 - vertexes]++;
					c++;
				}
			}
		}

		if (!c) continue;

		markLineVertexes(s);

		for (k = 0; k < s->linecount; k++)
		{
			line_t *li = s->lines[k];

			if (   li->frontsector == li->backsector
				&& lnsec[li - lines] < 0
				&& (vtxcheck[li->v1 - vertexes]
				||  vtxcheck[li->v1 - vertexes])
				&& !subfn_FindLineCrossing(li, LC_NORMAL))
			{
				Setspec(li, specSec, 5017);
			}
		}
	}
	// Find same sided lines connected to 1's lines and
	// then check for any lines crossing them
	for (lnarraycnt = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !lineflags[i]		// v220-26 add lineflags test
			&& ld->frontsector == ld->backsector
			&& Spans1sLines(ld))
		{
			linearray[lnarraycnt++] = i;
		}
	}

	for (i = 0; i < lnarraycnt; i++)
	{
		if ((ld = subfn_FindLineCrossing(&lines[linearray[i]], LC_SAMESECLINE|LC_OWNSECLINE|LC_NOVTXONLINE)) )
		{
			int		k, specSec = lines[linearray[i]].frontsector - sectors;

			s = ld->frontsector;

			CLR_vtxcheck();

			vtxcheck[ld->v1 - vertexes]++;
			vtxcheck[ld->v2 - vertexes]++;

			// Mark all lines connected to lines crossing the test line
			markLineVertexes(s);

			for (k = 0; k < s->linecount; k++)
			{
				line_t *li = s->lines[k];

				if (li->frontsector == li->backsector
					&& lnsec[li - lines] < 0
					&& (vtxcheck[li->v1 - vertexes]
					||  vtxcheck[li->v1 - vertexes])
					&& !subfn_FindLineCrossing(li, LC_NORMAL))
				{
					Setspec(li, specSec, 5018);
				}
			}
		}
	}
}
