//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_linegroups.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: general routines moved from r3d_mainspec.cpp
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"

static int	*lnmovunset;
static int	lnmovcount;

// CODE --------------------------------------------------------------------

// v220-8 accommodate moving sectors where a texture set at 0 should adopt
//        the lower facing sector's floor text. This is not supported in R3D
//        so tex num 1 is used if in the PNAMES list. This is very rare.
void LG_CheckMovTexUnset()
{
	sector_t	*fs, *bs;
	int			i;

	lnmovunset = NULL;
	lnmovcount = 0;

	for (i = 0; i < numlines; i++)
	{
		if (   lines[i].ml_flags & ML_TWOSIDED
			&& (fs = lines[i].frontsector) != (bs = lines[i].backsector)
			&& fs->specFlags & SF_MOVFLOOR
			&& bs->specFlags & SF_MOVFLOOR
			&& !vtoflags[lines[i].v1 - vertexes]
			&& !vtoflags[lines[i].v2 - vertexes])
		{
			if (  (origTEX[i] & 0x11) == 0x01
				&& fs->floorheight >= bs->floorheight)
			{
				if (lines[i].lnflags2 & LNF2_OWNEQUALTOTWO)
					linearray[lnmovcount++] = i;
			}
			else
			if (  (origTEX[i] & 0x11) == 0x10
				&& fs->floorheight <= bs->floorheight)
			{
				if (lines[i].lnflags2 & LNF2_OWNEQUALTOTWO)
					linearray[lnmovcount++] = i;
			}
		}
	}

	// v220-22 deal with this in LG_AnalyseMovTexUnset
	if (lnmovcount > 5)
	{
		lnmovunset = (int*)W_Malloc(lnmovcount * sizeof(int), LEVELID, 18);
		memcpy(lnmovunset, linearray, lnmovcount * sizeof(int));
	}
	else
	{
		lnmovcount = 0;
	}
}

// v220-22   check if any action is needed with lines set in lnmovunset
//           Specifically deals with dual nested movfloor structures
void LG_AnalyseMovTexUnset()
{
	if (!lnmovcount || ma_inf.noMovUnset) return;

	sector_t	*s;
	int			c, d, i, k;
	int			lnum, rc, box[4];
	int			cs = 0;

	CLR_hitline();
	SET_countsec(0);

	for (i = 0; i < lnmovcount; i++)
	{
		if (   !lineflags[lnmovunset[i]]
			&& !(lines[lnmovunset[i]].spflags & SPF_CONTROLLINE) )
		{
			hitline[lnmovunset[i]] = 1;
		}
	}

	while (1)
	{
		for (lnum = -1, i = 0; i < lnmovcount; i++)
		{
			if (hitline[lnmovunset[i]] == 1) { lnum = lnmovunset[i]; break; }
		}

		if (lnum < 0) break;


		s = origTEX[lnum] & 0x01 ? lines[lnum].backsector : lines[lnum].frontsector;

		if (subfn_CheckForRing(hitline, s, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			RDBOOL set = false;

			if (LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box, -1, MSIA_NONE))
			{
				// We need to establish a dual loop
				validcount++;

				for (c = i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (ss->sector == s)
					{
						seg_t	*sg = &segs[ss->firstline];
						line_t	*ld;

						for (k = 0; k < ss->linecount; k++, sg++)
						{
							if (sg->frontsector != s) { c = 0; break; }

							if ( (ld = sg->linedef) && ld->validcount != validcount)
							{
								ld->validcount = validcount;

								if (hitline[ld - lines] & (8|1))
									linearray[c++] = ld - lines;
							}
						}
					}
				}

				if (c > 2 && (d = nf_lineList(&lines[lnum], s, NULL, NULL, 0)) > 2)
				{
					s = origTEX[lnum] & 0x01 ? lines[lnum].frontsector : lines[lnum].backsector;

					for (i = 0; i < hitssecptr; i++)
					{
						subsector_t *ss = &subsectors[hitssec[i]];

						Setr3dsslfsec(ss, s - sectors, 999);
					}

					for (i = 0; i < c; i++)
					{
						Settype(&lines[linearray[i]], R3D_36, 6001);
						SetLFG(&lines[linearray[i]], LFG_NEWSET, 6001);
						hitline[linearray[i]] = 0;

						if (!cs)
						{
							addSec(lines[linearray[i]].frontsector - sectors, &cs);
							addSec(lines[linearray[i]].backsector - sectors, &cs);
						}
					}

					set = true;
				}
			}

			for (i = 0; i < lnmovcount; i++)
			{
				if (hitline[lnmovunset[i]] == 4)
				{
					hitline[lnmovunset[i]] = set ? 0 : 8;

					if (set)
					{
						Settype(&lines[lnmovunset[i]], R3D_36, 6002);
						SetLFG(&lines[lnmovunset[i]], LFG_NEWSET, 6002);
					}
				}
			}
		}
		else
		{
			hitline[lnum] = 8;
		}
	}

	for (i = 0; i < cs; i++)
		Setsecdone(&sectors[countsec[i]], 6002);

	W_Free(lnmovunset, LEVELID, 0);
}

// Remove ref in origTEX for AASHITTY tex's for closed static secs
void LG_RemoveAashitty()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->specFlags == 0x0080101C)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				origTEX[ld - lines] = 0;
			}
			sec->specFlags &= ~SF_HASORIGLOTEX;
		}
	}
}

void LG_UnclosedMovSectors()
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& SecCtrl(i)
			&& (sec->specFlags & 0x0003FF1C) == 0x0001000C
			&& sec->ceilingpic == skyflatnum)
		{
			sector_t *os;
			int k, count = 0, lnum;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (!(ld->spflags & SPF_CONTROLLINE))
				{
					os = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;

					if (!os || os->specFlags & SF_MOVSECTOR)
					{
						count = 0;
						break;
					}
					lnum = ld - lines;
					hitline[lnum] = 1;
					count++;
				}
			}
			if (count && !subfn_CheckForRing(hitline, sec, lnum, NULL, NULL, NULL, NULL))
			{
				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (!(ld->spflags & SPF_CONTROLLINE))
					{
						Setspec(ld, os - sectors, 6003);
						Settype(ld, R3D_3F, 6003);
						SetLFG(ld, LFG_LINESSECSET|LFG_NOCHANGE, 6003);
					}
				}
				Setsecdone3(sec, 6003);
			}
		}
		// v220-4 check for no low/high tex movceil sectors that should inherit
		//        an adjacent non-moving sector's floor and ceiling
		else if (!secdone[i]
			&& !SecCtrl(i)
			&& (sec->specFlags & 0xF7F7C) == 0x2000C
			&& sec->linecount)
		{
			RDBOOL	valid = true;
			int		k;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t	*ld = sec->lines[k];
				int		txn;

				if (   !ld->backsector
					|| (txn = origTEX[ld - lines]) == 0x22
					|| !Spans1sLines(ld))
				{
					valid = false; break;
				}

				if (   (txn == 0x02 && ld->frontsector->specFlags & SF_MOVCEIL)
					|| (txn == 0x20 && ld->backsector->specFlags & SF_MOVCEIL))
				{
					valid = false; break;
				}
			}

			if (valid)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];
					sector_t *s = ld->frontsector->specFlags & SF_MOVCEIL ? ld->backsector : ld->frontsector;

					if (!(s->specFlags & SF_MOVSECTOR))
					{
						if (origTEX[ld - lines] & 0x22)
							Settype(ld, R3D_2D, 6004);
						else
							Settype(ld, R3D_3F, 6004);

						Setspec(ld, s - sectors, 6004);
						SetLFG(ld, LFG_NEWSET, 6004);
					}
				}

				Setsecdone(sec, 6004);
			}
		}
	}
}

void LG_CheckFixedInMovSector()
{
	line_t	*lc, *ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->frontsector != ld->backsector) continue;
		if (lineflags[i] & (LFG_NOARBITER|LFG_3DTYPE)) continue;

// Check for static sector entwined in moving sector
		if (   lxarray[i] & 4
			&& !SecCtrl(ld->frontsector - sectors)
			&& (ld->frontsector->specFlags & 0x3FFFC) == 0x888C
			&& (lc = subfn_FindLineCrossing(ld, LC_NORMAL))
			&& (lc->frontsector->specFlags & 0x1FF1C) == 0x1001C)
		{
			ssline_t	*sl = &sslines[i];
			int			k;
			int			lsec = ld->frontsector - sectors;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->sector != lc->frontsector)
					ss->sector = lc->frontsector;

				SetOverride(ss, ld->frontsector - sectors, 6005);
			}
		}
		else
// Fix problem caused by 1's lines where their side 2 faces a sector.
		if (   !lxarray[i]	// v220-3
			&& lnsec[ld - lines] < 0
			&& ld->specType != R3D_0C	// Added v2.1.00 build_34
			&& ld->specType != R3D_2E	// Added v2.1.00 build_35
			&& !origTEX[i]
			&& Vtxs1(ld))
		{
			ssline_t	*sl = &sslines[i];
			int			k, secn = -1;
			int			lsec = ld->frontsector - sectors;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];
				int snum = ss->sector - sectors;

				if (snum != lsec)
				{
					seg_t	*sg = &segs[ss->firstline];
					int		m;

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (!sg->linedef
							|| sg->linedef == ld) continue;

						if (sg->linedef->frontsector
							== sg->linedef->backsector
							|| vtxarray[sg->linedef->v1 - vertexes] == 1
							|| vtxarray[sg->linedef->v2 - vertexes] == 1
							// test for bum lines like AV map15 line 2451
							|| subfn_SsecFacingLine(ss, sg->linedef) != snum)
						{
							secn = -2;
							break;
						}
					}

					if (secn == -2)
						break;
					else if (secn < 0)
						secn = snum;
					else if (secn != snum)
					{
						secn = -1;
						break;
					}
				}
			}

			if (secn >= 0)
			{
				Setspec(ld, secn, 6006);
			}
		}
	}
}


// Check sectors for unset step line plats
void LG_CheckUnsetStepLinePlats()
{
	sector_t *sec;
	int		 i, k, ring_count, linnb, bbox[4];

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		int count = 0;

		if (!secdone[i]
			&& !SecCtrl(i)
			&& sec->linecount > 2
			&& (sec->specFlags & 0x00FFFFFC) == 0x1C)
		{
			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->frontsector != ld->backsector)
				{
					hitline[ld - lines] = 1;
					count++;
				}
			}
		}

		if (count < 3) continue;

		if (subfn_CheckForRing(hitline, sec, sec->lines[0] - lines, &ring_count, bbox, &linnb, NULL)
			&& !linnb && ring_count == sec->linecount
			&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE))
		{
			int m, lnptr = 0;

			validcount++;

			for (k = 0; k < hitssecptr; k++)
			{
				subsector_t *ss = &subsectors[hitssec[k]];
				seg_t		*sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					line_t *ld = sg->linedef;

					if (!ld || ld->validcount == validcount) continue;

					ld->validcount = validcount;

					if (hitline[ld - lines]) continue;

					if (ld->frontsector == ld->backsector)
						linearray[lnptr++] = ld - lines;
				}
			}
			if (lnptr)
			{
				for (k = 0; k < lnptr; k++)
				{
					Setspec(&lines[linearray[k]], i, 6007);
					SetLFGSecSet(&lines[linearray[k]], 6007);
				}

				Setsecdone3(sec, 6007);
			}
		}
	}
}

// v220-31 Set flags used to determine whether or not a high fake ceiling
// takes precedence based on whether there are missing high textures.
// Similarly for low textures.
void LG_FakeMissing()
{
	int		i, k, side;

	if (R3D_ScriptedMap) return;

	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].heightsec >= 0)
		{
			sector_t *sc = &sectors[i], *os;
			sector_t *hs = &sectors[sectors[i].heightsec];

			if (sectors[i].ceilingheight < hs->ceilingheight)
			{
				line_t	 *ld;

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];

					if (   ld->ml_flags & ML_TWOSIDED
						&& !(ld->spflags & SPF_CONTROLLINE)
						&& ld->sidenum[1] >= 0)
					{
						side = ld->frontsector == sc ? 0 : 1;
						os = side ? ld->frontsector : ld->backsector;

						if (os->ceilingheight <= hs->ceilingheight) // v220-35 hs change was == $hschange
						{
							if (os->heightsec < 0 || sectors[os->heightsec].ceilingheight == hs->ceilingheight)
							{
								sc->specFlags2 |= SF2_FAKEDHIGHMISS;
								break;
							}
						}
						
						if (os->heightsec >= 0 && !sides[ld->sidenum[!side]].toptexture)
						{
							sc->specFlags2 |= SF2_FAKEDHIGHMISS;
							break;
						}
					}
				}
			}

			if (sectors[i].floorheight > hs->floorheight)
			{
				line_t	 *ld;

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];

					if (   ld->ml_flags & ML_TWOSIDED
						&& !(ld->spflags & SPF_CONTROLLINE)
						&& ld->sidenum[1] >= 0)
					{
						side = ld->frontsector == sc ? 0 : 1;
						
						if (!sides[ld->sidenum[side]].bottomtexture)
						{
							sc->specFlags2 |= SF2_FAKEDLOWMISS;
							break;
						}
					}
				}
			}
		}
	}
}
