//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_lscheck.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
//
// DESCRIPTION: additional checking functions
//
//-----------------------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"

#include "m_bams.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "doom/tables.h"

// CODE ----------------------------------------------------------------------

void subfn_CheckValidLFSector(sector_t *sec)
{
	int i, count = 0, secnum = -1;

	CLR_vtxcheck();
	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE)
			continue;

		if (LSS(ld))
		{
			if (sides[ld->sidenum[0]].texFlags != T_MID ||
				sides[ld->sidenum[1]].texFlags != T_MID)
				return;
			continue;
		}

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->mtflags & (R3D_MISHITEX|R3D_INVTEXANY) ||
			vtxarray[ld->v1 - vertexes] != 2 ||
			vtxarray[ld->v2 - vertexes] != 2 ||
			sides[ld->sidenum[0]].texFlags & (T_LOW|T_HI) ||
			sides[ld->sidenum[1]].texFlags & (T_LOW|T_HI))
			return;

		if (sides[ld->sidenum[0]].texFlags != T_MID &&
			sides[ld->sidenum[1]].texFlags != T_MID)
			return;

		hitline[ld - lines] = 1;
		if (ld->frontsector == sec)
		{
			vtxcheck[ld->v1 - vertexes] |= 1;
			vtxcheck[ld->v2 - vertexes] |= 2;
		}
		else
		{
			vtxcheck[ld->v1 - vertexes] |= 2;
			vtxcheck[ld->v2 - vertexes] |= 1;
		}
		count++;
	}
	if (count != 4) return;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (hitline[ld - lines] &&
			(vtxcheck[ld->v1 - vertexes] != 3 ||
			 vtxcheck[ld->v2 - vertexes] != 3))
			 return;
	}
	
	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (hitline[ld - lines])
		{
			Settype(ld, R3D_3F, 8001);
			if (secnum < 0)
				secnum = LS_FindAdjacentSecnum(ld, 3);
		}
	}

	if (secnum >= 0)
	{
		for (i = 0; i < sec->linecount; i++)
			if (!(sec->lines[i]->spflags & SPF_CONTROLLINE))
				Settype(sec->lines[i], R3D_3D, 8002);

		Setlfsec(sec, secnum, 8003);
		Setsecdone(sec, 8004);
	}
}

void LS_CheckValidLFSector()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].linecount > 6
			&& !secdone[i]
			&& SecCtrl(i))
		{
			subfn_CheckValidLFSector(&sectors[i]);
		}
	}
}

////////////////////////////////////////////////////////////////////////////
//
//	LS_CheckR3D_03Ladders
//
// Routine to deal with ladders in COD

int subfn_GetSecFromTouchingLine(line_t *ldf)
{
	int i;

	for ( i = 0; i < numlines; i++)
	{
		line_t *ld = &lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->frontsector != ld->backsector ||
			ld->frontsector == ldf->frontsector)
			continue;

		if (LS_CheckLineIntersect(ldf, ld, LC_NOCONDITION))
			return ld->frontsector - sectors;
	}
	return -1;
}

void LS_CheckR3D_03Ladders()
{
	line_t *ldf;
	int i, specSec;

	for ( i = 0; i < numsectors; i++)
	{
		sector_t *sec = &sectors[i];
		line_t *ld;
		int k, count1 = 1, count2 = 2;

		if (secdone[i] || sec->linecount != 3 || SecCtrl(i)) continue;

		CLR_vtxcheck();

		for (k = 0; k < 3; k++)
		{
			ld = sec->lines[k];

			if (!(ld->ml_flags & ML_TWOSIDED) ||
				ld->frontsector != ld->backsector)
				break;

			if (ld->specType == R3D_03)
			{
				ldf = ld;
				count1--;
			}
			else if (sec->lines[k]->specType == R3D_01)
				count2--;

			vtxcheck[ld->v1 - vertexes] |= 1;
			vtxcheck[ld->v2 - vertexes] |= 2;
		}

		if (count1 || count2) continue;

		specSec = subfn_GetSecFromTouchingLine(ldf);

		if (specSec >= 0)
			for (k = 0; k < 3; k++)
				Setspec(sec->lines[k], specSec, 8007);
	}
}

void LS_CheckR3D_04Lines(void)
{
	line_t *ld = lines;
	int i;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_04)
		{
			if (ld->frontsector->specFlags & SF_04_CLOSED &&
				ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				Settype(ld, R3D_0E, 8008);
				Setspec(ld, LS_FindAdjacentSecnum(ld, 2), 8010);
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////
//
// These lines, if enclosed by R3D_03 lines, inherit specSec from R3D_03.
//
void LS_CheckR3D_06Lines()
{
	subsector_t	*ssec = subsectors;
	sector_t	*lineSec;
	line_t		*ld;
	seg_t		*seg;
	int			i, k, specSec;

	SET_countsec(0);

	for (i = 0; i < numsubsectors; i++, ssec++)
	{
		if (countsec[ssec->sector - sectors]) continue;
		seg = &segs[ssec->firstline];
		for (k = 0; k < ssec->linecount; k++, seg++)
		{
			if ((ld = seg->linedef) != NULL
				&& ld->specType == R3D_06)
				goto checklines;
		}

		continue;

checklines:

		specSec = -1;
		seg = &segs[ssec->firstline];
		for (k = 0; k < ssec->linecount; k++, seg++)
		{
			if ((ld = seg->linedef) != NULL
				&& ld->specType == R3D_03 && ChkLptr(ld))
			{
				specSec = GetLptr(ld);
				lineSec = ld->frontsector;
				break;
			}
		}

		if (specSec < 0) continue;

		countsec[lineSec - sectors] = 1;

		for (k = 0; k < lineSec->linecount; k++)
		{
			ld = lineSec->lines[k];
			if (ld->specType == R3D_06 &&
				!ChkLptr(ld) &&
				ld->frontsector == lineSec)
				Setspec(ld, specSec, 8011);
		}
	}
}

RDBOOL subfn_CheckAllLinesConnect(sector_t *sec)
{
	line_t *ld;
	int i;

	CLR_vtxcheck();

	for ( i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (ld->specType != R3D_1E) continue;

		vtxcheck[ld->v1 - vertexes] |= 1;
		vtxcheck[ld->v2 - vertexes] |= 2;
	}

	for ( i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];
		if (ld->specType != R3D_1E) continue;

		if (vtxcheck[ld->v1 - vertexes] != 3 ||
			vtxcheck[ld->v2 - vertexes] != 3)
			return false;
	}

	return true;
}

// These lines set the floorheight UNLESS they are in a closed group
void LS_CheckR3D_1ELines(void)
{
	sector_t *sec;
	line_t *ld;
	int i, k;

	for (sec = sectors, i = 0; i < numsectors; sec++, i++)
	{
		if (!sec->tag)
		{
			for ( k = 0; k < sec->linecount; k++)
			{
				if ((ld = sec->lines[k])->specType == R3D_1E &&
					ld->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2))
				{
					if (!subfn_CheckAllLinesConnect(sec))
					{
						if (ld->mtflags & R3D_MISLTEX2)
						{
							Setr3dfh(ld->frontsector, ld->frontsector, 8013);
							Setr3dfh(ld->backsector , ld->frontsector, 8014);
						}
						else if (ld->mtflags & R3D_MISLTEX1)
						{
							Setr3dfh(ld->frontsector, ld->backsector, 8015);
							Setr3dfh(ld->backsector , ld->backsector, 8016);
						}
						break;
					}
				}
			}
		}
	}
}

//------------------------------------------------------------
//
// LS_CheckSpecialAnimSecs
//	Added rev 36 for vile map20 sec 400 and map22 sec 944 
//
//------------------------------------------------------------
void LS_CheckSpecialAnimSecs(void)
{
	sector_t	*sec;
	int			i, secnum;

	CLR_hitline();

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		line_t *ldf = NULL;
		int specSec = -1;

		if (!secdone[i] && secmarker[i] & 2)
		{
			secmarker[i] &= ~2;
		}
		else if (!secdone[i]
			&& !secmarker[i]
			&& MSFC(i, 0x0080001C)
			&& IsLiquidFloorFlat(sec)
			&& IsLiquidCeilFlat(sec)
			&& (secnum = XS_GetBoundarySecnum(sec)) >= 0)
		{
			if (MSFC(secnum, 0x0080809C)
				&& sectors[secnum].ceilingheight != sec->ceilingheight) 
			{
				Setr3dch(sec, &sectors[secnum], 8017);
				Setsecdone2(sec, 8018);
			}
			else if (!(sectors[secnum].specFlags & SF_MOVSECTOR)
				&& sectors[secnum].floorheight != sec->floorheight
				&& sectors[secnum].floorpic
				== sec->floorpic
				&& sectors[secnum].floorpic
				== sectors[secnum].ceilingpic)
			{
				Setr3dfh(sec, &sectors[secnum], 8019);
				Setsecdone2(sec, 8020);
			}
		}
	}
}

//================================================================================

int LS_BoundingBoxDiff(int out[4])
{
	return (
		(csbox[BOXLEFT ]>>16) - (out  [BOXLEFT ]>>16) +
		(csbox[BOXBASE ]>>16) - (out  [BOXBASE ]>>16) +
		(out  [BOXRIGHT]>>16) - (csbox[BOXRIGHT]>>16) +
		(out  [BOXTOP  ]>>16) - (csbox[BOXTOP  ]>>16)
		);
}

int subfn_ContainingSecnumOfPartialSec(sector_t *sec, line_t *ldc)
{
	int			i, cdiff = -1, diff;
	int			outer[4];
	sector_t	*other = sectors, *closest = NULL;

	sampleline = NULL;
	
	// Try all sectors that fit in the bounding box.
	for(i = 0; i < numsectors; i++, other++)
	{
		if (other == sec ||
			(SecONE(i) && SecONE(sec - sectors)))
			continue;

		memcpy(outer, secbbox[i].bounds, sizeof(outer));

		if (csbox[BOXLEFT ] >= outer[BOXLEFT ] &&
			csbox[BOXRIGHT] <= outer[BOXRIGHT] &&
			csbox[BOXBASE ] >= outer[BOXBASE ] &&
			csbox[BOXTOP  ] <= outer[BOXTOP  ])
		{
			diff = LS_BoundingBoxDiff(outer);
			if ((cdiff < 0 || diff <= cdiff) &&
				subfn_CheckContainingSectorLine(ldc, other))
			{
				closest = other;
				cdiff = diff;
			}
		}
	}
	if (closest)
		return closest - sectors;
	return -1;
}

int subfn_MarkConnectedLines(sector_t *sec, line_t *ldc)
{
	int i, repeat = true, count = 1;

	M_ClearBox(csbox);

	CLR_vtxcheck();
	vtxcheck[ldc->v1 - vertexes] = 1;
	vtxcheck[ldc->v2 - vertexes] = 1;
	hitline[ldc - lines] = 1;
	M_AddLineToBox(csbox, ldc);

	// Find all connected lines.
	while (repeat)
	{
		repeat = false;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			if (hitline[ld - lines]) continue;

			if (vtxcheck[ld->v1 - vertexes] ||
				vtxcheck[ld->v2 - vertexes])
			{
				vtxcheck[ld->v1 - vertexes] = 1;
				vtxcheck[ld->v2 - vertexes] = 1;
				hitline[ld - lines] = 1;
				count++;
				M_AddLineToBox(csbox, ld);
				repeat = true;
				break;
			}
		}
	}

	if (count < 4) return -1;

	// Mark lines inside marked lines
	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE ||
			hitline[ld - lines])
			continue;

		if (ld->bbox[BOXLEFT ] > csbox[BOXLEFT ] &&
			ld->bbox[BOXRIGHT] < csbox[BOXRIGHT] &&
			ld->bbox[BOXBASE ] > csbox[BOXBASE ] &&
			ld->bbox[BOXTOP  ] < csbox[BOXTOP  ])
		{
			hitline[ld - lines] = 1;
		}
	}

	return subfn_ContainingSecnumOfPartialSec(sec, ldc);
}

line_t *LS_FindOuterLine(sector_t *sec)
{
	line_t *li = NULL;
	int i, bounds[4], diff = DDMAXINT, d;

	memcpy(bounds, secbbox[sec - sectors].bounds, sizeof(bounds));

	for ( i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];
		if (!(ld->spflags & SPF_CONTROLLINE) &&
			!hitline[ld - lines])
		{
			if ((d = abs(ld->v1->x - bounds[BOXLEFT])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v1->x - bounds[BOXRIGHT])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v2->x - bounds[BOXLEFT])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v2->x - bounds[BOXRIGHT])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v1->y - bounds[BOXTOP])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v1->y - bounds[BOXBASE])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v2->y - bounds[BOXTOP])) < diff)
			{
				diff = d;
				li = ld;
			}
			if ((d = abs(ld->v2->y - bounds[BOXBASE])) < diff)
			{
				diff = d;
				li = ld;
			}
		}
	}

	return li;
}

void subfn_CheckSameR3D_00TexLinesWithCL(int secnum)
{
	sector_t *sec = &sectors[secnum];
	line_t *lc = NULL;
	int i, k, specSec;

	specSec = LS_GetContainingSecnum(sec, 1);

	// The quick way
	if (specSec >= 0)
	{
		if ((sec->specFlags & 0xC) == 0xC)
		{
			// b211-0 set subsecs directly to avoid later specSec tests
			for (i = 0; i < sec->linecount; i++)
			{
				if ( !(sec->lines[i]->spflags & SPF_CONTROLLINE) )
				{
					int		 lnum = sec->lines[i] - lines;
					ssline_t *sl = &sslines[lnum];

					for (k = 0; k < sl->num; k++)
					{
						subsector_t *ss = &subsectors[sl->list[k]];

						if (ss->floorsec < 0 && ss->ceilsec < 0)
							SetOverride(ss, specSec, 8035);
					}
					Settype(sec->lines[i], R3D_3D, 8036);
					SetLFG(sec->lines[i], LFG_NEWSET, 8036);
				}
			}
		}
		else	// set indirectly
		{
			for (i = 0; i < sec->linecount; i++)
			{
				if (!(sec->lines[i]->spflags & SPF_CONTROLLINE))
				{
					Setspec(sec->lines[i], specSec, 8035);
					Settype(sec->lines[i], R3D_3D, 8036);
				}
			}
		}
	}
	else	// The slow way (lines contained in more than one sector
	{
		for ( i = 0; i < sec->linecount; i++)
			if (!(sec->lines[i]->spflags & SPF_CONTROLLINE) &&
				(ChkLptr(sec->lines[i]) ||
				 sec->lines[i]->specType != R3D_04))
				return;

		CLR_hitline();

		for ( i = 0; i < sec->linecount; i++)
			if (!(sec->lines[i]->spflags & SPF_CONTROLLINE) &&
				(lc = subfn_FindLineCrossing(sec->lines[i], LC_NORMAL|LC_NOT_SHARED)) != NULL)
			{
				Settype(sec->lines[i], R3D_2E, 8037);
				hitline[sec->lines[i] - lines] = 2;
			}

		do
		{
			specSec = -1;
			{
				line_t *ld = LS_FindOuterLine(sec);

				if (ld != NULL)
					specSec = subfn_MarkConnectedLines(sec, ld);
			}

			for (i = 0; i < sec->linecount; i++)
				if (hitline[sec->lines[i] - lines] == 1)
				{
					if (sec->lines[i]->specType != R3D_2E)
					{
						Setspec(sec->lines[i], specSec, 8038);
						Settype(sec->lines[i], R3D_3D, 8039);
					}
					hitline[sec->lines[i] - lines] = 2;
				}
		}
		while(specSec >= 0);
	}

	Setsecdone(sec, 8040);
}

static RDBOOL subfn_PreCheckLineRings(int secnum)
{
	sector_t *sec = &sectors[secnum];
	int		 i, k, count = 0;

	if (sec->specFlags & SF_MOVSECTOR) return false;

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (Vtxs2(ld)
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& !ChkLFGSecSet(ld)
			&& !subfn_FindLineCrossing(ld, LC_NORMAL))
		{
			hitline[sec->lines[i] - lines] = 1;
			count++;
		}
	}

	if (count < 3) return false;

	count = 0;

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;
		int bbox[4], ring_count, success = false;

		if (hitline[lnum] != 1) continue;

		if (subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL))
		{
			int specSec = -1;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (hitline[ld - lines] == 4)
				{
					specSec = LS_FindAdjacentSecnum(ld, 7);
					if (specSec >= 0) break;
				}
			}
			if (specSec >= 0)
			{
				count++;

				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (hitline[ld - lines] == 4)
					{
						if (specSec != secnum) Setspec(ld, specSec, 8041);
						SetLFGSecSet(ld, 8042);
					}
				}
			}
			for (k = 0; k < sec->linecount; k++)
			{
				int n = sec->lines[k] - lines;

				if (hitline[n] == 4) hitline[n] = 8;
			}
		}
	}

	return count > 0;
}

void LS_CheckSameR3D_00TexLinesWithCL()
{
	int i;

	for (i = 0; i < numsectors; i++)
		if (!secdone[i] &&
			SecONE(i) &&			// Lines must all be in one sector
			secspflags[i] & SECSPF_MIDTEXCLOSED)	// Lines in sectors must all have mid tex both sides
		{
			if (!subfn_PreCheckLineRings(i))
				subfn_CheckSameR3D_00TexLinesWithCL(i);
		}
}

void LS_CheckR3D_06InheritanceFromR3D_03()
{
	sector_t *sec = sectors;
	int i;

	for ( i = 0; i < numsectors; i++, sec++)
		if (sec->linecount == 1 &&
			!secdone[i] &&
			sec->heightsec < 0 &&
			sec->lines[0]->specType == R3D_06)
		{
			Setspec(sec->lines[0], LS_FindAdjacentSecnum(sec->lines[0], 3), 8043);
			Setsecdone(sec, 8044);
		}
}

line_t *subfn_LookForFI_Line(sector_t *sec)
{
	int i;

	for ( i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			(!(ld->mtflags & R3D_MISHITEX) && origTEX[ld - lines] & 0x44))
			return NULL;
	}

	for ( i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->mtflags & R3D_MISHITEX &&
			ld->spflags & SPF_NOTEX &&
			Vtxs2(ld) && Vtfl3(ld))
		{
			vtxcheck[ld->v1 - vertexes] |= 1;
			vtxcheck[ld->v2 - vertexes] |= 2;
		}
	}

	for ( i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->mtflags & R3D_MISHITEX &&
			(vtxcheck[ld->v1 - vertexes] != 3 ||
			 vtxcheck[ld->v2 - vertexes] != 3) &&
			!(ld->mtflags & R3D_MISLOTEX) &&
			ld->spflags & SPF_NOTEX &&
			ld->ml_flags & ML_TWOSIDED &&
			ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec < 0 &&
			!(ld->spflags & SPF_CONTROLLINE) &&
			!LineHasMovSectors(ld) &&
			!(ld->mtflags & R3D_MISLOTEX) &&
			!(ld->mtflags & R3D_INVTEXANY) &&
			ld->frontsector->ceilingpic != skyflatnum &&
			ld->backsector->ceilingpic  != skyflatnum)
		{
			return ld;
		}
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////
//
//	LS_CheckR3D_22Lines
//  This is needed as sometimes a subsector does not have
//  the required line despite their being adjacent.  Where
//  this happens then it has to be filched from another
//  line in the chain.  The calling routine will keep trying
//  as there's no knowing at what point a connected line will
//  be set.

int subfn_FindR3D_22connected(line_t *ld)
{
	sector_t *sec = ld->frontsector;
	vertex_t *v1 = ld->v1, *v2 = ld->v2;
	int i;

	if (vtxarray[v1 - vertexes] > 1)
		for ( i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];
			if (ld->specType == R3D_22 && ChkLptr(ld) &&
				(ld->v1 == v1 || ld->v2 == v1))
				return GetLptr(ld);
		}

	if (vtxarray[v2 - vertexes] > 1)
		for ( i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];
			if (ld->specType == R3D_22 && ChkLptr(ld) &&
				(ld->v1 == v2 || ld->v2 == v2))
				return GetLptr(ld);
		}

	return -1;
}

void LS_CheckR3D_22Lines()
{
	int i, k, repeat;
	for (i = 0; i < numsectors; i++)
	{
		sector_t *sec = &sectors[i];
		int count = 0;

		if (secdone[i]) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			if (ld->specType == R3D_22)
			{
				if (CmpLptr(ld, -2)) continue;
				count++;
			}
		}

		if (!count) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];
			if (ld->specType == R3D_22 &&
				!CmpLptr(ld, -2))
			{
				Setspec(ld, LS_GetAdjacentLineSecnum(ld, 6), 8045);
			}
		}

		repeat = true, count = 0;
		while (repeat)
		{
			repeat = false;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (ld->specType == R3D_22 &&
					CmpLptr(ld, -1))
				{
					Setspec(ld, subfn_FindR3D_22connected(ld), 8046);
					if (!ChkLptr(ld))
					{
						count++;
						repeat = count < 10;
					}
				}
			}
		}
	}
}

//
// Check GLNodes routines.
//

sector_t *trace_sec;
int trace_count, trace_mask;

void subfn_TraceLine(line_t *ld)
{
	vertex_t *v1 = ld->v1, *v2 = ld->v2;
	int i;

	hitline[ld - lines] = 1;
	trace_count++;

	vtxcheck[v1 - vertexes] |= 1;
	vtxcheck[v2 - vertexes] |= 2;

	for (i = 0; i < trace_sec->linecount; i++)
	{
		ld = trace_sec->lines[i];

		if (!hitline[ld - lines] &&
			origTEX[ld - lines] == trace_mask &&
			(ld->v1 == v2 || ld->v2 == v1))
			subfn_TraceLine(ld);
	}
}

RDBOOL subfn_CheckFlatsWithR3D_1ALines(sector_t *sec)
{
	int i;
	trace_sec = sec;

	CLR_hitline();
	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];
		if (hitline[ld - lines]) continue;

		trace_count = 1;
		trace_mask = origTEX[ld - lines];
		subfn_TraceLine(ld);
		if (trace_count < 3)
				return false;
	}
	for (i = 0; i < sec->linecount; i++)
		if (!hitline[sec->lines[i] - lines] ||
			vtxcheck[sec->lines[i]->v1 - vertexes] != 3 ||
			vtxcheck[sec->lines[i]->v2 - vertexes] != 3)
			return false;
	return true;
}


void LS_CheckFlatsWithR3D_1ALoops()
{
	sector_t *sec;
	int i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		if (!secdone[i] &&
			!(sec->specFlags & SF_MOVSECTOR) &&
			!IsAnimFlat(sec->floorpic))
		{
			int f1, f2, k, valid = true, mask1 = -1, mask2 = -1;
			sector_t *ihs = NULL, *cs = NULL;
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				if (!(ld->ml_flags & ML_TWOSIDED) ||
					!(ld->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2)) ||
					vtxarray[ld->v1 - vertexes] != 2 ||
					vtxarray[ld->v2 - vertexes] != 2)
				{ valid = false; break; }
				if (ld->specType == R3D_1A)
				{
					if (mask1 == -1) mask1 = ld->mtflags;
					if (ld->mtflags != mask1)
					{ valid = false; break; }
					f1 = mask1 & R3D_MISLTEX1 ?
						ld->backsector->floorheight : ld->frontsector->floorheight;
				}
				else
				{
					if (mask2 == -1) mask2 = ld->mtflags;
					if (ld->mtflags != mask2)
					{ valid = false; break; }
					cs = mask2 & R3D_MISLTEX1 ? ld->backsector : ld->frontsector;
					if (!ihs)
					{
						ihs = cs;
						f2 = ihs->floorheight;
					}
					if (!cs || ihs != cs || ihs == sec)
					{ valid = false; break; }
				}
			}

			if (valid &&
				mask1 != -1 &&
				mask2 != -1 &&
				f2 >= f1 &&
				ihs &&
				subfn_CheckFlatsWithR3D_1ALines(sec))
			{
				for (k = 0; k < sec->linecount; k++)
				{
					if (sec->lines[k]->specType == R3D_1A)
					{
						Settype(sec->lines[k], R3D_31, 8059);
						Setspec(sec->lines[k], ihs - sectors, 8060);
					}
				}

				Setsecdone(sec, 8061);
			}
		}
}


void subfn_SetGeneralStepsArray()
{
	int i;
	line_t *ld = lines;

	CLR_hitline();

	// Set up array of general step line types
	for (i = 0; i < numlines; i++, ld++)
	{
		if (ld->ml_flags & ML_TWOSIDED &&
			ld->spflags & SPF_NOTEX &&
			LSS(ld) &&
			!(ld->frontsector->specFlags & SF_MOVSECTOR))
			hitline[i] = 1;
	}
}

static void set1sLineHeights(line_t *ld, sector_t *s)
{
	sector_t *needed;

	if (   s->specFlags & SF_MOVCEIL
		&& (needed = UT_FindLowestCeilingSector(s)))
	{
		int	i;

		for (i = 0; i < s->linecount; i++)
		{
			if ( !(s->lines[i]->ml_flags & ML_TWOSIDED) )
			{
				if (   s->lines[i]->v1 == ld->v1
					|| s->lines[i]->v1 == ld->v2
					|| s->lines[i]->v2 == ld->v1
					|| s->lines[i]->v2 == ld->v2)
				{
					s->lines[i]->adjacentsec = needed - sectors;
				}
			}
		}
	}
}

void LS_CheckMaskedDoors()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	short		tnum;
	int			i, tx;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& !(lineflags[i] & LFG_3DTYPE))
		{
			if ( (tx = origTEX[i]) & 0x22 )
			{
				// check if back sector needs to be drawn if a masked door is closed
				if (ld->specType == R3D_1D || ld->specType == R3D_2D || ld->specType == R3D_16)
				{
					if (   (bs = ld->backsector)->specFlags & SF_MOVSECTOR
						&& (tnum = sides[ld->sidenum[0]].midtexture) > 0
						&& (ld->spflags & SPF_MIDTRANSLUCENT || texArray[tnum]->txf_flags & TXF_MASKED))
					{
						bs->specFlags |= SF_MASKEDDOOR;
						set1sLineHeights(ld, bs);
					}
					else if ( (fs = ld->frontsector)->specFlags & SF_MOVSECTOR
						&& (tnum = sides[ld->sidenum[1]].midtexture) > 0
						&& (ld->spflags & SPF_MIDTRANSLUCENT || texArray[tnum]->txf_flags & TXF_MASKED) )
					{
						fs->specFlags |= SF_MASKEDDOOR;
						set1sLineHeights(ld, fs);
					}
				}
			}
			// v220-5 revised
			else if (tx & 0x44)	// top texture no mid texture
			{
				// Rare. Support for masked door using the toptexture.
				if (  (fs = ld->frontsector)->heightsec >= 0
					&& fs->specFlags & SF_MOVCEIL
					&& !(bs->specFlags & SF_MOVSECTOR)
					&& (tnum = sides[ld->sidenum[1]].toptexture) > 0
					&& texArray[tnum]->txf_flags & TXF_MASKED)
				{
					fs->specFlags |= SF_MASKEDDOOR;
					set1sLineHeights(ld, fs);
				}
				else if ( (bs = ld->backsector)->heightsec >= 0
					&& bs->specFlags & SF_MOVCEIL
					&& !(fs->specFlags & SF_MOVSECTOR)
					&& (tnum = sides[ld->sidenum[0]].toptexture) > 0
					&& texArray[tnum]->txf_flags & TXF_MASKED)
				{
					bs->specFlags |= SF_MASKEDDOOR;
					set1sLineHeights(ld, bs);
				}
			}
		}
	}
}

// Check if type R3D_03 for rendering purposes (follows analysis)
void LS_CheckLineType()
{
	line_t *ld = lines;
	int i;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (!(ld->ml_flags & ML_TWOSIDED))
			continue;

		if (ld->frontsector == ld->backsector &&
			ld->specType != R3D_03 &&
			ld->specType < R3D_40 &&
			((sides[ld->sidenum[0]].texFlags == T_MID
			&& !sides[ld->sidenum[1]].texFlags) ||
			 (sides[ld->sidenum[1]].texFlags == T_MID
			 && !sides[ld->sidenum[0]].texFlags)))
		{
			Settype(ld, R3D_03, 8062);
		}

		if (ld->specType == R3D_00 &&
			!LS_CheckMissing(ld) &&	// added v220-4
			(origTEX[i] & 0x22) == 0x22 &&
			texArray[sides[ld->sidenum[0]].midtexture]->txf_flags & TXF_MASKED &&
			ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec < 0)
		{
			Settype(ld, R3D_03, 8063);
		}
	}
}

void subfn_SetOVS(line_t *ld, int sid)
{
	subsector_t *ss = subsectors;
	sector_t *s1 = sid ? ld->backsector  : ld->frontsector;
	float v1[2];
	float v2[2];
	int i = 0, k;

	v1[VX] = FIX2FLT(ld->v1->x);
	v1[VY] = FIX2FLT(ld->v1->y);
	v2[VX] = FIX2FLT(ld->v2->x);
	v2[VY] = FIX2FLT(ld->v2->y);

	for (; i < numsubsectors; i++, ss++)
	{
		if (ss->sector != s1)
		{
			seg_t *sg = &segs[ss->firstline];
			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef != NULL && sg->linedef == ld)
				{
					if (sid &&
						subfn_SegFacingMidpoint(v2, v1, ss->midpoint.x, ss->midpoint.y))
					{
						SetOverride(ss, s1 - sectors, 8064);
						break;
					}
				}
			}
		}
	}
}

void subfn_CheckCoincidentLineArray(int lnum1, int lnum2, int sn1, int sn2)
{
	line_t *l1 = &lines[lnum1];
	line_t *l2 = &lines[lnum2];
	sector_t *s1, *s2;
	int h1, h2;

	if (!l1->backsector || !l2->backsector) return;

	s1 = sn1 ? l1->backsector : l1->frontsector;
	s2 = sn2 ? l2->backsector : l2->frontsector;
	h1 = s1->ceilingheight - s1->floorheight + sides[l1->sidenum[sn1]].rowoffset;
	h2 = s2->ceilingheight - s2->floorheight + sides[l2->sidenum[sn2]].rowoffset;

	if (h1 == h2) return;

	if (h1 > h2)
	{
		l2->spflags |= SPF_LINEONLINE;

		if (!(origTEX[l1 - lines] & 0x11) &&
			l1->frontsector->floorheight - l1->backsector->floorheight >= 8*FRACUNIT)
			subfn_SetOVS(l1, 0);
	}
	else
	{
		l1->spflags |= SPF_LINEONLINE;

		if (!(origTEX[l1 - lines] & 0x11) &&
			l1->backsector->floorheight - l1->frontsector->floorheight >= 8*FRACUNIT)
			subfn_SetOVS(l1, 1);
	}
}

static RDBOOL chkCoincidence(line_t *lc, line_t *ld)
{
	return ld->bbox[0] == lc->bbox[0]
		&& ld->bbox[1] == lc->bbox[1]
		&& ld->bbox[2] == lc->bbox[2]
		&& ld->bbox[3] == lc->bbox[3];
}

static void setSSECovr(line_t *ld, int snum)
{
	ssline_t	*sl = &sslines[ld - lines];
	int			i;

	ss_validcount++;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (!(ss->ssflags & SSF_SECTSET) && ss->ssvalidcount != ss_validcount)
		{
			float v1[2], v2[2];

			v1[VX] = FIX2FLT(ld->v1->x);
			v1[VY] = FIX2FLT(ld->v1->y);
			v2[VX] = FIX2FLT(ld->v2->x);
			v2[VY] = FIX2FLT(ld->v2->y);

			// Must be facing back of line
			if (subfn_SegFacingMidpoint(v2, v1, ss->midpoint.x, ss->midpoint.y))
			{
				SetOverride(ss, snum, 8074);
			}

			ss->ssvalidcount = ss_validcount;
		}
	}
}

void LS_CheckAllCoincidentLines()
{
	line_t		*ld = lines, *lc;
	int			i, k;
	byte		txi, txk;

	SET_linearray(-1);

	for (i = 0; i < numlines; i++, ld++)
	{
		if (ld->spflags & SPF_LOL_MASK) continue;

		if ((lc = subfn_FindLineOnLine(ld, 0)) == NULL) continue;

		// Remove any invalid references  This is only a problem with lines
		// having different front/back sectors otherwise specSec's remain valid.

		// NOTE has to be done here as LS_CheckAllCoincidentLines cannot
		// be located any further back - this may make a SSF_SECTSET
		// setting superfluous but otherwise is benign.

		if (ld->specType == R3D_16)
		{
			sector_t *fs = ld->frontsector;
			sector_t *bs = ld->backsector;

			if (!origTEX[i])
				Forcetype(ld, R3D_01, 8069);

			// We may need to force the ceiling sector
			// as the nodes will be buggered
			if (fs->ceilingheight == bs->ceilingheight)
			{
				sector_t *sec = fs->specFlags == 0x1C ? fs : bs;

				if ((sec->specFlags & 0xFFFFC) != 0x1C) continue;

				for (k = 0; k < sec->linecount; k++)
				{
					int			m, lnum = sec->lines[k] - lines;
					ssline_t	*sl = &sslines[lnum];
					int			side = sec->lines[k]->frontsector == sec ? 0 : 1;

					for (m = 0; m < sl->num; m++)
					{
						subsector_t	*ss = &subsectors[sl->list[m]];

						if (!(ss->ssflags & SSF_CSECFIXED))
						{
							seg_t		*sg = &segs[ss->firstline];
							int			x = (int)(ss->midpoint.x*FRACUNIT);
							int			y = (int)(ss->midpoint.y*FRACUNIT);

							if (P_PointOnLineSide(x, y, sec->lines[k]) == side)
							{
								Setr3dsscs(ss, sec-sectors, 46017);
								ss->ssflags |= SSF_CSECFIXED;
							}
						}
					}
				}
			}

			continue;
		}

		if (!(ld->ml_flags & ML_TWOSIDED) || !(lc->ml_flags & ML_TWOSIDED))
		{
			line_t *lt = ld->ml_flags & ML_TWOSIDED ? ld : lc;

			Setspec(lc, -1, 8070);
			Setspec(ld, -1, 8071);

			if (!(lt->ml_flags & ML_TWOSIDED) || origTEX[lt - lines] != 0x22)
				continue;

			lt->spflags |= SPF_LINEONLINE;

			continue;
		}

		k = lc - lines;
		txi = origTEX[i], txk = origTEX[k];
		
		if (   (txi == 0x22 && (txk == 0x01 || txk == 0x04))
			|| (txk == 0x22 && (txi == 0x01 || txi == 0x04)) )
		{
			ld->spflags |= SPF_LINEONLINE;
			continue;
		}
		
		if (   (ChkLptr(lc) || ChkLptr(ld))
			&& (ld->frontsector != ld->backsector
			||  lc->frontsector != lc->backsector))
		{
			// Must have same vertex values
			if (chkCoincidence(lc, ld))
			{
				if (   ld->frontsector != ld->backsector
					&& ld->frontsector->floorpic == ld->backsector->floorpic)
				{
					// Do nothing
				}
				else if (lc->frontsector != lc->backsector
					&& lc->frontsector->floorpic == lc->backsector->floorpic)
				{
					// Do nothing
				}
				else
				{
					Setspec(lc, -1, 8072);
					Setspec(ld, -1, 8073);
				}
			}
		}

		if (txi & 0x33 && txk & 0x33  && !(txi == 0x22 && txk == 0x22))
		{
			if (txi == 0x02 && txk == 0x02)
				subfn_CheckCoincidentLineArray(i, k, 0, 0);
			else
			if (txi == 0x02 && txk == 0x20)
				subfn_CheckCoincidentLineArray(i, k, 0, 1);
			else
			if (txi == 0x20 && txk == 0x02)
				subfn_CheckCoincidentLineArray(i, k, 1, 0);
			else
			if (txi == 0x20 && txk == 0x20)
				subfn_CheckCoincidentLineArray(i, k, 1, 1);
			else
			if (txi == 0x23 && txk == 0x22)
				lines[i].spflags |= SPF_LINEONLINE;
			else if (txi == 0x02 && !(txk == 0x06))
			{
				ld->spflags |= SPF_LINEONLINE2;

				if ( !(txk & 0x22) )
				{
					setSSECovr(ld, lc->frontsector - sectors);
				}
			}
		}
	}
}

void LS_CheckIfLineOnLineNeedsProjection()
{
	subsector_t *ss;
	line_t *ld, *li;
	int i, k, m;
	
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->ml_flags & ML_TWOSIDED &&
			ld->specType < R3D_40 &&
			(origTEX[i] == 0x02 || origTEX[i] == 0x20) &&
			ld->frontsector == ld->backsector &&
			Vtxs1(ld) &&
			(li = subfn_FindLineOnLine(ld, 2)) != NULL)
		{
			Setspec(ld, -1, 8075);
			ld->spflags |= SPF_LINEONLINE;

			// If subsec->sector is wrong then any spawned heights
			// will also be wrong and will need adjusting.
			for (k = 1, ss = subsectors; k < numsubsectors; k++, ss++)
			{
				sector_t *ns = NULL;
				if (ss->sector == ld->frontsector)
				{
					seg_t *seg = &segs[ss->firstline];
					for (m = 0; m < ss->linecount; m++, seg++)
					{
						if (seg->linedef == li &&
							subfn_SegFacingMidpoint(seg->fv1, seg->fv2,
							ss->midpoint.x, ss->midpoint.y))
						{
							ns = seg->frontsector;
							SetOverride(ss, ns - sectors, 8076);
							ma_inf.projecterr2 += 1;
							break;
						}
					}
					if (ns)
					{
						mobj_t *it;

						for (it = ns->thinglist; it; it = it->snext)
						{
							if (it->thinker.function == (void (__cdecl *)(void))P_MobjThinker
								&& !(it->flags2 & MF2_DONTDRAW)
								&& !(it->flags & MF_NOBLOCKMAP)
								&& !(it->flags & MF_SPAWNCEILING))
							{
								it->z = it->dropoffz = it->floorz = ns->floorheight;
								it->ceilingz = ns->ceilingheight;
								ma_inf.projecterr1 += 1;
							}
						}
					}
				}
			}
		}
	}
}

int subfn_CheckSLAS(sector_t *sec)
{
	int i, other = -1;

	if (sec->special != 16) return -1;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (ld->backsector &&
			ld->frontsector != ld->backsector)
		{
			if (origTEX[ld - lines])
			{
				return -1;
			}

			if (   ld->frontsector != sec
				&& !(ld->frontsector->specFlags & SF_NOLOTEX)
				&& !(ld->frontsector->specFlags & SF_DUALFMOVSEC))
			{
				return -1;
			}

			if (   ld->backsector != sec
				&& !(ld->backsector->specFlags & SF_NOLOTEX)
				&& !(ld->backsector->specFlags & SF_DUALFMOVSEC))
			{
				return -1;
			}

			if (ld->frontsector != sec &&
				IsLiquidFloorFlat(ld->frontsector) &&
				ld->frontsector->floorpic != sec->floorpic)
			{
				if (other == -1) other = ld->frontsector - sectors;
				if (other != ld->frontsector - sectors)
					return -1;
			}
			if (ld->backsector != sec &&
				IsLiquidFloorFlat(ld->backsector) &&
				ld->backsector->floorpic != sec->floorpic)
			{
				if (other == -1) other = ld->backsector - sectors;
				if (other != ld->backsector - sectors)
					return -1;
			}
		}
	}
	return other;
}

// Fix floorheight for moving secs with liquid flats under
// special circumstances.
void LS_CheckSpecialLinkedAnimSectors(void)
{
	sector_t *sec;
	int i, k, m, secnum;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& SecCtrl(i)
			&& IsLiquidFloorFlat(sec)
			&& (sec->specFlags & 0x400FFFFC) == 0x4009818C
			&& (secnum = subfn_CheckSLAS(sec)) >= 0
			&& sectors[secnum].special == sec->special)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
				{
					ssline_t *sl = &sslines[sec->lines[k] - lines];

					SetLFG(sec->lines[k], LFG_NEWSET, 990);
					
					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->sector == sec)
						{
							SetOverride(ss, secnum, 990);
						}
					}
				}
			}
			Setsecdone(sec, 8079);
			Setsecdone(&sectors[secnum], 8080);
		}
	}
}

// Check for moving floors in liquid moving floors
void LS_CheckMovLiqFloorInMovFloor()
{
	sector_t *sec;
	int		 i, k, m;
	int		 mask = SF_DUALFMOVSEC|SF_MOVFLOOR|SF_SECCLOSED|SF_NOHITEX;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& SecCtrl(i)
			&& sec->ceilingpic == skyflatnum
			&& (sec->specFlags & mask) == mask)
		{
			sector_t	*s = NULL;
			int			bbox[4];
			int			fh = sec->floorheight;

			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (ChkLFGSecSet(ld))
				{
					s = NULL;
					break;
				}

				hitline[ld - lines] = 1;

				if (ld->frontsector->specFlags & SF_MOVFLOOR
					&& IsLiquidFloorFlat(ld->frontsector)
					&& !SecCtrl(ld->frontsector - sectors)
					&& ld->frontsector->floorheight > fh)
				{
					s  = ld->frontsector;
					fh = s->floorheight;
				}
				if (ld->backsector
					&& ld->backsector->specFlags & SF_MOVFLOOR
					&& !SecCtrl(ld->backsector - sectors)
					&& IsLiquidFloorFlat(ld->backsector)
					&& ld->backsector->floorheight > fh)
				{
					s  = ld->backsector;
					fh = s->floorheight;
				}
			}
			if (!s) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				int ring_count;
				int lnum = sec->lines[k] - lines;

				if (hitline[lnum] == 1 &&
					subfn_CheckForRing(hitline, sec, lnum, &ring_count, bbox, NULL, NULL))
				{
					int secnum = s - sectors;

					for (m = 0; m < sec->linecount; m++)
					{
						line_t *ld = sec->lines[m];

						if (   hitline[ld - lines] == 4
							&& LineVertexOwnsSec(ld, secnum))
						{
							ring_count--;
						}
					}

					if (   !ring_count
						&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE)
						// Check that the floor inheriting the liquid floor
						// sector is entirely contained in the loop.
						&& !HitssecMidpointOutsideBox(bbox))
					{
						for (m = 0; m < hitssecptr; m++)
							Setr3dsslfsec(&subsectors[hitssec[m]], secnum, 8081);
					}

					for (m = 0; m < sec->linecount; m++)
					{
						line_t *ld = sec->lines[m];

						if (hitline[ld - lines] == 4)
							hitline[ld - lines] = 8;
					}

					if (!ring_count)
						Setsecdone3(sec, 8082);
				}
			}
		}
	}
}

