//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 Lee Killough
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_mapGeneral.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 10 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: initialises some arrays
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "sys_system.h"
#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "r3d_edit.h"
#include "r3d_compat.h"

//==========================================

void LS_SetSameSecInGroupFlag(void)
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		int k, gnum = -1;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t *ld = sec->lines[k];

			if (ld->frontsector == ld->backsector)
			{
				gnum = lgrpnum[ld - lines];
				break;
			}
		}

		if (gnum < 0) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t	*ld = sec->lines[k];
			int		chknum = lgrpnum[ld - lines];

			if (ld->frontsector == ld->backsector)
			{
				if (chknum != gnum)
				{
					gnum = -1;
					break;
				}
			}
			else
			{
				if (chknum == gnum)
				{
					gnum = -1;
					break;
				}
			}
		}
		if (gnum > 0)
		{
			secspflags[i] |= SECSPF_SAMESECLINES;
		}
	}
}

static RDBOOL subfn_CheckDoubleLoopHi(int lnum, sector_t *fs, sector_t *bs)
{
	sector_t	*sc, *ch, *os;
	line_t		*ld;
	int			a, c, i, rc, n;

	CLR_hitline();

	for (a = n = 0; a < 2; a++)
	{
		sc = a ? bs : fs;

		for (c = i = 0, os = NULL; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (ld->spflags & SPF_CONTROLLINE) continue;
			if ( !(ld->mtflags & (R3D_MISHTEX1|R3D_MISHTEX2)) ) continue;

			if (origTEX[ld - lines])
			{
				hitline[ld - lines] = 1; c++;
				continue;
			}

			ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
			if (ch->ceilingheight <= sc->ceilingheight) return false;
			if (ch->specFlags & SF_MOVCEIL) return false;
			if (ch->ceilingpic == skyflatnum) return false;

			if (!os)
			{
				os = ch;
			}
			else if (os != ch && os->ceilingheight != ch->ceilingheight)
			{
				return false;
			}

			hitline[ld - lines] = true; c++;
		}

		if (c < 3 || !os) return false;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL) && rc > 2)
		{
			for (i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					if (!origTEX[ld - lines])
					{
						if (ld - lines != lnum) linearray[n++] = ld - lines;
					}

					hitline[ld - lines] = 0;
				}
			}
		}
		else
		{
			return false;
		}
	}

	// This structure has two fixed ceilings
	subsector_t	*ss;
	ssline_t	*sl;
	seg_t		*sg;

	validcount++;

	for (a = 0; a < n; a++)
	{
		sl = &sslines[linearray[a]];

		for (c = 0; c < sl->num; c++)
		{
			ss = &subsectors[sl->list[c]];

			if (ss->ceilsec < 0)
			{
				ss->ceilsec = os - sectors;
				ss->lcsec = ss->sector - sectors;
			}

			for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
			{
				if ( !(ld = sg->linedef) ) continue;
				if (ld->validcount == validcount) continue;
				ld->validcount = validcount;

				if (!origTEX[ld - lines])
					Settype(ld, R3D_16, 21001);
				else
					Settype(ld, R3D_36, 21001);

				SetLFG(ld, LFG_NEWSET|LFG_FIXABS|LFG_FIXED, 21001);
			}
		}
	}

	Setsecdone(fs, 21001);
	Setsecdone(bs, 21001);

	return true;
}

// v220-10 Look for missing tex lines spanning 1's lines
void LS_SetLinesSpanning1sLines()
{
	sector_t	*fs, *bs;
	int			i, k;

	for (i = 0; i < numlines; i++)
	{
		if ( !(lines[i].ml_flags & ML_TWOSIDED) ) continue;

		// v220-17 add test
		// e.g btsx_e1 map12 L4447
		if (   lines[i].mtflags & (R3D_MISHTEX1|R3D_MISHTEX2)
			&& !(lines[i].mtflags & (R3D_MISLTEX1|R3D_MISLTEX2))
			&& (fs = lines[i].frontsector)->specFlags & SF_NOHITEX
			&& (bs = lines[i].backsector)->specFlags & SF_NOHITEX
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& !(bs->specFlags & SF_MOVSECTOR)
			&& SecCtrl(fs - sectors)
			&& SecCtrl(bs - sectors))
		{
			subfn_CheckDoubleLoopHi(i, fs, bs);
		}
		else
		if (   lines[i].lnflags2 & (LNF2_OWNBOTHGTRTWO|LNF2_OWNTWOANDPLUS)
			&& !(origTEX[i] & 0x11)
			&& !(lines[i].spflags & SPF_CONTROLLINE)
			&& (fs = lines[i].frontsector) != (bs = lines[i].backsector)
			&& fs->floorheight != bs->floorheight)
		{
			if (lines[i].backsector->floorheight > lines[i].frontsector->floorheight)
				fs = bs, bs = lines[i].frontsector;

			if ( (fs->specFlags & SF_MOVSECTOR) && !(bs->specFlags & SF_MOVSECTOR) )
			{
				RDBOOL	sd = fs == lines[i].frontsector ? 0 : 1;
				byte	mtx = sd ? 0x20 : 0x02; 
				byte	htx = sd ? 0x40 : 0x04;
				RDBOOL	t1 = origTEX[i] == mtx || origTEX[i] == 0x22;
				RDBOOL	t2 = origTEX[i] == htx || origTEX[i] == 0x44;
				int		c;

				// v220-13 ensure the sector that drops has missing lowtex lines
				for (c = k = 0; k < fs->linecount; k++)
				{
					line_t *ld = fs->lines[k];

					if (ld - lines == i) continue;

					if (   ld->ml_flags & ML_TWOSIDED
						&& !(ld->spflags & SPF_CONTROLLINE)
						&& ld->mtflags & R3D_MISLOTEX)
					{
						c++; break;
					}
				}

				if (!c) continue;

				for (c = k = 0; k < bs->linecount; k++)
				{
					line_t *ld = bs->lines[k];

					if (ld - lines == i) continue;
					if (ld->ml_flags & ML_TWOSIDED) { c = 0; break; }

					linearray[c++] = ld - lines;
				}

				if (c)
				{
					if ( !(lines[i].lnflags2 & LNF2_OWNBOTHGTRTWO) )
					{
						if (!t1 && !t2)	// Usually a map error unspotted with software rendering
						{
							lines[i].scpFlags |= SCPF_DRAWTOLIMITS;	// floor tex will be used
						}
					}
					else
					{
						lines[i].scpFlags |= SCPF_DRAWTOLIMITS;

						if (t2)	// If no midtex use existing line toptexture
						{
							sides[lines[i].sidenum[sd]].midtexture = sides[lines[i].sidenum[sd]].toptexture;
							origTEX[i] |= (byte)(sd ? 0x02 : 0x20);
						}
					}
				}
			}
		}
	}
}

void R3D_1AMissingTex()
{
	line_t	*ld, *ls;
	side_t	*sdld, *sdls;
	int		i, snum;
	byte	tchk;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_1A &&
			origTEX[i] & 0x11 &&
			ld->mtflags & (R3D_MISHTEX1|R3D_MISHTEX2) &&
			!(ld->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2|R3D_INVTEXANY)) &&
			Vtxs2(ld) &&
			vtxflags[ld->v1 - vertexes] != 3 &&
			vtxflags[ld->v2 - vertexes] != 3 &&
			(ls = subfn_FindLineConnected(ld, 6)) != NULL)
		{
			snum = (ld->mtflags & R3D_MISHTEX1) ? 0 : 1;
			tchk = origTEX[i] | (snum ? 64 : 4);
			sdls = &sides[ls->sidenum[snum]];

			if (sdls->toptexture && tchk == origTEX[ls - lines])
			{
				sdld = &sides[ld->sidenum[snum]];
				sdld->toptexture = sdls->toptexture;
				sdld->rowoffset = sdls->rowoffset;
				sdld->textureoffset = sdls->textureoffset;
				ld->spflags = ls->spflags;
				ld->mtflags = ls->mtflags;
				Settype(ld, ls->specType, 21002);
				ma_inf.misstexfixed++;
				if (texerror)
					pr_OFinsert("ERROR: Missing tex line %i \n",i);
			}
		}
	}
}

void R3D_1CMissingTex()
{
	line_t	*ld, *ls;
	side_t	*sdld, *sdls;
	int		i, snum, tyoff, offset, sign;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType == R3D_1C &&
			Vtfl3(ld) &&
			(ls = subfn_FindLineConnected(ld, 5)) != NULL)
		{
			snum = (origTEX[ls - lines] & 0x50) ? 1 : 0;

			sdld = &sides[ld->sidenum[snum]];
			sdls = &sides[ls->sidenum[snum]];

			sdld->toptexture = sdls->toptexture;
			sdld->bottomtexture = sdls->bottomtexture;
			sdld->rowoffset = sdls->rowoffset;
			// Texoff can't be calculated so a general rule must be
			// applied (in case it's got some daft value).  In the
			// case of line 3591 in 10sector map26 the texoff value
			// is correct despite no texture being present.
			tyoff = abs(sdld->textureoffset);
			sign  = sdld->textureoffset >= 0 ? 1 : -1;
			while (tyoff > 64*FRACUNIT) tyoff -= 64*FRACUNIT;
			offset = (tyoff % (8*FRACUNIT));
			ma_inf.misstexfixed++;
			if (texerror)
				pr_OFinsert("%s%5i - missing tex\n", texerrfix, i);
			if (offset) sdld->textureoffset = 8*offset*sign;
			sdld->texFlags = sdls->texFlags;
			ld->spflags = ls->spflags;
			ld->mtflags = ls->mtflags;
			Settype(ld, R3D_1B, 21003);
		}
	}
}

// Check for sectors where a lower tex can't be rendered
// but where one line backs a faked sector
void LS_CheckMapErrors()
{
	R3D_1AMissingTex();
	R3D_1CMissingTex();
}


int LS_ClearEmptySectorSpecials()
{
	int i, count = 0;

	for (i = 0; i < numsectors; i++)
	{
		if (!sectors[i].linecount)
		{
			count++;

			// Ensure errant specials don't get processed
			sectors[i].special = 0, sectors[i].tag = 0;
		}
	}
	return count;
}

// killough 1/30/98: Create xref tables for tags
// Hash the sector tags across the sectors and linedefs.
void LS_InitTagLists(void)
{
	register int i;

	for (i=numsectors; --i>=0; )		// Initially make all slots empty.
		sectors[i].firsttag = -1;
	for (i=numsectors; --i>=0; )		// Proceed from last to first sector
	{									// so that lower sectors appear first
		int j = (unsigned) sectors[i].tag % (unsigned) numsectors; // Hash func
		sectors[i].nexttag = sectors[j].firsttag;   // Prepend sector to chain
		sectors[j].firsttag = i;
	}

	// killough 4/17/98: same thing, only for linedefs

	for (i=numlines; --i>=0; )			// Initially make all slots empty.
		lines[i].firsttag = -1;
	for (i=numlines; --i>=0; )			// Proceed from last to first linedef
	{									// so that lower linedefs appear first
		int j = (unsigned) lines[i].tag % (unsigned) numlines; // Hash func
		lines[i].nexttag = lines[j].firsttag;   // Prepend linedef to chain
		lines[j].firsttag = i;
	}
}

void LS_Process3DLines(int analyse)
{
extern int flatListsCreated;

	flatListsCreated = 0;

	// Check for stacked sectors and append list of flats heights
	if (   texname3D
		&& (ma_inf.num3Dlines = LS_IdentifyR3D_40Lines(analyse)) > 0)
	{
		if (dbfile_active)
			Sys_Quit(false);

		LS_SetSSFlats3DArray();
		LS_Check3DScaledTexs();
	}
}

static void setLineViewGroupOwners(int* lvg)
{
	subsector_t	*ss;
	seg_t		*sg;
	int			i, k;
	
	if (!ma_inf.skykeepcount) return;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef)
			{
				lvg[sg->linedef - lines] = ss->viewgroup;
			}
		}
	}
}

// v220-35 supports subsector view groups to aid multiple 271/272 sky
//         selection used with the R3D_preseVars keyword fixskykeep
void LS_SetLineTaggedSectors2()
{
	int		i, k, m, ltag, s, nsp, sp, vg[16];
	int		vgcount = 0;
	int		*lvg = NULL;

	if (ma_inf.skykeepcount)
	{
		 lvg = (int*)calloc(numlines, sizeof(int));
		 setLineViewGroupOwners(lvg);
	}

	for (i = 0; i < numlines; i++)
	{
		sp = lines[i].special;

		if (sp != 271 && sp != 272) continue;

		lines[i].special = 0;

		// killough 10/98:Modified for scrolling support v220-35
		//
		// Support for sky textures being transferred from sidedefs.
		// Allows scrolling and other effects (but if scrolling is
		// used, then the same sector tag needs to be used for the
		// sky sector, the sky-transfer linedef, and the scroll-effect
		// linedef). Still requires user to use F_SKY1 for the floor
		// or ceiling texture, to distinguish floor and ceiling sky.

		ltag = lines[i].tag;

		if (ma_inf.noskychange)
		{
			int a, c;

			vgcount = 0;

			if ( (a = ma_inf.skykeepcount) )
			{
				for (c = 0; c < a; c++)
				{
					if (sides[lines[i].sidenum[0]].toptexture == ma_inf.skykeep[c][1])
					{
						vg[vgcount++] = ma_inf.skykeep[c][0];
					}
				}

				if (!vgcount) continue;	// nothing matched
			}
		}

		int matched = 0;

		for (s = -1; (s = P_FindSectorFromTag(ltag, s)) >= 0;)
		{
			if (sectors[s].ceilingpic != skyflatnum && sectors[s].floorpic != skyflatnum)
			{
				if (ge_inf.showskytagerrors)
					pr_OFinsert("special %i: S%i tag %i used with non-sky ceiling and floor and has been ignored\n",
					lines[i].special, s, ltag);
				continue;
			}

			if (!ma_inf.noskychange) { m = 0; goto setsec; }

			for (m = 0; m < vgcount; m++)
			{
				for (k = 0; k < numsubsectors; k++)
				{
					if (   subsectors[k].sector - sectors == s
						&& subsectors[k].viewgroup == vg[m])
					{
						matched++;
						goto setsec;
					}
				}

				continue;
setsec:;
				sectors[s].sidesky  = sides[lines[i].sidenum[0]].toptexture | PL_SKYFLAT;
				sectors[s].sky_offy = sides[lines[i].sidenum[0]].rowoffset;

				if (sectors[s].sky_offy) ma_inf.checkSkyOffy = true;	// v220-35 map fix: else skies can jump about if not all set

				if (sp == 272) sectors[s].specFlags2 |= SF2_HASTYPE272;

				// check if tagged to a scroller
				if (ltag && skl[s].sky_line < 0)
				{
					for (int n = -1; (n = P_FindLineFromLineTag(&lines[i],n)) >= 0;)
					{
						if ((nsp = lines[n].special) == 218 || nsp == 249 ||  nsp == 254)	// Check for a wall scorller
						{
							skl[s].sky_line = i; break;
						}
					}
				}
			}
		}

		if (ma_inf.noskychange && !matched)
		{
			for (s = -1; (s = P_FindSectorFromTag(ltag, s)) >= 0;)
			{
				if (sectors[s].ceilingpic != skyflatnum && sectors[s].floorpic != skyflatnum)
					continue;

				// check if tagged to a scroller
				if (ltag && skl[s].sky_line < 0)
				{
					for (int n = -1; (n = P_FindLineFromLineTag(&lines[i],n)) >= 0;)
					{
						if ((nsp = lines[n].special) == 218 || nsp == 249 ||  nsp == 254)	// Check for a wall scorller
						{
							skl[s].sky_line = i; break;
						}
					}
				}
			}
		}
	}

	if (lvg) free(lvg);
}

// v220-31 lines checked before running type 270 to ensure inheritance
//         Map error check and report for unset tags
void LS_SetLineTaggedSectors()
{
	int		i, ltag, s, secidx;
	int		has3D270 = -1;
	RDBOOL	error = false;

	for (i = 0; i < numlines; i++)
	{
		ltag = lines[i].tag;

		switch (lines[i].special)
		{
		// killough 3/7/98:
		// support for drawn heights coming from different sector
		case 242:
			if (!ltag) { error = true; break; }

			secidx = sides[*lines[i].sidenum].sector - sectors;

			for (s = -1; (s = P_FindSectorFromTag(lines[i].tag,s)) >= 0;)
			{
				sectors[s].heightsec = secidx;

				// v220-33 surrogate heightsec code removed
			}
			break;

		// killough 3/16/98:
		// Add support for setting floor lighting independently (e.g. lava)
		case 213:
			if (!ltag) { error = true; break; }

			secidx = sides[*lines[i].sidenum].sector - sectors;

			for (s = -1; (s = P_FindSectorFromTag(ltag, s)) >= 0;)
				sectors[s].floorlightsec = secidx;
			break;
			
		// killough 4/11/98: Add support for setting
		// ceiling lighting independently
		case 261:
			if (!ltag) { error = true; break; }

			secidx = sides[*lines[i].sidenum].sector - sectors;

			for (s = -1; (s = P_FindSectorFromTag(ltag, s)) >= 0;)
				sectors[s].ceilinglightsec = secidx;
			break;

		case 270:
			if (!ltag) { error = true; break; }

			if (has3D270 == -1)
			{
				if (!R3D_linemap)
				{
					pr_OFinsert("  WARNING: map has unsupported line special type 270\n");
					has3D270 = -2;
				}
				else
					has3D270 = i;
			}
			break;

		case 282:
			lines[i].special = 0;
			break;

		default:
			break;
		}

		if (error)
		{
			error = false;
			pr_OFinsert("MAP ERROR: L%i has special %i with an unset sector tag (special has been ignored)\n", i, lines[i].special);
			lines[i].special = 0;
			ma_inf.maperror++;
		}
	}

	if (has3D270 < 0) return;

	for (i = has3D270; i < numlines; i++)
	{
		if (lines[i].special == 270)
		{
			// Support for flats for 3D structures
			lines[i].special = 0;

			secidx = sides[*lines[i].sidenum].sector - sectors;

			for (s = -1; (s = P_FindSectorFromTag(lines[i].tag,s)) >= 0;)
			{
				sectors[s].sector3D = secidx;

				// v220-33 Extended support - this allows the floor/ceil tex to
				//         be x,y shifted and precise scaling values to be set
				if (lines[i].backsector)
				{
					// extended method
					sectors[secidx].flooroffx = sectors[secidx].ceiloffx = sides[lines[i].sidenum[0]].textureoffset;
					sectors[secidx].flooroffy = sectors[secidx].ceiloffy = sides[lines[i].sidenum[0]].rowoffset;
					sectors[secidx].scalingValueX = FIX2FLT(sides[lines[i].sidenum[1]].textureoffset)/10000.f;
					sectors[secidx].scalingValueY = FIX2FLT(sides[lines[i].sidenum[1]].rowoffset)/10000.f;

					if (   sectors[secidx].scalingValueX < 0.1f
						|| sectors[secidx].scalingValueY < 0.1f
						|| sectors[secidx].scalingValueX > 4.f
						|| sectors[secidx].scalingValueY > 4.f)
					{
						Con_Error("Impractical scaling values set with L%i x %f y %f- must be between 4.0 and 0,1\n"
							"40000 will convert to 4 (4 x bigger) 1000 will convert to 0.1 (10 x smaller)\n",
							i, sectors[secidx].scalingValueX, sectors[secidx].scalingValueY);
					}
				}
				else
				{
					short scale = lines[i].frontsector->lightlevel;

					// check for validity (authors don't always get this).
					if (scale < 1 || scale > 8) { lines[i].frontsector->lightlevel = 1; scale = 1; }

					// 1's line method
					sectors[secidx].scalingValueX = sectors[secidx].scalingValueY = 1.f/(float)scale;
				}

				// See if anything else needs to be transferred
				if (sectors[secidx].tag)
				{
					if (sectors[secidx].heightsec >= 0) sectors[s].heightsec = sectors[secidx].heightsec;
					if (sectors[secidx].floorlightsec >= 0) sectors[s].floorlightsec = sectors[secidx].floorlightsec;
					if (sectors[secidx].ceilinglightsec >= 0) sectors[s].ceilinglightsec = sectors[secidx].ceilinglightsec;
				}
			}
		}
	}
}

// Check switch activated doors actually have a switch. If not change the
// trigger type to push so it can be activated.  It would seem that this
// problem occurs when maps are authored to run in ZDoom as using this
// port they work, masking the fact that there is an authoring error.
void LS_CheckDoorType()
{
	int i;

	for (i = 0; i < numlines; i++)
	{
		if (  !lines[i].tag
			&& lines[i].special >= GenLockedBase
			&& lines[i].special <  GenCeilingBase)	// hence includes GenDoorBase
		{
			int Trig = (lines[i].special & TriggerType) >> TriggerTypeShift;

			if (Trig == SwitchOnce || Trig == SwitchMany)
			{
				lines[i].special |= 4;	// Change to PushOnce/PushMany
				pr_OFinsert("  LINE %i: GenDoor has no tag - trigger changed from switch to push\n", i);
			}
		}
	}
}

// R3D map error checks
int LS_CheckTaggedErrors(void)
{
	int i, s, count = 0;

	CLR_hitline();

	if (!R3D_ScriptedMap)	// v220-35 Do not clear sector tags with unset line specials
	{
		for (i = 0; i < numlines; i++)
		{
			line_t *ld = &lines[i];

			if (ld->special &&
				!ld->tag &&
				ld->ml_flags & ML_TWOSIDED &&
				ld->backsector->tag &&
				ld->backsector->tag != 666 &&
				ld->backsector->tag != 667)
			{
				int		k;
				RDBOOL	found = false;

				for (k = 0; k < numlines; k++)
				{
					if (!hitline[k] && lines[k].tag == ld->backsector->tag)
					{
						found = true;
						break;
					}
				}

				if (!found)
				{
					if (verbose)
						pr_OFinsert("  Map error fixed: Unassigned tag - setting line %i's tag to %i\n", i, lines[i].backsector->tag);

					count++;

					if (ld->special == ld->backsector->tag)
						// Likely map error - value put in wrong field
						ld->tag = ld->backsector->tag;
					else
						ld->tag = ld->backsector->tag = 0;

					hitline[i] = 1;
				}
			}
		}
	}

	// Some of this could be made redundant as written prior to movsec stuff
	for (i = 0; i < numlines; i++)
	{
		if (lines[i].special)
		{
			for (s = -1; (s = P_FindSectorFromTag(lines[i].tag,s)) >= 0;)
			{
				if (P_CheckLightSpecial(lines[i].special))
				{
					// Discriminate cases where lightlevels are transferred
					if (lines[i].special == 213 || lines[i].special == 261)
						sectors[s].specFlags |= (SF_HASSECLIGHT);
					else
						sectors[s].specFlags |= (SF_USESECLIGHT|SF_HASSECLIGHT);

					continue;
				}

				if (lines[i].tag && !P_CheckSpecial(lines[i].special))
					secspflags[s] |= SECSPF_HASACTION;

				switch(lines[i].special)
				{
				case  19: case  56: case  58: case  59:
				case  91: case  92: case  93: case  94:
				case  96: case 119: case 128: case 129:
				case 130: case 142: case 147:
					secspflags[s] |= SECSPF_RAISEFLR;		//raise floor (used with CheckHeights)
					break;
				case 39:	// teleport trigger
				case 97:	// teleport retrigger
				case 125:	// teleport monsteronly trigger
				case 126:	// as 125
				case 207:	// silent thing teleporters
				case 208:	// as 207
				case 243:	// silent line-line teleporter
				case 244:	// make fit within DCK's 256 linedef types
				case 262:	// add monster only
				case 263:	// silent thing,line,line rev types
				case 264:	// plus player/monster silent line
				case 265:	// reversed types
				case 266:	// add monster-only silent line-line
				case 267:	// as 266
				case 268:	// add monster-only silent
				case 269:	// as 268
					secspflags[s] |= SECSPF_TELEPTAG;	//teleport tags
					break;
				default:
					break;
				}
			}
		}
	}
	return count;
}

void LS_SetHASTEXTUREflags()
{
	sector_t *sec;
	int i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		for (k = 0; k < sec->linecount; k++)
		{
			if (!(sec->lines[k]->spflags & SPF_CONTROLLINE) &&
				origTEX[sec->lines[k] - lines] & 0x77)
			{
				sec->specFlags |= SF_HASTEXTURE;
				break;
			}
		}
	}
}

//============ Set Line Types


static int CheckIsRing(int lnum, byte *markline)
{
	line_t	 *ld = &lines[lnum];
	sector_t *s = ld->backsector;
	vertex_t *v = ld->v1;
	int		 i, isclosed = false, repeat = true;

	CLR_conarray();
	markline[lnum] = 2;
	conarray[lnum] = 1;

	while (repeat)
	{
		repeat = false;
		for (i = 0; i < s->linecount; i++)
		{
			int l = s->lines[i] - lines;

			if (markline[l] == 1)
			{
				if (s->lines[i]->v1 == v)
				{
					v = s->lines[i]->v2;
					markline[l] = 2;
					conarray[l] = 1;
					repeat = true;
				}
				else if (s->lines[i]->v2 == v)
				{
					v = s->lines[i]->v1;
					markline[l] = 2;
					conarray[l] = 1;
					repeat = true;
				}
			}
			else if (l == lnum && v == ld->v2)
			{
				isclosed = true;
				break;
			}
		}
	}
	if (!isclosed)
	{
		for (i = 0; i < s->linecount; i++)
			if (markline[s->lines[i] - lines] == 2)
				markline[s->lines[i] - lines] = 1;
	}
	return isclosed;
}

static RDBOOL CheckValidFloodlineSec(sector_t *liqs, sector_t *movs)
{
	int i;

	for (i = 0; i < movs->linecount; i++)
	{
		line_t *ld = movs->lines[i];

		// Test added b220-1
		if (ld->backsector && origTEX[ld - lines] & 0x11)
		{
			if (   ld->frontsector->floorheight < liqs->floorheight
				|| ld->backsector->floorheight  < liqs->floorheight)
			{
				return false;
			}
		}

		if (IsLiquidFloorFlat(ld->frontsector)
			&& ld->frontsector != liqs)
		{
			return false;
		}

		if (ld->backsector
			&& IsLiquidFloorFlat(ld->backsector)
			&& ld->backsector != liqs)
		{
			return false;
		}
	}
	return true;
}

void LS_Setlinetypes()
{
	line_t *ld;
	int i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (LSS(ld)
			&& ld->specType < R3D_40
			&& origTEX[i] == 0x22)
		{
			Settype(ld, Chk0304Type(ld), 21004);
		}
	}
}
