//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2011 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_misc.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//		LS_CheckDonut removed from this file
//
// DESCRIPTION: miscellaneous functions
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "dd_types.h"
#include "dd_wad.h"

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "m_misc.h"
#include "p_maputil.h"

extern int	stat_vispresent;

// sub functions ===============================================================

// Checks two door sectors for controlling sets of lines
RDBOOL subfn_CheckCompliantDoor(sector_t *s1, sector_t *s2)
{
	int i, count = 0;

	for (i = 0; i < s1->linecount; i++)
		if (s1->lines[i]->mtflags == (R3D_MISHTEX1|R3D_MISLTEX2))
			count++;

	if (count != 1) return false;

	for (i = 0; i < s2->linecount; i++)
		if (s1->lines[i]->mtflags == (R3D_MISHTEX1|R3D_MISLTEX2))
			count--;
	return (!count);
}

// Get rid of spurious specSec settings with single un-
// connected lines in sectors with animated tex floors;
// These are only valid if associated with lines having
// mid tex's each side.

void LS_DelSpecSecWithUnconnectedLinesInAnimSec()
{
	line_t	*ld;
	int		i, n;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   LSS(ld)
			&& IsLiquidFloorFlat(ld->frontsector)
			&& (n = GetLptr(ld)) >= 0
			&& Vtxs1(ld))
		{
			sector_t *csec = &sectors[n];
			int k, found = false;

			for (k = 0; k < csec->linecount; k++)
			{
				int type = csec->lines[k]->specType;

				if (   type == R3D_03
					|| type == R3D_04
					|| type == R3D_3D)
				{
					found = true;
					break;
				}
			}

			if (!found) Setspec(ld, -1, 10001);
		}
	}
}

void LS_CheckR3D_1AforR3D_3F()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (lineflags[i] & LFG_3DTYPE) continue;

		// Quick fix to problem - need to establish cause
		// Probably now not required - May 2004
		if (lineflags[ld - lines] & LFG_SEC2SET && !ChkLptr(ld))
		{
			lineflags[ld - lines] &= ~LFG_SEC2SET;
		}
		
		// These line types should not be rendered
		if (ld->specType == R3D_1A &&
			!LineHasMovSectors(ld))	// rev74 note incorrect code replaced
		{
			int mtflags = ld->mtflags & R3D_MISTEXANY;
			int itflags = ld->mtflags & R3D_INVTEXANY;
			if (mtflags == R3D_MISLTEX1 && itflags == R3D_INVHTEX2)
				Settype(ld, R3D_3F, 10002);
		}
	}
}

RDBOOL subfn_SetTexInSector(sector_t *sec, side_t *sd, RDBOOL invhflag)
{
	if (sd->bottomtexture ||
		sd->midtexture ||
		sd->toptexture)
	{
		sec->specFlags |= SF_SIDETEXTRUE;
	}
	if (!invhflag && sd->toptexture)
	{
		// Only set if texture exists and not invalid
		sec->specFlags |= SF_HITEXTRUE;
	}
	if (sd->bottomtexture ||
		sd->toptexture)
	{
		sec->specFlags |= SF_HILOTEXTRUE;
	}

	return false;
}

void LS_SetTexInSector()
{
	sector_t *sec = sectors;
	int i, k;

	for ( i = 0; i < numsectors; i++, sec++)
	{
		// Note: may have been set in NS_CheckSSLineFlag()
		if ( !(sec->specFlags & SF_SSLINETRUE) )
		{
			for (k = 0; k < sec->linecount; k++)
			{
				if (!(sec->lines[k]->spflags & SPF_CONTROLLINE) &&
					!(sec->lines[k]->ml_flags & ML_TWOSIDED))
				{
					sec->specFlags |= SF_SSLINETRUE;
					break;
				}
			}
		}

		for (k = 0; k < sec->linecount; k++)
		{
			if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
			{
				if (subfn_SetTexInSector(sec,
						&sides[sec->lines[k]->sidenum[0]],
						(sec->lines[k]->mtflags & R3D_INVHTEX1) != 0))
				{
					break;
				}
				if (sec->lines[k]->ml_flags & ML_TWOSIDED &&
					subfn_SetTexInSector(sec,
						&sides[sec->lines[k]->sidenum[1]],
						(sec->lines[k]->mtflags & R3D_INVHTEX2) != 0))
				{
					break;
				}
				
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////
//
// Transfer sec ptrs to unset or single lines
//

// Additional function added Sep 4 03
RDBOOL subfn_CheckR3D_04(sector_t *sec)
{
	int i, specSec = -1;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i], *ldc;

		if (ld->frontsector != sec ||
			ld->backsector != sec)
			return false;

		if (ChkLptr(ld))
		{
			if (specSec == -1)
				specSec = GetLptr(ld);
			else if (!CmpLptr(ld, specSec))
				return false;
		}

		if (ld->tyflags & TYF_MIDTEXLINE)
			continue;

		if ((ldc = subfn_FindCoincidentLine(ld, 0)) != NULL &&
			ldc->tyflags & TYF_MIDTEXLINE)
			continue;

		return false;
	}

	if (specSec == -1) return false;

	for (i = 0; i < sec->linecount; i++)
		Setspec(sec->lines[i], specSec, 10019);

	return true;
}


// Check for closed mov sectors with only mid texs
// that are set R3D_00
void LS_CheckReversedLineMovCeilSecs()
{
	sector_t *sec;
	int i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& sec->linecount == 4
			&& (sec->specFlags & 0x00FFFFFF) == 0x0002809C
			&& LS_ChkAllVtx2(sec, false, false))
		{
			line_t *ld = sec->lines[0];
			sector_t *s = ld->frontsector == sec ?
				ld->backsector : ld->frontsector;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->specType != R3D_00
					|| s != (ld->frontsector == sec ?
						ld->backsector : ld->frontsector))
				{
					s = NULL;
					break;
				}
			}
			if (s && !(s->specFlags & SF_MOVSECTOR))
			{
				for (k = 0; k < sec->linecount; k++)
				{
					Setspec(sec->lines[k], s-sectors, 10020);
					if (origTEX[sec->lines[k] - lines])
						Settype(sec->lines[k], R3D_1D, 10021);
				}

				Setsecdone2(sec, 10022);
			}
		}
	}
}

//========================================

static RDBOOL subfn_checkLoopSSECS(sector_t *s1, sector_t *s2)
{
	int i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& (sg->linedef->frontsector == s1 || sg->linedef->backsector == s1)
				&& (sg->linedef->frontsector == s2 || sg->linedef->backsector == s2) )
			{
				return true;
			}
		}
	}
	return false;
}

// b211-0 Look for closed sectors with no lower
// textures that totally own a similar sector
void LS_CheckRingSectors()
{
	sector_t	*sec, *s, *s1, *s2;
	ssline_t	*sl;
	line_t		*ld;
	int			box[4];
	int			c, i, k, m;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		// v220-10 eliminate (rare) occlusion problems with fake sectors
	/*	if (sec->heightsec >= 0 && sec->ceilingheight <= sec->floorheight)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->backsector)
				{
					ssline_t *sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
						subsectors[sl->list[m]].ssflags |= SSF_NOOCCLUDE;
				}
			}
		}*/

		if (   !secdone[i]
			&& sec->stptr < 0
			&& (sec->specFlags & 0x14) == 0x14)
		{
			RDBOOL valid = false;

			for (c = k = 0, s1 = s2 = NULL; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->specType >= R3D_40
					|| !(ld->mtflags & R3D_MISLOTEX)	// v220-5 test added
					|| !(s = ld->frontsector == sec ? ld->backsector  : ld->frontsector)
					|| s == sec)
				{
					s1 = NULL; break;
				}

				if ((s->specFlags & 0x14) == 0x14)
				{
					if (!s1) s1 = s;
					
					if (s1 == s) c++; else { s1 = NULL; break; }
				}
				else
				{
					if (!s2) s2 = s;
					if (s2 != s) { s1 = NULL; break; }
				}
			}

			if (!s1 || !s2) continue;

			CLR_hitline();

			for (k = 0; k < s1->linecount; k++)
			{
				ld = s1->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) { c++; continue; }

				hitline[ld - lines] = 1;
			}

			if (s1->linecount != c) continue;

			for (k = 0; k < s1->linecount; k++)
			{
				ld = s1->lines[k];

				if (   hitline[ld - lines] == 1
					&& subfn_CheckForRing(hitline, s1, ld - lines, NULL, box, NULL, NULL))
				{
					if (LS_MarkSubsecsInAreaGeneral(s1, hitline, 0, NULL, box, -1, MSIA_REVERSED)
						&& subfn_checkLoopSSECS(sec, s2))
					{
						valid = true; break;
					}

					for (m = 0; m < s1->linecount; m++)
					{
						ld = sec->lines[m];

						if (hitline[ld - lines] == 4)
							hitline[ld - lines] = 8;
					}
				}
			}

			if (valid)
			{
				RDBOOL type = (sec->specFlags & 0x1C) == 0x1C
					&& (s1->specFlags & 0x1C) == 0x1C;

				for (c = k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];
					
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->floorsec < 0 && ss->sector != s2)
						{
							if (type)
							{
								SetOverride(ss, s2 - sectors, 10023);
							//	ss->ssflags |= SSF_NOOCCLUDE;
								c++;
							}
							else
							{
								if (ss->sector->floorheight < s2->floorheight)
								{
									Setr3dssfs(ss, s2 - sectors, 10024); c++;
								}
							}
						}
					}
				}

				if (c)
				{
					for (k = 0; k < sec->linecount; k++)
					{
						line_t *ld = sec->lines[k];
						Settype(ld, (byte)(origTEX[ld - lines] ? R3D_00 : R3D_3F), 10025);
						SetLFG(ld, LFG_NEWSET|LFG_NOCHANGE, 10026);
					}

					Setsecdone(sec, 10027);
					Setsecdone(s1, 10028);
				}
			}
		}
	}
}

// b211-0 Check for rare but awkward and overcomplicated structures
//        with missing tex sectors contained in another missing tex
//        area referencing more than one sector.
//
// 1) LS_SetMiscStaticType
// 2) LS_SetMiscMoveType
// 3) LS_SetMiscLineType
//
static void subfn_SetMiscStaticType(sector_t *sec)
{
	sector_t *secn[2] = { NULL, NULL };
	sector_t *osec = NULL, *chk;
	int		 c, i, k, m;

	CLR_hitline();

	for (c = i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| origTEX[ld - lines]
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
			|| ld->specType >= R3D_40
			|| !Vtxs2(ld)
			|| lxarray[ld - lines]
			|| subfn_FindLineCrossing(ld, LC_NORMAL))
		{
			return;
		}

		chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

		if (!osec)
		{
			if (   chk->specFlags & (SF_MOVSECTOR|SF_NOLOTEX)
				|| chk->floorheight != sec->floorheight
				|| chk->ceilingheight != sec->ceilingheight)
			{
				return;
			}

			osec = chk;
		}
		else if (osec != chk)
		{
			return;
		}

		hitline[ld - lines] = 1;
		linearray[c++] = ld - lines;
	}

	if (!osec) return;

	validcount++;
	ss_validcount++;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	 *ld = sec->lines[i];
		ssline_t *sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				seg_t *sg = &segs[ss->firstline];

				ss->ssvalidcount = ss_validcount;

				if (ss->sector == osec) continue;

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (  !sg->linedef
						|| sg->linedef->validcount == validcount)
					{
						continue;
					}

					if ( !(sg->linedef->lnflags2 & LNF2_OWNEQUALTOTWO) )
						return;

					sg->linedef->validcount = validcount;
					
					if (hitline[sg->linedef - lines])
						continue;

					if (   !sg->backsector
						|| (sg->frontsector->specFlags & 0xF000C) != 0xC
						|| (sg->backsector->specFlags & 0xF000C) != 0xC
						|| sg->frontsector->floorheight   <= sec->floorheight
						|| sg->frontsector->ceilingheight != sec->ceilingheight
						|| sg->backsector->floorheight    <= sec->floorheight
						|| sg->backsector->ceilingheight  != sec->ceilingheight
						|| sg->frontsector->floorheight   != sg->backsector->floorheight)
					{
						return;
					}

					linearray[c++] = sg->linedef - lines;

					if (!secn[0])
					{
						secn[0] = sg->frontsector;
						secn[1] = sg->backsector;
						continue;
					}
					else
					{
						if (secn[0] == sg->frontsector || secn[0] == sg->backsector)
							continue;
					
						if (secn[1] == sg->frontsector || secn[1] == sg->backsector)
							continue;
					}

					return;
				}
			}
		}
	}

	if (!secn[0] || !secn[1]) return;

	for (i = 0; i < c; i++)
	{
		line_t	 *ld = &lines[linearray[i]];
		ssline_t *sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->floorsec < 0 && ss->sector != osec)
 				SetOverride(ss, osec - sectors, 10029);
		}

		SetLFG(ld, LFG_NEWSET, 10029);
	}

	Setsecdone(sec    , 10029);
	Setsecdone(secn[0], 10029);
	Setsecdone(secn[1], 10029);
}

void LS_SetMiscStaticType()
{
	sector_t *sec;
	int		 i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  (sec->specFlags & 0xF00FC) == 0x1C
			&& !secdone[i]
			&& sec->stptr < 0)
		{
			subfn_SetMiscStaticType(sec);
		}
	}
}

void LS_SetMiscMoveType()
{
	sector_t *sec;
	int		 c, i, k, m, n;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secdone[i] || !(SecCtrl(i)) ||  sec->stptr >= 0 || sec->heightsec >= 0) continue;

		if ( (sec->specFlags & 0x33040C) == 0x21000C)
		{
			subsector_t *ss;
			seg_t		*sg;
			sector_t	*os;
			ssline_t	*sl;
			int			lncnt = 0;

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t		*ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;
				if (ld->frontsector != ld->backsector) goto end_chkceilmov;
				linearray[c++] = ld - lines;
			}

			if (!c) goto end_chkceilmov;

			ss_validcount++;
			hitssecptr = 0;

			for (k = 0, os = NULL; k < c; k++)
			{
				sl = &sslines[linearray[k]];

				for (m = 0; m < sl->num; m++)
				{
					ss = &subsectors[sl->list[m]];

					if (ss->ssvalidcount == ss_validcount) continue;

					ss->ssvalidcount = ss_validcount;
					sg = &segs[ss->firstline];

					if (ss->sector == sec) hitssec[hitssecptr++] = sl->list[m];

					for (n = 0; n < ss->linecount; n++, sg++)
					{
						if (sg->frontsector && sg->frontsector != sec)
						{
							if (!os)
							{
								os = sg->frontsector;
								if ((os->specFlags & 0x40C) == 0xC) goto end_chkceilmov;
							}
							else if (os != sg->frontsector) goto end_chkceilmov;
						}
					}
				}
			}

			if (os)
			{
				for (k = 0; k < hitssecptr; k++)
					SetOverride(&subsectors[hitssec[k]], os - sectors, 10030);

				for (k = 0; k < c; k++)
					SetLFG(&lines[linearray[k]], LFG_NEWSET, 10030);

				Setsecdone(sec, 10030);
			}
end_chkceilmov:;
		}
		else
		if ( (sec->specFlags & 0x13001C) == 0x11001C)
		{
			sector_t *s = NULL, *chk;

			for (c = k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if ( !(ld->ml_flags & ML_TWOSIDED)
					|| ld->frontsector == ld->backsector
					|| !Vtxs2(ld)
					|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
					|| lineflags[ld - lines])
				{
					c = 0; break;
				}

				chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (!s) s = chk;

				if (   s != chk
					|| SecCtrl(s - sectors)
					|| (s->specFlags & 0x30001C) != 0x1C)
				{
					c = 0; break;
				}
				linearray[c++] = ld - lines;
			}

			if (c > 2)
			{
				CLR_hitline();

				for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

				for (k = 0, chk = NULL; k < c; k++)
				{
					ssline_t *sl = &sslines[linearray[k]];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];
						seg_t		*sg = &segs[ss->firstline];

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							if (sg->frontsector == s
								&& sg->backsector
								&& sg->backsector != s
								&& sg->backsector != sec
								&& !(sg->backsector->specFlags & 0x30004))
							{
								chk = sg->backsector; break;
							}
						}
						if (chk) break;
					}
					if (chk) break;
				}

				if (!chk) continue;

				for (k = 0; k < c; k++)
				{
					if (   hitline[linearray[k]] == 1
						&& subfn_CheckForRing(hitline, sec, linearray[k], NULL, NULL, NULL, NULL))
					{
						for (m = 0; m < sec->linecount; m++)
						{
							line_t *ld = sec->lines[m];

							n = ld - lines;

							if (hitline[n] == 4)
							{
								byte t = origTEX[n] & 0x22 ? R3D_16 : R3D_3F;

								Setspec(ld, chk - sectors, 10031);
								Settype(ld, t, 10031);
								SetLFG(ld, LFG_NEWSET, 10031);
							}
						}
						Setsecdone(sec, 10031);
						Setsecdone(chk, 10031);
					}
				}
			}
		}
	}
}

static RDBOOL vtxLinesIterator(int x, int y, vertex_t *v1, sector_t *sec)
{
	int			offset;
	const long	*list;
	
	if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
		return false;

	offset = y*bmapwidth + x;
	offset = *(blockmap + offset);
	list = blockmaplump + offset;

	// skip starting delimiter
	while( *(++list) >= 0)
	{
		line_t *ld = &lines[*list];

		if (ld->validcount == validcount)
			continue;		// line has already been checked

		ld->validcount = validcount;

		return ((ld->v1 == v1 || ld->v2 == v1)
			&& ld->frontsector == sec
			&& !(ld->ml_flags & ML_TWOSIDED));
	}
	return false;
}

static RDBOOL lineVertexConnected(line_t *ld, vertex_t *v1, sector_t *sec)
{
    int		xl, xh, yl, yh, bx, by;
	int		adjust = 256*FRACUNIT;

	validcount++;
	ld->validcount = validcount;	//Don't test self

	xl = MAPBS3(v1->x, bmaporgx, -adjust);
	xh = MAPBS3(v1->x, bmaporgx,  adjust);
	yl = MAPBS3(v1->y, bmaporgy, -adjust);
	yh = MAPBS3(v1->y, bmaporgy,  adjust);

	if (xl < 0) xl = 0;
	if (yl < 0) yl = 0;
	if (xh >= bmapwidth ) xh = bmapwidth  - 1;
	if (yh >= bmapheight) yh = bmapheight - 1;

	for (bx=xl ; bx<=xh ; bx++)
	{
		for (by=yl ; by<=yh ; by++)
		{
			if (vtxLinesIterator(bx, by, v1, sec))
				return true;
		}
	}
	return false;
}

typedef struct {
	sector_t *fs;
	sector_t *bs;
	sector_t *os;
	vertex_t *v;
	int		 lnum;
	int		 done;
} missceilvtx_t;

static missceilvtx_t *mcv = NULL;
static int mcvnum, mcvmaxnum;

// b221-0 1) Contained move floor line types
//        2) Cointained multi sec missing hitex flats
void LS_SetMiscLineType()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			i, k, sn;

	mcvnum = mcvmaxnum = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !(ld->ml_flags & ML_TWOSIDED)
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| lxarray[i]
			|| fs->stptr >= 0
			|| bs->stptr >= 0
			|| ld->specType >= R3D_40)
		{
			continue;
		}

		if (   ld->lnflags2 & LNF2_OWNEQUALTOTWO
			// general mov floor rule
			&& (fs->specFlags & SF_MOVFLOOR)^(bs->specFlags & SF_MOVFLOOR)
			&& !(fs->specFlags & (SF_DUALFMOVSEC|SF_MOVCEIL))
			&& !(bs->specFlags & (SF_DUALFMOVSEC|SF_MOVCEIL)))
		{
			// no lo, no hi sector tex rule
			if (   (fs->specFlags & 0x40C) == 0xC
				&& (bs->specFlags & 0x40C) == 0xC
				// discrimination rule
				&& !(fs->specFlags & SF_MOVSECTOR && fs->specFlags & SF_SECCLOSED)
				&& !(bs->specFlags & SF_MOVSECTOR && bs->specFlags & SF_SECCLOSED)
				// Both sectors must have the same containing sector
				&& (sn = LS_GetContainingSecnum(fs, 1)) >= 0
				&& LS_GetContainingSecnum(bs , 1) == sn)
			{
				Setspec(ld, sn, 999);
			}
			else	// look for lo floor 
			if (   Vtxs2(ld)
				&& (fs->specFlags & 0x7F041C) == 0x090014
				&& (bs->specFlags & 0xFF041C) == 0x80001C
				&& fs->floorheight > bs->floorheight
				&& !(origTEX[i] & 0x11))
			{
				ssline_t *sl = &sslines[i];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->sector == fs && ss->lfsec < 0)
						Setr3dsslfsec(ss, bs - sectors, 10032);
				}
			}
		}
		else
		if (   ld->lnflags2 & LNF2_OWNTWOANDPLUS
			&& !origTEX[i]
			&& fs->floorheight != bs->floorheight
			&& fs->ceilingheight != bs->ceilingheight
			&& (vtxarray[ld->v1 - vertexes] == 2
			||  vtxarray[ld->v2 - vertexes] == 2)
			&& !(fs->specFlags & (SF_MOVSECTOR|SF_NOHITEX))
			&& !(bs->specFlags & (SF_MOVSECTOR|SF_NOHITEX)) )
		{
			vertex_t *v = (vtxarray[ld->v1 - vertexes] == 2 ? ld->v2 : ld->v1);
			vertexowner_t *vt = &vertexowners[v - vertexes];

			for (k = 0; k < vt->num; k++)
			{
				sector_t *os = &sectors[vt->list[k]];

				if (lineVertexConnected(ld, v, os))
					continue;

				if ((os->specFlags & 0xF001C) == 0x1C)
				{
					if (mcvnum >= mcvmaxnum)
					{
						mcvmaxnum += 16;
						mcv = (missceilvtx_t*)W_ReallocCLR(mcv, mcvmaxnum * sizeof(missceilvtx_t), STATICID, 0);
					}

					mcv[mcvnum].fs = fs;
					mcv[mcvnum].bs = bs;
					mcv[mcvnum].os = os;
					mcv[mcvnum].v = v;
					mcv[mcvnum++].lnum = i;
				}
			}
		}
	}

	if (mcvnum)
	{
		for (i = 0; i < mcvnum; i++)
		{
			int	m, c = -1;	
			
			if (mcv[i].done) continue;

			mcv[i].done = true;

			for (k = i + 1; k < mcvnum; k++)
			{
				if (mcv[k].done) continue;

				if (   mcv[k].fs == mcv[i].fs
					&& mcv[k].bs == mcv[i].bs
					&& mcv[k].os == mcv[i].os)
				{
					mcv[k].done = true;
					if (c < 0) c = k; else { c = -1; break; }
				}
			}

			if (c > 0)
			{
				// The assumption here is that since a pair of lines each reference the same set
				// of sectors that the area between should adopt the lowest ceiling sector
				// of the reference line with missing textures.
				sector_t *s = (mcv[c].fs->ceilingheight < mcv[c].bs->ceilingheight ? mcv[c].fs : mcv[c].bs);
				sector_t *cmp = (s == mcv[c].fs ? mcv[c].bs : mcv[c].fs);

				for (k = 0; k < mcv[i].os->linecount; k++)
				{
					ssline_t *sl = &sslines[mcv[i].os->lines[k] - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ceilsec < 0 && (ss->sector == cmp || ss->sector == mcv[c].os))
							Setr3dsscs(ss, s - sectors, 10033);
					}
				}
			}
		}
		W_Free(mcv, STATICID, 0);
		mcv = NULL;
	}
}

//===========================================
//
// b211-0 check dual ring sectors

static RDBOOL validateList(int *sslist, int sscnt, RDBOOL setlfg)
{
	int i, k, ret = false;

	validcount++;

	for (i = 0; i < sscnt; i++)
	{
		subsector_t *ss = &subsectors[sslist[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (ld && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (setlfg)
				{
					SetLFG(ld, LFG_NEWSET|LFG_EXCLUDE, 10034);

					if (!origTEX[ld - lines])
						Settype(ld, R3D_3F, 10034);
					else if (ld->frontsector == ld->backsector
						&& origTEX[ld - lines] & 0x22)
						Settype(ld, R3D_03, 10034);
				}
				else
				{
					if (!hitline[ld - lines])
					{
						if (!secEquiv(ld->frontsector, ld, NULL))
							return false;
						ret = true;
					}
				}
			}
		}
	}
	return ret;
}

// v220-7 Check for lines sandwiched beteen two rings.
// This should be easy but is complicated by the way the nodebuilder creates
// subsectors with stuff like this.  The only way to get all the subsectors is
// to derive them using the bounding algo LS_MarkSubsecsInAreaGeneral - b211-0
void LS_CheckDualRingSectors()
{
	sector_t *sec;
	int		 i, k;
	int		 *sslist, sscnt;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !(sec->specFlags & SF_MOVSECTOR)
			&& (sec->specFlags & 0xC) != 0xC
			&& sec->linecount > 5
			&& sec->ceilingheight - sec->floorheight >= 4*FRACUNIT
			&& (sslist = getSSECSinDualRingSector(sec, &sscnt, 0, false, false)))
		{
			if (validateList(sslist, sscnt, false))
			{
				for (k = 0; k < sscnt; k++)
					SetOverride(&subsectors[sslist[k]], i, 10035);

				validateList(sslist, sscnt, true);

				Setsecdone(sec, 10035);
			}

			M_Free(sslist);
		}
	}
}

// v220-7
void LS_CheckDualRingMissingSectors()
{
	sector_t *sec, *chk, *s1, *s2;
	int		 c, i, k, lnum, rc, b1[4], b2[4], *box;
	byte	 *llist, *ptr;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   (sec->specFlags & 0x3001C) == 0x1C
			&& sec->linecount > 5
			&& sec->heightsec < 0
			&& sec->ceilingheight - sec->floorheight >= 4*FRACUNIT)
		{
			s1 = NULL, s2 = NULL;

			for (k = 0; k < sec->linecount; k++)
			{
				chk = sec->lines[k]->frontsector == sec ? sec->lines[k]->backsector : sec->lines[k]->frontsector;

				if (!chk || chk->floorheight >= sec->floorheight) { s1 = NULL; break; }

				if (!s1) { s1 = chk; continue; } else if (s1 == chk) continue;
				if (!s2) { s2 = chk; continue; } else if (s2 == chk) continue;

				s1 = NULL; break;
			}

			if (!s1 || !s2) continue;

			// v220-8 eliminate mov sectors
			if (   s1->ceilingpic != sec->ceilingpic
				|| s2->ceilingpic != sec->ceilingpic
				|| s1->specFlags & SF_MOVSECTOR
				|| s2->specFlags & SF_MOVSECTOR)
			{
				continue;
			}

			if (   sec->ceilingpic != skyflatnum
				&& (s1->ceilingheight != sec->ceilingheight || s2->ceilingheight != sec->ceilingheight))
			{
				continue;
			}

			for (lnum = -1, c = k = 0; k < sec->linecount; k++)
			{
				chk = sec->lines[k]->frontsector == sec ? sec->lines[k]->backsector : sec->lines[k]->frontsector;

				if (chk == s1) { lnum = sec->lines[k] - lines; hitline[lnum] = true; c++; }
			}

			if (  !subfn_CheckForRing(hitline, s1, lnum, &rc, b1, NULL, NULL)
				|| rc != c)
			{
				continue;
			}

			llist = (byte*)calloc(numlines, 1);

			for (lnum = -1, c = k = 0; k < sec->linecount; k++)
			{
				chk = sec->lines[k]->frontsector == sec ? sec->lines[k]->backsector : sec->lines[k]->frontsector;

				if (chk == s2) { lnum = sec->lines[k] - lines; llist[lnum] = true; c++; }
			}

			if (  !subfn_CheckForRing(llist, s2, lnum, &rc, b2, NULL, NULL)
				|| rc != c)
			{
				goto free_llist;
			}

			if (M_CompareBox(b1, b2))
			{
				if (!(s1->specFlags & SF_NOLOTEX)) goto free_llist;
				if (  s2->specFlags & SF_NOLOTEX ) goto free_llist;
				ptr = llist, box = b2, chk = s2;
			}
			else
			{
				if (  s1->specFlags & SF_NOLOTEX ) goto free_llist;
				if (!(s2->specFlags & SF_NOLOTEX)) goto free_llist;
				ptr = hitline, box = b1, chk = s1;
			}

			if (LS_MarkSubsecsInAreaGeneral(chk, ptr, 0, NULL, box, -1, MSIA_REVERSED))
			{
				RDBOOL	valid = true;
				int		m, lcnt = 0;

				validcount++;

				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t		*sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (sg->linedef && sg->linedef->validcount != validcount)
						{
							if (origTEX[sg->linedef - lines] & 0x55)	// was origTEXcopy ??
							{
								valid = false; break;
							}

							sg->linedef->validcount = validcount;
							linearray[lcnt++] = sg->linedef - lines;
						}
					}

					if (!valid) break;
				}

				if (valid)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];

						if (ss->sector->floorheight != s2->floorheight
							|| ss->sector->floorpic == s2->floorpic)
						{
							Setr3dssfs(ss, s2 - sectors, 10036);
						}

						Setr3dsscs(ss, i, 10036);
						Setr3dsslfsec(ss, i, 10036);
						ss->ssflags |= (SSF_NOCHANGE|SSF_FIXED);
					}

					for (k = 0; k < lcnt; k++)
						SetLFG(&lines[linearray[k]], LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 10033);

					Setsecdone(sec, 10036);
					Setsecdone(s1 , 10036);
					Setsecdone(s2 , 10036);
				}
			}
free_llist:;
			M_Free(llist);
		}
	}
}

// ------------------------------------------------------

static sector_t *compsec;

static RDBOOL PIT_MatchLines(line_t *ld, int *box)
{
	ssline_t	*sl;
	int			i, k;

	if ( !(ld->ml_flags & ML_TWOSIDED)
		|| ld->frontsector == ld->backsector
		|| box[BOXRIGHT] < ld->bbox[BOXLEFT ]
		|| box[BOXLEFT ] > ld->bbox[BOXRIGHT]
		|| box[BOXTOP  ] < ld->bbox[BOXBASE ]
		|| box[BOXBASE ] > ld->bbox[BOXTOP  ])
	{
		return true;
	}

	if ( !origTEX[ld - lines] ) ld->specType = R3D_01;	// Prevent being changed in LS_SetMissingTexLineType()

	sl = &sslines[ld - lines];

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount == ss_validcount) continue;

		ss->ssvalidcount = ss_validcount;

		if (ss->lcsec < 0 && ss->sector->ceilingheight > compsec->ceilingheight)
		{
			seg_t		*sg = &segs[ss->firstline];
			RDBOOL		valid = true;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->sidedef && sg->sidedef->midtexture)
				{
					valid = false;
					break;
				}
			}

			if (valid) Setr3dsslcsec(ss, compsec - sectors, 50007);
		}
	}

	return true;
}

// v220-4	This appears to be only of use with fenexam.wad and gallows.wad.
//			Thankfully the method used in these wads is not found in general
//			and given these are only demo wads this code is maybe not needed.
//			It has been included - just in case....
void  LS_PresetLCSEC()
{
	sector_t	*sec = sectors;
	int			i, k, m;

	for (i = 0; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& (sec->specFlags & 0xFFFFFC) == SF_NOHITEX
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& sec->ceilingpic != skyflatnum)
		{
			struct chk_t {
				sector_t *chk;
				line_t   *ld[128];
				int		 count;
				int		 box[4];
			} chksec[32];

			sector_t	*chk;
			int			box[4];
			int			chkseccnt = 0;

			M_ClearBox(box);

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (lineflags[ld - lines] || !(ld->ml_flags & ML_TWOSIDED) || ld->mtflags & R3D_MISLOTEX)
				{
					chkseccnt = 0; break;
				}

				M_AddLineToBox(box, ld);

				if (ld->mtflags & R3D_MISHITEX)
				{
					int	chknum = -1;
					chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					for (m = 0; m < chkseccnt; m++)
					{
						if (chk == chksec[m].chk) { chknum = m; break; }
					}

					if (chknum < 0)
					{
						if (chkseccnt >= 32) break;
						chksec[chkseccnt].chk = chk;
						chksec[chkseccnt].count = 0;
						M_ClearBox(chksec[chkseccnt].box);
						chknum = chkseccnt++;
					}

					if (chksec[chknum].count >= 128) { chkseccnt = 0; break; }
					chksec[chknum].ld[chksec[chknum].count] = ld;
					chksec[chknum].count++;
					M_AddLineToBox(chksec[chknum].box, ld);
				}
			}

			if (chkseccnt > 1)
			{
				int a, c, n;

				for (c = k = 0; k < chkseccnt; k++)
				{
					if (M_SameBox(box, chksec[k].box) && !(chksec[k].chk->specFlags & 0xC))
					{
						n = k;
						c++;
					}
				}

				if (c == 1)
				{
					RDBOOL	valid = true;
					int		bx, by, xl, xh, yl, yh;

					for (a = 0; a < chksec[n].chk->linecount; a++)
						if (!chksec[n].chk->lines[a]->backsector) { valid = false; break; }

					if (!valid) continue;

					compsec = sec;

					M_getBlockOffsets(chksec[n].box, 0, &xl, &xh, &yl, &yh);

					validcount++;
					ss_validcount++;

					for (bx = xl ; bx <= xh ; bx++)
					{
						for (by = yl ; by <= yh ; by++)
						{
							P_BlockLinesIterator(bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_MatchLines, box);
						}
					}
				}
			}
		}
	}
}

// v220-7
static void subfn_CheckSkyRings()
{
	sector_t	*sec;
	line_t		*ld;
	int			rc, box[4];
	int			c, i, k, m;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		sector_t *osec = NULL, *chk;

		if (   sec->ceilingpic == skyflatnum
			&& sec->floorpic == skyflatnum
			&& sec->floorheight == sec->ceilingheight
			&& sec->heightsec < 0
			&& sec->linecount > 5)
		{
			int lnum = -1;

			CLR_hitline();

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];
				
				if (   ld->specType < R3D_40
					&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
					&& !origTEX[ld - lines]
					&& ld->ml_flags & ML_TWOSIDED
					&& ld->frontsector != ld->backsector)
				{
					chk = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (   chk->floorpic == skyflatnum
						&& chk->ceilingpic == skyflatnum
						&& chk->floorheight < chk->ceilingheight)
					{
						if (!osec) osec = chk;
						else if (osec != chk)
						{
							osec = NULL; break;
						}
						if (lnum < 0) lnum = ld - lines;
						hitline[ld - lines] = 1; c++;
					}
				}
			}

			if (!osec || c < 3) continue;

			if (   subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL)
				&& rc == c
				&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_REVERSED|MSIA_SKY))
			{
				validcount++;
				ss_validcount++;

				for (c = k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					
					if (ss->ssvalidcount != ss_validcount)
					{
						seg_t *sg = &segs[ss->firstline];
												
						ss->ssvalidcount = ss_validcount;
						
						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (sg->linedef && sg->linedef->validcount != validcount)
							{
								sg->linedef->validcount = validcount;

								if (   !hitline[sg->linedef - lines]
									&& (origTEX[sg->linedef - lines] & 0x55
									|| !(sg->linedef->ml_flags & ML_TWOSIDED)
									|| sg->linedef->frontsector->ceilingpic != skyflatnum
									|| sg->linedef->backsector->floorpic != skyflatnum))
								{
									goto next_sec;
								}

								linearray[c++] = sg->linedef - lines;
							}
						}
					}
				}

				if (c)
				{
					for (k = 0; k < hitssecptr; k++)
					{
						SetOverride(&subsectors[hitssec[k]], osec - sectors, 10044);
						setSSECflag(&subsectors[hitssec[k]], SSF_NOCHANGE, 10044);
					}

					for (k = 0; k < c; k++)
					{
						ld = &lines[linearray[k]];

						if (!(origTEX[ld - lines]) || ld->frontsector != ld->backsector)
							Settype(ld, R3D_3F, 10044);
						else
							Settype(ld, R3D_03, 10044);

						SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 10044);
					}
				}
next_sec:;
			}
		}
	}
}

// v220-7
void LS_CheckSameSecOwner()
{
	sector_t	*sec, *os, *fs, *bs;
	line_t		*ld;
	byte		*mrkline = (byte*)M_Calloc(numlines, 0);
	fixed_t		diff;
	int			c, i, k, m, n;
	int			*has_sec = (int*)malloc(numlines * sizeof(int));
	int			count = 0;
	int			*sslist, sscnt;
	char		txn;
	RDBOOL		repeat;
	
	memset(has_sec, -1, numlines * sizeof(int));
	
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];
			
			if (   ld->frontsector == ld->backsector
				&& ld->specType < R3D_40
				&& ld->lnflags2 & LNF2_OWNEQUALTOONE
				&& Vtxs2(ld)
				&& !lxarray[ld - lines])
			{
				linearray[c++] = ld - lines;
				mrkline[ld - lines] = 1;
			}
		}
		
		if (c < 3) continue;
		
		repeat = true;
		os = NULL;
		
		while (repeat)
		{
			int		lnum = -1, box[4], rc;
			
			for (k = 0; k < c; k++)
			{
				if (mrkline[linearray[k]] == 1)
				{
					lnum = linearray[k]; break;
				}
			}
			
			if (lnum < 0) break;
			
			repeat = false;
			
			// At this stage we are only concerned with line rings
			if (subfn_CheckForRing(mrkline, sec, lnum, &rc, box, NULL, NULL))
			{
				int flag = MSIA_NONE;

				repeat = true;

				// Check validity - ring must have no lines with different secs inside
				if (!LS_MarkSubsecsInMultiSecLoop(0, NULL, rc, box, NULL, true, MSIA_TESTAREA))
					flag = MSIA_REVERSED;

				if (LS_MarkSubsecsInMultiSecLoop(0, NULL, rc, box, NULL, true, flag))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ( (ld = sg->linedef)
								&& hitline[ld - lines] != 4
								&& (ld->frontsector != sec || ld->backsector != sec))
							{
								goto end_check;
							}
						}
					}
				}
				else
					goto end_check;

				validcount++;
				ss_validcount++;

				for (k = 0; k < c; k++)
				{
					if (mrkline[linearray[k]] == 4)
					{
						ssline_t *sl = &sslines[linearray[k]];
						
						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];
							
							if (ss->ssvalidcount != ss_validcount)
							{
								seg_t *sg = &segs[ss->firstline];
								
								ss->ssvalidcount = ss_validcount;
								
								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if ( (ld = sg->linedef) && ld->validcount != validcount)
									{
										ld->validcount = validcount;

										if ((fs = ld->frontsector) != (bs = ld->backsector))
										{
											if (!os) os = sg->frontsector;
											else if (os != sg->frontsector) { os = NULL; goto end_check; }	

											txn = origTEX[ld - lines];

											if  ( !bs
												|| ((diff = fs->ceilingheight - bs->ceilingheight) && (!(txn & (diff > 0 ? 0x04 : 0x40)) || !(fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum)))
												|| ((diff = fs->floorheight   - bs->floorheight  ) && (!(txn & (diff > 0 ? 0x10 : 0x01)) || !(fs->floorpic   == skyflatnum && bs->floorpic   == skyflatnum))) )
											{
												os = NULL; goto end_check;
											}
										}
									}
								}
							}
						}
					}
				}
end_check:;

				for (k = 0; k < c; k++)
				{
					if (mrkline[linearray[k]] == 4)
					{
						mrkline[linearray[k]] = 8;

						if (os)
						{
							if (os == sec)
								mrkline[linearray[k]] = false;
							else
								has_sec[linearray[k]] = os - sectors;
						}
					}
				}
			}

			for (k = 0; k < c; k++)
			{
				if (mrkline[linearray[k]] != 8)
					mrkline[linearray[k]] = false;
			}
		}
	}

	SET_countsec(0);

	ss_validcount++;

	for (i = 0; i < numlines; i++)
	{
		if (has_sec[i] >= 0 && !countsec[has_sec[i]])
		{
			countsec[has_sec[i]] = true;

			if ( (sslist = getSSECSinDualRingSector(&sectors[has_sec[i]], &sscnt, 1, false, false)) )
			{
				for (k = 0; k < sscnt; k++)
				{
					subsector_t *ss = &subsectors[sslist[k]];
					
					if (ss->ssvalidcount != ss_validcount)
					{
						seg_t *sg = &segs[ss->firstline];
						
						if (ss->sector - sectors != has_sec[i]) SetOverride(ss, has_sec[i], 10045);

						setSSECflag(ss, SSF_NOCHANGE, 10045);
						
						ss->ssvalidcount = ss_validcount;
						
						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (sg->linedef)
							{
								if (sg->linedef->frontsector == sg->linedef->backsector)
								{
									int ln = sg->linedef - lines;

									if (origTEX[ln] & 0x22) Settype(sg->linedef, R3D_03, 10046);
								}

								SetLFG(sg->linedef, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 10046);
							}
						}
					}
				}

				M_Free(sslist);
			}
		}
	}

	M_Free(has_sec);

	// v220-7 check for sky inheritance
	subfn_CheckSkyRings();
}

void LS_SetVisSplit()
{
	line_t		*ld;
	int			i;

	// Set up identifier for masked walls where they are crossing
	// a line.  Any identified  will be drawn complete using the
	// line's vertexes in place of the segs.

	stat_vispresent = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& origTEX[i] == 0x22
			&& !lxarray[i]
			&& !(lineflags[i] & LFG_3DTYPE)
			&& !(ld->spflags & (SPF_MIDTRANSLUCENT|SPF_LOWTRANSLUCENT|SPF_LINEONLINE)) )
		{
			if (sslines[i].num/2 > 1)
			{
				ld->ml_flags |= ML_VISPRESENT;
				stat_vispresent++;
			}
		}
	}
}

// ----------------------------------------------

static void subfn_chkOrphanedMov(sector_t *s)
{
	subsector_t	*ss;
	sector_t	*ch, *os = NULL;
	line_t		*ld;
	ssline_t	*sl;
	int			c, i, k, n;

	for (c = n = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (s->lines[i]->spflags & SPF_CONTROLLINE) { c++; continue; }
		if (!ld->backsector) return;
		if (ld->frontsector == ld->backsector) return;
		if (origTEX[ld - lines] & 0x55) return;
		else linearray[n++] = ld - lines;

		ch = ld->frontsector == s ? ld->backsector : ld->frontsector;
		if (!os) { { if (ch->specFlags & SF_MOVSECTOR) return; } os = ch; }
		else if (os != ch) return;
	}

	if (!c || !n || s->linecount != n + c) return;
	if (!os || (os->specFlags & 0x4 && !(os->specFlags & 0x400)) ) return;

	ss_validcount++;
	hitssecptr = 0;

	for (i = 0; i < n; i++)
	{
		sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;
			if (ss->sector == os) continue;
			if (ss->sector != s) return;
			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		SetOverride(&subsectors[hitssec[i]], os - sectors, 10047);
	}

	for (i = 0; i < n; i++)
	{
		if (origTEX[linearray[i]])
			Settype(&lines[linearray[i]], R3D_03, 10047);
		else
			Settype(&lines[linearray[i]], R3D_3F, 10047);
		SetLFG(&lines[linearray[i]], LFG_NEWSET, 10047);
	}

	Setsecdone(s, 10047);
}

void LS_chkOrphanedMov()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0x3001C) == 0x1000C
			&& SecCtrl(i))
		{
			subfn_chkOrphanedMov(&sectors[i]);
		}
	}
}
