//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_miscEffects.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: miscellaneous functions dealing with effects
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "m_misc.h"
#include "p_maputil.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

void subfn_CheckUnsetSectorHeightSectors(sector_t *sec)
{
	sector_t *s;
	int i, k;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->frontsector == ld->backsector ||
			origTEX[ld - lines])
			continue;

		s = ld->frontsector == sec ?
			ld->backsector : ld->frontsector;

		if (s->special ||
			s->tag ||
			s->heightsec >= 0)
			continue;

		if (ld->frontsector->ceilingpic ==
			ld->backsector->ceilingpic &&
			ld->frontsector->floorpic ==
			ld->backsector->floorpic &&
			ld->frontsector->ceilingheight ==
			ld->backsector->ceilingheight &&
			ld->frontsector->floorheight ==
			ld->backsector->floorheight)
		{
			line_t *li;
			int outer[4];

			Setspec(ld, sec - sectors, 28001);

			if (secdone[s - sectors]) break;

			memcpy(outer, secbbox[s - sectors].bounds, sizeof(outer));

			// Find all step lines contained
			for (k = 0, li = lines; k < numlines; k++, li++)
			{
				if (LSS(li)
					&& li->specType < R3D_40
					&& li->bbox[BOXLEFT ] >= outer[BOXLEFT ]
					&& li->bbox[BOXRIGHT] <= outer[BOXRIGHT]
					&& li->bbox[BOXBASE ] >= outer[BOXBASE ]
					&& li->bbox[BOXTOP  ] <= outer[BOXTOP  ])
				{
					Setspec(li, sec - sectors, 28002);
					Setsecdone2(li->frontsector, 28003);
				}
			}			
			Setsecdone2(s, 28004);
		}
	}
}

// Check for fake sector inheritance where fake sector's ceiling
// has an animated flat.
void LS_CheckFakeSectorInAnimated()
{
	sector_t *sec;
	int i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
		if (!secdone[i] &&
			sec->special == 0 &&
			sec->heightsec >= 0 &&
			IsLiquidCeilFlat(&sectors[sec->heightsec]))
			subfn_CheckUnsetSectorHeightSectors(sec);
}

static RDBOOL isValidR3D_1A(line_t *ld)
{
	sector_t *s = ld->mtflags & R3D_MISLTEX1 ? ld->frontsector : ld->backsector;
	sector_t *r = ld->frontsector == s ? ld->backsector : ld->frontsector;
	int		 i;

	if (!r) return false;	// insurance only

	for (i = 0; i < s->linecount; i++)
	{
		line_t *li = s->lines[i];

		if (   li != ld
			&& li->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2)
			&& (li->v1 == ld->v1 || li->v1 == ld->v2 || li->v2 == ld->v1 || li->v2 == ld->v2))
		{
			sector_t *c = li->mtflags & R3D_MISLTEX1 ? li->backsector : li->frontsector;

			if (c && c->floorheight != r->floorheight)
			{
				return false;
			}
		}
	}

	return true;
}

// v220-7/10/17 modified
void LS_CheckR3D_1ALines()
{
	sector_t	*fs, *bs;
	line_t		*ld = lines;
	int			i, sf, bf;
	int			lcnt = 0;
	int			*lnum = (int*)malloc(numlines * sizeof(int));
	byte		mf;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (   lineflags[i]
			|| ld->specType != R3D_00
			|| !(ld->ml_flags & ML_TWOSIDED)
			|| !((mf = ld->mtflags) & R3D_MISLOTEX)
			|| mf & R3D_MISHITEX
			|| !Vtfl3(ld)
			|| LineHasMovSectors(ld)
			|| ChkLptr(ld)
			|| (fs = ld->frontsector)->heightsec >= 0
			|| (bs = ld->backsector)->heightsec  >= 0
			|| abs(fs->floorheight - bs->floorheight) <= 8*FRACUNIT)
		{
			continue;
		}

		sf = fs->specFlags & (SF_NOLOTEX|SF_SSLINETRUE);
		bf = bs->specFlags & (SF_NOLOTEX|SF_SSLINETRUE);

		if (!((sf == SF_NOLOTEX)^(bf == SF_NOLOTEX))) continue;

		if (   fs->R3Dfloorheight != DDMAXINT
			|| bs->R3Dfloorheight != DDMAXINT
			|| fs->R3Dfloorsec >= 0
			|| bs->R3Dfloorsec >= 0)
		{
			continue;
		}

		if (   mf == R3D_MISLTEX1
			&& sf == SF_NOLOTEX
			&& !(bs->specFlags & 0xC && !(bs->specFlags & 0x400))
			&& !(sides[ld->sidenum[0]].texFlags == T_MID && !sides[ld->sidenum[1]].texFlags)
			&& bs->ceilingheight - bs->floorheight >= 8*FRACUNIT)
		{
			if (isValidR3D_1A(ld))
			{
				Settype(ld, R3D_1A, 28009);
				if (vertexowners[ld->v1 - vertexes].num == 2 && vertexowners[ld->v2 - vertexes].num == 2) lnum[lcnt++] = i;
			}
		}
		else
		if (   mf == R3D_MISLTEX2
			&& bf == SF_NOLOTEX
			&& !(fs->specFlags & 0xC && !(fs->specFlags & 0x400))
			&& !(sides[ld->sidenum[1]].texFlags == T_MID && !sides[ld->sidenum[0]].texFlags)
			&& fs->ceilingheight - fs->floorheight >= 8*FRACUNIT)
		{
			if (isValidR3D_1A(ld))
			{
				Settype(ld, R3D_1A, 28009);
				if (vertexowners[ld->v1 - vertexes].num == 2 && vertexowners[ld->v2 - vertexes].num == 2) lnum[lcnt++] = i;
			}
		}
	}

	// v220-33 check for R3D_1A rings
	if (lcnt > 2)
	{
		int		*lloop = (int*)malloc(lcnt * sizeof(int));
		byte	*markline = (byte*)malloc(lcnt);

		CLR_hitline();

retry:;
		int llcnt = 0;
		fs = NULL;

		for (i = 0; i < lcnt; i++)
		{
			if (hitline[lnum[i]]) continue;

			ld = &lines[lnum[i]];

			if (!fs)
			{
				lloop[llcnt++] = lnum[i];
				fs = ld->frontsector->floorheight < ld->backsector->floorheight ? ld->frontsector : ld->backsector;
			}
			else
			{
				bs = ld->frontsector->floorheight < ld->backsector->floorheight ? ld->frontsector : ld->backsector;

				if (bs == fs) lloop[llcnt++] = lnum[i];
			}
		}

		if (llcnt > 2)
		{
			float	fpbox[4];
			int		box[4];
			int		truecnt;

			memset(markline, 0, llcnt);

			ld = &lines[lloop[0]];
			// get containing sector
			bs = ld->frontsector == fs ? ld->backsector : ld->frontsector;

			// check if a closed loop exists
			if ( (truecnt = subfn_CheckLoopClosed(0, llcnt, lloop, box, markline)) > 2)
			{
				M_SetFPBoxFromIntBox(fpbox, box);

				validcount++;

				for (i = 0; i < llcnt; i++)
				{
					if (markline[i]) { hitline[lloop[i]] = 4; lines[lloop[i]].validcount = validcount; }
				}

				if (   M_CompareFPBox(fpbox, bs->sibounds)
					&& LS_MarkSubsecsInAreaGeneral(fs, hitline, llcnt, lloop, box, -1, MSIA_NONE))
				{
					int a, m;

					// Check all lines contained
					for (a = i = 0; i < hitssecptr; i++)
					{
						subsector_t *ss = &subsectors[hitssec[i]];
						seg_t		*sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ( !(ld = sg->linedef) || ld->validcount == validcount) continue;

							ld->validcount = validcount;
							if (ld->frontsector != fs && ld->backsector != fs) { a = 0; break; }
							if (origTEX[ld - lines] & 0x11) { a = 0; break; }
							a++;
						}
					}

					if (a)
					{
						for (i = 0; i < hitssecptr; i++)
						{
							Setr3dsslfsec(&subsectors[hitssec[i]], bs - sectors, 26010);
						}
					}
				}

				for (i = 0; i < llcnt; i++)
				{
					if (markline[i]) hitline[lloop[i]] = 1;
				}
			}
			else
			{
				hitline[lloop[0]] = 1;
			}

			goto retry;
		}

		free(markline);
		free(lloop);
	}

	free(lnum);
}

static int subfn_Find1sLinesInBox(sector_t *sec, byte *mrkline, int *bbox, int *loopptr, int *looplist, int *nsec)
{
	int		midx = meanV(bbox[BOXLEFT], bbox[BOXRIGHT]);
	int		midy = meanV(bbox[BOXTOP ], bbox[BOXBASE ]);
	int		xl, xh, yl, yh, x, y;
	int		count = 0;

	M_getBlockOffsets(bbox, 0, &xl, &xh, &yl, &yh);

	validcount++;

	*nsec = -1;

	for (x = xl ; x <= xh ; x++)
	{
		for (y = yl ; y <= yh ; y++)
		{
			const long	*list;
			int			offset;

			if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
				continue;

			offset = y*bmapwidth + x;

			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				if (!(lines[*list].ml_flags & ML_TWOSIDED)
					&& lines[*list].frontsector != sec
					&& lines[*list].validcount != validcount)
				{
					int v1 = lines[*list].v1 - vertexes;
					int v2 = lines[*list].v2 - vertexes;

					lines[*list].validcount = validcount;

					if (vtxcheck[v1] == 1 && vtxcheck[v2] == 1
						&& !P_PointOnLineSide(midx, midy, &lines[*list]))
					{
						if (*nsec < 0)
							*nsec = lines[*list].frontsector - sectors;
						else if (*nsec != lines[*list].frontsector - sectors)
							return 0;
						looplist[(*loopptr)++] = *list;
						mrkline[*list] = 1;
						count++;
					}
				}
			}
		}
	}

	return count;
}

static void subfn_CheckLowFloorLoops(byte *mrkline, int *looplist)
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& sec->ceilingpic == skyflatnum
			&& !SecCtrl(i)
			&& !(sec->specFlags & (SF_SECCLOSED|SF_MOVSECTOR|SF_NOLOTEX|SF_NOHITEX))
			&& !IsLiquidFloorFlat(sec))
		{
			int k, m, valid = true, count = 0, bbox[4], loopptr = 0, nsec, ring_count;

			CLR_vtxcheck();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->ml_flags & ML_TWOSIDED)
				{
					if (ld->frontsector == ld->backsector
						|| origTEX[ld - lines] & 0x22
						|| vtxarray[ld->v1 - vertexes] < 2
						|| vtxarray[ld->v2 - vertexes] < 2)
					{
						valid = false;
						break;
					}

				}

				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;

				if (   vtxcheck[ld->v1 - vertexes] > 2
					|| vtxcheck[ld->v2 - vertexes] > 2)
				{
					valid = false;
					break;
				}
			}
			if (!valid) continue;

			M_ClearBox(bbox);

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (   vtxcheck[ld->v1 - vertexes] == 1
					&& vtxcheck[ld->v2 - vertexes] == 1)
				{
					count = 0;
					break;
				}

				M_AddLineToBox(bbox, ld);

				looplist[loopptr++] = ld - lines;
				mrkline[ld - lines] = 1;

				if (   vtxcheck[ld->v1 - vertexes] == 1
					|| vtxcheck[ld->v2 - vertexes] == 1)
				{
					count++;
				}
			}

			if (!count || !loopptr || count & 1) continue;

			// Find any 1's lines in another sector that bridge the gaps
			if (subfn_Find1sLinesInBox(sec, mrkline, bbox, &loopptr, looplist, &nsec) == count/2
				&& subfn_CheckForRingInList(mrkline, &loopptr, looplist, sec, looplist[0], &ring_count, bbox, NULL)
				&& ring_count == sec->linecount + count/2
				&& LS_MarkSubsecsInAreaGeneral(sec, mrkline, loopptr, looplist, bbox, nsec, MSIA_NONE))
			{
				int *hiloop = (int*)M_Malloc(numlines*sizeof(int), 57);
				int hilpptr = 0, csecnum;

				validcount++;
				Setsecdone3(sec, 28035);

				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t		*sg = &segs[ss->firstline];
					int			samesec = false, diff;

					Reset_SSEC(hitssec[k], 50001);

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						line_t *ld = sg->linedef;

						if (!ld) continue;

						if (ld->frontsector == ld->backsector)
						{
							samesec = true;
						}

						if (ld->validcount == validcount) continue;

						ld->validcount = validcount;

						if (ld->ml_flags & ML_TWOSIDED)
						{
							Setspec(ld, -1, 28036);

							if (   Vtfl3(ld)
								&& (ld->frontsector->ceilingpic == skyflatnum)
								 ^ (ld->backsector->ceilingpic == skyflatnum)
								&& (diff = ld->frontsector->ceilingheight
								 - ld->backsector->ceilingheight) != 0
								&& ((diff > 0 && ld->backsector->ceilingpic != skyflatnum)
								 || (diff < 0 && ld->frontsector->ceilingpic != skyflatnum)))
							{
								if (!(origTEX[ld - lines] & 0x44))
									Forcetype(ld, R3D_39, 28037);

								hiloop[hilpptr++] = ld - lines;
							}
							else
								Settype(ld, R3D_3D, 28038);
						}
					}

					if (samesec)
					{
						Setr3dssfs(ss, i, 28039);
						Setr3dsscs(ss, i, 28040);
						// v220-16 added SSF_NOCHANGE
						ss->ssflags |= (SSF_NOCHANGEFS|SSF_NOCHANGE);
					}
				}

				if (hilpptr && LS_SetSSECSinLoop(NULL, &csecnum, hilpptr, hiloop, MSIA_NONE))
				{
					for (k = 0; k < hitssecptr; k++)
						Setr3dsscs(&subsectors[hitssec[k]], csecnum, 28041);
				}

				M_Free(hiloop);
			}
		}
	}
}

// E.g Ambush - area bounded by S819
void LS_CheckUnclosedHighTexSectors()
{
	sector_t	*sec;
	byte		*mrkline  = (byte*)M_Calloc(numlines, 279);
	int			*looplist = (int *)M_Malloc(numlines * sizeof(int), 58);
	int			i;

	subfn_CheckLowFloorLoops(mrkline, looplist);

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i]
			&& !SecCtrl(i)
			&& !(sec->specFlags & (SF_SECCLOSED|SF_MOVSECTOR|SF_NOHITEX))
			&& sec->specFlags & SF_NOLOTEX
			&& sec->ceilingpic != skyflatnum
			&& !IsLiquidFloorFlat(sec))
		{
			int k, loopptr = 0;

			memset(mrkline, 0, numlines);

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];
				int lnum = ld - lines;

				if (ld->ml_flags & ML_TWOSIDED
					&& (ld->frontsector == ld->backsector
					|| origTEX[ld - lines] & 0x33
					|| !(origTEX[ld - lines] & 0x44)))
				{
					loopptr = 0;
					break;
				}

				mrkline[lnum] = 1;
				looplist[loopptr++] = lnum;
			}

			if (loopptr < 3) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				int		bbox[4], ring_count, nsec;
				int		lnum = sec->lines[k] - lines;

				if (mrkline[lnum] != 1) continue;

				if (subfn_CheckForRingInList(mrkline, &loopptr, looplist, sec, lnum, &ring_count, bbox, &nsec)
					&& nsec >= 0)
				{
					int m, n, valid = false;

					if (!LS_MarkSubsecsInAreaGeneral(sec, mrkline, loopptr, looplist, bbox, nsec, MSIA_NONE))
					{
						continue;
					}

					Setsecdone3(sec, 28042);

					validcount++;

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];
						seg_t		*sg = &segs[ss->firstline];

						for (n = 0; n < ss->linecount; n++, sg++)
						{
							line_t *ld = sg->linedef;

							if (!ld
								|| ld->validcount == validcount
								|| mrkline[ld - lines])
							{
								continue;
							}

							ld->validcount = validcount;

							if (!(ld->ml_flags & ML_TWOSIDED))
							{
								valid = false;
								break;
							}

							if (ld->frontsector == ld->backsector)
							{
								valid = true;
								continue;
							}

							if (   ld->frontsector->floorheight == ld->backsector->floorheight
								&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight
								&& !(ld->frontsector->specFlags & SF_MOVCEIL)
								&& !(ld->backsector->specFlags & SF_MOVCEIL))
							{
								valid = true;
								continue;
							}


							valid = false;
							break;
						}
					}

					if (valid)
					{
						for (m = 0; m < hitssecptr; m++)
						{
							subsector_t *ss = &subsectors[hitssec[m]];
							seg_t		*sg = &segs[ss->firstline];

							Reset_SSEC(hitssec[m], 50002);

							Setr3dsscs(ss, i, 28043);
							ss->ssflags |= SSF_NOCHANGECS;

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (sg->linedef && mrkline[sg->linedef - lines] == 4)
								{
									SetLFGSecSet(sg->linedef, 28044);
								}
							}
						}
					}
				}
			}
		}
	}
	M_Free(mrkline);
	M_Free(looplist);
}
