//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_missingtex.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
//
// DESCRIPTION: 1) checks texture validity with lines
//              2) sets line type (and removes the will to live)
//
//-----------------------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "r_data.h"
#include "r3d_edit.h"

// Save passing these
RDBOOL	texerror, texwarn;

char	*texerrfix  = {"  Tex error fixed: line "};

typedef struct
{
	RDBOOL	nomove_flag;
	int		ceilh_f, floorh_f;
	int		ceilh_b, floorh_b;
	byte	txn;
	byte	align;
} mt_var;

static mt_var mt;

//=======================================================================

static void subfn_ReportTexError(int line, int side1, int side2, int t1, int t2)
{
	ma_inf.misstexfixed++;

	if (!texerror) return;

	pr_OFinsert("%s%5i - side %i: missing", texerrfix, line, side1);
	if (t1) pr_OFdirect("   top  "); else pr_OFinsert(" bottom ");
	if (t2 >= 0)
	{
		pr_OFinsert("texture replaced with side %i's", side2);
		if (t2 == 0) pr_OFdirect(" bottom ");
		if (t2 == 1) pr_OFdirect("   mid  ");
		if (t2 == 2) pr_OFdirect("   top  ");
	}
	pr_OFdirect("texture.\n");
}

void subfn_ReportInvalidTexError(int line, int side, int t)
{
	ma_inf.invalidtex++;

	if (!texerror) return;

	pr_OFinsert("%s%5i - side %i: invalid", texerrfix, line, side);

	if (t == 0) pr_OFdirect(" bottom ");
	if (t == 1) pr_OFdirect("   mid  ");
	if (t == 2) pr_OFdirect("   top  ");

	pr_OFdirect("texture.\n");
}

// Check if missing textures can be substituted
static void subfn_SubstituteTex(line_t *ld)
{
	side_t	*sd1 = &sides[ld->sidenum[0]];
	side_t	*sd2 = &sides[ld->sidenum[1]];
	int		i = ld - lines;
	byte	lsflags = (byte)(ld->mtflags & R3D_MISTEXANY);

	if (sd1->sector->specFlags & SF_MOVSECTOR ||
		sd2->sector->specFlags & SF_MOVSECTOR)
		return;

	if (lsflags & R3D_MISLTEX1)	// No bottom texture side 1
	{
		if (sd1->midtexture)
		{
			sd1->bottomtexture = sd1->midtexture;
			sd1->texFlags |= T_LOW;
			subfn_ReportTexError(i, 1, 1, 0, 1);
			Clr_mtloflags(ld, 11001);
			Settype(ld, R3D_3E, 11001);
		}
		else
			Settype(ld, R3D_1A, 11002);
	}
	else if (lsflags & R3D_MISLTEX2)	// No bottom texture side 2
	{
		if (sd2->midtexture)
		{
			sd2->bottomtexture = sd2->midtexture;
			sd2->texFlags |= T_LOW;
			subfn_ReportTexError(i, 2, 2, 0, 1);
			Clr_mtloflags(ld, 11003);
			Settype(ld, R3D_3E, 11003);
		}
		else
			Settype(ld, R3D_1A, 11004);
	}

	if (lsflags & R3D_MISHTEX1)	// No top texture side 1
	{
		if (sd1->midtexture)
		{
			sd1->toptexture = sd1->midtexture;
			sd1->texFlags |= T_HI;
			Clr_mthiflags(ld, 11003);
			Settype(ld, R3D_3E, 11005);
			subfn_ReportTexError(i, 1, 1, 2, 1);
		}
		else
			Settype(ld, R3D_1A, 11006);
	}
	else if (lsflags & R3D_MISHTEX2)	// No top texture side 2
	{
		if (sd2->midtexture)
		{
			sd2->toptexture = sd2->midtexture;
			sd2->texFlags |= T_HI;
			subfn_ReportTexError(i, 2, 2, 2, 1);
			Clr_mthiflags(ld, 11004);
			Settype(ld, R3D_3E, 11007);
			Clr_mthiflags(ld, 11005);
			Settype(ld, R3D_3E, 11008);
		}
		else if (sd1->bottomtexture &&
			sd1->bottomtexture == sd1->toptexture)
		{
			sector_t *s = ld->frontsector;
			// Value needed is height * 2
			int height2 = texArray[sd1->toptexture]->height << (FRACBITS + 1);
			// An MBF hack (might be <= for all I know)
			if (!(ld->ml_flags & ML_DONTPEGTOP) &&
				height2 == s->ceilingheight - s->floorheight)
				Setr3dcs(ld->backsector, s - sectors, 11011);
			Clr_mthiflags(ld, 11006);
			Settype(ld, R3D_3E, 11009);
		}
		else
			Settype(ld, R3D_1A, 11010);
	}
}

RDBOOL CmpFloorsAndCeilings(line_t *ld)
{
	sector_t *fsec = ld->frontsector;
	sector_t *bsec = ld->backsector;

	if (!bsec) return false;

	return !(secspflags[fsec - sectors] & (SECSPF_CTRLLINES|SECSPF_HASACTION)) &&
		!(secspflags[bsec - sectors] & (SECSPF_CTRLLINES|SECSPF_HASACTION)) &&
		fsec->heightsec < 0 &&
		bsec->heightsec < 0 &&
		fsec->floorpic == bsec->floorpic  &&
		fsec->ceilingpic == bsec->ceilingpic &&
		fsec->lightlevel == bsec->lightlevel;
}

//============================================================
//
//	Analyse missing, middle and invalid textures
//	and set line mtflags and tyflags accordingly.
//
//============================================================

static RDBOOL lfsecSet(line_t *ld)
{
	if (ld->mtflags & R3D_MISLTEX1)
		return lfsecarray[ld->backsector - sectors] >= 0;

	return lfsecarray[ld->frontsector - sectors] >= 0;
}

static void texWarn(line_t *ld, int hi, int side)
{
	ma_inf.unrequiredtex++;
	if (side > 2) ma_inf.unrequiredtex++;

	if (texwarn)
		pr_OFinsert("  Tex warning: line %5i - %s tex %s unnecessary\n",
		ld - lines, hi ? "hi" : "lo",
		side == 0 ? "side 1" : side == 1 ? "side 2" : "both sides");
}

static RDBOOL CheckUnwantedTextureRefs(line_t *ld, side_t *sd1, side_t *sd2)
{
	// Check for types that indicate ceiling sec inheritance
	if (ld->frontsector == ld->backsector)
	{
		if (Vtxs2(ld) && (mt.txn & 0x44) == 0x44)
		{
			return false;
		}
	}
	// added v220-9
	else if (mt.nomove_flag && mt.floorh_f == mt.floorh_b)
	{
		if (mt.txn == 0x02 && mt.ceilh_b == mt.floorh_b)
		{
			sd1->midtexture = 0;
			mt.txn = 0;
			origTEX[ld - lines] = 0;
			return false;
		}

		if (mt.txn == 0x20 && mt.ceilh_f == mt.floorh_f)
		{
			sd2->midtexture = 0;
			mt.txn = 0;
			origTEX[ld - lines] = 0;
			return false;
		}
	}

	if (mt.nomove_flag)
	{
		if (Vtxs2(ld) && ld->lnflags2 & LNF2_OWNEQUALTOTWO)
		{
			sector_t *fs = ld->frontsector, *bs = ld->backsector;

			if (  (mt.txn & 0x66) == 0x04
				&& fs->ceilingpic == skyflatnum
				&& bs->ceilingpic != skyflatnum
				&& bs->ceilingheight > fs->ceilingheight)
			{
				sd1->toptexture = 0;
				mt.txn &= ~0x04;
				origTEX[ld - lines] = mt.txn;
				texWarn(ld, 1, 0);
				return false;
			}

			// v220-4 remove unwanted high textures
			if (  (mt.txn & 0x66) & 0x44
				&& bs->ceilingheight == fs->ceilingheight
				&& !(fs->specFlags & SF_MOVSECTOR)
				&& !(bs->specFlags & SF_MOVSECTOR))
			{
				if ((mt.txn & 0x66) == 0x04)
				{
					sd1->toptexture = 0;
					mt.txn &= ~0x04;
					origTEX[ld - lines] = mt.txn;
					texWarn(ld, 1, 0);
					return false;
				}

				if ((mt.txn & 0x66) == 0x40)
				{
					sd2->toptexture = 0;
					mt.txn &= ~0x40;
					origTEX[ld - lines] = mt.txn;
					texWarn(ld, 2, 0);
					return false;
				}
			}
		}

		if (IsLiquidFloorFlat(ld->frontsector))
		{
			if (mt.txn == 0x01
				&& ld->frontsector->floorpic == ld->backsector->floorpic
				&& mt.floorh_f > mt.floorh_b
				&& mt.ceilh_f == mt.ceilh_b
				&& Vtxs2(ld))
			{
				checkULTEX(ld, 1);
				sd1->bottomtexture =
				sd2->bottomtexture = 0;
				origTEX[ld - lines] = 0;
				texWarn(ld, 0, 0);
				return false;
			}

			// v220-10
			if (mt.txn == 0x10
				&& ld->frontsector->floorpic == ld->backsector->floorpic
				&& mt.floorh_f < mt.floorh_b
				&& mt.ceilh_f == mt.ceilh_b
				&& Vtxs2(ld))
			{
				checkULTEX(ld, 1);
				sd1->bottomtexture =
				sd2->bottomtexture = 0;
				origTEX[ld - lines] = 0;
				texWarn(ld, 0, 0);
				return false;
			}

			if (mt.txn == 0x40
				// rev69 accept a line like 10sector map17 2736
				// reject one like AV map 15 2722.
				&& Vtxs2(ld)	// rev69
				&& IsLiquidFloorFlat(ld->backsector)
				&& mt.floorh_f == mt.floorh_b
				&& mt.ceilh_f < mt.ceilh_b)
			{
				Settype(ld, R3D_33, 11014);
				return true;
			}
			if (mt.txn == 0x45
				&& !IsLiquidFloorFlat(ld->backsector)
				&& mt.floorh_f < mt.floorh_b
				&& mt.ceilh_f  > mt.ceilh_b)
			{
				Settype(ld, R3D_33, 11015);
				return true;
			}
		}

		if (mt.txn & 0x01 && !(mt.txn & 0x54)
			&& (sd1->bottomtexture || sd2->bottomtexture)
			&& mt.floorh_f > mt.floorh_b
			&& (mt.ceilh_b - mt.floorh_b) & FRACUNIT)
		{
			checkULTEX(ld, 1);
			sd1->bottomtexture =
			sd2->bottomtexture = 0;
			origTEX[ld - lines] = (mt.txn & 0x66);
			texWarn(ld, 0, 0);
			return false;
		}
		if (mt.txn & 0x10 && !(mt.txn & 0x45)
			&& (sd1->bottomtexture || sd2->bottomtexture)
			&& mt.floorh_b > mt.floorh_f
			&& (mt.ceilh_f - mt.floorh_f) & FRACUNIT)
		{
			checkULTEX(ld, 1);
			sd1->bottomtexture =
			sd2->bottomtexture = 0;
			origTEX[ld - lines] = (mt.txn & 0x66);
			texWarn(ld, 1, 0);
			return false;
		}

		if (mt.txn == 0x04
			&& sd1->toptexture)	// i.e. not AASHITTY
		{
			vertexowner_t *own1 = vertexowners + (ld->v1 - vertexes);
			vertexowner_t *own2 = vertexowners + (ld->v2 - vertexes);

			// If connected to more than two sectors then
			// texture is superfluous
			if (own1->num > 2 && own2->num > 2
				&& mt.floorh_f == mt.floorh_b
				&& mt.ceilh_f  == mt.ceilh_b)
			{
				sd1->toptexture = 0;
				mt.txn = 0;
				origTEX[ld - lines] = 0;
				texWarn(ld, 1, 0);
				return false;
			}
		}
	}
	
	if (!mt.nomove_flag)
	{
		// Added v2.1.00 build_48
		// Deals with invalid lo tex's causing rendering errors
		if (   vtxflags[ld->v1 - vertexes] == 1
			&& vtxflags[ld->v2 - vertexes] == 1)
		{
			int m1 = SF_MOVSECTOR|SF_MOVINVALID;
			int m2 = SF_MOVFLOOR|SF_MOVINVALID;

			if (  !(ld->backsector->specFlags & SF_MOVSECTOR)
				&& (ld->frontsector->specFlags & m1) == m2
				&& origTEX[ld - lines] == 0x01)
			{
				checkULTEX(ld, 1);
				sides[ld->sidenum[0]].bottomtexture = 0;
				sides[ld->sidenum[0]].texFlags = 0;
				origTEX[ld - lines] = 0;
				subfn_ReportInvalidTexError(ld - lines, 0, 0);
			}
			else if (  !(ld->frontsector->specFlags & SF_MOVSECTOR)
				&& (ld->backsector->specFlags & m1) == m2
				&& origTEX[ld - lines] == 0x10)
			{
				sides[ld->sidenum[1]].bottomtexture = 0;
				sides[ld->sidenum[1]].texFlags = 0;
				origTEX[ld - lines] = 0;
				subfn_ReportInvalidTexError(ld - lines, 1, 0);
			}
		}

		// b211-0
		// Deals with invalid hi and lo tex's causing rendering errors
		if (   ld->frontsector == ld->backsector
			&& !(mt.txn & 0x22)
			&& mt.txn & 0x55)
		{
			checkULTEX(ld, 3);
			sides[ld->sidenum[0]].bottomtexture = 0;
			sides[ld->sidenum[0]].toptexture = 0;
			sides[ld->sidenum[0]].texFlags = 0;
			sides[ld->sidenum[1]].bottomtexture = 0;
			sides[ld->sidenum[1]].toptexture = 0;
			sides[ld->sidenum[1]].texFlags = 0;
			mt.txn = 0;
			origTEX[ld - lines] = 0;
			return false;
		}

		// Fix lines like 7624 in cchest2 map24
		// Line must be connected to a single sided line one end.
		// The texture is taken from a line connected to the other.
		if (mt.txn & 0x11
			|| ld->frontsector->floorheight == ld->backsector->floorheight
			|| (ld->frontsector->specFlags & SF_MOVSECTOR
			&&  ld->backsector->specFlags & SF_MOVSECTOR)
			|| !(subfn_FindLineConnected(ld, 0)))
		{
			return false;
		}

		// Check connected lines
		if (ld->frontsector->specFlags & SF_MOVFLOOR)
		{
			line_t *lc = subfn_FindLineConnected(ld, 18);
			if (lc != NULL)
			{
				mt.txn |= 0x01;
				origTEX[ld - lines] |= 0x01;
				sd1->bottomtexture = sides[lc->sidenum[0]].bottomtexture;
				sd1->texFlags |= T_LOW;
				if (texerror)
				{
					pr_OFinsert("%s%5i - missing lo tex side 0 has been given lotex from line %i\n",
								texerrfix, ld - lines, lc - lines);
				}
				ma_inf.misstexfixed++;
			}
			return false;
		}
		else
		{
			line_t *lc = subfn_FindLineConnected(ld, 19);
			if (lc != NULL)
			{
				mt.txn |= 0x10;
				origTEX[ld - lines] |= 0x10;
				sd2->bottomtexture = sides[lc->sidenum[1]].bottomtexture;
				sd2->texFlags |= T_LOW;
				if (texerror)
				{
					pr_OFinsert("%s%5i - missing lo tex side 1 has been given lotex from line %i\n",
								texerrfix, ld - lines, lc - lines);
				}
				ma_inf.misstexfixed++;
			}
			return false;
		}
		return false;
	}

	// Remove superfluous texture references to simplify analysis.
	// Note: only done if both textures are present each side for
	// the upper pair or lower pair.  Each pair is mutually exclusive.
	// Note: Only one being present in a pair is significant.
	if (ld->frontsector == ld->backsector)
	{
		// rev 72 moved test inside, added sky test
		if (mt.txn & 0x55
			&& !(mt.txn & 0x44 && ld->frontsector->ceilingpic == skyflatnum))
		{
			byte ntxn = mt.txn & 0x22;

			if ((mt.txn & 0x11) == 0x11)
			{
				checkULTEX(ld, 1);
				sd1->bottomtexture =
				sd2->bottomtexture = 0;
				texWarn(ld, 0, 3);
			}
			else
				ntxn |= (mt.txn & 0x11);

			if ((mt.txn & 0x44) == 0x44)
			{
				checkULTEX(ld, 2);
				sd1->toptexture =
				sd2->toptexture = 0;
				texWarn(ld, 1, 3);
			}
			else
				ntxn |= (mt.txn & 0x44);

			origTEX[ld - lines] = mt.txn = ntxn;
		}
		return false;
	}
	
	if (!(mt.txn & 0x22))
	{
		RDBOOL lotest = sd1->bottomtexture && sd2->bottomtexture;
		RDBOOL hitest = sd1->toptexture && sd2->toptexture;

		// Remove unwanted textures
		if (lotest && mt.floorh_f == mt.floorh_b)
		{
			checkULTEX(ld, 1);
			origTEX[ld - lines] = mt.txn = mt.txn &= ~0x11;
			sd1->bottomtexture =
			sd2->bottomtexture = 0;
			sd1->texFlags &= ~T_LOW;
			sd2->texFlags &= ~T_LOW;

			if (mt.ceilh_f != mt.ceilh_b)
			{
				if (mt.ceilh_f > mt.ceilh_b &&
					sd2->toptexture)
				{
					origTEX[ld - lines] = mt.txn = mt.txn &= ~0x40;
					sd2->toptexture = 0;
					sd2->texFlags &= ~T_HI;
					texWarn(ld, 1, 1);
				}
				if (mt.ceilh_f < mt.ceilh_b &&
					sd1->toptexture)
				{
					checkULTEX(ld, 2);
					origTEX[ld - lines] = mt.txn = mt.txn &= ~0x04;
					sd1->toptexture = 0;
					sd1->texFlags &= ~T_HI;
					texWarn(ld, 1, 0);
				}
			}
		}

		if (hitest && mt.ceilh_f == mt.ceilh_b)
		{
			checkULTEX(ld, 2);
			origTEX[ld - lines] = mt.txn = mt.txn &= ~0x44;
			sd1->toptexture =
			sd2->toptexture = 0;
			sd1->texFlags &= ~T_HI;
			sd2->texFlags &= ~T_HI;

			if (mt.floorh_f != mt.floorh_b)
			{
				if (mt.floorh_f < mt.floorh_b &&
					sd2->bottomtexture)
				{
					origTEX[ld - lines] = mt.txn = mt.txn &= ~0x10;
					sd2->bottomtexture = 0;
					sd2->texFlags &= ~T_LOW;
					texWarn(ld, 0, 1);
				}
				if (mt.floorh_f > mt.floorh_b &&
					sd1->bottomtexture)
				{
					checkULTEX(ld, 1);
					origTEX[ld - lines] = mt.txn = mt.txn &= ~0x01;
					sd1->bottomtexture = 0;
					sd1->texFlags &= ~T_LOW;
					texWarn(ld, 0, 0);
				}
			}
		}

		if (lotest && mt.ceilh_f != mt.ceilh_b)
		{
			if (mt.ceilh_f > mt.ceilh_b &&
				sd2->toptexture)
			{
				origTEX[ld - lines] = mt.txn = mt.txn &= ~0x40;
				sd2->toptexture = 0;
				sd2->texFlags &= ~T_HI;
				texWarn(ld, 1, 1);
			}
			if (mt.ceilh_f < mt.ceilh_b &&
				sd1->toptexture)
			{
				origTEX[ld - lines] = mt.txn = mt.txn &= ~0x04;
				sd1->toptexture = 0;
				sd1->texFlags &= ~T_HI;
				texWarn(ld, 1, 0);
			}
		}

		if (hitest && mt.floorh_f != mt.floorh_b)
		{
			if (mt.floorh_f < mt.floorh_b &&
				sd2->bottomtexture)
			{
				origTEX[ld - lines] = mt.txn = mt.txn &= ~0x10;
				sd2->bottomtexture = 0;
				sd2->texFlags &= ~T_LOW;
				texWarn(ld, 0, 1);
			}
			if (mt.floorh_f > mt.floorh_b &&
				sd1->bottomtexture)
			{
				checkULTEX(ld, 1);
				origTEX[ld - lines] = mt.txn = mt.txn &= ~0x01;
				sd1->bottomtexture = 0;
				sd1->texFlags &= ~T_LOW;
				texWarn(ld, 0, 0);
			}
		}
	}
	else if (mt.txn == 0x22 &&
		ld->ml_flags & ML_DONTPEGBOTTOM &&
		mt.floorh_f != mt.floorh_b &&
		mt.ceilh_f == mt.ceilh_b)
	{
		return true;
	}

	// Check textures are on correct side.
	// NOTE subfn_FindLineConnected detects whether a missing tex line is connected to
	// a same sided line.  If so then the reversal is correct as this indicates a 
	// special effect such as found in AV map 19 sector 188 (the beam ends).
	// If not then it is considered an error.

	if (!(mt.txn & 0x22) &&	// rev41, was else if
		!Vtfl3(ld) &&	// Must not be connected each side to another miss tex line
		(mt.txn & 0x44) != 0x44 &&		// batman mainly
		(mt.txn & 0x11) != 0x11 &&		// batman mainly
		subfn_FindLineConnected(ld, 10) == NULL)
	{
		if (mt.ceilh_f != mt.ceilh_b &&
			mt.txn & 0x44)
		{
			int found_err = false;
			if (mt.ceilh_f < mt.ceilh_b)
			{
				if ((mt.txn & 0x44) == 0x04)
				{
					sd2->toptexture = sd1->toptexture;
					sd1->toptexture = 0;
					origTEX[ld - lines] = mt.txn = (mt.txn & 0x33)|0x40;
					found_err = true;
				}
			}
			else
			{
				if ((mt.txn & 0x44) == 0x40)
				{
					sd1->toptexture = sd2->toptexture;
					sd2->toptexture = 0;
					origTEX[ld - lines] = mt.txn = (mt.txn & 0x33)|0x04;
					found_err = true;
				}
			}

			if (found_err)
			{
				RDBOOL needed = ld->frontsector->ceilingpic != skyflatnum
						|| ld->backsector->ceilingpic != skyflatnum;

				if (texerror)
				{
					if (needed)
						pr_OFinsert("%s%5i - hi tex on wrong side.\n", texerrfix, ld - lines);
					else
						pr_OFinsert("%s%5i - hi tex superfluous in skysector and on wrong side.\n", texerrfix, ld - lines);
				}

				if (needed)
					ma_inf.misstexfixed++;
			}
		}

		if (mt.floorh_f != mt.floorh_b &&
			mt.txn & 0x11)
		{
			int found_err = false;
			if (mt.floorh_f > mt.floorh_b)
			{
				if ((mt.txn & 0x11) == 0x01)
				{
					checkULTEX(ld, 1);
					sd2->bottomtexture = sd1->bottomtexture;
					sd1->bottomtexture = 0;
					origTEX[ld - lines] = mt.txn = (mt.txn & 0x66)|0x10;
					found_err = true;
				}
			}
			else
			{
				if ((mt.txn & 0x11) == 0x10)
				{
					sd1->bottomtexture = sd2->bottomtexture;
					sd2->bottomtexture = 0;
					origTEX[ld - lines] = mt.txn = (mt.txn & 0x66)|0x01;
					found_err = true;
				}
			}

			if (found_err)
			{
				if (texerror)
					pr_OFinsert("%s%5i - lo tex on wrong side.\n", texerrfix, ld - lines);
				ma_inf.misstexfixed++;
			}
		}
	}
	return false;
}

static void setR3D_1Dssecs(sector_t *s, int num)
{
	sector_t	*needed = UT_FindLowestCeilingSector(s);
	int			c, i, k, bbox[4];

	if (needed && needed->ceilingpic != skyflatnum)
	{
		for (c = i = 0; i < s->linecount; i++)
		{
			if (!(s->lines[i]->ml_flags & ML_TWOSIDED)) return;

			linearray[c++] = s->lines[i] - lines;
		}
		
		if (nf_chkLineListRing(c, bbox, 0, NULL))
		{
			for (i = 0; i < s->linecount; i++)
			{
				int			lnum = s->lines[i] - lines;
				ssline_t	*sl = &sslines[lnum];

				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->ssvalidcount == ss_validcount) continue;

					ss->ssvalidcount = ss_validcount;

					if (ss->lcsec < 0)
					{
						int		x = (int)(ss->midpoint.x * FRACUNIT);
						int		y = (int)(ss->midpoint.y * FRACUNIT);

						if (M_ComparexyBox(x, y, bbox))
						{
							Setr3dsslcsec(ss, needed - sectors, num);
						}
					}
				}
			}
		}
	}
}

// should only be called with a 2's line
void subfn_SetLineMTFlags(line_t *ld, RDBOOL batman)
{
	int		origtf1, origtf2;
	RDBOOL	skyflag;
	RDBOOL	misslowtest;
	short	txS1, txS2;

	side_t *sd1 = &sides[ld->sidenum[0]], *sd2;

	if (   origTEX[ld - lines] == 0x55
		&& ld->backsector != ld->frontsector
		&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
		&& !(ld->backsector->specFlags & SF_MOVSECTOR))
	{
		ld->lnflags2 |= LNF2_MISSINGSPEC;
	}

	// Fixes HOM caused by misplaced texture
	if (!(ld->ml_flags & ML_TWOSIDED))
	{
		if (!sd1->toptexture &&
			!sd1->midtexture &&
			sd1->bottomtexture)
		{
			sd1->midtexture = sd1->bottomtexture;
			sd1->bottomtexture = 0;
			sd1->texFlags = 2;
			if (texerror)
				pr_OFinsert("%s%5i - midtex in bottom tex.\n", texerrfix, ld - lines);
			ma_inf.misstexfixed++;
		}
		return;
	}

	if ( (sd2 = &sides[ld->sidenum[1]]) < 0) return;	// insurance

	mt.nomove_flag = !LineHasMovSectors(ld);
	skyflag =  sd1->sector->ceilingpic == skyflatnum
			&& sd2->sector->ceilingpic == skyflatnum;

	// Get rid of invalid low textures with SF_MOVFIXED
	if (   origTEX[ld - lines] & 0x11
		&& (ld->frontsector->specFlags & SF_MOVFIXED)^(ld->backsector->specFlags & SF_MOVFIXED))
	{
		int txn = ld->frontsector->specFlags & SF_MOVFIXED ? 0x10 : 0x01;

		if (origTEX[ld - lines] & txn)
		{
			if (txn & 0x10)
			{
				sd2->texFlags &= ~T_LOW;
				sd2->bottomtexture = 0;
				ld->mtflags &= ~R3D_INVLTEX2;
			}
			else
			{
				sd1->texFlags &= ~T_LOW;
				sd1->bottomtexture = 0;
				ld->mtflags &= ~R3D_INVLTEX1;
			}

			origTEX[ld - lines] &= ~txn;
		}
	}

	if (ld->frontsector == ld->backsector)
	{
		int txn = origTEX[ld - lines] & 0x22;

		if (txn && txn != 0x22)
		{
			if (txn == 0x02)
				sd1->lsiflags |= LS_NOGLOW;
			else
				sd2->lsiflags |= LS_NOGLOW;
		}
	}

	// Get rid of all AASHITTY tex's with mov sectors if
	// no mid tex or other tex's are set - rev94
	if (ld->frontsector != ld->backsector
	//	&& !mt.nomove_flag
		&& origTEX[ld - lines] & 0x55 
		&& !(origTEX[ld - lines] & 0x22)
		&& !sd1->toptexture
		&& !sd1->bottomtexture
		&& !sd2->toptexture
		&& !sd2->bottomtexture)
	{
		// Don't report as error as author probably just wanted
		// to suppress any missing tex reports in an editor.
		sd1->texFlags = 0;
		sd2->texFlags = 0;
		origTEX[ld - lines] = 0;
	}

	mt.txn = origTEX[ld - lines];

	if (batman && (!mt.nomove_flag || skyflag) && mt.txn & 0x55 && !(mt.txn & 0x22))
	{
		// Batman is a very Doom engine specific TC and leads to problems
		// with most ports.  The main culprit is the liberal use of STONE3
		if (ld->frontsector->floorheight == ld->backsector->floorheight)
		{
			if (   sd1->bottomtexture
				&& !_stricmp(texArray[sd1->bottomtexture]->name, "STONE3"))
			{
				sd1->bottomtexture = 0;
				sd1->texFlags &= ~T_LOW;
				mt.txn &= ~0x01;
			}
			if (sd2->bottomtexture && !_stricmp(texArray[sd2->bottomtexture]->name, "STONE3"))
			{
				sd2->bottomtexture = 0;
				sd2->texFlags &= ~T_LOW;
				mt.txn &= ~0x10;
			}
		}

		if (skyflag || ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
		{
			if (sd1->toptexture && !_stricmp(texArray[sd1->toptexture]->name, "STONE3"))
			{
				sd1->toptexture = 0;
				sd1->texFlags &= ~T_HI;
				mt.txn &= ~0x04;
			}
			if (sd2->toptexture && !_stricmp(texArray[sd2->toptexture]->name, "STONE3"))
			{
				sd2->toptexture = 0;
				sd2->texFlags &= ~T_HI;
				mt.txn &= ~0x40;
			}
		}
		origTEX[ld - lines] = mt.txn;
	}

	mt.floorh_f = sd1->sector->heightsec >= 0 ? sectors[sd1->sector->heightsec].floorheight   : sd1->sector->floorheight;
	mt.ceilh_f  = sd1->sector->heightsec >= 0 ? sectors[sd1->sector->heightsec].ceilingheight : sd1->sector->ceilingheight;
	mt.floorh_b = sd2->sector->heightsec >= 0 ? sectors[sd2->sector->heightsec].floorheight   : sd2->sector->floorheight;
	mt.ceilh_b  = sd2->sector->heightsec >= 0 ? sectors[sd2->sector->heightsec].ceilingheight : sd2->sector->ceilingheight;

	// added v220-16
	if (skyflag && mt.nomove_flag && mt.txn & 0x22)
	{
		if (mt.txn & 0x02 && mt.floorh_b == mt.ceilh_b && mt.floorh_f < mt.floorh_b)
		{
			sd1->midtexture = 0;
			sd1->texFlags &= ~T_HI;
			mt.txn &= ~0x02;
			origTEX[ld - lines] = mt.txn;
			subfn_ReportInvalidTexError(ld - lines, 0, 1);
		}

		if (mt.txn & 0x20 && mt.floorh_f == mt.ceilh_f && mt.floorh_b < mt.floorh_f)
		{
			sd2->midtexture = 0;
			sd2->texFlags &= ~T_HI;
			mt.txn &= 0x20;
			origTEX[ld - lines] = mt.txn;
			subfn_ReportInvalidTexError(ld - lines, 1, 1);
		}
	}

	// Large floor diff lines with no texture.
	if (   mt.nomove_flag
		&& !origTEX[ld - lines]
		&& abs(mt.floorh_f - mt.floorh_b) > 1600*FRACUNIT
		&& !(sd1->sector->floorpic == skyflatnum
		&&   sd2->sector->floorpic == skyflatnum)
		&& (mt.ceilh_f == mt.ceilh_b || skyflag) )
	{
		if (mt.floorh_b > mt.floorh_f)
		{
			line_t *lc;

			if ( (lc = subfn_FindLineConnected(ld, 23)) != NULL
				&& sides[lc->sidenum[0]].bottomtexture)
			{
				sides[ld->sidenum[0]].bottomtexture = sides[lc->sidenum[0]].bottomtexture;
				sides[ld->sidenum[0]].texFlags |= T_LOW;
				origTEX[ld - lines] |= 0x01;
				if (texerror)
					pr_OFinsert("MAP ERROR: LINE %i miss lotex taken from line %i\n",ld-lines,lc-lines);
				ma_inf.misstexfixed++;
				return;
			}
		}
		else
		{
			line_t *lc;

			if ( (lc = subfn_FindLineConnected(ld, 24)) != NULL
				&& lc->ml_flags & ML_TWOSIDED
				&& sides[lc->sidenum[1]].bottomtexture)
			{
				sides[ld->sidenum[1]].bottomtexture = sides[lc->sidenum[1]].bottomtexture;
				sides[ld->sidenum[1]].texFlags |= T_LOW;
				origTEX[ld - lines] |= 0x10;
				if (texerror)
					pr_OFinsert("MAP ERROR: LINE %i miss lotex taken from line %i\n",ld-lines,lc-lines);
				ma_inf.misstexfixed++;
				return;
			}
		}
	}

	if (ld->frontsector->heightsec < 0 &&
		ld->backsector->heightsec < 0)
	{
		if (CheckUnwantedTextureRefs(ld, sd1, sd2)) return;
	}
	else if (mt.nomove_flag
		&& ld->frontsector->heightsec >= 0
		&& ld->backsector->heightsec < 0
		&& sectors[ld->frontsector->heightsec].ceilingheight
		!= ld->backsector->ceilingheight
		&& mt.txn & 0x01
		&& !(mt.txn & 0x06))
	{
		int v1 = ld->v1 - vertexes;
		int v2 = ld->v2 - vertexes;
		int n1 = vtxarray[v1];
		int n2 = vtxarray[v2];
		
		if (   (n1 == 2 && n2 > 2 && vertexowners[v1].num == 2 && vertexowners[v2].num > 2)
			|| (n2 == 2 && n1 > 2 && vertexowners[v2].num == 2 && vertexowners[v1].num > 2)
			|| (n2 >  2 && n1 > 2 && vertexowners[v2].num >  2 && vertexowners[v1].num > 2))
		{
			line_t *lc;

			if ( (lc = subfn_FindLineConnected(ld, 21)) != NULL)
			{
				sides[ld->sidenum[0]].toptexture = sides[lc->sidenum[0]].toptexture;
				sides[ld->sidenum[0]].texFlags |= T_HI;
				origTEX[ld - lines] |= 0x04;
				if (texerror || R3D_edit)
					pr_OFinsert("MAP ERROR: LINE %i miss hitex taken from line %i\n",ld-lines,lc-lines);
				ma_inf.misstexfixed++;
			}
		}
	}
	txS1 = sd1->texFlags;
	txS2 = sd2->texFlags;

	if (txS1 == T_MID && txS2 == T_MID)
	{
		if (sd1->midtexture == 
			sd2->midtexture)
		{
			ld->tyflags |= TYF_MIDTEXLINE;
			if (ld->frontsector == ld->backsector)
				ld->frontsector->specFlags |= SF_HASMIDTEXL;
		}
	}

	if (txS1 & T_MID && txS2 & T_MID)
	{
		if (sd1->rowoffset < 0 &&
			sd2->rowoffset < 0)
			ld->tyflags |= TYF_MIDTEXNEGY;
		if (sd1->rowoffset > 0 &&
			sd2->rowoffset > 0)
			ld->tyflags |= TYF_MIDTEXPOSY;
	}

	// Fix map error Apr 2002
	if (ld->tyflags & TYF_MIDTEXLINE &&
		!(ld->tyflags & (TYF_MIDTEXNEGY|TYF_MIDTEXPOSY)) &&
		ld->frontsector == ld->backsector &&
		((!sd1->rowoffset &&
		   sd2->rowoffset < 0) ||
		 (!sd2->rowoffset &&
		   sd1->rowoffset < 0)))
	{
		if (sd1->rowoffset < 0)
			sd2->rowoffset = sd1->rowoffset;
		else
			sd1->rowoffset = sd2->rowoffset;

		if (texerror)
			pr_OFinsert("%s%5i - incorrect offsets.\n", texerrfix, ld - lines);

		ma_inf.yofferror++;
	}

	if (!(mt.txn & 0x22))
		ld->tyflags |= TYF_NOMIDTEX;

	if (mt.txn & 0x01 && !sd1->bottomtexture)
	{
		ld->backsector->specFlags |= SF_HASORIGLOTEX;
	}

	if (mt.txn & 0x04 && !sd1->toptexture)
	{
		ld->backsector->specFlags |= SF_HASORIGHITEX;
	}

	if (mt.txn & 0x10 && !sd2->bottomtexture)
	{
		ld->frontsector->specFlags |= SF_HASORIGLOTEX;
	}

	if (mt.txn & 0x40 && !sd2->toptexture)
	{
		ld->frontsector->specFlags |= SF_HASORIGHITEX;
	}
// End set line types

// Identify invalid textures
	origtf1  = (mt.txn & 0x07) == 5 &&
			   !(ld->backsector->specFlags & SF_MOVSECTOR) &&
			   !sd1->toptexture && !sd1->bottomtexture;

	origtf2  = (mt.txn>>4) == 5 &&
			   !(ld->frontsector->specFlags & SF_MOVSECTOR) &&
			   !sd2->toptexture && !sd2->bottomtexture;

	if (mt.floorh_f >= mt.floorh_b && sd1->bottomtexture) ld->mtflags |= R3D_INVLTEX1;
	if (mt.floorh_f <= mt.floorh_b && sd2->bottomtexture) ld->mtflags |= R3D_INVLTEX2;
	if (mt.ceilh_f <= mt.ceilh_b && sd1->toptexture) ld->mtflags |= R3D_INVHTEX1;
	if (mt.ceilh_f >= mt.ceilh_b && sd2->toptexture) ld->mtflags |= R3D_INVHTEX2;

	// Delete invalid textures for certain line types to prevent flats problems.
	if (sd1->sector == sd2->sector && !sd1->midtexture && !sd2->midtexture)
	{
		if (ld->mtflags & R3D_INVLTEX1)
		{
			checkULTEX(ld, 1);
			sd1->bottomtexture = 0;
		}
		if (ld->mtflags & R3D_INVHTEX1 &&
			ld->mtflags & (R3D_INVHTEX2|R3D_INVLTEX1|R3D_INVLTEX2))
		{
			checkULTEX(ld, 2);
			sd1->toptexture = 0;
		}
		if (ld->mtflags & R3D_INVLTEX2)
		{
			sd2->bottomtexture = 0;
		}
		if (ld->mtflags & R3D_INVHTEX2 &&
			ld->mtflags & (R3D_INVHTEX1|R3D_INVLTEX1|R3D_INVLTEX2))
		{
			sd2->toptexture = 0;
		}
	}

	if (ld->mtflags & R3D_INVTEXANY
		&& mt.nomove_flag
		&& sd1->sector != sd2->sector
		&& sd1->rowoffset == sd2->rowoffset
		&& sd1->midtexture
		&& sd1->midtexture == sd2->midtexture
		&& mt.floorh_f == mt.floorh_b
		&& mt.ceilh_f  == mt.ceilh_b
		&& (!sd1->toptexture	|| ld->mtflags & R3D_INVHTEX1)
		&& (!sd2->toptexture	|| ld->mtflags & R3D_INVHTEX2)
		&& (!sd1->bottomtexture || ld->mtflags & R3D_INVLTEX1)
		&& (!sd2->bottomtexture || ld->mtflags & R3D_INVLTEX2))
	{
		// Delete all tex flags so that, for analysis purposes, when
		// testing using flags then apparently no texture exists.
		// Leave the texture values themselves as may be needed by
		// the renderer.
		Clr_allmtflags(ld, 11001);
		ld->tyflags = 0;
		Settype(ld, R3D_3D, 11017);
		return;
	}

	// Update/set texFlags

	if (!origtf1)
	{
		txS1 = 0;
		if (sd1->bottomtexture) txS1 |= T_LOW;
		if (sd1->midtexture   ) txS1 |= T_MID;
		if (sd1->toptexture   ) txS1 |= T_HI;
		sd1->texFlags = txS1;
	}

	if (!origtf2)
	{
		txS2 = 0;
		if (sd2->bottomtexture) txS2 |= T_LOW;
		if (sd2->midtexture   ) txS2 |= T_MID;
		if (sd2->toptexture   ) txS2 |= T_HI;
		sd2->texFlags = txS2;
	}

	if (txS1 == (T_LOW|T_HI)
		&& txS2 == (T_LOW|T_HI)
		&& mt.ceilh_f - mt.floorh_f >= 8*FRACUNIT
		&& mt.ceilh_b - mt.floorh_b >= 8*FRACUNIT
		&& !Vtxs2(ld))
	{
		Settype(ld, R3D_3E, 11018);
		if (ld->tag
			&& sd1->sector == sd2->sector
			&& Vtxs1(ld))
		{
			Setspec(ld, LS_GetAdjacentLineSecnum(ld, 7), 11003);
		}
		return;
	}

	if (mt.ceilh_f == mt.ceilh_b
		&& txS1 == T_HI
		&& txS2 == T_HI)
	{
		if (mt.ceilh_f - mt.floorh_f == FRACUNIT
			&& mt.floorh_f > mt.floorh_b)
		{
			return;
		}

		if (mt.ceilh_b - mt.floorh_b == FRACUNIT &&
			mt.floorh_b > mt.floorh_f)
		{
			return;
		}
	}

	if (!txS1 && !txS2)
		ld->spflags |= SPF_NOTEX;

	if (!sd1->bottomtexture &&
		!sd2->bottomtexture &&
		!sd1->toptexture &&
		!sd2->toptexture)
		ld->spflags |= SPF_NOULTEX;

	// Check for map inconsistencies
	if (ld->spflags & SPF_NOTEX &&
		ld->frontsector != ld->backsector &&
		(ld->backsector->linecount < 2 ||
		 ld->frontsector->linecount < 2))
	{
		pr_OFinsert("  Map warning: line %i references an invalid sector %i\n",
			ld - lines, ld->backsector->linecount < 2 ? ld->backsector - sectors : ld->frontsector - sectors);
		Settype(ld, R3D_0F, 11019);
		return;
	}

	misslowtest   = !(ld->mtflags & R3D_INVLOTEX) &&
					((txS1 == T_MID && txS2 == (T_MID|T_LOW)) ||
					 (txS1 == (T_MID|T_LOW) && txS2 == T_MID));

	// Check for valid definitions of missing textures
	if (sd1->sector != sd2->sector &&
		(mt.floorh_f != mt.floorh_b || mt.ceilh_f != mt.ceilh_b))
	{
		int cmpf = Vtxs2(ld) ? 8*FRACUNIT : 7*FRACUNIT;
		int cmpc = Vtxs2(ld) ? 9*FRACUNIT : FRACUNIT;

		if (!(txS1 & T_LOW) &&
			mt.floorh_f < mt.floorh_b &&
			(mt.floorh_b - mt.floorh_f > cmpf ||
			ld->backsector->specFlags & SF_MOVSECTOR))
		{
			if (!(misslowtest && !IsLiquidFloorFlat(ld->backsector) && IsLiquidFloorFlat(ld->frontsector)))
			{			
				ld->mtflags |= R3D_MISLTEX1;	//0x10
			}
		}
		else if (!(txS2 & T_LOW) &&
			mt.floorh_f > mt.floorh_b &&
			(mt.floorh_f - mt.floorh_b > cmpf ||
			ld->frontsector->specFlags & SF_MOVSECTOR))
		{
			if (!(misslowtest && !IsLiquidFloorFlat(ld->frontsector) && IsLiquidFloorFlat(ld->backsector)))
			{
				ld->mtflags |= R3D_MISLTEX2;	//0x40
			}
		}
		if (!(ld->frontsector->ceilingpic == skyflatnum &&
			  ld->backsector->ceilingpic == skyflatnum) &&
			  (!mt.txn || abs(mt.ceilh_f - mt.ceilh_b) <= MAX_SECHEIGHT))
		{
			if (!(txS1 & T_HI) && mt.ceilh_f - mt.ceilh_b >= cmpc)
			{
				ld->mtflags |= R3D_MISHTEX1;	//0x20
			}
			else if (!(txS2 & T_HI) && mt.ceilh_b - mt.ceilh_f >= cmpc )
			{
				ld->mtflags |= R3D_MISHTEX2;	//0x80
			}
		}
	}
	// Use array to store values prior to anything being reset
	mtarray[ld - lines] = ld->mtflags;

	if (!mt.txn &&
		ld->mtflags & R3D_MISTEXANY &&
		!LineHasMovSectors(ld))
	{
		if ((ld->mtflags == R3D_MISHTEX1 ||
			 ld->mtflags == R3D_MISHTEX2) &&
			abs(mt.ceilh_f - mt.ceilh_b ) < 8*FRACUNIT)
		{
			Clr_allmtflags(ld, 11002);
		}
		if ((ld->mtflags == R3D_MISLTEX1 ||
			 ld->mtflags == R3D_MISLTEX2) &&
			(abs(mt.floorh_f - mt.floorh_b) < 8*FRACUNIT ||
			 lfsecSet(ld)))
		{
			if (lfsecSet(ld))
				Settype(ld, R3D_3F, 11020);
			Clr_allmtflags(ld, 11020);
		}
	}
}

void LS_SetLineMTFlags()
{
	sector_t	*fs, *bs;
	byte		*testmtf = (byte*)calloc(numlines, 1);
	int			i;
	
	for (i = 0; i < numlines; i++)
	{
		if (   lines[i].ml_flags & ML_TWOSIDED
			&& lines[i].specType < R3D_40					// 3D lines are not changed
			&& !(lineflags[i] & (LFG_NEWSET|LFG_3DTYPE))	// b211-0
			&& !(lines[i].spflags & SPF_CONTROLLINE)
			&& !(lines[i].lnflags2 & LNF2_SLOPESECLN))		// $slopes
		{
			subfn_SetLineMTFlags(&lines[i], ge_inf.batman);
			testmtf[i] = 1;
		}
	}

	for (i = 0; i < numlines; i++)
	{
		if (testmtf[i])
		{
			// b211-0 reset any flags set with sky both sided
			if (lines[i].backsector)
			{
				if (lines[i].mtflags & (R3D_MISLTEX1|R3D_MISLTEX2))
				{
					fs = lines[i].frontsector, bs = lines[i].backsector;

					if (fs->floorpic == skyflatnum && bs->floorpic == skyflatnum)
					{
						lines[i].mtflags &= ~(R3D_MISLTEX1|R3D_MISLTEX2);
					}
				}
				if (lines[i].mtflags & (R3D_MISHTEX1|R3D_MISHTEX2))
				{
					fs = lines[i].frontsector, bs = lines[i].backsector;

					if (fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum)
					{
						lines[i].mtflags &= ~(R3D_MISHTEX1|R3D_MISHTEX2);
					}
				}
			}
		}
	}

	free(testmtf);
}

//============================================================================

// Abandon all hope ye who enter here.

// v220-10	chunks removed

void LS_SetMissingTexLineType()
{
	line_t	*ld, *li;
	int		i;
	RDBOOL	ht, ht1, ht2, tst1;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
	if ( !(ld->ml_flags & ML_TWOSIDED)
		|| ld->specType != R3D_00
		|| lineflags[i] & (LFG_LINESSECSET|LFG_NEWSET|LFG_3DTYPE))
	{
		continue;
	}

	if (!origTEX[i])
	{
		// v220-10 retain this for the moment
		if (ld->frontsector == ld->backsector)
		{
			line_t *ldc;

			if ((vtxarray[ld->v1 - vertexes] == 1)^
				(vtxarray[ld->v2 - vertexes] == 1) &&
				(vtxarray[ld->v1 - vertexes] > 2 ||
				 vtxarray[ld->v2 - vertexes] > 2) &&
				(ldc = subfn_FindLineConnected(ld, 2)) != NULL)
			{
				Setspec(ld, ldc->frontsector - sectors, 11027);
				Settype(ld, R3D_3F, 11159);
			}
			else
			{
				Settype(ld, R3D_01, 11160);
			}
		}
		else
		if (  (ld->frontsector->ceilingpic == skyflatnum)^(ld->backsector->ceilingpic == skyflatnum)
			&& ld->frontsector->heightsec >= 0
			&& ld->backsector->heightsec >= 0
			&& ld->frontsector->heightsec != ld->backsector->heightsec
			&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
			&& !(ld->backsector->specFlags  & SF_MOVSECTOR)
			&& ld->frontsector->floorheight == ld->backsector->floorheight
			&& (ht = ld->frontsector->ceilingheight - ld->backsector->ceilingheight)
			&& abs(ht) > 8*FRACUNIT
			&& ld->frontsector->ceilingheight - ld->frontsector->floorheight > 8*FRACUNIT
			&& ld->backsector->ceilingheight  - ld->backsector->floorheight  > 8*FRACUNIT)
		{
			if (ht > 0)
			{
				if (ld->backsector->ceilingpic == skyflatnum) Settype(ld, R3D_2C, 11131);
			}
			else if (ld->frontsector->ceilingpic == skyflatnum)	// ht < 0
				Settype(ld, R3D_2C, 11133);
		}
	}
	else
	{
		side_t	*sd1 = &sides[ld->sidenum[0]];
		side_t	*sd2 = &sides[ld->sidenum[1]];

		int		missingtex = 0;
		int		nummisstex = 0;

		short	lsflags = ld->mtflags;

		byte	txn = origTEX[i];

		fixed_t	floorh_f = sd1->sector->floorheight;
		fixed_t	ceilh_f  = sd1->sector->ceilingheight;
		fixed_t	floorh_b = sd2->sector->floorheight;
		fixed_t ceilh_b  = sd2->sector->ceilingheight;

		RDBOOL	nomidtex_flag = !sd1->midtexture && !sd2->midtexture;
		RDBOOL	notex_flag	  = (ld->spflags & SPF_NOTEX) > 0;
		RDBOOL	noULtex_flag  =	(ld->spflags & SPF_NOULTEX) > 0;

		RDBOOL	fsky_flag	  = ld->frontsector->ceilingpic == skyflatnum;
		RDBOOL	bsky_flag	  = ld->backsector->ceilingpic == skyflatnum;
		RDBOOL	sky_flag	  = fsky_flag && bsky_flag;

		RDBOOL	vtx2_flag	  = Vtxs2(ld);

		RDBOOL	vtx3_flag	  = Vtfl3(ld);

		RDBOOL	noanim_flag	  = !IsLiquidFloorFlat(ld->frontsector) &&
								!IsLiquidFloorFlat(ld->backsector);

		RDBOOL	nomove_flag	  = !LineHasMovSectors(ld);

		if (ld->mtflags & R3D_MISLTEX1) missingtex  = 1, nummisstex++;
		if (ld->mtflags & R3D_MISLTEX2) missingtex  = 2, nummisstex++;
		if (ld->mtflags & R3D_MISHTEX1) missingtex |= 4, nummisstex++;
		if (ld->mtflags & R3D_MISHTEX2) missingtex |= 8, nummisstex++;

		if ((missingtex == 4 || missingtex == 8)
			&& nomove_flag
			&& !(ld->mtflags & R3D_INVTEXANY)
			&& vtx2_flag
			&& vtx3_flag
			&& txn & 0x11
			&& !(txn & 0x66))
		{
			continue;
		}

// Doors 1
		if (!nomove_flag && txn == 0x20 &&
			ld->backsector->specFlags & SF_MOVCEIL &&
			!(ld->frontsector->specFlags & SF_MOVSECTOR) &&
			ceilh_f > ceilh_b &&
			ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec >= 0 &&
			sectors[ld->backsector->heightsec].ceilingheight ==
			ld->frontsector->ceilingheight &&
			sectors[ld->backsector->heightsec].floorheight ==
			ld->frontsector->floorheight)
		{
			Settype(ld, R3D_13, 11021);
			continue;
		}

		if ((missingtex & (4|8))
			&& txn
			&& !(txn & 0x22)	// added rev69
			&& !vtx2_flag
			&& !fsky_flag
			&& !bsky_flag
			&& subfn_FindLineConnected(ld, 0) != NULL)
		{
			if (Spans1sLines(ld)) continue;

			Settype(ld, R3D_3E, 11022);
			SetLFG(ld, LFG_NHCONNOBSLN, 11002);
		}

		if (ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec < 0)
		{
			if (!nomove_flag)
			{
				if (txn & 0x02 && !(ld->frontsector->specFlags & SF_MOVSECTOR))
				{
					if (ceilh_f > ceilh_b && !sd1->toptexture &&
						ld->backsector->specFlags & SF_MOVCEIL)
					{
						if (sky_flag) continue;	// v220-14

						setR3D_1Dssecs(ld->backsector, 11900);

						if (floorh_b > floorh_f)	// $v90
							Settype(ld, R3D_2D, 11023);
						else
							Settype(ld, R3D_1D, 11023);
						continue;
					}
					if (floorh_f < floorh_b && !sd1->bottomtexture &&
						ld->backsector->specFlags & SF_MOVFLOOR)
					{
						if (   sd2->bottomtexture
							// v220-22 add test
							|| !(ld->frontsector->specFlags & SF_NOLOTEX))
						{
							// v220-27 add test
							if (line_t *lc = subfn_FindLineOnLine(ld, 8))
							{
								SetMissinglo(lc, 11024);
								Settype(lc, R3D_03, 11024);
								SetMissinglo(ld, 11024);
								Settype(ld, R3D_03, 11024);
							}
							else
							{
								Settype(ld, R3D_2D, 11024);
							}
						}
						else
							Settype(ld, R3D_1D, 11024);

						continue;
					}
				}

				if (txn & 0x20 && !(ld->backsector->specFlags & SF_MOVSECTOR))
				{
					if (ceilh_f < ceilh_b && !sd2->toptexture &&
						ld->frontsector->specFlags & SF_MOVCEIL)
					{
						if (sky_flag) continue;	// v220-14

						setR3D_1Dssecs(ld->frontsector, 11901);

						if (floorh_f > floorh_b)	// $v90
							Settype(ld, R3D_2D, 11025);
						else
							Settype(ld, R3D_1D, 11025);
						continue;
					}
					if (floorh_f > floorh_b && !sd2->bottomtexture &&
						ld->frontsector->specFlags & SF_MOVFLOOR)
					{
						if (sd1->bottomtexture)	// $v90
							Settype(ld, R3D_2D, 11026);
						else
							Settype(ld, R3D_1D, 11026);
						continue;
					}
				}
			}
// End doors 1

			if (!lsflags && (txn == 0x40 || txn == 0x50))
			{
				if (nomove_flag
					// $added following tests Oct/05
					&& (subfn_FindLineCrossing(ld, LC_NORMAL) != NULL
					|| subfn_FindLineConnected(ld, 22)))
				{
				//	Settype(ld, R3D_3B, 11027);
				}
				continue;
			}
			if (notex_flag
				&& vtx2_flag
				&& !sky_flag
				&& noanim_flag
				&& nomove_flag
				&&
				((lsflags == R3D_MISLTEX1
				  && (floorh_b & FRACUNIT || (ceilh_b - floorh_b >= 8*FRACUNIT)))
				  ||
				 (lsflags == R3D_MISLTEX2 &&
				  (floorh_f & FRACUNIT || (ceilh_f - floorh_f >= 8*FRACUNIT))))
				)
			{
				continue;
			}

			// Added June 2003 - example PAR 1 9 - Sectors 74/75
			if (notex_flag && vtx2_flag &&
				noanim_flag &&
				lsflags == R3D_MISLTEX2 &&
				!(ld->frontsector->floorheight & FRACUNIT) &&
				!(ld->frontsector->ceilingheight & FRACUNIT) &&
				!SecACTN(ld->frontsector - sectors) &&
				!SecACTN(ld->backsector - sectors) &&
				ld->frontsector->ceilingpic != skyflatnum &&	//Assume
				ld->frontsector->ceilingpic ==
				ld->backsector->ceilingpic)
			{
				Settype(ld, R3D_3C, 11029);
				continue;
			}

			// Added June 2003
			if (lsflags == R3D_MISLTEX2)
			{
				if (sd1->texFlags == T_HI)
				{
					continue;
				}
				if (notex_flag && noanim_flag &&
					Vtxs2(ld) &&	// Added v2.1.00 build_30
					!Vtfl1(ld) &&	// Added b211-0 (Wim's map ChillerCastle2)
					ld->backsector->floorheight & FRACUNIT)
				{
						Settype(ld, R3D_2A, 11030);
						continue;
				}
			}

			// Added June 2003 - Map error detect
			if (lsflags == (R3D_MISHTEX2|R3D_INVHTEX1) &&
				sd1->texFlags == T_HI &&
				!sd2->texFlags)
			{
				Settype(ld, R3D_31, 11031);
				subfn_ReportTexError(i, 2, 1, 0, -1);
				continue;
			}

			// Added June 2003 - Map error detect
			if (notex_flag &&
				abs(ld->frontsector->ceilingheight -
				ld->backsector->ceilingheight) <= 4*FRACUNIT &&
				(lsflags == R3D_MISHTEX1 || lsflags == R3D_MISHTEX2) &&
				ld->frontsector->ceilingpic != skyflatnum &&
				ld->backsector->ceilingpic != skyflatnum)
			{
				if (lsflags == R3D_MISHTEX1) subfn_ReportTexError(i, 1, 0, 1, -1);
				if (lsflags == R3D_MISHTEX2) subfn_ReportTexError(i, 2, 0, 1, -1);
				continue;
			}

			// Added May 2003 - Deals with rising structures requiring
			// floor to be rendered at higher of the two floor heights.
			if (nomidtex_flag &&
				ld->frontsector != ld->backsector &&
				ld->frontsector->heightsec < 0 &&
				ld->backsector->heightsec  < 0 &&
				sd1->texFlags & T_HI &&
				sd2->texFlags & T_HI)
			{
				if (lsflags & R3D_INVHTEX1 &&
					!(lsflags & R3D_INVLTEX1) &&
					(lsflags & R3D_MISLTEX1) &&
					secspflags[ld->frontsector - sectors] & SECSPF_RAISEFLR) //Rising floor
				{
					Settype(ld, R3D_1A, 11032);
					continue;
				}

				else if (lsflags & R3D_INVHTEX2 &&
					!(lsflags & R3D_INVLTEX2) &&
					(lsflags & R3D_MISLTEX2) &&
					secspflags[ld->backsector - sectors] & SECSPF_RAISEFLR)
				{
					Settype(ld, R3D_1A, 11033);
					continue;
				}
			}

			if (notex_flag &&
				vtx2_flag &&
				vtx3_flag &&	// added for vile_geo
				lsflags == R3D_MISLTEX2 &&
				ld->backsector->specFlags & SF_SECCLOSED &&
				!IsLiquidFloorFlat(ld->frontsector) &&
				IsLiquidFloorFlat(ld->backsector))
			{
				if (floorh_f - floorh_b >= 4*FRACUNIT && // As above
					floorh_f - floorh_b <= MAX_SECHEIGHT)
				{
					Settype(ld, R3D_32, 11034);
					Setlfsec(ld->frontsector, ld->backsector - sectors, 11001);
				}
				continue;
			}

		}

		ht1 = ceilh_f - floorh_f <= 2*FRACUNIT;
		ht2 = ceilh_b - floorh_b <= 2*FRACUNIT;
		ht  = abs(ceilh_f - floorh_f) >= 8*FRACUNIT &&
			  abs(ceilh_b - floorh_b) >= 8*FRACUNIT;

		if (!ht1 && !ht2 && notex_flag &&
			!SecACTN(ld->frontsector - sectors) &&
			!SecACTN(ld->backsector - sectors) &&
			ld->frontsector->ceilingpic ==
			ld->backsector->ceilingpic)
		{
			// Note: R3D_00 is the default meaning missing walls will be drawn
			//		 in order to deal with map errors
			if (lsflags == R3D_MISLTEX2 &&
				ld->backsector->heightsec < 0 &&
				IsLiquidFloorFlat(ld->backsector) &&
				!IsLiquidFloorFlat(ld->frontsector))
			{
				if (ld->frontsector->specFlags & SF_SECCLOSED)
				{
					if (nomove_flag &&
						ld->frontsector->specFlags & SF_NOHITEX &&
						floorh_f > floorh_b)
					{
						Settype(ld, R3D_2A, 11035);
						Setlfsec(ld->frontsector, ld->backsector - sectors, 11002);
					}
					else if (vtxflags[ld->v1 - vertexes] == 3
						  || vtxflags[ld->v2 - vertexes] == 3)
						Settype(ld, R3D_3F, 11036);
					else
						Settype(ld, R3D_00, 11036);
				}
				else
				{
					subfn_ReportTexError(i, 2, 0, 0, -1);
					if (!Vtxs2(ld))
						Clr_allmtflags(ld, 11001);	//rev54
				}
				Setsecdone(ld->frontsector, 11007);
				Setsecdone(ld->backsector, 11008);
				continue;
			}
			if (lsflags == R3D_MISLTEX1 &&
				ld->frontsector->heightsec < 0 &&
				IsLiquidFloorFlat(ld->frontsector) &&
				!IsLiquidFloorFlat(ld->backsector))
			{
				if (ld->backsector->specFlags & SF_SECCLOSED)
				{
					if (nomove_flag &&
						ld->backsector->specFlags & SF_NOHITEX &&
						floorh_b > floorh_f)
					{
						// do nothing here
					}
					else
						Settype(ld, R3D_3F, 11037);
				}
				else
					subfn_ReportTexError(i, 1, 0, 0, -1);
				Setsecdone(ld->frontsector, 11009);
				Setsecdone(ld->backsector, 11010);
				continue;
			}
		}

		if (!ht1 && ht2 && notex_flag &&
			missingtex == (1|4) &&
			fsky_flag &&
			ld->backsector->ceilingpic != skyflatnum)
		{
			Settype(ld, R3D_37, 11038);
			SetLFG(ld, LFG_FIXED, 11038);	// b211-0
			continue;
		}

		if (ht && notex_flag && !sky_flag &&
			ld->frontsector->ceilingpic ==
			ld->backsector->ceilingpic)
		{
			if (missingtex == (1|8) &&
				ceilh_f - floorh_b < 4*FRACUNIT)
			{
				Settype(ld, R3D_3F, 11039);
				Setspec(ld, ld->frontsector - sectors, 11008);
				continue;
			}

			if (missingtex == (2|4) &&
				ceilh_b - floorh_f < 4*FRACUNIT)
			{
				Settype(ld, R3D_3F, 11040);
				Setspec(ld, ld->frontsector - sectors, 11009);
				continue;
			}
		}

		if (notex_flag && missingtex == 5 && !ht1 && ht2)
		{
			Settype(ld, R3D_3F, 11041);

			if (nomove_flag && vtx2_flag && vtx3_flag)
			{
				Setspec(ld, ld->frontsector - sectors, 11010);
				Setsecdone(ld->frontsector, 11011);
				Setsecdone(ld->backsector, 11012);
			}
			continue;
		}

		if (nummisstex == 1 && ht &&
			vtx3_flag &&
			!notex_flag && nomidtex_flag &&
			noanim_flag)
		{
			if (vtx2_flag)
			{
				// Floor height difference must be > tex height
				int fdiff = ((abs(floorh_f - floorh_b))>>16) -
					texArray[sd1->bottomtexture ?
							 sd1->bottomtexture :
							 sd2->bottomtexture]->height;

				if (missingtex == 1 &&
					sd1->texFlags == T_HI &&
					sd2->texFlags == (T_LOW|T_HI))
				{
					Settype(ld, R3D_30, 11042);
					if (fdiff) secspflags[ld->backsector - sectors] |= SECSPF_FLOORTEXHT;
					continue;
				}

				if (missingtex == 2 &&
					sd1->texFlags == (T_LOW|T_HI) &&
					sd2->texFlags == T_HI)
				{
					Settype(ld, R3D_30, 11043);
					if (fdiff) secspflags[ld->frontsector - sectors] |= SECSPF_FLOORTEXHT;
					continue;
				}

				if (missingtex == 4 &&
					sd1->texFlags == T_LOW &&
					sd2->texFlags == (T_LOW|T_HI))
				{
					if (sd2->bottomtexture == sd2->toptexture)
					{
						Settype(ld, R3D_30, 11044);
						if (fdiff) secspflags[ld->backsector - sectors] |= SECSPF_FLOORTEXHT;
					}
					else
						Settype(ld, R3D_00, 11045);
					continue;
				}

				if (missingtex == 8 &&
					sd1->texFlags == (T_LOW|T_HI) &&
					sd2->texFlags == T_LOW)
				{
					if (sd1->bottomtexture == sd1->toptexture)
					{
						Settype(ld, R3D_30, 11046);
						if (fdiff) secspflags[ld->frontsector - sectors] |= SECSPF_FLOORTEXHT;
					}
					else
						Settype(ld, R3D_00, 11047);
					continue;
				}
			}
			else
			{
				if (missingtex == 1 &&
					sd1->texFlags == T_HI &&
					sd2->texFlags == (T_LOW|T_HI))
				{
					Settype(ld, R3D_1A, 11048);
					continue;
				}

				if (missingtex == 2 &&
					sd1->texFlags == (T_LOW|T_HI) &&
					sd2->texFlags == T_HI)
				{
					Settype(ld, R3D_1A, 11049);
					continue;
				}

				if (missingtex == 4 &&
					sd1->texFlags == T_LOW &&
					sd2->texFlags == (T_LOW|T_HI))
				{
					Settype(ld, R3D_1A, 11050);
					continue;
				}

				if (missingtex == 8 &&
					sd1->texFlags == (T_LOW|T_HI) &&
					sd2->texFlags == T_LOW)
				{
					Settype(ld, R3D_1A, 11051);
					continue;
				}
			}
		}

		if (ht &&
			(sd1->texFlags == (T_LOW|T_HI) && sd2->texFlags == T_LOW)
			||
			(sd1->texFlags == T_LOW && sd2->texFlags == (T_LOW|T_HI))
			)
			continue;

// R3D_1D static
		if (!(secspflags[sd1->sector - sectors] & (SECSPF_CTRLLINES|SECSPF_HASACTION)) &&
			!(secspflags[sd2->sector - sectors] & (SECSPF_CTRLLINES|SECSPF_HASACTION)) &&
			ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec < 0)
		{
			if (ht &&
				sd1->sector != sd2->sector &&
				Vtxs1(ld) &&
				 ceilh_f == ceilh_b &&
				 floorh_f == floorh_b &&
				((sd1->texFlags == T_MID && !sd2->texFlags) ||
				 (sd2->texFlags == T_MID && !sd1->texFlags)))
			{
				Settype(ld, R3D_1D, 11052);
				continue;
			}

			if (nummisstex == 1 &&
				((tst1 = (missingtex == 1 && !ht1 && ht2))>0 ||
				 (missingtex == 2 && !ht2 && ht1)))
			{
				if (sd1->texFlags == T_MID &&
					sd2->texFlags == T_MID)
				{
					if (vtx3_flag)
					{
						Settype(ld, R3D_1D, 11053);
						continue;
					}
				}
				else
				{
					if (tst1)
					{
						if (sd2->texFlags == (T_LOW|T_HI) ||
							(sd2->texFlags == T_ALL && sd1->texFlags & T_MID))
						{
							if (sd1->texFlags & T_MID)
								Settype(ld, R3D_1D, 11054);

							continue;
						}
					}
					else //tst2
					{
						if ((sd1->texFlags == (T_LOW|T_HI) ||
							(sd1->texFlags == T_ALL && sd2->texFlags & T_MID)))
						{
							if (sd2->texFlags & T_MID)
								Settype(ld, R3D_1D, 11055);

							continue;
						}


						if (sd1->texFlags == (T_LOW|T_HI) &&
							sd2->texFlags == (T_HI|T_MID))
						{
							if (sd2->texFlags & T_MID)
								Settype(ld, R3D_1D, 11056);

							continue;
						}
					}
				}
			}
		}
// End R3D_1D static

		if (sky_flag
			&& lsflags == R3D_MISLTEX2
			&& ld->frontsector->floorpic == skyflatnum
			&& floorh_f - floorh_b > MAX_SECHEIGHT)	// rev99n
		{
			Settype(ld, R3D_3F, 11057);
			Setspec(ld, ld->backsector - sectors, 11014);
			continue;
		}

		if (ht && nomidtex_flag &&
			((missingtex == 8 &&
			  sd1->texFlags & T_LOW) ||
			 (missingtex == 4 &&
			  sd2->texFlags & T_LOW)) &&
			Vtxs1(ld))
			continue;

		if (ht && (missingtex == 5 || missingtex == 10))
		{
			if (notex_flag)
			{
				if (vtx2_flag && vtx3_flag)
				{
					if (noanim_flag)
					{
						if (!(missingtex ==  5 &&
							  ld->frontsector->specFlags & SF_MOVSECTOR) &&
							!(missingtex == 10 &&
							  ld->backsector->specFlags  & SF_MOVSECTOR))
						{
							Settype(ld, R3D_2B, 11058);
							if (missingtex == 10 &&
								!secspflags[ld->frontsector - sectors])
							{
								if (!ld->backsector->tag &&		// Guessing - see geo_vile map15 sector 1002
									!ld->backsector->special &&	// ditto
									ld->frontsector->ceilingpic != skyflatnum &&
									ld->backsector->ceilingpic != skyflatnum)
									Setspec(ld, ld->frontsector - sectors, 11015);
								Settype(ld, R3D_3F, 11059);
							}
							continue;
						}
					}
					else if (missingtex == 5 || missingtex == 10)
					{
						if (nomove_flag)
						{
							if (missingtex == 5)
							{
								if (IsLiquidFloorFlat(ld->frontsector))
									Clr_mtloflags(ld, 11007);
								if (fsky_flag)
									Clr_mthiflags(ld, 11008);
							}
							else
							{
								if (IsLiquidFloorFlat(ld->backsector))
									Clr_mtloflags(ld, 11009);
								if (bsky_flag)
									Clr_mthiflags(ld, 11010);
							}
						}
						if (ld->mtflags == lsflags)
						{
							Settype(ld, R3D_2B, 11060);
							continue;
						}
						if (ld->mtflags)
						{
							Settype(ld, R3D_1A, 11061);
							continue;
						}
					}
				}
				// Test for setting R3D_1B type rev34
				{
					int mask = SF_LINESCLOSED|SF_MOVFLOOR|SF_NOHITEX;

					if (!(ld->frontsector->specFlags & SF_MOVSECTOR) &&
						(ld->backsector->specFlags & mask) == mask &&
						!(ld->backsector->specFlags & (SF_MOVLOWER|SF_NOLOTEX)))
					{
						Settype(ld, R3D_1B, 11062);
						continue;
					}
				}
				if (notex_flag
					&& nomove_flag
					&& ld->frontsector->heightsec < 0	// rev41
					&& ld->backsector->heightsec < 0)
				{
					// b211-0 vnum test added
					int vnum1 = ld->v1 - vertexes;
					int vnum2 = ld->v2 - vertexes;

					if (   vtoflags[vnum1] & 1
						&& vtoflags[vnum2] & 1)
					{
						Settype(ld, R3D_00, 11063);
						Clr_allmtflags(ld, 11063);
					}
					else
						Settype(ld, R3D_3F, 11063);
				}
				else
					Settype(ld, R3D_00, 11064);
				continue;
			}
			if (sd1->texFlags == T_MID || sd2->texFlags == T_MID)	// should this be sd2 ?? commented out $v2.0.11
			{
				int vnum1 = ld->v1 - vertexes;
				int vnum2 = ld->v2 - vertexes;

				if (   !(sd1->texFlags & (T_LOW|T_HI))
					&& !(sd2->texFlags & (T_LOW|T_HI))
					&& nomove_flag
					&& !fsky_flag
					&& !bsky_flag)
				{
					if (!Spans1sLines(ld))	// v220-3
					{
						if (sd1->texFlags == T_MID)
						{
							if (ceilh_b  < ceilh_f ) ld->ml_flags |= ML_IGNOREMISSINGHI;
							if (floorh_b > floorh_f) SetMissinglo(ld, 11064);	// v220-6
						}
						if (sd2->texFlags == T_MID)
						{
							if (ceilh_b  > ceilh_f ) ld->ml_flags |= ML_IGNOREMISSINGHI;
							if (floorh_b < floorh_f) SetMissinglo(ld, 11065);	// v220-6
						}
					}
				}

				if (ld->ml_flags & (ML_IGNOREMISSINGHI|ML_IGNOREMISSINGLO))
				{
					if (   vtxarray[vnum1] == 1
						|| vtxarray[vnum2] == 1)
					{
						Settype(ld, R3D_1D, 11065);
					}
					else
					{
						if (   origTEX[ld - lines] == 0x22
							&& ld->frontsector != ld->backsector
							&& (li = subfn_FindLineOnLine(ld, 16)))
						{
							Settype(ld, R3D_03, 11065);
						}
						else
						{
							Settype(ld, R3D_00, 11065);
							// v220-4 reset flags
							if ( !(lineflags[ld - lines] & LFG_OVERRIDE) )
								ld->ml_flags &= ~ML_IGNOREMISSINGHI;
							ResetMissinglo(ld, 11065);
						}
					}
				}
				else if (vtoflags[vnum1] || vtoflags[vnum2])	// b211-0
					Settype(ld, R3D_00, 11066);
				else if (vtxarray[vnum1] == 1 || vtxarray[vnum2] == 1)
					Settype(ld, R3D_1D, 11065);
				else
					Settype(ld, R3D_1B, 11066);
				continue;
			}
		}
			
		// Map error check
		if (ht && (lsflags & R3D_INVTEXANY) == (R3D_INVLTEX2|R3D_INVHTEX2) &&
			Vtxs1(ld) &&
			!sd1->texFlags && sd2->texFlags == (T_LOW|T_HI))
		{
			sd1->toptexture = sd2->toptexture;
			sd2->toptexture = 0;
			sd1->bottomtexture = sd2->bottomtexture;
			sd2->bottomtexture = 0;
			Clr_allmtflags(ld, 11004);
			sd1->texFlags = 5;
			sd2->texFlags = 0;
			ma_inf.misstexfixed++;
			if (texerror)
				pr_OFinsert("%s%5i - textures swapped \n", texerrfix, i);
			continue;
		}

		// Map error check
		if (ht && (lsflags & R3D_INVTEXANY) == (R3D_INVLTEX1|R3D_INVHTEX1) &&
			Vtxs1(ld) &&
			!sd2->texFlags && sd1->texFlags == (T_LOW|T_HI))
		{
			checkULTEX(ld, 3);
			sd2->toptexture = sd1->toptexture;
			sd1->toptexture = 0;
			sd2->bottomtexture = sd1->bottomtexture;
			sd1->bottomtexture = 0;
			Clr_allmtflags(ld, 11005);
			sd2->texFlags = 5;
			sd1->texFlags = 0;
			ma_inf.misstexfixed++;
			if (texerror)
				pr_OFinsert("%s%5i - textures swapped \n", texerrfix, i);
			continue;
		}

		if (lsflags == 0x2C)
		{
			Settype(ld, R3D_17, 11067);
			continue;
		}

		if (lsflags == 0x28 &&
			!sd1->texFlags &&
			sd2->texFlags == (T_LOW|T_HI))
		{
			lsflags = 0x08;
			continue;
		}

		if (notex_flag && ht1 &&
			sd1->sector == sd2->sector &&
			SecACTN(sd1->sector - sectors) &&
			Vtxs1(ld))
		{
			Settype(ld, R3D_3F, 11068);
			Setspec(ld, LS_GetAdjacentLineSecnum(ld, 8), 11016);
			continue;
		}

		if (ht && sd1->texFlags == T_LOW && sd2->texFlags == T_LOW)
		{
			continue;
		}

		if (ht && sd1->texFlags == T_HI && sd2->texFlags == T_HI)
		{
			continue;
		}

		if (ld->frontsector->heightsec < 0 &&
			ld->backsector->heightsec < 0)
		{
			if (ht &&
				(missingtex == 4 || missingtex == 8) &&
				ld->spflags & SPF_NOTEX &&
				Vtxs1(ld))
			{
					Clr_allmtflags(ld, 11006);
					continue;
			}

			if (missingtex == 4 && !ht1 && ht2
				&& ld->tyflags & TYF_MIDTEXLINE)
			{
				if (Vtxs1(ld))
				{
					Clr_allmtflags(ld, 11007);
				}
				Settype(ld, R3D_1D, 11069);
				continue;
			}
		}

		if (ht && missingtex & (1|2) &&
			!(missingtex & (4|8)) &&
			!(lsflags & R3D_INVTEXANY) &&
			sd1->texFlags == T_MID &&
			sd1->texFlags == T_MID &&
			sd1->midtexture ==
			sd2->midtexture &&
			!IsAnimTex(sd1->midtexture))
		{
			Clr_allmtflags(ld, 11008);
			continue;
		}
		if (ht && missingtex & (1|2) &&
			!(missingtex & (4|8)) &&
			ld->mtflags & R3D_INVHITEX &&
			!(sd1->texFlags & T_MID) &&
			!(sd2->texFlags & T_MID))
		{
			continue;
		}

		// General checks

		if (ht && missingtex & (1|2) &&
			!vtx3_flag &&
			!(lsflags & R3D_INVTEXANY) &&
			sd1->texFlags == T_MID &&
			!(sd2->texFlags & (T_LOW|T_HI)) &&
			sd1->midtexture != sd2->midtexture)
		{
			sd1->bottomtexture = sd1->midtexture;
			sd1->texFlags |= T_LOW;
			ma_inf.misstexfixed++;
			if (texerror)
				pr_OFinsert("%s%5i - missing low tex\n", texerrfix, i);
			Clr_allmtflags(ld, 11010);
			continue;
		}

		// Added Apr 2004
		if (ld->mtflags == (R3D_MISHTEX1|R3D_MISLTEX2|R3D_INVLTEX1) &&
			vtx2_flag && vtx3_flag &&
			txn == 0x01 &&
			ld->frontsector->specFlags & SF_MOVSECTOR &&
			!(ld->backsector->specFlags & SF_MOVSECTOR))
		{
			Settype(ld, R3D_0B, 11098);
			continue;
		}

		if (missingtex == 8 &&
			vtx2_flag &&
			vtx3_flag &&
			!SecACTN(ld->frontsector - sectors) &&	//rev60
			!SecACTN(ld->backsector - sectors) &&	//rev60
			!(ld->mtflags & R3D_INVTEXANY) &&
			!sd1->texFlags &&
			sd2->texFlags == T_LOW &&
			ld->frontsector->floorpic ==
			ld->backsector->floorpic &&
			ld->frontsector->ceilingpic ==
			ld->backsector->ceilingpic)
		{
			Setr3dch(ld->backsector, ld->frontsector, 11017);
			if (ld->backsector->ceilinglightsec < 0)
				ld->backsector->ceilinglightsec = ld->frontsector - sectors;

			Settype(ld, R3D_38, 11106);
			continue;
		}

		if (nomidtex_flag &&
			!(vtxflags[ld->v1 - vertexes] > 2 &&
			  vtxflags[ld->v2 - vertexes] > 2) &&
			!(lsflags & R3D_INVTEXANY) &&
			ld->frontsector->ceilingpic != skyflatnum &&
			ld->backsector->ceilingpic != skyflatnum &&
			noanim_flag &&
			abs(ceilh_f - floorh_f) >= 8*FRACUNIT &&
			abs(ceilh_b - floorh_b) >= 8*FRACUNIT)
		{
			if (!sd2->texFlags)
			{
				if ((lsflags == R3D_MISHTEX1 &&
					  sd1->texFlags == T_LOW) ||
					(lsflags == R3D_MISLTEX1 &&
					 sd1->texFlags == T_HI))
				{
					Clr_allmtflags(ld, 11011);
					Settype(ld, R3D_00, 11107);
					continue;
				}
			}
			if (!sd1->texFlags)
			{
				if ((lsflags == R3D_MISHTEX2 &&
					  sd2->texFlags == T_LOW) ||
					(lsflags == R3D_MISLTEX2 &&
					 sd2->texFlags == T_HI))
				{
					Clr_allmtflags(ld, 11012);
					Settype(ld, R3D_00, 11108);
					continue;
				}
			}
		}

		if (nomidtex_flag)
		{
			ht1 = abs(floorh_f - floorh_b) >= 4*FRACUNIT && abs(floorh_f - floorh_b) < 8*FRACUNIT;
			ht2 = abs(ceilh_f  - ceilh_b ) >= 4*FRACUNIT && abs(ceilh_f  - ceilh_b ) < 8*FRACUNIT;

			if ((missingtex == 5 || missingtex == 10) &&
				 ht1 && ht2)
			{
				Settype(ld, R3D_3E, 11109);
				continue;
			}
			if ((missingtex == 1 || missingtex == 2) && ht1)
			{
				Settype(ld, R3D_3E, 11110);
				continue;
			}
			if ((missingtex == 4 || missingtex == 8) && ht2)
			{
				Settype(ld, R3D_3E, 11111);
				continue;
			}
		}

		if (ht && (lsflags & R3D_INVTEXANY) == 0x0B &&
			sd1->texFlags == (T_LOW|T_HI) && sd2->texFlags == T_HI &&
			vtx3_flag)
		{
			Settype(ld, R3D_2A, 11112);
			continue;
		}

		if (ht && (lsflags & R3D_INVTEXANY) == 0x0E &&
			sd1->texFlags == T_HI && sd2->texFlags == (T_LOW|T_HI) &&
			vtx3_flag)
		{
			Settype(ld, R3D_2A, 11113);
			continue;
		}

		if (ht && sd1->sector == sd2->sector &&
			vtxarray[ld->v1 - vertexes] > 1 &&
			vtxarray[ld->v2 - vertexes] > 1 &&
			((sd1->texFlags == T_ALL && sd2->texFlags == (T_LOW|T_HI)) ||
			 (sd2->texFlags == T_ALL && sd1->texFlags == (T_LOW|T_HI))))
		{
			Settype(ld, R3D_2F, 11114);	// Marked for further testing
			continue;
		}

		if (   lsflags & R3D_MISLTEX1
			&& sd1->midtexture
			&& !sd2->midtexture
			&& sd1->toptexture
			&& sd1->sector->ceilingheight > sd2->sector->floorheight
			&& !(sd1->sector->specFlags & SF_MOVSECTOR)
			&& !(sd2->sector->specFlags & SF_MOVSECTOR))
		{
			Settype(ld, R3D_2D, 11001);	// v220-18
			continue;
		}

		if (   lsflags & R3D_MISLTEX2
			&& sd2->midtexture
			&& !sd1->midtexture
			&& sd2->toptexture
			&& sd2->sector->ceilingheight > sd1->sector->floorheight
			&& !(sd1->sector->specFlags & SF_MOVSECTOR)
			&& !(sd2->sector->specFlags & SF_MOVSECTOR))
		{
			Settype(ld, R3D_2D, 11002);	// v220-18
			continue;
		}

		// Map error checks
		if (lsflags == (R3D_INVHTEX1|R3D_MISLTEX1) &&
			sd1->texFlags & T_MID)
		{
			if (nomove_flag)
			{
				subfn_SubstituteTex(ld);
				ld->mtflags &= ~(R3D_INVTEXANY|R3D_MISTEXANY);
			}
			continue;
		}

		if (nomidtex_flag &&
			lsflags == (R3D_MISHTEX1|R3D_MISLTEX1|R3D_INVHTEX2|R3D_INVLTEX2))
		{
			sd1->bottomtexture = sd2->bottomtexture;
			sd1->toptexture    = sd2->toptexture;
			sd2->bottomtexture = 0;
			sd2->toptexture    = 0;
			ld->mtflags &= ~(R3D_MISTEXANY|R3D_INVTEXANY);
			Settype(ld, R3D_3E, 11115);
			continue;
		}
		// End map error checks

		if (!ht1 && !ht2 &&
			(missingtex == 1 ||
			 missingtex == 2 ||
			 ld->frontsector->heightsec >= 0 ||
			 ld->backsector->heightsec  >= 0) &&
			sd1->texFlags == T_MID && sd2->texFlags == T_MID &&
			!sd1->rowoffset && !sd2->rowoffset)
		{
			Settype(ld, R3D_04, 11116);
			continue;
		}

		if (!ht1 && !ht2 &&
			nomidtex_flag &&
			(lsflags & R3D_MISTEXANY) == R3D_MISHTEX2 &&
			sd1->texFlags & T_LOW &&
			!sd2->texFlags)
		{
			if (nomove_flag)
			{
				subfn_SubstituteTex(ld);
			}
			continue;
		}

		if (ht1 && ht2 &&
			sd1->texFlags == T_MID &&
			sd2->texFlags == T_MID)
		{
			if (ld->frontsector == ld->backsector &&
				ld->frontsector->heightsec >= 0)
				Settype(ld, R3D_3E, 11119);

			continue;
		}

// DOORS 2
		if (ht &&
			sd1->texFlags == T_MID &&
			sd2->texFlags == T_MID &&
			sd1->midtexture == sd2->midtexture &&
			IsAnimTex(sd1->midtexture) &&
			vtxarray[ld->v1 - vertexes] > 2 &&
			vtxarray[ld->v2 - vertexes] > 2 &&
			!ld->mtflags &&
			ld->frontsector == ld->backsector &&
			SecACTN(ld->frontsector - sectors) &&
			sd1->rowoffset == sd2->rowoffset)
		{
			Settype(ld, R3D_2D, 11120);
			continue;
		}

		if (!ht1 && ht2 && !ld->mtflags)
		{
			if (sd1->texFlags == T_HI && !sd2->texFlags)
			{
				Settype(ld, R3D_00, 11121);	//Already is - just for clarity (or just in case!)
				continue;
			}
		}

		if (ht1 && !ht2 && floorh_f & FRACUNIT &&
			ld->mtflags == (R3D_MISLTEX2|R3D_MISHTEX2))
		{
			if (!sd1->texFlags && sd2->texFlags == T_MID)
			{
				Settype(ld, R3D_1D, 11122);
				continue;
			}
		}

		if (ht1 && !ht2 && ceilh_f & FRACUNIT &&
			ld->mtflags & R3D_INVHITEX)
		{
			if (sd2->texFlags == T_MID)
			{
				Settype(ld, R3D_1D, 11123);
				continue;
			}
		}

		if (!ht1 && !ht2 &&
			!(ld->mtflags & R3D_INVTEXANY) &&
			missingtex == (1|4) &&
			sd1->texFlags == T_MID &&
			sd2->texFlags == T_MID &&
			((vtxarray[ld->v1 - vertexes] == 1 &&
			  vtxarray[ld->v2 - vertexes] > 2) ||
			 (vtxarray[ld->v1 - vertexes] > 2 &&
			  vtxarray[ld->v2 - vertexes] == 1)))
		{
			Settype(ld, R3D_1D, 11124);
			continue;
		}

		if (!ht1 && ht2 && floorh_b & FRACUNIT &&
			ld->mtflags == (R3D_MISLTEX2|R3D_MISHTEX1))
		{
			if (sd1->texFlags == T_MID && sd2->texFlags == T_MID)
			{
				Settype(ld, R3D_1D, 11125);
				continue;
			}
		}
		// End doors

		if (nomove_flag)
		{
			if (!ht2 && !sd1->texFlags &&
				sd2->texFlags == T_MID &&
				lsflags == R3D_MISHTEX2)
			{
				if (   vtoflags[ld->v1 - vertexes] & 1
					&& vtoflags[ld->v2 - vertexes] & 1
					&& floorh_f == ceilh_f
					&& floorh_b < floorh_f
					&& ceilh_b > ceilh_f)
				{
					Settype(ld, R3D_2D, 11125);
					Setspec(ld, ld->backsector - sectors, 11125);
					continue;
				}
			}
			else if (!ht1 && !sd2->texFlags &&
				sd1->texFlags == T_MID &&
				lsflags  == R3D_MISHTEX1)
			{
				if (   vtoflags[ld->v1 - vertexes] & 1
					&& vtoflags[ld->v2 - vertexes] & 1
					&& floorh_b == ceilh_b
					&& floorh_f < floorh_b
					&& ceilh_f > ceilh_b)
				{
					Settype(ld, R3D_2D, 11125);
					Setspec(ld, ld->frontsector - sectors, 11125);
					continue;
				}
			}
		}

		if (!ht1 && ht2 && missingtex == 4 &&
			sd1->texFlags == T_MID && !sd2->texFlags)
		{
			Settype(ld, R3D_1D, 11126);
			continue;
		}

		if (nomidtex_flag && !ht1 && !ht2)
		{
			if ((missingtex == 5 || missingtex == 10))
			{
				if (!nomove_flag ||
					ld->frontsector->heightsec >= 0 ||
					ld->backsector->heightsec >= 0)
					continue;

				if (ld->backsector->ceilingpic != skyflatnum &&
					!Vtxs1(ld))
				{
					if (!sd1->texFlags && !sd2->texFlags)
						Settype(ld, R3D_2B, 11127);
					else
						Settype(ld, R3D_1B, 11128);
				}
				else
				{
					Settype(ld, R3D_1A, 11129);
					Clr_mthiflags(ld, 11011);
				}
				continue;
			}	

			if (nummisstex == 1 && !noULtex_flag)
			{
				if (missingtex == 1 && ceilh_f <= floorh_b)
				continue;
			}
		}

		if (notex_flag)
		{
			if (missingtex == 4 &&
				bsky_flag)
			{
				if (ceilh_b - floorh_b < 8*FRACUNIT)
					Settype(ld, R3D_3F, 11130);
				else
					Settype(ld, R3D_2C, 11131);
				continue;
			}
			if (missingtex == 8 &&
				fsky_flag)
			{
				if (ceilh_f - floorh_f < 8*FRACUNIT)
					Settype(ld, R3D_3F, 11132);
				else
					Settype(ld, R3D_2C, 11133);
				continue;
			}
		}

		if (!ht1 && !ht2 &&
			!(lsflags & R3D_INVTEXANY) &&
			(((missingtex == 1 ||
			   missingtex == 2) &&
			   ceilh_b == ceilh_f)
			||
			 ((missingtex == 4 ||
			   missingtex == 8) &&
			   floorh_f == floorh_b)))
		{
			if ((sd1->midtexture > 0)^(sd2->midtexture > 0))
			{
				// If a liquid flat's used then keep at R3D_00
				continue;
			}
			else if (ld->mtflags & R3D_MISTEXANY && !txn)
			{
				// Seems lines of length 1 or less (will be root 2 on a
				// diagonal to speed up processing) should be ignored with
				// regard to missing tex's.
				if (abs(ld->v1->x - ld->v2->x) <= FRACUNIT &&
					abs(ld->v1->y - ld->v2->y) <= FRACUNIT)
				{
					Clr_allmtflags(ld, 11013);
					continue;
				}
			}
		}

		if (!nomidtex_flag && noULtex_flag && ht1 && ht2 &&
			(sd1->texFlags & (T_LOW|T_HI) && sd2->texFlags & (T_LOW|T_HI)))
		{
			Settype(ld, R3D_0E, 11135);
			Setspec(ld, LS_FindAdjacentSecnum(ld, 2), 11022);
			continue;
		}

		if (sd1->texFlags == T_MID && sd2->texFlags == T_MID)
		{
			if (ht1 && !ht2 && missingtex == 10)
			{
				if (ld->frontsector->heightsec >= 0		// rev96
					&& ld->backsector->heightsec >= 0)
				{
					Settype(ld, R3D_13, 11136);
				}
				else
				{
					Settype(ld, R3D_1D, 11137);
					Setspec(ld, ld->backsector - sectors, 11023);
				}
				continue;
			}

			if (!ht1 && ht2 && missingtex == 5)
			{
				if (ld->frontsector->heightsec >= 0		// rev96
					&& ld->backsector->heightsec >= 0)
				{
					Settype(ld, R3D_13, 11138);
				}
				else
				{
					Settype(ld, R3D_1D, 11139);
					Setspec(ld, ld->frontsector - sectors, 11024);
				}
				continue;
			}
		}

		if ((ht1 || ht2) && nummisstex < 2 &&
			lsflags & (R3D_INVTEXANY|R3D_MISTEXANY))
		{
			if (!lsflags && nomidtex_flag)
				Settype(ld, R3D_1F, 11140);
			else if (!ht1 && lsflags & R3D_INVHTEX1 &&
				sd1->texFlags == T_ALL && !sd2->texFlags)
				Settype(ld, R3D_3E, 11141);
			else if (!nomidtex_flag)
			{
				if (!(lsflags & R3D_INVTEXANY) &&
					noULtex_flag)
				{
					if (sd1->texFlags == T_MID &&
						!sd2->texFlags &&
						lsflags & R3D_MISHTEX2)
						Settype(ld, R3D_12, 11142);
					else if (!sd1->texFlags &&
						sd2->texFlags == T_MID &&
						(lsflags & (R3D_MISLTEX2|R3D_MISHTEX2)) == (R3D_MISLTEX2|R3D_MISHTEX2))
						Settype(ld, R3D_1D, 11143);
					else if (sd1->texFlags == T_MID &&
						sd2->texFlags == T_MID &&
						(lsflags & (R3D_MISLTEX2|R3D_MISHTEX2)) == (R3D_MISLTEX2|R3D_MISHTEX2))
						Settype(ld, R3D_2D, 11144);
					else if (!ht2 && !sd1->texFlags &&
						sd2->texFlags == T_MID &&
						lsflags  == R3D_MISHTEX2)
						Settype(ld, R3D_11, 11145);
					else
					{
						int sn;

						if (Vtxs1(ld))
						{
							Settype(ld, R3D_1D, 11146);
							Clr_allmtflags(ld, 11014);
						}
						else if (!(sd1->texFlags == T_MID &&
							!sd2->texFlags &&
							missingtex == 2)
							&& !(sd2->texFlags == T_MID &&
							!sd1->texFlags &&
							missingtex == 1))
						{
							if (ceilh_f == floorh_f ||
								ceilh_b == floorh_b)	// Added Mar 2004
								continue;
							Settype(ld, R3D_0E, 11147);
						}
						else
							continue;

						sn = LS_FindAdjacentSecnum(ld, 3);

						// v220-4 do not set specsec with certain conditions (test added)
						if (sn != ld->frontsector - sectors && sn != ld->backsector - sectors)
							Setspec(ld, sn, 11025);
					}
				}
				else if (!ht1 && sd1->texFlags == (T_HI|T_MID) &&
					!(lsflags & (R3D_MISTEXANY|R3D_INVTEXANY)) &&
					(ld->ml_flags & (ML_DONTPEGBOTTOM|ML_DONTPEGTOP)) ==
					(ML_DONTPEGBOTTOM|ML_DONTPEGTOP) &&
					!sd2->texFlags)
				{
					Settype(ld, R3D_11, 11148);
				}
				else if (!ht1 && sd1->texFlags == (T_HI|T_MID) &&
					!sd2->texFlags)
				{
					Settype(ld, R3D_1D, 11149); //was R3D_13

					// added
					Setspec(ld, ld->frontsector - sectors, 11026);
					ld->backsector->specFlags |= SF_MASKEDDOOR;
				}
				else if (!ht1 && sd1->texFlags == (T_MID|T_LOW) &&
					sd2->texFlags == T_MID)
				{
					Settype(ld, R3D_2D, 11150);
				}
				else if (lsflags & R3D_INVHTEX1 &&
					sd1->texFlags & T_MID &&
					!sd2->texFlags)
				{
					Settype(ld, R3D_1D, 11151);
				}
			}
			if (!ld->specType &&
				notex_flag &&
				nummisstex == 1 &&
				!(lsflags & R3D_INVTEXANY))
			{
				if (!Vtxs1(ld))
					Settype(ld, R3D_3C, 11152);
				else if (nomove_flag)
				{
					subfn_SubstituteTex(ld);
					continue;
				}
			}
			if (ld->specType)
				continue;
		}

		if ((!(lsflags & (R3D_INVLTEX1|R3D_INVLTEX2)) &&
			(missingtex == 1 ||
			 missingtex == 2))
			 ||
			(!(lsflags & (R3D_INVHTEX1|R3D_INVHTEX2)) &&
			(missingtex == 4 ||
			 missingtex == 8)))
		{
			// Check if a texture substitution can be made.
			if (!(lsflags & R3D_INVTEXANY) &&
				(missingtex &  5 && sd1->texFlags & T_MID) ||
				(missingtex & 10 && sd2->texFlags & T_MID))
			{
				if (sd1->rowoffset < 0 &&
					sd1->rowoffset ==
					sd2->rowoffset &&
					ld->frontsector->ceilingheight ==
					ld->backsector->ceilingheight)
				{
					Settype(ld, R3D_04, 11153);
				}
				else
				if (sd1->rowoffset < 0 &&
					sd1->rowoffset ==
					sd2->rowoffset &&
					ld->ml_flags & ML_DONTPEGBOTTOM &&
					sd1->midtexture ==
					sd2->midtexture &&
					ld->frontsector->ceilingheight ==
					ld->backsector->ceilingheight)
				{
					Settype(ld, R3D_04, 11154);
				}
				else
				{
					if ((txn & 0x22) == 0x22)
					{
						Settype(ld, R3D_04, 11155);
						Clr_mthiflags(ld, 11155);
						Clr_mtloflags(ld, 11155);
						continue;
					}

					if (nomove_flag)
					{
						subfn_SubstituteTex(ld);
						continue;
					}
				}
			}
			else if (!notex_flag &&
				nomove_flag &&
				nomidtex_flag &&
				lsflags & R3D_INVTEXANY)
			{
				if (lsflags == (R3D_MISLTEX2|R3D_INVHTEX2) &&
					sd1->texFlags == T_HI)
					Settype(ld, R3D_1E, 11156);
				else if (lsflags == (R3D_MISLTEX1|R3D_INVHTEX1) &&
					sd2->texFlags == T_HI)
					Settype(ld, R3D_1E, 11157);
				else
					Settype(ld, R3D_1A, 11158);
			}
		}

		if (ld->specType) continue;

// 4)	Set basic and special line types

		if (notex_flag)
		{
			continue;
		}

		if (!ld->tag)
		{
			if (ceilh_b == floorh_b &&
				sd1->texFlags == T_ALL &&
				(!sd2->texFlags || sd2->texFlags == T_MID) &&
				// ceil test added $v113 (exclude COD map05 line 537)
				!(sd2->sector->specFlags & SF_MOVCEIL))
			{
				Settype(ld, R3D_10, 11167);
			}
			// Identify use of backsec sector for ceiling flats.
			else if (nomidtex_flag	 &&
				vtx2_flag			 &&
				floorh_f == floorh_b &&
				ceilh_f < ceilh_b    &&
				!sd1->bottomtexture  &&
				!sd2->bottomtexture  &&
				!sd1->toptexture     &&
				sd2->toptexture)
			{
				Settype(ld, R3D_0B, 11168);
			}
			else if (ceilh_b == floorh_b &&
				lsflags == R3D_MISHTEX1 &&
				sd1->texFlags == T_MID &&
				!sd2->texFlags)
			{
				Settype(ld, R3D_12, 11169);
			}
		}
		else if (!sd1->toptexture &&
			(sd1->midtexture ||
			sd1->bottomtexture) &&
			!sd2->texFlags)
		{
			if ((ld->ml_flags & ML_DONTPEGBOTTOM) &&
				ceilh_f == ceilh_b &&
				floorh_f > floorh_b)
				Settype(ld, R3D_0A, 11170);
			else if ((ld->ml_flags & ML_DONTPEGTOP) &&
				ceilh_f < ceilh_b &&
				floorh_f == floorh_b)
				Settype(ld, R3D_0A, 11171);
		}
		else if (!sd2->toptexture &&
			(sd2->midtexture ||
			sd2->bottomtexture) &&
			!sd1->texFlags)
		{
			if ((ld->ml_flags & ML_DONTPEGBOTTOM) &&
				ceilh_f == ceilh_b &&
				floorh_f != floorh_b)
				Settype(ld, R3D_0A, 11172);
			else if ((ld->ml_flags & ML_DONTPEGTOP) &&
				ceilh_f != ceilh_b &&
				floorh_f == floorh_b)
				Settype(ld, R3D_0A, 11173);
		}

		if (ld->specType) continue;

		if (!sd1->texFlags &&
			sd2->texFlags == T_MID &&
			missingtex == 10)
		{
			Settype(ld, R3D_19, 11174);
		}
		else if (!(lsflags & R3D_INVTEXANY))
		{
			// NOTE: Not R3D_1B (this is a missing tex defining the
			// ceiling flat with an unwanted lower tex - so
			// must be ignored.
			if (sd1->texFlags == T_MID && missingtex ==  5)
			{
				if (abs(ceilh_b - floorh_b) < 4*FRACUNIT)
					Settype(ld, R3D_1D, 11175);
				else
					Settype(ld, R3D_1A, 11176);
				ld->mtflags &= ~(R3D_MISLTEX1|R3D_MISLTEX2);
			}
			else if (sd2->texFlags == T_MID && missingtex == 10)
			{
				if (abs(ceilh_f - floorh_f) < 4*FRACUNIT)
					Settype(ld, R3D_1D, 11177);
				else
					Settype(ld, R3D_1A, 11178);
				ld->mtflags &= ~(R3D_MISLTEX1|R3D_MISLTEX2);
			}
		}
	}
	}
}

//========================================================================

RDBOOL subfn_CheckLinesInFakedSecs(sector_t *sec)
{
	int i, count1 = 0, count2 = 0;
	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->backsector == NULL ||
			!(ld->spflags & SPF_NOTEX))
			return false;

		count1 += Vtxs2(ld) > 0;
		count2 += (ld->frontsector->heightsec >= 0 ||
				   ld->backsector->heightsec >= 0);
	}
	return (count2 && count1 != sec->linecount);
}


void LS_CheckLinesInFakedSecs()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (subfn_CheckLinesInFakedSecs(sec))
		{
			for (k = 0; k < sec->linecount; k++)
				Settype(sec->lines[k], R3D_3F, 11179);
		}
	}
}

static RDBOOL secHasLoTex(sector_t *s)
{
	line_t	 *ld;
	int		 i;

	for (i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (   !(ld->spflags & SPF_CONTROLLINE)
			&& !(ld->mtflags & R3D_MISLOTEX)
			&& origTEX[ld - lines] & 0x11)
		{
			return true;
		}
	}

	return false;
}

// v220-18
void LS_CheckInvalidLoTexFinal()
{
	sector_t	*fs, *bs;
	line_t		*ld = lines;
	int			i, mf, ff, bf, sd;

	for (i = 0; i < numlines; i++, ld++)
	{
		if (  ((mf = ld->mtflags) == (R3D_MISLTEX1|R3D_INVLTEX2) || mf == (R3D_MISLTEX2|R3D_INVLTEX1))
			&& !lineflags[i]
			&& ld->specType == R3D_00
			&& Vtfl3(ld)
			&& (fs = ld->frontsector)
			&& (bs = ld->backsector) )
		{
			if (mf == (R3D_MISLTEX2|R3D_INVLTEX1)) fs = bs, bs = ld->frontsector;
			
			ff = ld->frontsector->specFlags;
			bf = ld->backsector->specFlags;

			if ( (ff & 0x3041C) == 0x14 && ((bf & 0x30410) == 0x410 || (bf & 0x3001C) == 0x10))
			{
				// Check that the owning sector has lo tex on standard lines
				if (!secHasLoTex(bs)) continue;

				sd = (mf == (R3D_MISLTEX1|R3D_INVLTEX2));

				subfn_ReportInvalidTexError(ld - lines, sd, 0);

				ld->mtflags &= ~R3D_INVLOTEX;
				origTEX[i] &= 0x66;
				sides[ld->sidenum[sd]].bottomtexture = 0;
				sides[ld->sidenum[sd]].texFlags &= ~T_LOW;
			}
		}
	}
}

static RDBOOL checkTexValidity(sector_t *sc, line_t *li, RDBOOL hi)
{
	sector_t	*ch = li->frontsector == sc ? li->backsector : li->frontsector;
	line_t		*ld;
	int			a, b, i;
	int			ht;
	byte		flag;

	if (!ch) return false;

	ht = hi ? sc->ceilingheight - ch->ceilingheight : sc->floorheight - ch->floorheight;

	if (abs(ht) <= 8*FRACUNIT) return false;

	flag = hi ? R3D_MISHITEX : R3D_MISLOTEX;

	for (a = b = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (li != ld && ld->mtflags & flag && ld->lnflags2 & LNF2_OWNEQUALTOTWO)
		{
			if (   (ld->frontsector == sc && ld->backsector == ch)
				|| (ld->frontsector == ch && ld->backsector == sc))
			{
				if (ld->v1 == li->v1 && vtxflags[ld->v2 - vertexes] == 3) a |= 1, b++;
				else
				if (ld->v1 == li->v2 && vtxflags[ld->v2 - vertexes] == 3) a |= 1, b++;
				else
				if (ld->v2 == li->v1 && vtxflags[ld->v1 - vertexes] == 3) a |= 2, b++;
				else
				if (ld->v2 == li->v2 && vtxflags[ld->v1 - vertexes] == 3) a |= 2, b++;
			}
		}
	}

	return a == 3 && b == 2;
}

// b211-0 Modified
// v220-6 Simplified
// v220-9 Modified
// v220-12 Modified
// v220-20 Modified
void LS_SetHiLoTexFlags()
{
	sector_t *sec;
	line_t   *ld;
	int      i, k;
	
	for (sec = sectors, i = 0; i < numsectors; sec++, i++)
	{
		int movfloor = sec->specFlags & SF_MOVFLOOR;
		int movceil  = sec->specFlags & SF_MOVCEIL;

		sec->specFlags |= (SF_NOLOTEX|SF_NOHITEX);	// Preset

		// Check flag validity
		for (k = 0; k < sec->linecount; k++)
		{
			if ( (ld = sec->lines[k])->spflags & SPF_CONTROLLINE) continue;

			if ( !(ld->ml_flags & ML_TWOSIDED) )
			{
				sec->specFlags |= SF_SSLINETRUE;
				continue;
			}

			if (!(sec->specFlags & (SF_NOHITEX|SF_NOLOTEX)) ) continue;

			if (ld->frontsector != ld->backsector)
			{
				side_t *sd1 = &sides[ld->sidenum[0]];
				side_t *sd2 = &sides[ld->sidenum[1]];

				if (sec->specFlags & SF_NOHITEX)
				{
					if ((sd1->toptexture
						 && ld->frontsector->ceilingheight > ld->backsector->ceilingheight)
						 ||
						(sd2->toptexture
						 && ld->frontsector->ceilingheight < ld->backsector->ceilingheight))
					{
						if (  !(sd1->toptexture && sd2->toptexture)
							&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
							&& !(ld->mtflags & R3D_MISHITEX)
							&& Vtfl1(ld)
							&& checkTexValidity(sec, ld, true))
						{
							if (sd1->toptexture) { sd1->toptexture = 0; origTEX[ld - lines] &= ~0x04; }
							if (sd2->toptexture) { sd2->toptexture = 0; origTEX[ld - lines] &= ~0x40; }

							if (ld->frontsector->ceilingheight > ld->backsector->ceilingheight) ld->mtflags |= R3D_MISHTEX1;
							if (ld->frontsector->ceilingheight < ld->backsector->ceilingheight) ld->mtflags |= R3D_MISHTEX2;
						}
						else
						{
							sec->specFlags &= ~SF_NOHITEX;
						}
					}
					else if (movceil)
					{
						if (sd1->toptexture || sd2->toptexture)
						{
							sec->specFlags &= ~SF_NOHITEX;
						}
					}
				}

				if (!(sec->specFlags & SF_NOLOTEX)) continue;

				if ((sd1->bottomtexture
					&& ld->frontsector->floorheight < ld->backsector->floorheight)
					||
					(sd2->bottomtexture
					&& ld->frontsector->floorheight > ld->backsector->floorheight))
				{
					if (  !(sd1->bottomtexture && sd2->bottomtexture)
						&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
						&& !(ld->mtflags & R3D_MISLOTEX)
						&& Vtfl1(ld)
						&& checkTexValidity(sec, ld, false))
					{
						if (sd1->bottomtexture) { sd1->bottomtexture = 0; origTEX[ld - lines] &= ~0x01; }
						if (sd2->bottomtexture) { sd2->bottomtexture = 0; origTEX[ld - lines] &= ~0x10; }

						if (ld->frontsector->floorheight > ld->backsector->floorheight) ld->mtflags |= R3D_MISLTEX1;
						if (ld->frontsector->floorheight < ld->backsector->floorheight) ld->mtflags |= R3D_MISLTEX2;
					}
					else
					{
						sec->specFlags &= ~SF_NOLOTEX;
					}
				}
				else if (movfloor)
				{
					if (sd1->bottomtexture || sd2->bottomtexture)
						sec->specFlags &= ~SF_NOLOTEX;
				}
			}
		}
	}
}

// v220-4 Checks for any missing tex even if height difference is less than 4
//        These are not set with the mtflags
RDBOOL LS_CheckMissing(line_t *ld)
{
	sector_t *fs = ld->frontsector;
	sector_t *bs = ld->backsector;
	int txn;

	if (  !bs
		|| fs == bs
		|| ld->specType >= R3D_40
		|| fs->stptr >= 0
		|| bs->stptr >= 0
		|| fs->heightsec >= 0
		|| bs->heightsec >= 0)
	{
		return false;
	}

	txn = origTEX[ld - lines];

	if (!(txn & 0x01) && bs->floorheight > fs->floorheight) return true;
	if (!(txn & 0x10) && bs->floorheight < fs->floorheight) return true;
	if (!(txn & 0x04) && bs->ceilingheight < fs->ceilingheight) return true;
	if (!(txn & 0x40) && bs->ceilingheight > fs->ceilingheight) return true;

	return false;
}

// v220-6 test for line loops with a movfloor one side where
//        all contained lines have no upper or lower textures.
static RDBOOL checkForSSline(sector_t *sec)
{
	int i, k, m;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (hitline[ld - lines] == 4)
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];
				seg_t		*sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (    (ld = sg->linedef)
						&& !(ld->ml_flags & ML_TWOSIDED))
					{
						return true;
					}
				}
			}
		}
	}

	return false;
}

static void LS_CheckLineLoops1()
{
	sector_t	*sec, *osc;
	line_t		*ld;
	int			ring_count, box[4];
	int			c, i, k, m, n;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->heightsec >= 0
			|| sec->stptr >= 0
			|| (sec->specFlags & 0xC) == 0xC)
		{
			continue;
		}

		CLR_hitline();

		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];
			n = ld - lines;

			if ( !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
				|| ld->spflags & SPF_CONTROLLINE
				|| lineflags[n]
				|| origTEX[n] & 0x55
				|| !(osc = ld->frontsector == sec ? ld->backsector : ld->frontsector)
				|| osc == sec
				|| !((osc->specFlags & SF_MOVFLOOR)^(sec->specFlags & SF_MOVFLOOR))
				|| osc->heightsec >= 0
				|| osc->stptr >= 0
				|| lxarray[n]
				// v220-7 added the following tests
				|| (osc->floorpic != sec->floorpic && !(sec->floorpic == skyflatnum || osc->floorpic == skyflatnum))
				|| abs(osc->floorheight - sec->floorheight) < 4*FRACUNIT)
			{
				continue;
			}

			hitline[ld - lines] = 1;
			c++;
		}

		if (c < 3) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (hitline[ld - lines] == 1)
			{
				RDBOOL	valid = true;

				if (subfn_CheckForRing(hitline, sec, ld - lines, &ring_count, box, NULL, NULL))
				{
					if (sec->specFlags & SF_MOVSECTOR && !checkForSSline(sec))
					{
						valid = false;
					}

					if (valid && LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_REVERSED))
					{
						int		fh = DDMININT, ch = DDMAXINT;
						int		lnum = ld - lines;
				
						for (m = 0; m < hitssecptr; m++)
						{
							subsector_t *ss = &subsectors[hitssec[m]];
							seg_t		*sg = &segs[ss->firstline];

							if (ss->sector->specFlags & SF_MOVSECTOR)
							{
								if (ss->sector->floorheight > fh)
									fh = ss->sector->floorheight;
							}
							else
							{
								if (ss->sector->ceilingheight < ch)
									ch = ss->sector->ceilingheight;
							}

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if ( (ld = sg->linedef)
									&& origTEX[ld - lines] & 0x55)
								{
									valid = false; break;
								}
							}

							if (!valid) break;
						}

						if (valid && fh < ch)
						{
							validcount++;

							for (m = 0; m < hitssecptr; m++)
							{
								subsector_t *ss = &subsectors[hitssec[m]];
								seg_t		*sg = &segs[ss->firstline];
								RDBOOL		sameceil = ss->sector->ceilingpic == sec->ceilingpic
											|| (ss->sector->ceilingheight > sec->ceilingheight);

								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if ( (ld = sg->linedef) && ld->validcount != validcount)
									{
										ld->validcount = validcount;

										if (origTEX[ld - lines])
										{
											if (sameceil)
											{
												Settype(ld, R3D_2D, 11180);
												Clr_mthiflags(ld, 11180);
												Clr_mtloflags(ld, 11180);
											}
											else
											{
												Clr_mtloflags(ld, 11181);
												SetMissinglo(ld, 11181);
												Settype(ld, R3D_03, 11181);
											}
										}
										else
											Settype(ld, R3D_3F, 11182);

										SetLFG(ld,LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS, 11182);
									}
								}

								if (sameceil)
								{
									SetOverride(ss, i, 11183);
								}
								else
								{
									Setr3dssfs(ss, i, 11183);
								}
							}
						}
					}

					for (m = 0; m < sec->linecount; m++)
					{
						ld = sec->lines[m];

						if (hitline[ld - lines] == 4)
							hitline[ld - lines] = 8;
					}
				}
			}
		}
	}
}

// v220-6 check for ring of missing tex lines enclosing a ring of same sector sided lines.
static void LS_CheckLineLoops3()
{
	line_t	*ld;
	int		*marksec = (int*)W_Calloc(numsectors * sizeof(int), STATICID, 0);
	int		c, i, k, m, n;

	if (R3D_maptype) return;

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& !lineflags[i]
			&& !origTEX[i]
			&& (ld->mtflags == 0x30 || ld->mtflags == 0xC0))
		{
			if (ld->mtflags == 0x30)
			{
				if (!(ld->frontsector->specFlags & 0xC))
				{
					hitline[i] = 1; c++;
					marksec[ld->frontsector - sectors]++;
				}
			}
			else if (!(ld->backsector->specFlags & 0xC))
			{
				hitline[i] = 1; c++;
				marksec[ld->backsector  - sectors]++;
			}
		}
	}

	if (c > 2)
	{
		sector_t *s;
		int		 secn, lnum, ring_count, box[4];

		for (i = 0; i < numsectors; i++)
		{
			if (marksec[i] < 3) continue;

			s = &sectors[i];

			for (k = 0; k < s->linecount; k++)
			{
				ld = s->lines[k];
				lnum = ld - lines;

				if (   hitline[lnum] == 1
					&& subfn_CheckForRing(hitline, s, lnum, &ring_count, box, NULL, NULL)
					&& ring_count > 2)
				{
					int  *marklin = (int *)malloc(numlines * sizeof(int));
					byte *linloop = (byte*)calloc(numlines, 1);

					for (c = m = 0; m < s->linecount; m++)
					{
						lnum = s->lines[m] - lines;
						if (hitline[lnum] == 4)
						{
							marklin[c] = lnum;
							linearray[c++] = lnum;
						}
					}

					secn = (lines[linearray[0]].frontsector == s ?
						lines[linearray[0]].backsector : lines[linearray[0]].frontsector) - sectors;

					if (LS_MarkSubsecsInMultiSecLoop(0, NULL, c, box, NULL, true, MSIA_NONE))
					{
						sector_t	*fs = NULL;
						byte		*lineloop = (byte*)calloc(numlines, 1);
						int			loopcount = 0, lrc, snum = -1;
						RDBOOL		valid = true;

						for (m = 0; m < hitssecptr; m++)
						{
							subsector_t *ss = &subsectors[hitssec[m]];
							seg_t		*sg = &segs[ss->firstline];

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if (sg->linedef)
								{
									if (hitline[sg->linedef - lines] != 4)
									{
										fs = sg->linedef->frontsector;

										if (   fs != sg->linedef->backsector
											|| fs == ld->frontsector
											|| fs == ld->backsector)
										{
											valid = false; break;
										}

										snum = sg->linedef - lines;

										if (!lineloop[snum])
										{
											lineloop[snum] = 1;
											loopcount++;
										}
									}
								}
							}

							if (!valid) break;
						}

						if (valid) valid = loopcount > 2;

						// Check the lines contained in the ring form a loop
						if (   valid
							&& (!subfn_CheckForRing(lineloop, fs, snum, &lrc, NULL, NULL, NULL)
							|| lrc != loopcount))
						{
							valid = false;
						}

						if (valid)
						{
							validcount++;

							for (m = 0; m < hitssecptr; m++)
							{
								subsector_t *ss = &subsectors[hitssec[m]];
								seg_t		*sg = &segs[ss->firstline];
							
								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if ( (ld = sg->linedef) && ld->validcount != validcount)
									{
										ld->validcount = validcount;

										if (!origTEX[ld - lines])
										{
											Settype(ld, R3D_16, 11190);
											SetMissinglo(ld, 11190);
											ld->ml_flags |= ML_IGNOREMISSINGHI;
										}
										else
											Settype(ld, R3D_03, 11190);

										SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS, 11191);
									}
								}
								
								if (ss->sector == fs)
								{
									SetFloorlightsec(ss, secn, 11192);
									Setr3dsslcsec(ss, secn, 11192);
									Setr3dsslfsec(ss, secn, 11192);
								}
							}
						}

						free(lineloop);
					}

					for (m = 0; m < s->linecount; m++)
					{
						ld = s->lines[m];
						lnum = ld - lines;

						if (hitline[lnum] == 4) hitline[lnum] = 8;
					}
					
					free(marklin);
				}
			}
		}
	}

	W_Free(marksec, STATICID, 0);
}

// v220-6
// v220-9 LS_CheckLineLoops2() removed
void LS_CheckLineLoops()
{
	LS_CheckLineLoops1();
	LS_CheckLineLoops3();
}

// v220-10 get rid of unwanted hi textures in sky sector
void LS_DeleteUnwantedHiTextures()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			c, i;

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& origTEX[i] & 0x44
			&& (origTEX[i] & 0x44) != 0x44)
		{
			fs = ld->frontsector, bs = ld->backsector;

			if (   fs == bs
				|| fs->specFlags & SF_MOVCEIL
				|| bs->specFlags & SF_MOVCEIL)
			{
				continue;
			}

			if (   bs->ceilingpic == skyflatnum
				&& origTEX[i] & 0x40
				&& fs->ceilingheight > bs->ceilingheight)
			{
				sides[ld->sidenum[1]].toptexture = 0;
				origTEX[i] &= ~0x40; c++;
				if (texerror) pr_OFinsert("Unwanted toptexture side 1 deleted - L%i\n", i);
			}
			else
			if (   fs->ceilingpic == skyflatnum
				&& origTEX[i] & 0x04
				&& fs->ceilingheight < bs->ceilingheight)
			{
				sides[ld->sidenum[0]].toptexture = 0;
				origTEX[i] &= ~0x04; c++;
				if (texerror) pr_OFinsert("Unwanted toptexture side 0 deleted - L%i\n", i);
			}
		}
	}

	if (c && !texerror)
		pr_OFinsert("Deleted %i unwanted toptextures in sky sectors\n", c);
}
