//-----------------------------------------------------------------------------
//
// Copyright (C) 2013  G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceMissingNew.cpp
//	First release December 2013 (v220-22) gmj
//
// DESCRIPTION: Deals with missing texture loops that do not comply
//              with the standard analyse functions.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_MAINspec.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static vertex_t *getVtxOwner(line_t *ld, int *secidx)
{
	vertexowner_t	*vo = NULL;
	vertex_t		*vx = NULL;

	*secidx = -1;

	if (   vtxcheck[ld->v1 - vertexes] == 1
		&& vtxcheck[ld->v2 - vertexes] == 2
		&& vertexowners[ld->v1 - vertexes].num > 2
		&& vertexowners[ld->v2 - vertexes].num == 2)
	{
		vo = &vertexowners[ld->v1 - vertexes];
		vx = ld->v1;
	}
	else
	if (   vtxcheck[ld->v1 - vertexes] == 2
		&& vtxcheck[ld->v2 - vertexes] == 1
		&& vertexowners[ld->v2 - vertexes].num > 2
		&& vertexowners[ld->v1 - vertexes].num == 2)
	{
		vo = &vertexowners[ld->v2 - vertexes];
		vx = ld->v2;
	}

	if (vo)
	{
		int i;

		for (i = 0; i < vo->num; i++)
		{
			if ((sectors[vo->list[i]].specFlags & 0x3000C) == 0xC)
			{
				*secidx = vo->list[i];
				break;
			}
		}
	}

	return *secidx >= 0 ? vx : NULL;
}

void TM2_TraceHiLoops(void)
{
	sector_t	*sc;
	int			i;

	if (R3D_maptype) return;	// This trick type not supported

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0x3041C) == 0x14)
		{
			line_t			*ld;
			int				*linemem;
			int				box[4], rc;
			int				c, d, k, m;

			sc = &sectors[i];

			CLR_hitline();

			for (c = d = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;
				if (!ld->backsector) continue;	// insurance
				hitline[ld - lines] = 1;

				if (ld->frontsector->ceilingheight == ld->backsector->ceilingheight) continue;

				if (origTEX[ld - lines] & 0x44)
				{
					if (vtxflags[ld->v1 - vertexes] > 0 && vtxflags[ld->v2 - vertexes] > 0) continue;
					if (vtxflags[ld->v1 - vertexes] > 0 || vtxflags[ld->v2 - vertexes] > 0) d++;
				}
				else	// !(origTEX[ld - lines] & 0x44)
				{
					if (vtxflags[ld->v1 - vertexes] > 0 && vtxflags[ld->v2 - vertexes] > 0)
					{
						linearray[c++] = ld - lines;
					}
				}
			}

			if (c < 3 || !d) continue;

			linemem = (int*)malloc(c * sizeof(int));
			memcpy(linemem, linearray, c * sizeof(int));

			CLR_vtxcheck();

			for (k = 0; k < c; k++)
			{
				ld = &lines[linemem[k]];
				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
			}

			while (1)
			{
				vertex_t		*vx1, *vx2;
				sector_t		*ch, *os;
				int				secidx;
				int				lnum = -1;

				for (k = 0, vx1 = NULL; k < c; k++)
				{
					ld = &lines[linemem[k]];
					
					if (hitline[ld - lines] == 1 && (vx1 = getVtxOwner(ld, &secidx)) )
					{
						os = &sectors[secidx]; lnum = ld - lines; break;
					}
				}

				if (!vx1) break;

				if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
				{
					sector_t	*ns;
					vertex_t	*vc;
					line_t		*li;
					int			lac;
					int			chksec[2];

					for (k = 0, ns = NULL; k < c; k++)
					{
						if (hitline[linemem[k]] != 4) continue;

						ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

						if (!ns) ns = ch; else if (ns != ch) ns = NULL;
						if (!ns) break;
					}

					if (!ns) goto tidy_up;

					for (k = 0; k < c; k++)
					{
						li = &lines[linemem[k]];

						if (hitline[li - lines] != 4) continue;
						if (ld == li) continue;
						if ( (vx2 = getVtxOwner(li, &secidx)) ) break;
					}

					if (!vx2 || secidx != os - sectors) break;

					vc = NULL;
					lac = 0;

					for (k = 0; k < c; k++)
					{
						if (hitline[linemem[k]] == 4) linearray[lac++] = linemem[k];
					}

					for (k = 0; k < os->linecount; k++)
					{
						ld = os->lines[k];

						if (ld->v1 == vx1) { vc = ld->v2; linearray[lac++] = ld - lines; break; }
						if (ld->v2 == vx1) { vc = ld->v1; linearray[lac++] = ld - lines; break; }
					}

					if (!vc) goto tidy_up;

					validcount++;

					for (k = 0; k < os->linecount; k++)
					{
						if (vc == vx2) break;
						if ((ld = os->lines[k])->validcount == validcount) continue;
						if (ld->v1 == vc) { vc = ld->v2; M_AddLineToBox(box, ld); linearray[lac++] = ld - lines; ld->validcount = validcount; k = 0; continue; }
						if (ld->v2 == vc) { vc = ld->v1; M_AddLineToBox(box, ld); linearray[lac++] = ld - lines; ld->validcount = validcount; k = 0; continue; }
						break;
					}

					if (vc != vx2) goto tidy_up;

					chksec[0] = sc - sectors, chksec[1] = os - sectors;

					if (LS_MarkSubsecsInMultiSecLoop(2, chksec, lac, box, NULL, NULL, MSIA_NONE))
					{
						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t	*ss = &subsectors[hitssec[k]];
							seg_t		*sg = &segs[ss->firstline];

							Setr3dsscs(ss, ns - sectors, 53001);

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if (   (ld = sg->linedef)
									&& !lineflags[ld - lines])
								{
									byte txn = origTEX[ld - lines];

									if (!txn)
										Settype(ld, R3D_3F, 53001);
									else if (ld->frontsector == ld->backsector)
										Settype(ld, R3D_03, 53002);
									else if (ld->backsector
										&& txn & 0x22
										&& !(txn & 0x55)
										&& ld->frontsector->floorheight == ld->backsector->floorheight)
										Settype(ld, R3D_03, 53003);
									else
										Settype(ld, R3D_36, 53004);

									SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 53005);
								}
							}
						}

						Setsecdone(sc, 53006);
						Setsecdone(os, 53007);
					}
tidy_up:;
					for (k = 0; k < sc->linecount; k++)
					{
						ld = sc->lines[k];

						if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}

			free(linemem);
		}
	}
}

// ================================================================

void TM2_CheckLoLoopIn0xCsectors()
{
	sector_t	*ch, *sc, *os;
	line_t		*ld;
	int			c, i, k;

	if (R3D_maptype) return;	// This trick type not supported

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i] && (sectors[i].specFlags & 0x3041C) == 0x1C && !SecCtrl(i))
		{
			sc = &sectors[i];

			for (c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (!ld->backsector) continue;
				if (ld->frontsector == ld->backsector) continue;
				if (origTEX[ld - lines]) continue;
				if (!Vtxs2(ld)) continue;

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (!os)
				{
					if (secdone[ch - sectors]) { c = 0; break; }
					if (ch->floorheight >= sc->floorheight) { c = 0; break; }
					if ((ch->specFlags & 0x3040C) != 0xC) { c = 0; break; }
					os = ch;
				}
				else if (os != ch)
				{
					c = 0; break;
				}

				linearray[c++] = ld - lines;
			}

			if (c > 2 && os)
			{
				int			rc;

				CLR_hitline();

				for (k = 0; k < c; k++)
				{
					hitline[linearray[k]] = 1;
				}

				while (1)
				{
					int lnum = -1;

					for (k = 0; k < c; k++)
					{
						if (hitline[linearray[k]] == 1) { lnum = linearray[k]; break; }
					}

					if (lnum < 0) break;

					if (subfn_CheckForRing(hitline, sc, lnum, &rc, NULL, NULL, NULL) && rc > 2)
					{
						subsector_t	*ss;
						ssline_t	*sl;
						int			m, n;

						hitssecptr = 0;
						ss_validcount++;

						for (k = 0; k < c; k++)
						{
							sl = &sslines[linearray[k]];

							for (m = 0; m < sl->num; m++)
							{
								ss = &subsectors[sl->list[m]];

								if (ss->sector != sc) continue;

								if (ss->ssvalidcount != ss_validcount)
								{
									seg_t *sg = &segs[ss->firstline];

									for (n = 0; n < ss->linecount; n++, sg++)
									{
										if (  (ld = sg->linedef)
											&& ld->frontsector != ld->backsector
											&& hitline[ld - lines] != 4)
										{
											goto abort;
										}
									}

									hitssec[hitssecptr++] = sl->list[m];
									ss->ssvalidcount = ss_validcount;
								}
							}
						}

						// No lines are marked as this is preliminary.
						// It might have to be changed using a later routine.
						for (k = 0; k < hitssecptr; k++)
						{
							ss = &subsectors[hitssec[k]];
							Setr3dsslfsec(ss, os - sectors, 53008);
						}			
					}
					else
					{
						hitline[lnum] = 0;
					}
				}
			}
		}
abort:;
	}
}

// ================================================================

// If same sided lines cross over 0xC sectors with missing textures
// then chaos will reign.  Thankfully such horrors are rare as this
// kind of thing causes nodes builder havoc.
void TM2_CheckMissLoNestedInMissHi()
{
	sector_t	*ch, *sc, *os;
	line_t		*ld;
	int			*lnmem;
	int			*lnmem1;
	int			*lnmem2;
	int			mem1c, mem2c;
	int			c, d, i, k;

	if (R3D_maptype) return;	// This trick type not supported

	for (i = 0; i < numsectors; i++)
	{
		if (!secdone[i] && (sectors[i].specFlags & 0x3041C) == 0x1C && !SecCtrl(i))
		{
			sc = &sectors[i];

			for (c = d = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->frontsector == ld->backsector)
				{
					// Crossing same sided lines can cause havoc
					// with the nodes builder
					if (lxarray[ld - lines] & 0x4) d++;
					continue;
				}

				if (!ld->backsector
				|| origTEX[ld - lines] & 0x55
				|| !Vtxs2(ld)
				|| ld->frontsector->floorheight != ld->backsector->floorheight)
				{
					c = 0; break;
				}

				linearray[c++] = ld - lines;
			}

			if (c < 6 || !d) continue;

			lnmem = (int*)malloc(c * sizeof(int));
			memcpy(lnmem, linearray, c * sizeof(int));

			// Look for loop in loop where the outer has missing hi
			// and the inner does not

			lnmem1 = (int*)malloc(c * sizeof(int));
			lnmem2 = (int*)malloc(c * sizeof(int));

			mem1c = mem2c = 0;

			for (k = 0, os = NULL; k < c; k++)
			{
				ld = &lines[lnmem[k]];

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (ch->ceilingheight > sc->ceilingheight) goto tidy_up;

				if (ch->ceilingheight < sc->ceilingheight)
				{
					if (!os)
					{
						if ( (!(ch->specFlags & 0x400) && ch->specFlags & 0xC) ) goto tidy_up;
						os = ch;
					}
					else if (os != ch)
					{
						goto tidy_up;
					}

					lnmem1[mem1c++] = ld - lines;
				}
				else
				{
					lnmem2[mem2c++] = ld - lines;
				}
			}

			if (!os || mem1c < 3 || mem2c < 3) goto tidy_up;

			CLR_hitline();

			for (k = 0; k < c; k++) hitline[lnmem[k]] = 1;

			while (1)
			{
				sector_t	*is = NULL;	// inner sector
				int			box1[4], box2[4], rc;
				int			lnum;

				for (k = 0, lnum = -1; k < mem1c; k++)
				{
					if (hitline[lnmem1[k]] == 1) { lnum = lnmem1[k]; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, sc, lnum, &rc, box1, NULL, NULL) && rc > 2)
				{
					int valid = true;

					for (k = 0, lnum = -1; k < mem2c; k++)
					{
						if (hitline[lnmem2[k]] == 1) { lnum = lnmem2[k]; break; }
					}

					if (lnum >= 0)
					{
						if (subfn_CheckForRing(hitline, sc, lnum, &rc, box2, NULL, NULL) && rc > 2)
						{
							if (!M_CompareBox(box2, box1)) goto tidy_up;
						}
						else
						{
							hitline[lnum] = 0;
						}
					}
					else
					{
						goto tidy_up;
					}

					for (k = 0; k < mem2c; k++)
					{
						if (hitline[lnmem2[k]] == 4)
						{
							hitline[lnmem2[k]] = 8;

							if (valid)
							{
								ld = &lines[lnmem2[k]];
								ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
								if (!is) is = ch; else if (is != ch) valid = false;
							}
						}
					}

					if (!valid || !is) goto tidy_up;

					if (LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box1, -1, MSIA_NONE))
					{
						subsector_t	*ss;
						seg_t		*sg;
						int			cntsec = 0;
						int			m;

						validcount++;

						for (c = k = 0; k < hitssecptr; k++)
						{
							ss = &subsectors[hitssec[k]];
							sg = &segs[ss->firstline];

							for (m = 0; m < ss->linecount; m++, sg++)
							{
								if ( (ld = sg->linedef) && ld->validcount != validcount)
								{
									ld->validcount = validcount;
									linearray[c++] = ld - lines;

									if (hitline[ld - lines] == 4) continue;

									if (ld->frontsector == ld->backsector) continue;
									if (!ld->backsector) continue;
									if (origTEX[ld - lines] & 0x55) 	goto tidy_up;

									if (!Vtxs2(ld)) goto tidy_up;

									if (   ld->frontsector->ceilingheight != sc->ceilingheight
										|| ld->backsector->ceilingheight != sc->ceilingheight
										|| ld->frontsector->floorheight > sc->floorheight
										|| ld->backsector->floorheight > sc->floorheight)
									{
										goto tidy_up;
									}
								}
							}

							if (!secdone[ss->sector - sectors]) addSec(ss->sector - sectors, &cntsec);
						}

						for (k = 0; k < mem1c; k++)
						{
							if (hitline[lnmem1[k]] == 4) hitline[lnmem1[k]] = 0;
						}

						// correction for nodes build issues
						{
							subsector_t	*ss;
							int			*storela = (int*)malloc(c * sizeof(int));
							int			*storess = (int*)malloc(hitssecptr * sizeof(int));
							int			storeptr = hitssecptr;

							memcpy(storela, linearray, c * sizeof(int));
							memcpy(storess, hitssec, hitssecptr * sizeof(int));

							for (k = 0; k < mem2c; k++)
							{
								if (hitline[lnmem2[k]] == 8) hitline[lnmem2[k]] = 4;
							}

							valid = true;

							if (LS_MarkSubsecsInAreaGeneral(is, hitline, 0, NULL, box2, -1, MSIA_NONE))
							{
								// first set the whole area
								for (k = 0; k < storeptr; k++)
								{
									ss = &subsectors[storess[k]];

									Setr3dsscs(ss, os - sectors, 53009);
									Setr3dssfs(ss, sc - sectors, 53010);
								}

								for (k = 0; k < hitssecptr; k++)
								{
									ss = &subsectors[hitssec[k]];

									Setr3dssfs(ss, is - sectors, 53011);
								}
							}
							else
							{
								valid = false;
							}

							for (k = 0; k < mem2c; k++)
							{
								if (hitline[lnmem2[k]] == 4)
								{
									hitline[lnmem2[k]] = 0;
								}
							}

							memcpy(linearray, storela, c * sizeof(int));

							free(storela);
							free(storess);

							if (!valid) goto tidy_up;
						}

						for (k = 0; k < c; k++)
						{
							ld = &lines[linearray[k]];

							if (origTEX[ld - lines] & 0x22)
								Settype(ld, R3D_03, 53012);
							else
								Settype(ld, R3D_3F, 53013);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 53014);
						}

						for (k = 0; k < cntsec; k++)
						{
							SetSecidxDone(countsec[k], 53015);
						}
					}

					for (k = 0; k < mem1c; k++)
					{
						if (hitline[lnmem1[k]] == 8) hitline[lnmem1[k]] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}

tidy_up:;
			free(lnmem);
			free(lnmem1);
			free(lnmem2);
		}
	}
}

// ================================================================

void TM2_CheckMissHiInDoubleMovCeil()
{
	line_t	*ld;
	int		rc, box[4];
	int		c, d, i, k, n;

	if (R3D_maptype) return;	// This trick type not supported

	for (c = d = i = 0; i < numlines; i++)
	{
		if (  !lineflags[i]
			&& lines[i].backsector
			&& lines[i].frontsector != lines[i].backsector
			&& !(origTEX[i] & 0x44)
			&& lines[i].backsector->specFlags & 0x20000
			&& lines[i].frontsector->specFlags & 0x20000)
		{
			linearray[c++] = i;
			addSec(lines[i].frontsector - sectors, &d);
			addSec(lines[i].backsector - sectors, &d);
		}
	}

	if (c < 3) return;

	CLR_hitline();
	CLR_vtxcheck();

	for (i = 0; i < c; i++)
	{
		ld = &lines[linearray[i]];

		vtxcheck[ld->v1 - vertexes]++;
		vtxcheck[ld->v2 - vertexes]++;
	}

	for (i = 0; i < c; i++)
	{
		ld = &lines[linearray[i]];

		if (   vtxcheck[ld->v1 - vertexes] == 2
			&& vtxcheck[ld->v2 - vertexes] == 2)
		{
			hitline[linearray[i]] = 1;
		}
	}

	for (i = 0; i < d; i++)
	{
		while(1)
		{
			sector_t	*ch, *os, *sc = &sectors[countsec[i]];
			int			lnum = -1;

			for (k = 0; k < sc->linecount; k++)
			{
				n = sc->lines[k] - lines;

				if (hitline[n] == 1) { lnum = n; break; }
			}

			if (lnum < 0) break;

			if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL))
			{
				RDBOOL reversed = false;

				if (!LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE))
				{
					if (!LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_REVERSED))
						goto end_check;

					reversed = true;
				}

				for (k = 0; k < hitssecptr; k++)
				{
					if (   subsectors[hitssec[k]].ceilsec >= 0
						|| subsectors[hitssec[k]].lcsec >= 0)
					{
						goto end_check;
					}
				}

				if (!LS_OnlySSLinesInSSECarray(sc, hitline)) goto end_check;

				if (reversed)
				{
					os = sc;
				}
				else
				{
					int rep = 0;


					for (rep = 0; rep < 2; rep++)
					{
						for (k = n = 0, os = NULL; k < sc->linecount; k++)
						{
							ld = sc->lines[k];

							if (hitline[ld - lines] == 4)
							{
								ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
								if (!os) os = ch; else if (os == ch) { hitline[ld - lines] = 0; n++; }
							}
						}

						if (n > 2) break;
					}

					if (n < 2) goto end_check;
				}

				if (!os) goto end_check;

				for (k = 0; k < hitssecptr; k++)
				{
					Setr3dsslcsec(&subsectors[hitssec[k]], os - sectors, 53016);
				}

end_check:;
				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];

					if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
				}
			}
			else
			{
				hitline[lnum] = 0;
			}
		}
	}
}

// ===============================================================

typedef struct {
	int		 *llist;
	int		 lcount;
	int		 box[4];
	int		 done;
	__int64	 area;
} nestedbox_t;

static int __cdecl BoxSorter(const void *e1, const void *e2)
{
	__int64 lc1 = ((nestedbox_t*)e1)->area;
	__int64 lc2 = ((nestedbox_t*)e2)->area;
	if (lc1 > lc2) return -1;
	if (lc1 < lc2) return  1;
	return 0;
}

// Checks that all lines found in a hitssec list of ssecs are valid.
// If so returns a list of all lines found in linearray.
static int onlySSLinesInSSECarray(sector_t *sc)
{
	line_t	*ld;
	int		c, i, k;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (!(ld = sg->linedef) || ld->validcount == validcount)
				continue;

			ld->validcount = validcount;

			linearray[c++] = ld - lines;

			if (   hitline[ld - lines] != 4
				&& hitline[ld - lines] != 64)
			{
				if (!ld->backsector)
				{
					if (ld->frontsector != sc)
						return 0;
				}
				else if (ld->frontsector != ld->backsector)
					return 0;
			}
		}
	}

	return c;
}

// Look for textureless mov ceiling
void TM2_CheckMisHiAsymetricMovCeil()
{
	sector_t *ch, *sc, *os, *fs, *bs;
	line_t	 *ld;
	int		 boxcount;
	int		 c, i, k, m;

	if (R3D_maptype) return;	// This trick type not supported

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || secspflags[i] & SECSPF_CTRLLINES) continue;

		if ( (sectors[i].specFlags & 0x3041C) == 0x2001C && sectors[i].linecount > 5)
		{
			nestedbox_t *nbt;
			nestedbox_t	*nbptr;
			int			numbt;
			int			linesset = 0;

			sc = &sectors[i];

			for (c = k = 0, os = NULL; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) { linesset++; continue; }
				if (lineflags[ld - lines]) { c = 0; break; }
				if (origTEX[ld - lines] & 0x44) { c = 0; break; }

				// find lines with mislotex flags where the floor diff is <= 8
				if (   (bs = ld->backsector)
					&& (fs = ld->frontsector) != bs
					&& abs(fs->floorheight - bs->floorheight) > 8*FRACUNIT)
				{
					linesset++; continue;
				}

				if (fs == bs) { linesset++;  continue; }
				if (fs->specFlags & SF_MOVFLOOR || bs->specFlags & SF_MOVFLOOR) { linesset++;  continue; }

				ch = fs == sc ? bs : fs;

				if (ch->specFlags & SF_MOVCEIL || (ch->specFlags & SF_NOHITEX && !(ch->specFlags & 0x400)) )
				{ c = 0; break; }

				if (!os) { os = ch; }
				else if (ch != os && ch->ceilingheight != os->ceilingheight)
				{ c = 0; break; }

				// Exceptional case where contained ceiling is lower than surrounding
				if (ch->ceilingheight > sc->ceilingheight)
				{
					linearray[c++] = ld - lines;
				}
			}

			if (!os || c < 6) continue;	// Looking for nested ring type

			numbt = (c/3) + 1;	// max needed, will normally be less
			nbt = (nestedbox_t*)calloc(numbt, sizeof(nestedbox_t));

			CLR_hitline();
			boxcount = 0;

			for (k = 0; k < c; k++) hitline[linearray[k]] = 1;

			// get line boxes
			while(1)
			{
				int			n, rc;
				int			lnum = -1;

				for (k = 0; k < c; k++) { if (hitline[linearray[k]] == 1) { lnum = linearray[k]; break; } }

				if (lnum < 0) break;

				nbptr = &nbt[boxcount++];

				if (subfn_CheckForRing(hitline, sc, lnum, &rc, nbptr->box, NULL, NULL) && rc > 2)
				{
					nbptr->llist = (int*)malloc(rc * sizeof(int));
					nbptr->lcount = rc;
					nbptr->done = false;
					nbptr->area = (nbptr->box[BOXRIGHT] - nbptr->box[BOXLEFT])
								* (nbptr->box[BOXTOP]   - nbptr->box[BOXBASE]);

					for (k = n = 0; k < c; k++)
					{
						if (hitline[linearray[k]] == 4)
						{
							nbptr->llist[n++] = linearray[k];
							hitline[linearray[k]] = 0;
						}
					}

					if (n != rc) Con_Error("Program error: rc %i n %i\n", rc, n);
				}
				else
				{
					hitline[lnum] = 0;
				}
			}

			if (boxcount < 2) goto abort;
			
			qsort(nbt, boxcount, sizeof(nestedbox_t), BoxSorter);

			while (1)
			{
				int outer = -1, inner = -1;
				int	*ssecmem, *tmpssecmem;
				int	ssec_cnt, tmpssec_cnt;

				for (k = 0; k < boxcount; k++)
				{
					if (!nbt[k].done) { nbt[k].done = true; outer = k; break; }
				}

				if (outer < 0) break;

				nbptr = &nbt[outer];

				for (k = 0; k < boxcount; k++)
				{
					if (!nbt[k].done && M_CompareBox(nbt[k].box, nbptr->box))
					{ nbt[k].done = true; inner = k; break; }
				}

				if (inner < 0) continue;

				// check there are no further loops inside the current outer loop
				// these could be handled but it is currently not necessary
				for (k = 0; k < boxcount; k++)
				{
					if (!nbt[k].done && M_CompareBox(nbt[k].box, nbptr->box)) goto abort;
				}

				// inner loop
				CLR_hitline();
				nbptr = &nbt[inner];

				for (k = 0; k < nbptr->lcount; k++)
				{
					hitline[nbptr->llist[k]] = 4;
				}

				if (!LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, nbptr->box, -1, MSIA_REVERSED))
					continue;

				for (k = 0; k < nbptr->lcount; k++)
				{
					hitline[nbptr->llist[k]] = 64;
				}

				// remember ssecs inside
				ssec_cnt = hitssecptr;
				ssecmem = (int*)malloc(ssec_cnt * sizeof(int));
				memcpy(ssecmem, hitssec, ssec_cnt * sizeof(int));

				// outer loop
				nbptr = &nbt[outer];

				for (k = 0; k < nbptr->lcount; k++)
				{
					hitline[nbptr->llist[k]] = 4;
				}

				if (!LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, nbptr->box, -1, MSIA_NONE))
				{
					free(ssecmem);
					continue;
				}
				
				tmpssec_cnt = hitssecptr;
				tmpssecmem = (int*)malloc(tmpssec_cnt * sizeof(int));
				memcpy(tmpssecmem, hitssec, tmpssec_cnt * sizeof(int));

				hitssecptr = 0;

				for (k = 0; k < tmpssec_cnt; k++)
				{
					for (m = 0; m < ssec_cnt; m++)
					{
						if (tmpssecmem[k] == ssecmem[m]) goto next_chk;
					}

					hitssec[hitssecptr++] = tmpssecmem[k];
next_chk:;
				}

				// check our current list of ssecs is valid and if so set the lines
				if ( (c = onlySSLinesInSSECarray(sc)) )
				{
					byte txn;

					for (k = 0; k < c; k++)
					{
						txn = origTEX[linearray[k]];
						ld = &lines[linearray[k]];

						if (!txn)
						{
							if (ld->backsector && ld->frontsector->floorheight == ld->backsector->floorheight)
								Settype(ld, R3D_3F, 53018);
							else // draw missing lower tex
							{
								Settype(ld, R3D_00, 53019);
								Clr_mtloflags(ld, 53019);
							}
						}
						else if (ld->backsector
							&& txn & 0x22
							&& !(txn & 0x55)
							&& ld->frontsector->floorheight == ld->backsector->floorheight)
							Settype(ld, R3D_03, 53020);
						else
							Settype(ld, R3D_36, 53021);

						SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 53022);
					}
					
					linesset += c;

					// Since no upper textures are to be drawn the rendered ceiling height can
					// be fixed. We do not want to see a disconnected rising ceiling flat.
					for (k = 0; k < hitssecptr; k++)
					{
						Setr3dsscs(&subsectors[hitssec[k]], os - sectors, 53023);
					}
				}

				free(tmpssecmem);
				free(ssecmem);
			}

			if (sc->linecount <= linesset) SetSecidxDone(i, 53034);

abort:;
			for (k = 0; k < boxcount; k++) free(nbt[k].llist);
			free(nbt);
		}
	}
}

// ===============================================================

static int pmld_linecount;
static sector_t *pmld_sector;

// Note: validcount must be set by calling code as well as
//       ss_validcount and both pmld_linecount and pmld_sector.
static void findLoopFromSsecs(int lineidx)
{
	ssline_t	*sl = &sslines[lineidx];
	line_t		*ld;
	int			i, k;

	lines[lineidx].validcount = validcount;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			seg_t *sg = &segs[ss->firstline];

			ss->ssvalidcount = ss_validcount;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (   sg->frontsector == pmld_sector
					&& (ld = sg->linedef)
					&& ld->validcount != validcount)
				{
					linearray[pmld_linecount++] = ld - lines;
					findLoopFromSsecs(ld - lines);
				}
			}
		}
	}
}

void TM2_PartialMissingLoMov()
{
	int		c, i, k, m;

	if (R3D_maptype) return;	// This trick type not supported

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].heightsec >= 0) continue;

		if (!(sectors[i].specFlags & SF_MOVFLOOR) && !SecCtrl(i))
		{
			sector_t	*ch, *sc = &sectors[i];
			line_t		*ld;
			fixed_t		fht, chht, scfht = sc->floorheight;
			int			rc, box[4];
			int			*lnmem = (int*)malloc(sc->linecount * sizeof(int));
			int			lnmemcnt = 0;

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				if (lineflags[ld - lines]) continue;
				if (!ld->backsector) continue;
				if (ld->frontsector == ld->backsector) continue;

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
				if (!ch) continue;
				if (  !(ch->specFlags & SF_MOVFLOOR)
					|| (ch->specFlags & 0x4 && !(ch->specFlags & 0x400)) )
				{
					continue;
				}
				if (ch == ld->frontsector && origTEX[ld - lines] & 0x10) continue;
				if (ch == ld->backsector  && origTEX[ld - lines] & 0x01) continue;

				chht = ch->floorheight;

				if (   ch->siflags & SIF_MOVHIGH1
					&& (fht = P_FindHighestFloorSurrounding(ch)) > scfht
					&& fht > chht)
				{
					goto isvalid;
				}

				if (   ch->siflags & SIF_MOVHIGH2
					&& (fht = P_FindNextHighestFloor(ch, chht)) > scfht
					&& fht > chht)
				{
					goto isvalid;
				}

				continue;
isvalid:;
				lnmem[lnmemcnt++] = ld - lines;

				ss_validcount++;
				validcount++;
				// globals used with re-entrant function findLoopFromSsecs()
				pmld_linecount = 0;
				pmld_sector = sc;

				linearray[pmld_linecount++] = ld - lines;
				findLoopFromSsecs(ld - lines);

				if (pmld_linecount > 2)
				{
					CLR_hitline();

					for (m = 0; m < pmld_linecount; m++) hitline[linearray[m]] = 1;

					if (   subfn_CheckForRing(hitline, sc, linearray[0], &rc, box, NULL, NULL)
						&& rc == pmld_linecount
						&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& (c = onlySSLinesInSSECarray(sc)) )
					{
						byte txn;

						for (m = 0; m < c; m++)
						{
							txn = origTEX[linearray[m]];
							ld = &lines[linearray[m]];

							if (!txn)
								Settype(ld, R3D_3F, 53025);
							else if (ld->backsector
								&& txn & 0x22
								&& !(txn & 0x55)
								&& ld->frontsector->floorheight == ld->backsector->floorheight)
								Settype(ld, R3D_03, 53026);
							else
								Settype(ld, R3D_36, 53027);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 53028);
						}

						for (m = 0; m < hitssecptr; m++)
						{
							Setr3dsslfsec(&subsectors[hitssec[m]], ch - sectors, 53029);
						}
					}
				}
			}

			// second check for complete sector loops
			// This needs to be done because there is no guarantee that getting
			// a loop, other than in simple cases, will work by iterating ssecs.
			for (c = k = 0; k < lnmemcnt; k++)
			{
				ld = &lines[lnmem[k]];
				if (!lineflags[ld - lines]) c++;
			}

			if (c < 3) { free(lnmem); continue; }

			CLR_vtxcheck();

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				if (lineflags[ld - lines]) continue;
				if (ld->frontsector == ld->backsector) continue;

				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
			}

			CLR_hitline();

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				if (lineflags[ld - lines]) continue;
				if (ld->frontsector == ld->backsector) continue;

				if (vtxcheck[ld->v1 - vertexes] == 2 && vtxcheck[ld->v2 - vertexes] == 2)
				{
					hitline[ld - lines] = 1;
				}
			}

			while (1)
			{
				int lnum = -1;

				for (k = 0; k < lnmemcnt; k++)
				{
					ld = &lines[lnmem[k]];

					if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
				{
					if (   LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE)
						&& (c = onlySSLinesInSSECarray(sc)) )
					{
						byte txn;

						ch = lines[lnum].frontsector == sc ? lines[lnum].backsector : lines[lnum].frontsector;

						for (m = 0; m < c; m++)
						{
							txn = origTEX[linearray[m]];
							ld = &lines[linearray[m]];

							if (!txn)
								Settype(ld, R3D_3F, 53030);
							else if (ld->backsector
								&& txn & 0x22
								&& !(txn & 0x55)
								&& ld->frontsector->floorheight == ld->backsector->floorheight)
								Settype(ld, R3D_03, 53031);
							else
								Settype(ld, R3D_36, 53032);

							SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 53033);
						}

						for (m = 0; m < hitssecptr; m++)
						{
							Setr3dsslfsec(&subsectors[hitssec[m]], ch - sectors, 53034);
						}
					}

					for (k = 0; k < sc->linecount; k++)
					{
						ld = sc->lines[k];

						if (hitline[ld - lines] == 4) hitline[ld - lines] = 0;
					}
				}
				else
				{
					hitline[lnum] = 0;
				}
			}

			free(lnmem);
		}
	}
}

// ================================================================

static void Subfn_MissingLoLoopsMovInside(int secn)
{
	sector_t	*sc = &sectors[secn];
	int			c, i, k, lnum = -1;

	CLR_hitline();

	// check no subsec settings have been made
	for (c = i = 0; i < sc->linecount; i++)
	{
		line_t		*ld = sc->lines[i];
		ssline_t	*sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->floorsec >= 0 || ss->ceilsec >= 0 || (ss->lfsec >= 0 && ss->lfsec != secn))
				return;
		}

		if (   (ld->mtflags == 0x41 && ld->frontsector == sc)
			|| (ld->mtflags == 0x14 && ld->backsector == sc))
		{
			if (lnum < 0) lnum = ld - lines;
			hitline[ld - lines] = 1; c++;
		}
	}

	if (c < 3) return;

	fixed_t		box[4];
	int			rc = 0;

	// Only interested in a single loop
	if (!subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) || rc != c) return;
	// Only interested in secn being on the outside of a loop
	if (!LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_REVERSED)) return;

	subsector_t	*ss;
	seg_t		*sg;

	// Check no lines or subsectors in the loop have been previously processed
	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		if (   ss->floorsec >= 0
			|| ss->ceilsec >= 0
			|| (ss->lfsec >= 0 && ss->lfsec != secn)
			|| ss->sector == sc)	// not a reversed loop
		{
			return;
		}

		for (k = 0, sg = &segs[ss->firstline]; k <ss->linecount; k++, sg++)
		{
			if (sg->linedef && lineflags[sg->linedef - lines]) return;
		}
	}

	// Should all be OK if we get here. Set the contained ssecs lfsec
	for (i = 0; i < hitssecptr; i++) Setr3dsslfsec(&subsectors[hitssec[i]], secn, 53035);

	Setsecdone(sc, 53036);

	// Mark everything as processed
	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		sc = ss->sector;
		if (!secdone[sc - sectors])
		{
			Setsecdone(sc, 53036);
		}

		for (k = 0, sg = &segs[ss->firstline]; k <ss->linecount; k++, sg++)
		{
			if (sg->linedef && !lineflags[sg->linedef - lines])
				SetLFG(sg->linedef, LFG_NEWSET|LFG_FIXED, 53037);
		}
	}
}

// v220-33
void TM2_SingleMissingLowLoop()
{
	int		c, i, k, m;

	if (R3D_maptype) return;	// This trick type not supported

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].heightsec >= 0) continue;

		if (sectors[i].specFlags != 0x00008198 || SecCtrl(i)) continue;

		// See if sector has a loop of missing lo linwes
		for (c = k = 0; k < sectors[i].linecount; k++)
		{
			if ( (m = sectors[i].lines[k]->mtflags) == 0x14 || m == 0x41)
			{
				Subfn_MissingLoLoopsMovInside(i);
				break;
			}
		}
	}
}