//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_tricks.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
//
// DESCRIPTION: some obscure tricks support.  Some barely work with
//              software renderers so don't expect too much.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// Check for chained missing tex lines conferring floor sec.
// b211-0 modified
void TR_TrickChainedFloorMissing()
{
	subsector_t *ss;
	sector_t	*sec, *s;
	line_t		*ld;
	int			i, k, m;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->ceilingheight > sec->floorheight
			&& !(sec->specFlags & (SF_MOVSECTOR|SF_HASTEXTURE|SF_SECCLOSED))
			&& !IsLiquidFloorFlat(sec))
		{
			int secnum = -1, sn, mt = 0, count = 0;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (lineflags[ld - lines]) { secnum = -1; break; }	// v220-9 added
				if (ChkLptr(ld)) { secnum = -1; break; }

				if (!origTEX[ld - lines] && ld->mtflags)
				{
					if (!mt) mt = ld->mtflags;

					if (   mt != ld->mtflags
						|| !(mt & R3D_MISLOTEX)
						|| mt & R3D_MISHITEX)
					{
						secnum = -1; break;
					}

					sn = ld->frontsector == sec ?
						ld->backsector - sectors : ld->frontsector - sectors;

					if (secnum == -1) secnum = sn;

					if (secnum != sn) { secnum = -1; break; }

					count++;
				}
			}

			if (secnum < 0 || count < 5) continue;

			s = &sectors[secnum];

			if (sec->ceilingpic == skyflatnum)
			{
				for (k = 0, ss = subsectors; k < numsubsectors; k++, ss++)
				{
					seg_t	*sg = &segs[ss->firstline];
					int		ht;

					ld = NULL;

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if (   sg->backsector
							&& sg->linedef
							&& (sg->frontsector == s || sg->backsector == s)
							&& sg->frontsector->ceilingpic == sg->backsector->ceilingpic)
						{
							ht = MIN_OF(sg->frontsector->floorheight, sg->backsector->floorheight);
							ld = sg->linedef; break;
						}
					}

					if (ld && ld->specType != R3D_16)
					{
						sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (sg->linedef && sg->backsector)
							{
								if (   sg->backsector->ceilingpic != skyflatnum
									&& sg->frontsector->floorheight == ht)
								{
									Settype(ld, R3D_16, 26001);
									break;
								}
							}
						}
					}
				}
			}
			else
			{
				for (k = 0; k < s->linecount; k++)
				{
					ld = s->lines[k];
					if (ld->mtflags == mt)
					{
						Setspec(ld, i, 26002);
						Settype(ld, R3D_3F, 26002);
					}
				}

				for (k = 0, ss = subsectors; k < numsubsectors; k++, ss++)
				{
					if (ss->sector == s)
					{
						seg_t *sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if (  (ld = sg->linedef)
								&& LSS(ld)
								&& ld->frontsector == s)
							{
								float v1[2];
								float v2[2];

								v1[VX] = FIX2FLT(ld->v1->x);
								v1[VY] = FIX2FLT(ld->v1->y);
								v2[VX] = FIX2FLT(ld->v2->x);
								v2[VY] = FIX2FLT(ld->v2->y);

								// Check facing ld->sidenum[0]
								if (subfn_SegFacingMidpoint(v2, v1, ss->midpoint.x, ss->midpoint.y))
								{
									Setr3dsslfsec(ss, secnum, 26003);
									break;
								}
							}
						}
					}
				}
			}
		}
	}
}

// =================================================

void subfn_MissingBoundTo1sLine(sector_t *s)
{
	subsector_t *ss;
	sector_t	*os;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			chksec[2], box[4], rc;
	int			c, i, k;

	for (c = i = 0; i < s->linecount; i++)
	{
		if ( !(s->lines[i]->ml_flags & ML_TWOSIDED) )
		{
			ld = s->lines[i]; c++;
		}

		if (c > 1) return;
	}

	if (c != 1) return;

	// Next stage is to look for a specific sector type
	sl = &sslines[ld - lines];

	for (i = 0, os = NULL; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->frontsector
				&& sg->frontsector != s
				&& (sg->frontsector->specFlags & 0x3041C) == 0xC
				&& sg->frontsector->floorheight > s->floorheight)
			{
				os = sg->frontsector; break;
			}
		}

		if (os) break;
	}

	if (!os || os->ceilingheight < s->ceilingheight) return;

	// check validity of os.
	// All of its lines plus the seed line should form a ring

	CLR_hitline();

	linearray[0] = ld - lines;

	for (i = 0, c = 1; i < os->linecount; i++)
	{
		linearray[c] = os->lines[i] - lines;
		if ( (origTEX[linearray[c]] & 0x22) == 0x22) return;	// Only works in Doom if one side set
		c++;
	}

	chksec[0] =  s - sectors; chksec[1] = os - sectors;

	if (  (rc = nf_chkLineListRing(c, box, 0, hitline))
		&& LS_MarkSubsecsInMultiSecLoop(2, chksec, rc, box, NULL, false, MSIA_NONE))
	{
		validcount++;

		// Check the subsector lines
		for (c = i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( !(ld = sg->linedef) ) continue;
				if (ld->validcount == validcount) continue;

				ld->validcount = validcount;

				linearray[c++] = ld - lines;

				if (sg->linedef && hitline[sg->linedef - lines] != 4)
				{
					if (!sg->linedef->backsector)
					{
						if (sg->linedef->frontsector != s) return;
					}
					else if (sg->frontsector != sg->backsector
						&& sg->frontsector != s)
					{
						return;	// if not in loop front must belong to s
					}
				}
			}
		}

		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];

			if (ss->sector != s) SetOverride(ss, s - sectors, 26004);
		}

		for (i = 0; i < c; i++)
		{
			ld = &lines[linearray[i]];

			if (ld->ml_flags & ML_TWOSIDED)
			{
				if (origTEX[linearray[i]] == 0x22)
					Settype(ld, R3D_03, 26005);
				else if (origTEX[linearray[i]])
					Settype(ld, R3D_02, 26005);	// special line handling type
				else if (ld->frontsector == ld->backsector)
					Settype(ld, R3D_3F, 26005);
			}
			else if (ld->specType != R3D_00)
			{
				Settype(ld, R3D_00, 26005);
			}

			SetLFG(ld, LFG_FIXED|LFG_NEWSET|LFG_FIXABS, 26005);
		}

		Setsecdone(s, 26006);
	}
}

// v220-10 Search sectors with a single 1's line used for trick rendering.
void TR_MissingBoundTo1sLine()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		// Note: only movsec, closed and tex hi/lo flags set at this point
		if ((sectors[i].specFlags & 0x3041C) == 0x40C && !SecCtrl(i))
			subfn_MissingBoundTo1sLine(&sectors[i]);
	}
}

// =========================================================

static void subfn_floorsec(sector_t *s)
{
	ssline_t	*sl;
	sector_t	*os;
	line_t		*ld;
	seg_t		*sg;
	int			fs, lidx;
	int			c, i, k, m;

	for (c = i = 0, fs = -1; i < s->linecount; i++)
	{
		if ((ld = s->lines[i])->spflags & SPF_CONTROLLINE) continue;
		lidx = ld - lines;
		if (lineflags[lidx]) return;
		if (origTEX[lidx]) return;
		if (lxarray[lidx]) return;
		if (s->lines[i]->frontsector != s->lines[i]->backsector) return;
		linearray[c++] = lidx;
	}

	if (c < 2) return;

	hitssecptr = 0;
	validcount++;
	ss_validcount++;

	// Get a list of ssecs referencing the sector's lines
	for (i = 0, os = NULL; i < c; i++)
	{
		sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
			{
				if (!sg->frontsector || sg->frontsector == s) continue;
				if (sg->frontsector->specFlags & SF_MOVFLOOR) continue;
				if (!os)
				{
					os = sg->frontsector;
					if (os->specFlags & SF_NOLOTEX && !(os->specFlags & SF_SSLINETRUE))
						return;
				}
				else
				if (os != sg->frontsector) return;
			}

			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	if (!os) return;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		if (ss->sector != os) SetOverride(ss, os - sectors, 26007);
	}

	for (i = 0; i < c; i++)
	{
		Settype(&lines[linearray[i]], R3D_3F, 26008);
		SetLFG(&lines[linearray[i]], LFG_NEWSET|LFG_FIXED, 26008);
	}

	Setsecdone(s, 26009);
}


// v220-20	This purely presets lfsec with a movefloor contained in a missing lowtex ring
void TR_MovFloorInside()
{
	sector_t	*ch, *sc, *ms, *os;
	line_t		*ld;
	int			lnum, rc, box1[4], box2[4];
	int			a, b, i, k, m;

	// -----------------

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].stptr >= 0 || (sectors[i].specFlags & 0x30014) != 0x10) continue;

		sc = &sectors[i];
		M_ClearBox(box1);
		M_ClearBox(box2);

		for (a = b = k = 0, ms = os = NULL; k < sc->linecount; k++)
		{
			ld = sc->lines[k];

			if (lineflags[ld - lines]) { a = 0; break; }
			if (ld->spflags & SPF_CONTROLLINE) { a = 0; break; }

			ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
			if (!ch) { a = 0; break; }
			if (ch == sc) continue;
			if ( (ch->specFlags & 0x830418) == 0x810018)
			{
				if (!ms) ms = ch; else if (ch != ms) { a = 0; break; }
				M_AddLineToBox(box2, ld);
				a++;
			}
			else if (ld->mtflags & R3D_MISLOTEX)
			{
				if (!os)
				{
					if (ch->specFlags & SF_MOVFLOOR) { a = 0; break; }
					if (ch->floorheight >= sc->floorheight + 8*FRACUNIT) { a = 0; break; }
					os = ch;
				}
				else if (os != ch)
				{
					if (   os->floorheight != ch->floorheight
						|| os->floorpic != ch->floorpic)
					{ a = 0; break;
					}
				}

				M_AddLineToBox(box1, ld);
				linearray[b++] = ld - lines;
			}
			else { a = 0; break; }
		}

		if (a < 3 || b < 3 || !M_FitBox(box2, box1) || os->floorheight >= ms->floorheight) continue;

		CLR_hitline();

		for (k = 0; k < b; k++) hitline[linearray[k]] = 1;

		// Since NOLOTEX is not set then low textures must be present on the mov sec being referenced

		lnum = linearray[0];

		if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box1, NULL, NULL)
			&& rc > 2
			&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box1, -1, MSIA_NONE))
		{
			// Check all lines contained
			validcount++;

			for (a = k = 0; k < hitssecptr; k++)
			{
				subsector_t *ss = &subsectors[hitssec[k]];
				seg_t		*sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if ( !(ld = sg->linedef) || ld->validcount == validcount) continue;

					ld->validcount = validcount;
					if (ld->frontsector != sc && ld->backsector != sc) { a = 0; break; }
					a++;
				}
			}

			if (a)
			{
				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];

					if (ss->sector == ms)
						Setr3dsslfsec(ss, os - sectors, 26010);
				}
			}
		}
	}
}

// v220-18 changed
void TR_MovFloorContained_X()
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (secdone[i] || sectors[i].stptr >= 0) continue;

		if ( (sectors[i].specFlags &0x23FFEC) == 0x21000C)
		{
			subfn_floorsec(&sectors[i]);
		}
	}
}