//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_utils.c
//
// PointIsOnLine() (C) Lee Killough 1998
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Removed dualRingSector(int, int*);
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// DESCRIPTION: base utilities used by the analyser
//
//-----------------------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "r_things.h"
#include "r3d_edit.h"

extern int	blacktexnum, notexTEXNUM;
char		*addflags = NULL;	//b211-0
ssecinfo_t	ssinf;

//============================================================================

RDBOOL addSec(int secidx, int *count)
{
	int i;

	if (secidx < 0 || secidx >= numsectors)
	{
#ifdef _DEBUG
		Con_Error("program error addSec");
#endif
		return false;
	}

	for (i = 0; i < *count; i++)
	{
		if (countsec[i] == secidx) return false;
	}

	countsec[ (*count)++ ] = secidx;
	return true;
}

RDBOOL findSec(int secidx, int count)
{
	int i;

	for (i = 0; i < count; i++)
	{
		if (countsec[i] == secidx) return true;
	}

	return false;
}

void SET_countsec(int num)
{
	memset(countsec, num, numsectors*sizeof(int));
}

void SET_linearray(int num)
{
	memset(linearray, num, (numlines + 1)*sizeof(int));
}

void CLR_hitline()
{
	memset(hitline, 0, numlines);
}

void CLR_excludeline()
{
	memset(excludeline, 0, numlines);
}

void CLR_vtxcheck(void)
{
	memset(vtxcheck, 0, numvertexes);
}

void CLR_conarray(void)
{
	memset(conarray, 0, numlines);
}

void CLR_ssecarray(void)
{
	memset(ssecarray, 0, numsubsectors);
}

RDBOOL LSS(line_t *ld)
{
	return ld->ml_flags & ML_TWOSIDED	// test to screen out special fx lines
		&& ld->frontsector == ld->backsector;
}

fixed_t meanV(__int64 a, __int64 b)
{
	return (int)((a + b)/2);
}

// Masked Sector specFlags compare
int MSFC(int secnum, int flags)
{
	int mask = flags & SF_LINESCLOSED ? SF_TESTMASK : SF_TESTMASK2;
	return (sectors[secnum].specFlags & mask) == flags;
}

// Test for same sided line only
byte Chk0304Type(line_t *ld)
{
	if (origTEX[ld - lines] == 0x22)
	{
		side_t *sd1 = &sides[ld->sidenum[0]];
		side_t *sd2 = &sides[ld->sidenum[1]];
		int txn1 = sd1->midtexture;
		int txn2 = sd2->midtexture;

		if (!(ld->ml_flags & (ML_BLOCKING|ML_DONTPEGTOP)) &&
			txn1 == txn2 &&
			texArray[txn1]->height <= 16 &&
			sd2->rowoffset ==
			sd1->rowoffset)
		{
			return R3D_03;
		}
	}

	return R3D_04;
}

RDBOOL IsLiquidFloorFlat(sector_t *sec)
{
	if (!sec->floorpic) return false;
	return R_CheckFlatFlag(sec->floorpic, FLT_SPLASH);
}

RDBOOL IsLiquidCeilFlat(sector_t *sec)
{
	return R_CheckFlatFlag(sec->ceilingpic, FLT_SPLASH);
}

RDBOOL IsAnimFlat(int picnum)
{
	return R_CheckFlatFlag(picnum, TXF_ANIMPIC);
}

RDBOOL IsAnimTex(int picnum)
{
	if(picnum < 0 || picnum >= numtextures) return 0;
	return (texArray[picnum]->txf_flags & TXF_ANIMPIC) > 0;
}

RDBOOL subfn_SecClosed(sector_t *sec, byte type)
{
	int i;
	int closed = 0;

	if (sec->linecount < 3) return false;

	CLR_vtxcheck();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->specType == type && sides[ld->sidenum[0]].sector == sec)
			if (!ld->mtflags &&
				ld->tyflags & TYF_MIDTEXLINE)
			{
				vtxcheck[ld->v1 - vertexes] |= 1;
				vtxcheck[ld->v2 - vertexes] |= 2;
				closed++;
			}
	}

	if (closed < 3) return false;

	for (i = 0; i < numvertexes; i++)
		if (vtxcheck[i] && vtxcheck[i] != 3)
			return 0;

	return true;
}

RDBOOL SecCtrl(int secnum)
{
	return (secspflags[secnum] & SECSPF_CTRLLINES) > 0;
}

RDBOOL SecACTN(int secnum)
{
	return (secspflags[secnum] & SECSPF_HASACTION) > 0;
}

RDBOOL SecONE(int secnum)
{
	return (secspflags[secnum] & SECSPF_LINONESEC) > 0;
}

RDBOOL SecMIDC(int secnum)
{
	return (secspflags[secnum] & SECSPF_MIDTEXCLOSED) > 0;
}

RDBOOL Vtfl(line_t *ld)
{
	return vtxflags[ld->v1 - vertexes] > 0
		&& vtxflags[ld->v2 - vertexes] > 0;
}

RDBOOL Vtfl3(line_t *ld)
{
	return vtxflags[ld->v1 - vertexes] == 3
		&& vtxflags[ld->v2 - vertexes] == 3;
}

RDBOOL Vtfl1(line_t *ld)
{
	return vtxflags[ld->v1 - vertexes] == 1
		&& vtxflags[ld->v2 - vertexes] == 1;
}

RDBOOL Vtxs1(line_t *ld)
{
	return vtxarray[ld->v1 - vertexes] == 1
		&& vtxarray[ld->v2 - vertexes] == 1;
}

RDBOOL Vtxs2(line_t *ld)
{
	return vtxarray[ld->v1 - vertexes] == 2
		&& vtxarray[ld->v2 - vertexes] == 2;
}

int GetLptr(line_t *ld)
{
	return lnsec[ld - lines];
}

RDBOOL ChkLptr(line_t *ld)
{
	return lnsec[ld - lines] >= 0;
}

RDBOOL CmpLptr(line_t *ld, int num)
{
	return lnsec[ld - lines] == num;
}

RDBOOL Spans1sLines(line_t *ld)
{
	return vtoflags[ld->v1 - vertexes] == 1
		&& vtoflags[ld->v2 - vertexes] == 1;
}

static RDBOOL secHasSSline(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		if (!(sec->lines[i]->ml_flags & ML_TWOSIDED))
			return true;
	}

	return false;
}

// v220-10
RDBOOL subfn_EquivalentSec(sector_t *a, sector_t *b, RDBOOL comptex)
{
	if (!a || !b) return false;
	if (a == b) return true;

	if (comptex)
	{
		if (   a->ceilingpic != b->ceilingpic
			|| a->floorpic != b->floorpic
			|| SecCtrl(a - sectors)
			|| SecCtrl(b - sectors))
		{
			return false;
		}
	}

	return a->ceilingheight == b->ceilingheight
		&& a->floorheight   == b->floorheight
		&& !(a->specFlags & SF_MOVSECTOR)
		&& !(b->specFlags & SF_MOVSECTOR);
}

// v220-10
void subfn_setSSEC_LFG(int num, RDBOOL clear_mtflags)
{
	line_t	*ld;
	int		i, k;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;
				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (clear_mtflags)
				{
					Clr_allmtflags(ld, num);
				}
				else if (!origTEX[ld - lines]
					&& ld->mtflags & R3D_MISLOTEX
					&& !(ld->mtflags & R3D_MISHITEX))
				{
					// will be set in MX_UnsetR3D_16 if necessary
					continue;
				}

				SetLFG(ld, LFG_NEWSET|LFG_FIXED, num);
			}
		}
	}
}

// v220-7 If a ssec has two lines and four different sectors (front and back)
//        and one of the frontsecs matches sec sent then return true
RDBOOL chk2LineSSEC(subsector_t *ss, sector_t *sec)
{
	if (ss->linecount == 2)
	{
		sector_t	*s[4];
		seg_t		*sg = &segs[ss->firstline];

		s[0] = sg->frontsector, s[1] = sg->backsector;

		if (!s[1] || s[0] == s[1]) return false;

		sg++;

		s[2] = sg->frontsector, s[3] = sg->backsector;

		if (!s[3] || s[2] == s[3]) return false;

		if (s[0] == s[2] || s[0] == s[3] || s[1] == s[2] || s[1] == s[3])
			return false;

		return (sec == s[0] || sec == s[2]);
	}

	return false;
}

int findContainingSec(sector_t *sec)
{
	int i, csec = XS_GetBoundarySecnum(sec);

	if (csec >= 0 && sectors[csec].specFlags & SF_NOLOTEX)
	{
		sector_t *s, *con = &sectors[csec];

		csec = -1;

		for (i = 0; i < con->linecount; i++)
		{
			line_t	 *ld = con->lines[i];

			if (origTEX[ld - lines]) continue;

			s = ld->frontsector == con ?
				ld->backsector : ld->frontsector;

			if (   !(s->specFlags & SF_NOLOTEX)
				|| secHasSSline(s))
			{
				csec = s - sectors;
				break;
			}
		}
	}

	return csec;
}

// Looks for same sided lines in a ssec array with a liquid floor flat
// v2.1.00 build_51
RDBOOL LQSSLines(sector_t *s, int *ovrsec)
{
	line_t	*ld;
	int		i, k, count = 0, secnum;

	if (!IsLiquidFloorFlat(s)) return false;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef)
				&& hitline[ld - lines] != 4)
			{
				if (ld->frontsector != ld->backsector
					|| ld->frontsector->floorpic != s->floorpic)
				{
					return false;
				}
				count++;
			}
		}
	}

	if (count && (secnum = findContainingSec(s)) >= 0) { *ovrsec = secnum; return true; }

	return false;
}

// b211-0
// if type == 1 only lower is checked
// if type == 2 only upper is checked
// default upper and lower checked
RDBOOL LS_ChkMissing(line_t *ld, int type)
{
	sector_t *fs, *bs;

	if (   ld->ml_flags & ML_TWOSIDED
		&& (fs = ld->frontsector) != (bs = ld->backsector))
	{
		int cdiff = fs->ceilingheight - bs->ceilingheight;
		int fdiff = fs->floorheight   - bs->floorheight;

		if (type != 1 && !(fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum))
		{
			if (   (cdiff > 0 && !sides[ld->sidenum[0]].toptexture)
				|| (cdiff < 0 && !sides[ld->sidenum[1]].toptexture))
			{
				return true;
			}
		}
		
		if (type != 2 && !(fs->floorpic == skyflatnum && bs->floorpic == skyflatnum))
		{
			if (   (fdiff > 0 && !sides[ld->sidenum[1]].bottomtexture)
				|| (fdiff < 0 && !sides[ld->sidenum[0]].bottomtexture))
			{
				return true;
			}
		}
	}
	return false;
}

sector_t *UT_FindLowestCeilingSector(sector_t* sec)
{
    sector_t	*other, *needed = NULL;
    fixed_t		height = 32000*FRACUNIT;
    int			i;

	for (i = 0; i < sec->linecount; i++)
	{
		if ((other = getNextSector(sec->lines[i],sec)) &&
			other->ceilingheight < height)
		{
			height = other->ceilingheight;
			needed = other;
		}
	}

    return needed;
}

RDBOOL LineSecEquiv(line_t *ld)
{
	vertexowner_t	*vo;
	sector_t		*f = ld->frontsector;
	sector_t		*b = ld->backsector;
	int				i;

	if (   !b
		|| f == b
		|| f->specFlags & SF_MOVSECTOR
		|| b->specFlags & SF_MOVSECTOR
		|| vtoflags[ld->v1 - vertexes]
		|| vtoflags[ld->v2 - vertexes])
	{
		return false;
	}

	if (   f->floorheight   != b->floorheight
		|| f->ceilingheight != b->ceilingheight)
	{
		return false;
	}

	vo = &vertexowners[ld->v1 - vertexes];

	for (i = 0; i < vo->num; i++)
	{
		if (   vo->list[i] != f - sectors
			&& vo->list[i] != b - sectors)
		{
			if ((sectors[vo->list[i]].specFlags & 0xC) != 0xC)
				return false;
		}
	}

	vo = &vertexowners[ld->v2 - vertexes];

	for (i = 0; i < vo->num; i++)
	{
		if (   vo->list[i] != f - sectors
			&& vo->list[i] != b - sectors)
		{
			if ((sectors[vo->list[i]].specFlags & 0xC) != 0xC)
				return false;
		}
	}
	return true;
}

// Sent a line ring pointed to by hitline returns list of ssecs
// inside as long as all the ssec lines belong to those in the ring
// or are same sector sided and belong to the sector.
//
RDBOOL SetAndCheckSSECSInLoop(sector_t *sec, sector_t *os)
{
	int		i, k, m;
	RDBOOL	valid = true;

	hitssecptr = 0;

	ss_validcount++;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!valid) break;

		if (hitline[ld - lines] == 4)
		{
			ssline_t *sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t	*ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;

				ss->ssvalidcount = ss_validcount;

				if (ss->sector == sec || (os && ss->sector == os))
				{
					seg_t *sg = &segs[ss->firstline];

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						if ( (ld = sg->linedef)
							&& hitline[ld - lines] != 4
							&& !(ld->frontsector == ld->backsector
							&& ld->frontsector == sec) )
						{
							if (!os || sg->frontsector != os)
							{
								valid = false; break;
							}
						}
					}
					hitssec[hitssecptr++] = ss - subsectors;
				}
			}
		}
	}

	return (valid && hitssecptr > 0);
}

RDBOOL secEquiv(sector_t *sec, line_t *ld, RDBOOL *diffsec)
{	
	sector_t *f = ld->frontsector;
	sector_t *b = ld->backsector;

	if (!b) return false;

	if (f == sec && b == sec) return true;

	if (   f->specFlags     != b->specFlags
		|| f->floorheight   != b->floorheight
		|| f->ceilingheight != b->ceilingheight)
	{
		return false;
	}

	if (diffsec) *diffsec = true;
	return true;
}

RDBOOL ssUnset1(sector_t *sc)
{
	return sc->R3Dceilsec < 0
		&& sc->R3Dfloorsec < 0
		&& sc->R3Dceilheight == DDMININT
		&& sc->R3Dfloorheight == DDMAXINT
		&& sc->floodline < 0;
}

RDBOOL ssUnset2(subsector_t *ss)
{
	return (ss->floorsec < 0 || ss->floorsec == ss->sector - sectors)
		&& (ss->ceilsec  < 0 || ss->ceilsec  == ss->sector - sectors)
		&& ss->lcsec < 0
		&& ss->lfsec < 0;
}

// Line 2's flag must have been checked before calling
RDBOOL chkSkyFloor(line_t *ld)
{
	if (ld->frontsector->floorheight > ld->backsector->floorheight
		&& ld->frontsector->floorpic == skyflatnum)
	{
		return true;
	}

	if (ld->frontsector->floorheight < ld->backsector->floorheight
		&& ld->backsector->floorpic == skyflatnum)
	{
		return true;
	}

	return false;
}

// Note: SSF_NOCHANGE not checked
RDBOOL SSECunset(subsector_t *ss)
{
	if (!ssUnset2(ss)) return false;
	return ssUnset1(ss->sector);
}

RDBOOL ChkSSECunset(subsector_t *ss)
{
	return	!(ss->ssflags & SSF_NOCHANGE)
		&& SSECunset(ss);
}

RDBOOL SecUnset(sector_t *sec)
{
	int i, k;

	if (!ssUnset1(sec)) return false;

	ss_validcount++;

	for (i = 0; i < sec->linecount; i++)
	{
		int		 lnum = sec->lines[i] - lines;
		ssline_t *sl = &sslines[lnum];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->sector == sec && !ssUnset2(ss))
				return false;
		}
	}
	return true;
}

RDBOOL ssecAllSameSec(sector_t *sec, int lnum)
{
	ssline_t *sl = &sslines[lnum];
	int		 i, k;

	if (!ssUnset1(sec)) return false;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->sector == sec)
		{
			seg_t *sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->frontsector != sec) return false;
			}
		}
	}

	return true;
}

float Get_LineLength(line_t *li)
{
	float fx = FIX2FLT(li->v2->x - li->v1->x),
		  fy = FIX2FLT(li->v2->y - li->v1->y);
	return (float)sqrt(fx*fx + fy*fy);
}

sector_t *GetLineSecFacingLine(line_t *ld, line_t *ref)
{
	int midx = meanV(ref->v1->x, ref->v2->x);
	int midy = meanV(ref->v1->y, ref->v1->y);

	return P_PointOnLineSide(midx, midy, ld) ? ld->backsector : ld->frontsector;
}

// Returns true if two lines are connected and form a single straight line
// based on the map's granularity rather than the more obvious test (180 deg).
// The latter is too accurate given vertices are truncated to an integer
// value and may not, therefore, give the required result.
static int subfn_CheckLSV(vertex_t *v1, vertex_t  *v2, vertex_t  *v3)
{
	int x  = v1->x, y = v1->y;
	int dx = (v2->x - v3->x) >> FRACBITS;			// Line vector
	int dy = (v2->y - v3->y) >> FRACBITS;
	int a  = (v2->x >> FRACBITS) - (x >> FRACBITS);	// point-->v1 vector
	int b  = (v2->y >> FRACBITS) - (y >> FRACBITS);
	
	// Check if common vertex lies on line drawn from other two vertices.
	
	a *= dy;
	b *= dx;
	a -= b;
	b = dx + dy;
	if (((a-b)^(a+b)) < 0) return true;
	dy -= dx;
	b = a+dy;
	a -= dy;
	return (a^b) < 0;
}

// Was named subfn_CheckLineSlope - changed to be more explicit
int	subfn_CheckConnectedLineSlope(line_t *ld, line_t *li)
{
	if (ld->v1 == li->v1) return subfn_CheckLSV(ld->v1, ld->v2, li->v2);
	if (ld->v1 == li->v2) return subfn_CheckLSV(ld->v1, ld->v2, li->v1);
	if (ld->v2 == li->v1) return subfn_CheckLSV(ld->v2, ld->v1, li->v2);
	if (ld->v2 == li->v2) return subfn_CheckLSV(ld->v2, ld->v1, li->v1);
	return false;
}

int ChkLFGSecSet(line_t *ld)
{
	// b211-0 LFG_NEWSET added
	return (lineflags[ld - lines] & (LFG_LINESSECSET|LFG_NEWSET|LFG_3DTYPE)) > 0;
}

void subfn_resetSector(int secnum, byte type)
{
	sector_t *s = &sectors[secnum];
	int i;

	Setr3dch_byH(s, DDMININT, 9002);
	Setr3dfh_byH(s, DDMAXINT, false, 9003);
	Setr3dcs(s, -1, 9004);
	Setr3dfs(s, -1, 9005);
	Setlfsec(s, -1, 9006);
	for (i = 0; i < s->linecount; i++)
	{
		line_t *ld = s->lines[i];

		Setspec (ld, -1, 9007);
		lineflags[ld - lines] &= ~LFG_SEC2SET;

		if (type != 0xFF)
		{
			if (type != R3D_01)
			{
			/*	if (ld->tyflags & SPF_NOTEX &&
					ld->mtflags)
					Settype(ld, R3D_3F, 9008);*/
				// Error meant this was never logically set
				// Should have been spflags not tyflags.
				// Assume not important - do nothing
			}
			else if (!origTEX[ld - lines])
				Settype(ld, R3D_01, 9009);
		}
	}
}

void LS_SetSectorsClosedFlag()
{
	sector_t *sec;
	line_t	 *ld;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		RDBOOL	valid = true;

		if (sec->linecount < 3) continue;

		CLR_vtxcheck();

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (ld->frontsector == ld->backsector && !Vtxs2(ld)) continue;	// v220-8
			if (ld->lnflags2 & LNF2_BADSSECLN) continue;	// b211-0

			vtxcheck[ld->v1 - vertexes]++;
			vtxcheck[ld->v2 - vertexes]++;
		}

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (ld->frontsector == ld->backsector && !Vtxs2(ld)) continue;	// v220-8
			if (ld->lnflags2 & LNF2_BADSSECLN) continue;	// b211-0

			if (vtxcheck[ld->v1 - vertexes] != 2)
			{
				if (   vtxcheck[ld->v1 - vertexes] > 2
					&& vertexowners[ld->v1 - vertexes].num == 1)
					continue;
				valid = false;
				break;
			}
			if (vtxcheck[ld->v2 - vertexes] != 2)
			{
				if (   vtxcheck[ld->v2 - vertexes] > 2
					&& vertexowners[ld->v2 - vertexes].num == 1)
					continue;
				valid = false;
				break;
			}
		}
		if (valid) sec->specFlags |= SF_SECCLOSED;
	}
}

// Initialises vtxarray and vtxflags arrays
void LS_InitVtxarray(void)
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType >= R3D_40) continue;

		vtxarray[ld->v1 - vertexes]++;
		vtxarray[ld->v2 - vertexes]++;
	}
}

// b211-0 Setup line crossing array
void LS_InitLXarray(void)
{
	line_t	*lc, *ld, *lf;
	byte	*lmem = (byte*)M_Calloc(numlines, 0);
	RDBOOL	hdr_printed = false;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		// v220-4 single sided line crossing check added
		if (!ld->backsector)
		{
			if (   !lmem[i]
				&& (lc = subfn_FindLineCrossing(ld, LC_SINGLESIDED))
				&& lc->frontsector != ld->frontsector)
			{
				if (!lc->backsector) lmem[lc - lines] = true;

				if (!hdr_printed)
				{
					pr_OFinsert("----------------------------------------------------\n"
						"!!! WARNING: Potentially serious map error found in %s !!!\n"
						"  The map has single sided lines crossing other lines in different sectors.\n"
						"  If a severe rendering problem is located near these lines then please report\n"
						"  this to the map author or check whether an updated map is available.\n", mapinfo->id);
					hdr_printed = true;
				}

				pr_OFinsert("MAP ERROR: 1s line %i (frontsec %i) is crossing a %is line %i (frontsec %i)\n",
					i, ld->frontsector - sectors, lc->backsector ? 2 : 1, lc - lines, lc->frontsector - sectors);

				ma_inf.maperror++;
			}
		}
		else
		if (ld->frontsector == ld->backsector)
		{
			if ( (lc = subfn_FindLineCrossing(ld, LC_NORMAL)) )
			{
				lxarray[i] |= 4;

				if (   ld->lnflags2 & LNF2_OWNEQUALTOONE
					&& !(origTEX[ld - lines] & 0x22)
					&& origTEX[lc - lines] == 0x22
					&& lc->frontsector == lc->backsector)
				{
					lxarray[i] |= 16;
				}

				// v220-22
				if ( !(ld->frontsector->specFlags & SF_MOVFLOOR)
					&& lc->frontsector == lc->backsector
					&& lc->frontsector->specFlags & SF_MOVFLOOR)
				{
					lxarray[i] |= 64;
				}

				if (subfn_FindLineCrossing(ld, LC_NOVTXONLINEGEN|LC_DIFFSECLINE) != NULL)
				{
					lxarray[i] |= 1;

					if (subfn_FindLineCrossing(ld, LC_NOVTXONLINEGEN|LC_DIFF_FLR_HT) != NULL)
						lxarray[i] |= 2;
				}
			}
			else if ( subfn_FindLineCrossing(ld, LC_VTXISONLINE|LC_VTXNOTSHARED) != NULL)
			{
				lxarray[i] |= 8;
			}
			
			if ( (lf = subfn_FindLineCrossing(ld, LC_SUPERIMPOSED)) != NULL)
			{
				lxarray[i] |= 32;
			}
		}
	}

	M_Free(lmem);

	if (hdr_printed)
		pr_OFdirect("----------------------------------------------------\n");
}

// b211-0 Added
void LS_CheckUnmergedVtxs()
{
	line_t	*ld, *li;
	int		i, show = ArgExists("-vtx");

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType < R3D_40)
		{
			if (vtxarray[ld->v1 - vertexes] == 1
				&& (li = subfn_UnmergedVTXline(ld, 2)))
			{
				ld->lnflags2 |= LNF2_VTXUNMERGED;

				if (show)
					pr_OFinsert("LINE %i vtx 0 is not merged with line %i\n",
						i, li - lines);
			}
			else if (vtxarray[ld->v2 - vertexes] == 1
				&& (li = subfn_UnmergedVTXline(ld, 4)))
			{
				ld->lnflags2 |= LNF2_VTXUNMERGED;

				if (show)
					pr_OFinsert("LINE %i vtx 1 is not merged with line %i\n",
						i, li - lines);
			}
		}
	}
}

static void setVtxflags(line_t *ld)
{
	if (   ld->ml_flags & ML_TWOSIDED
		&& !(ld->lnflags2 & LNF2_SLOPESECLN)
		&& ld->specType < R3D_40
		&& ld->frontsector != ld->backsector)
	{
		sector_t *fs = ld->frontsector;
		sector_t *bs = ld->backsector;
		int		 diff;
		RDBOOL	 found = false;

		if (   (diff = fs->ceilingheight - bs->ceilingheight)
			&& !(fs->ceilingpic == skyflatnum
			&&   bs->ceilingpic == skyflatnum) )
		{
			if (diff > 0 && !sides[ld->sidenum[0]].toptexture)
				found = true;
			else
			if (diff < 0 && !sides[ld->sidenum[1]].toptexture)
				found = true;
		}
		if (   !found
			&& (diff = fs->floorheight - bs->floorheight)
			&& !(fs->floorpic == skyflatnum
			&&   bs->floorpic == skyflatnum) )
		{
			if (diff > 0 && !sides[ld->sidenum[1]].bottomtexture)
				found = true;
			else
			if (diff < 0 && !sides[ld->sidenum[0]].bottomtexture)
				found = true;
		}
		if (found)
		{
			vtxflags2[ld->v1 - vertexes]++;
			vtxflags2[ld->v2 - vertexes]++;

			if (vtxflags[ld->v1 - vertexes] & 1)
				vtxflags[ld->v1 - vertexes] |= 2;
			else
				vtxflags[ld->v1 - vertexes] |= 1;

			if (vtxflags[ld->v2 - vertexes] & 1)
				vtxflags[ld->v2 - vertexes] |= 2;
			else
				vtxflags[ld->v2 - vertexes] |= 1;
		}
	}
}

// b211-0 was called from CheckUnwantedTextureRefs(sec)
// Modified to reset all lines and moved to precede LS_SetLineMTFlags()
void LS_InitVtflags()
{
	line_t	*ld;
	int		i, k;

	memset(vtxflags, 0, numvertexes);

	for (i = 0; i < numlines; i++)
	{
		if (lines[i].specType >= R3D_40)
			lineflags[i] |= LFG_NEWSET;
		else
			setVtxflags(&lines[i]);
	}

	// v220-7 revised. Check for certain types of missing uppertex
	// lines spanning 1's lines.
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& origTEX[i] == 0x22
			&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight
			&& Spans1sLines(ld)
			&& !(ld->frontsector->specFlags & SF_MOVCEIL)
			&& !(ld->backsector->specFlags  & SF_MOVCEIL)
			&& ld->frontsector->ceilingpic != skyflatnum
			&& ld->backsector->ceilingpic  != skyflatnum
			&& vtoflags[ld->v1 - vertexes] == 1
			&& vtoflags[ld->v2 - vertexes] == 1
			&& vtxflags[ld->v1 - vertexes] == 1
			&& vtxflags[ld->v2 - vertexes] == 1 )
		{
			short txn = 0;

			for (k = 0; k < ld->frontsector->linecount; k++)
			{
				if ( !(ld->frontsector->lines[k]->ml_flags & ML_TWOSIDED) )
				{
					txn = sides[ld->frontsector->lines[k]->sidenum[0]].midtexture;
					break;
				}
			}

			if (txn)
			{
				sides[ld->sidenum[1]].toptexture = txn;
				origTEX[ld - lines] |= 0x40;

				SetLFG(ld, LFG_FIXED|LFG_BADLINE, 60002);
				pr_OFinsert("MAP ERROR FIXED: L%i top tex added\n", i);
				ma_inf.maperror++;
			}
		}
	}
}

// b211-0 add
void LS_SetMultiLoopFlag()
{
	sector_t *sec;
	int		 c, i, k, m;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   sec->ceilingheight - sec->floorheight <= FRACUNIT
			&& sec->ceilingpic == skyflatnum)
		{
			int rc, bbox[16][4];
			CLR_hitline();

			for (k = 0; k < sec->linecount; k++)
			{
				hitline[sec->lines[k] - lines] = 1;
			}

			for (c = k = 0; k < sec->linecount; k++)
			{
				int lnum = sec->lines[k] - lines;

				if (hitline[lnum] == 1
					&& subfn_CheckForRing(hitline, sec, lnum, &rc, bbox[c], NULL, NULL))
				{
					if (++c >= 16) break;
				}
			}

			if (c > 1)
			{
				for (k = 0; k < c; k++)
				{
					for (m = k; m < c; m++)
					{
						if (M_CompareBox(bbox[k], bbox[m]) || M_CompareBox(bbox[m], bbox[k]))
						{
							sec->specFlags |= (SF_MULTILOOP|SF_NOOCCLUDE); goto next;
						}
					}
				}
			}
		}
next:;
	}
}

// Set sector LINESCLOSED flag
void LS_SetLinesclosedFlag()
{
	sector_t *sec;
	int		 i, k, count = 0;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->linecount > 2 &&
			sec->heightsec < 0)
		{
			sector_t *s1 = NULL, *s2 = NULL;
			int valid = true;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (LSS(ld))
					continue;

				if (!(ld->ml_flags & ML_TWOSIDED)
					|| !Vtxs2(ld)
					|| ld->frontsector->heightsec >= 0
					|| ld->backsector->heightsec >= 0)
				{
					valid = false;
					break;
				}
				count++;
			}
			if (!valid || !count) continue;

			CLR_vtxcheck();

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->frontsector == ld->backsector)
					continue;

				if (!s1) s1 = ld->frontsector;
				if (!s2) s2 = ld->backsector;
				if (s1 != ld->frontsector ||
					s2 != ld->backsector)
				{
					valid = false;
					break;
				}
				vtxcheck[ld->v1 - vertexes] |= 1;
				vtxcheck[ld->v2 - vertexes] |= 2;
			}

			if (!valid || !s1 || !s2) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->frontsector == ld->backsector)
					continue;

				if (vtxcheck[ld->v1 - vertexes] != 3 ||
					vtxcheck[ld->v2 - vertexes] != 3)
				{
					valid = false;
					break;
				}
			}

			if (valid)
				sec->specFlags |= SF_LINESCLOSED;
		}
	}
}

// Set flag to prevent mov secs sticking
// This is caused by mobj's being stuck in the ceiling where adjacent
// move sectors are being adjusted. (See T_MovePlane()) - rev85
// Also used for analysis
void LS_SetDualMovSecFlag()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->specFlags & SF_MOVFLOOR)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->backsector
					&& ld->specType < R3D_40
					&& ld->backsector != ld->frontsector
					&& ld->frontsector->specFlags & SF_MOVFLOOR
					&& ld->backsector->specFlags & SF_MOVFLOOR)
				{
					ld->frontsector->specFlags |= SF_DUALFMOVSEC;
					ld->backsector->specFlags |= SF_DUALFMOVSEC;
					break;
				}
			}
		}

		if (!(sec->specFlags & SF_MOVSECTOR)
			&& sec->ceilingheight == sec->floorheight
			&& sec->ceilingpic != skyflatnum)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if (ld->backsector
					&& ld->specType == R3D_16
					&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
					&& !(ld->backsector->specFlags & SF_MOVSECTOR))
				{
					SetNOocclude(sec, 9001);
					break;
				}
			}
		}
	}
}

// Check for faked moving sectors with R3D_11 lines
void LS_SetFakeMoveFlag()
{
	sector_t *sec;
	int		 i, k;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!secdone[i] &&
			sec->specFlags & SF_MOVSECTOR &&
			sec->heightsec >= 0)
		{
			for (k = 0; k < sec->linecount; k++)
			{
				if (sec->lines[k]->specType == R3D_11)
				{
					sec->specFlags |= SF_MOVFKR3D_11;
					Setsecdone(sec, 9010);
					break;
				}
			}
		}
	}
}

RDBOOL LS_SameSecLineCondition(sector_t *sec, int condition)
{
	int i;

	if (!(secspflags[sec - sectors] & SECSPF_LINONESEC))
		return false;

	if (condition)
	{
		for (i = 0; i < sec->linecount; i++)
		{
			if (sec->lines[i]->spflags & SPF_CONTROLLINE)
				continue;

			if (condition & NO_LINECROSSING)
			{
				if (subfn_FindLineCrossing(sec->lines[i], LC_NORMAL) != NULL)
				{
						return false;
				}
			}
			if (condition & NO_SINGLEVTX1)
			{
				if (vtxarray[sec->lines[i]->v1 - vertexes] == 1
					|| vtxarray[sec->lines[i]->v2 - vertexes] == 1)
				{
					return false;
				}
			}
			if (condition & NO_UNCONNLINE)
			{
				if (Vtxs1(sec->lines[i]))
				{
					return false;
				}
			}
			if (condition & NO_COINLINE)
			{
				if (subfn_FindCoincidentLine(sec->lines[i], 0) != NULL)
				{
					return false;
				}
			}
		}
	}

	return true;
}

// Example: Eternall MAP25 sec 874
RDBOOL LS_AllSameSecLine(sector_t *sec)
{
	int  i;

	if (!addflags) addflags = (char*)W_Calloc(numsectors, STATICID, 0);

	if (addflags[sec - sectors])
		return addflags[sec - sectors] > 0;

	if (!sec->linecount)
	{
		addflags[sec - sectors] = -1;
		return false;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		if (!(sec->lines[i]->spflags & SPF_CONTROLLINE))
		{
			if (   sec->lines[i]->specType >= R3D_40
				|| !(sec->lines[i]->ml_flags & ML_TWOSIDED)
				|| (sec->lines[i]->frontsector
				!=  sec->lines[i]->backsector  ) )
			{
				addflags[sec - sectors] = -1;
				return false;
			}
		}
	}

	addflags[sec - sectors] = 1;
	return true;
}


static RDBOOL subfn_CheckLineLoopsAndTex(int secnum)
{
	sector_t *sec = &sectors[secnum], *fsec = NULL, *bsec = NULL;
	int i, bottomtex = 0;

	if (SecCtrl(secnum) || SecONE(secnum))
		return false;

	fsec = sec->lines[0]->frontsector;
	bsec = sec->lines[0]->backsector;

	if (!bsec || fsec == bsec)
		return false;

	bottomtex = sides[sec->lines[0]->sidenum[0]].bottomtexture;

	if (!bottomtex)
		return false;

	CLR_vtxcheck();

	for ( i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->ml_flags & ML_TWOSIDED) ||
			ld->specType >= R3D_40 ||
			ld->mtflags ||
			ld->frontsector == ld->backsector ||
			ld->frontsector != fsec ||
			ld->backsector != bsec ||
			sides[ld->sidenum[0]].texFlags != T_LOW ||
			sides[ld->sidenum[1]].texFlags ||
			sides[ld->sidenum[0]].bottomtexture != bottomtex)
		{
			return false;
		}

		vtxcheck[ld->v1 - vertexes] |= 1;
		vtxcheck[ld->v2 - vertexes] |= 2;
	}

	for ( i = 0; i < sec->linecount; i++)
		if (vtxcheck[sec->lines[i]->v1 - vertexes] != 3 ||
			vtxcheck[sec->lines[i]->v2 - vertexes] != 3)
			return false;

	return true;
}

static RDBOOL subfn_CheckOuterLinesClosed(sector_t *sec, line_t *ld)
{
	int i, count = 1;
	RDBOOL repeat = true;

	CLR_vtxcheck();
	vtxcheck[ld->v1 - vertexes] |= 1;
	vtxcheck[ld->v2 - vertexes] |= 2;
	hitline[ld - lines] = 1;

	// Find all connected lines.
	while (repeat)
	{
		repeat = false;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			if (ld->spflags & SPF_CONTROLLINE ||
				hitline[ld - lines]) continue;

			if (vtxcheck[ld->v1 - vertexes])
			{
				if (vtxcheck[ld->v1 - vertexes] & 1)
				{
					vtxcheck[ld->v1 - vertexes] |= 2;
					vtxcheck[ld->v2 - vertexes] |= 1;
				}
				else
				{
					vtxcheck[ld->v1 - vertexes] |= 1;
					vtxcheck[ld->v2 - vertexes] |= 2;
				}

				hitline[ld - lines] = 1;
				count++;
				repeat = true;
				break;
			}
			if (vtxcheck[ld->v2 - vertexes])
			{
				if (vtxcheck[ld->v2 - vertexes] & 1)
				{
					vtxcheck[ld->v2 - vertexes] |= 2;
					vtxcheck[ld->v1 - vertexes] |= 1;
				}
				else
				{
					vtxcheck[ld->v2 - vertexes] |= 1;
					vtxcheck[ld->v1 - vertexes] |= 2;
				}

				hitline[ld - lines] = 1;
				count++;
				repeat = true;
				break;
			}
		}
	}
	if (count < 3) return false;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE ||
			!hitline[ld - lines])
			continue;

		if (vtxcheck[ld->v1 - vertexes] != 3 ||
			vtxcheck[ld->v2 - vertexes] != 3)
			return false;
	}
	return true;
}

static RDBOOL subfn_CheckMidTexOuterClosed(int secnum)
{
	line_t	*li;
	int		i, count = 0;

	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		line_t *ld = sectors[secnum].lines[i];

		if (ld->spflags & SPF_CONTROLLINE)
			continue;

		if (ld->frontsector != ld->backsector ||
			ld->specType ||
			origTEX[ld - lines] != 0x22)
			return false;

		count++;
	}

	if (count < 3) return false;

	CLR_hitline();
	// Must now check that the lines (or outer lines) are closed.
	if ((li = LS_FindOuterLine(&sectors[secnum])) == NULL)
		return false;

	return subfn_CheckOuterLinesClosed(&sectors[secnum], li);
}

void LS_SetTexAndSamesecFlags()
{
	sector_t *sec = sectors;
	int i;

	for (i = 0; i < numsectors; i++, sec++)
	{
		if (sec->linecount)
		{
			if (sec->heightsec < 0 && LS_AllSameSecLine(sec))
			{
				secspflags[i] |= SECSPF_LINONESEC;
				if (subfn_CheckMidTexOuterClosed(i))
					secspflags[i] |= SECSPF_MIDTEXCLOSED;
			}
			else if (subfn_CheckLineLoopsAndTex(i))
				secspflags[i] |= SECSPF_LNLOOPTEX;
		}
	}
}

RDBOOL LS_ChkAllVtx2(sector_t *sec, RDBOOL ctrl, RDBOOL mislotex)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ctrl
			&& ld->spflags & SPF_CONTROLLINE)
			continue;

		if (mislotex
			&& (origTEX[ld - lines]
				|| !(ld->mtflags & R3D_MISLOTEX)
				|| ld->mtflags & R3D_MISHITEX
				|| !Vtfl3(ld)))
			return false;

		if (!Vtxs2(ld))
			return false;
	}
	return true;
}

// killough 3/15/98:
//
// A fast function for testing intersections between a point x,y and linedefs.
// Radius (r) sets the area searched around x,y and must be at least 1
// Note: r is in map units (not fixed_t)
RDBOOL PointIsOnLine(fixed_t x, fixed_t y, line_t *l, int r)
{
	if (r < 1) r = 1;
	
	int dx = l->dx >> FRACBITS;
	int  a = (l->v1->x >> FRACBITS) - (x >> FRACBITS);

	// test x bounds of linedef and thing for intersection.
	if (abs(a + a + dx) - abs(dx) > r + r) return false;

	int dy = l->dy >> FRACBITS;
	int  b = (l->v1->y >> FRACBITS) - (y >> FRACBITS);

	// test y bounds of linedef and thing for intersection.
	if (abs(b + b + dy) - abs(dy) > r + r) return false;

	// Next, make sure that at least one thing crosshair intersects linedef's
	// extension. Requires only 3-4 multiplications, the rest adds/subs/
	// shifts/xors (writing the steps out this way leads to better codegen).
	
	a *= dy;
	b *= dx;
	a -= b;
	b = dx + dy;
	b *= r;
	if (((a-b)^(a+b)) < 0)
		return true;
	dy -= dx;
	dy *= r;
	b = a + dy;
	a -= dy;
	return (a^b) < 0;
}

// b211-0  XYonLine is a high precision version of PointIsOnLine
//	If r == 1 then max deviation is equiv to mapunit/FRACUNIT
RDBOOL XYonLine(fixed_t x, fixed_t y, line_t *l, fixed_t r)
{
	if (r < 1) r = 1;

	// v220-4 add box test
	if (   l->bbox[BOXLEFT ] < x + r
		&& l->bbox[BOXRIGHT] > x - r
		&& l->bbox[BOXBASE ] < y + r
		&& l->bbox[BOXTOP  ] > y - r)
	{
		__int64 dx = l->dx;
		__int64 dy = l->dy;
		__int64  a = l->v1->x - x;
		__int64  b = l->v1->y - y;
		
		a *= dy;
		b *= dx;
		a -= b;
		b = dx + dy;
		b *= r;
		if (((a-b)^(a+b)) < 0)
			return true;
		dy -= dx;
		dy *= r;
		b = a+dy;
		a -= dy;
		return (a^b) < 0;
	}
	return false;
}

RDBOOL VertexIsOnLine(vertex_t *vc, line_t *l)
{
	return XYonLine(vc->x, vc->y, l, 1);	// highest precision
}

// v220-5 modified to allow diffsec test
// v220-6 if on more than on line returns -1
int FindIfThingIsOnLine(mobj_t *th, RDBOOL diffsec)
{
	subsector_t	*ss = th->subsector;
	seg_t		*sg = &segs[ss->firstline];
	line_t		*ld, *li = NULL;
	int			i;

	// NOTE: validcount cannot always be used here

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (  (ld = sg->linedef)
			&& XYonLine(th->x, th->y, ld, 1))
		{
			if (  !diffsec
				|| (ld->frontsector != th->subsector->sector
				&&  ld->backsector  != th->subsector->sector) )
			{
				if (!li)
				{
					li = ld;
				}
				else
				{
					li = NULL; break;
				}
			}
		}
	}
	return li ? li - lines : -1;
}

// v220-31
int FindIfThingIsOn3DLine(mobj_t *th)
{
	subsector_t	*ss = th->subsector;
	seg_t		*sg = &segs[ss->firstline];
	line_t		*ld;
	int			i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (  (ld = sg->linedef)
			&& ld->specType >= R3D_40
			&& XYonLine(th->x, th->y, ld, 1))
		{
			return ld - lines;
		}
	}

	return -1;
}

// v220-6 changed so if diameter (not radius) fits in one axis then allowed
// v220-13 code added for shadows with fixed things
RDBOOL thingFitsInSector(mobj_t *iter, sector_t *sc)
{
	if (sc)
	{
		float *b = sc->sibounds;
		float d  = FIX2FLT(R_MaxRadius(iter, false, 1)) * 2.0f;

		if (d < 16.0f) d = 16.0f;	// Use a minimum value

		return ((b[BOXRIGHT] - b[BOXLEFT] > d) || (b[BOXTOP] - b[BOXBASE]) > d);
	}
	else	// v220-13 Used for shadows
	{
		line_t		*ld;
		int			i, r = R_MaxRadius(iter, false, 2);

		if (r < 0x200000) r = 0x200000;	// 32 FRACUNITS

		for (sc = iter->subsector->sector, i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (   ld->ml_flags & ML_TWOSIDED
				&& ld->frontsector != ld->backsector
				&& XYonLine(iter->x, iter->y, ld, r))
			{
				int side = P_PointOnLineSide(iter->x, iter->y, ld);

				if (   ( side && ld->frontsector->floorheight < ld->backsector->floorheight)
					|| (!side && ld->frontsector->floorheight > ld->backsector->floorheight) )
				{
					return false;
				}
			}
		}
	}

	return true;
}

RDBOOL subfn_CheckRingFromArray(byte *mrklinarray, int lnum)
{
	line_t		*ld = &lines[lnum];
	vertex_t	*v = ld->v1;
	int			i;
	RDBOOL		isclosed = false, repeat = true;

	mrklinarray[lnum] = 2;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < numlines; i++)
		{
			if (mrklinarray[i] != 1) continue;

			if (lines[i].v1 == v)
			{
				v = lines[i].v2;
				mrklinarray[i] = 2;
				repeat = true;
			}
			else if (lines[i].v2 == v)
			{
				v = lines[i].v1;
				mrklinarray[i] = 2;
				repeat = true;
			}
		}
	}

	if (v == ld->v2) isclosed = true;

	for (i = 0; i < numlines; i++)
	{
		if (mrklinarray[i] == 2)
		{
			if (isclosed)
				mrklinarray[i] = 4;
			else
				mrklinarray[i] = 1;
		}
	}
	return isclosed;
}

// All the list elements must form a closed ring
RDBOOL subfn_CheckRingFromList(int *list, int elements, int *box)
{
	line_t		*ld = &lines[list[0]];
	vertex_t	*v = ld->v1;
	byte		*mrkline = (byte*)M_Calloc(elements, 0);
	int			c, i;
	RDBOOL		repeat = true;

	mrkline[0] = 1; c = 1;
	if (box) M_ClearBox(box);

	while (repeat)
	{
		repeat = false;

		for (i = 1; i < elements; i++)
		{
			if (mrkline[i]) continue;

			if (lines[list[i]].v1 == v)
			{
				v = lines[list[i]].v2;
				if (box) M_AddLineToBox(box, &lines[list[i]]);
				mrkline[i] = 1; c++;
				repeat = true;
			}
			else if (lines[list[i]].v2 == v)
			{
				v = lines[list[i]].v1;
				if (box) M_AddLineToBox(box, &lines[list[i]]);
				mrkline[i] = 1; c++;
				repeat = true;
			}
		}
	}

	M_Free(mrkline);

	return (c == elements && v == ld->v2);
}

static sector_t *subfn_FindSector(line_t *ld, int x1, int y1)
{
	subsector_t *ssec;

	ssec = R_PointInSubsector(x1, y1);
	if (ssec->sector && ssec->sector != ld->frontsector && ssec->sector != ld->backsector)
	{
		return ssec->sector;
	}

	return NULL;
}

int subfn_SegFacingMidpoint(float v1[2], float v2[2], float mx, float my)
{
	// The dot product.
	return ((v1[VY] - v2[VY])*(v1[VX] - mx) +
			(v2[VX] - v1[VX])*(v1[VY] - my) > 0);
}

int subfn_SsecFacingLine(subsector_t *ss, line_t *ld)
{
	float v1[2];
	float v2[2];

	v1[VX] = FIX2FLT(ld->v1->x);
	v1[VY] = FIX2FLT(ld->v1->y);
	v2[VX] = FIX2FLT(ld->v2->x);
	v2[VY] = FIX2FLT(ld->v2->y);

	if (subfn_SegFacingMidpoint(v1, v2, ss->midpoint.x, ss->midpoint.y))
		return ld->frontsector - sectors;
	return ld->backsector - sectors;
}

// The global "sidelist" is set depending on the direction of the line relative to the first line
// tested. This allows for LS_MarkSubsecsInMultiSecLoop being able determine whether a point is
// inside or outside the line ring.
RDBOOL subfn_CheckForRing(byte *mrklinarray, sector_t *sec, int lnum, int *ring_count, int *bbox, int *linnb, int *seccheck)
{
	line_t		*ld = &lines[lnum];
	sector_t	*s1 = NULL;
	sector_t	*s2 = NULL;
	vertex_t	*v = ld->v1;
	int			i, c = 0;
	RDBOOL		isclosed = false, repeat = true;

	mrklinarray[lnum] = 2;

	if (seccheck) (*seccheck) = -1;

	sidelist[c++] = ld->sidenum[0];

	while (repeat)
	{
		repeat = false;
		for (i = 0; i < sec->linecount; i++)
		{
			line_t	*li = sec->lines[i];
			int		k = li - lines;

			if (mrklinarray[k] == 1)
			{
				if (li->v1 == v)
				{
					if (seccheck && !s1 && Vtxs2(li))
					{
						s1 = ld->frontsector;
						s2 = ld->backsector;
					}
					v = li->v2;
					mrklinarray[k] = 2;
					sidelist[c++] = -1;
					repeat = true;
				}
				else if (li->v2 == v)
				{
					if (seccheck && !s1 && Vtxs2(li))
					{
						s1 = ld->frontsector;
						s2 = ld->backsector;
					}
					v = li->v1;
					mrklinarray[k] = 2;
					sidelist[c++] = li->sidenum[0];
					repeat = true;
				}
			}
		}
	}

	if (v == ld->v2)
	{
		isclosed = true;

		if (   s1	// if seccheck is NULL then s1 will not be set
			&& !(s1->specFlags & SF_MOVFLOOR &&
			s2 && s2->specFlags & SF_MOVFLOOR))
		{
			for (i = 0; i < sec->linecount; i++)
			{
				line_t	*li = sec->lines[i];

				if (mrklinarray[li - lines] == 2)
				{
					if ( !(li->frontsector == s1
						&& li->backsector  == s2)
						&&
						 !(li->frontsector == s2
						&& li->backsector  == s1))
					{
						// By putting the line num in seccheck the front and back
						// sectors can be inspected (if required) by the caller.
						// Currently it is only the fact that another sector is
						// present that is of interest (see SECLOOPS [1c]).
						// This test is needed to eliminate sectors such as
						// ash2ash sec 92.
						if (!(li->frontsector->specFlags & SF_MOVFLOOR
							&& li->backsector
							&& li->backsector->specFlags & SF_MOVFLOOR))
						{
							*seccheck = li - lines;	// No need to get all lines
							break;
						}
					}
				}
			}
		}
	}

	if (ring_count) (*ring_count) = 0;
	if (linnb     ) (*linnb     ) = 0;
	if (bbox      ) M_ClearBox(bbox);

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;

		if (mrklinarray[lnum] == 2)
		{
			if (isclosed)
			{
				mrklinarray[lnum] = 4;
				if (ring_count) (*ring_count)++;
				if (linnb && !(sec->lines[i]->ml_flags & ML_TWOSIDED)) (*linnb)++;
				if (bbox) M_AddLineToBox(bbox, sec->lines[i]);
			}
			else
			{
				mrklinarray[lnum] = 1;
			}
		}
	}
	return isclosed;
}

// v220-16  Function that uses hitline and preset linearray by default
int subfn_LineRing(sector_t *sc, int count, int *box)
{
	int i, rc;

	if (count < 3) return 0;

	CLR_hitline();

	for (i = 0; i < count; i++)
	{
		hitline[linearray[i]] = 1;
	}

	if (subfn_CheckForRing(hitline, sc, linearray[0], &rc, box, NULL, NULL) && rc > 2)
		return rc;

	return 0;
}

static int subfn_SearchForBridgingLine(line_t *lb, vertex_t *v1, vertex_t *v2, int slidx)
{
	int		xl, xh, yl, yh, x, y;

	if (!lb) return -1;

	M_getBlockOffsets(lb->bbox, 0, &xl, &xh, &yl, &yh);

	for (x = xl ; x <= xh ; x++)
	{
		for (y = yl ; y <= yh ; y++)
		{
			const long	*list;
			int			offset;

			if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
				continue;

			offset = y*bmapwidth + x;

			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				if (lines[*list].v1 == v1 && lines[*list].v2 == v2)
				{
					sidelist[slidx] = -1;
					return *list;
				}
				if (lines[*list].v2 == v1 && lines[*list].v1 == v2)
				{
					sidelist[slidx] = lines[*list].sidenum[0];
					return *list;
				}
			}
		}
	}

	return -1;
}

RDBOOL subfn_CheckForRingInList(byte *mrklinarray, int *loopptr, int *looplist, sector_t *sec, int lnum, int *ring_count, int *bbox, int *nsec)
{
	line_t		*ld = &lines[lnum];
	vertex_t	*v1 = ld->v1, *v2;
	RDBOOL		isclosed = false, repeat = true;
	int			i, c = 0, bnum;
	line_t		*lb = &lines[lnum];

	mrklinarray[lnum] = 2;

	if (nsec) *nsec = -1;

	sidelist[c++] = ld->sidenum[0];

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < *loopptr; i++)
		{
			int k = looplist[i];

			if (mrklinarray[k] == 1)
			{
				line_t	*li = &lines[k];

				if (li->v1 == v1)
				{
					v1 = li->v2;
					mrklinarray[k] = 2;
					sidelist[c++] = -1;
					repeat = true;
				}
				else if (li->v2 == v1)
				{
					v1 = li->v1;
					mrklinarray[k] = 2;
					sidelist[c++] = li->sidenum[0];
					repeat = true;
				}
			}
		}
	}
	
	isclosed = (v1 == ld->v2);

	// Check if the vertex is connected to a line belonging to another sector.
	// The line must connect each end to separate lines in the looplist.
	if (nsec && !isclosed)
	{
		v1 = ld->v1, v2 = ld->v2;

		for (i = 0; i < *loopptr; i++)
		{
			if(mrklinarray[looplist[i]] == 2) mrklinarray[looplist[i]] = 1;
		}

		mrklinarray[ld - lines] = 2;
		repeat = true;

		while (repeat)
		{
			repeat = false;

			for (i = 0; i < *loopptr; i++)
			{
				int k = looplist[i];

				if (mrklinarray[k] == 1)
				{
					line_t	*li = &lines[k];

					if (li->v1 == v1)
					{
						v1 = li->v2;
						mrklinarray[k] = 2;
						lb = li;
						repeat = true;
					}
					else if (li->v2 == v1)
					{
						v1 = li->v1;
						mrklinarray[k] = 2;
						lb = li;
						repeat = true;
					}
					if (li->v1 == v2)
					{
						v2 = li->v2;
						mrklinarray[k] = 2;
						lb = li;
						repeat = true;
					}
					else if (li->v2 == v2)
					{
						v2 = li->v1;
						mrklinarray[k] = 2;
						lb = li;
						repeat = true;
					}
				}
			}
		}
		if ((bnum = subfn_SearchForBridgingLine(lb, v1, v2, c)) >= 0
			&& lines[bnum].ml_flags & ML_TWOSIDED
			&& lines[bnum].frontsector != lines[bnum].backsector)
		{
			int x = meanV(lb->v1->x, lb->v2->x);
			int y = meanV(lb->v1->y, lb->v2->y);
			int side = P_PointOnLineSide (x, y, &lines[bnum]);

			*nsec = side ? lines[bnum].backsector - sectors : lines[bnum].frontsector - sectors;

			looplist[(*loopptr)++] = bnum;
			mrklinarray[bnum] = 2;
			isclosed = true;
		}
		else
		{
			for (i = 0; i < *loopptr; i++)
			{
				if(mrklinarray[looplist[i]] == 2) mrklinarray[looplist[i]] = 8;
			}
		}
	}

	if (ring_count) (*ring_count) = 0;
	if (bbox      ) M_ClearBox(bbox);

	for (i = 0; i < *loopptr; i++)
	{
		line_t *li = &lines[looplist[i]];

		if (mrklinarray[li - lines] == 2)
		{
			if (isclosed)
			{
				mrklinarray[li - lines] = 4;
				if (ring_count) (*ring_count)++;
				if (bbox      ) M_AddLineToBox(bbox, li);
			}
			else
			{
				mrklinarray[li - lines] = 1;
			}
		}
	}

	return isclosed;
}

// Specific test for missing hi-tex loop
// The marked loop lines are sent in a byte sized array.  If conarray then must be indicated. 
RDBOOL LS_CheckFillCeilArea(sector_t *sec, byte *hitline, int ring_count, int callnum)
{
	sector_t	*osec = NULL;
	RDBOOL		useconarray = (hitline == conarray);
	int			i, count = 0, num = useconarray  ? 1 : 4;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];
		int		lnum = ld - lines;

		if (!ld->backsector) continue;

		if (hitline[lnum] == num
			&& !(origTEX[lnum] & 0x44)
			&& ld->frontsector->ceilingheight
			!= ld->backsector->ceilingheight)
		{
			if (!ld->backsector) return false;

			osec = ld->frontsector == sec ? ld->backsector : ld->frontsector;

			if (osec->ceilingheight < sec->ceilingheight)
			{
				count++;
			}
			else
				return false;
		}
	}

	if (count != ring_count) return false;

	if (!useconarray)
	{
		CLR_conarray();

		for (i = 0; i < sec->linecount; i++)
		{
			if (hitline[sec->lines[i] - lines] == 4)
				conarray[sec->lines[i] - lines] = 1;
		}
	}

	return LS_FillGenArea(sec, osec - sectors, FGA_SSCEILSEC, callnum);
}

// Test that can be used on a sector basis
int LS_ChkLineLoops(sector_t *sec, RDBOOL vtx2only, RDBOOL misslotex, RDBOOL clrhitline)
{
	int		count = 0, ring_count = 0, isclosed = false;
	int		i, lnum;

	if (clrhitline) CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		lnum = sec->lines[i] - lines;

		if (   sec->lines[i]->spflags & SPF_CONTROLLINE
			|| hitline[lnum])
		{
			continue;
		}

		if (misslotex)
		{
			if (   sec->lines[i]->specType == R3D_00
				&& !lineflags[lnum]
				&& Vtfl3(sec->lines[i])
				&& Vtxs2(sec->lines[i])
				&& !(origTEX[lnum] & 0x11))
			{
				hitline[sec->lines[i] - lines] = 1;
				count++;
			}
		}
		else
		if (vtx2only)
		{
			if (vtxarray[sec->lines[i]->v1 - vertexes] == 2
				&& vtxarray[sec->lines[i]->v2 - vertexes] == 2)
			{
				hitline[sec->lines[i] - lines] = 1;
				count++;
			}
		}
		else	// check all lines in sector
		{
			hitline[sec->lines[i] - lines] = 1;
			count++;
		}
	}

	if (count > 2)
	{
		for (i = 0; i < sec->linecount; i++)
		{
			lnum = sec->lines[i] - lines;

			if (hitline[lnum] == 1)
			{
				if ((isclosed = subfn_CheckForRing(hitline, sec, lnum, &ring_count, NULL, NULL, NULL)) > 0)
					break;
			}
		}

		if (!clrhitline && isclosed)
		{
			for (i = 0; i < sec->linecount; i++)
			{
				lnum = sec->lines[i] - lines;

				if (hitline[lnum] == 1) hitline[lnum] = 0;
			}
		}
	}

	return isclosed ? ring_count : 0;
}

// Checks for lines in a sector being closed head to tail in sequence
// Lines must have been premarked in markline array
RDBOOL LS_CheckRingInSec(byte *markline, line_t *ld, sector_t *s)
{
	vertex_t *v = ld->v1;

	RDBOOL isclosed = false, repeat = true;

	int		lnum = ld - lines;
	int		i;

	//--------------------------
	
	markline[lnum] = 2;

	while (repeat)
	{
		repeat = false;
		for (i = 0; i < s->linecount; i++)
		{
			int n = s->lines[i] - lines;

			if (markline[n] == 1)
			{
				if (lines[n].v2 == v)
				{
					v = lines[n].v1;
					markline[n] = 2;
					repeat = true;
					break;
				}
			}
		}
	}
	
	isclosed = v == ld->v2;

	if (!isclosed)
	{
		for (i = 0; i < s->linecount; i++)
		{
			if (markline[s->lines[i] - lines] == 2)
				markline[s->lines[i] - lines] = 1;
		}
	}
	return isclosed;
}

// The movsec list must first have been initialised
void AddToMovsecList(int secnum)
{
	int i;

	for(i = 1; i <= movsec_list[0]; i++)
		if (movsec_list[i] == secnum) return;

	movsec_list[0]++;
	movsec_list[movsec_list[0]] = secnum;
}

int getNextTag()
{
	int i, t = 0;

	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].tag > t)
			t = sectors[i].tag;
	}

	return t + 1;
}

void SetboxFloat(int *ibox, float *fbox)
{
	float xl = FIX2FLT(ibox[BOXLEFT ]);
	float xr = FIX2FLT(ibox[BOXRIGHT]);
	float yb = FIX2FLT(ibox[BOXBASE ]);
	float yt = FIX2FLT(ibox[BOXTOP  ]);

	if(xl < fbox[BOXLEFT ]) fbox[BOXLEFT ] = xl;
	if(xr > fbox[BOXRIGHT]) fbox[BOXRIGHT] = xr;
	if(yb < fbox[BOXBASE ]) fbox[BOXBASE ] = yb;
	if(yt > fbox[BOXTOP  ]) fbox[BOXTOP  ] = yt;

}

void Reset_SEC(int secnum)
{
	sector_t *s = &sectors[secnum];
	int		 i;

	s->R3Dceilheight  = DDMININT;
	s->R3Dfloorheight = DDMAXINT;
	s->R3Dceilsec  = -1;
	s->R3Dfloorsec = -1;

	for (i = 0; i < s->linecount; i++)
		Setspec(s->lines[i], -1, 9011);
}

void Reset_SSEC(int ssecnum, int num)
{
	subsector_t *ss = &subsectors[ssecnum];
	seg_t		*sg;
	int			i;

	ss->sector->R3Dceilheight  = DDMININT;
	ss->sector->R3Dfloorheight = DDMAXINT;
	ss->sector->R3Dceilsec = -1;
	ss->sector->R3Dfloorsec = -1;

	if (!(ss->ssflags & SSF_NOCHANGECS))
		Setr3dsscs(ss, -1, num);

	if (!(ss->ssflags & SSF_NOCHANGEFS))
		Setr3dssfs(ss, -1, num);

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if (sg->linedef) Setspec(sg->linedef, -1, num);
	}
}

void Clear_SSEC(subsector_t *ss)
{
	SetOverride(ss, -1, 9016);
	ss->sector->R3Dceilheight  = DDMININT;
	ss->sector->R3Dfloorheight = DDMAXINT;
	ss->sector->R3Dceilsec = -1;
	ss->sector->R3Dfloorsec = -1;
}

RDBOOL LineHasMovSectors(line_t *ld)
{
	if (ld->frontsector->specFlags & SF_MOVSECTOR)
		return true;
	if (ld->backsector &&
		ld->backsector->specFlags & SF_MOVSECTOR)
		return true;
	return false;
}

RDBOOL LineVertexConMovOrLiqSec(line_t *ld)
{
	int i, k, vtx[2];

	vtx[0] = ld->v1 - vertexes;
	vtx[1] = ld->v2 - vertexes;

	for (i = 0; i < 2; i++)
	{
		vertexowner_t *own = &vertexowners[vtx[i]];

		for (k = 0; k < own->num; k++)
		{
			sector_t *s = &sectors[own->list[i]];

			if (IsLiquidFloorFlat(s)) return true;

			if (s->specFlags & SF_MOVSECTOR) return true;
		}
	}
	
	return false;
}

RDBOOL LineVertexOwnsSec(line_t *ld, int secnum)
{
	vertexowner_t	*own = vertexowners + (ld->v1 - vertexes);
	int				i;

	if (ld->lnflags2 & LNF2_OWNEQUALTOONE)
		return own->list[0] == secnum;

	for (i = 0; i < own->num; i++)
		if (secnum == own->list[i])
			return true;
	
	own = vertexowners + (ld->v2 - vertexes);

	for (i = 0; i < own->num; i++)
		if (secnum == own->list[i])
			return true;

	return false;
}

RDBOOL HitssecMidpointOutsideBox(int *bbox)
{
	int i;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		int midx = (int)(ss->midpoint.x * FRACUNIT);
		int midy = (int)(ss->midpoint.y * FRACUNIT);

		if (   midx <= bbox[BOXLEFT ]
			|| midx >= bbox[BOXRIGHT]
			|| midy <= bbox[BOXBASE ]
			|| midy >= bbox[BOXTOP  ])
		{
			return true;
		}
	}
	return false;
}

RDBOOL idenLine(line_t *ld)
{
	int txn = origTEX[ld - lines];

	if (ld->lnflags2 & LNF2_SLOPESECLN) return false;

	if (   txn == 0x22
		&& ld->frontsector->floorheight   == ld->backsector->floorheight
		&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
	{
		return true;
	}

	return ld->frontsector->floorheight   == ld->backsector->floorheight
		&& (ld->frontsector->ceilingheight == ld->backsector->ceilingheight
		||  (!(txn & 0x44)
		&&   ld->frontsector->ceilingpic  == skyflatnum
		&&   ld->backsector->ceilingpic   == skyflatnum))
		&& !(!(txn & 0x11)
		&&  ld->frontsector->specFlags & SF_MOVFLOOR
		^   ld->backsector->specFlags  & SF_MOVFLOOR)
		&& !(!(txn & 0x44)
		&&  ld->frontsector->specFlags & SF_MOVCEIL
		^   ld->backsector->specFlags  & SF_MOVCEIL)
		&& ld->frontsector->lightlevel    == ld->backsector->lightlevel
		&& ld->frontsector->floorpic      == ld->backsector->floorpic
		&& ld->frontsector->ceilingpic    == ld->backsector->ceilingpic
		&& !(ld->frontsector->specFlags & SF_HASSECLIGHT)
		&& !(ld->backsector->specFlags & SF_HASSECLIGHT);
}

static void setLIB(libox_t *lib, sector_t *sec, line_t *lt)
{
	if (!(lt->ml_flags & ML_TWOSIDED))
	{
		lib->sslinecnt++;
		return;
	}

	if (lt->frontsector != lt->backsector)
	{
		lib->dsdiffsec++;
		if (idenLine(lt))
			lib->dssametype++;
	}
	else
	{
		if (lt->frontsector == sec)
			lib->dscompsec++;
		else
		{
			int fl = lt->frontsector->specFlags & SF_MOVSECTOR;

			lib->dssamesec++;
			lib->dssametype++;

			if (fl)
			{
				if (fl & SF_MOVFLOOR) lib->movfsec++;
				if (fl & SF_MOVCEIL ) lib->movcsec++;
			}

			if (lnsec[lt - lines] < 0)
			{
				lib->sameunset++;
			}
		}
	}
}

int subfn_setSSEClib(libox_t *lib, sector_t *sec, int *secstore)
{
	int i, k, m, secptr = 0;
	int fht = sec->floorheight;
	int cht = sec->ceilingheight;

	lib->diff_f = 0;
	lib->diff_c = 0;

	memset(lib, 0, sizeof(libox_t));
	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (!ld || ld->validcount == validcount)
				continue;

			ld->validcount = validcount;

			if (!lib->diff_f)
			{
				if (ld->frontsector->floorheight != fht
					|| 
					(ld->ml_flags & ML_TWOSIDED
					&& ld->backsector->floorheight != fht)
					|| 
					(ld->frontsector == ld->backsector
					&& ld->frontsector->floorpic != sec->floorpic
					&& ld->frontsector->specFlags & SF_MOVCEIL
					&& !(ld->frontsector->specFlags & SF_MOVFLOOR))
					|| 
					(ld->frontsector == ld->backsector
					&& ld->frontsector->floorpic != sec->floorpic
					&& !(origTEX[ld - lines] & 0x55)
					&& ld->frontsector->floorheight == sec->floorheight
					&& ld->frontsector->ceilingheight == sec->ceilingheight
					&& !(ld->frontsector->specFlags & SF_MOVSECTOR)))
				{
					lib->diff_f = true;
				}
			}

			if (ld->frontsector == ld->backsector)
			{
				int found = false, secn = ld->frontsector - sectors;

				for (m = 0; m < secptr; m++)
				{
					if (secstore[m] == secn)
					{
						found = true; break;
					}
				}

				if (!found) secstore[secptr++] = secn;
			}

			setLIB(lib, sec, ld);
		}
	}
	return secptr;
}

RDBOOL LS_MissingHiTex(subsector_t *ss)
{
	sector_t	*fs, *bs;
	seg_t		*sg = &segs[ss->firstline];
	int			i;

	for (i = 0; i < ss->linecount; i++)
	{
		if (sg->linedef
			&& !(origTEX[sg->linedef - lines] & 0x44)
			&& sg->linedef->ml_flags & ML_TWOSIDED
			&& (fs = sg->linedef->frontsector)
			!= (bs = sg->linedef->backsector)
			&& fs->ceilingheight != bs->ceilingheight
			&& !(fs->ceilingpic == skyflatnum
			&&   bs->ceilingpic == skyflatnum))
			return true;
	}
	return false;
}

// Inspects all lines totally inside the box (not if a vertex is on the edge)
// or if it's one sided and adds the line's sector to the secptr array.
// The number of members in the array is stored to ptr. 
void LS_SectorsFromLinesInBox(int *box, int *secptr, int *ptr)
{
	int		xl, xh, yl, yh, x, y;

	M_getBlockOffsets(box, 0, &xl, &xh, &yl, &yh);

	for (x = xl ; x <= xh ; x++)
	{
		for (y = yl ; y <= yh ; y++)
		{
			const long	*list;
			int			offset;

			if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
				continue;

			offset = y*bmapwidth + x;
			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				line_t *lt = &lines[*list];

				if (lt->validcount == validcount)
					continue;		// line has already been checked

				lt->validcount = validcount;

				if ( !(lt->ml_flags & ML_TWOSIDED) ) continue;

				if (   lt->bbox[BOXLEFT ] > box[BOXLEFT ]
					&& lt->bbox[BOXRIGHT] < box[BOXRIGHT]
					&& lt->bbox[BOXBASE ] > box[BOXBASE ]
					&& lt->bbox[BOXTOP  ] < box[BOXTOP  ])
				{
					int		i, n, secn = lt->frontsector - sectors;

					for (i = 0; i < 2; i++, secn = lt->backsector - sectors)
					{
						RDBOOL	found = false;

						for (n = 0; n < *ptr; n++)
						{
							if (secptr[n] == secn) { found = true; break; }
						}

						if (!found) secptr[(*ptr)++] = secn;
					}
				}
			}
		}
	}
}

int LS_LinesInBox(sector_t *sec, byte *mrkline, int *linelist, int *bbox, libox_t *lib, RDBOOL vtxinside)
{
	int		xl, xh, yl, yh, x, y;
	int		lineptr = 0;

	if (lib) memset(lib, 0, sizeof(libox_t));

	validcount++;

	M_getBlockOffsets(bbox, 0, &xl, &xh, &yl, &yh);

	for (x = xl ; x <= xh ; x++)
	{
		for (y = yl ; y <= yh ; y++)
		{
			const long	*list;
			int			offset;

			if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
				continue;

			offset = y*bmapwidth + x;

			offset = *(blockmap + offset);
			list = blockmaplump + offset;

			// skip starting delimiter
			while( *(++list) >= 0)
			{
				line_t *lt = &lines[*list];

				if (lt->validcount == validcount)
					continue;		// line has already been checked

				if (mrkline && mrkline[*list] == 4)
					continue;

				lt->validcount = validcount;

				if (   lt->bbox[BOXLEFT ] < bbox[BOXLEFT ]
					|| lt->bbox[BOXRIGHT] > bbox[BOXRIGHT]
					|| lt->bbox[BOXBASE ] < bbox[BOXBASE ]
					|| lt->bbox[BOXTOP  ] > bbox[BOXTOP  ])
				{
					int found = false;

					if (vtxinside && lt->frontsector == lt->backsector)
					{
						ssline_t *sl = &sslines[lt - lines];
						int		 i, k;

						for (i = 0; i < sl->num; i++)
						{
							subsector_t *ss = &subsectors[sl->list[i]];
							seg_t		*sg = &segs[ss->firstline];

							for (k = 0; k < ss->linecount; k++, sg++)
							{
								if (sg->linedef
									&& sg->linedef == lt)
								{
									if (   sg->v1->x >= bbox[BOXLEFT ]
										&& sg->v1->x <= bbox[BOXRIGHT]
										&& sg->v1->y >= bbox[BOXBASE ]
										&& sg->v1->y <= bbox[BOXTOP  ]
										&& sg->v2->x >= bbox[BOXLEFT ]
										&& sg->v2->x <= bbox[BOXRIGHT]
										&& sg->v2->y >= bbox[BOXBASE ]
										&& sg->v2->y <= bbox[BOXTOP  ])
									{
										found = true;
										break;
									}
								}
								if (found) break;
							}
						}
					}

					if (!found) continue;
				}

				if (lib)
				{
					setLIB(lib, sec, lt);
				}

				if (linelist)
					linelist[lineptr++] = lt - lines;
				else
					lineptr++;
			}
		}
	}

	return lineptr;
}

void LS_SubsecsInBox(int *bbox)
{
	subsector_t *ss;
	int			i;
	
	hitssecptr = 0;

	for(i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		ssec_midx = (int)(ss->midpoint.x * FRACUNIT);
		ssec_midy = (int)(ss->midpoint.y * FRACUNIT);

		if (ssec_midx > bbox[BOXLEFT ] &&
			ssec_midx < bbox[BOXRIGHT] &&
			ssec_midy > bbox[BOXBASE ] &&
			ssec_midy < bbox[BOXTOP  ])
		{
			if (ss->sector) hitssec[hitssecptr++] = i;
		}
	}
}

// If sec is true then test for one's lines can be made
RDBOOL LS_OnlySSLinesInSSECarray(sector_t *sec, byte *mrklinearray)
{
	int i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& mrklinearray[sg->linedef - lines] != 4
				&& mrklinearray[sg->linedef - lines] != 64)
			{
				if (!sg->linedef->backsector)
				{
					if (sg->linedef->frontsector != sec)
						return false;
				}
				else if (sg->linedef->frontsector != sg->linedef->backsector)
					return false;
			}
		}
	}

	return true;
}

RDBOOL LS_NoHiTexInSSECarray()
{
	int i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef)
			{
				if (origTEX[sg->linedef - lines] & 0x44)
					return false;
			}
		}
	}

	return true;
}

RDBOOL LS_SsecAllSegsFrontSame(sector_t *s)
{
	ssline_t	*sl;
	int			i, k;

	hitssecptr = 0;
	ss_validcount++;

	for (i = 0; i < s->linecount; i++)
	{
		sl = &sslines[s->lines[i] - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;

				if (ss->sector == s)
					hitssec[hitssecptr++] = sl->list[k];
			}
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t	*ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->frontsector && sg->frontsector != s)
				return false;
		}
	}

	return true;
}

RDBOOL LS_LinesInside(byte *mrklinearray, sector_t *s)
{
	line_t	*ld;
	int		i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) != NULL
				&& ld->frontsector != ld->backsector
				&& mrklinearray[ld - lines] != 4)
			{
				if (ld->frontsector != s || ld->backsector != s)
					return true;
			}
		}
	}
	return false;
}

// v220-9 added. Lines in linearray will be used
// Returns true if a line not belonging to the loop is found inside the loop ring
RDBOOL LS_LinesInsideLoop(sector_t *sec, int loopcount)
{
	int		i, k, m;
	int		mrkloop[256];
	int		mrkcount = 0;

	CLR_hitline();
	hitssecptr = 0;
	ss_validcount++;

	for (i = 0; i < loopcount; i++) hitline[linearray[i]] = 1;

	for (i = 0; i < loopcount; i++)
	{
		ssline_t *sl = &sslines[linearray[i]];
		line_t	 *ld;

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			if (ss->sector == sec)
			{
				seg_t *sg = &segs[ss->firstline];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if ((ld = sg->linedef) && !(hitline[ld - lines] & 1))
					{
						if (ld->frontsector == sec && ld->backsector == sec)
						{
							if ( !(hitline[ld - lines] & 2) )
							{
								if (mrkcount > 255) return true;
								mrkloop[mrkcount++] = ld - lines;
								hitline[ld - lines] |= 2;
							}
						}
						else
							return true;
					}
				}

				hitssec[hitssecptr++] = sl->list[k];
			}
		}
	}

	for (i = 0; i < mrkcount; i++)
	{
		ssline_t *sl = &sslines[mrkloop[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			if (ss->sector != sec) return true;

			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	return false;
}

RDBOOL LS_SSEC_FloorSet()
{
	int		i;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		if (   ss->floorsec >= 0
			|| ss->sector->R3Dfloorheight != DDMAXINT
			|| ss->sector->R3Dfloorsec >= 0)
		{
			return true;
		}
	}
	return false;
}

// v220-7 check osec is owned by same sector sided lines and that these lines exist..
RDBOOL LS_HasOsecSSLinesInSSECarray(sector_t *sec, sector_t *osec, byte *mrklinearray, line_t **ld, int *ldcount)
{
	int		c, i, k;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   sg->linedef
				&& sg->linedef->validcount != validcount)
			{
				ld[(*ldcount)++] = sg->linedef;

				sg->linedef->validcount = validcount;

				if (mrklinearray[sg->linedef - lines] != 4)
				{
					if (   sg->linedef->frontsector != osec
						|| sg->linedef->backsector  != osec)
					{
						return false;
					}
					c++;
				}
			}
		}
	}

	return c > 0;
}

RDBOOL LS_LinesInSSECarray(byte *mrklinearray, int flags, int *count)
{
	int i, k;

	if (count) *count = 0;

	// v220-8 add test to eliminate any enclosed slopes
	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (ss->sector->specFlags2 & (SF2_LOSLOPE|SF2_HISLOPE)) return false;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (!sg->linedef) continue;

			if (   sg->linedef->lnflags2 & LNF2_SLOPESECLN
				|| sg->linedef->specType >= R3D_40)
			{
				return false;
			}
		}
	}

	if (flags & LSS_SKYTEST)
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (sg->linedef
					&& mrklinearray[sg->linedef - lines] != 4)
				{
					sector_t *fs = sg->linedef->frontsector;
					sector_t *bs = sg->linedef->backsector;

					if (!Vtxs2(sg->linedef)
						|| !(sg->linedef->ml_flags & ML_TWOSIDED)
						|| (fs->ceilingpic != skyflatnum
						&&  bs->ceilingpic != skyflatnum))
					{
						return false;
					}
				}
			}
		}
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (flags & LSS_SSECSET && !ChkSSECunset(ss))
			return true;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef && mrklinearray[sg->linedef - lines] != 4)
			{
				if (count)
				{
					if (   (ss->floorsec >= 0 && ss->floorsec == ss->ceilsec)
						|| sg->linedef->frontsector != sg->linedef->backsector
						|| origTEX[sg->linedef - lines])
					{
						return true;
					}

					(*count)++; continue;
				}

				if (flags & LSS_NONE)
				{
					return true;
				}
				else if (flags & LSS_SKYTEST)
				{
					sector_t *fs = sg->linedef->frontsector;
					sector_t *bs = sg->linedef->backsector;

					if (fs && bs && fs != bs
						&& (fs->ceilingpic != skyflatnum
						||  bs->ceilingpic != skyflatnum))
					{
						if (fs->ceilingpic == skyflatnum
							&& bs->ceilingheight < fs->ceilingheight)
						{
							return true;
						}
						if (bs->ceilingpic == skyflatnum
							&& fs->ceilingheight < bs->ceilingheight)
						{
							return true;
						}
					}
				}
				else if (flags & LSS_HASULTEX)
				{
					if (origTEX[sg->linedef - lines] & 0x55)
					{
						return true;
					}
				}
				else if (flags & LSS_HASTEX)
				{
					if (origTEX[sg->linedef - lines])
					{
						return true;
					}
				}
				else
				{
					sector_t *fs = sg->linedef->frontsector;
					sector_t *bs = sg->linedef->backsector;

					if (fs == bs)
					{
						if (!(flags & LSS_EXCSAMESD))
						{
							return true;
						}
					}
					else if (sg->linedef->ml_flags & ML_TWOSIDED)
					{
						if (flags & LSS_EXCSAMEHT
							&& fs->floorheight == bs->floorheight
							&& fs->ceilingheight == bs->ceilingheight)
						{
							continue;
						}
						if (flags & LSS_DIFFSEC)
						{
							return true;
						}
						if (flags & LSS_DIFFIDEN
							&& idenLine(sg->linedef))
						{
							return true;
						}
					}
				}
			}
		}
	}
	return false;
}

int LS_MarkLinesInSSECarray(sector_t *sec, byte *mrklinearray, byte *exclude)
{
	int i, k, lnptr = 0;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (exclude[hitssec[i]] || !ChkSSECunset(ss)) continue;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *li = sg->linedef;

			if (!li || (exclude[hitssec[i]] && li->validcount == validcount))
			{
				continue;
			}

			if (mrklinearray[li - lines] != 4)
			{
				if (li->frontsector == li->backsector)
				{
					if (   li->validcount != validcount
						&& li->frontsector != sec)
					{
						linearray[lnptr++] = li - lines;
					}
				}
				else if (li->ml_flags & ML_TWOSIDED)
				{
					if (origTEX[li - lines] & 0x11
						&& sg->frontsector->specFlags & SF_MOVFLOOR)
					{
						exclude[hitssec[i]] = true;
						continue;
					}
					if (origTEX[li - lines] & 0x55
						&& sg->frontsector->specFlags & SF_MOVCEIL)
					{
						exclude[hitssec[i]] = true;
						continue;
					}
					if (!origTEX[li - lines] || idenLine(li))
					{
						if (li->validcount != validcount)
							linearray[lnptr++] = li - lines;
					}
					else if (li->frontsector != sec
						&& sg->linedef->backsector != sec)
					{	
						return 0;
					}
				}
			}
			li->validcount = validcount;
		}
	}
	return lnptr;
}

void LS_ExcludeSubsecsInside(sector_t *sec, byte *sseclist)
{
	int			i, k, lnum, ring_count, bbox[4];
	int			hittemp   = hitssecptr;
	int			*ssectemp = (int*)M_Calloc(hitssecptr * sizeof(int), 310);

	memcpy(ssectemp, hitssec, hitssecptr * sizeof(int));
	memcpy(conarray, hitline, numlines);	// v220-13

	for (i = 0; i < sec->linecount; i++)
	{
		if (hitline[sec->lines[i] - lines] == 4)
		{
			conarray[sec->lines[i] - lines] = 8;
		}
	}

	for (i = 0; i < sec->linecount; i++)
	{
		lnum = sec->lines[i] - lines;

		if (conarray[lnum] == 1
			&& subfn_CheckForRing(conarray, sec, lnum, &ring_count, bbox, NULL, NULL))
		{
			int valid = false;

			if (LS_MarkSubsecsInAreaGeneral(sec, conarray, 0, NULL, bbox, -1, MSIA_TESTINNER))
			{
				for (k = 0; k < hitssecptr; k++)
				{
					sseclist[hitssec[k]] |= 2;
				}

				valid = true;
			}

			for (k = 0; k < sec->linecount; k++)
			{
				if (conarray[sec->lines[k] - lines] == 4)
				{
					conarray[sec->lines[k] - lines] = 8;
					if (valid) hitline[sec->lines[k] - lines] = 8;
				}
			}
		}
	}
	
	hitssecptr = hittemp;
	memcpy(hitssec, ssectemp, hitssecptr * sizeof(int));
	M_Free(ssectemp);
}

// Looplist must be set - mrkline and bbox are optional
int subfn_CheckLoopClosed(int lptr, int loopcnt, int *looplist, int *bbox, byte *mrkline)
{
	line_t		*li = &lines[looplist[lptr]];
	vertex_t	*v1 = li->v1, *v2 = li->v2;
	RDBOOL		repeat = true;
	byte		*mrkloop = (byte*)M_Calloc(loopcnt, 311);
	int			i, count = 0;

	if (mrkline) memcpy(mrkloop, mrkline, loopcnt);

	mrkloop[lptr] = 1;

	if (bbox) M_ClearBox(bbox);
	if (bbox) M_AddLineToBox(bbox, li);

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < loopcnt; i++)
		{
			if (!mrkloop[i])
			{
				li = &lines[looplist[i]];

				if (li->v1 == v1)
				{
					v1 = li->v2;
					mrkloop[i] = 1;
					if (bbox) M_AddLineToBox(bbox, li);
					count++;
					repeat = true;
				}
				else if (li->v2 == v1)
				{
					v1 = li->v1;
					mrkloop[i] = 1;
					if (bbox) M_AddLineToBox(bbox, li);
					count++;
					repeat = true;
				}
			}
		}
	}

	if (v1 == v2 && mrkline)
	{
		for (i = 0; i < loopcnt; i++)
		{
			if (mrkloop[i] == 1) mrkline[i] = 1;
		}
	}

	M_Free(mrkloop);
	
	return (v1 == v2) ? count + 1: 0;
}

sector_t *getLoopSecFromBox(line_t *ld, int *bbox)
{
	int midx = meanV(bbox[BOXLEFT], bbox[BOXRIGHT]);
	int midy = meanV(bbox[BOXTOP ], bbox[BOXBASE ]);

	return P_PointOnLineSide(midx, midy, ld) ? ld->backsector : ld->frontsector;
}

// rev 170s
RDBOOL LS_SetSSECSinLoop(sector_t *s, int *secnum, int loopcnt, int *looplist, int flags)
{
	int	bbox[4];

	if (   loopcnt
		&& subfn_CheckLoopClosed(0, loopcnt, looplist, bbox, NULL) == loopcnt)
	{
		if (!s && secnum)
		{
			s = getLoopSecFromBox(&lines[looplist[0]], bbox);
			*secnum = (s == NULL) ? -1 : s - sectors;
		}
		return s && LS_MarkSubsecsInLoop(s, loopcnt, looplist, bbox, flags, NULL);
	}
	return false;
}

// LS_SetSSINF =========================================
//
//

static RDBOOL infTexMissing(line_t *ld, RDBOOL cdiff, RDBOOL fdiff)
{
	sector_t *fs = ld->frontsector;
	sector_t *bs = ld->backsector;
	side_t   *sd;

	// special case
	if (ld->lnflags2 & LNF2_MISSINGSPEC)
	{
		ssinf.missingspec++;

		if (cdiff)	// get lowest ceiling
		{
			ssinf.specceil = fs->ceilingheight > bs->ceilingheight ?
				bs - sectors : fs - sectors;
		}

		if (fdiff)	// get lowest floor
		{
			ssinf.specfloor = fs->floorheight > bs->floorheight ?
				bs - sectors : fs - sectors;
		}

		return true;
	}

	if (!(origTEX[ld - lines] & 0x55))
	{
		if (   fs->specFlags & SF_MOVSECTOR
			|| bs->specFlags & SF_MOVSECTOR)
		{
			return true;
		}
		return false;
	}

	if (origTEX[ld - lines] & 0x11)
	{
		if (fs->floorheight != bs->floorheight)
		{
			if (   fs->specFlags & SF_MOVFLOOR
				|| bs->specFlags & SF_MOVFLOOR)
				return false;

			if (fs->floorheight > bs->floorheight)
				sd = &sides[ld->sidenum[1]];
			else
				sd = &sides[ld->sidenum[0]];

			if (!sd->bottomtexture) return true;
		}
	}

	if (origTEX[ld - lines] & 0x44)
	{
		if (fs->ceilingheight != bs->ceilingheight)
		{
			if (   fs->specFlags & SF_MOVCEIL
				|| bs->specFlags & SF_MOVCEIL)
				return false;

			if (fs->ceilingheight < bs->ceilingheight)
				sd = &sides[ld->sidenum[1]];
			else
				sd = &sides[ld->sidenum[0]];

			if (!sd->toptexture) return true;
		}
	}
	

	return false;
}

RDBOOL twoSKY(line_t *ld, RDBOOL hi)
{
	if (ld->ml_flags & ML_TWOSIDED)
	{
		if (hi) return ld->frontsector->ceilingpic == skyflatnum
			&& ld->backsector->ceilingpic == skyflatnum;
		
		return ld->frontsector->floorpic == skyflatnum
			&& ld->backsector->floorpic == skyflatnum;
	}
	return false;
}

// Returns any hi or lo tex found in flags for a two sided line
// having different front and back sectors.
int LS_TestLineTex(line_t *ld)
{
	sector_t	*fs = ld->frontsector;
	sector_t	*bs = ld->backsector;
	int			cdiff, fdiff;
	int			flags = 0;

	if (!bs || fs == bs) return 0;

	cdiff = fs->ceilingheight - bs->ceilingheight;
	fdiff = fs->floorheight   - bs->floorheight;

	if (   cdiff
		&& fs->ceilingpic == skyflatnum
		&& bs->ceilingpic == skyflatnum)
	{
		cdiff = 0;
	}

	if (   fdiff
		&& fs->floorpic == skyflatnum
		&& bs->floorpic == skyflatnum)
	{
		fdiff = 0;
	}

	if (cdiff)
	{
		if (abs(cdiff) > 8*FRACUNIT) flags |= LT_DIFFHI;

		if (   (cdiff > 0 && sides[ld->sidenum[0]].toptexture)
			|| (cdiff < 0 && sides[ld->sidenum[1]].toptexture))
		{
			flags |= LT_HI;
		}
		else
		{
			flags |= LT_MISHI;
		}
	}

	if (fdiff)
	{
		if (abs(fdiff) > 8*FRACUNIT) flags |= LT_DIFFLO;

		if (   (fdiff > 0 && sides[ld->sidenum[1]].bottomtexture)
			|| (fdiff < 0 && sides[ld->sidenum[0]].bottomtexture))
		{
			flags |= LT_LO;
		}
		else
		{
			flags |= LT_MISLO;
		}
	}
	return flags;
}

RDBOOL checkMissingTexLine(line_t *ld)
{
	sector_t *fs = ld->frontsector;
	sector_t *bs = ld->backsector;

	if (bs)
	{
		int cdiff = twoSKY(ld,1) ? 0 : fs->ceilingheight - bs->ceilingheight;
		int fdiff = twoSKY(ld,0) ? 0 : fs->floorheight   - bs->floorheight;
		int txn   = origTEX[ld - lines];

		if (abs(fdiff) > 1024*FRACUNIT) fdiff = 0;
		if (abs(cdiff) > 1024*FRACUNIT) cdiff = 0;

		return (cdiff && !(txn & 0x44)) || (fdiff && !(txn & 0x11));
	}
	return false;
}

// b211-0 Added
RDBOOL invalidTEX_movfType(line_t *ld)
{
	int txn, txt, fmovf, bmovf;

	if (!(ld->ml_flags & ML_TWOSIDED)) return false;

	txn = origTEX[ld - lines];

	fmovf = ld->frontsector->specFlags & SF_MOVFLOOR;
	bmovf = ld->backsector->specFlags  & SF_MOVFLOOR;

	txt = txn & 0x11;

	if (txt == 0x11 && !(fmovf || bmovf)) return true;

	if ((fmovf^bmovf))
	{
 		if (txt == 0x01 && !bmovf) return true;
 		if (txt == 0x10 && !fmovf) return true;
	}

	return false;
}

// b211-0 Added
void LS_SetSECDONE(line_t *ld, int num)
{
	if (!secdone[ld->frontsector - sectors]) Setsecdone3(ld->frontsector, num);
	if (ld->backsector
		&& !secdone[ld->backsector - sectors]) Setsecdone3(ld->backsector, num);
}

// v220-21 added
static RDBOOL isMissingLoop(sector_t *ts, line_t *ld)
{
	sector_t	*sc = ld->frontsector == ts ? ld->backsector : ld->frontsector;

	if (sc && (sc->specFlags & 0xF0C) == 0x408)
	{
		int			c = 0;
		int			*lnarray = (int*)malloc(sc->linecount * sizeof(int));

		if ( (c = nf_lineList(ld, sc, NULL, lnarray, 1)) > 2)
		{
			sector_t	*os;
			int			i;

			for (i = 0; i < c; i++)
			{
				os = lines[lnarray[i]].frontsector == sc ? lines[lnarray[i]].backsector : lines[lnarray[i]].frontsector;

				if (sc->floorheight - os->floorheight < 8*FRACUNIT)
				{
					free(lnarray);
					return false;
				}
			}

			free(lnarray);
			return true;
		}

		free(lnarray);
		return false;
	}

	return false;
}

// b211-0 Added
// Note: validcount must be set before calling
void LS_SetLines(sector_t *sec, subsector_t *ss, RDBOOL noR3D_16, RDBOOL forceR3D_16, RDBOOL checkTypeR3D_16, int num)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		line_t	*li = sg->linedef;
		int		txn;

		if (!li || li->validcount == validcount) continue;

		li->validcount = validcount;

		if (li->ml_flags & ML_TWOSIDED && Vtxs2(li))
		{
			asyflags[li->sidenum[0]] = 0;
			asyflags[li->sidenum[1]] = 0;
		}

		if (li->spflags & SPF_CONTROLLINE) continue;

		txn = origTEX[li - lines];

		// v220-6 add test
		if (   li->backsector
			&& !(txn & 0x11)
			&& txn & 0x22
			&& !(li->frontsector->specFlags & SF_MOVFLOOR)
			&& !(li->backsector->specFlags  & SF_MOVFLOOR) )
		{
			if (asyflags[li->sidenum[0]] & ASY_L) txn |= 0x01;
			if (asyflags[li->sidenum[1]] & ASY_L) txn |= 0x10;
		}

		// v220-8 added
		if (  !txn
			&& li->ml_flags & ML_TWOSIDED
			&& (li->mtflags == 0x20 || li->mtflags == 0x80)
			&& li->frontsector != li->backsector
			&& Vtxs2(li)
			&& (li->frontsector->specFlags & 0x1000C) == 0xC
			&& (li->backsector->specFlags  & 0x1000C) == 0xC
			&& li->frontsector->ceilingpic == li->backsector->ceilingpic
			&& li->frontsector->ceilingpic != skyflatnum)
		{
			Settype(li, R3D_16, num);
		}
		else if (!(li->ml_flags & ML_TWOSIDED)
			// Test for stuff like TVR! map21
			|| (!txn
			&& Vtxs2(li)
			&& !LineSecEquiv(li)
			&& !(li->frontsector->specFlags & SF_MOVSECTOR)
			&& !(li->backsector->specFlags  & SF_MOVSECTOR)
			&& (IsLiquidFloorFlat(li->frontsector)
			||  IsLiquidFloorFlat(li->backsector))))
		{
			Settype(li, R3D_00, num);
		}
		// Test for stuff like TVR! map05 sec 3, 4, 5
		else if (li->mtflags == 0x41
			&& Vtxs2(li)
			&& origTEX[li - lines] == 0x01
			&& IsLiquidFloorFlat(li->frontsector)
			&& !IsLiquidFloorFlat(li->backsector)
			&& (li->frontsector->specFlags & 0xFF0004) == (SF_MOVFLOOR|SF_MOVLOWER)
			&& (li->backsector->specFlags & 0xF000C) == 0xC)
		{
			Settype(li, R3D_14, num);
		}
		else
		{
			byte settype = R3D_16;

			if (!forceR3D_16)
			{
				if (txn & 0x22)
				{
					if (!(txn & 0x11)
						&& li->frontsector->floorheight != li->backsector->floorheight)
					{
						if (Vtxs2(li) && (txn == 0x02 || txn == 0x20)
							&& li->frontsector->ceilingheight == li->backsector->ceilingheight)
						{
							settype = R3D_16;
						}
						else
						{
							// v220-22 add sector test - previously R3D_1D was always set
							sector_t *ts = li->frontsector == ss->sector ? li->backsector : li->frontsector;

							if (ts && (ts->specFlags & SF_NOLOTEX || isMissingLoop(ts, li)))
								settype = R3D_1D;
							else
								settype = R3D_2D;

							if (txn == 0x22 && Vtxs2(li))	// v220-6
								SetMissinglo(li, num);
						}
					}
					else
					if (!(txn & 0x44)
						&& li->frontsector->ceilingheight != li->backsector->ceilingheight)
					{
						if (Vtxs2(li) && (txn == 0x02 || txn == 0x20))
							settype = R3D_16;
						else
							settype = R3D_1D;
					}
					else
					{
						settype = R3D_03;
					}
				}
				else if (txn & 0x55)
				{
					settype = R3D_00;
				}
				else if (li->frontsector == li->backsector)
					settype = R3D_01;
				else if (li->backsector->ceilingheight == li->frontsector->ceilingheight
					&& li->backsector->floorheight == li->frontsector->floorheight
					&& !(li->frontsector->specFlags & SF_MOVSECTOR)
					&& !(li->backsector->specFlags & SF_MOVSECTOR))
				{
					settype = R3D_01;
				}
				else if (noR3D_16 || (checkTypeR3D_16 && !txn))
				{
					int mf = li->frontsector->specFlags & SF_MOVFLOOR;
					int mb = li->backsector->specFlags  & SF_MOVFLOOR;

					// v220-4 extra test added with mf and mb
					if (   mf^mb
						&& checkTypeR3D_16
						&& 
						(  (mf && li->frontsector->specFlags & SF_MOVLOWER && li->frontsector->floorheight > li->backsector->floorheight)
						|| (mb && li->backsector->specFlags  & SF_MOVLOWER && li->frontsector->floorheight < li->backsector->floorheight) )
						)
					{
						settype = R3D_16;
					}
					else if (mf || mb)
					{
						// v220-10 add tests for misslo tex
						if (mf && !mb && li->mtflags == R3D_MISLTEX1)
							settype = R3D_36;
						else if (!mf && mb && li->mtflags == R3D_MISLTEX2)
							settype = R3D_36;
						else
							settype = R3D_01;
					}
					else
						settype = R3D_00;
				}
			}

			Settype(li, settype, num);
		}

		lnsec[li - lines] = -1;

		if (   vertexowners[li->v1 - vertexes].num <= 2
			&& vertexowners[li->v2 - vertexes].num <= 2)
		{
			SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, num);
		}
		else if (!(li->ml_flags & ML_TWOSIDED)
			|| !((li->frontsector == sec) ^ (li->backsector == sec)) )
		{
			if (!(lxarray[li - lines] & 15))
				SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, num);
		}
	}
}

// b211-0 Added
RDBOOL LS_SetSSINF(sector_t *sec, byte *array, byte *exssecs, byte *bridgelines, RDBOOL exline, RDBOOL invalidtex)
{
	subsector_t *ss;
	sector_t	*fs, *bs;
	seg_t		*sg;
	int			i, k;

	memset(&ssinf, 0, sizeof(ssecinfo_t));

	ssinf.othersec = ssinf.specceil = ssinf.specfloor = -1;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		if (exssecs && exssecs[hitssec[i]]) continue;

		ss = &subsectors[hitssec[i]];

		if (ss->sector->specFlags2 & (SF2_LOSLOPE|SF2_HISLOPE))
			return false;

		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *li = sg->linedef;

			if (li && li->validcount != validcount)
			{
				int lnum = li - lines;
				int txn  = origTEX[lnum];

				li->validcount = validcount;

				if (bridgelines && bridgelines[li - lines]) continue;
				if (exline && excludeline[li - lines]) continue;	// excludeline is global

				fs = li->frontsector;
				bs = li->backsector;

				if (array[lnum] == 4)
				{
					if (txn & 0x44) ssinf.boundhashi++;
					if (txn & 0x22) ssinf.boundhasmid++;
					if (txn & 0x11) ssinf.boundhaslo++;

					if (txn)
						ssinf.boundshastex |= txn;
					else
					{
						int flags = LS_TestLineTex(li);

						if (flags & LT_MISHI && flags & LT_DIFFHI) ssinf.boundsmisshi++;
						if (flags & LT_MISLO && flags & LT_DIFFLO) ssinf.boundsmisslo++;
					}
				}
				else
				{
					ssinf.anylines++;
					if (fs == bs)
					{
						ssinf.anysameseclines++;
						if (fs->specFlags & SF_MOVFLOOR)
							ssinf.anysamemovef = true;
					}
				}

				if (!bs)
				{
					ssinf.lines++;
				}
				else if (fs == bs)
				{
					ssinf.sameseclines++;

					if (fs != sec)
					{
						if (!(li->lnflags2 & LNF2_OWNEQUALTOONE) || !Vtxs2(li))
							ssinf.othersslinechk = true;

						ssinf_otherlineptr[ssinf.othersslinecnt++] = lnum;

						if (txn)
							ssinf.othersslinetex++;
					}
				}
				else
				if (array[lnum] != 4)
				{
					int cdiff = twoSKY(li,1) ? 0 : fs->ceilingheight - bs->ceilingheight;
					int fdiff = twoSKY(li,0) ? 0 : fs->floorheight   - bs->floorheight;

					if (  (fs == sec || bs == sec)
						&& txn & 0x44	// whether invalid or not
						&& fs->ceilingpic == skyflatnum
						&& bs->ceilingpic == skyflatnum)
					{
						ssinf_ownrcptr[ssinf.ownringcnt++] = lnum;
						ssinf.irowntex++;
						continue;
					}

					if (   !fdiff
						&& !cdiff
						&& !txn
						&& Vtxs2(li)
						&& (fs->specFlags & SF_MOVFLOOR || bs->specFlags & SF_MOVFLOOR))
					{
						ssinf.fmovfloor = true;
					}

					if ( (cdiff || fdiff) && (fs == sec || bs == sec))
					{
						if (   fs->floorpic == skyflatnum
							&& bs->floorpic != skyflatnum
							&& fs->floorheight > bs->floorheight)
						{
							ssinf_ownrcptr[ssinf.ownringcnt++] = lnum;
							if (txn) ssinf.irowntex++;
							continue;
						}

						if (   fs->floorpic != skyflatnum
							&& bs->floorpic == skyflatnum
							&& fs->floorheight < bs->floorheight)
						{
							ssinf_ownrcptr[ssinf.ownringcnt++] = lnum;
							if (txn) ssinf.irowntex++;
							continue;
						}
					}

					if (abs(fdiff) > 1024*FRACUNIT) fdiff = 0;
					if (abs(cdiff) > 1024*FRACUNIT) cdiff = 0;

					// Set any inside ring
					if (cdiff && txn & 0x44) ssinf.irhitex++;
					if (fdiff && txn & 0x11) ssinf.irlotex++;

					if ( (cdiff || fdiff) && (fs == sec || bs == sec))
					{
						// ignore if a texture is missing
						if ((cdiff || fdiff) && !infTexMissing(li, cdiff, fdiff))
						{
							ssinf_ownrcptr[ssinf.ownringcnt++] = lnum;
						}
						else
							ssinf.lines++;

					}
					else
					{
						if (txn & 0x22 && fdiff) ssinf.midtex++;

						if (cdiff)
						{
							sector_t *s = cdiff > 0 ? fs : bs;

							if (fs->specFlags & SF_MOVCEIL || bs->specFlags & SF_MOVCEIL)
							{
								if (txn & 0x44)
									ssinf.hitex++;
							}
							else if (sg->sidedef->toptexture && sg->frontsector == s)
								ssinf.hitex++;
							else if (sg->sidedef->texFlags)
								ssinf.misshi++;
						}
						if (fdiff)
						{
							sector_t *s = fdiff > 0 ? bs : fs;

							if (fs->specFlags & SF_MOVFLOOR || bs->specFlags & SF_MOVFLOOR)
							{
								if (txn & 0x11)
									ssinf.lotex++;
							}
							else if (sg->sidedef->bottomtexture && sg->frontsector == s)
							{
							//	if (!(invalidtex && txn == 0x11))
									ssinf.lotex++;
							}
							else if (sg->sidedef->texFlags)
								ssinf.misslo++;
						}

						if (fdiff && ssinf.othersec >= -1)
						{
							int msk, othersec;

							if (fdiff < 0)
							{
								msk = 0x01; othersec = fs - sectors;
							}
							else
							{
								msk = 0x10; othersec = bs - sectors;
							}

							if (origTEX[lnum] & msk)
							{
								if (ssinf.othersec == -1)
									ssinf.othersec = othersec;
								else if (ssinf.othersec != othersec)
									ssinf.othersec = -2;

							}

							if (ssinf.othersec >= 0)
							{
								ssinf_otherlorcptr[ssinf.otherloringcnt++] = li - lines;
							}
						}

						ssinf.lines++;
					}
				}
			}
		}
	}
	return true;
}

// v220-7 Check for two line rings one nested in the other
//        lines must be in linearray. ringcount if set returns the count of all ines found
int dualRingNestedSector(sector_t *sec, int cnt, int *ringcount)
{
	int		i, rc, box[3][4], bn = 0, lnum;
	RDBOOL	repeat;

	if (ringcount) (*ringcount) = 0;

	if (cnt < 3) return false;

	CLR_hitline();

	for (i = 0; i < cnt; i++) hitline[linearray[i]] = 1;

	repeat = true;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < sec->linecount; i++)
		{
			lnum = sec->lines[i] - lines;

			if (   hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, sec, lnum, &rc, box[bn], NULL, NULL))
			{
				if (bn == 2) return false;
				if (ringcount) (*ringcount) += rc;
				repeat = true; bn++;
			}
			else
				hitline[lnum] = 0;
		}
	}

	return bn == 1 || (bn == 2 && (M_FitBox(box[0], box[1]) || M_FitBox(box[1], box[0])));
}

// v220-8 Check for nested line loops and if found return the largest loop,
//
//        The lines to be checked should be in linearray and the
//        number in cnt.
//        cnt must be > 5 for any possibility of at least one loop
//        being nested inside another.
//        box must be dimensioned to 4 ints.
//        Returns number of nested rings found with largest having
//        the lines stored to ring if at least 2 loops were found
//        thus ring must be preset and suitably dimensioned,
//
int largestRingNestedSector(sector_t *sec, int cnt, int *ring, int *count, int *box)
{
	typedef struct {
		int	box[4];
		int line;
	} obox_t;

	obox_t	*obox;
	int		obox_count = 0, obox_maxcount = 8;

	int		*lamem;
	int		c, i;
	RDBOOL	repeat = true;

	//-------------------------------

	*count = 0;

	if (cnt < 6) return 0;

	CLR_hitline();

	for (i = 0; i < cnt; i++) hitline[linearray[i]] = 1;

	lamem = (int*)malloc(cnt * sizeof(int));
	memcpy(lamem, linearray, cnt * sizeof(int));

	obox = (obox_t*)malloc(obox_maxcount * sizeof(obox_t));

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < sec->linecount; i++)
		{
			c = sec->lines[i] - lines;

			if (   hitline[c] == 1
				&& subfn_CheckForRing(hitline, sec, c, NULL, box, NULL, NULL))
			{
				repeat = true;

				if (obox_maxcount >= obox_count)
				{
					obox_maxcount += 8;
					obox = (obox_t*)realloc(obox, obox_maxcount * sizeof(obox_t));
				}

				memcpy(obox[obox_count].box, box, 4 * sizeof(int));
				obox[obox_count].line = c;
				obox_count++;
			}
			else
				hitline[c] = 8;
		}
	}

	if (obox_count > 1)	// get lines in the largest ring
	{	
		int num = 0;

		for (i = 1; i < obox_count; i++)
		{
			if (M_FitBox(obox[num].box, obox[i].box)) num = i;
		}

		// get the true number of boxes that fit this ring
		for (c = i = 0; i < obox_count; i++)
		{
			if (i != num && M_FitBox(obox[i].box, obox[num].box)) c++;
		}

		obox_count = c + 1;	// include self

		if (obox_count > 1)
		{
			CLR_hitline();

			for (i = 0; i < cnt; i++) hitline[lamem[i]] = 1;

			if (subfn_CheckForRing(hitline, sec, obox[num].line, NULL, NULL, NULL, NULL))
			{
				for (c = i = 0; i < sec->linecount; i++)
				{
					line_t *ld = sec->lines[i];

					if (hitline[ld - lines] == 4) ring[c++] = ld - lines;
				}

				*count = c;
				memcpy(box, obox[num].box, 4 * sizeof(int));
			}
		}
	}

	M_Free(obox);
	M_Free(lamem);

	return obox_count;
}

int *getSSECSinDualRingSector(sector_t *sec, int *sscnt, int test, RDBOOL set_outer, RDBOOL return_all)
{
	int c, d, i, k, n, rc, box[2][4];
	int *sslist = NULL;
	int cln = sec->linecount;
	RDBOOL cmp;

	*sscnt = 0;
	
	if (test == 2) goto test_direct;

	if (test == 0)
	{
		for (c = i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			if (ld->spflags & SPF_CONTROLLINE) { cln--; continue; }

			if (   !(ld->ml_flags & ML_TWOSIDED)
				|| !Vtxs2(ld)
				|| ld->specType >= R3D_40
				|| secEquiv(sec, ld, NULL)
				|| ld->frontsector->specFlags & SF_MOVSECTOR
				|| ld->backsector->specFlags & SF_MOVSECTOR
				|| subfn_FindLineCrossing(ld, LC_NORMAL))
			{
				return NULL;
			}

			c++;
		}

		if (c != cln) return NULL;
	}
	
	CLR_hitline();

	for (c = i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (test == 1)
		{
			if (ld->spflags & SPF_CONTROLLINE) continue;

			if (ld->frontsector != ld->backsector)
			{
				hitline[ld - lines] = 1; c++;
			}
		}
		else
			hitline[ld - lines] = 1;
	}

	if (test == 1 && c < 3) return NULL;

test_direct:;

	for (c = d = i = 0; i < sec->linecount; i++)
	{
		n = sec->lines[i] - lines;

		if (   hitline[n] == 1
			&& subfn_CheckForRing(hitline, sec, n, &rc, box[c], NULL, NULL))
		{
			int a = 0;

			d += rc;	// keep a running total of the line ring count

			if (!c)		// Differentiate lines in first ring (ring 0)
			{
				for (k = 0; k < sec->linecount; k++)
				{
					n = sec->lines[k] - lines;

					if (hitline[n] == 4) hitline[n] = 16, a++;
				}
			}

			if (++c > 1) break;

			if (d == cln) break;
		}
	}

	if (test == 2 && c != 2) return NULL;

	if (test == 1 && c == 1)	// single ring case
	{
		// set linearray
		for (c = i = 0; i < sec->linecount; i++)
		{
			n = sec->lines[i] - lines;

			if (hitline[n] == 16) linearray[c++] = n;
		}

		if (LS_MarkSubsecsInLoop(sec, c, NULL, box[0], MSIA_NONE, NULL))
		{
			sslist = (int*)M_Malloc(hitssecptr * sizeof(int), 0);

			memcpy(sslist, hitssec, hitssecptr * sizeof(int));
			*sscnt = hitssecptr;
			return sslist;
		}

		return NULL;
	}

	if (c != 2) return NULL;

	if (   (test == 1 || d == cln)					// test OR all lines accounted for
		&& ( (cmp = M_CompareBox(box[0], box[1]))	// and one box fits in the other
		||  M_CompareBox(box[1], box[0])) )
	{
		int *r1ssec = NULL, r1ssptr, r2ssptr;
		int flag0, flag1;

		//---------------------------

		// Establish whether ring 1 is facing in or out
		// ring 0, tested further, on must face the other way
		if (cmp)
			flag0 = MSIA_REVERSED, flag1 = MSIA_NONE;
		else
			flag0 = MSIA_NONE, flag1 = MSIA_REVERSED;

		// set linearray
		for (c = i = 0; i < sec->linecount; i++)
		{
			n = sec->lines[i] - lines;

			if (hitline[n] == 4) linearray[c++] = n;
		}

		// Get all the subsectors inside ring 1
		if (   LS_MarkSubsecsInLoop(sec, c, NULL, box[1], flag1, NULL)
			&& (r1ssptr = hitssecptr) )
		{
			r1ssec = (int*)M_Malloc(r1ssptr * sizeof(int), 0);
			memcpy(r1ssec, hitssec, r1ssptr * sizeof(int));
		}
		else
			return NULL;

		// Set the linearray values for ring 0
		for (c = i = 0; i < sec->linecount; i++)
		{
			n = sec->lines[i] - lines;

			if (hitline[n] == 16) linearray[c++] = n;
		}

		// Get all the subsectors inside ring 0
		if (!LS_MarkSubsecsInLoop(sec, c, NULL, box[0], flag0, NULL))
		{
			M_Free(r1ssec); return NULL;
		}

		// Create a list of all the subsectors between the rings
		if ( (r2ssptr = hitssecptr) )
		{
			int *p1 = r1ssec, *p2 = hitssec;

			// Put the larger area's ssec count in r2ssptr
			if (r2ssptr < r1ssptr)
			{
				r2ssptr = r1ssptr, r1ssptr = hitssecptr;
				p1 = hitssec, p2 = r1ssec;
			}

			sslist = (int*)M_Malloc(r2ssptr * sizeof(int), 0);

			if (return_all)
			{
				*sscnt = r2ssptr;
				memcpy(sslist, p2, r2ssptr * sizeof(int));
			}
			else
			{
				for (i = 0; i < r2ssptr; i++)
				{
					RDBOOL	valid = true;
					int		sn = p2[i];

					for (k = 0; k < r1ssptr; k++)
					{
						if (p1[k] == sn) { valid = false; break; }
					}

					if (valid) sslist[(*sscnt)++] = sn;
				}
			}
		}

		if (*sscnt == 0 && sslist)
			M_Free(sslist); sslist = NULL;

		M_Free(r1ssec);
	}

	return sslist;
}

// the osx structure values must be set together with linearray
RDBOOL U_checkDoubleLoop(sector_t *sc, int type)
{
	subsector_t	*ss;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			d, i, k, m, p, rc, box[2][4];
	int			szBi = sizeof(osx.innerbox);
	int			szBo = sizeof(osx.outerbox);

	CLR_hitline();

	for (d = i = 0; i < osx.lcount; i++) hitline[linearray[i]] = 1;

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < osx.lcount; i++)
		{
			if (hitline[linearray[i]] == 1) { lnum = linearray[i]; break; }
		}

		if (lnum < 0) break;

		if (d > 1) return false;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box[d], NULL, NULL))
		{
			for (i = 0; i < osx.lcount; i++)
			{
				if (hitline[linearray[i]] == 4) hitline[linearray[i]] = (d ? 16 : 8);
			}
			++d;
		}
		else
		{
			return false;
		}
	}

	if (d != 2) return false;

	if      (M_CompareBox(box[0], box[1])) { p =  8; memcpy(osx.innerbox, box[0], szBi); memcpy(osx.outerbox, box[1], szBo); }
	else if (M_CompareBox(box[1], box[0])) { p = 16; memcpy(osx.innerbox, box[1], szBi); memcpy(osx.outerbox, box[0], szBo); }
	else return false;

	ss_validcount++;
	validcount++;
	osx.inner = NULL;

	// look inside the smaller box
	if (type == 1)
	{
		for (i = 0; i < osx.lcount; i++)
		{
			if (hitline[linearray[i]] == p)
			{
				sl = &sslines[linearray[i]];

				for (k = 0; k < sl->num; k++)
				{
					ss = &subsectors[sl->list[k]];

					if (ss->ssvalidcount == ss_validcount) continue;

					ss->ssvalidcount = ss_validcount;

					if (ss->sector != sc)
					{
						for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
						{
							if (   (ld = sg->linedef)
								&& !hitline[ld - lines]
								&& ld->validcount != validcount)
							{
								if (ld->frontsector != ld->backsector)
								{
									return false;
								}

								ld->validcount = validcount;
								if (!osx.inner) osx.inner = ld->frontsector;
								else if (osx.inner != ld->frontsector) return false;
							}
						}
					}
				}
			}
		}
	}
	else
	{
		CLR_conarray();

		for (i = 0; i < osx.lcount; i++)
		{
			if (hitline[linearray[i]] == p)
			{
				conarray[linearray[i]] = 4;
			}
		}

		if (!LS_MarkSubsecsInAreaGeneral(sc, conarray, 0, NULL, osx.innerbox, -1, MSIA_REVERSED))
		{
			return false;
		}

		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];

			for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
			{
				if ( (ld = sg->linedef) && ld->validcount != validcount)
				{
					ld->validcount = validcount;
					if (hitline[ld - lines]) continue;
					if (ld->frontsector != ld->backsector) return false;
					if (!osx.inner) osx.inner = ld->frontsector;
					else if (osx.inner != ld->frontsector) return false;
				}
			}
		}
	}

	if (!type && !osx.inner) return false;

	p = p == 8 ? 16 : 8;

	// get (type 1) all subsectors inside the larger box
	if (type == 1)
	{
		sector_t	*ch;
		int			res = false, chksec[1];
		
		chksec[0] = sc - sectors;

		// Since LS_MarkSubsecsInMultiSecLoop uses linearray we need to
		// copy it first so that it can be restored falling calling it.
		osx.intList = (int*)malloc(sizeof(int) * osx.lcount);
		memcpy(osx.intList, linearray, sizeof(int) * osx.lcount);

		// look outside the larger box
		for (osx.listcnt = i = 0, osx.outer = NULL; i < osx.lcount; i++)
		{
			if (hitline[osx.intList[i]] == p)
			{
				ld = &lines[osx.intList[i]];
				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
				if (!osx.outer) osx.outer = ch;
				else if (osx.outer != ch) goto restore_and_exit;
				linearray[osx.listcnt++] = osx.intList[i];
			}
		}
		
		res = LS_MarkSubsecsInMultiSecLoop(1, chksec, osx.listcnt, osx.outerbox, NULL, false, MSIA_NONE);

		if (res)
		{
			for (i = 0; i < osx.lcount; i++)
			{
				if (hitline[osx.intList[i]]) hitline[osx.intList[i]] = 4;
			}

			res = LS_OnlySSLinesInSSECarray(sc, hitline);
		}

restore_and_exit:;
		memcpy(linearray, osx.intList, sizeof(int) * osx.lcount);
		free(osx.intList); osx.intList = NULL;
		return res;
	}

	ss_validcount++;

	// look outside the larger box
	for (i = 0, osx.outer = NULL; i < osx.lcount; i++)
	{
		if (hitline[linearray[i]] == p)
		{
			sl = &sslines[linearray[i]];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;
				
				ss->ssvalidcount = ss_validcount;

				if (ss->sector != sc)
				{
					ss->ssvalidcount = ss_validcount;
					if (!osx.outer) osx.outer = ss->sector;
					else if (osx.outer != ss->sector) return false;
				}
			}
		}
	}

	return osx.outer != NULL;
}
