//-----------------------------------------------------------------------------
//
// Copyright (C) 2008 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r_precache.c
//
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
//  Texture handling simplified
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION:	texture setup and precache functions
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "de_base.h"
#include "de_console.h"
#include "de_network.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_misc.h"
#include "de_render.h"
#include "Doom\p_local.h"

#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// Model cacheing array

typedef struct {
	mobjinfo_t	*minf;
	modeldef_t	*mfw;
	modeldef_t	*cfm;
	RDBOOL		checked;
} moinf_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	M_CacheText_hu_font_b(char* string);
void	GL_ReloadTextureUnclamped(int idx);
int		Def_GetMobjNum(char *id);
void	R_PreCacheModelSkins(void);
int		R_PrecacheSkinsForMoType(int type);
int		R_CheckModelSkins(modeldef_t *modef);
int		P_GetSwitchPair(int txn);
int		R_CacheLevelTitle(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int totalskins;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------


//===========================================================================
// subfn_PrecacheSprites	b211-0 cacheing extended
//	Prepares the specified texture and all the other textures in the
//	same animation group.
//===========================================================================

static int subfn_CacheRelatedModels(moinf_t *moinf, int moinf_num)
{
	ded_mobj_t	*dmo;
	int			stnum;
	int			c, i, mnum;

	for (c = i = 0, dmo = defs.mobjs; i < stats.num_mobjs; i++, dmo++)
	{
		if (   dmo->cache
			&& (mnum = Def_GetMobjNum(dmo->id)) >= 0)
		{
			mobjinfo_t	*inf = mobjinfo + mnum;

			if (   (stnum = inf->spawnstate) > 0
				&& stnum < moinf_num
				&& !moinf[stnum].minf
				&& states[stnum].model)
			{
				moinf[stnum].minf = inf;	// v220-4 use moinf_num
				c++;
			}
		}
	}
	return c;
}

static void skinError(int dmednum, int missing, RDBOOL stnum)
{
	Con_Error("%s NUMBER [%i] model has %i skin%s missing.\n"
				"Check models have been properly installed",
		stnum ? "STATE" : "DOOMED", dmednum, missing, missing > 1 ? "s" : "");
}

RDBOOL R_PrecacheStateModel(int statenum, int dmednum)
{
	modeldef_t	*mf;
	int			c;

	if (   !statenum
		|| !(mf = (modeldef_t*) states[statenum].model)
		|| !R_ModelHasSkins(mf))
	{
		return false;
	}
	
	if ( (c = R_CheckModelSkins(mf)) )
		skinError(dmednum, c, false);

	return true;
}

static void subfn_PrecacheSprites()	// $precache
{
	static int firsttime = true;

	thinker_t	*th;
	modeldef_t	*mf;
	byte		*hasspr = (byte*)W_Calloc(numsprites, STATICID, 113);
	int			moinf_num = MAX_OF(count_states.num, NUMSTATES);
	moinf_t		*moinf  = (moinf_t*)W_Calloc(moinf_num * sizeof(moinf_t), STATICID, 113);
	int			stnum;

	int			i, k;
	int			sprite_count = 0;
	int			model_count = 0;

	int			*motype = NULL;
	int			motypecount = 0;
	int			motypemaxcount = 0;

	mobj_t		**moremove = NULL;
	int			moremovecount = 0;
	int			moremovemaxcount = 0;

	char		buf[128];

	//-----------------------------------------------

	// Identify and validate type
	for(th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*mo    = (mobj_t*) th;
			int		type   = mo->type;
			int		sprnum;

			//---------------------------------------------

			if (type < 0 || type >= stats.num_mobjs)
				Con_Error("Thing type %i not recognised", type);

			if ( (sprnum = mo->sprite) < 0 || sprnum >= numsprites)
				Con_Error("ERROR: invalid sprite with thing mapnum %i", MAPNUM(mo));

			if (hasspr[sprnum] & 3) continue;

			if (mo->flags2 & MF2_DONTDRAW) continue;

			if (sprnum == SPR_TNT1)
			{
				sprites[sprnum].sprflags |= SPR_NOMODEL;
				hasspr[sprnum] |= 1;
				continue;
			}

			if (!sprites[sprnum].spriteframes)
			{
				pr_OFinsert("PWAD ERROR: T%i is invalid (its sprite has no frames) and will be ignored\n", MAPNUM(mo));

				if (moremovecount >= moremovemaxcount)
				{
					moremovemaxcount += 16;
					moremove = (mobj_t**)M_Realloc(moremove, sizeof(mobj_t*) * moremovemaxcount, 43);	;
				}
				mo->type = MT_UNKNOWNTHING;	// Don't want this respawning
				moremove[moremovecount++] = mo;
				continue;
			}

			if (   mo->ddflags & DDMF_ALWAYSLIT
				|| mo->frame & FF_FULLBRIGHT)
			{
				if (!(hasspr[sprnum] & 2))
				{
					hasspr[sprnum] |= 2;	// Sprite will need cacheing
					sprite_count++;
				}
			}

			if (sprites[sprnum].sprflags & SPR_NOMODEL)
			{
				if (!(hasspr[sprnum] & 2))
				{
					hasspr[sprnum] |= 2;	// Sprite will need cacheing
					sprite_count++;
				}
			}
			else if (CFX
				&& (stnum = mo->info->spawnstate)
				&& !moinf[stnum].checked)
			{
				moinf[stnum].checked = true;

				if (  !ge_inf.spritemodels
					&& hasDEH
					&& stnum < DH_CHECKNUM
					&& DEHCheck[stnum] & DEHF_STATE)
				{
					sprites[sprnum].sprflags |= SPR_NOMODEL;
					hasspr[sprnum] |= 2;	// Sprite will need cacheing
					sprite_count++;
				}
				else if (states[stnum].model
					&& (mf = (modeldef_t*) states[stnum].model)
					&& R_ModelHasSkins(mf))
				{
					model_t *mdl;

					if (mo->type >= 0)
					{
						if (motypecount >= motypemaxcount)
						{
							motypemaxcount += 64;
							motype = (int*)M_Realloc(motype, motypemaxcount * sizeof(int), 44);
						}

						motype[motypecount++] = mo->type;
					}
					else
					{
						moinf[stnum].minf = mo->info;
					}

					model_count++;

					if (mo->corelink && mo->corelink->state->model)
					{
						stnum = mo->corelink->state - states;
						moinf[stnum].cfm = (modeldef_t*) states[stnum].model;
						model_count++;
					}

					// v220-9 precache lod skins directly.
					if (   mo->lodstate > 0
						&& (mf = GetStateModel(states + mo->lodstate, mo->selector))
						&& mf->sub[0].model
						&& (mdl = modellist[mf->sub[0].model]) )
					{
						int num = mdl->info.numSkins;

						for (i = 0; i < num; i++)
							GL_PrepareSkin(mdl, i, NULL);
					}
				}
			}
		}
	}

	if (moremove)
	{
		for (i = 0; i < moremovecount; i++) P_RemoveMobj(moremove[i], 5);
		M_Free(moremove);
	}

	// v220-33
	if (CFX && firsttime)
	{
		extern int	reduceTo;

		if (ArgCheckWith("-dsst", 1))
		{
			int val = atoi(ArgNext());

			if (val >= 1 && val <= 4)
			{
				reduceTo = 0x1<<( (5 - val) + 6);
				skin_inf.downscaleskintex = true;
				Con_Printf("DSST set: Skins with both x,y dimensions >= %i will be reduced to half size\n", reduceTo);
			}
		}
	}

	if (CFX && !R3D_edit)
	{
		moinf_t	*mi;
		int		n;

		// Cache the main weapons
		for (i = S_PUNCH, n = 0, mi = moinf; i <= S_EXPLODE3; i++)
		{
			if (hasDEH && DEHCheck[i] & DEHF_STATE && states[i].sprite)
			{
				sprites[states[i].sprite].sprflags |= SPR_NOMODEL;
			}
			else if (states[i].model
				&& R_ModelHasSkins((modeldef_t*) states[i].model))
			{
				mi[n++].mfw = (modeldef_t*) states[i].model;
				model_count++;
			}
		}

		// Add related or any unspawned scripted models
		model_count += subfn_CacheRelatedModels(moinf, moinf_num);
	}

	timecache[ITC_CACHE2] = 0;

	// v220-35 if not nomonsters just cache all defined model skins
	if (!(nomonsters && ArgExists("-z")) && model_count)
	{
		sprintf(buf, "Precaching %i model skins...", totalskins);
		Con_InitProgress(buf, totalskins);
		Con_TimingStart();

		R_PreCacheModelSkins();

		Con_HideProgress();
		timecache[ITC_CACHE2] = Con_TimingEnd(NULL);
	}

	// Check validity of skins
	if (model_count)
	{
		mobjinfo_t	*inf; 
		moinf_t		*mi;
		int			c, num;
		int			st_n = stats.num_mobjs;

	//	if (nomonsters) test removed v220-35
		{
			sprintf(buf, "Precaching skins for %i models...", model_count);

			Con_InitProgress(buf, model_count);
			Con_TimingStart();
		}

		for (i = 0, num = 0; i < motypecount; i++)
		{
			R_PrecacheSkinsForMoType(motype[i]);
		/*	if (nomonsters)*/ Con_Progress(++num, PBARF_PERCENT);
		}

		M_Free(motype);

		for (i = 0, mi = moinf; i < st_n; i++, mi++)
		{
			if (mi->mfw)
			{
				if ( (c = R_CheckModelSkins(mi->mfw)) ) skinError(i, c, true);
			/*	if (nomonsters)*/ Con_Progress(++num, PBARF_PERCENT);
			}
		}

		for (i = 0, mi = moinf; i < moinf_num; i++, mi++)
		{
			if (mi->cfm)
			{
				if ( (c = R_CheckModelSkins(mi->cfm)) )
					skinError(i, c, true);
			}

			if ( (inf = mi->minf) )
			{
				R_PrecacheStateModel(inf->spawnstate  , inf->doomednum);
				R_PrecacheStateModel(inf->seestate    , inf->doomednum);
				R_PrecacheStateModel(inf->painstate   , inf->doomednum);
				R_PrecacheStateModel(inf->meleestate  , inf->doomednum);
				R_PrecacheStateModel(inf->missilestate, inf->doomednum);
				R_PrecacheStateModel(inf->deathstate  , inf->doomednum);
				R_PrecacheStateModel(inf->xdeathstate , inf->doomednum);
				R_PrecacheStateModel(inf->raisestate  , inf->doomednum);

			/*	if (nomonsters)*/ Con_Progress(++num, PBARF_PERCENT);
			}
		}

	//	if (nomonsters)
		{
			Con_HideProgress();
			timecache[ITC_CACHE2] = Con_TimingEnd(NULL);
		}
	}

	if (sprite_count)
	{
		int sprnum;

		if (!model_count)
		{
			sprintf(buf, "Precaching skins for %i sprites...", sprite_count);

			Con_InitProgress(buf, sprite_count);
			Con_TimingStart();
		}

		sprite_count = 0;

		for (th = thinkercap.next; th != &thinkercap; th = th->next)
		{
			if (th->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				mobj_t	*mo = (mobj_t*) th;

				if ((sprnum = mo->sprite) < 0)
					Con_Error("subfn_PrecacheSprites: sprnum < 0 with spawnstate %i", mo->info->spawnstate);

				if (hasspr[sprnum] & 2)
				{
					if (!model_count) Con_Progress(++sprite_count, PBARF_PERCENT); // percent display

					for (i = 0; i < sprites[sprnum].numframes; i++)
					{
						spriteframe_t *sf = &sprites[sprnum].spriteframes[i];

						for (k = 0; k < 8; k++)
						{
							if (!spritelumps[sf->lump[k]].tex) GL_PrepareSprite(sf->lump[k], NULL, 6);
						}
					}

					hasspr[sprnum] = 0;
				}
			}
		}

		if (!model_count) Con_HideProgress();

		timecache[ITC_CACHE2] += Con_TimingEnd(NULL);
	}

	Con_Printf("Precached  : %i graphic textures\n", skin_inf.graphicstexcount);
	Con_Printf("Graphics   : texture memory used %.2fMB\n", float(skin_inf.graphicstexmemoryused)/1000000.f);

	if (CFX)
	{
		Con_Printf("Precached  : %i model skins : skincount %i\n", skin_inf.skintexcount, totalskins);
		Con_Printf("Model skins: texture memory used %.2fMB\n", float(skin_inf.skintexmemoryused)/1000000.f);

		if (firsttime && skin_inf.downscaledskintexcount)
			Con_Printf("Downscaled %i skin textures\n", skin_inf.downscaledskintexcount);
	}

	firsttime = false;

	W_Free(moinf , STATICID, 511);
	W_Free(hasspr, STATICID, 511);
}

// ============= end sprite cacheing

//===========================================================================
// R_SetupSky - called once at map start
//===========================================================================
void R_SetupSky(void)
{
	sector_t	*s;
	int			i, k;

	if(!mapinfo) Con_ErrorD("R_SetupSky: mapinfo unset");

	// Check for sky from sidedef inheritance
	for (i = 0, s = sectors; i < numsectors; i++, s++)
	{
		if (  (s->ceilingpic == skyflatnum
			|| s->floorpic == skyflatnum)
			&& !(s->sidesky & PL_SKYFLAT)
			&& (sectors[i].specFlags & 0xC) == 0xC)
		{
			sector_t *os = NULL;

			for (k = 0; k < s->linecount; k++)
			{
				line_t	 *l = s->lines[k];
				sector_t *c = l->frontsector == s ? l->backsector : l->frontsector;

				if (!os) os = c;
				if (!os || c != os) { os = NULL; break; }
			}

			if (   os
				&& (os->ceilingpic == skyflatnum
				||  os->floorpic == skyflatnum)
				&& os->skyvar >= 0)
			{
				s->sidesky = os->sidesky;
				s->skyvar  = os->skyvar;

				if (os->specFlags2 & SF2_HASTYPE272)
					s->specFlags2 |= SF2_HASTYPE272;

				if (os->specFlags2 & SF2_SKYSCROLLY)
					s->specFlags2 |= SF2_SKYSCROLLY;
			}
		}
	}

	Init_SkyParams();
}


// Texture precache sub routines v220-10 revised ===============

static void markGroupTex(int txn, byte *texmark)
{
	int i, n;

	if ( (n = texArray[txn]->anim_group) )
	{
		for (i = 0; i < numtextures; i++)
		{
			if (texArray[i]->anim_group == n)
				texmark[i] = true;
		}
	}
	else
	{
		texmark[txn] = true;
	}
}

static short checkTEX(int txn, byte *texmark)
{
	int n;

	if (txn <= 0 || txn >= numtextures) return 0;

	markGroupTex(txn, texmark);

	if (   texArray[txn]->txf_flags & TXF_SWITCH
		&& (n = P_GetSwitchPair(txn)) )
	{
		markGroupTex(n, texmark);
	}

	return (short)txn;
}

// Used only if ge_inf.infflags & 1
extern int	gamenum;
static char	**farray;
static char	**tarray;
static byte	mapDone[100];
static int	wadTextCount;
static int	wadFlatCount;
static int __cdecl nameSorter(const void *e1, const void *e2)
{
	return _stricmp(*(char**)e1, *(char**)e2);
}

static RDBOOL initTexReportArrays()
{
	if (ge_inf.infflags & 1)
	{
		if (!farray)
		{
			farray = (char**)W_Calloc(numlumps*sizeof(char*), STATICID, 0);
			tarray = (char**)W_Calloc(numlumps*sizeof(char*), STATICID, 0);
		}

		if (mapDone[gamenum]) return 0;
		mapDone[gamenum] = 1;
		return 1;
	}
	return 0;
}

void printTexReport()
{
	if (!farray) return;

	int i, k, m = 18;

	for (i = 0; i < 100; i++) { if (mapDone[i]) goto tex_report; }

	return;

tex_report:
	// list maps
	pr_OFinsert("Texture report for ");
	for (i = 0; i < 100; i++) if (mapDone[i]) {  pr_OFinsert("MAP%.2i ", i); m += 6; }
	pr_OFrepeat("\n", "=", m, 1);

	if (wadFlatCount > 1) qsort(farray, wadFlatCount, sizeof(char*), nameSorter);
	if (wadTextCount > 1) qsort(tarray, wadTextCount, sizeof(char*), nameSorter);

	// List all flats
	pr_OFinsert("PWAD sector flats for maps (total %i) ", wadFlatCount);
	pr_OFrepeat(NULL, "=", 7, 1);
	for (i = 0; i < wadFlatCount; i++) pr_OFinsert("%.8s\n", farray[i]);

	// List all textures
	pr_OFinsert("PWAD side textures for maps (total %i) ", wadTextCount);
	pr_OFrepeat(NULL, "=", 7, 1);
	for (i = 0; i < wadTextCount; i++)
	{
		pr_OFinsert("%.8s\n", tarray[i]);

		{
			int idx = R_CheckTextureNumForName(tarray[i]);

			if (idx > 0)
			{
				texture_t	*tex = texArray[idx];

				if (tex->patchcount > 0)
				{
					if (   tex->patchcount == 1
						&& !strnicmp(tarray[i], lumpinfo[tex->patches[0].patch].name, 8)
						&& !tex->patches[0].originx
						&& !tex->patches[0].originy)
					{
						continue;
					}

					for (k = 0; k < tex->patchcount; k++)
					{
						pr_OFinsert("\t%8s x %3i y %3i\n",
							lumpinfo[tex->patches[k].patch].name,
							tex->patches[k].originx, tex->patches[k].originy);
					}
				}
			}
		}
	}

	for (i = m = 0; i < wadFlatCount; i++)
	{
		for (k = 0; k < wadTextCount; k++)
		{
			if (!_stricmp(farray[i], tarray[k])) m++;
		}
	}

	if (!m) return;

	// List all textures used for flats
	pr_OFinsert("PWAD flats from textures (total %i) ", m);
	pr_OFrepeat(NULL, "=", 7, 1);

	for (i = 0; i < wadFlatCount; i++)
	{
		for (k = 0; k < wadTextCount; k++)
		{
			if (!_stricmp(farray[i], tarray[k]))
					pr_OFinsert("%.8s\n", farray[i]);
		}
	}
	pr_OFrepeat(NULL, "=", 54, 1);
}

// end tex report

//===========================================================================
// R_PrecacheFlat
//	Prepares the specified flat and all the other flats in the same
//	animation group. Has the consequence that all lumps inside the 
//	F_START...F_END block obtain a flat_t record.
//===========================================================================
static void R_PrecacheFlat(int num, RDBOOL report)
{
	flat_t	*f2, *f = R_GetFlat(num);
	int		i = numlumps, n = f->group;
	int		k, w, h;
	int		chk, type;

	GL_PrepareFlat(num);

	if (report)
	{
		char	*p = lumpinfo[num].name;

		if (   (chk = W_CheckPWADNumForName(p)) >= 0
			&& ((type = DoomLumpIdentify(chk, &w, &h, DLI_ANY)) == LMP_FLAT
			|| type == LMP_PATCH || type == LMP_PNG || type == LMP_JPG))
		{
			if (w != 64 || h != 64)
			{
				for (k = 0; k < wadTextCount; k++)
				{
					if (!_stricmp(p, tarray[k])) goto noflat;
				}
				tarray[wadTextCount++] = p;
			}
			else
			{
				for (k = 0; k < wadFlatCount; k++)
				{
					if (!_stricmp(p, farray[k])) goto noflat;
				}
				farray[wadFlatCount++] = p;
			}
		}
	}

noflat:;

	if (!n) return;

	// Flats should be enclosed inside F_START, F_END markers, per
	// the specs. The WAD loader pre-merges all F_START, F_END blocks.
	while(--i >= 0)
	{
		if (   lumpinfo[i].groupMarker == LGT_FLATS
			&& (f2 = R_GetFlat(i))
			&& n == f2->group)
		{
			if (report)
			{
				char	*p = lumpinfo[i].name;

				if (   (chk = W_CheckPWADNumForName(p)) >= 0
					&& ((type = DoomLumpIdentify(chk, &w, &h, DLI_ANY)) == LMP_FLAT
					|| type == LMP_PATCH || type == LMP_PNG || type == LMP_JPG))
				{
					if (w != 64 || h != 64)
					{
						for (k = 0; k < wadTextCount; k++)
						{
							if (!_stricmp(p, tarray[k])) goto nextflat;
						}
						tarray[wadTextCount++] = p;
					}
					else
					{
						for (k = 0; k < wadFlatCount; k++)
						{
							if (!_stricmp(p, farray[k])) goto nextflat;
						}
						farray[wadFlatCount++] = p;
					}
				}
			}
nextflat:;
			if (!f2->gltex) GL_PrepareFlat(i);
		}
	}
}

static void subfn_PrecacheGraphics()
{
	byte	*texmark = (byte*)M_Calloc(numtextures, 314);
	int		counttex;
	int		i, report = initTexReportArrays();

	//-------------------------

	R_CacheLevelTitle();	// Patch may need cacheing
	Con_InitProgress("Precaching flats...", numsectors);
	Con_TimingStart();

	// Precache flats.
	for (i = 0; i < numsectors; i++)
	{
		R_PrecacheFlat(sectors[i].floorpic, report);
		R_PrecacheFlat(sectors[i].ceilingpic, report);
		Con_Progress(i, PBARF_PERCENT);
	}

	Con_HideProgress();

	// Precache the map name
	M_CacheText_hu_font_b(mapinfo->mapname);

	for (i = 0; i < numsides; i++)
	{
		sides[i].bottomtexture = checkTEX(sides[i].bottomtexture, texmark);
		sides[i].midtexture    = checkTEX(sides[i].midtexture   , texmark);
		sides[i].toptexture    = checkTEX(sides[i].toptexture   , texmark);
	}

	if (report)
	{
		int		k;

		for (i = 0; i < numtextures; i++)
		{
			if (texmark[i])
			{
				texture_t	*tex = texArray[i];

				if (tex->patchcount > 0)
				{
					char	*p;

					if (tex->oldname[0])
						p = (char*)tex->oldname;
					else
						p = (char*)tex->name;

					if (W_CheckPWADNumForName(p) >= 0)
					{
						for (k = 0; k < wadTextCount; k++)
						{
							if (!_stricmp(p, tarray[k])) goto nexttex;
						}

						tarray[wadTextCount++] = p;
					}
					else
					{
						for (k = 0; k < tex->patchcount; k++)
						{
							char	*np = lumpinfo[tex->patches[k].patch].name;

							if (strnicmp(p, np, 8) && W_CheckPWADNumForName(np) >= 0)
							{
								for (k = 0; k < wadTextCount; k++)
								{
									if (!_stricmp(p, tarray[k])) goto nexttex;
								}

								tarray[wadTextCount++] = p;
								goto nexttex;
							}
						}
					}
				}
			}
nexttex:;
		}
	}

	for (i = counttex = 0; i < numtextures; i++)
	{
		if (texmark[i])
		{
			if (texArray[i]->gltex && texArray[i]->txf_flags & TXF_CLAMPS) GL_DeleteTexIndex(&texArray[i]->gltex, "texclamps");
			if (!texArray[i]->gltex) counttex++;
		}
	}

	Con_InitProgress("Precaching textures...", counttex);

	// Precache textures
	for (i = counttex = 0; i < numtextures; i++)
	{
		if (texmark[i] && !texArray[i]->gltex)
		{
			int trn = texturetranslation[i];

			Con_Progress(++counttex, PBARF_PERCENT);

			// Prepare this texture - not the translated one
			texturetranslation[i] = i;

			texArray[i]->gltex = GL_PrepareTexture(i);

			// Restore the translated value
			texturetranslation[i] = trn;
		}
	}

	// v220-31 unclamp masked ladder line mid textures
	if (R3D_maptype)
	{
		for (i = 0; i < numlines; i++)
		{
			if (lines[i].ovflags & OVF_3DLADDERLINE)
			{
				// Ladderlines have the same midtex both sides
				short tex = sides[lines[i].sidenum[0]].midtexture;

				// v220-31 check ladderline texture, if masked, is not clamped
				if (tex && (texArray[tex]->txf_flags & (TXF_MASKED|TXF_NOCLAMP)) == TXF_MASKED)
				{
					GL_ReloadTextureUnclamped(tex);
				}
			}
		}
	}

	Con_HideProgress();
	M_Free(texmark);
	timecache[ITC_CACHE1] = Con_TimingEnd(NULL);
}

//===========================================================================
// R_PrecacheLevel
//===========================================================================
void R_PrecacheLevel(int flags)
{
	//	Prepare textures, flats
	if (flags & CACHE_GRAPHICS)
		subfn_PrecacheGraphics();

	//	Prepare models, sprites.
	if (flags & CACHE_SPRITES && !ge_inf.noobjects)
		subfn_PrecacheSprites();
}

static void checkAlpha(int tx)
{
	int i, n = texArray[tx]->anim_group;

	if (n)
	{
		for (i = 0; i < numtextures; i++)
		{
			if (texArray[i]->anim_group == n)
			{
				int idx = texturetranslation[i];

				texturetranslation[idx] = i;
				GL_PrepareTexture(i);

				if (rendtexmask)
					GL_PrepareTextureNoAlpha(i);

				texturetranslation[i] = idx;
			}
		}
	}
	else
	{
		GL_PrepareTexture(tx);
		if (rendtexmask) GL_PrepareTextureNoAlpha(tx);
	}
}

// v220-6 check if any non-alpha variants of alpha textures will be needed
void R_PrecacheNoAlphaTex()
{
	seg_t *sg;
	int   i, tx;

	for (i = 0, sg = segs; i < numsegs; i++, sg++)
	{
		if (sg->linedef && sg->sidedef)
		{
			int midtex = sg->sidedef->midtexture;

			if ( !sg->backsector )
			{
				if (midtex) checkAlpha(midtex);

				continue;
			}

			if ( (tx = sg->sidedef->bottomtexture) )
			{
				if (sg->backsector->heightsec < 0)
					checkAlpha(tx);
			}
			
			if ( (tx = sg->sidedef->toptexture) )
			{
				if (sg->linedef->lnflags2 & LNF2_NOALPHA)
				{
					checkAlpha(tx);
				}
				else if (!(sg->backsector->specFlags & SF_NOOCCLUDE)
					&& !(!midtex && sg->backsector->specFlags & SF_MASKEDDOOR)
					&& sg->backsector->heightsec < 0)
				{
					checkAlpha(tx);
					// speed test on future pass
					sg->linedef->lnflags2 |= LNF2_NOALPHA;
				}
			}
		}
	}
}

// v220-31
void SCP_precacheModels()
{
	extern int	skinCount;
	extern int	*preSpawnCache;
	extern int	psc_num;

	int			i, dnum, type;

	// ------------------------

	skinCount = 0;

	if (!psc_num) return;

	pr_OFdirect("Precache check for as yet unspawned scripted Things =========\n");

	for (i = 0; i < psc_num; i++)
	{
		dnum = preSpawnCache[i];
		type = P_GetTypeFromDoomednum(dnum);

		if (type < stats.num_mobjs)	// b211-0
		{
			int n = skinCount;
			int found = R_PrecacheSkinsForMoType(type);

			if (!found)
				pr_OFinsert("Model not found with doomednum %i\n", dnum);

			if (n < skinCount) pr_OFinsert("  Cached doomednum %i\n", dnum);
		}
	}

	pr_OFdirect("End precache check ===========================================\n");
}