//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_sprite.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
//  Clamping method revised for masked walls
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b211-1)
//  old sort code removed
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Shadow rendering
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_refresh.h"
#include "de_render.h"
#include "de_play.h"
#include "de_graphics.h"

#include "doom\p_local.h"
#include "m_misc.h"
#include "r_model.h"

#include "r3d_OGLfunc.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void Rend_RenderVisSprite(vissprite_t *spr);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

float	maxSpriteAngle = 60.0f;
int		pspOffX = 0;
int		pspOffY = 0;
int		rendSharp = 1;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// The colors to modify when sprites are lit.
static rendpoly_vertex_t slRGB1, slRGB2; 
static fvertex_t slViewVec;
static vissprite_t *slSpr;

// CODE --------------------------------------------------------------------

//===========================================================================
// Rend_Draw3DPlayerSprites
//===========================================================================
void Rend_Draw3DPlayerSprites(void)
{
	int i;

	// Clear Z buffer so models are not clipped by nearby polygons.
	glClear(GL_DEPTH_BUFFER_BIT);
	glTranslatef(-fvx, -fvy, -fvz);	// Translate to player's position

	// sub frames - weapon/muzzle flash usually
	for(i = 0; i < DDMAXPSPRITES; i++)
	{
		if (vispsprites[i].visflags & VIS_MODEL)
			Rend_RenderVisModel(vispsprites + i, true);
	}
}

//===========================================================================
// Rend_DrawPlayerSprites
//	Draws 2D player sprites unless 3D types exist 
//===========================================================================
void Rend_DrawPlayerSprites(void)
{
	spriteinfo_t	info;
	ddpsprite_t		*psp;
//	float			offx = pspOffX / 16.0f;
	float			offy = pspOffY / 16.0f;
	int				i;

	// Cameramen have no psprites.
//	if (ppt.viewplayer->flags & DDPF_CAMERA) return;

	for (i = 0, psp = ppt.viewplayer->psprites; i < DDMAXPSPRITES; i++, psp++)
	{
		if (!psp->stateptr || psp->flags & DDPSPF_RENDERED)  // Used?	
			continue;

		psp->flags |= DDPSPF_RENDERED;

		if (ge_inf.batman)	// More fucking about
		{
			static int lastplrweapon = -1;

			if (ppt.viewplrmo->player->readyweapon != lastplrweapon)
				lastplrweapon = ppt.viewplrmo->player->readyweapon;

			if (lastplrweapon == 1 && psp->stateptr->sprite != 3) return;
		}

		R_GetSpriteInfo(psp->stateptr->sprite, 
			psp->stateptr->frame & FF_FRAMEMASK, &info, 2);

		if (psp->stateptr->frame & FF_FULLBRIGHT || LevelFullBright)
		{
			glColor4f(1.0f, 1.0f, 1.0f, psp->alpha);
		}
		else
		{
			glColor4f (psp->plight, psp->plight, psp->plight, psp->alpha);
		}

		GL_DrawPSprite(psp->x - info.xoffset,
					   psp->y - info.yoffset + offy,
					   info.flip, info.splnum, false, 0);
	}
}

//===========================================================================
// Rend_RenderVisModel
//===========================================================================
void Rend_RenderVisModel(vissprite_t *vismdl, RDBOOL psprite)
{
	submodeldef_t	*smf;
	int				i;

	if (!vismdl->mo.mf) return;

	if (whitefog && vismdl->mo.vmoflags & (HAS_NOFOG|HAS_FOGUNSET))
		glDisable(GL_FOG);

	// Render each submodel making up the model.
	for (i = 0; i < MAX_FRAME_MODELS; i++)
	{
		if ( (smf = &vismdl->mo.mf->sub[i]) && smf->model)
			Rend_RenderModel(vismdl, vismdl->mo.mf, smf, i, psprite);
	}

	if (whitefog && vismdl->mo.vmoflags & (HAS_NOFOG|HAS_FOGUNSET))
		glEnable(GL_FOG);
}

//===========================================================================
// Rend_RenderVISWall - Draw translucent/masked/fog wall
//  v220-3 fog wall support added
//===========================================================================
static void Rend_RenderVISWall(vissprite_wall_t *w, int visflags)
{
	if (visflags & VIS_FOGWALL)
	{
		glDisable(GL_TEXTURE_2D);
		glDepthMask(GL_FALSE);
		glBegin(GL_QUADS);
		glColor4ubv(w->vertices[0].color);
		glVertex3f(w->vertices[0].pos[VX], w->bottom, w->vertices[0].pos[VY]);
		glVertex3f(w->vertices[0].pos[VX], w->top, w->vertices[0].pos[VY]);
		glColor4ubv(w->vertices[1].color);
		glVertex3f(w->vertices[1].pos[VX], w->top, w->vertices[1].pos[VY]);
		glVertex3f(w->vertices[1].pos[VX], w->bottom, w->vertices[1].pos[VY]);
		glEnd();
		glEnable(GL_TEXTURE_2D);
		glDepthMask(GL_TRUE);
		return;
	}

	GL_Bind(w->texture);

	if (visflags & VIS_NOFOG) glDisable(GL_FOG);

	if (visflags & VIS_TRANS)
	{
		glDepthMask(GL_FALSE);
	}
	else if (!(visflags & VIS_NOALPHACUT))
	{
		glAlphaFunc(GL_GEQUAL, 0.5f);
		glDisable(GL_BLEND);
	}

	// Draw one quad.
	glBegin(GL_QUADS);
	glColor4ubv(w->vertices[0].color);

	glTexCoord2f(w->texc[0][VX], w->texc[1][VY]);
	glVertex3f(w->vertices[0].pos[VX], w->bottom, w->vertices[0].pos[VY]);

	glTexCoord2fv(w->texc[0]);
	glVertex3f(w->vertices[0].pos[VX], w->top, w->vertices[0].pos[VY]);

	glColor4ubv(w->vertices[1].color);

	glTexCoord2f(w->texc[1][VX], w->texc[0][VY]);
	glVertex3f(w->vertices[1].pos[VX], w->top, w->vertices[1].pos[VY]);

	glTexCoord2fv(w->texc[1]);
	glVertex3f(w->vertices[1].pos[VX], w->bottom, w->vertices[1].pos[VY]);

	glEnd();

	if (visflags & VIS_NOFOG) glEnable(GL_FOG);

	if (visflags & VIS_TRANS)
	{
		glDepthMask(GL_TRUE);
	}
	else if (!(visflags & VIS_NOALPHACUT))
	{
		glAlphaFunc(GL_GREATER, 0);
		glEnable(GL_BLEND);
	}
}

//===========================================================================
// Rend_Draw_vissprites
//	Render sprites, 3D models, masked wall segments and halos, ordered 
//	back to front
//
//  b220-0 changed to use revised sort code
//  b220-3 translucent_water added
//
//===========================================================================

void Rend_Draw_vissprites (int select)
{
	uint	i;

	for (i = 0; i < visspritecnt; i++)
	{
		// Draw order: furthest to nearest
		vissprite_t *spr = vispointer[i];

		if (!spr) Con_Error("Unset vissprite [%i %i]", i, visspritecnt);

		if (spr->visflags & VIS_RENDERED) continue;

		// $translucent_water
		if (select)
		{
			if (spr->visflags & VIS_WALL || !(spr->mo.vmoflags & HAS_WATERPLANE))
			{
				if (select > 0)
				{
					if (spr->mo.vmoflags & HAS_UNDERWATER) continue;
				}
				else
				{
					if (!(spr->mo.vmoflags & HAS_UNDERWATER)) continue;
				}
			}
		}

		if (spr->visflags & VIS_WALL)
		{
			Rend_RenderVISWall(&spr->wall, spr->visflags);
		}
		// v220-31 If using the menu or console, limit rendering to the nearest
		//         1500 Things to reduce stress. All of the code, even up to
		//         calling the gl functions, could be retained without causing
		//         much of a slowdown. This would be pointless, however, given
		//         the intention is to render nothing. It is the gl functions
		//         themselves that present the greatest time penalty.
		//
		//         This has been added in case a user has started nuts.wad
		//         or similar on a slow machine and is struggling to exit.
		else if ( !((menu_active || con_freezegame) && visspritecnt - i > 1500) )
		{
			if (spr->mo.mf)
			{
				Rend_RenderVisModel(spr, false);	// Model
			}
			else// if (spr->mo.splnum >= 0)
			{
				Rend_RenderVisSprite(spr);	// Standard sprite
			}
		}

		spr->visflags |= VIS_RENDERED;
	}
}


//===========================================================================
// Rend_SpriteTexCoord
//===========================================================================
void Rend_SpriteTexCoord(int splnum, int x, int y)
{
	glTexCoord2f(spritelumps[splnum].tc[VX] * x, spritelumps[splnum].tc[VY] * y);
}

//===========================================================================
// Rend_SpriteLighter
//===========================================================================
static RDBOOL Rend_SpriteLighter(lumobj_t *lum, fixed_t dist)
{
	fvertex_t	lightVec;
	float		fdist, zfactor;
	float		directness, side;
	float		d, inleft, inright;
	float		factor;
	int			i, temp;

	if(!dist) return true;

	if(    slRGB1.rgba[0] == 0xff
		&& slRGB1.rgba[1] == 0xff
		&& slRGB1.rgba[2] == 0xff 
		&& slRGB2.rgba[0] == 0xff
		&& slRGB2.rgba[1] == 0xff
		&& slRGB2.rgba[2] == 0xff)
	{
		return false; // No point in continuing, light is already white.
	}

	lightVec.x = FIX2FLT(slSpr->mo.gx - lum->th->x);
	lightVec.y = FIX2FLT(slSpr->mo.gy - lum->th->y);

	fdist = FIX2FLT(dist);

	d = (FIX2FLT(slSpr->mo.gz) + FIX2FLT(slSpr->mo.gzt))/2;

	zfactor = (d - FIX2FLT(lum->z) + lum->center) / dlMaxRad;

	if(zfactor < 0) zfactor = -zfactor;
	if(zfactor > 1) return true; // Too high or low.
	zfactor = 1-zfactor;
	// Enlarge the full-lit area.
	zfactor *= 2;
	if(zfactor > 1) zfactor = 1;
	
	lightVec.x /= fdist;
	lightVec.y /= fdist;

	// Also include the effect of the distance to zfactor.
	fdist /= (lum->radius*2) /*dlMaxRad*/;
	fdist = 1-fdist;
	//fdist *= 4;
	if(fdist > 1) fdist = 1;
	zfactor *= fdist;

	// Now the view vector and light vector are normalized.
	directness = slViewVec.x * lightVec.x + slViewVec.y * lightVec.y; // Dot product.
	side = -slViewVec.y * lightVec.x + slViewVec.x * lightVec.y;
	// If side > 0, the light comes from the right.
	if(directness > 0)
	{
		// The light comes from the front.
		if(side > 0)
		{
			inright = 1;
			inleft = directness;
		}
		else 
		{
			inleft = 1;
			inright = directness;
		}
	}
	else
	{
		// The light comes from the back.
		if(side > 0)
		{
			inright = side;
			inleft = 0;
		}
		else 
		{
			inleft = -side;
			inright = 0;
		}
	}
	inright *= zfactor;
	inleft *= zfactor;

	if (lum->flags&LUMF_PLAYER)
		factor = 1.f - M_AccurateDistanceFP(FIX2FLT(lum->th->x) - slSpr->mo.v1[VX], FIX2FLT(lum->th->y) - slSpr->mo.v1[VY])/(float)lum->radius;
	else
		factor = 1.f;

	if(inleft > 0)
	{
		for(i=0; i<3; i++)
		{
			temp = (int)(slRGB1.rgba[i] + inleft*lum->rgb[i]);
			if(temp > 0xff) temp = 0xff;
			if (factor < 1.f) temp = MAX_OF((byte)(temp*factor), slRGB1.rgba[i]);
			slRGB1.rgba[i] = temp;
		}
	}
	if(inright > 0)
	{
		for(i=0; i<3; i++)
		{
			temp = (int)(slRGB2.rgba[i] + inright*lum->rgb[i]);
			if(temp > 0xff) temp = 0xff;
			if (factor < 1.f) temp = MAX_OF((byte)(temp*factor), slRGB2.rgba[i]);
			slRGB2.rgba[i] = temp;
		}
	}
	return true;
}

//===========================================================================
// Rend_ScaledAmbientLight
//===========================================================================
static void Rend_ScaledAmbientLight(byte *out, byte *ambient, float mul)
{
	int i;

	if(mul < 0) mul = 0;
	if(mul > 1) mul = 1;

	for(i=0; i<3; i++)
	{
		if(out[i] < ambient[i]*mul)
			out[i] = (byte)(ambient[i]*mul);
	}
}

//===========================================================================
// Rend_RenderVisSprite
//===========================================================================
static void Rend_RenderVisSprite(vissprite_t *spr)
{
	int			splnum = spr->mo.splnum;
	int			offset = 0;
	int			ll = 0;
	float		i, bot, top;
	DGLubyte	alpha;
	RDBOOL		additiveBlending = false, flip;
	RDBOOL		uselight = false;
	RDBOOL		restore_blend = false;
	byte		*dc_translation;

	if (whitefog && spr->mo.vmoflags & (HAS_NOFOG|HAS_FOGUNSET))
		glDisable(GL_FOG);

	// Do we need to translate any of the colors?
	if(spr->mo.flags & DDMF_TRANSLATION)
	{
		dc_translation = translationtables-256 + spr->mo.pclass*((8-1)*256) +
			((spr->mo.flags & DDMF_TRANSLATION) >> (DDMF_TRANSSHIFT-8));
		// We need to prepare a translated version of the sprite.
		GL_SetTranslatedSprite(splnum, dc_translation);
	}
	else
	{
		// Get the texture rgb. No translation required.
		GL_SetSprite(splnum, NULL, 3);
	}

	// Set the lighting and alpha.
	if (spr->visflags & VIS_TRANSSPR)
	{
		alpha = 180;
		additiveBlending = true;
	}
	else if (spr->visflags & VIS_STEALTH)
	{
		alpha = 85;
	}
	else if (spr->mo.flags & DDMF_BRIGHTSHADOW)
	{
		alpha = 204;	// 80 %.
		additiveBlending = true;
	}
	else if(spr->mo.flags & DDMF_SHADOW)
		alpha = 85;		// One third.
	else if(spr->mo.flags & DDMF_ALTSHADOW)
		alpha = 160;	// Two thirds.
	else
		alpha = 255;

	// Sprite has a custom alpha multiplier?
	if(spr->mo.alpha >= 0) alpha = (byte)(alpha * spr->mo.alpha);

	if (spr->visflags & VIS_SHADOW)
	{
		glColor4ub(0, 0, 0, alpha);
	}
	else if (spr->mo.lightlevel < 0)
	{
		glColor4ub(0xff, 0xff, 0xff, alpha);
	}
	else
	{
		ll = r_ambient > spr->mo.lightlevel? r_ambient : spr->mo.lightlevel;

		slRGB1.vpdist = spr->distance;

		RL_VertexColors(&slRGB1, 1, ll, false);

		// Add extra light using dynamic lights.
		if ( (numLuminous && !(spr->visflags & VIS_NOTLIT) && alpha == 255) )
		{
			float len;

			uselight = true;

			memcpy(slRGB2.rgba, slRGB1.rgba, 3);

			// Global variables as parameters... ugly.
			slSpr = spr;
			slViewVec.x = FIX2FLT(spr->mo.gx - viewx);
			slViewVec.y = FIX2FLT(spr->mo.gy - viewy);

			len = (float)(sqrt(slViewVec.x*slViewVec.x + slViewVec.y*slViewVec.y));

			if (len)
			{
				slViewVec.x /= len;
				slViewVec.y /= len;
				DL_RadiusIterator(spr->mo.gx, spr->mo.gy, dlMaxRadft, false, Rend_SpriteLighter);
			}

			// Check floor and ceiling for glow. They count as ambient light.
			if(spr->mo.vmoflags & (HAS_CEILGLOW|HAS_FLOORGLOW))
			{
				len = 1 - (FIX2FLT(spr->mo.gz) - spr->mo.secfloor) / glowHeight;

				for(i = 0; i < 2; i++)
					Rend_ScaledAmbientLight(slRGB1.rgba, spr->mo.floorglow, len);

				len = 1 - (spr->mo.secceil - FIX2FLT(spr->mo.gzt)) / glowHeight;

				for(i = 0; i < 2; i++)
					Rend_ScaledAmbientLight(slRGB2.rgba, spr->mo.ceilglow, len);
			}				
		}

		slRGB1.rgba[CA] = alpha;
		glColor4ubv(slRGB1.rgba);
	}

	// Should the texture be flipped?
	flip = spr->mo.flip;

	top = spr->mo.top;
	bot = top - (float)spritelumps[splnum].height;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	// First off translate to servo values (if any are set)
	if (spr->mo.visoff[VX] != 0 || spr->mo.visoff[VY] != 0 || spr->mo.visoff[VZ] != 0)
	{
		glTranslatef(spr->mo.visoff[VX], spr->mo.visoff[VZ], spr->mo.visoff[VY]);
	}

	if (spr->mo.viewaligned)
	{
		float centerx = FIX2FLT(spr->mo.gx);
		float centery = FIX2FLT(spr->mo.gy);
		float centerz = (top+bot)/2;

		glTranslatef(centerx, centerz, centery);	// Translate to sprite.centre
		glRotatef(vpitch, viewsidex, 0, viewsidey);	// rotate perpendicular to the view plane.
		glTranslatef(-centerx, -centerz, -centery);	// Translate back
	}

	if(additiveBlending)
	{
		// Change the blending mode.
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	}
	else if (rendSharp
		&& alpha == 255
		&& !(spr->visflags & VIS_NOALPHACUT)
		&& spritelumps[splnum].alpha)
	{
		glAlphaFunc(GL_GEQUAL, 0.5f);
		glDisable(GL_BLEND);
		restore_blend = true;
	}

	// Render the sprite.
	glBegin(GL_QUADS);
	Rend_SpriteTexCoord(splnum, flip, 1);
	glVertex3f(spr->mo.v1[VX], bot, spr->mo.v1[VY]);
	Rend_SpriteTexCoord(splnum, flip, 0);
	glVertex3f(spr->mo.v1[VX], top, spr->mo.v1[VY]);

	if(uselight)
	{
		slRGB2.rgba[CA] = alpha;
		glColor4ubv(slRGB2.rgba);
	}

	Rend_SpriteTexCoord(splnum, !flip, 0);
	glVertex3f(spr->mo.v2[VX], top, spr->mo.v2[VY]);
	Rend_SpriteTexCoord(splnum, !flip, 1);
	glVertex3f(spr->mo.v2[VX], bot, spr->mo.v2[VY]);
	glEnd();

	// Restore the original matrices
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	if(additiveBlending)
	{
		// Change to normal blending.
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else if (restore_blend)
	{
		glAlphaFunc(GL_GREATER, 0);
		glEnable(GL_BLEND);
	}

	if (whitefog && spr->mo.vmoflags & (HAS_NOFOG|HAS_FOGUNSET))
		glEnable(GL_FOG);
}
