//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_direc.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Change made to Dir_ConvertSlashes()
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Directory and file system related stuff
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <direct.h>
#include <ctype.h>

#include "de_base.h"
#include "de_system.h"
#include "de_misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// b211-0 Added
// When using a wild card with findfirst/findnext (e.g. "*.wad")
// then files with extensions like .wad-OK will also be selected.
// This function's been added to check file extensions.
RDBOOL Dir_ChExt(char *fname, char *ext)
{
	char file_ext[MAXEXT];

	_splitpath(fname, NULL, NULL, NULL, file_ext);

	return !_stricmp(file_ext, ext);
}

void Dir_GetDir(directory_t *dir)
{
	dir->drive = _getdrive();
	_getcwd(dir->path, 255);
	if(dir->path[strlen(dir->path)-1] != '\\')
		strcat(dir->path, "\\");
}

int Dir_ChDir(directory_t *dir)
{
	_chdrive(dir->drive);
	return !_chdir(dir->path); // Successful if == 0.
}

void Dir_MakeDir(char *path, directory_t *dir)
{
	char temp[256];

	Dir_FileDir(path, dir);
	Dir_FileName(path, temp);
	strcat(dir->path, temp);
	Dir_ValidDir(dir->path); // Make it a well formed path.
}

//===========================================================================
// Dir_FileDir
//	Translates the given filename (>,} => basedir).
//===========================================================================
void Dir_FileDir(char *str, directory_t *dir)
{
	char temp[MAXPATH], pth[MAXPATH];

	pth[0] = 0;	// b211-0

	M_TranslatePath(str, pth);
	_fullpath (temp, pth, MAXPATH - 1);
	_splitpath(temp, dir->path, pth, 0, 0);
	strcat(dir->path, pth);
	dir->drive = toupper(dir->path[0]) - 'A' + 1;
}

void Dir_FileName(char *str, char *name)
{
	char ext[MAXEXT];

	_splitpath(str, 0, 0, name, ext);
	strcat(name, ext);
}

int Dir_FileID(char *str)
{
	int i, id = 0x5c33f10e;
	char temp[MAXPATH], *ptr;

	memset(temp, 0, sizeof(temp));
	// Do a little scrambling with the full path name.
	_fullpath(temp, str, MAXPATH - 1);
	_strupr(temp);	
	for(i=0, ptr=temp; *ptr; ptr++, i++)
		((unsigned char*) &id)[i % 4] += *ptr;
	return id;
}

//===========================================================================
// Dir_IsAbsolute
//	Returns true if the the second character is a ':' (drive).
//===========================================================================
int Dir_IsAbsolute(const char *str)
{
	if (str && strlen(str) > 2)
	{
		char	szdrive[MAXDRIVE];

		_splitpath(str, szdrive, NULL, NULL, NULL);
		
		return (int)strlen(szdrive);
	}

	return 0;
}

//===========================================================================
// Dir_ValidDir
//	Appends a backslash, if necessary. Also converts forward slashes into
//	backward ones. Does not check if the directory actually exists, just
//	that it's a well-formed path name.
//===========================================================================
void Dir_ValidDir(char *str)
{
	int i, len = strlen(str);

	// DOS-ify slashes.
	for(i = 0; i < len; i++) if(str[i] == '/') str[i] = '\\';

	// Remove whitespace from the end.
	for(i = len - 1; isspace(str[i]) && i >= 0; i--) str[i] = 0; 

	// Make sure it's valid.
	if(str[len - 1] != '\\') strcat(str, "\\");
}

//===========================================================================
// Dir_ConvertSlashes (v220-13 as per original DD source)
//===========================================================================
void Dir_ConvertSlashes(char *modifiableBuffer)
{  
	uint i;

	for (i = 0; i < strlen(modifiableBuffer); i++)
	{
        if( modifiableBuffer[i] == '/' ) modifiableBuffer[i] = '\\';
    }
}
