//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: ui_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  modified for use with R3D
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  vid-arrays added
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  extensively revised
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
//        ge_inf.usefmodmidi added
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  User Interface, ties to the console routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "de_base.h"
#include "de_audio.h"
#include "de_console.h"
#include "de_system.h"
#include "de_graphics.h"
#include "de_ui.h"
#include "de_misc.h"

#include "rend_sky.h"
#include "s_mus.h"
#include "r3d_compat.h"
#include "r3d_edit.h"
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define SCROLL_TIME		5

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	CP_ToggleList(ui_object_t *ob);
void	GL_CheckVSYNC(void);
void	R_ChangeFlatGlow(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

int		UI_ListItemHeight(uidata_list_t *listdata);
int		UI_ListButtonHeight(ui_object_t *ob);
int		UI_ListThumbPos(ui_object_t *ob);
void	UI_StrCpyLen(char *dest, char *src, int max_width);
int		UI_MouseInsideBox(int x, int y, int w, int h);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		mididev;
extern int		midi_error;
extern int		playerdrag;
extern int		default_FOV;
extern short	iMidiFlags;
extern RDBOOL	vsyncSupported;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

char		showStatus[32];			// v220-13
uint		musicpos = 0xFFFFFFFF;	// v220-13
RDBOOL		ui_active = false;		// The user interface is active
RDBOOL		ui_showmouse = true;
RDBOOL		ui_setfov = false;
RDBOOL		ui_use43;
ui_page_t	*ui_page = 0;			// Currently active page
int			ui_fonthgt;				// Height of the UI font
DGLuint		ui_cursor;	// Cursor and screen textures
DGLubyte	*screen_buffer;
int			ui_cx, ui_cy;			// Cursor position.
int			ui_rest_cx, ui_rest_cy;
int			ui_rest_start;			// Start time of current resting
int			ui_rest_time = TICSPERSEC/2;	// 500 ms (sort of)
int			ui_rest_offset_limit = 2;
int			ui_moved;				// True if the mouse has been moved

ui_color_t ui_colors[NUM_UI_COLORS] =
{
	// The Risen3D colors. (Search using R3Dcolours)
	{ 1.0f,   1.0f,   0.0f },	// UIC_TEXT
	{ 0.8f,   0.8f,   0.8f },	// UIC_GREY
	{ 0.0f,   1.0f,   1.0f },	// UIC_WARN
	{ 0.3f,   0.9f,   0.5f },	// UIC_MUSIC
	{ 1.0f,   0.8f,   0.8f },	// UIC_LIST
	{ 0.0f,   0.0f,   0.3f },	// UIC_SHADOW
	{ 0.1f,   0.1f,   0.1f },	// UIC_BG_LIGHT
	{ 0.9f,   0.9f,   0.9f },	// UIC_BG_MEDIUM
	{ 0.3f,   0.3f,   0.3f },	// UIC_BG_DARK
	{ 1.0f,   1.0f,   1.0f },	// UIC_BRD_HI
	{ 0.4f,   0.4f,   0.4f },	// UIC_BRD_MED
	{ 0.0f,   0.0f,   0.0f },	// UIC_BRD_LO
	{ 0.882f, 0.785f, 0.420f },	// UIC_BRD_RHI
	{ 0.551f, 0.455f, 0.262f },	// UIC_BRD_RMED
	{ 0.220f, 0.126f, 0.105f },	// UIC_BRD_RLOW
	{ 0.882f, 0.882f, 0.882f },	// UIC_BRD_WHI
	{ 0.551f, 0.551f, 0.551f },	// UIC_BRD_WMED
	{ 0.220f, 0.220f, 0.220f },	// UIC_BRD_WLOW
	// colours added for automap v220-22
	{ 1.0f,   0.0f,   0.0f },	// UIC_RED
	{ 0.0f,   1.0f,   1.0f }	// UIC_CYAN
};

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int  ui_anisotropic;
static int	ui_mididev;
const int val_vid_cap[] = { 100, 111, 125, 143, 167, 200, 250, 333 };

// CODE --------------------------------------------------------------------

void Set_UI_cursor()
{
	static int	dcursor = -1;	// v220-3
	byte		*image;

	if (ui_cursor) return;

	if (dcursor == -1)
		dcursor = W_CheckRisen3DWADNumForName("DDCURSOR", 1);

	// Load cursor.
	image = (byte*)M_CacheLumpNum(dcursor, NULL, true);

	GL_NewTexture(&ui_cursor);
	GL_TexImage(GL_RGBA, 16, 32, 0, image);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	M_Free(image);
}

void UI_Set_showStatus()
{
	char *ptr = (gamestate == GS_LEVEL) ? mapinfo->id : "INTERVAL";

	if (midi_selected && midi_error)
	{
		if (midi_error == 1)
			sprintf(showStatus, "%s %s FILE ERROR", ptr, playStatus);
		else
			sprintf(showStatus, "ERROR: BAD FILE or MIDI DEVICE");
	}
	else
		sprintf(showStatus, "%s %s", ptr, playStatus);
}

//===========================================================================
// UI_Init
//===========================================================================
void UI_Init(void)
{
	if (ui_active) return;

	ui_active = true;
	ui_anisotropic = anisotropic;
	ui_mididev = mididev;

	UI_Set_showStatus();

	// Setup state.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, current_w, current_h, 0, -1, 1);
	
	FR_SetFont(1);	// Change font.
	
	ui_fonthgt = FR_TextHeight("W");

	// Should the mouse cursor be visible?
	ui_showmouse = mouseInited;

	Set_UI_cursor();

	ui_cx = current_w/2;
	ui_cy = current_h/2;
	ui_moved = false;
}

void setClampPeriod()
{
	extern uint clamp_period;

	// clamp limits
	if (cfg.vid_cap <  4) cfg.vid_cap =  4;
	if (cfg.vid_cap > 11) cfg.vid_cap = 11;

	clamp_period = 14 - (uint)cfg.vid_cap;
}

//===========================================================================
// UI_End
//===========================================================================
void UI_End(void)
{
	event_t rel;

	if (!ui_active) return;

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	ui_active = false;		// Restore old state.
	mididev = ui_mididev;	// Catch case where option changed but not acted upon

	S_LocalMenuSound(sfx_pstop);
	GL_FreeScreenBuffer();
	GL_DeleteTexIndex(&ui_cursor, "ui_cursor");

	if (ui_setfov || (ui_use43 != cfg.use43 && !cMode.w_checkwin))
	{
		GL_SetRatioAndFov();
		ui_setfov = false;
	}

	GL_CheckVSYNC();
	R_InitZDist();		// v220-9 may have changed
	setClampPeriod();	// v220-13 may have changed
	R_ChangeFlatGlow();	// v220-33 may have changed

	// v220-31 User may have changed current stored settings
	if (set_light_defaults)
	{
		if ( !(R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_SCPLIGHT) )
		{
			ge_inf.flags &= ~(GEF_TORCHLIGHT|GEF_VTXLIGHT|GEF_SINECOMPRESS);
			if (cfg.useTorchLight) ge_inf.flags |= GEF_TORCHLIGHT;
			if (cfg.setVtxLight	 ) ge_inf.flags |= GEF_VTXLIGHT;
			if (cfg.sineCompress ) ge_inf.flags |= GEF_SINECOMPRESS;
		}
		set_light_defaults = 0;
	}

	if (!R3D_edit)
	{
		if ( !(R3D_ScriptedMap || !CFX) )
		{
			ge_inf.num_modelLight = modelLights;
		}

		if ( !(R3D_ScriptedMap || !BDMFX || ma_inf.noscriptset & NSF_WATERTRANS) )
		{
			ge_inf.flags &= ~GEF_WATERTRANS;
			if (water_trans) ge_inf.flags |= GEF_WATERTRANS;
			ge_inf.was_water_opacity = water_opacity;
		}

		if (!R3D_ScriptedMap)
		{
			if (cfg.footsteps		) ge_inf.flags |= GEF_FOOTSTEPS; else ge_inf.flags &= ~GEF_FOOTSTEPS;
			if (cfg.weaponsounds	) ge_inf.flags |= GEF_WEAPRSOUND; else ge_inf.flags &= ~GEF_WEAPRSOUND;
			if (cfg.weapsmoke		) ge_inf.flags |= GEF_WEAPSMOKE; else ge_inf.flags &= ~GEF_WEAPSMOKE;
			if (cfg.splashsprite	) ge_inf.flags |= GEF_SPLASHSPRITE; else ge_inf.flags &= ~GEF_SPLASHSPRITE;
			if (useDynLights		) ge_inf.flags |= GEF_DYNLIGHTS; else ge_inf.flags &= ~GEF_DYNLIGHTS;
		}
	}

	stats.fpsnum = 35;
	stats.framecount = 0;

	if (ui_anisotropic != anisotropic)
	{
		Con_Execute("texreset", false);
		FR_SetFont(1);	// v220-6 reset the font to the correct size
	}

	// Check if the shift key was (possibly) released
	// while the UI was eating all the input events.
	if (!shiftDown)
	{
		rel.type = ev_keyup;
		rel.data1 = DDKEY_RSHIFT;
		DD_PostEvent(&rel);
	}
}

//===========================================================================
// UI_InitPage
//===========================================================================
void UI_InitPage(ui_page_t *page, ui_object_t *objects)
{
	int i;
	ui_object_t *deffocus = NULL;
	ui_object_t meta;

	page->objects = objects;
	page->focus = -1;
	page->capture = -1;				// No capture.
	page->drawer = UIPage_Drawer;
	page->responder = UIPage_Responder;
	page->ticker = UIPage_Ticker;
	page->timer = 0;
	page->count = UI_CountObjects(objects);

	for(i = 0; i < page->count; i++)
	{
		objects[i].flags &= ~UIF_FOCUS;

		if (   objects[i].type == UI_TEXT 
			|| objects[i].type == UI_BOX
			|| objects[i].type == UI_META)
		{
			objects[i].flags |= UIF_NO_FOCUS;
		}

		if(objects[i].flags & UIF_DEFAULT)
			deffocus = objects + i;

		objects[i].timer = 0;		// Reset timer.
	}

	if(deffocus)
	{
		page->focus = deffocus - objects;
		deffocus->flags |= UIF_FOCUS;
	}
	else
	{
		// Find an object for focus.
		for(i=0; i<page->count; i++)
		{
			if(!(objects[i].flags & UIF_NO_FOCUS))
			{
				// Set default focus.
				page->focus = i;
				objects[i].flags |= UIF_FOCUS;
				break;
			}
		}
	}
	// Meta effects.
	for(i = 0, meta.type = 0; i < page->count; i++)
	{
		if(!meta.type && objects[i].type != UI_META) continue;

		if(objects[i].type == UI_META)
		{
			// This will be the meta for now.
			memcpy(&meta, objects + i, sizeof(meta));
			// Neutralize the actual object.
			objects[i].group = UIG_NONE;
			objects[i].flags |= UIF_HIDDEN;
			objects[i].relx = objects[i].rely = 0;
			objects[i].relw = objects[i].relh = 0;
			continue;
		}

		// Apply the meta.
		if(meta.group != UIG_NONE) objects[i].group = meta.group;
		objects[i].relx += meta.relx;
		objects[i].rely += meta.rely;
		objects[i].relw += meta.relw;
		objects[i].relh += meta.relh;
	}
}

//===========================================================================
// UI_AvailableWidth
//	The width of the available page area, in pixels.
//===========================================================================
int UI_AvailableWidth(void)
{
	return current_w - UI_BORDER*4;
}

//===========================================================================
// UI_AvailableHeight
//===========================================================================
int UI_AvailableHeight(void)
{
	return current_h - UI_TITLE_HGT - UI_BORDER*4;
}

//===========================================================================
// UI_ScreenX
//	Convert a relative coordinate to a screen coordinate.
//===========================================================================
int UI_ScreenX(int relx)
{
	return UI_BORDER*2 + (relx * UI_AvailableWidth()) / 1000;		
}

//===========================================================================
// UI_ScreenY
//	Convert a relative coordinate to a screen coordinate.
//===========================================================================
int UI_ScreenY(int rely)
{
	return UI_BORDER*2 + UI_TITLE_HGT + (rely * UI_AvailableHeight()) / 1000; 
}

//===========================================================================
// UI_ScreenW
//===========================================================================
int UI_ScreenW(int relw)
{
	return (relw * UI_AvailableWidth()) / 1000;
}

//===========================================================================
// UI_ScreenH
//===========================================================================
int UI_ScreenH(int relh)
{
	return (relh * UI_AvailableHeight()) / 1000;
}

//===========================================================================
// UI_SetPage
//	Change and prepare the active page.
//===========================================================================
void UI_SetPage(ui_page_t *page)
{
	int i;
	ui_object_t *ob;

	ui_page = page;	
	if(!page) return;
	// Init objects.
	for(i=0, ob=page->objects; i<page->count; i++, ob++) 
	{
		// Calculate real coordinates.
		ob->x = UI_ScreenX(ob->relx);
		ob->w = UI_ScreenW(ob->relw);
		ob->y = UI_ScreenY(ob->rely); 
		ob->h = UI_ScreenH(ob->relh);
		// Update edit box text.
		if(ob->type == UI_EDIT)
		{
			memset(ob->text, 0, sizeof(ob->text));
			strncpy(ob->text, ((uidata_edit_t*)ob->data)->ptr, 255);
		}
		// Stay-down button state.
		if(ob->type == UI_BUTTON2)
		{
			if( *(char*) ob->data)
				ob->flags |= UIF_ACTIVE;
			else
				ob->flags &= ~UIF_ACTIVE;
		}
		// List box number of visible items.
		if(ob->type == UI_LIST)
		{
			uidata_list_t	*dat = (uidata_list_t*)ob->data;
			int				relh = UI_ScreenH(354);

			dat->numvis = dat->numtodisplay;//(ob->h - 2*UI_BORDER) / ui_fonthgt;//UI_ListItemHeight(dat);
			ob->h = dat->numvis * ui_fonthgt + 2*UI_BORDER;

			while (ob->h > relh)
			{
				dat->numvis--;//(ob->h - 2*UI_BORDER) / ui_fonthgt;//UI_ListItemHeight(dat);
				ob->h = dat->numvis * ui_fonthgt + 2*UI_BORDER;
			}

			if(dat->selection < dat->first)
				dat->first = dat->selection;
			if(dat->selection > dat->first+dat->numvis-1)
				dat->first = dat->selection - dat->numvis + 1;
			UI_InitColumns(ob);
		}
	}
	// The mouse has not yet been moved on this page.
	ui_moved = false;
}

int UI_Responder(event_t *ev)
{
	if (!ui_active) return false;
	if (!ui_page) return false;

	// Check for Shift events.
	switch(ev->type)
	{
	case ev_mouse:
		if(ev->data1 || ev->data2) ui_moved = true;
		ui_cx += ev->data1;
		ui_cy += ev->data2;
		if(ui_cx < 0) ui_cx = 0;
		if(ui_cy < 0) ui_cy = 0;
		if(ui_cx >= current_w) ui_cx = current_w-1;
		if(ui_cy >= current_h) ui_cy = current_h-1;
		break;

	default:
		break;
	}
	// Call the page's responder.
	ui_page->responder(ui_page, ev);
	// If the UI is active, all events are eaten by it.
	return true;
}

void UI_Ticker(void)
{
	if(!ui_active) return;
	if(!ui_page) return;
	// Call the active page's ticker.
	ui_page->ticker(ui_page);
}

void UI_Drawer(void)
{
	static int vol = -1, nowtime;

	if(!ui_active) return;
	if(!ui_page) return;
	// Call the active page's drawer.
	ui_page->drawer(ui_page);
	// Draw mouse cursor.
	UI_DrawMouse(ui_cx, ui_cy);

	if (vol < 0)
	{
		vol = sfx_volume;
		nowtime = Sys_GetTime();
	}
	else
	if (vol != sfx_volume)
	{
		if (Sys_GetTime() - nowtime > 5.0f)
		{
			vol = sfx_volume;
			nowtime = Sys_GetTime();
			S_LocalMenuSound(sfx_swtchn);
		}
	}
}

int UI_CountObjects(ui_object_t *list)
{
	int count;

	for(count=0; list->type != UI_NONE; list++, count++);
	return count;
}

//===========================================================================
// UI_FlagGroup
//===========================================================================
void UI_FlagGroup(ui_object_t *list, int group, int flags, int set)
{
	for(; list->type; list++)
	{
		if(list->group == group)
		{
			switch(set)
			{
			case UIFG_CLEAR:
				list->flags &= ~flags;
				break;
				
			case UIFG_SET:
				list->flags |= flags;
				break;
				
			case UIFG_XOR:
				list->flags ^= flags;
				break;
			}
		}
	}
}

//===========================================================================
// UI_FindObject
//	All the specified flags must be set.
//===========================================================================
ui_object_t *UI_FindObject(ui_object_t *list, int group, __int64 flags)
{
	for(; list->type; list++)
	{
		if (   list->group == group
			&& (list->flags & flags) == flags) return list;
	}

	return NULL;
}

// Ob must be on the current page! It can't be NULL.
void UI_Focus(ui_object_t *ob)
{
	int i;

	if(!ob) Con_ErrorD("UI_Focus: Tried to set focus on NULL");

	// Can the object receive focus?
	if(ob->flags & UIF_NO_FOCUS) return;
	ui_page->focus = ob - ui_page->objects;

	for(i=0; i<ui_page->count; i++)
	{
		if(i == ui_page->focus)
			ui_page->objects[i].flags |= UIF_FOCUS;
		else
			ui_page->objects[i].flags &= ~UIF_FOCUS;
	}
}

// Ob must be on the current page!
// If ob is NULL, capture is ended.
void UI_Capture(ui_object_t *ob)
{
	if(!ob)
	{
		// End capture.
		ui_page->capture = -1;
		return;
	}
	if(!ob->responder) return;	// Sorry, pal...
	// Set the capture object.
	ui_page->capture = ob - ui_page->objects;
	// Set focus.
	UI_Focus(ob);
}

//---------------------------------------------------------------------------
// Default Callback Functions
//---------------------------------------------------------------------------
int UIPage_Responder(ui_page_t *page, event_t *ev)
{
	ui_object_t *ob;
	int			i, k;

	if (page->capture >= 0)
	{
		// There is an object that has captured input.
		ob = page->objects + page->capture;

		// Capture objects must have a responder!
		// This object gets to decide what happens.
		return ob->responder(ob, ev);
	}

	// Check for ESC or TAB keys
	if (ev->type == ev_keydown || ev->type == ev_keyrepeat)
	{
		// No repeats with ESC
		if (ev->data1 == DDKEY_ESCAPE && ev->type == ev_keydown)
		{
			UI_SetPage(page->previous);
			if(!ui_page) UI_End();		// If no page is set then exit the UI
			return true;				// The event was used.
		}

		if (ev->data1 == DDKEY_TAB)		// TAB is used for navigation
		{
			// Remove the focus flag from the current focus object.
			page->objects[page->focus].flags &= ~UIF_FOCUS;

			k = 0;

			do	// Move focus
			{
				page->focus += shiftDown? -1 : 1;
				// Check range.
				if(page->focus < 0) 
					page->focus = page->count - 1;
				else if(page->focus >= page->count)
					page->focus = 0;
			}
			
			while (++k < page->count && page->objects[page->focus].flags & (UIF_DISABLED|UIF_NO_FOCUS|UIF_HIDDEN|UIF_NOTDRAWN));

			// Flag the new focus object.
			page->objects[page->focus].flags |= UIF_FOCUS;

			return true;	// The event was used.
		}
	}

	if (shiftDown) return true;	// Only wanted with SHIFT/TAB

	// Call responders until someone uses the event
	// Start with the focused object
	for (i = 0; i < page->count; i++)
	{
		// Determine the index of the object to process
		if ((k = page->focus + i) < 0) k += page->count;	// Wrap to top
		if (k >= page->count) k -= page->count;				// Wrap to bottom

		ob = page->objects + k;

		// Check the flags of this object.
		if (  !ob->responder								// Must have a responder
			|| ob->flags & (UIF_HIDDEN|UIF_DISABLED))		// These flags prevent response
		{
			continue;
		}

		// Don't allow ENTER with a list. This is not obvious until
		// trying to use control panel with no mouse or joystick.
		// Unlike buttons, list entries don't change data.
		if (   ev->data1 == DDKEY_ENTER
			&& ob->flags & (UIF_LIST|UIF_RES_LISTF|UIF_RES_LISTW))
		{
			return false;
		}

		if (ob->responder(ob, ev))
		{
			UI_Focus(ob);	// Focus on object that uses the event
			return true;
		}
	}

	return false;	// The event was not used
}

//===========================================================================
// UIPage_Ticker
//	Call the ticker routine for each object.
//===========================================================================
void UIPage_Ticker(ui_page_t *page)
{
	int i;
	ui_object_t *ob;

	// Call the ticker of each object, unless they're hidden or paused.
	for(i = 0, ob = page->objects; i < page->count; i++, ob++)
	{
		if(ob->flags & UIF_PAUSED || ob->flags & UIF_HIDDEN) continue;
		if(ob->ticker) ob->ticker(ob);
		// Advance object timer.
		ob->timer++;
	}
	page->timer++;

	// Check mouse resting.
	if(abs(ui_cx - ui_rest_cx) > ui_rest_offset_limit
		|| abs(ui_cy - ui_rest_cy) > ui_rest_offset_limit)
	{
		// Restart resting period.
		ui_rest_cx = ui_cx;
		ui_rest_cy = ui_cy;
		ui_rest_start = page->timer;
	}
}

void UIPage_Drawer(ui_page_t *page)
{
	int i;
	float t;
	ui_object_t *ob;
	ui_color_t focuscol;

	UI_DrawTitle(page);				// Draw title

	// Draw each object, unless they're hidden
	for(i=0, ob=page->objects; i<page->count; i++, ob++)
	{
		if(ob->flags & UIF_HIDDEN || !ob->drawer) continue;

		ob->drawer(ob);	
		
		if (   ob->flags & UIF_FOCUS
			&& (ob->type != UI_EDIT || !(ob->flags & UIF_ACTIVE)))
		{
			t = (page->timer % 30) / 15.0f;
			if(t > 1) t = 2-t;
			UI_MixColors(UI_COL(UIC_TEXT), UI_COL(UIC_BRD_MED), &focuscol, t);				
			// Draw a focus rectangle.
			UI_DrawRect(ob->x-1, ob->y-1, ob->w+2, ob->h+2, 1, UI_COL(UIC_SHADOW),
				&focuscol, UI_COL(UIC_TEXT));
		}
	}
}

RDBOOL UI_CheckListOb(ui_object_t *ob)
{
	switch(ob->flags & 0xFF)
	{
	case UIEF_MIDIDEV: return compat[c_nomusic] || ge_inf.usefmodmidi || !midi_selected;
	case UIEF_MODVIS : return !loadModels || R3D_ScriptedMap;
	}
	return false;
}

RDBOOL UI_CheckLightingChange()
{
	if (   cfg.useTorchLight  && !(ge_inf.flags & GEF_TORCHLIGHT) ) return true;
	if ( !(cfg.useTorchLight) &&   ge_inf.flags & GEF_TORCHLIGHT  ) return true;
	if (   cfg.setVtxLight  && !(ge_inf.flags & GEF_VTXLIGHT) ) return true;
	if ( !(cfg.setVtxLight) &&   ge_inf.flags & GEF_VTXLIGHT  ) return true;
	if (   cfg.sineCompress  && !(ge_inf.flags & GEF_SINECOMPRESS) ) return true;
	if ( !(cfg.sineCompress) &&   ge_inf.flags & GEF_SINECOMPRESS  ) return true;
	return false;
}

// v220-3 / v220-4
static RDBOOL chkObjectFlags(ui_object_t *ob)
{
	switch(ob->flags & 0xFF)
	{
	case UIEF_EDITINH	: return (!BDMFX || !r_texglow || R3D_edit);
	case UIEF_SETGLOW	: return (R3D_edit || !BDMFX);
	case UIEF_SCPINH	: return R3D_ScriptedMap;
	case UIEF_TXTTRANS2	: return (!BDMFX || (!ge_inf.devmap && (R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_WATERTRANS)));
	case UIEF_TRANSONLY	: return !BDMFX || !water_trans || (!ge_inf.devmap && (R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_WATERTRANS));
	case UIEF_MODONLY	: return (R3D_edit || R3D_ScriptedMap || !BDMFX);
	case UIEF_MUSIC		: return compat[c_nomusic];
	case UIEF_SOUND		: return compat[c_nosound];
	case UIEF_SHADOW	: return !BDMFX || R3D_slopetypes;
	case UIEF_OBSHADOW	: return !BDMFX || R3D_slopetypes || !cfg.useShadows;
	case UIEF_TXTNODET2	: return !has_detail;
	case UIEF_NODETAIL	: return !has_detail || !r_detail;
	case UIEF_CHANGE	: return ui_anisotropic == anisotropic;
	case UIEF_DLIGHTS	: return (!useDynLights || R3D_ScriptedMap || !CFX || R3D_edit);
	case UIEF_DLITLEVEL	: return (!useDynLights || !CFX || R3D_edit);
	case UIEF_MOUSE		: return ge_inf.nomouse;
	case UIEF_AMBIENT	: return !R3D_edit;
	case UIEF_PARTICLES	: return !CFX;
	// v220-4
	case UIEF_SCPALLOW	: return (R3D_edit || !R3D_ScriptedMap);
	case UIEF_FOOTSTEP	: return (!BDMFX || compat[c_nosound] || R3D_ScriptedMap);
	case UIEF_WEAPSND	: return (!BDMFX || compat[c_nosound] || R3D_ScriptedMap);
	case UIEF_SPRITE	: return (loadModels & 2);
	// v220-5
	case UIEF_ASPECT	: return !valid_screen_ratio;
		// v220-9
	case UIEF_DARKEN	: return R3D_ScriptedMap && !ge_inf.devmap;
	case UIEF_NOANISO	: return maxAniso < 2.0f;	// v220-9
	case UIEF_MODVIS	: return !loadModels || R3D_ScriptedMap;
	case UIEF_NOSKYBOX	: return R3D_ScriptedMap || !has_skybox;
	case UIEF_HASVSYNC	: return !vsyncSupported;
		// v220-13
	case UIEF_VIDCAP    : return vidVsync;
	case UIEF_SLDMUSIC  : return compat[c_nomusic];
	case UIEF_FMODMUSIC	: return compat[c_nomusic] || (midi_selected && !ge_inf.usefmodmidi);
	case UIEF_TOGGLE	: return compat[c_nomusic] || !(ge_inf.UseList & 1) || gamestate != GS_LEVEL;
	case UIEF_MIDI		:
		{
			uint n = strlen(playStatus);

			return compat[c_nomusic] || !midi_selected || (n > 3 && strcmp(&playStatus[n - 4], " MUS"));
		}
	case UIEF_MIDIDEV	: return compat[c_nomusic] || ge_inf.usefmodmidi || !midi_selected || mididev == ui_mididev;
	case UIEF_MIDMUSIC	: return compat[c_nomusic] || ge_inf.usefmodmidi || !midi_selected || iMidiFlags & 1;
	case UIEF_MIDIRESET	: return compat[c_nomusic] || ge_inf.usefmodmidi || !midi_selected || iMidiFlags & 2;
		// v220-31
	case UIEF_LDEFS		:
		if (!UI_CheckLightingChange()) return true;	// else fall through
	case UIEF_LSINE		:
	case UIEF_LTORCH	:
	case UIEF_LVTX		:
		return (R3D_edit || (!ge_inf.devmap && (R3D_ScriptedMap || ma_inf.noscriptset & NSF_SCPLIGHT)) );
	case UIEF_SPLASH:
		return (compat[c_nosound] || !BDMFX || !cfg.footsteps);
	case UIEF_SMOKE:
		return (!loadModels);
	default:
		return false;
	}
	return false;	// happy compiler
}

void UIText_Drawer(ui_object_t *ob)
{
	char	uitmpstr[40];
	char	*ptr = NULL;
	float	alpha;
	float	newalpha = 0.5f;
	int		textcolour = UIC_TEXT;

	if (chkObjectFlags(ob)) return;

	switch(ob->flags & 0xFF)
	{
	case UIEF_TXTNODETAIL:
		if (!has_detail)
			ptr = "DETAIL WAD NOT LOADED - OPTIONS DISABLED";
		break;
	case UIEF_TXTNODET2:
		if (!has_detail) return;
	case UIEF_TXTSOUND:
		if (compat[c_nosound])
			ptr = "SOUND DISABLED IN THE LAUNCHER";
		break;
	case UIEF_TXTMUSIC:
		if (compat[c_nomusic])
		{
			ptr = "MUSIC DISABLED - OPTIONS NOT AVAILABLE";
		}
		else if (music_playing[0])
		{
			ptr = music_playing;
			textcolour = UIC_MUSIC;
			newalpha = 1.0f;
		}
		else
		{
			ptr = "No map music available";
			textcolour = UIC_MUSIC;
		}
		break;
	case UIEF_TXTMODONLY:
		if (!BDMFX)
			ptr = "BASIC DOOM MODE - DYNLIGHTS DISABLED";
		else if (ma_inf.forcedyn)
			ptr = "Dynlight types detected - DYNLIGHTS ALWAYS ON";
		break;
	case UIEF_TXTSETGLOW:
		if (R3D_edit)
			ptr = "EDIT MODE - GLOW DISABLED";
		else if (!BDMFX)
			ptr = "BASIC DOOM MODE - GLOW DISABLED";
		break;
	case UIEF_TXTMODVIS:
		if (R3D_ScriptedMap)
			ptr = "SCRIPTED MODEL MAP - SPRITE SWAP DISABLED";
		if (!loadModels)
			ptr = "No models loaded - further settings disabled";
		break;
	case UIEF_TXTSHADOW:
		if (!BDMFX)
			ptr = "BASIC DOOM MODE - SHADOW OPTIONS DISABLED";
		if (R3D_slopetypes)
			ptr = "MAP HAS SLOPES - SHADOW OPTIONS DISABLED";
		break;
	case UIEF_TXTTRANSLUC:
		if (!BDMFX)
			ptr = "BASIC DOOM MODE - TRANSLUCENT WATER DISABLED";
		else
		if (!ge_inf.devmap && (R3D_ScriptedMap || ma_inf.noscriptset & NSF_WATERTRANS))
			ptr = "WATER TRANSLUCENCY IS MAP DEFINED";
		else if (R3D_edit)
			ptr = "Edit Mode: WATER TRANSLUCENCY IS DISABLED";
		break;
	case UIEF_TXTTRANS2:
		if (R3D_edit || (!ge_inf.devmap && (R3D_ScriptedMap || ma_inf.noscriptset & NSF_WATERTRANS)))
			ptr = "";
		break;
	case UIEF_TXTLIGHTS:
		if (!CFX || !useDynLights) return;
		if (R3D_ScriptedMap)
			ptr = "SCRIPTED MODEL MAP - LIGHT NUMBER OPTION DISABLED";
		break;
	case UIEF_TXTDARKEN:
		if (R3D_ScriptedMap && !ge_inf.devmap)
			ptr = "SCRIPTED MAP: SCRIPT SETS LIGHT OPTIONS";
		break;
	case UIEF_TXTLTORCH:
		if (!ge_inf.devmap && (R3D_ScriptedMap || ma_inf.noscriptset & NSF_SCPLIGHT))
			ptr = "LIGHT OPTIONS HAVE BEEN PRESET";
		break;
	case UIEF_TXTLDEFS:
		if (!UI_CheckLightingChange())
		{
			ptr = "Current settings match defaults";
		}
		else
		{
			if (!ge_inf.devmap && (R3D_ScriptedMap || ma_inf.noscriptset & NSF_SCPLIGHT))
				ptr = "Light defaults cannot be set with this map";
			else if (R3D_edit)
				ptr = "Light defaults cannot be set in edit mode";
		}
		break;
	case UIEF_TXTFOOTSTEP:
		if (compat[c_nosound]) return;

		if (!BDMFX)
			ptr = "BASIC DOOM MODE - FOOTSTEPS DISABLED";
		else
		if (R3D_ScriptedMap)
			ptr = "SCRIPTED FOOTSTEPS ENABLED";
		break;
	case UIEF_TXTWEAPSND:
		if (compat[c_nosound]) return;

		if (!BDMFX)
			ptr = "BASIC DOOM MODE - WEAPON READY SOUNDS DISABLED";
		else
		if (R3D_ScriptedMap)
			ptr = "SCRIPTED WEAPON SOUNDS ENABLED";
		break;
		break;
	case UIEF_TXTSPLASH:
		if (!BDMFX)
			ptr = "BASIC DOOM MODE - SPLASHES DISABLED";
		else
		if (!cfg.footsteps)
			ptr = "Splash requires footsteps (see audio)";
		break;
	case UIEF_TXTSMOKE:
		if (!loadModels)
			ptr = "MODELS NOT LOADED - SMOKE DISABLED";
		break;
	case UIEF_TXTAMBIENT:
		if (!R3D_edit)
			ptr = "AMBIENT LIGHT CAN ONLY BE CHANGED IN EDIT MODE";
		break;
	case UIEF_TXTSCPALLOW:
		if (R3D_ScriptedMap) return;
		break;
	case UIEF_TXTMOUSE:
		if (ge_inf.nomouse)
		{
			if (ge_inf.nomouse != 1)
				ptr = "MOUSE FAILED TO INITIALISE";
			else
				ptr = "USER DISABLED MOUSE AT STARTUP";
		}
		break;
	case UIEF_TXTHUDSIZE:
		if (R3D_ScriptedMap)
			ptr = "HUD SIZE FIXED IN SCRIPTED MAP";
		break;
	case UIEF_TXTHUDTYPE:
		{
			static const char *size[] = { "", "ON SCR", "RIGHT", "MID", "LEFT", "MAX" };
			static const char *type[] = { "", "Risen3D", "DOOM small right", "DOOM small centre", "DOOM small left", "DOOM" };

			if (R3D_ScriptedMap) return;

			if (ob->relx == 590)
				ptr = (char*)size[cfg.sbarscale];
			else
				ptr = (char*)type[cfg.sbarscale];

			newalpha = 1.0f;
		}
		break;
	case UIEF_SCPNOSKY:
		if (R3D_ScriptedMap)
				ptr = "SKY SETTINGS FIXED IN SCRIPTED MAP";
		else if (!has_skybox)
		{
			if (!CFX)
				ptr = "BASIC DOOM MODE - SKYBOX NOT SUPPORTED";
			else
				ptr = "No skyboxes loaded";
		}
		break;
	case UIEF_SETVSYNC:
		if (!vidVsync || !vsyncSupported)
		{
			ptr = "VIDEO OPTIONS (With VSYNC off FPS cap can be set from 100 to 333)";
			newalpha = 1.0f;
		}
		break;
	case UIEF_CHKVSYNC:
		if (!vsyncSupported)
				ptr = "VSYNC not supported";
		break;
	case UIEF_DEFFOV:
		ptr = uitmpstr;
		sprintf(ptr, "Base FOV: default %d", default_FOV);
		newalpha = 1.0f;
		break;
	default:
		if (ob->action == CP_ToggleList)
		{
			extern char	*listName;

			if (compat[c_nomusic]) return;

			ptr = (listName && ge_inf.UseList & 2) ? listName : showStatus;
			newalpha = 1.0f;
		}
		break;
	}

	if (ptr)
	{
		alpha = newalpha;
	}
	else
	{
		ptr = ob->text;

		if (ob->flags & UIF_DISABLED)
		{
			alpha = 0.2f;
		}
		else
			alpha = 1.0f;
	}

	UI_TextOutEx(ptr, ob->x, ob->y + ob->h/2, false, true, UI_COL(textcolour), alpha);
}

int UIButton_Responder(ui_object_t *ob, event_t *ev)
{
	if (   chkObjectFlags(ob)
		|| (ob->action == CP_ToggleList && (!(ge_inf.UseList & 1) || gamestate != GS_LEVEL)) )
	{
		ob->flags |= UIF_NOTDRAWN;
		return false;
	}

	ob->flags &= ~UIF_NOTDRAWN;

	if(ob->flags & UIF_CLICKED)
	{
		if(ev->type == ev_mousebup || ev->type == ev_keyup)
		{
			UI_Capture(0);
			ob->flags &= ~UIF_CLICKED;		
			if(UI_MouseInside(ob) || ev->type == ev_keyup)
			{
				// Activate?
				if(ob->action) ob->action(ob);
			}
			return true;
		}
	}
	else if((ev->type == ev_mousebdown && UI_MouseInside(ob))
		||  (ev->type == ev_keydown && IS_ACTKEY(ev->data1)))
	{
		if(ob->type == UI_BUTTON)
		{
			// Capture input.
			UI_Capture(ob);
			ob->flags |= UIF_CLICKED;
		}
		else
		{
			int flags = (int)(ob->flags & UIF_ACTIVE);

			// Stay-down buttons change state.
			ob->flags ^= (int)UIF_ACTIVE;

			if (ob->data)
			{	*(int*) ob->data = (ob->flags & UIF_ACTIVE) != 0;
			}

			if (ob->action)		// Call the action function.
			{	ob->action(ob);
			}

			if (flags ^ (int)(ob->flags & UIF_ACTIVE))
				S_LocalMenuSoundAtVolume(flags ? sfx_swtchx : sfx_swtchn, 0.4f);

		}
		ob->timer = 0;
		return true;
	}
	return false;
}

void UIButton_Drawer(ui_object_t *ob)
{
	int act = (ob->flags & UIF_ACTIVE) != 0;
	int click = (ob->flags & UIF_CLICKED) != 0;
	int down = act || click;
	ui_color_t back;
	float t = ob->timer/15.0f;
	float alpha = 1.0f;

	//------------------------------------------

	if (chkObjectFlags(ob)) return;

	if (ob->action == CP_ToggleList)
	{
		if (gamestate != GS_LEVEL || !(ge_inf.UseList & 1))
			return;
	}

	if (ob->flags & UIF_DISABLED)
	{
		alpha = .2f;
	}

	// Mix the background color.
	if(!click || t > .5f) t = .5f;
	if(act && t > .1f) t = .1f;
	UI_MixColors(UI_COL(UIC_TEXT), UI_COL(UIC_SHADOW), &back, t);
	UI_Gradient(ob->x+UI_BUTTON_BORDER, ob->y+UI_BUTTON_BORDER, 
		ob->w-UI_BUTTON_BORDER*2, ob->h-UI_BUTTON_BORDER*2,	&back, 0, alpha, 0);

	UI_DrawRectEx((float)ob->x, (float)ob->y, ob->w, ob->h, (float)UI_BUTTON_BORDER, 
		UI_COL(down? UIC_BRD_LOW : UIC_BRD_HI), 
		UI_COL(UIC_BRD_MED), 
		UI_COL(down? UIC_BRD_HI : UIC_BRD_LOW),	alpha);

	UI_TextOutEx(
		ob->action == CP_ToggleList ? "TOGGLE" : ob->text,
		down*2 + ob->x 
		+ (ob->flags & UIF_LEFT_ALIGN? UI_BUTTON_BORDER*2 : ob->w/2), 
		down*2 + ob->y + ob->h/2,
		!(ob->flags & UIF_LEFT_ALIGN), true, 
		UI_COL(UIC_TEXT), alpha);
}

int UIList_Responder(ui_object_t *ob, event_t *ev)
{
	uidata_list_t *dat = (uidata_list_t*)ob->data;

	// v220-3 always process UIF_RES_LISTW|UIF_RES_LISTF
	int		oldsel = (ob->flags & (UIF_RES_LISTW|UIF_RES_LISTF)) ? -1 : dat->selection;

	int		i,buth, barh;
	int		used = false;

	//--------------------------

	if (!dat->count || UI_CheckListOb(ob)) return false;

	if(ob->flags & UIF_CLICKED)
	{
		// We've captured all input.
		if(ev->type == ev_mousebup)
		{
			dat->button[1] = false;
			// Release capture.
			UI_Capture(0);
			ob->flags &= ~UIF_CLICKED;
		}
		else if(ev->type == ev_mouse)
		{
			// Calculate the new position.
			buth = UI_ListButtonHeight(ob);
			barh = ob->h - 2*(UI_BORDER + buth);
			if(barh - buth)
			{
				dat->first = ((ui_cy - ob->y - UI_BORDER - (buth*3)/2) * 
					(dat->count - dat->numvis) + (barh-buth)/2) / (barh-buth);
			}
			else
			{
				dat->first = 0;
			}
			// Check that it's in range.
			if(dat->first < 0) dat->first = 0;
			if(dat->first > dat->count - dat->numvis)
				dat->first = dat->count - dat->numvis;
		}
		// We're eating everything.
		return true;
	}
	
	if (ev->type == ev_keydown || ev->type == ev_keyrepeat)
	{
		used = true;	// assume

		switch(ev->data1)
		{
		case DDKEY_UPARROW:
			if(dat->selection > 0) dat->selection--;
			break;

		case DDKEY_DOWNARROW:
			if(dat->selection < dat->count-1) dat->selection++;
			break;
		
		case DDKEY_HOME:
			dat->selection = 0;
			break;

		case DDKEY_END:
			dat->selection = dat->count-1;
			break;

		default:
			used = false;
		}
	}
	else if(ev->type == ev_mousebdown)
	{
		if(!UI_MouseInside(ob)) return false;

		// Now we know we're going to eat this event.
		used = true;
		buth = UI_ListButtonHeight(ob);

		// Clicked in the item section?
		if(UI_MouseInsideBox(ob->x + UI_BORDER, ob->y + UI_BORDER,
			ob->w - 2*UI_BORDER - (dat->count >= dat->numvis? UI_BAR_WDH : 0), 
			ob->h - 2*UI_BORDER))
		{
			dat->selection = dat->first + (ui_cy - ob->y - UI_BORDER) / ui_fonthgt;
			//	UI_ListItemHeight(dat);
			if(dat->selection >= dat->count) dat->selection = dat->count-1;
		}
		else if(dat->count < dat->numvis)
		{
			// No scrollbar.
			return true;
		}
		// Clicked the Up button?
		else if(UI_MouseInsideBox(ob->x + ob->w - UI_BORDER - UI_BAR_WDH,
			ob->y + UI_BORDER, UI_BAR_WDH, buth))
		{
			// The Up button is now pressed.
			dat->button[0] = true;
			ob->timer = SCROLL_TIME;	// Ticker does the scrolling.
			return true;
		}
		// Clicked the Down button?
		else if(UI_MouseInsideBox(ob->x + ob->w - UI_BORDER - UI_BAR_WDH,
			ob->y + ob->h - UI_BORDER - buth, UI_BAR_WDH, buth))
		{
			// The Down button is now pressed.
			dat->button[2] = true;
			ob->timer = SCROLL_TIME;	// Ticker does the scrolling.
			return true;
		}
		// Clicked the Thumb?
		else if(UI_MouseInsideBox(ob->x + ob->w - UI_BORDER - UI_BAR_WDH,
			UI_ListThumbPos(ob), UI_BAR_WDH, buth))
		{
			dat->button[1] = true;
			// Capture input and start tracking mouse movement.
			UI_Capture(ob);
			ob->flags |= UIF_CLICKED;
			return true;
		}
		else return false;
	}
	else if(ev->type == ev_mousebup)
	{
		// Release all buttons.
		for(i = 0; i < 3; i++) dat->button[i] = false;
		// We might have left the object's area, so let other objects
		// process this event, too.
		return false;	
	}
	else 
	{
		// We are not going to use this event.
		return false;
	}
	// Adjust the first visible item.
	if(dat->selection < dat->first)
		dat->first = dat->selection;
	if(dat->selection > dat->first+dat->numvis-1)
		dat->first = dat->selection - dat->numvis + 1;
	// Call action function?
	if(oldsel != dat->selection && ob->action) ob->action(ob);
	return used;
}

void UIList_Ticker(ui_object_t *ob)
{
	uidata_list_t *dat = (uidata_list_t*)ob->data;

	if(ob->timer >= SCROLL_TIME && (dat->button[0] || dat->button[2]))
	{
		// Next time in 3 ticks.
		ob->timer = 0;
		// Send a direct event.
		if(dat->button[0] && dat->first > 0) dat->first--;
		if(dat->button[2] && dat->first < dat->count-dat->numvis) dat->first++;
	}
}

void UIList_Drawer(ui_object_t *ob)
{
	uidata_list_t		*dat = (uidata_list_t*)ob->data;
	uidata_listitem_t	*items = (uidata_listitem_t*)dat->items;

	char	buf[256], tmp[256];
	char	*ptr, *endptr;

	int		ihgt, maxw = ob->w - 2*UI_BORDER;
	int		i, c, x, y, barw;
	int		maxh = ob->h - 2*UI_BORDER, buth;
	int		dis;
	float	alpha;

	RDBOOL	drawbkgnd;

	//-------------------------------------------

	if (UI_CheckListOb(ob)) { ob->flags &= ~UIF_FOCUS; return; }

	dis =  (ob->flags & UIF_DISABLED) != 0;

	alpha = dis ? .2f : 1.0f;

	// The background.
	UI_Gradient(ob->x + UI_BORDER, ob->y + UI_BORDER, 
		ob->w - UI_BORDER*2, ob->h - UI_BORDER*2,
		UI_COL(UIC_SHADOW), 0, alpha, 0);

	// The borders.
	UI_DrawRectEx((float)ob->x, (float)ob->y, ob->w, ob->h, (float)UI_BORDER,
		UI_COL(UIC_BRD_LOW), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_HI),
		alpha);

	// The title.
	UI_TextOutEx(ob->text, ob->x, ob->y - UI_BORDER - ui_fonthgt, false, false,
		UI_COL(UIC_TEXT), alpha);

	// Is a scroll bar necessary?
	if (dat->numvis < dat->count)
	{
		barw = UI_BAR_WDH;
		maxw -= barw;
		buth = UI_ListButtonHeight(ob);
		x = ob->x + ob->w - UI_BORDER - barw;
		y = ob->y + UI_BORDER;
		UI_Gradient(x, y, barw, maxh, UI_COL(UIC_TEXT), 0, alpha*.2f, alpha*.2f);

		// Up Button.	
		UI_DrawButton(x, y, barw, buth, UI_BAR_BORDER,
			!dat->first? alpha*.2f : alpha, NULL,
			dat->button[0], dis, UIBA_UP);

		// Thumb Button.
		UI_DrawButton(x, UI_ListThumbPos(ob), barw, buth, UI_BAR_BORDER,
			alpha, NULL,
			dat->button[1], dis, UIBA_NONE);

		// Down Button.
		UI_DrawButton(x, y+maxh-buth, barw, buth, UI_BAR_BORDER,
			dat->first+dat->numvis>=dat->count? alpha*.2f : alpha, NULL,
			dat->button[2], dis, UIBA_DOWN);
	}

	x = ob->x + UI_BORDER;
	y = ob->y + UI_BORDER;

	// Draw columns?
	for (c = 0; c < UI_MAX_COLUMNS; c++)
	{
		if(!dat->column[c] || dat->column[c] > maxw-2*UI_BORDER) continue;

		UI_Gradient(x + UI_BORDER + dat->column[c] - 2,
			ob->y + UI_BORDER, 1, maxh, UI_COL(UIC_TEXT), 0, 
			alpha*.5f, alpha*.5f);
	}

	ihgt = ui_fonthgt;//UI_ListItemHeight(dat);
	
	if (ob->flags & (UIF_RES_LISTW|UIF_RES_LISTF))
	{
		drawbkgnd = (ob->flags & UIF_FOCUS) > 0;
	}
	else
		drawbkgnd = true;

	for (i = dat->first; i < dat->count && i < dat->first + dat->numvis; i++, y += ihgt)
	{
		// Draw a background
		if (drawbkgnd && i == dat->selection)
			UI_Gradient(x, y, maxw, ihgt, UI_COL(UIC_TEXT), 0, alpha*.6f, alpha*.2f);

		// The text, clipped w/columns.
		ptr = items[i].text;

		for (c = 0; c < UI_MAX_COLUMNS; c++)
		{
			endptr = strchr(ptr, '\t');
			memset(tmp, 0, sizeof(tmp));

			if(endptr) strncpy(tmp, ptr, endptr-ptr); else strcpy(tmp, ptr);

			memset(buf, 0, sizeof(buf));

			UI_StrCpyLen(buf, tmp, maxw-2*UI_BORDER - dat->column[c]);
			UI_TextOutEx(buf, x + UI_BORDER + dat->column[c], y + ihgt/2, 
				false, true, UI_COL(UIC_TEXT), alpha);

			if(!endptr) break;

			ptr = endptr + 1;
		}
	}
}

//===========================================================================
// UI_SliderButtonWidth
//===========================================================================
int UI_SliderButtonWidth(ui_object_t *ob)
{
	uidata_slider_t *dat = (uidata_slider_t*)ob->data;
	int width = ob->h - UI_BAR_BORDER*2;
	if(width < UI_BAR_BORDER*3) width = UI_BAR_BORDER*3;
	return width;
}

//===========================================================================
// UI_SliderThumbPos
//===========================================================================
int UI_SliderThumbPos(ui_object_t *ob)
{
	uidata_slider_t *dat = (uidata_slider_t*)ob->data;
	float range = dat->max - dat->min, useval;
	int butw = UI_SliderButtonWidth(ob);

	if(!range) range = 1;	// Should never happen.
	useval = dat->value - dat->min;

	if (!dat->floatmode || dat->step > 1.0f)
	{
		useval = (float)((int)useval);
	}

	if (useval < 0) useval = 0;
	if (useval > range) useval = range;

	dat->value = useval + dat->min;

	return (int)(ob->x + UI_BAR_BORDER + butw
		+ useval/range*(ob->w - UI_BAR_BORDER*2 - butw*3));
}

//===========================================================================
// UISlider_Responder
//===========================================================================
int UISlider_Responder(ui_object_t *ob, event_t *ev)
{
	uidata_slider_t	*dat = (uidata_slider_t*)ob->data;
	float			oldvalue = dat->value;
	RDBOOL			used = false;
	int				i, butw, inw;

	if (chkObjectFlags(ob))
	{
		ob->flags |= UIF_NOTDRAWN;
		return false;
	}

	ob->flags &= ~UIF_NOTDRAWN;

	if (ob->flags & UIF_CLICKED)
	{
		// We've captured all input.
		if(ev->type == ev_mousebup)
		{
			dat->button[1] = false; // Release thumb.
			UI_Capture(0);
			ob->flags &= ~UIF_CLICKED;
		}
		if(ev->type == ev_mouse)
		{
			// Calculate new value from the mouse position.
			butw = UI_SliderButtonWidth(ob);
			inw = ob->w - 2*UI_BAR_BORDER - 3*butw;
			if(inw > 0)
			{
				dat->value = dat->min + (dat->max - dat->min)
					* (ui_cx - ob->x - UI_BAR_BORDER - (3*butw)/2)/(float)inw;
			}
			else
			{
				dat->value = dat->min;
			}
			if (!dat->floatmode) dat->value = dat->value + .5f;
			if (dat->value < dat->min) dat->value = dat->min;
			if (dat->value > dat->max) dat->value = dat->max;
			if (ob->action) ob->action(ob);

			// v220-13
			if (oldvalue != dat->value && dat == (uidata_slider_t*)&sld_musicpos)
			{
				musicpos = (uint)(dat->value);
			}
		}

		return true;
	}
	else if(ev->type == ev_keydown || ev->type == ev_keyrepeat)
	{
		used = true;
		switch(ev->data1)
		{
		case DDKEY_HOME:
			dat->value = dat->min;
			break;

		case DDKEY_END:
			dat->value = dat->max;
			break;

		case DDKEY_LEFTARROW:
			dat->value -= dat->step;
			if(dat->value < dat->min) dat->value = dat->min;
			break;

		case DDKEY_RIGHTARROW:
			dat->value += dat->step;
			if(dat->value > dat->max) dat->value = dat->max;
			break;

		default:
			used = false;
		}
	}
	else if(ev->type == ev_mousebdown)
	{
		if(!UI_MouseInside(ob)) return false;

		used = true;
		butw = UI_SliderButtonWidth(ob);

		// Where is the mouse cursor?
		if(UI_MouseInsideBox(ob->x, ob->y, butw + UI_BAR_BORDER, ob->h))
		{
			// The Left button is now pressed.
			dat->button[0] = true;
			ob->timer = SCROLL_TIME;	// Ticker does the scrolling.
			return true;
		}
		if(UI_MouseInsideBox(ob->x + ob->w - butw - UI_BAR_BORDER, ob->y,
			butw + UI_BAR_BORDER, ob->h))
		{
			// The Right button is now pressed.
			dat->button[2] = true;
			ob->timer = SCROLL_TIME;	// Tickes does the scrolling.
			return true;
		}
		if(UI_MouseInsideBox(UI_SliderThumbPos(ob), ob->y, butw, ob->h))
		{
			// Capture input and start tracking mouse movement.
			dat->button[1] = true;
			UI_Capture(ob);
			ob->flags |= UIF_CLICKED;
			return true;
		}
	}
	else if(ev->type == ev_mousebup)
	{
		// Release all buttons.
		for(i = 0; i < 3; i++) dat->button[i] = false;
		// We might be outside the object's area (or none of the buttons
		// might even be pressed), so let others have this event, too.
		return false;
	}
	else
	{
		// We are not going to use this event.
		return false;
	}
	// Did the value change?
	if (oldvalue != dat->value && ob->action)
	{
		ob->action(ob);
	}

	return used;
}

//===========================================================================
// UISlider_Ticker
//===========================================================================
float datAlign(float step, float value, int change)
{
	int v;

	if (!step) return value;

	if (!value)
		v = 0;
	else if (value >= 0)
		v = (int)((value + step/2) / step);
	else
		v = (int)((value - step/2) / step);

	if (change < 0) v -= 1;
	if (change > 0) v += 1;

	return (float)v * step;
}

void UISlider_Ticker(ui_object_t *ob)
{
	uidata_slider_t *dat = (uidata_slider_t*)ob->data;

	if (ob->timer >= SCROLL_TIME)
		ob->timer = 0;

	if (chkObjectFlags(ob)) return;

	if (  !ob->timer
		&& ob->action
		&& (dat->button[0] || dat->button[2]))
	{
		float ov = dat->value;

		dat->value = datAlign(dat->step, dat->value,
			dat->button[2] ? 1 : -1);

		if(dat->value < dat->min) dat->value = dat->min;
		if(dat->value > dat->max) dat->value = dat->max;

		if (ov != dat->value)
			ob->action(ob);
	}
}

//===========================================================================
// UISlider_Drawer
//===========================================================================
void UISlider_Drawer(ui_object_t *ob)
{
	uidata_slider_t *dat = (uidata_slider_t*)ob->data;
	int		inwidth = ob->w - UI_BAR_BORDER*2;
	int		inheight = ob->h - UI_BAR_BORDER*2;
	int		butw = UI_SliderButtonWidth(ob);
	int		x, y, thumbx;
	int		dis =  (ob->flags & UIF_DISABLED) != 0
				|| ((ob->flags & 0xFF) == UIEF_SHADOW && (!BDMFX || R3D_slopetypes || !cfg.useShadows));

	float	alpha = dis ? .2f : 1;
	char	buf[80];

	//--------------------------

	if (chkObjectFlags(ob)) return;

	// The borders.
	UI_DrawRectEx((float)ob->x, (float)ob->y, ob->w, ob->h, (float)UI_BAR_BORDER, 
		UI_COL(UIC_BRD_LOW), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_HI), alpha);
	x = ob->x + UI_BAR_BORDER;
	y = ob->y + UI_BAR_BORDER;

	// The background.
	UI_Gradient(x, y, inwidth, inheight, UI_COL(UIC_SHADOW), 0, alpha, 0);

	// The left button.
	UI_DrawButton(x, y, butw, inheight, UI_BAR_BORDER, 
		alpha * (dat->value == dat->min? .2f : 1), NULL, 
		dat->button[0], dis, UIBA_LEFT);

	// The right button.
	UI_DrawButton(x + inwidth - butw, y, butw, inheight, UI_BAR_BORDER,
		alpha * (dat->value == dat->max? .2f : 1), NULL, 
		dat->button[2], dis, UIBA_RIGHT);

	// The thumb.
	UI_DrawButton(thumbx = UI_SliderThumbPos(ob), y, butw, inheight, 
		UI_BAR_BORDER, alpha, NULL, dat->button[1], dis, UIBA_NONE);

	if (ob->flags & UIF_NOVALUES) return;

	// The value.
	if (dat->zerotext && dat->value == dat->min) 
		strcpy(buf, dat->zerotext);
	else
	{
		if(dat->floatmode)	// v220-33 add support for zbias needing 4 decimal places
		{
			sprintf(buf, ob->flags & UIF_INT ? "%.0f" : (ob->flags & 0xFF) == UIEF_ZBIAS ? "%.4f" : "%.2f", dat->value);
			
			if ((ob->flags & 0xFF) == UIEF_DARKEN)
			{
				ge_inf.maxLightDistance = dat->value;
				sprintf(buf, "%i", (int)dat->value);
			}
		}
		else
		{
			if ((ob->flags & 0xFF) == UIEF_VIDCAP)
				sprintf(buf, "%i", val_vid_cap[(int)dat->value - 4]);	// vid_cap
			else
				sprintf(buf, "%i", (int)dat->value);
		}
	}

	UI_TextOutEx(buf, x + (dat->value < (dat->min + dat->max)/2?
		inwidth - butw - UI_BAR_BORDER - FR_TextWidth(buf) 
		: butw + UI_BAR_BORDER), y + inheight/2, false, true, 
		UI_COL(UIC_TEXT), alpha);
}

//===========================================================================
// UI_DrawScreen v220-3
//  Draw the graphic used with adjusting gamma etc.
//===========================================================================
void UI_DrawScreen(ui_object_t *ob)
{
	if (screen_buffer)
	{
		float	x = (float)UI_ScreenX(ob->relx);
		float	y = (float)UI_ScreenY(ob->rely);
		float	w = (float)UI_ScreenW(ob->relw);
		float	h = (float)UI_ScreenH(ob->relh);

		glDisable(GL_TEXTURE_2D);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		// Use window co-ordinates.
		glOrtho(0, current_w, 0, current_h, -1, 1);

		// Set the lower left hand starting co-ordinate
		glRasterPos2f(x, (float)current_h - (y + h));
		// Zoom to the size required
		glPixelZoom(w / (float)current_w, h / (float)current_h);

		// Tell glDrawPixels how to unpack the screen buffer
		glPixelStorei(GL_UNPACK_ALIGNMENT  , 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH , 0);
		glPixelStorei(GL_UNPACK_SKIP_ROWS  , 0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

		// Draw the pixels to the frame's back buffer
		glDrawBuffer(GL_BACK);
		glDrawPixels(current_w, current_h, GL_BGR,	// v220-8 changed from GL_RGB
							GL_UNSIGNED_BYTE, screen_buffer);

		// Reset to default
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glPixelZoom(1.0f, 1.0f);
		glEnable(GL_TEXTURE_2D);
	}
}

//---------------------------------------------------------------------------
// Helper Functions
//---------------------------------------------------------------------------

void UI_InitColumns(ui_object_t *ob)
{
	uidata_list_t *dat = (uidata_list_t*)ob->data;
	uidata_listitem_t *list = (uidata_listitem_t*)dat->items;
	int i, c, w, sep, last, maxw;
	char *endptr, *ptr, temp[256];
	int width[UI_MAX_COLUMNS];
	int numcols = 1;

	memset(dat->column, 0, sizeof(dat->column));
	memset(width, 0, sizeof(width));
	for(i=0; i<dat->count; i++)
	{
		ptr = list[i].text;
		for(c=0; c<UI_MAX_COLUMNS; c++)
		{
			if(c+1 > numcols) numcols = c + 1;
			endptr = strchr(ptr, '\t');
			memset(temp, 0, sizeof(temp));
			if(endptr) 
				strncpy(temp, ptr, endptr-ptr);
			else
				strcpy(temp, ptr);
			w = FR_TextWidth(temp);
			if(w > width[c]) width[c] = w;
			if(!endptr) break;
			ptr = endptr + 1;
		}
	}
	// Calculate the total maximum width.
	for(w=i=0; i<UI_MAX_COLUMNS; i++) 
	{
		w += width[i];		
		if(width[i]) last = width[i];
	}
	// Calculate the offset for each column.
	maxw = ob->w - 4*UI_BORDER - (dat->count > dat->numvis? UI_BAR_WDH : 0);
	sep = maxw - w;
	if(numcols > 1) sep /= numcols - 1;
	if(sep < 0) sep = 0;
	for(c=i=0; i<numcols; i++)
	{
		dat->column[i] = c;
		c += sep + width[i];
	}
}

int UI_ListItemHeight(uidata_list_t *listdata)
{
	int h = listdata->itemhgt;
	if(h < ui_fonthgt) h = ui_fonthgt;
	return h;
}

int UI_ListButtonHeight(ui_object_t *ob)
{
	int barh = ob->h - 2*UI_BORDER;
	int buth = UI_BAR_WDH;
	if(buth > barh/3) buth = barh/3;
	return buth;
}

int UI_ListThumbPos(ui_object_t *ob)
{
	uidata_list_t *dat = (uidata_list_t*)ob->data;
	int buth = UI_ListButtonHeight(ob);
	int barh = ob->h - 2*(UI_BORDER + buth);

	if(dat->count <= dat->numvis) return 0;
	return ob->y + UI_BORDER + buth + ((barh-buth) * dat->first) /
		(dat->count - dat->numvis);
}

void UI_StrCpyLen(char *dest, char *src, int max_width)
{
	int i, width;
	
	for(i=0, width=0; src[i]; i++)
	{
		dest[i] = src[i];
		width += FR_CharWidth(src[i]);
		if(width > max_width)
		{
			dest[i] = 0;
			break;
		}
	}
}

int UI_MouseInsideBox(int x, int y, int w, int h)
{
	return (ui_cx >= x && ui_cx <= x+w && ui_cy >= y && ui_cy <= y+h);
}

//===========================================================================
// UI_MouseInside
//	Returns true if the mouse is inside the object.
//===========================================================================
int UI_MouseInside(ui_object_t *ob)
{
	return UI_MouseInsideBox(ob->x, ob->y, ob->w, ob->h);
}

//===========================================================================
// UI_MouseResting
//	Returns true if the mouse hasn't been moved for a while.
//===========================================================================
int UI_MouseResting(ui_page_t *page)
{
	if(!ui_moved) return false;
	return page->timer - ui_rest_start >= ui_rest_time;
}

void UI_MixColors(ui_color_t *a, ui_color_t *b, ui_color_t *dest, float amount)
{
	dest->red = (1-amount)*a->red + amount*b->red;
	dest->green = (1-amount)*a->green + amount*b->green;
	dest->blue = (1-amount)*a->blue + amount*b->blue;	
}

void UI_ColorA(ui_color_t *color, float alpha)
{
	glColor4f(color->red, color->green, color->blue, alpha);
}

void UI_Color(ui_color_t *color)
{
	glColor3f(color->red, color->green, color->blue);
}

void UI_DrawTitleEx(char *text, int height)
{
	UI_Gradient(0, 0, current_w, height, 
		UI_COL(UIC_BG_MEDIUM), UI_COL(UIC_BG_LIGHT), .8f, 1);
	UI_Gradient(0, height, current_w, UI_BORDER, 
		UI_COL(UIC_SHADOW),	UI_COL(UIC_BG_DARK), 1, 0);
	UI_TextOutEx(text, UI_BORDER, height/2, false, true, 
		UI_COL(UIC_TEXT), 1);
}

void UI_DrawTitle(ui_page_t *page)
{
	// Draw background gradient
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	UI_Color(UI_COL(UIC_BG_DARK)); // Top color.
	glVertex2i(0, 0);
	glVertex2i(current_w, 0); 
	UI_Color(UI_COL(UIC_BG_LIGHT)); // Bottom color.
	glVertex2i(current_w, current_h);
	glVertex2i(0, current_h);
	glEnd();
	glEnable(GL_BLEND);

	if (ui_showmouse)
		UI_DrawTitleEx("Risen3D v220-"BUILDTEXT" Control Panel", UI_TITLE_HGT);
	else
		UI_DrawTitleEx("Move: TAB/SHIFT-TAB  Select: Button - ENTER,  List - Arrow UP/DOWN then TAB", UI_TITLE_HGT);

	if (paused)
	{
		static float alpha;
		static float inc = .1f;

		alpha += inc;

		if (alpha > 1.f) { alpha = 1.f; inc = -.02f; }
		if (alpha < .4f) { alpha = .4f; inc =  .02f; }

		UI_TextOutEx(
			"**PAUSED**",
			current_w/2, 
			UI_TITLE_HGT/2,
			true, true, 
			UI_COL(UIC_TEXT), alpha);
	}
}

void UI_Gradient(int x, int y, int w, int h, ui_color_t *top, ui_color_t *bottom, 
				 float top_alpha, float bottom_alpha)
{
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	UI_ColorA(top, top_alpha);
	glVertex2i(x  , y);
	glVertex2i(x+w, y);
	UI_ColorA(bottom ? bottom : top, bottom_alpha);
	glVertex2i(x+w, y+h);
	glVertex2i(x  , y+h);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

//===========================================================================
// UI_HorizGradient
//===========================================================================
void UI_HorizGradient
	(int x, int y, int w, int h, ui_color_t *left, ui_color_t *right, 
	 float left_alpha, float right_alpha)
{
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	UI_ColorA(left, left_alpha);
	glVertex2i(x, y+h);
	glVertex2i(x, y);
	UI_ColorA(right? right : left, right_alpha);
	glVertex2i(x+w, y);
	glVertex2i(x+w, y+h);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void UI_Line(int x1, int y1, int x2, int y2, ui_color_t *start, 
			 ui_color_t *end, float start_alpha, float end_alpha)
{
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_LINES);
	UI_ColorA(start, start_alpha);
	glVertex2i(x1, y1);
	UI_ColorA(end? end : start, end_alpha);
	glVertex2i(x2, y2);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

// Draw shadowed text.
void UI_TextOutEx(char *text, int x, int y, int horiz_center, int vert_center,
				   ui_color_t *color, float alpha)
{
	// Center, if requested.
	if(horiz_center) x -= FR_TextWidth(text)/2;
	if(vert_center ) y -= FR_TextHeight(text)/2;

	// Shadow.
	UI_ColorA(UI_COL(UIC_SHADOW), .6f*alpha);
	FR_TextOutOffset(text, (float)x + UI_SHADOW_OFFSET, (float)y + UI_SHADOW_OFFSET);

	// Actual text.
	UI_ColorA(color, alpha);
	FR_TextOut(text, x, y);
}

void UI_DrawRectEx(float x, float y, int w, int h, float brd,
				 ui_color_t *hi, ui_color_t *med, ui_color_t *low, float alpha)
{
	float x_w = x + (float)w;
	float y_h = y + (float)h;
	float xbrd = x + brd;
	float ybrd = y + brd;
	float xbrd_h = x_w - brd;
	float ybrd_h = y_h - brd;

	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	// Top.
	UI_ColorA(hi, alpha);
	glVertex2f(xbrd_h, ybrd);
	glVertex2f(xbrd, ybrd);
	glVertex2f(x, y);
	UI_ColorA(med, alpha);
	glVertex2f(x_w, y);
	// Right.
	UI_ColorA(med, alpha);
	glVertex2f(xbrd_h, ybrd);
	glVertex2f(x_w, y);
	UI_ColorA(low, alpha);
	glVertex2f(x_w, y_h);
	glVertex2f(xbrd_h, ybrd_h);
	// Bottom.
	UI_ColorA(low, alpha);
	glVertex2f(xbrd, ybrd_h);
	glVertex2f(xbrd_h, ybrd_h);
	glVertex2f(x_w, y_h);
	UI_ColorA(med, alpha);
	glVertex2f(x, y_h);
	// Left.
	UI_ColorA(hi, alpha);
	glVertex2f(x, y);
	glVertex2f(xbrd, ybrd);
	UI_ColorA(med, alpha);
	glVertex2f(xbrd, ybrd_h);
	glVertex2f(x, y_h);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void UI_DrawRectSolid(float x, float y, int w, int h, float brd,
				 ui_color_t *hi, ui_color_t *med, ui_color_t *low, float alpha)
{
	UI_DrawRectEx(x, y, w, h, brd, hi, med, low, alpha);

	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	// Centre
	UI_ColorA(low, alpha);
	glVertex2f(x + w - brd, y + h - brd);
	glVertex2f(x + brd, y + h - brd);
	glVertex2f(x + brd, y + brd);
	glVertex2f(x + w - brd, y + brd);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void UI_DrawRectFill(float x, float y, int w, int h, float brd,
				 ui_color_t *hi, ui_color_t *med, ui_color_t *low, float alpha)
{
	UI_DrawRectEx(x, y, w, h, brd, hi, med, low, alpha);

	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	// Centre
	UI_ColorA(hi, alpha);
	glVertex2f(x + w - brd, y + h - brd);
	glVertex2f(x + brd, y + h - brd);
	glVertex2f(x + brd, y + brd);
	UI_ColorA(med, alpha);
	glVertex2f(x + w - brd, y + brd);
	glEnd();
	glEnable(GL_TEXTURE_2D);
}

void UI_DrawRect(int x, int y, int w, int h, int brd,
				 ui_color_t *hi, ui_color_t *med, ui_color_t *low)
{
	UI_DrawRectEx((float)x, (float)y, w, h, (float)brd, hi, med, low, 1);
}

void UI_DrawTriangle(float x, float y, float radius, ui_color_t *hi, 
					 ui_color_t *med, ui_color_t *low, float alpha)
{
	float xrad = radius*.866f;	// cos(60)
	float yrad = radius/2;		// sin(60)

	glDisable(GL_TEXTURE_2D);
	glBegin(GL_TRIANGLES);
	
	y += radius/4;

	// Upper left triangle.
	UI_ColorA(radius > 0? hi : med, alpha);
	glVertex2f(x, y);
	glVertex2f(x - xrad, y + yrad);
	UI_ColorA(radius > 0? hi : low, alpha);
	glVertex2f(x, y - radius);

	// Upper right triangle.
	UI_ColorA(low, alpha);
	glVertex2f(x, y);
	glVertex2f(x, y - radius);
	UI_ColorA(med, alpha);
	glVertex2f(x + xrad, y + yrad);

	// Bottom triangle.
	if(radius < 0) UI_ColorA(hi, alpha);
	glVertex2f(x, y);
	glVertex2f(x + xrad, y + yrad);
	UI_ColorA(radius > 0? low : med, alpha);
	glVertex2f(x - xrad, y + yrad);

	glEnd();
	glEnable(GL_TEXTURE_2D);
}

//===========================================================================
// UI_DrawHorizTriangle
//	A horizontal triangle, pointing left or right. Positive radius 
//	means left.
//===========================================================================
void UI_DrawHorizTriangle(float x, float y, float radius, ui_color_t *hi, 
						  ui_color_t *med, ui_color_t *low, float alpha)
{
	float yrad = radius*.866f;	// cos(60)
	float xrad = radius/2;		// sin(60)

	glDisable(GL_TEXTURE_2D);
	glBegin(GL_TRIANGLES);
	
	x += radius/4;

	// Upper left triangle.
	UI_ColorA(radius > 0? hi : med, alpha);
	glVertex2f(x, y);
	if(radius < 0) UI_ColorA(low, alpha);
	glVertex2f(x - radius, y);
	glVertex2f(x + xrad, y - yrad);

	// Lower left triangle.
	UI_ColorA(radius > 0? med : hi, alpha);
	glVertex2f(x, y);
	if(radius < 0) UI_ColorA(hi, alpha);
	glVertex2f(x + xrad, y + yrad);
	UI_ColorA(radius > 0? low : med, alpha);
	glVertex2f(x - radius, y);

	// Right triangle.
	UI_ColorA(radius > 0? med : hi, alpha);
	glVertex2f(x, y);
	UI_ColorA(radius > 0? hi : med, alpha);
	glVertex2f(x + xrad, y - yrad);
	UI_ColorA(radius > 0? low : hi, alpha);
	glVertex2f(x + xrad, y + yrad);

	glEnd();
	glEnable(GL_TEXTURE_2D);
}

//===========================================================================
// UI_DefaultButtonBackground
//===========================================================================
void UI_DefaultButtonBackground(ui_color_t *col, RDBOOL down)
{
	UI_MixColors(UI_COL(UIC_TEXT), UI_COL(UIC_SHADOW), col, 
		down? .1f : .5f);
}

//===========================================================================
// UI_DrawButton
//===========================================================================
void UI_DrawButton(int x, int y, int w, int h, int brd, float alpha, 
				   ui_color_t *background, int down, int disabled,
				   int arrow)
{
	int inside = MIN_OF(w - brd*2, h - brd*2);
	int boff = down? 2 : 0;
	ui_color_t back;

	if(!background)
	{
		// Calculate the default button color.
		UI_DefaultButtonBackground(&back, down);
		background = &back;
	}

	UI_DrawRectEx((float)x, (float)y, w, h, (float)brd, 
		UI_COL(down? UIC_BRD_LOW : UIC_BRD_HI), 
		UI_COL(UIC_BRD_MED), 
		UI_COL(down? UIC_BRD_HI : UIC_BRD_LOW), alpha);
	UI_Gradient(x + brd, y + brd, w - brd*2, h - brd*2, background, 0, 
		disabled? .2f : 1, 0);
	
	switch(arrow)
	{
	case UIBA_UP:
	case UIBA_DOWN:
		UI_DrawTriangle((float)(x + w/2 + boff), (float)(y + h/2 + boff), 
			inside/2.75f * (arrow == UIBA_DOWN? -1 : 1), 
			UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW),
			alpha * (disabled? .2f : 1));
		break;

	case UIBA_LEFT:
	case UIBA_RIGHT:
		UI_DrawHorizTriangle((float)(x + w/2 + boff), (float)(y + h/2 + boff),
			inside/2.75f * (arrow == UIBA_RIGHT? -1 : 1),
			UI_COL(UIC_BRD_HI), UI_COL(UIC_BRD_MED), UI_COL(UIC_BRD_LOW),
			alpha * (disabled? .2f : 1));
		break;
	}
}

void UI_ChangeMidiDev(ui_object_t *ob)
{
	void	MC_ChangeMidiDev(void);

	MC_ChangeMidiDev();
	UI_Set_showStatus();

	ui_mididev = mididev;
	ob->flags &= ~UIF_FOCUS;
}

void UI_DrawMouse(int x, int y)
{
	if(!ui_showmouse || !ui_cursor || paused) return;

	x--; y--;
	GL_Bind(ui_cursor);
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
	glTexCoord2i(0, 0); glVertex2i(x   , y);
	glTexCoord2i(1, 0); glVertex2i(x+16, y);
	glTexCoord2i(1, 1); glVertex2i(x+16, y+32);
	glTexCoord2i(0, 1); glVertex2i(x   , y+32);
	glEnd();
}
