//-----------------------------------------------------------------------------
//
// Copyright (C) 2007 - 2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_traceMissing.c
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: sets rendering sector and/or line types by inspecting
//				missing textures.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "r3d_MAINspec.h"
#include "r3d_edit.h"
#include "p_maputil.h"
#include "m_misc.h"
   

// MACROS ------------------------------------------------------------------

#define			DRLISTLIMIT	128

// TYPES -------------------------------------------------------------------

typedef struct {
	int		*looplist;
	int		loopcount;
	int		box[4];
	uint	area;
	int		secnum;
	byte	hilo;
	byte	ignore;
	byte	inside;
	byte	nested;
} loopvar_t;

typedef struct {
	loopvar_t	*lvar;
	int			count;
	int			maxcount;
} ringinf_t;

// for setLineRings
typedef struct {
	int	maxcount;
	int	count;
	int box[4];
	int	sz;
	int *list;

} linestore_t;

typedef struct  {
	int		*hiring;
	int		himaxcnt;
	int		*loring;
	int		lomaxcnt;
} rssring_t;

typedef struct  {

	int		hiringcnt;
	int		loringcnt;

	int		eliminate;
	int		movskyceil;
	int		anylines;
	int		uncompliantlines;
	int		sameseclines;
	int		sameheightlines;
	int		notexlines;
	int		boundshitex;
	int		boundsmidtex;
	int		boundslotex;
	int		boundmoveflags;
	int		nohitexinside;	// v220-4
	int		hitex;
	int		midtex;
	int		lotex;
	int		markhilo;
	int		setallownsec;
	int		totalinside;
	int		hiliqfinside;
	int		hiliqfsec;
	// v220-4
	int		lowfloorsequal1;
	int		lowfloorsequal2;

} rssinf_t;


// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static ringinf_t	rinf;
static linestore_t	lst[DRLISTLIMIT];

static rssring_t	rssring;
static rssinf_t		rssinf;

static int		*ringtmp;
static int		*ringhi;
static int		*ringlo;
static int		*ringss;
static int		*linelist;

static byte		*secmark;
static byte		*eliminatesec;
static int		*linememory;
static int		*lnarray2;
static int		*tempsseclist;
static byte		*excludessecs;
static byte		*hitlinebak;
static byte		*misstexlines;
static byte		*bridgelines;
static byte		*linemark;
static byte		*vtfcount;

static int		ringlocnt;
static int		ringhicnt;
static int		ringtmpcnt;
static int		ringsscnt;
static int		linememcnt;

static int		lnarray2cnt;

static int		chkseccnt;
static int		chksec[MAX_SEC_PER_VTX];


//--------------------------------------------------------

static void setCHKSEC(sector_t *s)
{
	int		i;

	for (i = 0; i < chkseccnt; i++)
	{
		if (chksec[i] == s - sectors) return;
	}

	if (chkseccnt < MAX_SEC_PER_VTX)
		chksec[chkseccnt++] = s - sectors;
}

// This cannot be guaranteed, unfortunately, as it depends on the nodesbuilder.
// But works most of the time and not a big problem if it fails (used for small
// infills to approve appearance).
static int getSecLinesFromSSlines(sector_t *s1, sector_t *s2, int lnum)
{
	sector_t *s = NULL;
	ssline_t *sl = &sslines[lnum];
	int		 i, k;
	RDBOOL	 ret = false;

	for (i = 0; i < sl->num; i++)
	{
		subsector_t *ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount != ss_validcount)
		{
			seg_t *sg = &segs[ss->firstline];

			ss->ssvalidcount = ss_validcount;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				line_t *ld = sg->linedef;

				if (!ld || ld->validcount == validcount) continue;

				ld->validcount = validcount;

				if (   !hitline[ld - lines]
					&& !(ld->spflags & SPF_CONTROLLINE)
					&& !Vtxs1(ld)
					&& ld->frontsector != s1
					&& ld->frontsector != s2
					&& ld->backsector  != s1
					&& ld->backsector  != s2)
				{
					int x = (int)(ss->midpoint.x * FRACUNIT);
					int y = (int)(ss->midpoint.y * FRACUNIT);
					// Has to be done this way as one can't depend on the
					// seg frontsector assignments.
					int	side = P_PointOnLineSide(x, y, ld);
					sector_t *chk = side ? ld->backsector : ld->frontsector;

					if (!s) s = chk; else if (s != chk) return -1;

					ringtmp[ringtmpcnt++] = ld - lines;
					hitlinebak[ld - lines] = 1;
				}
			}
		}
	}
	return s ? s - sectors : -1;
}

// Can also be used to check whether a ring can be completed with a bridging line
static int checkForRingInList(byte *lnarray, int ringcount, int *ringlist, int lnum, int *bbox, int *vnum, int ht, RDBOOL sametype)
{
	line_t		*ld = &lines[lnum];
	vertex_t	*v1;
	RDBOOL		isclosed = false, repeat = true;
	int			mthi, mtlo;
	int			c, i;

	lnarray[lnum] = 2;

	if (sametype)
	{
		mtlo = lines[lnum].mtflags & R3D_MISLOTEX;
		mthi = lines[lnum].mtflags & R3D_MISHITEX;
	}

	if (vnum)
	{
		if (   vtxcheck[ld->v2 - vertexes] == 1
			|| vtxcheck[ld->v2 - vertexes] == 3
			|| vtxcheck[ld->v2 - vertexes] == 6)
		{
			v1 = ld->v1;
			vnum[0] = ld->v2 - vertexes;
		}
		else
		{
			v1 = ld->v2;
			vnum[0] = ld->v1 - vertexes;
		}
	}
	else	// don't care
		v1 = ld->v1;

	while (repeat)
	{
		repeat = false;

		for (c = i = 0; i < ringcount; i++)
		{
			int k = ringlist[i];

			if (lnarray[k] == 1)
			{
				line_t	*li = &lines[k];

				if (sametype)
				{
					if (li->mtflags & R3D_MISLOTEX && !mtlo) continue;
					if (li->mtflags & R3D_MISHITEX && !mthi) continue;
				}

				if (vnum && !c)
				{
					if (   vtxcheck[li->v1 - vertexes] == 3
						|| vtxcheck[li->v1 - vertexes] == 6
						|| vtxcheck[li->v2 - vertexes] == 3
						|| vtxcheck[li->v2 - vertexes] == 6)
					{
						continue;
					}
				}
				if (li->v1 == v1)
				{
					if (   ht == DDMININT
						|| li->frontsector->ceilingheight == ht
						|| (li->backsector && li->backsector->ceilingheight == ht))
					{
						v1 = li->v2;
						lnarray[k] = 2;
						c++; repeat = true;
					}
				}
				else if (li->v2 == v1)
				{
					if (   ht == DDMININT
						|| li->frontsector->ceilingheight == ht
						|| (li->backsector && li->backsector->ceilingheight == ht))
					{
						v1 = li->v1;
						lnarray[k] = 2;
						c++; repeat = true;
					}
				}
				if (vnum && repeat)
				{
					if (   vtxcheck[v1 - vertexes] == 3
						|| vtxcheck[v1 - vertexes] == 6)
					{
						vnum[1] = v1 - vertexes;
						return false;
					}
				}
			}
		}
	}

	isclosed = (v1 == ld->v2);

	if (isclosed)
	{
		M_ClearBox(bbox);
	}
	
	for (c = i = 0; i < ringcount; i++)
	{
		if (lnarray[ringlist[i]] == 2)
		{
			if (isclosed)
			{
				lnarray[ringlist[i]] = 4; c++;
				M_AddLineToBox(bbox, &lines[ringlist[i]]);
			}
			else
			{
				lnarray[ringlist[i]] = 1;
			}
		}
	}

	return c;
}

// Store loop line inf to lvar array
static void setRINF(int secnum, int ringcount, int *bbox, RDBOOL hi)
{
	loopvar_t	*lvptr;
	int			c, i;
	int			*rp = hi ? ringhi    : ringlo;
	int			rc  = hi ? ringhicnt : ringlocnt;

	//--------------------------------------

	if (rinf.count >= rinf.maxcount)
	{
		rinf.maxcount += 32;

		rinf.lvar = (loopvar_t*)W_ReallocCLR(rinf.lvar, rinf.maxcount * sizeof(loopvar_t), TEMPID, 0);
	}
	
	lvptr = &rinf.lvar[rinf.count++];

	// Store the info about the loop
	memcpy(lvptr->box, bbox, sizeof(lvptr->box));
	lvptr->hilo      = (byte)hi;
	lvptr->secnum    = secnum;
	lvptr->loopcount = ringcount;
	lvptr->area      = ((bbox[BOXRIGHT]>>16) - (bbox[BOXLEFT]>>16)) *
				       ((bbox[BOXTOP  ]>>16) - (bbox[BOXBASE]>>16));

	lvptr->looplist  = (int*)W_Malloc(ringcount * sizeof(int), TEMPID, 26);

	for (c = i = 0; i < rc; i++)
	{
		if (hitline[rp[i]] == 4)
		{
			if (c >= ringcount) Con_Error("Program Error: setRINF %i\n", ringcount);

			if (hi && !lvptr->ignore)
			{
				line_t *ld = &lines[rp[i]];

				if (   vtxcheck[ld->v1 - vertexes] == 4
					|| vtxcheck[ld->v2 - vertexes] == 4)
				{
					lvptr->ignore = true;
				}
			}

			hitline[rp[i]] = 8;

			lvptr->looplist[c++] = rp[i];
		}
	}
}

static int __cdecl AreaSorter(const void *e1, const void *e2)
{
	uint lc1 = ((loopvar_t*)e1)->area;
	uint lc2 = ((loopvar_t*)e2)->area;
	if (lc1 > lc2) return -1;
	if (lc1 < lc2) return  1;
	return 0;
}

static RDBOOL chkContainedSSINF(sector_t *sec, int lnum, byte *lnarray, int *inf_flags, RDBOOL testself)
{
	sector_t *s = lines[lnum].frontsector == sec ?
						lines[lnum].backsector : lines[lnum].frontsector;
	RDBOOL	skysec = sec->ceilingpic == skyflatnum;
	int		i, k, h1, h2;

	if (!s) return false;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (!ld || ld->validcount == validcount)
				continue;

			ld->validcount = validcount;

			if (lnarray[ld - lines] == 4 || !(ld->ml_flags & ML_TWOSIDED))
			{
				if (inf_flags)
				{
					if (    ld->ml_flags & ML_TWOSIDED
						&& (ld->frontsector->specFlags & SF_MOVSECTOR
						||	ld->backsector->specFlags  & SF_MOVSECTOR))
					{
						(*inf_flags) |= 4;
					}
					else if (testself)	// inner loop
					{
						if (   ld->ml_flags & ML_TWOSIDED
							&& !(origTEX[ld - lines])
							&& Vtxs2(ld))
						{
							(*inf_flags) |= 2;
						}
						else
						{
							(*inf_flags) |= 4;
						}
					}
					else
					{
						if ( !(ld->ml_flags & ML_TWOSIDED)
							|| origTEX[ld - lines]
							|| !Vtxs2(ld))
						{
							(*inf_flags) |= 1;
						}
					}
				}

				continue;
			}

			// Only invalid textures are allowed
			if ((origTEX[ld - lines] & 0x11) == 0x01
				&& ld->frontsector->floorheight
				>  ld->backsector->floorheight)
			{
				return false;
			}

			if ((origTEX[ld - lines] & 0x11) == 0x10
				&& ld->frontsector->floorheight
				<  ld->backsector->floorheight)
			{
				return false;
			}

			if (!testself) continue;

			if (ld->frontsector == sec || ld->backsector == sec)
				return false;

			h1 = ld->frontsector->floorheight;
			h2 = ld->backsector->floorheight;

			if (   origTEX[ld - lines] & 0x55
				&& MIN_OF(h1, h2) < sec->floorheight)
			{
				return false;
			}
			
			if (h1 != h2
				&& origTEX[ld - lines] & 0x11
				&& MIN_OF(h1, h2) != s->floorheight)
			{
				return false;
			}
		}

		if (skysec && ss->sector->ceilingpic == skyflatnum)
			continue;
		
		if (ss->sector->ceilingheight > sec->ceilingheight)
			return false;
	}

	return true;
}

static RDBOOL vtxLinesIterator2(int x, int y, vertex_t *v1, sector_t *sec)
{
	int			offset;
	const long	*list;
	
	if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
		return false;

	offset = y*bmapwidth + x;
	offset = *(blockmap + offset);
	list = blockmaplump + offset;

	// skip starting delimiter
	while( *(++list) >= 0)
	{
		line_t *ld = &lines[*list];

		if (ld->validcount == validcount)
			continue;		// line has already been checked

		ld->validcount = validcount;

		if (   hitline[ld - lines]
			&& (ld->v1 == v1 || ld->v2 == v1)
			&& ld->frontsector != sec
			&& ld->backsector  != sec)
			return true;
	}
	return false;
}

static RDBOOL findLineVertexConnected(line_t *ld, vertex_t *v1, sector_t *sec)
{
    int		xl, xh, yl, yh, bx, by;

	validcount++;
	ld->validcount = validcount;	//Don't test self

	M_getBlockOffsets(ld->bbox, 0, &xl, &xh, &yl, &yh);

	for (bx=xl ; bx<=xh ; bx++)
	{
		for (by=yl ; by<=yh ; by++)
		{
			if (vtxLinesIterator2(bx, by, v1, sec))
				return true;
		}
	}
	return false;
}

static void containedPASS1(sector_t *sec)
{
	sector_t	*s;
	line_t		*ld, *li;
	int			c0, c1, c2, lnum, rc;
	int			k, m, n;
	int			usefscs = 0;
	RDBOOL		done = false;
		
	CLR_hitline();

	for (c0 = c1 = c2 = k = 0; k < sec->linecount; k++)
	{
		ld = sec->lines[k];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			if (   ld->frontsector == ld->backsector
				|| ld->specType >= R3D_40
				|| twoSKY(ld, 1))
			{
				return;
			}

			if (!origTEX[ld - lines])
			{
				int cdiff, fdiff;

				if (!(ld->lnflags2 & LNF2_OWNEQUALTOTWO))
				{
					return;
				}

				s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				fdiff = s->floorheight - sec->floorheight;
				cdiff = s->ceilingheight - sec->ceilingheight;

				if ((fdiff > 0 && cdiff > 0) || (fdiff < 0 && cdiff < 0))
				{
					usefscs++;
				}

				if (  (s->specFlags & 0xC) != 0xC
					|| s->specFlags & SF_MOVCEIL)
				{
					return;
				}
				c0++;
			}
			else
				c1++;

			hitline[ld - lines] = 1;
		}
		else c2++;
	}

	if (c0 > 2 && c1 && c2 && (c1 + c2) > 2)
	{
		int		bbox[4];

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k]; lnum = ld - lines;

			if (hitline[lnum] != 1) continue;

			s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

			if (subfn_CheckForRing(hitline, s, lnum, &rc, bbox, NULL, NULL))
			{
				RDBOOL	loopdone = false;

				if (   LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, bbox, -1, MSIA_NONE)
					&& LS_SetSSINF(s, hitline, NULL, NULL, false, false))
				{
					RDBOOL movf = (s->specFlags & SF_MOVSECTOR) == SF_MOVFLOOR;

					// v220-6 check floorheights with a textureless contained sector
					if (   !ssinf.boundhashi
						&& !ssinf.boundhaslo
						&& (!ssinf.anylines || ssinf.anylines == ssinf.anysameseclines)
						&& (sec->specFlags & 0x3009C) == 0x10
						&& (s->specFlags & 0x3009C) == 0x1C
						&& s->floorheight - sec->floorheight > 1024*FRACUNIT)
					{
						s->ceilinglightsec = sec - sectors;
						done = loopdone = true;
						goto exit_loop;
					}

					if (   !ssinf.boundhashi
						&& !ssinf.boundhaslo
						&& (!ssinf.anylines || ssinf.anylines == ssinf.anysameseclines))
					{
						validcount++;
						
						for (m = 0; m < hitssecptr; m++)
						{
							subsector_t *ss = &subsectors[hitssec[m]];
							seg_t		*sg = &segs[ss->firstline];

							if (ssinf.anylines)
								SetOverride(ss, sec - sectors, 46004);
							else if (movf && ssinf.boundsmisshi == rc)
								Setr3dsscs(ss, sec - sectors, 46005);
							else if (usefscs == rc)
								Setr3dssfs(ss, sec - sectors, 46006);

							for (n = 0; n < ss->linecount; n++, sg++)
							{
								if ( !(li = sg->linedef) || li->validcount == validcount)
									continue;

								li->validcount = validcount;

								if (usefscs == rc)
								{
									if (movf && !origTEX[li - lines])
										Settype(li, R3D_14, 46007);
									else
										Settype(li, R3D_00, 46008);
								}
								else if (origTEX[li - lines] & 0x22)
								{
									Settype(li, R3D_03, 46009);
								}
								else if (li->backsector->specFlags & SF_MOVFLOOR
									||  li->frontsector->specFlags & SF_MOVFLOOR)
								{
									if (   !ssinf.anylines
										&& (li->backsector->specFlags & SF_MOVFLOOR)
										^ (li->frontsector->specFlags & SF_MOVFLOOR)
										&& !(li->frontsector->specFlags & SF_MOVCEIL)
										&& !(li->backsector->specFlags & SF_MOVCEIL)
										&& ((li->backsector->specFlags & SF_MOVFLOOR
										&&  IsLiquidFloorFlat(li->backsector)
										&&  (li->backsector->specFlags & 0xC) == 0xC)
										||  (li->frontsector->specFlags & SF_MOVFLOOR
										&&  IsLiquidFloorFlat(li->frontsector)
										&&  (li->frontsector->specFlags & 0xC) == 0xC)) )
									{
										Settype(li, R3D_16, 46010);
									}
									else if (!ssinf.anylines
										&& !(origTEX[li - lines] & 0x44)
										&& li->ml_flags & ML_TWOSIDED
										&& li->frontsector->ceilingheight
										!= li->backsector->ceilingheight
										&& !(li->frontsector->ceilingpic == skyflatnum
										&&   li->backsector->ceilingpic == skyflatnum) )
										Settype(li, R3D_16, 46011);
									else
										Settype(li, R3D_01, 46012);
								}
								else
								{
									Settype(li, R3D_16, 46013);
								}

								SetLFG(li, LFG_NEWSET|LFG_FIXED, 46014);
							}
						}
						done = loopdone = true;
					}
				}

exit_loop:;
				for (m = 0; m < s->linecount; m++)
				{	
					lnum = s->lines[m] - lines;

					if (hitline[lnum] == 4) hitline[lnum] = 8;
				}

				if (loopdone) Setsecdone3(s, 46015);
			}
		}
	}
	if (done)
	{
		for (k = 0; k < sec->linecount; k++)
		{
			lnum = sec->lines[k] - lines;

			if (!hitline[lnum])
			{
				Settype(sec->lines[k], R3D_00, 46016);
				SetLFG(sec->lines[k], LFG_NEWSET|LFG_FIXED, 46017);
			}
		}
		Setsecdone3(sec, 46018);
	}
}

//============================================================

// SKYFLOORS ======================

static void setSKYFloors(sector_t *sec)
{
	line_t		*ld;
	int			c0 = 0, c1 = 0;
	int			k, m, n, p;
	int			lnum, bbox[4];
		
	CLR_hitline();

	for (k = 0; k < sec->linecount; k++)
	{
		int		ht1, ht2;
		
		ld = sec->lines[k]; lnum = ld - lines;

		if (   lineflags[lnum] & (LFG_NEWSET|LFG_3DTYPE)
			|| ld->spflags & SPF_CONTROLLINE
			|| !(ld->ml_flags & ML_TWOSIDED))
		{
			continue;
		}

		if (!twoSKY(ld, 1)) return;

		if (ld->frontsector == ld->backsector)
		{
			c1++; continue;
		}
		
		ht1 = ld->frontsector->floorheight;
		ht2 = ld->backsector->floorheight;

		if (ht1 > ht2)
		{
			if (!sides[ld->sidenum[1]].bottomtexture
				&& ld->backsector->floorpic != skyflatnum)
			{
				c0++; hitline[lnum] = 1;
			}
		}
		if (ht1 < ht2)
		{
			if (!sides[ld->sidenum[0]].bottomtexture
				&& ld->frontsector->floorpic != skyflatnum)
			{
				c0++; hitline[lnum] = 1;
			}
		}
	}

	if (c0 < 3 || !c1) return;

	for (k = 0; k < sec->linecount; k++)
	{
		lnum = sec->lines[k] - lines;

		if (   hitline[lnum] == 1
			&& subfn_CheckForRing(hitline, sec, lnum, NULL, bbox, NULL, NULL))
		{
			sector_t *s = lines[lnum].frontsector == sec ?
				lines[lnum].backsector : lines[lnum].frontsector;

			for (m = 0; m < sec->linecount; m++)
			{
				ssline_t	*sl;
				
				ld = sec->lines[m]; lnum = ld - lines;

				sl = &sslines[lnum];

				for (n = 0; n < sl->num; n++)
				{
					subsector_t *ss = &subsectors[sl->list[n]];
					seg_t		*sg = &segs[ss->firstline];

					if (ss->sector != sec && ss->floorsec < 0)
						Setr3dssfs(ss, sec - sectors, 46019);

					for (p = 0; p < ss->linecount; p++, sg++)
					{
						if (sg->linedef && !origTEX[sg->linedef - lines])
						{
							Settype(sg->linedef, R3D_01, 46016);
							SetLFG(sg->linedef, LFG_FIXED|LFG_NEWSET, 46020);
						}
					}
				}

				if (hitline[lnum] == 4)
				{
					hitline[lnum] = 8;
					Settype(sec->lines[m], R3D_01, 46021);
					SetLFG(&lines[lnum], LFG_FIXED|LFG_NEWSET, 46022);
				}
			}
			Setsecdone3(s, 46023);
			Setsecdone3(sec, 46024);
		}
	}
}

//============================================================

static int getContainedSecnum(sector_t *sec, byte *lnarray)
{
	RDBOOL	isliq = IsLiquidFloorFlat(sec);
	int		i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];
		int		lnum = ld - lines;

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   ld->ml_flags & ML_TWOSIDED
			&& lnarray[lnum] == 4
			&& ld->lnflags2 & LNF2_OWNEQUALTOTWO)
		{
			sector_t *s = ld->frontsector == sec ?
				ld->backsector : ld->frontsector;

			if (!isliq && IsLiquidFloorFlat(s))
			{
				return s - sectors;
			}
			if (   isliq
				&& s->ceilingheight == sec->ceilingheight
				&& s->ceilingpic == sec->ceilingpic)
			{
				return s - sectors;
			}
		}
	}
	return -1;
}


static void checkContainingLoopsPASS1()
{
	sector_t	*sec;
	int			i;
	
	linememcnt = 0;

	// PASS 1 Check for contained types with an outer ring having 1's lines
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& sec->linecount > 5
			&& !(sec->specFlags & 0xC)		// Always false if sector has 1's lines
			&& sec->ceilingheight - sec->floorheight >= 4*FRACUNIT
			&& !(sec->specFlags & SF_MOVSECTOR)
			&& sec->heightsec < 0
			&& sec->stptr < 0)
		{
			containedPASS1(sec);
		}
	}

	// 2) Check skyfloors - these are quite specific when associated with
	//    skyflat ceilings and same sector lines
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (  !secdone[i]
			&& sec->ceilingpic == skyflatnum
			&& sec->floorpic   == skyflatnum
			&& sec->ceilingheight - sec->floorheight >= 4*FRACUNIT)
		{
			setSKYFloors(sec);
		}
	}
}

static RDBOOL secHeightCheck(line_t *ld)
{
	return ld->frontsector == ld->backsector
		&& (abs(ld->frontsector->ceilingheight - ld->frontsector->floorheight) >= 1024*FRACUNIT
		||  abs(ld->backsector->ceilingheight  - ld->backsector->floorheight ) >= 1024*FRACUNIT);
}

static void RINF_SetArray()
{
	sector_t	*sec;
	int			cht, fht, diff, bbox[4];
	int			ahi, alo, ahicnt, alocnt;
	int			c, i, k;
	RDBOOL		movflr, movclg;
	RDBOOL		skyhi, skylo;
	RDBOOL		liqfloormovetest;

	memset(&rinf, 0, sizeof(ringinf_t));

	// 1) Mark all lines in the sectors with compliant missing textures
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   secdone[i]
			|| eliminatesec[i]
			|| sec->linecount < 3
			|| sec->heightsec >= 0
			|| sec->stptr >= 0)
		{
			continue;
		}

		skyhi  = sec->ceilingpic == skyflatnum;
		cht    = sec->ceilingheight;
		fht    = sec->floorheight;

		if ((sec ->specFlags & 0xC) == 0xC
			&& skyhi
			&& !IsLiquidFloorFlat(sec) // $v67
			&& cht - fht > MAX_SECHEIGHT)
		{
			continue;
		}

		if (   cht - fht < 4*FRACUNIT
			|| (skyhi && cht - fht > 1600*FRACUNIT))
		{
			continue;
		}

		skylo  = sec->floorpic == skyflatnum;
		movflr = (sec->specFlags & SF_MOVFLOOR) > 0;
		movclg = (sec->specFlags & SF_MOVCEIL ) > 0;

		CLR_hitline();
		CLR_vtxcheck();

		ringhicnt = ringlocnt = 0;

		ahi = -1, ahicnt = 0;	// used to allow one missing tex 
		alo = -1, alocnt = 0;
		
		liqfloormovetest = sec->specFlags & SF_MOVFLOOR && IsLiquidFloorFlat(sec);

		for (k = 0; k < sec->linecount; k++)
		{
			sector_t	*s;
			int			sid;
			int			flags;
			int			limit;
			line_t		*ld = sec->lines[k];
			RDBOOL		vtfvalid = vtfcount[ld->v1 - vertexes] == 2
								&& vtfcount[ld->v2 - vertexes] == 2;

			if (   !(ld->ml_flags & ML_TWOSIDED)
			//	|| subfn_FindCoincidentLine(ld, 0)		// $v62 test added
				|| lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)
				|| ld->specType >= R3D_40
				|| ld->spflags & SPF_CONTROLLINE
				|| ld->frontsector == ld->backsector
				|| (!Vtfl3(ld) && !vtfvalid))
			{
				continue;
			}

			sid = sec == ld->frontsector ? 0 : 1;
			s  = sid ? ld->frontsector : ld->backsector;

			// Test added to ensure order of processing is correct $v73
			if (liqfloormovetest && (s->specFlags & 0x3001C) == 0x1C)
			{
				continue;
			}

			flags = LS_TestLineTex(ld);

			// $v65 test added
			if ((s ->specFlags & 0x3000C) == 0xC
				&& s->ceilingpic == skyflatnum
				&& s->ceilingheight - s->floorheight > MAX_SECHEIGHT)
			{
				SetLFG(ld, LFG_NEWSET|LFG_FIXED, 60001);
				continue;
			}

			if (   (s->ceilingheight - s->floorheight < 4*FRACUNIT && !(flags & LT_MISLO))
				|| (s->ceilingpic == skyflatnum && s->ceilingheight - s->floorheight > 1600*FRACUNIT)
				|| (movclg && s->specFlags & SF_MOVCEIL)
				|| s->heightsec >= 0)
			{
				continue;
			}

			if (   (diff = cht - s->ceilingheight) > 8*FRACUNIT
				&& !((skyhi || skylo) && s->ceilingpic == skyflatnum))
			{
				if (!sides[ld->sidenum[sid]].toptexture)
				{
					vtxcheck[ld->v1 - vertexes]++;
					vtxcheck[ld->v2 - vertexes]++;
					ringhi[ringhicnt++] = ld - lines;
					hitline[ld - lines] = 1;
				}
				else
				{
					if ((origTEX[ld - lines] & 0x44) != 0x44)
					{
						ahi = ld - lines;
						ahicnt++;
					}
				}
			}

			limit = s->specFlags & SF_MOVFLOOR ? 7*FRACUNIT : 8*FRACUNIT;	// $v73

			if (   (diff = fht - s->floorheight) < -limit
				&& !(skylo && s->floorpic == skyflatnum))
			{
				if (!sides[ld->sidenum[sid]].bottomtexture)
				{
					ringlo[ringlocnt++] = ld - lines;
					hitline[ld - lines] = 1;
				}
				else
				{
					if ((origTEX[ld - lines] & 0x11) != 0x11)
					{
						alo = ld - lines;
						alocnt++;
					}
				}
			}
		}
		
		if (ringhicnt < 3 && ringlocnt < 3) continue;

		if (ringhicnt)
		{
			// Check for a single texture - the doom engine ignores only 1 set
			if (ahicnt == 1) { hitline[ahi] = 1; ringhi[ringhicnt++] = ahi; }

			for (k = 0; k < ringhicnt; k++)
			{
				if (   hitline[ringhi[k]] == 1
					&& (c = checkForRingInList(hitline, ringhicnt, ringhi, ringhi[k], bbox, NULL, DDMININT, false)) )
				{
					setRINF(i, c, bbox, 1);
				}
			}
		}

		if (ringlocnt)
		{
			// Check for a single texture - the doom engine ignores only 1 set
			if (alocnt == 1) { hitline[alo] = 1; ringlo[ringlocnt++] = alo; }

			for (k = 0; k < ringlocnt; k++)
			{
				if (   hitline[ringlo[k]] == 1
					&& (c = checkForRingInList(hitline, ringlocnt, ringlo, ringlo[k], bbox, NULL, DDMININT, false)) )
				{
					setRINF(i, c, bbox, 0);
				}
			}
		}
	}

	// sort by area
	if (rinf.count) qsort(rinf.lvar, rinf.count, sizeof(loopvar_t), AreaSorter);
}

static RDBOOL eliminateSUBSSECS(sector_t *sec, int linecnt, RDBOOL movfloor, RDBOOL hi, int primary_secnum, int callnum)
{
	int		i, k, tempptr, bbox[4];
	RDBOOL	secsent = (sec != NULL);
	RDBOOL	linesfound = false;

	if (linecnt < 3) return false;

	memcpy(tempsseclist, hitssec, hitssecptr * sizeof(int));
	tempptr = hitssecptr;

	CLR_hitline();

	for (i = 0; i < linecnt; i++) hitline[ringtmp[i]] = 1;

	for (i = 0; i < linecnt; i++)
	{
		if (   hitline[ringtmp[i]] == 1
			&& checkForRingInList(hitline, linecnt, ringtmp, ringtmp[i], bbox, NULL, DDMININT, false))
		{
			int			flag = MSIA_NONE;
			int			count;
			RDBOOL		exclines = false;
			int			mask = 4;

			if (!secsent)
			{
				int	txn = origTEX[ringtmp[i]];

				sec = lines[ringtmp[i]].frontsector;

				if (txn & 0x11) mask |= 1;
				if (txn & 0x44) mask |= 2;
			}
			else
			{
				mask |= hi ? 2 : 1;
			}

			for (count = k = 0; k < linecnt; k++)
			{
				if (hitline[ringtmp[k]] == 4)
				{
					linelist[count++] = ringtmp[k];
				}
			}

			if (!LS_MarkSubsecsInLoop(sec, count, linelist, bbox, MSIA_TESTAREA, NULL))
				flag = MSIA_REVERSED;

			// This should not fail
			if (LS_MarkSubsecsInLoop(sec, count, linelist, bbox, flag, NULL))
			{
				RDBOOL rctest = callnum == 1
								&& flag == MSIA_NONE
								&& !hi
								&& primary_secnum >= 0
								&& movfloor
								&& sec->ceilingpic == skyflatnum;

				// Check for special case where exclusion is not wanted
				if (rctest)
				{
					LS_SetSSINF(sec, hitline, NULL, NULL, false, false);

					if (   ssinf.boundshastex == 0x01
						&& ssinf.anylines
						&& ssinf.anylines == ssinf.anysameseclines
						&& ssinf.anysamemovef
						&& !ssinf.ownringcnt
						&& !ssinf.lotex)
					{
						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];

							// v220-10 if a textureless sec is present check the floor
							//         is set to the the containing sector.
							if (   ss->sector != sec
								&& !(sec->specFlags & SF_MOVFLOOR)
								&& (ss->sector->specFlags & 0x40C) == 0xC)
							{
								Setr3dssfs(ss, sec - sectors, 46031);
							}

							Setr3dsslfsec(ss, primary_secnum, 46031);

							subsectors[hitssec[k]].ssflags |= SSF_RENDLFSEC;	// v220-4
						}
					}
				}

				for (k = 0; k < hitssecptr; k++)
				{
					excludessecs[hitssec[k]] |= mask;
				}

				exclines = linesfound = true;
			}

			for (k = 0; k < linecnt; k++)
			{
				if (hitline[ringtmp[k]] == 4)
				{
					hitline[ringtmp[k]] = 8;
					if (exclines) excludeline[ringtmp[k]] = true;
				}
			}
		}
	}

	hitssecptr = tempptr;
	memcpy(hitssec, tempsseclist, hitssecptr * sizeof(int));

	return linesfound;
}

static int setLineExclude(sector_t *sec, RDBOOL hi)
{
	line_t	*ld;
	int		i, k, diff, linecnt = 0;

	validcount++;

	// Set up ringtmp for any internal rings enclosing
	// subsecs that my need to be excluded.
	for (i = 0; i < hitssecptr; i++)
	{
		sector_t	*fs, *bs;
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (excludessecs[hitssec[i]] & 4) continue;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (!(ld = sg->linedef) || ld->validcount == validcount)
				continue;

			ld->validcount = validcount;

			if (   ld->ml_flags & ML_TWOSIDED
				&& (fs = ld->frontsector) != (bs = ld->backsector)
				&& (fs == sec || bs == sec))
			{
				if (hi)
				{
					diff = fs->ceilingheight - bs->ceilingheight;

					if (!(fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum))
					{
						if (   (diff > 0 && sides[ld->sidenum[0]].toptexture)
							|| (diff < 0 && sides[ld->sidenum[1]].toptexture))
						{
							ringtmp[linecnt++] = ld - lines;
						}
					}
				}
				else
				{
					diff = fs->floorheight - bs->floorheight;

					if (!(fs->floorpic == skyflatnum && bs->floorpic == skyflatnum))
					{
						if (   (diff > 0 && sides[ld->sidenum[1]].bottomtexture)
							|| (diff < 0 && sides[ld->sidenum[0]].bottomtexture))
						{
							ringtmp[linecnt++] = ld - lines;
						}
					}
				}
			}
		}
	}
	return linecnt;
}

// v220-9 check that the line ring is a legal map construction
static RDBOOL chkLegal(sector_t *s, loopvar_t *lvar, int lvcount, int anytex)
{
	sector_t	*sc, *os = NULL;
	line_t		*li;
	int			fh[] = { DDMAXINT, DDMAXINT, DDMAXINT };
	int			c, d, i, k;

	// First check the boundary lines
	for (c = d = i = 0; i < lvar->loopcount; i++)
	{		
		if (origTEX[lvar->looplist[i]] & 0x55) continue;

		li = &lines[lvar->looplist[i]];
		sc = li->frontsector == s ? li->backsector : li->frontsector;

		if (sc)
		{
			if (!anytex)
			{
				if ( !(li->frontsector->specFlags & 0x10004)
					|| li->frontsector->specFlags & 0x400)
				{
					if (!os) os = li->frontsector, d++;
					else if (os == li->frontsector) d++;
				}

				if (   li->backsector	// should be
					&& !(li->backsector->specFlags & 0x10004))
				{
					if (!os) os = li->backsector, d++;
					else if (os == li->backsector) d++;
				}
			}

			for (k = 0; k < 3; k++)
			{
				if (fh[k] == sc->floorheight) goto next_line;
			}

			for (k = 0; k < 3; k++)
			{
				if (fh[k] == DDMAXINT) { fh[k] = sc->floorheight; c++; break; }
			}
		}
next_line:;
	}

	if (c < 2) return true;

	if (os && lvar->loopcount == d)
	{
		validcount++;

		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];

			if ((ss->sector->specFlags & 0xC) == 0xC)
			{
				seg_t *sg = &segs[ss->firstline];

				for (k = 0; k < ss->linecount; k++, sg++)
				{
					if ( (li = sg->linedef)
						&& li->validcount != validcount)
					{
						li->validcount = validcount;
						Settype(li, R3D_3F, 46000);
						SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 46197);
					}
				}

				SetOverride(ss, os - sectors, 46197);
			}
		}
	}
	else
	{
		for (c = i = 0; i < lvar->loopcount; i++)
		{
			line_t	*li = &lines[lvar->looplist[i]];
					
			if (li->specType != R3D_00)
				Settype(li, R3D_00, 46198);

			SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 46198);
		}
	}

	return false;
}

static void setRINF_SSINF(sector_t *sec, loopvar_t *lvar, int lvcount, sector_t *os, RDBOOL primary)
{
	sector_t	*s = NULL, *fs, *bs, *inner = NULL;
	line_t		*li;
	int			i, k;
	int			lnum;
	int			linecnt = 0;
	RDBOOL		movfloor = (sec->specFlags & SF_MOVFLOOR) != 0;
	RDBOOL		test = true;
	// v220-4
	RDBOOL		liqflr = os ? IsLiquidFloorFlat(os) : false;
	fixed_t		csflrht = os ? os->floorheight : DDMAXINT;
	int			chkflrht = DDMAXINT;

	//----------------------------

	memset(&rssinf, 0, sizeof(rssinf_t));

	validcount++;

	// First check the boundary lines
	for (i = 0; i < lvar->loopcount; i++)
	{
		li = &lines[lvar->looplist[i]];

		if (primary) excludeline[li-lines] = true;

		if (li->ml_flags & ML_TWOSIDED)
		{
			int flags = LS_TestLineTex(li);

			if (origTEX[li - lines] & 0x22) rssinf.boundsmidtex++;
			if (flags & LT_HI) rssinf.boundshitex++;
			if (flags & LT_LO) rssinf.boundslotex++;
		}
		
		fs = li->frontsector; bs = li->backsector;

		if (test)
		{
			if (lvar->hilo)
			{
				if (!s) s = fs->ceilingheight > bs->ceilingheight ? fs : bs;
				else
				if (s != (fs->ceilingheight > bs->ceilingheight ? fs : bs))
				{ s = NULL; test = false; }
			}
			else
			{
				if (!s) s = fs->floorheight > bs->floorheight ? fs : bs;
				else
				if (s != (fs->floorheight > bs->floorheight ? fs : bs))
				{ s = NULL; test = false; }
			}
		}

		li->validcount = validcount;
	}

	if (!s) return;

	// Set ringtmp - used by eliminateSUBSSECS
	linecnt = setLineExclude(s, lvar->hilo);

	if (!eliminateSUBSSECS(s, linecnt, movfloor, lvar->hilo, -1, 0))
	{
		int txn;

		linecnt = 0; validcount++;

		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (!(li = sg->linedef) || li->validcount == validcount)
					continue;

				li->validcount = validcount;

				if (!(li->ml_flags & ML_TWOSIDED) || excludeline[li - lines])
				{
					continue;
				}

				if (   Vtxs2(li)
					&& !Vtfl(li)
					&& !(li->spflags & SPF_CONTROLLINE)
					&& (txn = origTEX[li - lines]) & 0x55
					&& li->frontsector != li->backsector)
				{
					ringtmp[linecnt++] = li - lines;
				}

				if (li->frontsector->specFlags & SF_MOVSECTOR)
					rssinf.boundmoveflags |= (fs->specFlags & 0x730000);
				if (li->backsector->specFlags & SF_MOVSECTOR)
					rssinf.boundmoveflags |= (fs->specFlags & 0x730000);
			}
		}

		if (linecnt > 2)
		{
			int primary_secnum = -1;

			// First part of check for movsec in movsec
			// Second part done in eliminateSUBSSECS()
			if (primary && movfloor && !lvcount)
			{
				line_t		*l = &lines[lvar->looplist[0]];
				sector_t	*s = l->frontsector - sectors == lvar->secnum ?
												l->backsector : l->frontsector;

				if (   s && !(s->specFlags & SF_MOVSECTOR)
					&& s->floorheight > sectors[lvar->secnum].floorheight)
				{
					primary_secnum = s - sectors;
				}
			}

			rssinf.eliminate = eliminateSUBSSECS(NULL, linecnt, movfloor, lvar->hilo, primary_secnum, 1);
		}
	}

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (chkflrht == DDMAXINT && csflrht > ss->sector->floorheight + 8*FRACUNIT)
			chkflrht = ss->sector->floorheight;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			int txn, flags;

			if (!(li = sg->linedef) || li->validcount == validcount)
				continue;

			li->validcount = validcount;
			lnum = li - lines;

			if ( !(li->ml_flags & ML_TWOSIDED)  ) continue;

			txn = origTEX[lnum];

			fs = li->frontsector; bs = li->backsector;

			if (liqflr && fs != os && bs != os)	// The containing sector
			{
				if (ss->sector->floorheight == chkflrht && fs->floorheight == bs->floorheight)
				{
					rssinf.lowfloorsequal1++;
				}
				else if (!txn && (fs->specFlags & SF_MOVFLOOR)^(bs->specFlags & SF_MOVFLOOR))
				{
					// Check textureless excluded lines with a mov floor
					rssinf.lowfloorsequal2++;
				}
			}

			if (!primary && excludeline[lnum]) continue;

			if (bs)
			{
				rssinf.totalinside++;

				if (fs->floorheight > sec->floorheight)
				{
					inner = fs;
				}
				else if (bs->floorheight > sec->floorheight)
				{
					inner = bs;
				}
				else inner = NULL;
			}

			if (excludeline[lnum])
			{
				if (txn & 0x11) rssinf.markhilo |= 1;
				if (txn & 0x44) rssinf.markhilo |= 2;

				continue;
			}

			rssinf.anylines++;

			if (fs == bs)
			{
				rssinf.sameseclines++; continue;
			}

			if (   fs->ceilingpic == skyflatnum
				&& bs->ceilingpic == skyflatnum
				&& (fs->specFlags & SF_MOVCEIL)^(bs->specFlags & SF_MOVCEIL)
				&& !(fs->specFlags & SF_MOVFLOOR)
				&& !(bs->specFlags & SF_MOVFLOOR)
				&& !IsLiquidFloorFlat(fs)
				&& !IsLiquidFloorFlat(bs))
			{
				rssinf.movskyceil++;
			}

			if (!(txn & 0x44) && lvar->hilo)
			{
				if (   (lvar->secnum == fs - sectors && bs->specFlags & SF_NOHITEX)
					|| (lvar->secnum == bs - sectors && fs->specFlags & SF_NOHITEX) )
				{
					rssinf.nohitexinside++;
				}
			}

			if (   !txn
				&& !(fs->specFlags & SF_MOVSECTOR)
				&& !(bs->specFlags & SF_MOVSECTOR)
				&& fs->ceilingheight == bs->ceilingheight)
			{
				sector_t	*s;
				int			fdiff;

				if ( (fdiff = fs->floorheight - bs->floorheight) == 0)
				{
					rssinf.sameheightlines++; continue;
				}

				s = fdiff > 0 ? fs : bs;

				if (   Vtxs2(li)
					&& IsLiquidFloorFlat(s))
				{
					rssinf.hiliqfinside++;
					rssinf.hiliqfsec = s - sectors;
					continue;
				}
			}

			flags = LS_TestLineTex(li);

			if (flags & LT_HI)
			{
				rssinf.hitex++;

				if (rssinf.hiringcnt >= rssring.himaxcnt)
				{
					rssring.himaxcnt += 32;
					rssring.hiring = (int*)W_Realloc(rssring.hiring, rssring.himaxcnt * sizeof(int), STATICID, 0);
				}
				rssring.hiring[rssinf.hiringcnt++] = lnum;
			}

			if (flags & LT_LO)
			{
				rssinf.lotex++;

				if (rssinf.loringcnt >= rssring.lomaxcnt)
				{
					rssring.lomaxcnt += 32;
					rssring.loring = (int*)W_Realloc(rssring.loring, rssring.lomaxcnt * sizeof(int), STATICID, 0);
				}
				rssring.loring[rssinf.loringcnt++] = lnum;
			}
			else if (inner)
			{
				if (fs == inner || bs == inner)
					rssinf.setallownsec++;
			}

			if (origTEX[li - lines] & 0x22) rssinf.midtex++;

			if (origTEX[lnum])
				rssinf.uncompliantlines++;
			else if (li->frontsector != sec && li->backsector != sec)
				rssinf.notexlines++;
		}
	}
}

static void SetSECDONE(line_t *ld, int num)
{
	if (!secdone[ld->frontsector - sectors]) Setsecdone3(ld->frontsector, 46032);
	if (ld->backsector
		&& !secdone[ld->backsector - sectors]) Setsecdone3(ld->backsector, 46033);
}

static RDBOOL checkfsegComplies(subsector_t *ss)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		line_t *ld = sg->linedef;

		if (ld && (origTEX[ld - lines] & 0x55) == 0x55)
		{
			if (sg->sidedef->midtexture) return true;
		}
	}
	return false;
}

typedef struct {
	int	boundslotex;
	int	boundshitex;
	int	anytex;
	int	lotex;
	int	hitex;
	int	inlotex;
	int	inhitex;
	int	samesec;
	int	movesec;
} baseinf_t;

static baseinf_t	baseinf;

typedef struct {
	// 0 - used for lines with same sec as primary ring lines
	// 1 - used for lines in other sectors
	int		*ln[2];
	int		lncnt[2];
	int		lnmaxcnt[2];

} innerinf_t;

static innerinf_t	innerinf;

static void setInner(line_t *ld, int idx)
{
	if (origTEX[ld - lines] & 0x11)
		baseinf.inlotex++;
	if (origTEX[ld - lines] & 0x44)
		baseinf.inhitex++;

	if (innerinf.lncnt[idx] >= innerinf.lnmaxcnt[idx])
	{
		innerinf.lnmaxcnt[idx] += 64;
		innerinf.ln[idx] = (int*)W_Realloc(innerinf.ln[idx], innerinf.lnmaxcnt[idx] * sizeof(int), STATICID, 0);
	}
	innerinf.ln[idx][innerinf.lncnt[idx]++] = ld - lines;
}

static int setBaseinf(loopvar_t *lvar, loopvar_t *lvmem, int lvcount)
{
	subsector_t *ss;
	sector_t	*sc, *fs, *bs;
	line_t		*ld;
	seg_t		*sg;
	int			msec;
	int			i, k;
	int			count = 0;

	linememcnt = 0;

	validcount++;

	memset(&baseinf, 0, sizeof(baseinf_t));

	innerinf.lncnt[0] = innerinf.lncnt[1] = 0;

	// Set the outer ring values
	for (i = 0; i < lvar->loopcount; i++)
	{
		ld = &lines[lvar->looplist[i]];

		ld->validcount = validcount;

		baseinf.anytex |= origTEX[ld - lines];

		linememory[linememcnt++] = ld - lines;

		if (origTEX[ld - lines] & 0x11) baseinf.boundslotex++;
		else
		if (origTEX[ld - lines] & 0x44) baseinf.boundshitex++;
	}

	// Set the inner ring values
	for (i = 0; i < lvcount; i++)
	{
		loopvar_t *lpv = &lvmem[i];

		for (k = 0; k < lpv->loopcount; k++)
		{
			ld = &lines[lpv->looplist[k]];

			baseinf.anytex |= origTEX[ld - lines];

			ld->validcount = validcount; count++;

			linememory[linememcnt++] = ld - lines;

			if ( lpv->hilo && origTEX[ld - lines] & 0x11) baseinf.lotex++;
			else
			if (!lpv->hilo && origTEX[ld - lines] & 0x44) baseinf.hitex++;
		}
	}

	sc = &sectors[lvar->secnum];
	msec = sc->tag > 0 && (sc->specFlags & 0x3001C) == 0x1001C;

	// Check if anything else is inside
	for (i = 0; i < hitssecptr; i++)
	{
		if (lvar->inside && excludessecs[hitssec[i]]) continue;

		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			ld = sg->linedef;

			if (!ld || ld->validcount == validcount) continue;

			ld->validcount = validcount; count++;

			baseinf.anytex |= origTEX[ld - lines];

			linememory[linememcnt++] = ld - lines;

			fs = ld->frontsector; bs = ld->backsector;

			if (fs == bs) baseinf.samesec++;
			else
			// v220-18 test added
			if (   bs
				&& msec
				&& sc->tag == fs->tag
				&& fs->tag == bs->tag
				&& (lvar->secnum == fs - sectors || lvar->secnum == bs - sectors)
				&& fs->floorheight == bs->floorheight
				&& fs->ceilingheight == bs->ceilingheight
				&& (fs->specFlags & 0x3001C) == 0x1001C
				&& (bs->specFlags & 0x3001C) == 0x1001C)
			{
				baseinf.movesec++;
			}

			if (fs == sc || bs == sc) setInner(ld, 0); else setInner(ld, 1);
		}
	}
	return count;
}

static void setLines(loopvar_t *lvar, RDBOOL setfloor, RDBOOL boundstex, int num)
{
	int		i, count = lvar->loopcount;
	byte	settype = setfloor ? boundstex ? R3D_00 : R3D_01 : R3D_16;

	for (i = 0; i < count; i++)
	{
		line_t *ld = &lines[lvar->looplist[i]];

		Settype(ld, settype, num);
		SetLFG (ld, LFG_NEWSET|LFG_FIXED, num);
	}

	for (i = 0; i < innerinf.lncnt[0]; i++)
	{
		line_t *ld = &lines[innerinf.ln[0][i]];

		if (!origTEX[ld - lines]) Settype(ld, R3D_01, num);
		SetLFG (ld, LFG_NEWSET|LFG_FIXED, num);
	}
}

typedef struct{
	subsector_t **list;
	int			sscnt;
	int			ssmaxcnt;
} sslist_t;

static sslist_t sslist;

static int subfn_setSSEClist()
{
	int	i;

	sslist.sscnt = 0;

	for (i = 0; i < hitssecptr; i++)
	{
		if (sslist.sscnt >= sslist.ssmaxcnt)
		{
			sslist.ssmaxcnt += 32;
			sslist.list = (subsector_t**)W_Realloc(sslist.list, sslist.ssmaxcnt * sizeof(subsector_t*), STATICID, 0);
		}
		sslist.list[sslist.sscnt++] = &subsectors[hitssec[i]];
	}

	return sslist.sscnt;
}

// b211-0
static void chkSharedLineSSECS(subsector_t *ss, int osec)
{
	seg_t	*sg = &segs[ss->firstline];
	short	floorpic = sectors[osec].floorpic;
	int		ht = sectors[osec].floorheight;
	int		i, k;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (   sg->linedef
			&& !origTEX[sg->linedef - lines]
			&& sg->linedef->mtflags & (R3D_MISLTEX1|R3D_MISLTEX2))
		{
			ssline_t *sl = &sslines[sg->linedef - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *chk = &subsectors[sl->list[k]];

				if (chk->sector != ss->sector
					&& !(chk->sector->specFlags & SF_MOVSECTOR)
					&& chk->floorsec < 0
					&& chk->lfsec < 0
					&& chk->sector->floorpic == floorpic
					&& chk->sector->floorheight == ht)
				{
					Setr3dsslfsec(chk, chk->sector - sectors, 46190);
					Setr3dssfs(chk, ss->sector - sectors, 46191);
				}
			}
		}
	}
}

static RDBOOL allVtxs2(loopvar_t *lvar)
{
	int i;

	for (i = 0; i < lvar->loopcount; i++)
		if ( !Vtxs2(&lines[lvar->looplist[i]]) ) return false;
	return true;
}

static RDBOOL rejectSEC(sector_t *sec)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if ( !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
			|| !Vtxs2(ld)
			|| origTEX[ld - lines] & 0x22)
			return false;
	}

	return true;
}

// v220-10
static RDBOOL isLVARline(loopvar_t *lvar, int idx)
{
	int i;

	for (i = 0; i < lvar->loopcount; i++)
	{
		if (lvar->looplist[i] == idx) return true;
	}

	return false;
}

static RDBOOL boundBySSlines(loopvar_t *lvar)
{
	line_t	*ld;
	int		i, k, m;

	// check contained by same sector sided lines
	for (i = 0; i < lvar->loopcount; i++)
	{
		ssline_t *sl = &sslines[lvar->looplist[i]];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			seg_t		*sg = &segs[ss->firstline];

			for (m = 0; m < ss->linecount; m++, sg++)
			{
				if ( (ld = sg->linedef)
					&& !isLVARline(lvar, ld - lines)
					&& ld->frontsector != ld->backsector)
				{
					if (  !ld->backsector
						|| ld->backsector->floorheight   != ld->frontsector->floorheight
						|| ld->backsector->ceilingheight != ld->frontsector->ceilingheight)
					{
						return false;
					}
				}
			}
		}
	}

	return true;
}

// v220-5
static int SS_HasLotex(subsector_t *ss)
{
	line_t	*ld;
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->sidedef)
		{
			if (sg->sidedef->bottomtexture) return true;

			// v220-10 tex other side and movsec?
			if (   sg->linedef
				&& origTEX[sg->linedef - lines] & 0x11
				&& sg->backsector->specFlags & SF_MOVFLOOR)
			{
				return -1;
			}
		}

		if ( (ld = sg->linedef) && ld->backsector)
		{
			int		side = ld->frontsector == sg->sidedef->sector ? 1 : 0;
			side_t	*sd = &sides[ld->sidenum[side]];

			if (sd->bottomtexture && sd->sector->floorheight < sg->sidedef->sector->floorheight)
				return 1;

			if (   sd->bottomtexture
				&& !(sd->sector->specFlags & SF_MOVFLOOR)
				&& sg->sidedef->sector->specFlags & SF_MOVFLOOR)
			{
				return -1;
			}
		}
	}

	return 0;
}

// v220-5 If a subsector referencing s has seg frontsecs
//        not equal to s then return true.
static RDBOOL linesInside(sector_t *s)
{
	ssline_t *sl;
	int		 i, k, m;

	for (i = 0; i < s->linecount; i++)
	{
		line_t *li = s->lines[i];

		if (li->spflags & SPF_CONTROLLINE) continue;

		sl = &sslines[li - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];
			seg_t		*sg;

			if (ss->sector != s) continue;

			for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
			{
				if (sg->frontsector && sg->frontsector != s)
					return true;
			}
		}
	}

	return false;
}

static RDBOOL allSecsInsideAre0x1C(sector_t *in, sector_t *out)
{
	subsector_t *ss;
	seg_t		*sg;
	int			i, k;

	//------------------------------

	if (!out) return false;

	if (  (in->specFlags & 0x3001C) != 0x1C
		|| out->ceilingheight > in->ceilingheight
		|| out->floorheight   < in->floorheight)
	{
		return false;
	}

	if (   out->ceilingheight != in->ceilingheight
		&& out->floorheight != in->floorheight)
	{
		return false;
	}

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t *ld = sg->linedef;

			if (!ld || ld->frontsector == ld->backsector) continue;
			if (!ld->backsector) return false;
			if (ld->frontsector != out && (ld->frontsector->ceilingheight != in->ceilingheight || (ld->frontsector->specFlags & 0x3001C) != 0x1C)) return false;
			if (ld->backsector  != out && (ld->backsector->ceilingheight  != in->ceilingheight || (ld->backsector->specFlags  & 0x3001C) != 0x1C)) return false;
		}
	}

	return true;
}

// v220-6
static RDBOOL chkLineType(line_t *li)
{
	if (!li->backsector || origTEX[li - lines] & 0x22) return true;
	if (li->frontsector->specFlags & SF_MOVFLOOR && li->mtflags & R3D_INVLTEX1) return true;
	if (li->backsector->specFlags  & SF_MOVFLOOR && li->mtflags & R3D_INVLTEX2) return true;
	return false;
}

static line_t *findBestLine(loopvar_t *lvar)
{
	sector_t	*sc = &sectors[lvar->secnum], *ch, *os1 = NULL, *os2 = NULL;
	line_t		*ld, *ld1 = NULL, *ld2 = NULL;
	int			a, b, i;

	for (a = b = i = 0; i < lvar->loopcount; i++)
	{
		ld = &lines[lvar->looplist[i]];
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (!ch) continue;

		if (!os1)
		{
			os1 = ch; ld1 = ld; a++;
		}
		else if (os1 == ch)
		{
			a++;
		}
		else if (!os2)
		{
			os2 = ch; ld2 = ld; b++;
		}
		else if (os2 == ch)
		{
			b++;
		}
	}

	if ( (ld = a > b ? ld1 : ld2) ) return ld;

	return &lines[lvar->looplist[0]];
}

static RDBOOL noHiTexInSector(sector_t *sc)
{
	line_t	*ld;
	int		i;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (!(ld->ml_flags & ML_TWOSIDED)) return false;

		if (origTEX[ld - lines] & 0x44)
		{
			byte mask = ld->frontsector == sc ? 0x04 : 0x40;

			if (origTEX[ld - lines] & mask) return false;
		}
	}

	return true;
}

static void setRINFssecs(loopvar_t *lvar, RDBOOL nofloorsec, RDBOOL noceilsec, loopvar_t *lvmem, RDBOOL missboth, int lvcount)
{
	sector_t *sec = &sectors[lvar->secnum];
	int		 fsecnum, csecnum, flag = MSIA_NONE;
	int		 i, seccnt;
	byte	 settype = 0xFF;
	RDBOOL	 valid;

	if (!LS_MarkSubsecsInLoop(sec, lvar->loopcount, lvar->looplist, lvar->box, MSIA_TESTAREA, NULL))
		flag = MSIA_REVERSED;

	// This should not fail
	if (LS_MarkSubsecsInLoop(sec, lvar->loopcount, lvar->looplist, lvar->box, flag, &seccnt))
	{
		line_t	 *ref = findBestLine(lvar);
		sector_t *srf = flag == MSIA_NONE ? sec : ref->frontsector == sec ? ref->backsector : ref->frontsector;
		int		 os = ref->frontsector - sectors == lvar->secnum ?
					ref->backsector - sectors : ref->frontsector - sectors;
		int		 msec;

		if (!(i = setBaseinf(lvar, lvmem, lvcount)) && !lvar->ignore && missboth) return;

		if (i && ((msec = !lvar->hilo && baseinf.movesec == i) || i == baseinf.samesec) )
		{
			if ((lvar->hilo || msec) && !lvcount && flag == MSIA_NONE && !baseinf.anytex && !lvar->ignore && !missboth)
			{
				for (i = 0; i < hitssecptr; i++)
				{
					if (msec)
					{
						Setr3dsslfsec(&subsectors[hitssec[i]], os, 46034);
						Setr3dsscs(&subsectors[hitssec[i]], os, 46034);
					}
					else
					{
						SetOverride(&subsectors[hitssec[i]], os, 46034);
					}
				}
				for (i = 0; i < linememcnt; i++)
				{
					Settype(&lines[linememory[i]], R3D_3F, 46034);
					SetLFG(&lines[linememory[i]], LFG_NEWSET|LFG_FIXED, 46034);
				}
				return;
			}
			
			if (i == baseinf.samesec) return;
		}

		if (   flag == MSIA_NONE
			&& missboth
			&& !(baseinf.anytex & 0x55)
			&& !lvcount
			&& sectors[os].specFlags & SF_MOVFLOOR
			&& (sectors[os].specFlags & 0xC) != 0xC
			&& (sectors[lvar->secnum].specFlags & (SF_MOVFLOOR|0xC)) == (SF_MOVFLOOR|0xC))
		{
			for (i = 0; i < hitssecptr; i++)
			{
				SetOverride(&subsectors[hitssec[i]], os, 46034);
			}

			for (i = 0; i < linememcnt; i++)
			{
				if (origTEX[linememory[i]])
					Settype(&lines[linememory[i]], R3D_03, 46035);
				else
					Settype(&lines[linememory[i]], R3D_3F, 46035);
				SetLFG(&lines[linememory[i]], LFG_NEWSET|LFG_FIXED, 46035);
			}

			return;
		}
		else
		if (   !lvcount
			&& !lvar->hilo
		//	&& flag == MSIA_NONE
		//	&& sec->specFlags & SF_MOVFLOOR
			&& !baseinf.lotex
			&& !baseinf.hitex
			&& !baseinf.inlotex
			&& !baseinf.inhitex
			&& baseinf.boundslotex > 1
			&& innerinf.lncnt[0])
		{
			int cnt = subfn_setSSEClist();

			for (i = 0; i < cnt; i++)
			{
				Setr3dssfs(sslist.list[i], lvar->secnum, 46036);
				Setr3dsslfsec(sslist.list[i], os, 46036);
			}

			setLines(lvar, true, baseinf.boundslotex > 1 || baseinf.boundshitex > 1, 46037);

			if (lvcount)
			{
				for (i = 0; i < lvcount; i++)
				{
					if (lvmem[i].inside)
						setLines(&lvmem[i], true, false, 46038); 
				}
			}

			return;
		}

		setRINF_SSINF(sec, lvar, lvcount, &sectors[os], true);

		if (!chkLegal(srf, lvar, lvcount, rssinf.notexlines)) return;

		// Check for types that shouldn't be set
		if (!lvcount && flag == MSIA_REVERSED && !rssinf.totalinside)
		{
			RDBOOL doset = lvar->hilo;

			if (!doset)
			{
			//	line_t		*ref = &lines[lvar->looplist[0]];
				sector_t	*s1  = &sectors[lvar->secnum];
				sector_t	*s2  = ref->frontsector == s1 ?
					ref->backsector : ref->frontsector;

				doset = !(s2->specFlags & SF_MOVFLOOR) && s1->specFlags & SF_MOVFLOOR;
			}

			// Illegal structure
			if (doset)
			{
				for (i = 0; i < lvar->loopcount; i++)
				{
					line_t *li = &lines[lvar->looplist[i]];
						
					Settype(li, R3D_00, 46039);
					SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 46039);
				}
				return;
			}
		}
		
		valid =	   !lvar->nested
				|| !rssinf.anylines
				|| (lvar->hilo && rssinf.anylines == rssinf.nohitexinside)
				|| rssinf.sameseclines
				|| rssinf.sameheightlines
				|| rssinf.notexlines
				|| rssinf.boundmoveflags & SF_MOVFLOOR
				|| ( lvar->hilo && rssinf.loringcnt && !rssinf.hiringcnt)
				|| (!lvar->hilo && rssinf.hiringcnt && !rssinf.loringcnt);


		if (!valid)
		{
			if (   !lvar->hilo
				&& lvar->nested
				&& !rssinf.eliminate
				&& rssinf.movskyceil == rssinf.anylines
				&& !rssinf.lotex
				&& !rssinf.hitex
				&& !(rssinf.boundmoveflags & SF_MOVSECTOR))
			{
				int i;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					Setr3dssfs(ss, os, 99999);
				}
			}
		}
		else
		{
		//	line_t	*ref = &lines[lvar->looplist[0]];
			RDBOOL	R3D_16type;
			RDBOOL	lowliqfloor = false;
			RDBOOL	chkfloortex = false;
			RDBOOL	chkceiltex  = false;
			RDBOOL	testtex =  !rssinf.uncompliantlines
							&& rssinf.anylines
							&& rssinf.sameseclines;
			int		lowfloorset = -1;
			byte	settype;

			//-----------------------------------------

			// v220-6 set/reset nested textureless loops.
			// Note: either the outer line's sector floorheights must be different
			//       or the ceilingheights but not both
			if (   flag == MSIA_NONE
				&& lvar->nested
				&& !lvar->inside
				&& allSecsInsideAre0x1C(sec, &sectors[os]))
			{
				// v220-10 since these usually need to inherit an adjacent
				//         ssec->floorsec value they are dealt with near the end.
				//         As such the code here has been removed (other than return).

				return;
			}

			if (lvar->hilo)
			{
				// Lowest ceiling
				sec = ref->frontsector->ceilingheight > ref->backsector->ceilingheight ?
						ref->backsector : ref->frontsector;

				if (sec->ceilingheight - sec->floorheight <= FRACUNIT) return;

				chkceiltex = testtex && !rssinf.hitex;
			}
			else
			{
				// Highest floor
				sec = ref->frontsector->floorheight < ref->backsector->floorheight ?
						ref->backsector : ref->frontsector;

				if (sec->ceilingheight - sec->floorheight <= FRACUNIT) return;

				chkfloortex = testtex && !rssinf.lotex;

			}

			if (flag == MSIA_REVERSED)
			{
				lowliqfloor = IsLiquidFloorFlat(&sectors[lvar->secnum]);
			}

			settype = R3D_16;	// default

			// This is set in two stages as, partly set, is used for test following.
			// If that test fails then it is fully set.
			R3D_16type   = !lowliqfloor
						&& !rssinf.anylines
						&& !rssinf.uncompliantlines
						&& !rssinf.sameseclines
						&& !rssinf.sameheightlines
						&& !rssinf.notexlines
						&& !( lvar->hilo && rssinf.loringcnt && !rssinf.hiringcnt)
						&& !(!lvar->hilo && rssinf.hiringcnt && !rssinf.loringcnt);

			if (!lvcount && lvar->ignore && R3D_16type && flag == MSIA_NONE && seccnt > 1)
			{
				Setr3dfh(&sectors[lvar->secnum], sec, 46043);
				settype = R3D_00;
			}
			else
			{
				if (R3D_16type)
				{
					if (flag == MSIA_NONE && seccnt > 1)
					{
						R3D_16type = false;
					}
					else
					{
						if (flag == MSIA_REVERSED
							&& sec->specFlags & SF_MOVFLOOR
							&& !rssinf.boundshitex
							&& !rssinf.boundslotex
							&& !rssinf.boundsmidtex)
						{
							settype = R3D_01;
						}
						else
							settype = 0xFF;
					}
				}
			}

			// Check for special movfloor cases
			if (   flag == MSIA_NONE
				&& !lvar->ignore
				&& !rssinf.boundsmidtex
				&& rssinf.setallownsec == rssinf.totalinside)
			{
				validcount++;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];
					seg_t		*sg = &segs[ss->firstline];
					int			k;

					for (k = 0; k < ss->linecount; k++, sg++)
					{
						line_t *li = sg->linedef;

						if (!li || li->validcount == validcount) continue;

						li->validcount = validcount;

						if (!lvar->hilo)
						{
							if (!(li->ml_flags & ML_TWOSIDED))
							{
								Settype(li, R3D_00, 46043);
							}
							else if (!(li->frontsector->specFlags & SF_MOVFLOOR)
								&& !(li->backsector->specFlags & SF_MOVFLOOR)
								&& abs(li->frontsector->floorheight - li->backsector->floorheight) <= 8*FRACUNIT)
							{
								Settype(li, R3D_00, 46044);
							}
							else
								Settype(li, R3D_16, 46043);
						}
						else
							Settype(li, R3D_16, 46044);

						SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 46044);
					}
				}
			}
			else if (   flag == MSIA_REVERSED
				&& sec->specFlags & SF_MOVFLOOR
				&& rssinf.notexlines
				&& !(rssinf.setallownsec == rssinf.anylines && !rssinf.boundslotex)
				&& !rssinf.markhilo)
			{
				sector_t *cs = NULL;
				int		 csec;

				if (sec->ceilingpic == skyflatnum)
				{
					int ht = sec->ceilingheight;

					for (i = 0; i < hitssecptr; i++)
					{
						subsector_t *ss = &subsectors[hitssec[i]];

						if (ss->sector->ceilingheight == ht
							&& ss->sector->ceilingpic != skyflatnum)
						{
							cs = ss->sector; ht = ss->sector->ceilingheight;
						}
					}
				}

				csec = cs ? cs - sectors : -1;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					Setr3dssfs(ss, lvar->secnum, 46045);
					if (SecCtrl(ss->sector - sectors))
					{
						Setr3dsslfsec(ss, sec - sectors, 46045);
						Setr3dsslcsec(ss, csec, 46045);
					}
				}
			}
			else
			if (!R3D_16type || rssinf.markhilo)
			{
				sector_t	*fsec = sec, *csec = sec;
				int			limit = fsec->specFlags & SF_MOVFLOOR ? 7*FRACUNIT : 8*FRACUNIT;	// $v73
				RDBOOL		repsecsky, movfloor = false, movceil = false, noR3D_16 = false;
				RDBOOL		setfloorsequal = false;
				RDBOOL		samesec =  ssinf.anylines
									&& ssinf.anylines == ssinf.anysameseclines;

				// Test added b220-1
				if (   srf == sec
					&& (sec->specFlags & (SF_MOVSECTOR|0x1C)) == 0x1C
					&& rejectSEC(srf))
				{
					return;
				}

				// v220-4 add low floor floodfill where contained sectors have a movfloor
				if (rssinf.anylines && !rssinf.loringcnt)
				{
					setfloorsequal =
						(!rssinf.lowfloorsequal2 && rssinf.lowfloorsequal1 == rssinf.anylines && rssinf.uncompliantlines)	// a)
						||
						(  rssinf.lowfloorsequal1 && rssinf.lowfloorsequal2
						&& rssinf.lowfloorsequal1 == rssinf.setallownsec && rssinf.lowfloorsequal2 == rssinf.anylines);		// b)
				}

				validcount++;	// for LS_SetLines()

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (   ss->sector->specFlags & SF_MOVFLOOR
						&& linesInside(ss->sector)) // v220-5 added
					{
						movfloor = noR3D_16 = true;
					}

					if (ss->sector->specFlags & SF_MOVCEIL)
						movceil = noR3D_16 = true;
					else if (ss->sector->ceilingheight == csec->ceilingheight
						&& ss->sector->ceilingpic != skyflatnum)
					{
						csec = ss->sector;
					}
				}

				// v220-13 Check for conflicting ceiling candidates
				if (csec != sec)
				{
					sector_t	*hcsec = NULL;

					for (i = 0; i < hitssecptr; i++)
					{
						subsector_t *ss = &subsectors[hitssec[i]];

						if ( !(ss->sector->specFlags & (SF_MOVCEIL|SF_NOHITEX))
							&& ss->sector->ceilingheight > sec->ceilingheight)
						{
							hcsec = ss->sector;
						}
					}

					if (hcsec)
					{
						for (i = 0; i < lvar->loopcount; i++)
						{
							line_t			*lc = &lines[lvar->looplist[i]];
							vertexowner_t	*vo = &vertexowners[lc->v1 - vertexes];
							int				n;

							for (n = 0; n < vo->num; n++)
							{
								if (vo->list[n] == hcsec - sectors)
								{
									csec = hcsec;
									goto hcsec_end;
								}
							}

							vo = &vertexowners[lc->v2 - vertexes];

							for (n = 0; n < vo->num; n++)
							{
								if (vo->list[n] == hcsec - sectors)
								{
									csec = hcsec;
									goto hcsec_end;
								}
							}
						}
					}
				}
hcsec_end:;
				// v220-18 Check for competing floor candidates
				int	 lls = os >= 0
						&& os != srf - sectors
						&& sectors[os].floorheight == srf->floorheight
						&& (sectors[os].specFlags & 0x3FF14) == 0x14
						&& (srf->specFlags & 0x3FF14) == 0x14;

				fsecnum = lls ? srf - sectors : fsec - sectors;
				csecnum = csec - sectors;
				
				repsecsky = sec->ceilingpic == skyflatnum;

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];
					RDBOOL		movef = (ss->sector->specFlags & SF_MOVFLOOR) > 0;
					RDBOOL		movec = (ss->sector->specFlags & SF_MOVCEIL ) > 0;
					RDBOOL		fchange = !nofloorsec && abs(ss->sector->floorheight - sec->floorheight) > limit;
					RDBOOL		cchange = !noceilsec  && ss->sector->ceilingheight != sec->ceilingheight;
					RDBOOL		invmovfloor = !rssinf.boundmoveflags
								&& !lvcount
								&& sec->specFlags & SF_MOVFLOOR
								&& sec->specFlags & SF_MOVINVALID;
					RDBOOL		hiliqf = rssinf.hiliqfinside
								&& rssinf.hiliqfinside == rssinf.anylines - rssinf.notexlines;
					RDBOOL		validlowithhi = rssinf.boundslotex
								&& lvcount
								&& flag == MSIA_NONE
								&& rssinf.anylines
								&& fsecnum == csecnum
								&& lvar->secnum != fsecnum
								&& !movef
								&& !movec
								&& !rssinf.lotex
								&& !rssinf.midtex
								&& !rssinf.hitex
								&& !hiliqf;

					int			mark = excludessecs[hitssec[i]];

					if (hiliqf) noR3D_16 = false;

					if (rssinf.boundslotex)
					{
						if (!hiliqf && !(mark & 1) && ss->sector->specFlags & SF_NOLOTEX)
						{
							int mask = 0x4013001C;

							if ( (ss->sector->specFlags & mask) == 0x4011001C)
								Setr3dsslfsec(ss, lvar->secnum, 46046);
							else
								Setr3dssfs(ss, lvar->secnum, 46046);
						}

						if (!(mark & 2)
							&& (chkceiltex || (ss->sector->specFlags & SF_NOHITEX) && !(ss->sector->specFlags & SF_SSLINETRUE)) )	// v220-20
						{
							Setr3dsscs(ss, csecnum, 46047);

							if (   validlowithhi
								&& ss->floorsec < 0
								&& !(mark & 1))
							{
								Setr3dssfs(ss, lvar->secnum, 46047);
							}
							// v220-15 check for fixed sector in donut floor
							else if (ss->sector - sectors == lvar->secnum
								&& (ss->sector->specFlags & 0x80301C) == 0x800018
								&& (fsec->specFlags & 0x440F0FFC) == 0x40010410
								&& fsec->floorheight < sectors[lvar->secnum].floorheight
								&& ss->floorsec < 0
								&& ss->lfsec < 0)
							{
								Setr3dsslfsec(ss, fsec - sectors, 46047);
							}
						}
					}
					else if (rssinf.boundshitex)
					{
						// v220-5 add SS_HasLotex() test
						// Modified v220-7
						if ( !(mark & 1) || chkfloortex)
						{
							int ret = SS_HasLotex(ss);

							if (ret == 0)
							{
								// v220-10 test added
								if (ss->sector->floorheight < sectors[fsecnum].floorheight)
								{
									Setr3dssfs(ss, fsecnum, 46048);
								}

								// v220-18 test added
								if (   rssinf.boundshitex == 1
									&& ss->sector->ceilingheight > sectors[fsecnum].ceilingheight
									&& noHiTexInSector(ss->sector))
								{
									Setr3dsslcsec(ss, fsecnum, 46048);
									if (lowfloorset < 0) lowfloorset = lvar->secnum;
								}
							}
							else if (ret < 0)
								Setr3dsslfsec(ss, fsecnum, 46048);
						}

						if (!(mark & 2)
							&& ss->sector->ceilingheight > sectors[lvar->secnum].ceilingheight)
						{
							Setr3dsscs(ss, lvar->secnum, 46048);
						}
					}
					else if (rssinf.sameheightlines)
					{
						if (lvar->hilo)
						{
							if (!(mark & 2)) Setr3dsscs(ss, csecnum, 46049);
						}
						else
						{
							if (!(mark & 1))
							{
								if (fsec->specFlags & SF_MOVFLOOR)
								{
									if (   IsLiquidFloorFlat(ss->sector)
										&& checkfsegComplies(ss))
									{
										Setr3dsslfsec(ss, fsecnum, 46050);
										ss->ssflags |= SSF_CHKFLOORDIFF;
									}
								}
								// v220-6
								// Don't set lowfloor if the boundary has midtextures
								// or if the replacement sector has lower textures
								// e.g; EPIC    map05 sec 691 : not wanted
								//    ; BATMAN  map03 sec 125 : not needed (although does not cause problem if set)
								//    ; Requiem map21 sec  64 : needed
								else if (!rssinf.boundsmidtex
									&& sectors[fsecnum].specFlags & SF_NOLOTEX
									&& ss->sector - sectors != fsecnum
									&& ss->floorsec != fsecnum)
								{
									Setr3dssfs(ss, fsecnum, 46051);
								}
							}
						}
					}
					else
					{
						if (setfloorsequal) Setr3dsslfsec(ss, sec-sectors, 46052);

						if (!(mark & 2))
						{
							if (chkceiltex)
							{
								if (   lvar->hilo
									// if sky use the default sector
									|| (ss->sector->ceilingpic == skyflatnum && !repsecsky))
								{
									if (movceil)
										Setr3dsslcsec(ss, csecnum, 46053);
									else
										Setr3dsscs(ss, csecnum, 46053);
								}
							}
							else
							{
								if (movec)
								{
									if (!(ss->sector->specFlags & SF_NOHITEX)
										&& sectors[csecnum].ceilingpic != skyflatnum)
										Setr3dsslcsec(ss, csecnum, 46054);
								}
								else if (cchange && ss->sector->specFlags & SF_NOHITEX)
									Setr3dsscs(ss, csecnum, 46054);
							}
						}
						
						if (!(mark & 1) && !setfloorsequal)
						{
							if (chkfloortex)
							{
								if (!lvar->hilo)
								{
									// v220-10 movef with lowtex test added
									if (movfloor || (movef && !(fsec->specFlags & SF_NOLOTEX)) )
										Setr3dsslfsec(ss, fsecnum, 46055);
									// Test added v220-3
									else if ((sectors[fsecnum].specFlags & 0x1C) != 0x1C)
										Setr3dssfs(ss, fsecnum, 46056);
								}
							}
							else
							{
								if (movef)
								{
									if (!(ss->sector->specFlags & SF_NOLOTEX))
										Setr3dsslfsec(ss, fsecnum, 46057);
								}
								else if (fchange && ss->sector->specFlags & SF_NOLOTEX)
								{
									if (sectors[fsecnum].specFlags & SF_MOVFLOOR)
									{
										if (ss->sector->specFlags & SF_MOVINVALID)
											Setr3dsslfsec(ss, fsecnum, 46058);
									}
									else
										Setr3dssfs(ss, fsecnum, 46058);
								}
							}
						}

						// v220-16 Prevent being changed in r3d_tracenormal (this could be a subset of a set)
						if (ss->ceilsec >= 0 || ss->floorsec >= 0) ss->ssflags |= SSF_NOTRACEN;
					}

					if (lowliqfloor
						|| (invmovfloor && ss->sector->floorheight < sec->floorheight))
					{
						if (   lowliqfloor
							&& ss->sector->floorheight > sectors[lvar->secnum].floorheight
							&& (ss->sector->specFlags & 0xB001C) == 0x9001C
							&&  ss->sector->ceilingpic == skyflatnum)
						{
							// v220-22 change from Setr3dssfs to Setr3dsslfsec
							Setr3dsslfsec(ss, lvar->secnum, 46059);
						}
						else if (ss->lfsec < 0)	// only if not already set
						{
							// Don't draw liquid over liquid in a sky sector $v67
							if ((ss->sector->specFlags & 0xF001C) == 0x1C
								&& IsLiquidFloorFlat(ss->sector)
								&& ss->sector->floorpic == sectors[lvar->secnum].floorpic
								&& ss->sector->ceilingpic == skyflatnum
								&& sectors[lvar->secnum].ceilingpic == skyflatnum)
							{
								SetOverride(ss, lvar->secnum, 46059);
							}
							else
							{
								Setr3dsslfsec(ss, lvar->secnum, 46059);
								if (lowfloorset < 0) lowfloorset = lvar->secnum;
							}

							if (   lowliqfloor
								&& ss->sector->floorheight > sectors[lvar->secnum].floorheight
								&& (ss->sector->specFlags & 0x3001C) == 0x1C
								&& !allVtxs2(lvar))
							{
								chkSharedLineSSECS(ss, lvar->secnum);
							}
						}
					}

					// v220-10 new test added.
					if (   lvar->hilo == 1
						&& lvar->secnum == srf - sectors
						&& !(srf->specFlags  & SF_MOVSECTOR)
						&& (sec->specFlags & SF_MOVSECTOR) == SF_MOVSECTOR)
					{
						Setr3dsslfsec(ss, sec - sectors, 46059);
					}

					if (srf->ceilingheight == srf->floorheight && !(srf->specFlags & SF_MOVSECTOR))
					{
						LS_SetLines(sec, ss, false, true, false, 46060); 
					}
					// v220-10 new test added
					else if (lvar->hilo == 1
						&& (srf->specFlags   & SF_MOVSECTOR) == SF_MOVFLOOR
						&& (sec->specFlags & SF_MOVSECTOR) == SF_MOVFLOOR
						&& srf->ceilingheight > sec->ceilingheight)
					{
						LS_SetLines(sec, ss, false, true, false, 46061); 
					}
					else
					{
						LS_SetLines(sec, ss, noR3D_16, false, !lvcount && !lvar->hilo, 46062);
					}
				}

				if (lowfloorset >= 0)
						Setsecdone(&sectors[lowfloorset], 46063);
			}
			else
			{
				// v220-10 yet another test needed
				if (   settype == R3D_01
					&& (srf->specFlags & 0x1F001C) == 0x9001C)
				{
					sector_t	*chk;
					RDBOOL		ring_valid = true;

					for (i = 0; i < lvar->loopcount; i++)
					{
						line_t	*li = &lines[lvar->looplist[i]];

						if (   !li->backsector
							|| li->backsector == li->frontsector
							|| !(li->frontsector->specFlags & SF_MOVFLOOR)
							|| !(li->backsector->specFlags & SF_MOVFLOOR)
							|| !(chk = li->frontsector == srf ? li->backsector : li->frontsector)
							|| (chk->specFlags & SF_NOLOTEX && !(chk->specFlags & SF_SSLINETRUE)) )
						{
							ring_valid = false; break;
						}
					}

					if (ring_valid && boundBySSlines(lvar))
						settype = R3D_14;
				}

				for (i = 0; i < lvar->loopcount; i++)
				{
					line_t	*li = &lines[lvar->looplist[i]];
					byte	type = settype;

					if (type == 0xFF)
					{
						int txn = origTEX[lvar->looplist[i]];

						if (txn & 0x22)
						{
							if (txn & 0x55 && !invalidTEX_movfType(li))
								type = R3D_03;
							else if (lvcount
								|| (!(txn & 0x55)
								&& (li->frontsector->specFlags & SF_MOVFLOOR)
								^  (li->backsector->specFlags  & SF_MOVFLOOR)) )
							{
								if (   li->frontsector->specFlags & SF_MOVFIXED
									|| li->backsector->specFlags  & SF_MOVFIXED)
								{
									type = R3D_16;
								}
								else
								{
									type = R3D_1D;
								}
							}
							else
								type = R3D_16;
						}
						else if (invalidTEX_movfType(li))
							type = R3D_1D;
						else if (txn & 0x11 &&
							(abs(li->frontsector->ceilingheight - li->backsector->ceilingheight) > MAX_SECHEIGHT
							|| (li->frontsector->specFlags & SF_MOVSECTOR ^ li->backsector->specFlags & SF_MOVSECTOR)))
						{
							type = R3D_00;
						}
						// v229-10 $checkout
						else if (li->frontsector->specFlags & SF_MOVFLOOR
							&&   li->backsector->specFlags  & SF_MOVFLOOR)
						{
							type = R3D_1A;
						}
						else
							type = R3D_16;
					}

					if (   type == R3D_16
						&& !origTEX[li - lines]
						&& li->frontsector->ceilingheight == li->backsector->ceilingheight
						&& !(li->frontsector->specFlags & SF_MOVCEIL)
						&& !(li->backsector->specFlags  & SF_MOVCEIL))
					{
						type = R3D_14;
					}

					Settype(li, type, 46072);

					if (type != R3D_03)
						SetLFG(li, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 46073);
				}
			}
			return;
		}
	}
}

static int getFHT(loopvar_t	*lpv)
{
	line_t		*lc;

	if (lpv->hilo) return DDMAXINT;

	lc = &lines[lpv->looplist[0]];

	if (lc->frontsector->floorheight < lc->backsector->floorheight)
		return lc->frontsector->floorheight;
		
	return lc->backsector->floorheight;
}

static int getCHT(loopvar_t	*lpv)
{
	line_t		*lc;

	if (!lpv->hilo) return DDMININT;

	lc = &lines[lpv->looplist[0]];

	if (lc->frontsector->ceilingheight > lc->backsector->ceilingheight)
		return lc->frontsector->ceilingheight;
		
	return lc->backsector->ceilingheight;
}

static RDBOOL fn_missBoth(loopvar_t *lpv)
{
	int i;

	for (i = 0; i < lpv->loopcount; i++)
	{
		int flags = LS_TestLineTex(&lines[lpv->looplist[i]]);

		if (!(flags & LT_MISLO) || !(flags & LT_MISHI))
			return false;
	}
	return true;
}

// RINF entry point
// Deal with missing tex rings
static void enactRINF()
{
	int		c, dc, df, i, k, m;

	RINF_SetArray();

	if (!rinf.count) return;

	for (i = 0; i < rinf.count; i++)
	{
		loopvar_t	*lvar = &rinf.lvar[i];
		int			*box  = lvar->box;

		// Check this is not inside another missing tex box
		for (k = 0; k < rinf.count && k != i; k++)
		{
			if (M_CompareBox(box, rinf.lvar[k].box))
			{
				lvar->inside = true;
			}
		}
	}

	for (i = 0; i < rinf.count; i++)
	{
		loopvar_t	*lvar = &rinf.lvar[i], *lvarin;
		loopvar_t	*lvmem[128];
		int			fht = getFHT(lvar), cht = getCHT(lvar);
		int			lvcount = 0;
		int			hilo = 0;

		//--------------------------------------------------

		if (!lvar->inside)
		{
			RDBOOL missboth = fn_missBoth(lvar); 

			for (c = dc = df = k = 0; k < rinf.count; k++)
			{
				lvarin = &rinf.lvar[k];

				if (lvarin->inside && M_CompareBox(lvarin->box, lvar->box))
				{
					lvmem[lvcount++] = lvarin; c++;

					if (fht != DDMAXINT && !lvarin->hilo && getFHT(lvarin) == fht)
					{
						df++;
					}

					if (cht != DDMININT && lvarin->hilo && getFHT(lvarin) == fht)
					{
						dc++;
					}
				}
			}

			// Check if inner loops are nested (either will or won't be - never a mix)
			// All inner loops are in the primary outer loop
			// but we need to check whether all inner loops are nested.

			lvar->nested = true;	// Assume nested until proven otherwise

			if (lvcount > 1)
			{
				// Start with the smallest
				for (k = lvcount; --k > 0 ;)
				{
					lvarin = lvmem[k];

					// Note: it is possible to have the same box twice (as separate rings
					// 	for hi and lo). As such M_FitBox is used for comparison.
					if (!M_FitBox(lvmem[k]->box, lvmem[k-1]->box))
					{

						lvar->nested = false; break;
					}
				}
			}

			if (lvar->nested)
			{
				for (k = 0; k < lvcount; k++)
					lvmem[k]->nested = true;
			}

			// v220-4 check outer box for an SF_MOVINVALID type  e.g, batman map30 sec 42
			if (   !lvar->inside
				&& !lvar->hilo
				&& !lvcount
				&& lvarin->loopcount
				&& !(sectors[lvar->secnum].specFlags & SF_MOVSECTOR)
				&& !LS_MarkSubsecsInLoop(&sectors[lvar->secnum], lvar->loopcount, lvar->looplist, lvar->box, MSIA_TESTAREA, NULL))
			{
				sector_t *sec = &sectors[lvar->secnum];
				sector_t *s1, *s2 = NULL;
				line_t	 *ld;

				for (k = 0; k < lvarin->loopcount; k++)
				{
					ld = &lines[lvarin->looplist[k]];

					s1 = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!s2)
					{
						if (  (s1->specFlags & 0x1FFFF8) != 0x110018
							|| s1->floorheight <= sectors[lvar->secnum].floorheight
							|| s1->floorpic != sectors[lvar->secnum].floorpic)
						{
							break;
						}
						s2 = s1;
					}
					if (s2 != s1) { s2 = NULL; break; }
				}

				if (s2) 
				{
					for (k = 0; k < lvarin->loopcount; k++)
					{
						ld = &lines[lvarin->looplist[k]];

						Settype(ld, R3D_17, 46600);
						SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46600);
					}

					continue;
				}
			}

			// v220-4 check if an outer ring's box is the same size as its outer box
			//        If so it will mean that the outer box's limits have the values
			//        of vertices shared by both rings
			if (   !lvar->inside
				&& !lvcount
				&& !missboth
				&& lvar->loopcount == sectors[lvar->secnum].linecount
				&& (sectors[lvar->secnum].specFlags & 0xFFFFC) == 0x1C)
			{
				sector_t *sec = &sectors[lvar->secnum];
				sector_t *s1, *s2 = NULL, *s3 = NULL;
				RDBOOL   valid = true;

				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					s1 = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!s2) s2 = s1;
					if (s2 != s1) { valid = false; break; }

				}

				if (valid && s2 && (s2->specFlags & 0x3000C) == 0x10008)
				{
					int		 nbox[4];

					M_ClearBox(nbox);

					for (k = 0; k < s2->linecount; k++)
					{
						line_t *ld = s2->lines[k];

						if (!(ld->spflags & SPF_CONTROLLINE))
						{
							s1 = ld->frontsector == s2 ? ld->backsector : ld->frontsector;

							if (s1 != sec)
							{
								if (!s3) s3 = s1;
								if (s3 != s1) { valid = false; break; }
							}

							M_AddLineToBox(nbox, ld);
						}
					}

					if (   valid
						&& s3
						&& M_SameBox(lvar->box, nbox)
						&& !(s3->specFlags & SF_MOVSECTOR)
						&& s3->specFlags & SF_NOHITEX
						&& s3->floorheight <= sec->floorheight
						&& s3->floorheight <= s2->floorheight)
					{
						// OK.  So we have a moving sector enclosing a static textureless
						//      sector.  We can only assume that the map author wants the
						//      walls of the moving sector not to be drawn despite those
						//      lines common to both sectors having a low texture

						for (k = 0; k < s2->linecount; k++)
						{
							line_t *ld = s2->lines[k];

							if (!(ld->spflags & SPF_CONTROLLINE))
							{
								if (ld->frontsector == sec || ld->backsector == sec)
									Setspec(ld, s3 - sectors, 46600);
								Settype(ld, R3D_3F, 46600);
								SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_FIXABS, 46601);
							}
						}
						Setsecdone(sec, 46601);
						Setsecdone( s2, 46601);
						Setsecdone( s3, 46601);
					}

					continue;
				}
			}

			// Start with the smallest
			for (k = lvcount; --k >= 0 ;)
			{
				lvarin = lvmem[k];

				if (   lvarin->inside
					&& M_CompareBox(lvarin->box, lvar->box)
					&& (origTEX[lvarin->looplist[0]] & 0x55))
				{
					sector_t *sec = &sectors[lvarin->secnum];
					int		 mark = lvarin->hilo ? 1:2;
					int		 flag = MSIA_NONE;

					if (!LS_MarkSubsecsInLoop(sec, lvarin->loopcount, lvarin->looplist, lvarin->box, MSIA_TESTAREA, NULL))
						flag = MSIA_REVERSED;

					// This should not fail
					if (LS_MarkSubsecsInLoop(sec, lvarin->loopcount, lvarin->looplist, lvarin->box, flag, NULL))
					{
						line_t	*ref = &lines[lvarin->looplist[0]];

						setRINF_SSINF(sec, lvar, lvcount, NULL, false);

						if (lvarin->hilo)			// Missing ceiling (low wall tex present)
						{
							// Highest floor
							sec = ref->frontsector->floorheight < ref->backsector->floorheight ?
							 ref->backsector : ref->frontsector;
						}
						else						// Missing floor (high wall tex present)
						{
							// Lowest ceiling
							sec = ref->frontsector->ceilingheight > ref->backsector->ceilingheight ?
							 ref->backsector : ref->frontsector;
						}

						for (m = 0; m < hitssecptr; m++)
						{
							subsector_t *ss = &subsectors[hitssec[m]];

							if (excludessecs[hitssec[m]] & 4) continue;

							if (rssinf.sameseclines || rssinf.sameheightlines)
							{
								// Ensure the floor or ceil is correct
								if (lvarin->hilo)	// if ceiling missing then check the floor
								{
									if (ss->sector->floorheight != sec->floorheight)
										Setr3dssfs(ss, sec - sectors, 46074);
								}
								else				// if floor missing then check the ceiling
								{
									if (ss->sector->ceilingheight != sec->ceilingheight)
										Setr3dsscs(ss, sec - sectors, 46075);
								}
							}

							excludessecs[hitssec[m]] |= mark;
						}
					}
				}
			}

			if (lvar->loopcount)
			{
				setRINFssecs(lvar, c == 1 && df == 1, c == 1 && dc == 1, lvarin, missboth, lvcount);
			}

			// check for special types
			{
				sector_t *s = NULL, *sec = &sectors[lvar->secnum];
				int		 box[2][4];

				CLR_hitline();

				for (c = k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (   ld->lnflags2 & LNF2_MISSINGSPEC
						&& Vtxs2(ld)
						&& !Vtfl(ld))
					{
						sector_t *chk = ld->frontsector == sec ?
							ld->backsector : ld->frontsector;

						if (!s) s = chk; else if (s != chk) { c = 0; break; }
						hitline[ld - lines] = 1; c++;
					}
				}

				if (c < 3 || s->floorheight > sec->floorheight) continue;

				for (k = 0; k < s->linecount; k++)
				{
					int lnum = s->lines[k] - lines;

					if (hitline[lnum] == 1
						&& subfn_CheckForRing(hitline, s, lnum, NULL, box[0], NULL, NULL))
					{
						M_ClearBox(box[1]);

						for (m = 0; m < s->linecount; m++)
						{
							lnum = s->lines[m] - lines;

							if (hitline[lnum] == 4) M_AddLineToBox(box[1], s->lines[m]);
						}

						if (M_CompareBox(box[1], lvar->box))
						{
							RDBOOL	valid = false;

							for (m = lvcount; --m >= 0 ;)
							{
								lvarin = lvmem[m];

								if (   lvarin->inside
									&& !lvarin->hilo
									&& M_CompareBox(lvarin->box, lvar->box))
								{
									valid = true;
								}
							}

							if (valid
								&& LS_MarkSubsecsInAreaGeneral(s, hitline, 0, NULL, box[0], -1, MSIA_NONE))
							{
								validcount++;	// for LS_SetLines()

								for (m = 0; m < hitssecptr; m++)
								{
									subsector_t *ss = &subsectors[hitssec[m]];

									Setr3dssfs(ss, s - sectors, 46076);
									LS_SetLines(s, ss, true, false, false, 46077);
								}
							}
						}
						for (m = 0; m < s->linecount; m++)
						{
							lnum = s->lines[m] - lines;

							if (hitline[lnum] == 4) hitline[lnum] = 8;
						}
					}
				}
			}
		}
	}
}

//--------------------------------------------------------

static int setLineRings(sector_t *s, int limit, RDBOOL needvtx2, RDBOOL nomidtex)
{
	int c, d, i, k, rc, lnum;

	ringhicnt = ringlocnt = 0;

	CLR_hitline();

	for (c = i = 0; i < s->linecount; i++)
	{
		line_t *li = s->lines[i];

		if (li->spflags & SPF_CONTROLLINE) continue;

		if (   li->frontsector == li->backsector
			|| (nomidtex && origTEX[li - lines])
			|| !(li->ml_flags & ML_TWOSIDED)
			|| lxarray[li - lines] & 4
			|| (needvtx2 && !Vtxs2(li))
			|| chkSkyFloor(li))
		{
			return 0;
		}

		hitline[li - lines] = 1; c++;
	}

	if (c < 6) return 0;

	for (c = i = 0; i < s->linecount; i++)
	{
		if (c >= limit) return 0;
		
		lnum = s->lines[i] - lines;

		if (   hitline[lnum] == 1
			&& subfn_CheckForRing(hitline, s, lnum, &rc, lst[c].box, NULL, NULL))
		{
			lst[c].sz =	M_AreaBox(lst[c].box);
			lst[c].count = rc;

			if (lst[c].maxcount < rc)
			{
				lst[c].maxcount = rc;
				lst[c].list = (int*)W_Realloc(lst[c].list, rc * sizeof(int), STATICID, 0);
			}

			for (d = k = 0; k < s->linecount; k++)
			{
				int lnum = s->lines[k] - lines;

				if (hitline[lnum] == 4)
				{
					lst[c].list[d++] = lnum;
					hitline[lnum] = 8;
				}
			}

			if (d != lst[c].count) Con_ErrorD("setLineRings: count mismatch");

			c++;
		}
	}

	return c;
}

static sector_t *hifloorsec, *hifloorchk;

static RDBOOL highFloorSec(line_t *ld)
{
	if (ld->backsector)
	{
		int fdiff = ld->frontsector->floorheight - ld->backsector->floorheight;

		if (fdiff > 0 && ld->backsector->specFlags & SF_NOLOTEX)
		{
			if (!hifloorsec)
			{
				hifloorsec = ld->frontsector;
			}
			else if (hifloorsec != ld->frontsector)
			{
				hifloorchk = ld->frontsector;
				return false;
			}
			return true;
		}
		if (fdiff < 0 && ld->frontsector->specFlags & SF_NOLOTEX)
		{
			if (!hifloorsec)
			{
				hifloorsec = ld->backsector;
			}
			else if (hifloorsec != ld->backsector)
			{
				hifloorchk = ld->backsector;
				return false;
			}
			return true;
		}
	}
	return false;
}

static RDBOOL allSameSecLinesInside(RDBOOL chkhifloor, byte *linecheck)
{
	int			i, k;

	hifloorsec = hifloorchk = NULL;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef && sg->linedef->validcount != validcount)
			{
				sg->linedef->validcount = validcount;

				if (linecheck)
				{
					if (   linecheck[sg->linedef - lines]
						|| sg->linedef->frontsector == sg->linedef->backsector)
					{
						continue;
					}

					return false;
				}

				if (chkhifloor && highFloorSec(sg->linedef))
				{
					if (   !hifloorchk
						&& !(sg->linedef->frontsector->specFlags & SF_MOVFLOOR)
						&& !(sg->linedef->backsector
						&& sg->linedef->backsector->specFlags & SF_MOVFLOOR))
					{
						continue;
					}

					return false;
				}

				if (!hitline[sg->linedef - lines]
					&& sg->linedef->frontsector != sg->linedef->backsector)
				{
					if (!chkhifloor)
					{
						if (  !sg->linedef->backsector
							|| origTEX[sg->linedef - lines])
						{
							continue;
						}
						if (highFloorSec(sg->linedef))
						{
							continue;
						}
						else
						{
							if (   hifloorsec
								&& hifloorchk
								&& hifloorchk->floorheight
								== hifloorsec->floorheight)
							{
								continue;
							}
						}
					}
					return false;
				}
			}
		}
	}

	return linecheck != NULL || chkhifloor ? true : hifloorsec != NULL;
}

// The ring line marker must be preset in hitlinebak
static void excludeSSEC(sector_t *sec, int *ptr, int count, RDBOOL *recheck)
{
	int		c, i, k, lnum, bbox[4];

	//----------------------------

	if (count < 3) return;

	memset(hitlinebak, 0, numlines);

	for (i = 0; i < count; i++) hitlinebak[ptr[i]] = 1;

	for (i = 0; i < count; i++)
	{
		lnum = ptr[i];

		if (   hitlinebak[lnum] == 1
			&& checkForRingInList(hitlinebak, count, ptr, lnum, bbox, NULL, DDMININT, false))
		{
			int tempptr = hitssecptr;
			int flag = MSIA_NONE;

			memcpy(tempsseclist, hitssec, hitssecptr*sizeof(int));

			if (!LS_MarkSubsecsInAreaGeneral(sec, hitlinebak, 0, NULL, bbox, -1, MSIA_TESTAREA))
			{
				if (!recheck) return;

				flag = MSIA_REVERSED;
			}

			for (c = k = 0; k < count; k++)
			{
				if (hitlinebak[ptr[k]] == 4) ringtmp[c++] = ptr[k];
			}

			if (LS_MarkSubsecsInLoop(sec, c, ringtmp, bbox, flag, NULL))
			{
				byte mask;

				if (recheck)	// else uses values sent
				{
					for (k = 0; k < count; k++)
					{
						if (hitlinebak[ptr[k]] == 4) excludeline[ptr[k]] = true;
					}

					LS_SetSSINF(sec, hitlinebak, excludessecs, bridgelines, true, false);

					if (ssinf.lotex) mask  = 1;
					if (ssinf.hitex) mask |= 2;

					for (k = 0; k < hitssecptr; k++)
						excludessecs[hitssec[k]] = mask;

					*recheck = true;
				}
			}

			hitssecptr = tempptr;
			memcpy(hitssec, tempsseclist, hitssecptr*sizeof(int));

			for (k = 0; k < count; k++)
			{
				if (hitlinebak[ptr[k]] == 4) hitlinebak[ptr[k]] = 8;
			}
		}
	}
}

static int __cdecl lstboxSize(const void *ls1, const void *ls2)
{
	int	sz1 = (*(linestore_t**)ls1)->sz; 
	int	sz2 = (*(linestore_t**)ls2)->sz; 

	return  sz1 > sz2 ? -1 : sz1 < sz2 ? 1 : 0;
}

// Deal with multiple loops
static RDBOOL DualRingsectors0x1C(sector_t *sec)
{
	sector_t	*chk, *outer = NULL;
	line_t		*ld;
	int			c, i, rcount;
	byte		*lineptr = NULL;

	//-----------------------

	if (sec->ceilingpic != skyflatnum && sec->ceilingheight - sec->floorheight > MAX_SECHEIGHT) return false;

	// Check for same sided double nested loops
	for (c = i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (   ld->frontsector == ld->backsector
			&& !(ld->spflags & SPF_CONTROLLINE)
			&& !(lxarray[ld - lines] & 4)
			&& Vtxs2(ld))
		{
			c++;
		}
	}

	if (c > 5 && (rcount = setLineRings(sec, DRLISTLIMIT, true, false)) > 1)
	{
		int		*bbox = lst[0].box, area, largest;
		RDBOOL	valid = true;

		area = lst[0].sz;
		largest = 0;

		for (i = 1; i < rcount; i++)
		{
			if (lst[i].sz > area)
			{
				bbox = lst[i].box;
				area = lst[i].sz;
				largest = i;
			}
		}

		for (i = 0; i < rcount && valid; i++)
		{
			if (i != largest) valid = M_CompareBox(lst[i].box, bbox);
		}

		if (valid)
		{
			subsector_t *ss;
			seg_t		*sg;
			line_t		*ld, *li;
			ssline_t	*sl;
			sector_t	*s, *fs, *bs;
			int			k, m, n, x, y;

			// What's expected here is that each ring will access
			// one (and only one) containing sector. Two of each
			// may access the same sector but each ring is treated
			// individually.

			for (i = 0, valid = true; i < rcount; i++)
			{
				s = NULL;
				validcount++;
				ss_validcount++;

				// Get a ring's referencing sector
				for (k = 0; k < lst[i].count && valid; k++)
				{
					li = &lines[lst[i].list[k]];
					sl = &sslines[li - lines];

					for (m = 0; m < sl->num && valid; m++)
					{
						ss = &subsectors[sl->list[m]];
						
						if (ss->ssvalidcount == ss_validcount)
							continue;

						x = (int)(ss->midpoint.x * FRACUNIT);
						y = (int)(ss->midpoint.y * FRACUNIT);

						ss->ssvalidcount = ss_validcount;

						sg = &segs[ss->firstline];

						for (n = 0; n < ss->linecount && valid; n++, sg++)
						{
							if (!(ld = sg->linedef) || ld->validcount == validcount)
							{
								continue;
							}

							ld->validcount = validcount;
							
							if (origTEX[ld - lines] & 0x55
								&& (fs = ld->frontsector) != (bs = ld->backsector)
								&& fs != li->frontsector
								&& bs != li->frontsector)
							{
								int	side = P_PointOnLineSide(x, y, ld);

								chk = side ? ld->backsector ?
									ld->backsector : ld->frontsector : ld->frontsector;

								if (!s) s = chk; else if (s != chk) { valid = false; break; }
							}
						}
					}
				}
				if (valid && s)
				{
					validcount++;
					ss_validcount++;

					for (k = 0; k < lst[i].count && valid; k++)
					{
						li = &lines[lst[i].list[k]];
						sl = &sslines[li - lines];

						for (m = 0; m < sl->num && valid; m++)
						{
							ss = &subsectors[sl->list[m]];
							sg = &segs[ss->firstline];

							SetOverride(ss, s - sectors, 46078);

							for (n = 0; n < ss->linecount; n++)
							{
								int txn;

								if (!(ld = sg->linedef) || ld->validcount == validcount)
								{
									continue;
								}

								ld->validcount = validcount;

								txn = origTEX[ld - lines];

								if (!txn)
									Settype(ld, R3D_01, 46079);
								else if (txn & 0x22 && !(txn & 0x55))
									Settype(ld, R3D_03, 46080);
								else
									Settype(ld, R3D_00, 46081);

								SetLFG(ld, LFG_FIXED|LFG_NEWSET, 46082);

								if (!secdone[ld->frontsector - sectors])
								{
									Setsecdone3(ld->frontsector, 46083);
									secmark[ld->frontsector - sectors] = 1;
								}

								if (ld->backsector
									&& !secdone[ld->backsector - sectors])
								{
									Setsecdone3(ld->backsector, 46080);
									secmark[ld->backsector - sectors] = 1;
								}
							}
						}
					}
				}
			}
		}
	}

	return secdone[sec - sectors] ? true : false;
}

static sector_t *getNext(line_t* ld, sector_t* sec)
{
	return	ld->frontsector != sec ?
		ld->frontsector : ld->backsector != sec ?
		ld->backsector  : NULL;
}

static sector_t *lowestCeilingSurrounding(sector_t *sec)
{
    sector_t	*other, *osec = NULL, *chk = NULL;
    fixed_t		height = sec->ceilingheight;
    int			i;

	for (i = 0; i < sec->linecount; i++)
	{
		if (sec->lines[i]->spflags & SPF_CONTROLLINE) continue;

		if (  (other = getNext(sec->lines[i], sec)) )
		{
			int lnum = sec->lines[i] - lines;

			if (other->ceilingheight < height)
				osec = other; height = other->ceilingheight;

			if (origTEX[lnum] & 0x11) chk = other;
		}
	}

	if (!osec) return NULL;

    return chk && chk->ceilingheight == osec->ceilingheight ? chk : osec;
}

static sector_t *backsec;

static RDBOOL validBack(line_t *ld, sector_t *sec, int type)
{
	sector_t *chk = ld->frontsector == sec ?
		ld->backsector : ld->frontsector;

	if (   !chk
		|| !(chk->specFlags & type)
		|| chk->ceilingheight == sec->ceilingheight)
	{
		return false;
	}

	if (!backsec) backsec = chk;

	return backsec == chk;
}

static void chkOuterMissing(sector_t *sec, line_t *ld)
{
	sector_t *fs = ld->frontsector;
	sector_t *bs = ld->backsector;

	if (!bs || fs == sec || bs == sec) return;

	if (bs && fs != bs)
	{
		int diff = fs->ceilingheight - bs->ceilingheight;

		if (diff && !(fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum))
		{
			int side = diff < 0;
			if (!sides[ld->sidenum[side]].toptexture)
			{
				ringtmp[ringtmpcnt++] = ld - lines;
				return;
			}
		}

		if (    (diff = fs->floorheight - bs->floorheight)
			&& !(fs->floorpic == skyflatnum && bs->floorpic == skyflatnum))
		{
			int side = diff > 0;
			if (!sides[ld->sidenum[side]].bottomtexture)
				ringtmp[ringtmpcnt++] = ld - lines;
		}
	}
}

static sector_t *chkRingMissing(sector_t *sec, int o)
{
	sector_t	*s1, *s2;
	int			c, i, n = ringtmp[0];

	// Not all the lines in the outer ring will be in ringtmp as they
	// were derived from subsectors.  There is no guarantee that a
	// subsector between the rings will own lines in both line rings.
	if (!sec || ringtmpcnt < 2)
	{
		ringtmpcnt = 0;
		return backsec;
	}

	s1 = lines[o].frontsector != sec ?
		 lines[o].frontsector : lines[o].backsector;
	s2 = lines[n].frontsector != s1 ?
		 lines[n].frontsector : lines[n].backsector;

	memset(hitlinebak, 0, numlines);

	for (i = 0; i < ringtmpcnt; i++)
		hitlinebak[ringtmp[i]] = 1;

	for (c = i = 0; i < s1->linecount; i++)
	{
		int lnum = s1->lines[i] - lines;

		if (hitlinebak[lnum] == 1)
		{
			c++;	// Used as a check for lines sent
		}
		else if ( s1 && s2
			&& (s1->lines[i]->frontsector == s2
			||  s1->lines[i]->backsector  == s2) )
		{
			hitlinebak[lnum] = 1;
		}
	}

	if (c == ringtmpcnt
		&& subfn_CheckForRing(hitlinebak, s1, ringtmp[0], NULL, NULL, NULL, NULL))
	{

		for (i = 0, ringtmpcnt = 0; i < s1->linecount; i++)
		{
			int lnum = s1->lines[i] - lines;

			if (hitlinebak[lnum] == 4) ringtmp[ringtmpcnt++] = lnum;
		}

		return s2;
	}

	ringtmpcnt = 0; return backsec;
}

static int __cdecl heightSorterFloor(const void *e1, const void *e2)
{
	int sh1 = sectors[*(int*) e1].floorheight;
	int sh2 = sectors[*(int*) e2].floorheight;
	if (sh1 > sh2) return -1;
	if (sh1 < sh2) return  1;
	return 0;
}

static int __cdecl heightSorterCeil(const void *e1, const void *e2)
{
	int sh1 = sectors[*(int*) e1].ceilingheight;
	int sh2 = sectors[*(int*) e2].ceilingheight;
	if (sh1 > sh2) return -1;
	if (sh1 < sh2) return  1;
	return 0;
}

static void lookForTexLineRings(int secidx)
{
	sector_t	*s;
	line_t		*ld;
	int			c, d, e;	// counters
	int			i, k, m;	// loop indexes
	int			tempptr, diff, lnum, bbox[4];
	int			*seclist;
	int			seclistcnt;
	RDBOOL		setlfssecs = false;
	RDBOOL		setlcssecs = false;
	RDBOOL		midring = false;

	memset(hitlinebak, 0, numlines);

	validcount++;

	for (c = d = e = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (   (ld = sg->linedef)
				&& ld->validcount != validcount)
			{
				sector_t *fs = ld->frontsector;
				sector_t *bs = ld->backsector;
				int		 txn = origTEX[ld - lines];

				ld->validcount = validcount;

				if (bs && fs != bs)
				{
					if (txn == 0x01 || txn == 0x10)
					{
						if (   fs->ceilingheight == bs->ceilingheight
							&& (diff = fs->floorheight - bs->floorheight)
							&& !(fs->floorpic == skyflatnum && bs->floorpic == skyflatnum) )
						{
							int secnum, side = diff > 0;

							if (sides[ld->sidenum[side]].bottomtexture)
							{
								hitlinebak[ld - lines] = 1, c++;
								secnum = side ? fs - sectors : bs - sectors;

								if (!(secmark[secnum] & 8))
									secmark[secnum] |= 8, d++;
							}
						}
					}

					if (txn == 0x02 || txn == 0x20)
					{
						if (   fs->floorheight == bs->floorheight
							&& fs - sectors != secidx
							&& bs - sectors != secidx
							&& (diff = fs->ceilingheight - bs->ceilingheight)
							&& !(fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum) )
						{
							int secnum, side = diff < 0;

							if (sides[ld->sidenum[side]].midtexture)
							{
								hitlinebak[ld - lines] = 1, c++;
								secnum = side ? fs - sectors : bs - sectors;

								if (!(secmark[secnum] & 16))
									secmark[secnum] |= 16, e++;
							}
						}
					}
				}
			}
		}
	}

	if (c < 3) return;

	tempptr = hitssecptr;
	memcpy(tempsseclist, hitssec, hitssecptr*sizeof(int));

	seclist = (int*)M_Malloc(MAX_OF(d, e) * sizeof(int), 78);

	// 1) Do Floors ---------------------------------------

	for (seclistcnt = i = 0; i < numsectors; i++)
	{
		if (secmark[i] & 8)
		{
			if (seclistcnt >= d) Con_ErrorD("lookForTexLineRings: program error LOW");
			seclist[seclistcnt++] = i;
			secmark[i] &= ~8;
		}
	}

	if (seclistcnt)
	{
		qsort(seclist, seclistcnt, sizeof(int), heightSorterFloor);

		for (i = 0; i < seclistcnt; i++)
		{
			s = &sectors[seclist[i]];

			for (k = 0; k < s->linecount; k++)
			{
				int rc;

				ld = s->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (ld->frontsector == s && sides[ld->sidenum[0]].bottomtexture)
					continue;

				if (ld->backsector == s && sides[ld->sidenum[1]].bottomtexture)
					continue;

				lnum = ld - lines;

				if (hitlinebak[lnum] == 1
					&& subfn_CheckForRing(hitlinebak, s, lnum, &rc, bbox, NULL, NULL)
					&& LS_MarkSubsecsInAreaGeneral(s, hitlinebak, 0, NULL, bbox, -1, MSIA_NONE))
				{
					for (m = 0; m < hitssecptr; m++)
					{
						Setr3dsslfsec(&subsectors[hitssec[m]], seclist[i], 46092);
					}
					
					setlfssecs = true;
				}
			}
		}
	}

	// 2) Do ceilings ------------------------------------------

	for (seclistcnt = i = 0; i < numsectors; i++)
	{
		if (secmark[i] & 16)
		{
			if (seclistcnt >= e) Con_ErrorD("lookForTexLineRings: program error MID");
			seclist[seclistcnt++] = i;
		}
		secmark[i] &= ~16;
	}

	if (seclistcnt)
	{
		qsort(seclist, seclistcnt, sizeof(int), heightSorterCeil);

		for (i = 0; i < seclistcnt; i++)
		{
			s = &sectors[seclist[i]];

			for (k = 0; k < s->linecount; k++)
			{
				int midcount = 0, ring_count = 0;

				ld = s->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (ld->frontsector == s && sides[ld->sidenum[0]].midtexture)
					continue;

				if (ld->backsector == s && sides[ld->sidenum[1]].midtexture)
					continue;

				lnum = ld - lines;

				if (hitlinebak[lnum] == 1
					&& subfn_CheckForRing(hitlinebak, s, lnum, &ring_count, bbox, NULL, NULL)
					&& LS_MarkSubsecsInAreaGeneral(s, hitlinebak, 0, NULL, bbox, -1, MSIA_NONE))
				{
					for (m = 0; m < s->linecount; m++)
					{
						line_t	*li = s->lines[m];
						int		lnum = li - lines;

						if (hitlinebak[lnum] == 4)
						{
							if (   (li->frontsector == s && sides[li->sidenum[1]].midtexture)
								|| (li->backsector == s  && sides[li->sidenum[0]].midtexture))
							{
								midcount++;
							}
						}
					}

					for (m = 0; m < hitssecptr; m++)
					{
						subsector_t *ss = &subsectors[hitssec[m]];
						int			sn = ss->sector - sectors;

						if (sn != secidx && sn != seclist[i])
						{
							Setr3dsslcsec(ss, sn, 46093);
						}
						Setr3dsscs(ss, seclist[i], 46094);
					}
					
					if (midcount == ring_count) midring = true;
					setlcssecs = true;
				}
			}
		}
	}

	hitssecptr = tempptr;
	memcpy(hitssec, tempsseclist, hitssecptr*sizeof(int));

	if (setlfssecs || setlcssecs)
	{
		validcount++;

		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];
			seg_t		*sg = &segs[ss->firstline];

			if (setlcssecs && ss->lcsec < 0)
				Setr3dsslcsec(ss, secidx, 46095);

			if (setlfssecs && ss->lfsec < 0)
				Setr3dsslfsec(ss, secidx, 46096);

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if ( (ld = sg->linedef) && ld->validcount != validcount)
				{
					ld->validcount = validcount;

					if (!origTEX[ld - lines])
						Settype(ld, R3D_01, 46097);
					else if (midring)
						Settype(ld, R3D_16, 46098);
					else
						Settype(ld, R3D_00, 46099);

					if (!secdone[ld->frontsector - sectors])
						Setsecdone(ld->frontsector, 46100);

					if (   ld->backsector
						&& !secdone[ld->backsector - sectors])
						Setsecdone(ld->backsector, 46101);

					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46102);
				}
			}
		}
	}

	M_Free(seclist);
}

static RDBOOL noSideSky(subsector_t *ss)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->frontsector && sg->frontsector->ceilingpic == skyflatnum) return false;
		if (sg->backsector  && sg->backsector->ceilingpic  == skyflatnum) return false;
	}
	return true;
}

static sector_t *hasDualRings(sector_t *sec, int type)
{
	sector_t	*osec = NULL, *osec1 = NULL;
	RDBOOL		validBackHi = true;
	RDBOOL		validBackLo = true;
	int			cht = sec->ceilingheight;
	int			fht = sec->floorheight;
	int			ccnt = 0, fcnt = 0;
	int			flinelo = -1, flinehi = -1;
	int			box[4];
	int			a, i, k, m;


	if (!(type & (SF_NOHITEX|SF_NOLOTEX)) ) return NULL;
	if (type & SF_NOHITEX && sec->ceilingpic == skyflatnum) return NULL;
	if (type & SF_NOLOTEX && sec->floorpic   == skyflatnum) return NULL;

	backsec = NULL;
	ringhicnt = ringlocnt = 0;
	memset(hitlinebak, 0, numlines);

	CLR_hitline();

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (    ld->ml_flags & ML_TWOSIDED
			&& !(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)))
		{
			sector_t *chk = ld->frontsector == sec ?
				ld->backsector : ld->frontsector;

			if (   type == SF_NOHITEX
				&& (chk->ceilingheight != cht || chk->specFlags & SF_NOHITEX))
			{
				if (validBackHi) validBackHi = validBack(ld, sec, type);

				if (!osec || !osec1)
				{
					sector_t *chk = ld->frontsector == sec ?
						ld->backsector : ld->frontsector;

					if (   !(chk->specFlags & type)
						&& chk->ceilingheight < sec->ceilingheight)
					{
						if (!origTEX[ld - lines]
							&& chk->floorheight == sec->floorheight)
						{
							osec = chk;
						}
						else if (origTEX[ld - lines] & 0x11)
						{
							osec1 = chk;
						}
					}
				}
				if (chk->ceilingheight != cht)
				{
					if (flinehi < 0) flinehi = ld - lines;
					ringhi[ringhicnt++] = ld - lines;
					hitline[ld - lines] = 1;
				}
				hitlinebak[ld - lines] = 1; ccnt++;
			}
			if (   type == SF_NOLOTEX
				&& (chk->floorheight != fht || chk->specFlags & SF_NOLOTEX))
			{
				if (validBackLo) validBackLo = validBack(ld, sec, type);
				if (chk->floorheight != fht)
				{
					if (flinelo < 0) flinelo = ld - lines;
					ringlo[ringlocnt++] = ld - lines;
					hitline[ld - lines] = 1;
				}
				hitlinebak[ld - lines] = 1; fcnt++;
			}
		}
	}

	if (osec && osec1 && osec1->ceilingheight == osec->ceilingheight)
		osec = osec1;

	if (!validBackHi && !osec && backsec && lowestCeilingSurrounding(sec) == backsec)
	{
		RDBOOL valid = false;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t *ld = sec->lines[i];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			if ( !(ld->ml_flags & ML_TWOSIDED)
				|| ld->frontsector == ld->backsector
				|| ld->frontsector->ceilingpic != ld->backsector->ceilingpic)
			{
				valid = false; break;
			}

			if (   origTEX[ld - lines] & 0x11	// check for AASHITTY
				&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight)
			{
				if (   !sides[ld->sidenum[0]].toptexture
					&& !sides[ld->sidenum[1]].toptexture)
				{
					valid = true;
				}
			}
		}
		if (valid)
		{
			// Setr3dch_byH(sec, backsec->ceilingheight, 46103);
			// sec->lightlevel = backsec->lightlevel;
			// v220-9 changed to;
			Setr3dch(sec, backsec, 46103);
		}
	}

	if (ringhicnt > 2 || ccnt)
	{
		if (   (a = !validBackHi && osec != NULL)
			|| (validBackHi
			&& backsec
			&& (osec = lowestCeilingSurrounding(backsec))) )
		{
			if (!a)
			{
				RDBOOL isset = false;

				for (i = 0; i < sec->linecount; i++)
				{
					int lnum = sec->lines[i] - lines;

					if (   hitlinebak[lnum] == 1
						&& subfn_CheckForRing(hitlinebak, sec, lnum, NULL, box, NULL, NULL))
					{
						if (LS_MarkSubsecsInAreaGeneral(sec, hitlinebak, 0, NULL, box, -1, MSIA_NONE))
						{
							for (k = 0; k < hitssecptr; k++)
							{
								Setr3dsscs(&subsectors[hitssec[k]], osec - sectors, 46104);
							}
						}
						for (k = 0; k < sec->linecount; k++)
						{
							line_t *ld = sec->lines[k];

							lnum = ld - lines;

							if (hitlinebak[lnum] == 4)
							{
								hitlinebak[lnum] = 8;
							}
						}
						isset = true;
					}
				}
				if(isset) return sec;
			}
		}

		if (ringhicnt < 3) return NULL;

		backsec = NULL;

		if (   subfn_CheckForRing(hitline, sec, flinehi, NULL, box, NULL, NULL)
			&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE))
		{
			int		oldflinehi = flinehi;
			int		box2[4];
			RDBOOL	setR3D16 = false;
			flinehi = -1;

			validcount++;
			ss_validcount++;
			ringtmpcnt = 0;

			for (a = i = 0; i < sec->linecount; i++)
			{
				int lnum = sec->lines[i] - lines;

				if (hitline[lnum] == 4)
				{
					ssline_t *sl = &sslines[lnum];

					sec->lines[i]->validcount = validcount;	// Dont test self

					// We need to check that this is not inside yet another
					// missing tex sector
					for (k = 0; k < sl->num; k++)
					{
						int			ssecidx = sl->list[k];
						subsector_t *ss = &subsectors[ssecidx];
						seg_t		*sg;
						line_t		*ld;
						RDBOOL		found = false;

						if (ss->ssvalidcount == ss_validcount) continue;

						ss->ssvalidcount = ss_validcount;

						for (m = 0; m < hitssecptr; m++)
						{
							if (hitssec[m] == ssecidx)
							{ found = true; break; }
						}
						if (found) continue;

						// not in list so must be an outer subsec
						// check the lines for missing textures
						// If any are found then abort
						sg = &segs[ss->firstline];

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ( !(ld = sg->linedef)
								|| ld->validcount == validcount)
							{
								continue;
							}

							ld->validcount = validcount;

							chkOuterMissing(sec, ld);
						}
					}

					if (!backsec)
					{
						line_t *ld = sec->lines[i];

						if (!Vtfl(ld))
						{
							if (ld->frontsector != sec) backsec = ld->frontsector;
							else if (ld->backsector
								&& ld->backsector != sec) backsec = ld->backsector;
						}
					}
					hitline[lnum] = 8;
				}
				else if (!hitline[lnum] && Vtxs2(sec->lines[i]))
				{
					if (flinehi < 0) flinehi = lnum;
					hitline[lnum] = 1; a++;
				}
			}

			if (!backsec && (backsec = chkRingMissing(sec, oldflinehi)))
			{
				for (i = 0; i < ringtmpcnt; i++)
				{
					int lnum = ringtmp[i];

					// v229-10 test added
					if ( !(origTEX[lnum] & 0x44)
						&& (lines[lnum].backsector->specFlags & 0xFFFFC) == 0x20014
						&& lines[lnum].frontsector->ceilingheight > lines[lnum].backsector->ceilingheight
						&& lines[lnum].backsector->ceilingheight == lines[lnum].backsector->floorheight)
					{
						Settype(&lines[lnum], R3D_36, 46107);
					}
					else
					{
						Settype(&lines[lnum], R3D_16, 46107);
					}

					// v220-22 new flag added
					SetLFG(&lines[lnum], LFG_NEWSET|LFG_FIXED|LFG_DUAL, 46108);
				}
				return sec;
			}

			if (!backsec)
			{
				backsec = lines[oldflinehi].frontsector == sec ?
					lines[oldflinehi].backsector : lines[oldflinehi].frontsector;

				if (sec->heightsec < 0)
					setR3D16 = true;	// all outer lines had lower missing
			}

			if (   a
				&& backsec
				&& subfn_CheckForRing(hitline, sec, flinehi, NULL, box2, NULL, NULL)
				&& M_CompareBox(box2, box))
			{
				if (!allSameSecLinesInside(false, NULL)) return NULL;

				for (i = 0; i < ringtmpcnt; i++)
				{
					line_t *ld = &lines[ringtmp[i]];

					Settype(ld, R3D_16, 46109);
					SetLFG(&lines[ringtmp[i]], LFG_NEWSET|LFG_FIXED, 46110);
				}

				for (i = 0; i < hitssecptr; i++)
				{
					subsector_t *ss = &subsectors[hitssec[i]];

					if (noSideSky(ss))
						Setr3dsscs(ss, backsec - sectors, 46111);

					if (hifloorsec)
					{
						Setr3dssfs(ss, hifloorsec - sectors, 46112);
					}
					else if (sec->specFlags & SF_NOLOTEX)
						Setr3dssfs(ss, sec - sectors, 46113);
				}

				lookForTexLineRings(sec - sectors);

				if (!setR3D16) return sec;

				for (i = 0; i < sec->linecount; i++)
				{
					int lnum = sec->lines[i] - lines;

					if (hitline[lnum] == 8)	// the outer line
					{
						Settype(sec->lines[i], R3D_16, 46112);
						SetLFG(sec->lines[i], LFG_NEWSET|LFG_FIXED, 46114);
					}
				}
				return sec;
			}
		}
	}
	return NULL;
}

static RDBOOL sectors0x1C(sector_t *sec)
{
	sector_t	*s; 
	int			ring_count, firstline = -1, bbox[4];
	int			c, i;
	RDBOOL		valid;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->lnflags2 & (LNF2_OWNTWOANDPLUS|LNF2_OWNEQUALTOTWO)
			&& !(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)))
		{
			sector_t *s = ld->frontsector == sec ?
				ld->backsector : ld->frontsector;

			if (  !s
				|| MSFC(s - sectors, SF_SECCLOSED|SF_NOHITEX|SF_NOLOTEX)
				|| s->heightsec >= 0
				|| s->specFlags & SF_MOVSECTOR
				|| chkSkyFloor(ld))
			{
				return false;
			}

			hitline[ld - lines] = 1;
		}
		else
		{
			return false;
		}
	}

	CLR_hitline();
	valid = true;
	s = NULL;

	for (c = i = 0; i < sec->linecount; i++)
	{
		line_t		*li = sec->lines[i];
		sector_t	*fs = li->frontsector;
		sector_t	*bs = li->backsector;

		if (li->spflags & SPF_CONTROLLINE) continue;

		if (   !(origTEX[li - lines] & 0x22)
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& !(bs->specFlags & SF_MOVSECTOR)
			&& fs->floorheight == bs->floorheight
			&&
			(   fs->ceilingheight == bs->ceilingheight
			|| (fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum)) )
		{
			if (origTEX[li - lines] & 0x11)
				Settype(li, R3D_00, 46115);
			else
				Settype(li, R3D_3F, 46116);	// v220-10
		}
		else
		{
			int flags;

			if (   fs->ceilingpic != skyflatnum
				&& bs->ceilingpic != skyflatnum
				&& fs->ceilingheight != bs->ceilingheight)
			{
				sector_t *chk = fs == sec ? bs : fs;

				if (valid && chk->ceilingheight < sec->ceilingheight)
				{
					if (!s) s = chk;
					else if (s != chk) { valid = false; s = NULL; }
				}
				else
				{
					valid = false; s = NULL;
				}

				if (valid)
				{
					if (firstline < 0) firstline = li - lines;
					hitline[li - lines] = 1; c++;
				}
			}
			
			flags = LS_TestLineTex(li);

			if (   flags & LT_MISLO
				&& li->lnflags2 & LNF2_OWNEQUALTOTWO
				&& !(flags & (LT_MISHI|LT_DIFFLO)) )
				Settype(li, R3D_00, 46117);
			else if (flags & (LT_MISHI|LT_MISLO))
			{
				if (lxarray[li - lines] & 32)
					Settype(li, R3D_00, 46117);
				else
					Settype(li, R3D_16, 46117);
			}
			else
				Settype(li, R3D_00, 46117);

			if (fs->floorheight >= fs->ceilingheight) SetNOocclude(fs, 46117);
			if (bs->floorheight >= bs->ceilingheight) SetNOocclude(bs, 46118);
		}
		SetLFG(sec->lines[i], LFG_FIXED|LFG_NEWSET, 46118);
	}

	if (s && c > 2)
	{
		int k;

		for (i = 0; i < sec->linecount; i++)
		{
			line_t	*ld = sec->lines[i];

			if (hitline[ld - lines] == 1
				&& subfn_CheckForRing(hitline, sec, ld - lines, &ring_count, bbox, NULL, NULL))
			{
				if (   LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE)
					&& LS_SetSSINF(sec, hitline, NULL, NULL, false, false))
				{
					if (ssinf.lines && !ssinf.hitex)
					{
						for (k = 0; k < hitssecptr; k++)
						{
							Setr3dsscs(&subsectors[hitssec[k]], s - sectors, 46119);
						}
					}
				}

				for (k = 0; k < sec->linecount; k++)
				{
					int lnum = sec->lines[k] - lines;

					if (hitline[lnum] == 4) hitline[lnum] = 8;
				}
			}
		}
	}

	return true;
}

// Missing rings ============================================


// Some lines may not have a texture (map error)
// This filters out errors
static RDBOOL hasLOHItexInRing(sector_t *sec, RDBOOL hi)
{
	int i, count = 0;

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;

		if (hitline[lnum] == 4)
		{
			if (origTEX[lnum] & 0x22)
				count++;
			else
			{
				if ( hi && origTEX[lnum] & 0x44) count++;
				if (!hi && origTEX[lnum] & 0x11) count++;
			}
		}
	}
	return count > 1;	// A single line set is a map error, so ignore
}

static int chkTexInRing(sector_t *sec)
{
	int i, c0 = 0, c1 = 0;

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;

		if (hitline[lnum] == 4)
		{
			if (origTEX[lnum] & 0x44) c0++;
			if (origTEX[lnum] & 0x11) c1++;
		}
	}
	return (c0 > 1)|((c1 > 1)<<1);
}

static void update_lineArray(sector_t *sec, byte *array, RDBOOL isset)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		int lnum = sec->lines[i] - lines;

		if (array[lnum] == 4)
		{
			array[lnum] = 8;

			if (isset)
			{
				SetLFG(sec->lines[i], LFG_NEWSET, 46120);
				misstexlines[lnum] = false;
			}
		}
	}
}

static void markLinesForMissing(RDBOOL setlfg, byte settype, int num)
{
	int i, k;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef && sg->linedef->validcount != validcount)
			{
				int lnum = sg->linedef - lines;

				sg->linedef->validcount = validcount;

				if (settype != 0xFF)
				{
					Settype(sg->linedef, settype, 46121 + num);
					SetLFG(sg->linedef, LFG_FIXED, 46122 + num);
				}
				else if (!origTEX[lnum])
				{
					byte type = R3D_01;

					if (sg->linedef->backsector->ceilingheight
						> sg->linedef->frontsector->ceilingheight
						&& sg->linedef->backsector->ceilingpic == skyflatnum
						&& sg->linedef->frontsector->ceilingpic != skyflatnum)
					{
						type = R3D_16;
					}
					else
					if (sg->linedef->frontsector->ceilingheight
						> sg->linedef->backsector->ceilingheight
						&& sg->linedef->frontsector->ceilingpic == skyflatnum
						&& sg->linedef->backsector->ceilingpic != skyflatnum)
					{
						type = R3D_16;
					}
					else
					{
						if (   sg->linedef->backsector->floorheight
							!= sg->linedef->frontsector->floorheight)
						{
							RDBOOL p1 = IsLiquidFloorFlat(sg->linedef->frontsector);
							RDBOOL p2 = IsLiquidFloorFlat(sg->linedef->backsector);

							if (p1^p2) type = R3D_3E;
						}
					}

					Settype(sg->linedef, type, 46123 + num);
					SetLFG(sg->linedef, LFG_FIXED, 46124 + num);
				}

				if (setlfg)
					SetLFG(sg->linedef, LFG_NEWSET, 46125 + num);

				misstexlines[lnum] = false;
			}
		}
	}
}

static int getSecnum(sector_t *sec, RDBOOL hi)
{
	sector_t *s = NULL, *fs, *bs;
	int		 i, diff, txn, cht = DDMININT, fht = DDMAXINT;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*ld = sec->lines[i];
		int		lnum = ld - lines;

		if (hitline[lnum] == 4)
		{
			fs = ld->frontsector, bs = ld->backsector;

			if (bridgelines[lnum]) continue;

			if (hi)
			{
				if ( (diff = fs->ceilingheight - bs->ceilingheight) )
				{
					txn = diff > 0 ? 0x04 : 0x40;

					if (!(origTEX[lnum] & txn))
					{
						sector_t *c = diff > 0 ? bs : fs;

						if (c->ceilingheight > cht) s = c, cht = c->ceilingheight;
					}
				}
			}

			if (!hi)
			{
				if ( (diff = fs->floorheight - bs->floorheight) )
				{
					txn = diff < 0 ? 0x01 : 0x10;

					if (!(origTEX[lnum] & txn))
					{
						sector_t *f = diff < 0 ? bs : fs;

						if (f->floorheight < fht) s = f, fht = f->floorheight;
					}
				}
			}
		}
	}
	return s ? s - sectors : -1;
}

static line_t *findBridge(sector_t *sec, vertex_t *v1, vertex_t *v2)
{
	int i;

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (ld->lnflags2 & LNF2_OWNBOTHGTRTWO)
		{
			if (ld->v1 == v1 && ld->v2 == v2) return ld;
			if (ld->v1 == v2 && ld->v2 == v1) return ld;
		}
	}
	return NULL;
}

static line_t *checkBridgeLine(sector_t *sec, int *vn, RDBOOL hi)
{
	vertex_t	*v1, *v2;
	line_t		*ld;
	int			i, vtxmarkptr = 0;

	v1 = &vertexes[vn[0]], v2 = &vertexes[vn[1]];

	// Find a line, in the same sector, that bridges any of the vtxcheck 1 lines
	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   hitline[ld - lines] != 1
			&& ld->lnflags2 & LNF2_OWNBOTHGTRTWO)
		{
			if (ld->v1 == v1 && ld->v2 == v2) return ld;
			if (ld->v1 == v2 && ld->v2 == v1) return ld;
		}
	}

	// Look for line in another (owned) sector
	{
		vertexowner_t	*vo = &vertexowners[vn[0]];
		int				msk = hi ? SF_NOHITEX : SF_NOLOTEX;

		for (i = 0; i < vo->num; i++)
		{
			sector_t *s = &sectors[vo->list[i]];

			if (s != sec
				&& sec->specFlags & msk
				&& (ld = findBridge(s, v1, v2)) )
			{
				chkseccnt = 2;
				return ld;
			}
		}
	}
	return NULL;
}

static RDBOOL ownerCompare(vertexowner_t *vo1, vertexowner_t *vo2)
{
	int i, k;

	for (i = 0; i < vo1->num; i++)
	{
		int		n = vo1->list[i];
		RDBOOL	match = false;

		for (k = 0; k < vo2->num; k++)
		{
			if (vo2->list[k] == n) { match = true; break; }
		}
		if (!match) return false;
	}

	return true;
}

static RDBOOL validBridgeLine(sector_t *sec, line_t *ld)
{
	vertexowner_t *vo1 = &vertexowners[ld->v1 - vertexes];
	vertexowner_t *vo2 = &vertexowners[ld->v2 - vertexes];
	int		i;

	if (ownerCompare(vo1, vo2)) return true;

	// If the line shares a vertex with another bridge line (in the same
	// sector) and the other end's owners match then it is valid.
	for (i = 0; i < sec->linecount; i++)
	{
		line_t	*li = sec->lines[i];

		if (li->spflags & SPF_CONTROLLINE) continue;

		if (li->lnflags2 & LNF2_OWNBOTHGTRTWO && li != ld)
		{
			sector_t	*fs = li->frontsector;
			sector_t	*bs = li->backsector;

			if (bs
				&& fs != bs
				&& (fs == ld->frontsector && bs == ld->backsector
				||  bs == ld->frontsector && fs == ld->backsector)
				&& fs->floorheight == bs->floorheight
				&& fs->ceilingheight == bs->ceilingheight)
			{
				vertexowner_t *vo3 = &vertexowners[li->v1 - vertexes];
				vertexowner_t *vo4 = &vertexowners[li->v2 - vertexes];
				vertexowner_t *vo5 = NULL, *vo6;
				if      (li->v1 == ld->v1 && li->v2 != ld->v2) { vo5 = vo2, vo6 = vo4; }
				else if (li->v1 == ld->v2 && li->v2 != ld->v1) { vo5 = vo1, vo6 = vo4; }
				else if (li->v2 == ld->v1 && li->v1 != ld->v2) { vo5 = vo2, vo6 = vo3; }
				else if (li->v2 == ld->v2 && li->v1 != ld->v1) { vo5 = vo1, vo6 = vo3; }

				if (vo5) return (ownerCompare(vo5, vo6));
			}
		}
	}

	return false;
}

static RDBOOL analyseRing(sector_t *sec, byte *lnarray, int *lnum, int *lptr, int *bbox, RDBOOL *bridge, RDBOOL hi)
{
	int a, c, i, k, nextline = -1;

	*lptr = 0;
	*bridge = false;
	
	CLR_vtxcheck();

	for (a = i = 0; i < sec->linecount; i++)
	{
		int hline = sec->lines[i] - lines;

		if (lnarray[hline] == 1)
		{
			linelist[a++] = hline;
			vtxcheck[lines[hline].v1 - vertexes]++;
			vtxcheck[lines[hline].v2 - vertexes]++;
		}
		else if (!lnarray[hline])
		{
			vtxcheck[lines[hline].v1 - vertexes]+=2;
			vtxcheck[lines[hline].v2 - vertexes]+=2;
		}
	}

	if (a < 3) return false;

	for (i = 0; i < a; i++)
	{
		int hline = linelist[i];

		if (lnarray[hline] == 1)
		{
			if (   vtxcheck[lines[hline].v1 - vertexes] == 3
				|| vtxcheck[lines[hline].v2 - vertexes] == 3
				|| vtxcheck[lines[hline].v1 - vertexes] == 6
				|| vtxcheck[lines[hline].v2 - vertexes] == 6)
			{
				nextline = hline;
				break;
			}
		}
	}

	if (nextline >= 0)
	{
		line_t	*ld;
		int		vnum[2];

		vnum[0] = vnum[1] = -1;

		if (checkForRingInList(lnarray, a, linelist, nextline, bbox, vnum, DDMININT, false))
		{
			for (c = k = 0; k < a; k++)
			{
				int n = linelist[k];

				if (lnarray[n] == 4) linearray[c++] = n;
			}

			*lptr = c;
			*lnum = nextline;

			return true;
		}
		else if (vnum[0] >= 0 && vnum[1] >= 0 && vnum[0] != vnum[1])
		{
			if (  (ld = checkBridgeLine(sec, vnum, hi))
				&& validBridgeLine(sec, ld))
			{
				M_ClearBox(bbox);
				M_AddLineToBox(bbox, ld);

				for (c = k = 0; k < a; k++)
				{
					int n = linelist[k];

					if (lnarray[n] == 2)
					{
						lnarray[n] = 4;
						linearray[c++] = n;
						M_AddLineToBox(bbox, &lines[n]);
					}
				}

				linearray[c++] = ld - lines;
				lnarray[ld - lines] = 4;
				bridgelines[ld - lines] = true;

				*lptr = c;
				*lnum = nextline;
				*bridge = true;

				if (chkseccnt == 2)
				{
					int x = meanV(bbox[BOXLEFT], bbox[BOXRIGHT]);
					int y = meanV(bbox[BOXTOP ], bbox[BOXBASE ]);
					int	side = P_PointOnLineSide(x, y, ld);

					chksec[1] = side ?
						ld->backsector - sectors : ld->frontsector - sectors;
				}
				return true;
			}
			else
			{
				for (k = 0; k < a; k++)
				{
					int n = linelist[k];

					if (lnarray[n] == 2) lnarray[n] = 1;
				}
			}
		}
		else
			lnarray[nextline] = 8;
	}
	return false;
}

static void trace_Missing(sector_t *sec, RDBOOL hi)
{
	int		i, bridge, lnum, lptr, bbox[4];
	
	chkseccnt = 1, chksec[0] = sec - sectors;

	CLR_excludeline();

	while (analyseRing(sec, hitline, &lnum, &lptr, bbox, &bridge, hi))
	{
		RDBOOL	isset = false, clear = false, recheck = false;

		if (   LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, lptr, bbox, NULL, false, MSIA_NONE)
			&& LS_SetSSINF(sec, hitline, excludessecs, bridgelines, true, false))
		{
			int		origlo, orighi;

			if (!clear) clear = ssinf.anylines > 0;

			if (bridge || hasLOHItexInRing(sec, !hi) || ssinf.sameseclines)
			{
				if (ssinf.ownringcnt)
					excludeSSEC(sec, ssinf_ownrcptr, ssinf.ownringcnt, &recheck);

				if (recheck) LS_SetSSINF(sec, hitline, excludessecs, bridgelines, true, false);

				orighi = ssinf.hitex;
				origlo = ssinf.lotex;

				// We are only interested if other sector lines are inside
				if (ssinf.lines || ssinf.sameseclines || ssinf.othersslinecnt)
				{
					int a, b, c, secnum = getSecnum(sec, hi);

					if ( hi && !ssinf.hitex)
					{
						for (a = b = c = i = 0; i < hitssecptr; i++)
						{
							subsector_t	*ss = &subsectors[hitssec[i]];
							byte		mask = excludessecs[hitssec[i]];

							if (!orighi && !(mask & 2))
								Setr3dsscs(&subsectors[hitssec[i]], secnum, 46126), a++;

							if (ssinf.specfloor >= 0 && ss->sector != sec)
								Setr3dssfs(&subsectors[hitssec[i]], ssinf.specfloor, 46127), c++;
							else if (!origlo && !(mask & 1) && !ssinf.misslo)
								Setr3dssfs(&subsectors[hitssec[i]], sec - sectors, 46128), b++;
						}

						isset = true;
					}
					if (!hi && !ssinf.lotex)
					{
						for (a = b = c = i = 0; i < hitssecptr; i++)
						{
							subsector_t	*ss = &subsectors[hitssec[i]];
							byte		mask = excludessecs[hitssec[i]];

							if (!orighi && !(mask & 2) && !ssinf.misshi)
								Setr3dsscs(&subsectors[hitssec[i]], sec - sectors, 46129), a++;

							if (ssinf.specceil >= 0 && ss->sector != sec)
								Setr3dsscs(&subsectors[hitssec[i]], ssinf.specceil, 46130), c++;
							else if (!origlo && !(mask & 1))
								Setr3dssfs(&subsectors[hitssec[i]], secnum, 46131), b++;
						}

						isset = true;
					}
				}
			}

			if (clear || isset) markLinesForMissing(isset, 0xFF, 200);
		}
		update_lineArray(sec, hitline, isset|clear);
	}
}

static RDBOOL skyLines(line_t *ld)
{
	return ld->backsector
		&& !(ld->backsector->ceilingpic == skyflatnum
		&&  ld->frontsector->ceilingpic == skyflatnum)
		&& !(ld->backsector->floorpic == skyflatnum
		&&  ld->frontsector->floorpic == skyflatnum);
}

static RDBOOL skyCeil(int lnum)
{
	line_t *li = &lines[lnum];

	return li->backsector
		&& li->backsector->ceilingpic == skyflatnum
		&& li->frontsector->ceilingpic == skyflatnum;
}

static int onlySameSecInside(int linecnt)
{
	line_t	*ld;
	int		c, i, k;

	validcount++;

	for (i = 0; i < linecnt; i++)
		lines[ringtmp[i]].validcount = validcount;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (!(ld = sg->linedef) || ld->validcount == validcount) continue;

			if (ld->frontsector != ld->backsector) return -1;

			c++;
		}
	}
	return c;
}

static void checkUnsetMissingLines()
{
	sector_t	*sec, *fs, *bs;
	line_t		*ld; 
	int			c, i, k, m, lnum, bbox[4];
	
	linememcnt = 0;

	memset(misstexlines, 0, numlines);

	CLR_hitline();
	SET_countsec(0);
	memset(hitlinebak, 0, numlines);

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !(lineflags[i] & (LFG_NEWSET|LFG_3DTYPE))
			&& ld->ml_flags & ML_TWOSIDED
			&& Vtfl3(ld)
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& fs->ceilingheight - fs->floorheight > 4*FRACUNIT
			&& bs->ceilingheight - bs->floorheight > 4*FRACUNIT
			&& !IsLiquidFloorFlat(fs)
			&& !IsLiquidFloorFlat(bs)
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& !(bs->specFlags & SF_MOVSECTOR)
			&& !(ld->spflags & SPF_CONTROLLINE))
		{
			RDBOOL mh = fs->ceilingheight != bs->ceilingheight
				&& !(origTEX[i] & 0x44)
				&& !(fs->ceilingpic == skyflatnum && bs->ceilingpic == skyflatnum);
			RDBOOL ml = fs->floorheight != bs->floorheight
				&& !(origTEX[i] & 0x11)
				&& !(fs->floorpic == skyflatnum && bs->floorpic == skyflatnum);

			if (mh && !ml)
			{
				misstexlines[i] = true;
				linememory[linememcnt++] = i;
				hitline[i] = 1;
			}
			else if (!mh && !ml)
			{
				hitlinebak[i] = true;
			}
		}
	}

	// Bridging lines
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (hitlinebak[i])
		{
			// check it has an owner that has no missing tex
			vertexowner_t	*own = &vertexowners[ld->v1 - vertexes];
			RDBOOL			valid = false;

			for (k = 0; k < own->num; k++)
			{
				if (!countsec[own->list[k]])
				{
					valid = true; break;
				}
			}

			if (!valid)
			{
				own = &vertexowners[ld->v2 - vertexes];

				for (k = 0; k < own->num; k++)
				{
					if (!countsec[own->list[k]])
					{
						valid = true; break;
					}
				}
			}

			if (valid)
			{
				linememory[linememcnt++] = i;
				hitline[i] = 1;
			}
		}
	}

	for (i = 0; i < linememcnt; i++)
	{
		lnum = linememory[i];

		if (   hitline[lnum] != 1
			|| origTEX[lnum] & 0x44
			|| lineflags[lnum] & (LFG_NEWSET|LFG_3DTYPE)
			|| eliminatesec[lines[lnum].frontsector - sectors]
			|| (lines[lnum].backsector && eliminatesec[lines[lnum].backsector - sectors])
			|| skyCeil(lnum))
		{
			continue;
		}

		ld = &lines[lnum]; fs = ld->frontsector; bs = ld->backsector;

		sec = fs->ceilingheight < bs->ceilingheight ? fs : bs;

		// v220-7 parm 8 added (sametype)
		if (checkForRingInList(hitline, linememcnt, linememory, lnum, bbox, NULL, sec->ceilingheight, true))
		{
			sector_t	*use = sec;
			sector_t	*chk;
			int			xmid = meanV(bbox[BOXLEFT], bbox[BOXRIGHT]);
			int			ymid = meanV(bbox[BOXTOP ], bbox[BOXBASE ]);
			int			lncount = 0;
			int			ht = sec->ceilingheight;
			int			side;
			RDBOOL		maybeR3D_16 = true;

			chkseccnt = 0;

			for (c = k = 0; k < linememcnt; k++)
			{
				if (hitline[linememory[k]] == 4)
				{
					linearray[lncount++] = linememory[k];

					ld = &lines[linememory[k]];

					chk = ld->frontsector->ceilingheight < ld->backsector->ceilingheight ?
								ld->frontsector : ld->backsector;

					if (   maybeR3D_16
						&& chk != sec
						&& chk->ceilingheight != sec->ceilingheight)
					{
						maybeR3D_16 = false;
					}

					// Find lowest ceiling
					if (chk->ceilingheight < ht)
					{
						use = chk; ht = chk->ceilingheight;
					}

					side = P_PointOnLineSide(xmid, ymid, ld);

					chk = side ? ld->backsector : ld->frontsector;

					if (chk == use)	// v220-4
						chk = !side ? ld->backsector : ld->frontsector;

					setCHKSEC(chk);
				}
			}

			if (   chkseccnt
				&& LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, lncount, bbox, NULL, false, MSIA_NONE))
			{
				int	samesecinside;
				int linecnt = 0;
				int	flags = (chkseccnt == 1) ? sectors[chksec[0]].specFlags & 0xC : 0;	// v220-4

				for (k = 0; k < linememcnt; k++)
				{
					if (hitline[linememory[k]] == 4) ringtmp[linecnt++] = linememory[k];
				}
				
				samesecinside = 0;
				
				if (!(flags & 0xC) && (samesecinside = onlySameSecInside(linecnt)) <= 0)
				{
					memcpy(hitlinebak, hitline, numlines);
					
					eliminateSUBSSECS(sec, linecnt, false, 1, -1, 2);

					memcpy(hitline, hitlinebak, numlines);
				}

				validcount++;

				for (k = 0; k < hitssecptr; k++)
				{
					subsector_t *ss = &subsectors[hitssec[k]];
					seg_t		*sg = &segs[ss->firstline];
					RDBOOL		setR3D_16 = maybeR3D_16;
					RDBOOL		setR3D_00 = false;

					ss->ssflags |= SSF_NOCHANGEFS;

					if (samesecinside > 0)
					{
						setR3D_16 = false;
						Setr3dsscs(ss, sec - sectors, 46137);
					}
					else
					{
						// v220-4 modified
						if (!excludessecs[hitssec[k]])
						{
							if (flags & 0x8)
								Setr3dsscs(ss, sec - sectors, 46137);
							else
								Setr3dsslcsec(ss, sec - sectors, 46137);
						}

						if (ss->sector->ceilingheight <= sec->ceilingheight)
						{
							setR3D_00 = true;
						}
					}

					for (m = 0; m < ss->linecount; m++, sg++)
					{
						ld = sg->linedef;

						if (!ld || ld->validcount == validcount) continue;

						if (ss->sector->ceilingpic != skyflatnum
							&& (origTEX[ld - lines] == 0x02 || origTEX[ld - lines] == 0x20))
						{
							sector_t *chk = ld->frontsector == ss->sector ?
								ld->backsector : ld->frontsector;

							if (chk && chk->ceilingpic == skyflatnum)
								continue;
						}

						ld->validcount = validcount;

						if (setR3D_00 && !(setR3D_16 && !origTEX[ld - lines]))	// $v94
						{
							Settype(ld, R3D_00, 46139);
							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46139);
						}
						else if (setR3D_16)
						{
							Settype(ld, R3D_16, 46140);
							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46140);
						}
						else if (!origTEX[ld - lines])
						{
							Settype(ld, R3D_01, 46141);
							SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46141);
						}
						else
						{
							SetLFG(ld, LFG_NEWSET, 46142);
							Setr3dsscs(ss, use - sectors, 46142);
						}

						misstexlines[ld - lines] = false;
					}
				}
			}

			for (k = 0; k < linememcnt; k++)
			{
				if (hitline[linememory[k]] == 4) hitline[linememory[k]] = 8;
			}
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!(secmark[i] & 2) || IsLiquidFloorFlat(sec)) continue;

		CLR_hitline();

		for (c = k = 0; k < sec->linecount; k++)
		{
			lnum = sec->lines[k] - lines;

			if (misstexlines[lnum] && !skyLines(sec->lines[k]))
			{
				hitline[lnum] = 1; c++;
			}
		}

		if (c < 3) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			line_t	*ld = sec->lines[k];
			
			lnum = ld - lines;

			if (hitline[lnum] == 1
				&& subfn_CheckForRing(hitline, sec, lnum, NULL, bbox, NULL, NULL))
			{
				RDBOOL isset = false;
				RDBOOL hi = !(origTEX[lnum] & 0x44)
					&& ld->backsector->ceilingheight != ld->frontsector->ceilingheight;
				int secnum = getSecnum(sec, hi);

				if (sec->specFlags & SF_MOVSECTOR
					|| MSFC(i, SF_SECCLOSED|SF_NOHITEX|SF_NOLOTEX))
				{
					for (m = 0; m < sec->linecount; m++)
					{
						lnum = sec->lines[m] - lines;

						if (hitline[lnum] == 4)
							Settype(&lines[lnum], R3D_16, 46155);
					}
					isset = true;
				}
				else
				{
					int type = chkTexInRing(sec);

					if (   type
						&& type != 3
						&& LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE))
					{
						for (m = 0; m < hitssecptr; m++)
						{
							if (!(type & 1))
								Setr3dsscs(&subsectors[hitssec[m]], secnum, 46156);
							else
								Setr3dssfs(&subsectors[hitssec[m]], secnum, 46157);
						}
						isset = true;
					}
				}
				update_lineArray(sec, hitline, isset);
			}
		}
	}
}

static RDBOOL multiSecLoop(int *array, int count, int pos, int *bbox)
{
	line_t		*ld = &lines[array[pos]];
	vertex_t	*v1 = ld->v1;
	RDBOOL		isclosed = false, repeat = true;
	int			c = 0, i;

	M_ClearBox(bbox);

	misstexlines[pos] = 2;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < count; i++)
		{
			if (misstexlines[i] == 1)
			{
				line_t	*li = &lines[array[i]];

				if (li->v1 == v1)
				{
					v1 = li->v2; misstexlines[i] = 2; c++; repeat = true;
				}
				else if (li->v2 == v1)
				{
					v1 = li->v1; misstexlines[i] = 2; c++; repeat = true;
				}
			}
		}
	}

	isclosed = (v1 == ld->v2);

	for (i = 0; i < count; i++)
	{
		if (misstexlines[i] == 2)
		{
			misstexlines[i] = isclosed ? 4 : 1;

			M_AddLineToBox(bbox, &lines[array[i]]);
		}
	}

	return isclosed;
}

static RDBOOL noHeight(line_t *ld)
{
	sector_t *s = ld->frontsector;

	if (s->ceilingheight <= s->floorheight) return true;

	s = ld->backsector;

	if (s && s->ceilingheight <= s->floorheight) return true;

	return false;
}

static void setMSLType(line_t *ld, RDBOOL setR3D_16, int num)
{
	int		txn = origTEX[ld - lines];
	byte	type = R3D_00;

	if (ld->ml_flags & ML_TWOSIDED)
	{
		if (!txn)
		{
			int flags = LS_TestLineTex(ld);

			if (   flags & LT_MISHI && flags & LT_MISLO
				&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
				&& !(ld->backsector->specFlags & SF_MOVSECTOR))
			{
			//	Settype(ld, (byte)(!num ? R3D_00 : R3D_3F), 46158);	// v220-4 was R3D_16
				// v220-4 do nothing
			}
			else if (!setR3D_16 || noHeight(ld))
				type = R3D_01;
		}
		else if (txn & 0x22)
		{
			if (txn & 0x22 && !(txn & 0x11)
				&& ld->frontsector->floorheight
				!= ld->backsector->floorheight)
				type = R3D_1D;
			else
				type = R3D_03;
		}

		SetLFG(ld, LFG_FIXED|LFG_NEWSET, num);
	}

	Settype(ld, type, num);
	LS_SetSECDONE(ld, num);
}

static line_t *connectedLineIndex(int *lnarray, int count, int index, vertex_t *v1)
{
	int	i;

	for (i = 0; i < count; i++)
	{
		if (i != index && (lines[lnarray[i]].v1 == v1 || lines[lnarray[i]].v2 == v1))
		{
			return &lines[lnarray[i]];
		}
	}
	return NULL;	// Shouldn't happen
}

// Given the rarity of the error this fixes in released maps it has been added primarily
// as an editing aid but has the benefit that it can fix map errors as well.
static void substituteTEX(int *lnarray, int count, int index, RDBOOL hi)
{
	line_t		*ld = &lines[lnarray[index]], *l1, *l2;
	sector_t	*fs = ld->frontsector, *bs = ld->backsector;

	if (!bs) return;

	// Get the lines both sides (needed to calculate textureoffset)
	l1 = connectedLineIndex(lnarray, count, index, ld->v1);
	l2 = connectedLineIndex(lnarray, count, index, ld->v2);

	if (l1 && l2)
	{
		int		sidd = hi ? fs->floorheight   < bs->floorheight   ? 0 : 1 :
							fs->ceilingheight > bs->ceilingheight ? 0 : 1;

		int		sid1 = (origTEX[l1 - lines] & 0x50) ? 1 : 0;
		int		sid2 = (origTEX[l2 - lines] & 0x50) ? 1 : 0;
		int		i, txn, n1 = 0, n2 = 0;

		side_t	*sdld = &sides[ld->sidenum[sidd]];
		side_t	*sdl1 = &sides[l1->sidenum[sid1]];
		side_t	*sdl2 = &sides[l2->sidenum[sid2]];

		RDBOOL	arbitrate;

		//-------------------------------------------

		if (hi)	// Missing hi ring so we're checking the low texture
		{
			if (!sdl1->bottomtexture || !sdl2->bottomtexture) return;
			arbitrate = sdl1->bottomtexture != sdl2->bottomtexture;
		}
		else
		{
			if (!sdl1->toptexture || !sdl2->toptexture) return;
			arbitrate = sdl1->toptexture != sdl2->toptexture;
		}

		if (arbitrate)
		{
			// To arbitrate the texture use the line with the lowest vertex count
			for (i = 0; i < sdld->sector->linecount; i++)
			{
				line_t *li =  sdld->sector->lines[i];

				if (li == ld || li == l1 || li == l2) continue;

				if (li->v1 == l1->v1 || li->v1 == l1->v2 || li->v2 == l1->v1 || li->v2 == l1->v2) n1++; 
				if (li->v1 == l2->v1 || li->v1 == l2->v2 || li->v2 == l2->v1 || li->v2 == l2->v2) n2++; 
			}
		}

		if (n1 > n2) sdl1 = sdl2, l1 = l2;

		if (hi)
		{
			txn = sdld->bottomtexture = sdl1->bottomtexture;
		}
		else
		{
			txn = sdld->toptexture = sdl1->toptexture;
		}

		// Calculate the offset
		{
			int len = ((int)M_AccurateLength(ld))<<FRACBITS;
			
			if (ld->v1 == l1->v2 || ld->v1 == l1->v1)
				sdld->textureoffset = sdl1->textureoffset + len;
			else
				sdld->textureoffset = sdl1->textureoffset - len;
		}

		sdld->rowoffset = sdl1->rowoffset;
		sdld->texFlags  = sdl1->texFlags;
		ld->spflags     = l1->spflags;
		ld->mtflags     = l1->mtflags;

		origTEX[ld - lines] = origTEX[l1 - lines];
	}
}

static void setMultiInf(RDBOOL hi)
{
	line_t	*ld;
	int		i, k, m;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			RDBOOL	boundsline = false;
			int		flags;

			if (!(ld = sg->linedef) || ld->validcount == validcount) continue;

			ld->validcount = validcount;

			for (m = 0; m < linememcnt; m++)
			{
				if (misstexlines[m] == 4 && linememory[m] == ld - lines)
				{
					boundsline = true; break;
				}
			}

			if (boundsline) continue;

			if (!origTEX[ld - lines] && ld->ml_flags & ML_TWOSIDED && Vtxs2(ld) && Vtfl3(ld))
			{
				flags = LS_TestLineTex(ld);

				// Detecting this type (missing tex line inside a multisec missing
				// tex loop ring) is the only time that LNF2_NOCHANGE is used. It
				// prevents it being set to R3D_3F at a later point.
				if (!hi && flags & LT_MISHI)
				{
					Settype(ld, R3D_16, 46000);
					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46000);
					ld->lnflags2 |= LNF2_NOCHANGE;
				}
				if (hi && flags & LT_MISLO)
				{
					Settype(ld, R3D_16, 46000);
					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46000);
					ld->lnflags2 |= LNF2_NOCHANGE;
				}
			}
		}
	}
}

static RDBOOL mostMissing(sector_t *sec)
{
	int c, i;

	for (c = i = 0; i < sec->linecount; i++)
	{
		line_t   *ld = sec->lines[i];
		sector_t *os = ld->frontsector == sec ? ld->backsector : ld->frontsector;

		if (  !os
			|| os == sec
			|| origTEX[ld - lines]
			|| os->floorpic == sec->floorpic
			|| os->floorheight < sec->floorheight)
		{
			return true;
		}

		if (os->floorheight > sec->floorheight) c++;
	}

	return c >= sec->linecount - 1;
}

static void check_lnc_Lines(int lidx, sector_t *ms, sector_t *fs)
{
	subsector_t *ss;
	sector_t	*ch;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k;
	// ----------------------

	sl = &sslines[lidx];
	hitssecptr = 0;
	validcount++;
	ss_validcount++;

	for (c = i = 0, ch = NULL; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];

		if (ss->ssvalidcount == ss_validcount) continue;

		ss->ssvalidcount = ss_validcount;

		if (ss->sector != ms && ss->sector != fs) return;

		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( !(ld = sg->linedef) ) continue;
			if (ld->validcount == validcount) continue;
			if (lineflags[ld - lines]) return;

			ld->validcount = validcount;
			linearray[c++] = ld - lines;
			if (ld - lines == lidx) continue;

			if (!sg->backsector) return;
			if ( !(sg->backsector->specFlags & SF_SSLINETRUE) ) return;
			if (sg->backsector->specFlags & SF_MOVSECTOR) return;
			if (sg->backsector->floorheight <= ss->sector->floorheight) return;

			if (!ch)
			{
				ch = sg->backsector;
			}
			else if (ch != sg->backsector)
			{
				if (ch->floorheight != sg->backsector->floorheight) return;
			}
		}

		hitssec[hitssecptr++] = ss - subsectors;
	}

	if (!ch) return;

	for (i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		SetOverride(ss, ch - sectors, 46196);
	}

	for (i = 0; i < c; i++)
	{
		ld = &lines[linearray[i]];

		if (!origTEX[ld - lines])
			Settype(ld, R3D_3F, 46196);
		else
			Settype(ld, R3D_36, 46196);

		SetLFG(ld, LFG_FIXABS, 46196);
	}

	Setsecdone(fs, 46196);
	Setsecdone(ms, 46196);
}

// Rare - but used extensively in Eternal map28
static void checkMultisecLoops(RDBOOL hi)
{
	sector_t	*sec;
	line_t		*ld;
	int			c, i, k, bbox[4];
	
	linememcnt = 0;

	CLR_hitline();
	CLR_vtxcheck();

	memset(misstexlines, 0, numlines);

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int txf = hi ? origTEX[i] & 0x44 : origTEX[i] & 0x11;

		if (   ld->ml_flags & ML_TWOSIDED
			&& !(lineflags[i] & (LFG_FIXABS|LFG_3DTYPE))	// v220-6 exclude lines set in LS_CheckLineLoops1
			&& !txf
			&& Vtfl3(ld)
			&& ld->frontsector != ld->backsector
			&& ld->specType < R3D_40
			&& ld->frontsector->stptr < 0
			&& ld->backsector->stptr < 0
			&& vtxarray[ld->v1 - vertexes] > 1
			&& vtxarray[ld->v2 - vertexes] > 1
			&& !(ld->spflags & SPF_CONTROLLINE))
		{
			RDBOOL mark = false;
			int	fdiff = ld->frontsector->floorheight != ld->backsector->floorheight;
			int cdiff = ld->frontsector->ceilingheight != ld->backsector->ceilingheight;

			if (hi)
			{
				mark = cdiff
					&& !(ld->frontsector->ceilingpic == skyflatnum
					&&   ld->backsector->ceilingpic  == skyflatnum);
			}
			else
			{
				mark = fdiff
					&& !(ld->frontsector->floorpic == skyflatnum
					&&   ld->backsector->floorpic  == skyflatnum);
			}

			if (mark)
			{
				vtxcheck[ld->v1 - vertexes]++;
				vtxcheck[ld->v2 - vertexes]++;
				hitline[i] = 1; c++;
			}
			else if (!hi && !fdiff && !cdiff
				&& IsLiquidFloorFlat(ld->frontsector)
				&& IsLiquidFloorFlat(ld->backsector ) )
			{
				mark = true;
				hitline[i] = 1; c++;
				bridgelines[i] = true;
			}
		}
	}

	if (c < 3) return;

	memset(linememory, 0, numlines * sizeof(int));

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   hitline[i] == 1
			&& vtxcheck[ld->v1 - vertexes] == 2
			&& vtxcheck[ld->v2 - vertexes] == 2)
		{
			misstexlines[linememcnt] = 1;
			linememory[linememcnt++] = i;
		}
	}

	if (linememcnt > 3)
	{
		line_t	**sslinemem = (line_t**)W_Calloc(numlines * sizeof(line_t*), STATICID, 0); 	
		int		pic = 0;
		RDBOOL	repeat = true;

		while(repeat)
		{
			repeat = false;

			for (i = 0; i < linememcnt; i++)
			{
				if (misstexlines[i] != 1) continue;

				if (   lines[linememory[i]].lnflags2 & LNF2_OWNGTRTHANTWO
					&& !bridgelines[linememory[i]]
					&& multiSecLoop(linememory, linememcnt, i, bbox))
				{
					sector_t	*s, *chk;
					int			k, m, newcount, seccount;
					int			ht = hi ? DDMAXINT : DDMININT;
					int			ht2 = DDMAXINT;
					int			csec = -1, csec2 = -1, lc = 0;
					int			csecsec[2][2] = { { -1, 0 }, { -1, 0 } };
					
					chkseccnt = 0;

					for (seccount = k = 0; k < linememcnt; k++)
					{
						if (misstexlines[k] == 4)
						{
							RDBOOL found = false;

							ld = &lines[linememory[k]];

							if (hi)
							{
								s = ld->frontsector->ceilingheight > ld->backsector->ceilingheight ?
									ld->frontsector : ld->backsector;

								chk = s == ld->frontsector ? ld->backsector : ld->frontsector;

								lc++;

								for (m = 0; m < 2; m++)
								{
									if (csecsec[m][0] == chk - sectors) { csecsec[m][1]++; break; }
									if (csecsec[m][0] == -1) { csecsec[m][0] = chk - sectors; csecsec[m][1]++; break; }
								}

								if (chk->ceilingheight < ht)
								{
									csec = chk - sectors; ht = chk->ceilingheight;
								}

								// If found this will take precedence
								if (!origTEX[ld - lines] && chk->ceilingheight < ht2)
								{
									csec2 = chk - sectors; ht2 = chk->ceilingheight;
								}
							}
							else
							{
								s = ld->frontsector->floorheight < ld->backsector->floorheight ?
									ld->frontsector : ld->backsector;

								chk = s == ld->frontsector ? ld->backsector : ld->frontsector;

								if (chk->floorheight > ht)
								{
									csec = chk - sectors; ht = chk->floorheight;
								}
							}

							setCHKSEC(s);
						}
					}

					if (chkseccnt > 1)
					{
						int		other;
						int		misscnt = 0, laptr = 0;
						int		missmask = hi ? 0x11 : 0x44;

						//-------------------------------------

						// v220-10 mechanism added using csesec to establish if only two owning
						// sectors are accessed and, if so, to allow the dominant one to prevail.
						// As configured just one line is ignored (all the rest must agree)
						// which is why only two dimensions are needed (as there can be many
						// sectors referenced).

						if (lc > 2)
						{
							if (csecsec[0][1] == lc - 1) csec = csecsec[0][0], csec2 = -1;
							else
							if (csecsec[1][1] == lc - 1) csec = csecsec[1][0], csec2 = -1;
						}

						for (c = k = 0; k < linememcnt; k++)
						{
							if (misstexlines[k] == 4)
							{
								if (!(origTEX[linememory[k]] & missmask))
								{
									misscnt++; laptr = c;
								}
								linearray[c++] = linememory[k];
							}
						}

						if (misscnt == 1)
						{
							substituteTEX(linearray, c, laptr, hi);
						}

						if (csec2 >= 0) csec = csec2;

						if (LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, c, bbox, &other, false, MSIA_NONE))
						{
							RDBOOL		valid = true;
							RDBOOL		check_valid = false;

							// v220-5 pre-check added
							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];
								seg_t		*sg = &segs[ss->firstline];

								if ( hi && ss->sector->specFlags & SF_NOHITEX) continue;
								if (!hi && ss->sector->specFlags & SF_NOLOTEX) continue;

								check_valid = true; break;
							}

							if (check_valid)
							{
								int			c1 = 0, c2 = 0, c3 = 0;

								validcount++;

								// Validate this
								for (k = 0; k < hitssecptr; k++)
								{
									subsector_t *ss = &subsectors[hitssec[k]];
									seg_t		*sg = &segs[ss->firstline];

									if ( hi && ss->sector->specFlags & SF_NOHITEX) continue;
									if (!hi && ss->sector->specFlags & SF_NOLOTEX) continue;

									for (m = 0; m < ss->linecount; m++, sg++)
									{
										if ( (ld = sg->linedef)
											&& ld->validcount != validcount)
										{
											ld->validcount = validcount;

											// Check for area bounded entirely by 1's lines
											// or textureless lines spanning 1's lines
											if (!(ld->ml_flags & ML_TWOSIDED))
											{
												c1++;
											}
											else
											if (   !origTEX[ld - lines]
												&& vtoflags[ld->v1 - vertexes]
												&& vtoflags[ld->v2 - vertexes])
											{
												c2++;
											}

											c3++;

											if (   ld->ml_flags & ML_TWOSIDED
												&& !hitline[ld - lines]
												&& Vtxs2(ld))
											{
												int		fn = ld->frontsector - sectors;
												int		bn = ld->backsector - sectors;
												int		n;
												RDBOOL	found = false;

												for (n = 0; n < chkseccnt; n++)
												{
													if (chksec[n] == fn || chksec[n] == bn)
													{
														found = true; break;
													}
												}

												if (!found)
												{
													valid = false;
												}
											}
										}
									}
								}

								if (c1 + c2 == c3) valid = false;
							}

							if (valid)
							{
								RDBOOL	nohilotex = csec >= 0 ? (sectors[csec].specFlags & 0x40C) == 0xC : false;
								RDBOOL	wasset = false;
								int sslinecount = 0;

								setMultiInf(hi);

								validcount++;

								for (k = 0; k < hitssecptr; k++)
								{
									subsector_t *ss = &subsectors[hitssec[k]];
									seg_t		*sg = &segs[ss->firstline];

									if (hi)
									{
										// b211-0 SF_NOHITEX test added
										if (ss->sector->specFlags & SF_NOHITEX)
										{
											// v220-4 nohilotex added
											if (!nohilotex)
											{
												if (!(sectors[csec].specFlags & SF_MOVCEIL)) Setr3dsscs(ss, csec, 46166);	// v220-35
											}
											// V220-16 test added
											else if (ss->sector->ceilingheight > sectors[csec].ceilingheight
											//	&& ss->sector->ceilingpic == skyflatnum
												&& sectors[csec].ceilingpic != skyflatnum)
											{
												Setr3dsscs(ss, csec, 46199);
											}

											wasset = true;

											// Added vexpRC4
											// ss->floorsec test added v220-17a
											if ( !(ss->floorsec >= 0 && sectors[csec].floorheight >= sectors[ss->floorsec].floorheight)
												&& ss->sector->floorheight >= sectors[csec].floorheight
												&& !SecCtrl(csec))
											{
												Setr3dsslfsec(ss, csec, 46166);
											}
										}
									}
									else
									{
										// v220-6 add movfloor test
										if (ss->sector->specFlags & SF_MOVFLOOR)
											Setr3dsslfsec(ss, csec, 46167);
										else
										{
											// v220-8 add mostMissing test
											if (mostMissing(ss->sector))
												Setr3dssfs(ss, csec, 46167);
											else if (ss->sector->R3Dfloorsec < 0)
												Setr3dfs(ss->sector, csec, 46167);
										}

										wasset = true;
									}

									inhibitSSEC(ss, 46167);

									for (m = 0; m < ss->linecount; m++, sg++)
									{
										if ( (ld = sg->linedef) && ld->validcount != validcount)
										{
											ld->validcount = validcount;

											sslinemem[sslinecount++] = ld;
										}
									}
								}

								// v220-4 fix bug preventing line type and flags being set
								//        Also ensure PS_CheckUnsetSupportLines (ref 43005) is filtered out
								if (wasset)
								{
									for (k = 0; k < sslinecount; k++)
									{
										RDBOOL isset = false;

										for (m = 0; m < linememcnt; m++)
										{
											if (misstexlines[m] == 4 && linememory[m] == sslinemem[k] - lines)
											{
												setMSLType(sslinemem[k], true, 100001); isset = true;
											}
										}

										if (sslinemem[k]->lnflags2 & LNF2_OWNEQUALTOONE)
										{
											if (!origTEX[sslinemem[k] - lines])
												Settype(sslinemem[k], R3D_3F, 46167);
											SetLFG(sslinemem[k], LFG_LINESSECSET, 46167);
										}
										else if (!isset && Vtxs2(sslinemem[k]) && sslinemem[k]->specType != R3D_16)
											setMSLType(sslinemem[k], false, 100002); isset = true;
									}
								}

								if (sectors[csec].specFlags & SF_MOVSECTOR)
								{
									for (k = 0; k < numsubsectors; k++)
									{
										if (subsectors[k].sector - sectors == csec)
											inhibitSSEC(&subsectors[k], 46168);
									}
								}
							}

							for (c = k = 0; k < linememcnt; k++)
							{
								if (misstexlines[k] == 4)
									misstexlines[k] = 8;
							}
						}
					}

					for (newcount = k = 0; k < linememcnt; k++)
					{
						if (misstexlines[k] == 1)
						{
							misstexlines[newcount] = 1;
							linememory[newcount++] = linememory[k];
						}
					}

					if (newcount > 3)
					{
						linememcnt = newcount;
						repeat = true;
					}
					break;
				}
			}
		}

		W_Free(sslinemem, STATICID, 0);
	}

	if (!hi) return;

	/*	We may well have complete single sector loops left which remain
		unprocessed either because they are single sector or because
		they have lines sharing a vertex with another missing sector's
		loop.These can be inspected but, at this stage 3 rules must apply;

		1)	A vertex must be shared
		2)	A vertex must be shared with a different sector.
		3)	No lines can be inside a valid loop.

		The only reason we can get away with this is because nested sectors,
		which are sorted next, don't share vertexes.  In fact this is the
		best point to get these multi vertex owned lines out of the way.
	*/

	for (i = 0; i < numlines; i++)
	{
		if (   lineflags[i] & (LFG_NEWSET|LFG_3DTYPE)
			|| !(lines[i].ml_flags & ML_TWOSIDED)
			|| eliminatesec[lines[i].frontsector - sectors]
			|| eliminatesec[lines[i].backsector - sectors])
		{
			hitline[i] = 0;
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (!(sec->specFlags & SF_NOHITEX)) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			int		lnum, m;

			ld = sec->lines[k];

			lnum = ld - lines;

			if (hitline[lnum] == 1)
			{
				sector_t *s = ld->frontsector->ceilingheight > ld->backsector->ceilingheight ?
					ld->frontsector : ld->backsector;

				if (s != sec) break;

				s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

				if (subfn_CheckForRing(hitline, sec, lnum, NULL, bbox, NULL, NULL))
				{
					RDBOOL valid = false, hasmidtex = false;

					for (m = 0; m < sec->linecount; m++)
					{
						ld = sec->lines[m];
						lnum = ld - lines;

						if (hitline[lnum] == 4)
						{
							vertex_t *v1 = NULL;

							if (vtxcheck[ld->v1 - vertexes] == 4)
								v1 = ld->v1;
							else
							if (vtxcheck[ld->v2 - vertexes] == 4)
								v1 = ld->v2;

							if ((valid = v1 && findLineVertexConnected(ld, v1, sec)) )
								break;
						}
					}

					if (valid
						&& (valid = LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, bbox, -1, MSIA_NONE)) )
					{
						LS_SetSSINF(sec, hitline, NULL, NULL, false, false);

						if ( (valid = !ssinf.anylines) )
						{
							for (m = 0; m < hitssecptr; m++)
								Setr3dsscs(&subsectors[hitssec[m]], s - sectors, 46168);
						}
					}

					for (m = 0; m < sec->linecount; m++)
					{
						ld = sec->lines[m];
						lnum = ld - lines;

						if (hitline[lnum] == 4)
						{
							hitline[lnum] = 8;

							if (valid) setMSLType(ld, true, 100003);
						}
					}
				}
				else
					hitline[lnum] = 8;
			}
		}
	}
}

// We are loking for a particular type
static void setPairedMovsecs(sector_t *sec)
{
	sector_t	*cs, *osec = NULL;
	line_t		*ld;
	int			flag, bbox[4], chksec;
	int			secnum = sec - sectors;
	int			i, k;
	int			linecnt, lnum, rc;

	if (!SecCtrl(secnum)) return;

	CLR_hitline();

	for (linecnt = i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   !(ld->ml_flags & ML_TWOSIDED)
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO)
			|| !Vtxs2(ld)
			|| ((ld->frontsector->specFlags & 0xFFFBFC) != (ld->backsector->specFlags & 0xFFFBFC)) )
		{
			return;
		}

		if (!osec)
		{
			osec = ld->frontsector == sec ? ld->backsector : ld->frontsector;

			// FRACUNIT diff expected else would create (an even worse) HOM
			if (abs(osec->floorheight - sec->floorheight) > FRACUNIT)
				return;
		}

		hitline[ld - lines] = 1; linememory[linecnt++] = ld - lines;
	}

	if (linecnt < 3) return;

	for (i = 0; i < sec->linecount; i++)
	{
		lnum = sec->lines[i] - lines;

		if (hitline[lnum] == 1) break;
	}

	// Note: only one chance
	if (   hitline[lnum] != 1
		|| !subfn_CheckForRing(hitline, sec, lnum, &rc, bbox, NULL, NULL)
		|| rc != linecnt)
	{
		return;	// criteria not met
	}

	flag = MSIA_NONE, chksec = -1;

	if (!LS_MarkSubsecsInLoop(sec, linecnt, linememory, bbox, MSIA_TESTAREA, NULL))
		flag = MSIA_REVERSED;

	if (!LS_MarkSubsecsInLoop(sec, linecnt, linememory, bbox, flag, NULL)) return;

	if (!LS_SetSSINF(sec, hitline, NULL, NULL, false, false)) return;

	if (ssinf.anylines) return;	// No internal lines allowed

	// Get the containing sector
	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (   hitline[ld - lines] == 4
			&& (chksec = LS_FindAdjacentSecnum(ld, 10)) >= 0)
			break;
	}

	if (chksec < 0 || !sampleline) return;

	cs = sampleline->frontsector - sectors == chksec ?
			sampleline->backsector : sampleline->frontsector;

	ss_validcount++;

	for (i = 0; i < sec->linecount; i++)
	{
		lnum = sec->lines[i] - lines;

		if (hitline[lnum] == 4)
		{
			ssline_t *sl = &sslines[lnum];

			Settype(sec->lines[i], R3D_01, 46169);
			SetLFG(sec->lines[i], LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE, 46170);

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (ss->ssvalidcount == ss_validcount) continue;

				ss->ssvalidcount = ss_validcount;
				Setr3dssfs(ss, cs - sectors, 46171);
				Setr3dsscs(ss, cs - sectors, 46172);
				Setr3dsslfsec(ss, chksec, 46173);
			}
		}
	}

	Setsecdone3(sec , 46174);
	Setsecdone3(osec, 46175);
	Setsecdone3(cs  , 46176);
}

static void setVtflag()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& !(ld->lnflags2 & LNF2_SLOPESECLN)
			&& ld->specType < R3D_40
			&& ld->frontsector != ld->backsector)
		{
			sector_t *fs = ld->frontsector;
			sector_t *bs = ld->backsector;
			int		 diff;
			RDBOOL	 found = false;

			if (   (diff = fs->ceilingheight - bs->ceilingheight)
				&& !(fs->ceilingpic == skyflatnum
				&&   bs->ceilingpic == skyflatnum) )
			{
				if (diff > 0 && !sides[ld->sidenum[0]].toptexture)
					found = true;
				else
				if (diff < 0 && !sides[ld->sidenum[1]].toptexture)
					found = true;
			}
			if (   !found
				&& (diff = fs->floorheight - bs->floorheight)
				&& !(fs->floorpic == skyflatnum
				&&   bs->floorpic == skyflatnum) )
			{
				if (diff > 0 && !sides[ld->sidenum[1]].bottomtexture)
					found = true;
				else
				if (diff < 0 && !sides[ld->sidenum[0]].bottomtexture)
					found = true;
			}
			if (found)
			{
				vtfcount[ld->v1 - vertexes]++;
				vtfcount[ld->v2 - vertexes]++;
			}
		}
	}
}


void TM_TraceMissing()
{
	vertexowner_t	*v1, *v2;
	sector_t		*fs, *bs;
	sector_t		*sec, *osec;
	line_t			*ld;
	int				c, i, k;
	RDBOOL			nochkbridge, skyhi, skylo;
	RDBOOL			misshi, misslo, hisec, liqsec, abort;

	//------------------------------------------

	ringtmp  = (int*)W_Malloc(maxsecsz*sizeof(int) * 8, TEMPID, 27);
	ringhi   = (int*)W_Malloc(maxsecsz*sizeof(int)    , TEMPID, 28);
	ringlo   = (int*)W_Malloc(maxsecsz*sizeof(int)    , TEMPID, 29);
	ringss   = (int*)W_Malloc(maxsecsz*sizeof(int)    , TEMPID, 30);
	linelist = (int*)W_Calloc(maxsecsz*sizeof(int)    , TEMPID, 0);

	secmark      = (byte*)W_Calloc(numsectors, TEMPID, 0);
	eliminatesec = (byte*)W_Calloc(numsectors, TEMPID, 0);
	linememory   = (int *)W_Malloc(numlines * sizeof(int), TEMPID, 31);
	lnarray2     = (int *)W_Malloc(numlines * sizeof(int), TEMPID, 32);
	tempsseclist = (int *)W_Malloc(numsubsectors * sizeof(int), TEMPID, 33);
	excludessecs = (byte*)W_Calloc(numsubsectors, TEMPID, 0);
	hitlinebak   = (byte*)W_Malloc(numlines, TEMPID, 34);
	misstexlines = (byte*)W_Malloc(numlines, TEMPID, 35);
	bridgelines  = (byte*)W_Calloc(numlines, TEMPID, 0);
	linemark     = (byte*)W_Malloc(numlines, TEMPID, 36);
	vtfcount     = (byte*)W_Calloc(numvertexes, TEMPID, 0);

	setVtflag();
	
	for (i = 0; i < numlines; i++)
	{
		// v220-18 Look for movsec line sandwiched in two texless sectors
		if (   (bs = lines[i].backsector)
			&& (fs = lines[i].frontsector) != bs
			&& origTEX[i] & 0x22
			&& !(origTEX[i] & 0x55))
		{
			if (   fs->heightsec < 0
				&& bs->heightsec < 0
				&& Spans1sLines(&lines[i]))
			{
				if (  (fs->specFlags & 0xFFFFC) == 0x1001C
					&& SecCtrl(fs - sectors)
					&& (bs->specFlags & 0xFFFFC) == 0x1C
					&& !SecCtrl(bs - sectors))
				{
					check_lnc_Lines(i, fs, bs);
					continue;
				}
				else
				if (  (bs->specFlags & 0xFFFFC) == 0x1001C
					&& SecCtrl(bs - sectors)
					&& (fs->specFlags & 0xFFFFC) == 0x1C
					&& !SecCtrl(fs - sectors))
				{
					check_lnc_Lines(i, bs, fs);
					continue;
				}
			}

			// Eliminate any sectors belonging to a midtexline with no height
			if (   (v1 = &vertexowners[lines[i].v1 - vertexes])->num > 3 
				&& (v2 = &vertexowners[lines[i].v2 - vertexes])->num == v1->num
				&& fs->ceilingpic != skyflatnum
				&& bs->ceilingpic != skyflatnum
				&& !(fs->specFlags & SF_MOVSECTOR)
				&& !(bs->specFlags & SF_MOVSECTOR)
				&& ((fs->ceilingheight <= fs->floorheight)
				||  (bs->ceilingheight <= bs->floorheight)) )
			{
				ma_inf.eliminate_sectors = true;
				for (k = 0; k < v1->num; k++) eliminatesec[v1->list[k]] = true;
			}
		}
	}

	// 1) check for connected missing hi lines in different sectors
	checkMultisecLoops(true);

	// 2) check for connected missing lo lines in different sectors
	checkMultisecLoops(false);

	memset(misstexlines, 0, numlines);

	// 3) Check for contained loops
	checkContainingLoopsPASS1();

	memset(misstexlines, 0, numlines);
	memset(excludeline, 0, numlines);

	// 4) Check for nested loops
	enactRINF();

	memset(misstexlines, 0, numlines);

	CLR_excludeline();

	// 5) Deal with dual ring sectors
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   secdone[i]
			|| sec->linecount < 3
			|| sec->specFlags & SF_MOVSECTOR
			|| sec->stptr >= 0)
		{
			secmark[i] = 1;
			continue;
		}
		
		if ((sec->specFlags & 0xF001C) == 0x1C)
		{
			if (DualRingsectors0x1C(sec))
				secmark[i] = 1;
		}
		else
		if ( (osec = hasDualRings(sec, sec->specFlags & (SF_NOHITEX|SF_NOLOTEX)))
			&& osec != sec )
		{
			if (osec != sec)
			{
				secmark[osec - sectors] |= 4;
			}
			else
			{
				for (k = 0; k < sec->linecount; k++)
				{
					ld = sec->lines[k];

					if (ld->spflags & SPF_CONTROLLINE) continue;

					if (!origTEX[ld - lines])
						Settype(ld, R3D_01, 46177);
					else
						Settype(ld, R3D_00, 46178);

					SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46179);
				}
				Setsecdone(sec, 46180);
				secmark[i] = 1;
			}
		}
	}

	for (i = 0; i < numsectors; i++)
	{
		if (secmark[i] & 4)
		{
			sec = &sectors[i];

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (!origTEX[ld - lines])
					Settype(ld, R3D_01, 46181);
				else
					Settype(ld, R3D_00, 46182);

				SetLFG(ld, LFG_NEWSET|LFG_FIXED, 46183);
			}

			Setsecdone(sec, 46184);
			secmark[i] = 1;
		}
	}

	// 6) Deal with textureless sectors
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secdone[i]) continue;

		if ((sec->specFlags & 0x4013001C) == 0x4011001C)
		{
			setPairedMovsecs(sec);
		}
		else if (MSFC(i, SF_SECCLOSED|SF_NOHITEX|SF_NOLOTEX)
			&& sec->ceilingheight - sec->floorheight > FRACUNIT
			&& !IsLiquidFloorFlat(sec)
			&& sec->heightsec < 0
			&& sectors0x1C(sec))
		{
			Setsecdone(sec, 46185);
			secmark[i] = 1;
		}
	}

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (secmark[i]) continue;

		// b220-1 test added
		if (   sec->floorheight <= sec->ceilingheight
			&& sec->specFlags & 0xFFFFC)	// v220-3 test added i.e. must be unclosed,
		{									// static and not have any have hi&lo textures to pass
			continue;
		}

		CLR_hitline();

		// First get missing lines
		for (k = 0, abort = false; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			if (   ld->ml_flags & ML_TWOSIDED
				&& origTEX[ld - lines] & 0x22
				&& !(origTEX[ld - lines] & 0x55)
				&& ld->backsector->floorheight != ld->frontsector->floorheight
				&& ld->backsector->ceilingheight != ld->frontsector->ceilingheight)
			{
				abort = true; break;
			}
		}

		if (abort) continue;
		
		nochkbridge = true;
		skyhi  = sec->ceilingpic == skyflatnum;
		skylo  = sec->floorpic   == skyflatnum;
		hisec  = sec->ceilingheight - sec->floorheight > 1024*FRACUNIT;
		liqsec = IsLiquidFloorFlat(sec);

		ringhicnt = ringlocnt = 0;
		misshi = misslo = false;

		// First get missing lines
		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (   ld->ml_flags & ML_TWOSIDED
				&& !(lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE))
				&& ld->specType < R3D_40
				&& ld->frontsector != ld->backsector)
			{
				int			side = sec == ld->backsector;
				side_t		*sdf = &sides[ld->sidenum[side]];
				sector_t	*s = ld->frontsector == sec ?
								ld->backsector : ld->frontsector;

				if (   secdone[s - sectors]
					|| s->linecount < 3
					|| s->specFlags & SF_MOVSECTOR
					|| s->stptr >= 0
					|| (liqsec && IsLiquidFloorFlat(s))
					|| s->ceilingheight - s->floorheight < 4*FRACUNIT)
				{
					continue;
				}

				if (   !sdf->toptexture
					&& sec->ceilingheight > s->ceilingheight
					&& !(s->specFlags & SF_MOVCEIL)
					&& !(skyhi && s->ceilingpic == skyflatnum) )
				{
					ringhi[ringhicnt++] = ld - lines;
					hitline[ld - lines] |= 1;
					misstexlines[ld - lines] = true;
					if (ld->lnflags2 & LNF2_OWNGTRTHANTWO)
					{
						misshi = true;
						nochkbridge = false;
					}
				}

				if (   !sdf->bottomtexture
					&& sec->floorheight < s->floorheight
					&& !(s->specFlags & SF_MOVFLOOR)
					&& !(skylo && s->floorpic == skyflatnum) )
				{
					ringlo[ringlocnt++] = ld - lines;
					hitline[ld - lines] |= 2;
					misstexlines[ld - lines] = true;
					if (ld->lnflags2 & LNF2_OWNGTRTHANTWO)
					{
						misslo = true;
						nochkbridge = false;
					}
				}
			}
		}

		if (ringhicnt < 3 && ringlocnt < 3)
		{
			continue;
		}

		secmark[i] |= 2;

		if (hisec && !nochkbridge && (misshi^misslo))
		{
			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->backsector
					&& ld->backsector != ld->frontsector)
				{
					if (misshi
						&& !(hitline[ld - lines] & 1)
						&& !(origTEX[ld - lines] & 0x44))
					{
						ringhi[ringhicnt++] = ld - lines;
						nochkbridge = true;
					}
					if (misslo
						&& !(hitline[ld - lines] & 2)
						&& !(origTEX[ld - lines] & 0x11))
					{
						ringlo[ringlocnt++] = ld - lines;
						nochkbridge = true;
					}
				}
			}
		}

		if (ringhicnt > 2)
		{
			CLR_hitline();
			for (k = 0; k < ringhicnt; k++) hitline[ringhi[k]] = 1;
			trace_Missing(sec, 1);
		}

		if (ringlocnt > 2)
		{
			CLR_hitline();
			for (k = 0; k < ringlocnt; k++) hitline[ringlo[k]] = 1;
			trace_Missing(sec, 0);
		}
	}

	// 7) Deal with loops with a 1's line separately
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		RDBOOL	valid = false;

		if (secmark[i] & 1) continue;

		for (c = k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			if (   !origTEX[ld - lines]
				|| lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)
				|| !vtxflags[ld->v1 - vertexes]
				|| !vtxflags[ld->v2 - vertexes])
			{
				valid = false; break;
			}

			if (ld->ml_flags & ML_TWOSIDED)
			{
				if (origTEX[ld - lines] & 0x22)
				{
					valid = false; break;
				}
			}
			else
				valid = true;
		}

		if (!valid) continue;

		for (c = k = 0; k < sec->linecount; k++)
		{
			if (sec->lines[k]->ml_flags & ML_TWOSIDED)
			{
				if (   vtoflags[ld->v1 - vertexes]
					&& vtoflags[ld->v2 - vertexes])
					Settype(sec->lines[k], R3D_00, 46169);
				else
					Settype(sec->lines[k], R3D_16, 46185);

				SetLFG(sec->lines[k], LFG_NEWSET|LFG_FIXED, 46186);
			}
		}
		secmark[i] = 1;
	}

	// 8) Set dual ring mov floor secs
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& !IsLiquidFloorFlat(sec)
			&& (sec->specFlags & 0x13001C) == 0x0011001C
			&&  sec->stptr < 0
			&& (c = setLineRings(sec, DRLISTLIMIT, true, false)) > 1)
		{
			linestore_t	*lstptr[DRLISTLIMIT];
			byte		boxmrk[DRLISTLIMIT];
			int			a, m;

			memset(boxmrk, 0, DRLISTLIMIT);

			for (k = 0; k < DRLISTLIMIT; k++) lstptr[k] = &lst[k];

			qsort(lstptr, c, sizeof(linestore_t*), lstboxSize);

			for (k = 0; k < c; k++)
			{
				if (!boxmrk[k])
				{
					int	bi = -1;

					for (a = m = 0; m < c; m++)
					{
						if (m != k && M_CompareBox(lst[m].box, lst[k].box))
						{
							boxmrk[m] = true;

							bi = m, a++;
						}
					}

					if (a) boxmrk[k] = true;

					if (a == 1)
					{
						Setsecdone(sec, 46187);

						// get the lines
						for (m = 0; m < lst[k].count; m++)
						{
							int lnum = lst[k].list[m];

							if (origTEX[m] == 0x22)
								Settype(&lines[lnum], R3D_1D, 46188);
							else if (origTEX[lnum] & 0x22)
							{
								if (   lines[lnum].frontsector->specFlags & SF_MOVINVALID
									|| lines[lnum].backsector->specFlags & SF_MOVINVALID)
									Forcetype(&lines[lnum], R3D_16, 46188);
								else
									Forcetype(&lines[lnum], R3D_03, 46188);
							}
							else
								Settype(&lines[lnum], R3D_01, 46188);

							SetLFG(&lines[lnum], LFG_NEWSET, 46189); 
						}

						for (m = 0; m < lst[bi].count; m++)
						{
							int lnum = lst[bi].list[m];	// v220-5 bug fix (was lst[k])

							if (origTEX[lnum] == 0x22)
								Settype(&lines[lnum], R3D_1D, 46190);
							else if (origTEX[lnum] & 0x22)
							{
								if (   lines[lnum].frontsector->specFlags & SF_MOVINVALID
									|| lines[lnum].backsector->specFlags & SF_MOVINVALID)
									Forcetype(&lines[lnum], R3D_16, 46190);
								else
									Forcetype(&lines[lnum], R3D_03, 46190);
							}
							else
								Settype(&lines[lnum], R3D_01, 46190);

							SetLFG(&lines[lnum], LFG_NEWSET, 46191); 
						}
					}
				}
			}
		}
	}

	// 8) Set the rest of the lines
	checkUnsetMissingLines();

	// 9) Check unset missing in movfloor sectors (rare) $v66
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (   !secdone[i]
			&& (sec->specFlags & 0x3000C) == 0x10004
			&&  sec->stptr < 0)
		{
			int		m, fht = DDMAXINT, box[4];
			RDBOOL	hasutex = false, movfloor = false;

			chkseccnt = 0;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& !lineflags[ld - lines]
					&& !(origTEX[ld - lines] & 0x33)
					&& ld->mtflags & R3D_MISLOTEX)
				{
					int		 sid = ld->frontsector == sec ? 1 : 0;
					sector_t *sc = sid ? ld->backsector : ld->frontsector;

					if (sides[ld->sidenum[!sid]].texFlags) hasutex = true;

					if (sc->specFlags & SF_MOVFLOOR) movfloor = true;

					if (fht == DDMAXINT)
						fht = sc->floorheight;
					else if (sc->floorheight != fht)
					{
						c = 0; break;
					}

					setCHKSEC(sc);
					linearray[c++] = ld - lines;
				}
			}

			if (c < 3 || (movfloor && chkseccnt > 1)) continue;

			CLR_hitline();

			if (c == nf_chkLineListRing(c, box, 0, hitline))
			{
				int flag = MSIA_NONE;

				if (!LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_TESTAREA))
					flag = MSIA_REVERSED;

				if (   LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, flag)
					&& LS_OnlySSLinesInSSECarray(NULL, hitline))
				{
					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						if (hasutex)
							Setr3dssfs(ss, chksec[0], 46192);
						else
							SetOverride(ss, i, 46193);

						for (m = 0; m < ss->linecount; m++, sg++)
						{
							if ( (ld = sg->linedef) )
							{
								if (!origTEX[ld - lines])
									Settype(ld, R3D_3F, 46194);
								else
								if (ld->frontsector == ld->backsector)
									Settype(ld, R3D_03, 46194);

								if (hitline[ld - lines] == 4
									&& origTEX[ld - lines])
								{
									SetMissinglo(ld, 46194);
								}

								SetLFG(ld, LFG_NEWSET, 46194);
							}
						}
					}
					Setsecdone(sec, 46194);
				}
			}
		}
	}

	W_FreeALL(TEMPID);	// These types can now be freed
}

void TM_FindFlatRingSectors()
{
	int		b1, b2;
	int		i, num;
	int		*ptr, cntr;

	for (i = 0; i < numsectors; i++)
	{
		if (  (sectors[i].specFlags & 0x3041C) == 0x1C
			&& sectors[i].floorpic != skyflatnum
			&& sectors[i].linecount > 5)
		{
			if ( (num = setLineRings(sectors + i, DRLISTLIMIT, false, true)) > 1)
			{
				linestore_t	*lstptr[DRLISTLIMIT];
				byte		boxmrk[DRLISTLIMIT];
				sector_t	*os, *is;
				int			mask = 0xC;
				int			k, m, n;

				memset(boxmrk, 0, DRLISTLIMIT);

				for (k = 0; k < DRLISTLIMIT; k++) lstptr[k] = &lst[k];

				qsort(lstptr, num, sizeof(linestore_t*), lstboxSize);

				while(1)
				{
					RDBOOL	invalid = false;

					// Start with largest box
					for (k = 0, b1 = -1; k < num - 1; k++)
					{
						if (!boxmrk[k]) { b1 = k; boxmrk[b1] = true; break; }
					}

					if (b1 < 0) break;

					// Find first box that fits inside largest selected
					for (k = b1 + 1, b2 = -1; k < num; k++)
					{
						// Check that one line ring fits inside the other
						if (!boxmrk[k] && M_CompareBox(lstptr[k]->box, lstptr[b1]->box))
						{
							// If more than one loop exists inside then mark as invalid
							if (b2 >= 0) invalid = true;

							b2 = k; boxmrk[b2] = true;
						}
					}

					if (b2 < 0 || invalid) goto end_list;

					// set pointers to outer loop
					ptr  = lstptr[b1]->list; cntr = lstptr[b1]->count;

					// get the outer sector
					os = lines[ptr[0]].frontsector - sectors == i ? lines[ptr[0]].backsector : lines[ptr[0]].frontsector;

					// set pointers to inner loop
					ptr  = lstptr[b2]->list; cntr = lstptr[b2]->count;

					// get the inner sector
					is = lines[ptr[0]].frontsector - sectors == i ? lines[ptr[0]].backsector : lines[ptr[0]].frontsector;

					if (sectors[i].ceilingheight - sectors[i].floorheight >= 8*FRACUNIT)
					{
						if (is->ceilingheight - is->floorheight >= 8*FRACUNIT)
							goto end_list;
					}

					// All sectors must be static
					if (is->specFlags & SF_MOVSECTOR || os->specFlags & SF_MOVSECTOR)
					{
						goto end_list;
					}
					// We're primarily looking for low missing textures..
					if (   os->floorheight == sectors[i].floorheight
						&& is->floorheight == sectors[i].floorheight)
					{
						goto end_list;
					}
					// with raised floors inside the reference sector...
					if (   is->floorheight < os->floorheight
						|| sectors[i].floorheight < os->floorheight)
					{
						goto end_list;
					}
					// but are not worried whether there are any hi missing or not
					if (   os->ceilingheight == sectors[i].ceilingheight
						|| (os->ceilingpic == skyflatnum
						&& sectors[i].ceilingpic == skyflatnum))
					{
						mask = 0x30004;
					}
					// Check the outer sector is valid for inheritance
					if (os->specFlags & mask && !(os->specFlags & 0x400))
					{
						goto end_list;
					}

					hitssecptr = 0;
					ss_validcount++;

					for (k = 0; k < cntr; k++)
					{
						ssline_t *sl = &sslines[ptr[k]];

						for (m = 0; m < sl->num; m++)
						{
							subsector_t *ss = &subsectors[sl->list[m]];

							if (ss->ssvalidcount != ss_validcount)
							{
								seg_t	*sg = &segs[ss->firstline];

								ss->ssvalidcount = ss_validcount;

								for (n = 0; n < ss->linecount; n++, sg++)
								{
									if (sg->linedef && hitline[sg->linedef - lines] != 8)
									{
										goto end_list;
									}
								}

								hitssec[hitssecptr++] = sl->list[m];
							}
						}
					}

					for (k = 0; k < hitssecptr; k++)
						SetOverride(&subsectors[hitssec[k]], os - sectors, 46195);

					for (k = 0; k < cntr; k++)
					{
						Settype(&lines[ptr[k]], R3D_3F, 46195);
						SetLFG(&lines[ptr[k]], LFG_NEWSET|LFG_FIXED, 46195);
					}

					// set pointers back to outer loop
					ptr  = lstptr[b1]->list; cntr = lstptr[b1]->count;

					for (k = 0; k < cntr; k++)
					{
						Settype(&lines[ptr[k]], R3D_3F, 46195);
						SetLFG(&lines[ptr[k]], LFG_NEWSET|LFG_FIXED, 46195);
					}
end_list:;
				}
			}
		}
	}
}
